/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: runtime_util.c,v 1.2 2003/02/17 15:56:12 frassi Exp $
 */


#include "defs.h"
#include "runtime_util.h"
#include "util.h" /* upeek */
#include "pars_funct.h" /* get_sysc_name, is_both */
#include <assert.h>



#ifndef Y_TAB_H
#define Y_TAB_H

/* file header per wids.lex */
#include "y.tab.h"

#endif /* !Y_TAB_H */

/* costante ritornata all'automa per indicare che la condizione  fallita */
#define NOEV 0

/* se il valore di func_not_eval == 1 allora eval_funct  fallita */
int func_not_eval = 0;


/*****************************************************************
 ****************** Funzioni per le condizioni *******************
 *****************************************************************/

extern var_table_t var_table;          /* tabella contenente le variabili globali e le variabili locali 
					  alle sequenze nel file delle specifiche (escluse le variabili locali alle macro) */
extern macro_var_table_t macro_var_table;   /* tabella contenente le variabili locali alle macro */

extern int in_macro;                   /* indica se il parser sta esaminando una macro */

int
get_type_values(char *var_name, int num_macro)
{
  var_entry *var = NULL;
  
  if (in_macro)
    var = getmacrovar(macro_var_table, var_name, num_macro);
  
  if (var == NULL)
    {
      /* la variabile  globale oppure  locale ad una regola */
      var = getvar(var_table, var_name);
    }
  
  assert( var != NULL );
  
  return var->type_values;
}

var_t
get_type_var(char *var_name, int num_macro)
{
  var_entry *var = NULL;
  
  if (in_macro)
    var = getmacrovar(macro_var_table, var_name, num_macro);
  
  if (var == NULL)
    {
      /* la variabile  globale oppure  locale ad una regola */
      var = getvar(var_table, var_name);
    }
  
  assert( var != NULL );
  
  return var->type_var;
}




cond_funct_ptr 
eq(arg_node_t *left_op, arg_node_t *right_op, int num_macro)
{
  if (left_op->arg)
    {
      if (right_op->arg) return (*arg_arg)(left_op, right_op, num_macro);
      else               return (*arg_funct)(left_op, right_op, num_macro);
    }
  
  else
    {
      if (right_op->arg) return (*funct_arg)(left_op, right_op, num_macro);
      else               return (*funct_funct)(left_op, right_op, num_macro);
    }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"eq: ERROR in eq\n");
  exit(-1);
}

cond_funct_ptr 
arg_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE == SIMPLE_VALUE (chi la utilizza una cond. del genere ?) */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->type_value != r->arg->type_value) 
	    {
	      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (SIMPLE_VALUE == SIMPLE_VALUE)\n");
	      exit(-1);
	    }
	  switch (l->arg->type_value) {
	  case STRING_VALUE:
	    return eq_str_str;
	    break;
	  case CHARACTER_VALUE: 
	    return eq_chr_chr;
	    break;
	  case LONG_VALUE:
	    return eq_long_long;
	    break;
	  }
	}
      
      /* SIMPLE_VALUE == HANDLER */
      if (r->arg->type_content == HANDLER)
	{
	  
	  if (r->arg->content.handler->var_name) 
	    {
	      /* SIMPLE_VALUE == VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro))
		    {
		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (SIMPLE_VALUE == VAR)\n");
		      exit(-1);
		    }
		  
		  switch (l->arg->type_value) {
		  case STRING_VALUE:
		    return eq_str_var;
		    break;
		  case CHARACTER_VALUE: 
		    return eq_chr_var;
		    break;
		  case LONG_VALUE:
		    return eq_long_var;
		    break;
		  }
		}
	      /* SIMPLE_VALUE == HASH_VAR */
	      else
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (SIMPLE_VALUE == HASH_VAR)\n");
		      exit(-1);
		    }
		  
		  switch (l->arg->type_value) {
		  case STRING_VALUE:
		    return eq_str_hashvar;
		    break;
		  case CHARACTER_VALUE: 
		    return eq_chr_hashvar;
		    break;
		  case LONG_VALUE:
		    return eq_long_hashvar;
		    break;
		  }
		}

	    }
	  /* SIMPLE_VALUE == SEGNAPOSTO */
	  else
	    {
	      /* il valore del segnaposto viene preso a run-time ed in handler_t non ho l'effettivo valore */
	      if (l->arg->type_value !=
		  r->arg->content.handler->type_label)
		{
		  fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (SIMPLE_VALUE == LABEL)\n");
		  exit(-1);
		} 
	      
	      switch (l->arg->type_value) {
	      case STRING_VALUE:
		return eq_str_label;
		break;
	      case CHARACTER_VALUE: 
		return eq_chr_label;
		break;
	      case LONG_VALUE:
		return eq_long_label;
		break;
	      }
	    }
	}
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      /* HANDLER == SIMPLE_VALUE */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->content.handler->var_name) 
	    {
	      /* VAR == SIMPLE_VALUE */
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (VAR == SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  switch (r->arg->type_value) {
		  case STRING_VALUE:
		    return eq_var_str;
		    break;
		  case CHARACTER_VALUE: 
		    return eq_var_chr;
		    break;
		  case LONG_VALUE:
		    return eq_var_long;
		    break;
		  }
		}
	      /* HASH_VAR == SIMPLE_VALUE */
	      else
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (HASH_VAR == SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  switch (r->arg->type_value) {
		  case STRING_VALUE:
		    return eq_hashvar_str;
		    break;
		  case CHARACTER_VALUE: 
		    return eq_hashvar_chr;
		    break;
		  case LONG_VALUE:
		    return eq_hashvar_long;
		    break;
		  }
		}
	    }
	  /* SEGNAPOSTO == SIMPLE_VALUE */
	  else
	    {
	      if (r->arg->type_value !=
		  l->arg->content.handler->type_label)
		{
		  fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (LABEL == SIMPLE_VALUE)\n");
		  //printf("\n%d   %d\n",
		  // l->arg->content.handler->type_label,r->arg->type_value);
		  exit(-1);
		}
	      
	      switch (r->arg->type_value) {
	      case STRING_VALUE:
		return eq_label_str;
		break;
	      case CHARACTER_VALUE:
		return eq_label_chr;
		break;
	      case LONG_VALUE:
		return eq_label_long;
		break;
	      }
	    }
	}
      else /* HANDLER == HANDLER */
	{
	  if ((l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{ 
		  /* VAR == VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (VAR == VAR)\n");
			  exit(-1);
			}
		      
		      switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
		      case STRING_VALUE:
			return eq_var_var_str;
			break;
		      case CHARACTER_VALUE: 
			return eq_var_var_chr;
			break;
		      case LONG_VALUE:
			return eq_var_var_long;
			break;
		      }
		      
		    }
		  /* VAR == HASHVAR */ 
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (VAR == HASHVAR)\n");
			  exit(-1);
			}
		      return eq_var_hashvar;
		    }
		}
	      else 
		{
		  /* HASHVAR == VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar) 
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (HASHVAR == VAR)\n");
			  exit(-1);
			}
		      return eq_hashvar_var;
		    }
		    /* HASHVAR == HASHVAR */
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (HASHVAR == HASHVAR)\n");
			  exit(-1);
			}
		      return eq_hashvar_hashvar;
		    }
		}
	      
	    }
	  else if ( !(l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* SEGNAPOSTO == VAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      /* printf("tipo var = %d\n", get_type_values(r->arg->content.handler->var_name, num_macro));
			 printf("tipo sysc = %d\n", l->arg->content.handler->type_label);
			 printf("num arg = %d\n", l->arg->content.handler->label_ref.n_arg + 1); */

		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (LABEL == VAR)\n");
		      exit(-1);
		    }
		  
		  switch (get_type_values(r->arg->content.handler->var_name, num_macro)) {
		  case STRING_VALUE:
		    return eq_label_var_str;
		    break;
		  case CHARACTER_VALUE: 
		    return eq_label_var_chr;
		    break;
		  case LONG_VALUE:
		    return eq_label_var_long;
		    break;
		  }
		  
		}
	      else
		{
		  /* SEGNAPOSTO == HASHVAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (LABEL == HASHVAR)\n");
		      exit(-1);
		    }
		  
		  return eq_label_hashvar;
		}
	    }
	  else if ( (l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* VAR == SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (VAR == LABEL)\n");
		      exit(-1);
		    }
		  
		  switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
		  case STRING_VALUE:
		    return eq_var_label_str;
		    break;
		  case CHARACTER_VALUE: 
		    return eq_var_label_chr;
		    break;
		  case LONG_VALUE:
		    return eq_var_label_long;
		    break;
		  }
		  
		}
	      else
		{
		  /* HASHVAR == SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (HASHVAR == LABEL)\n");
		      exit(-1);
		    }
		  
		  return eq_hashvar_label;
		}
	    }
	  else if ( !(l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      /* SEGNAPOSTO == SEGNAPOSTO */

	      if ( l->arg->content.handler->type_label != 
		   r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"arg_arg: ERROR: type mismatch in condition (LABEL == LABEL)\n");
		  exit(-1);
		}
	      
	      switch ( l->arg->content.handler->type_label ) {
	      case STRING_VALUE:
		return eq_label_label_str;
		break;
	      case CHARACTER_VALUE: 
		return eq_label_label_chr;
		break;
	      case LONG_VALUE:
		return eq_label_label_long;
		break;
	      default:
		fprintf(stderr,"arg_arg: ERROR: type not permitted in condition (LABEL == LABEL)\n");
		exit(-1);
		break;
	      }
	      
	    }
	}
      
    }
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"arg_arg: ERROR: undefined case\n");
  exit(-1); 
}

/****************************************************************************/

cond_funct_ptr 
arg_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE == FUNCT */
      if ( l->arg->type_value != funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"arg_funct: ERROR: type mismatch in condition (SIMPLE_VALUE == FUNCT)\n");
	  exit(-1);
	}
      
      switch (l->arg->type_value) {
      case STRING_VALUE:
	return eq_str_funct;
	break;
      case CHARACTER_VALUE: 
	return eq_chr_funct;
	break;
      case LONG_VALUE:
	return eq_long_funct;
	break;
      }
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      if (l->arg->content.handler->var_name) 
	{
	  /* VAR == FUNCT */
	  
	  if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"arg_funct: ERROR: type mismatch in condition (VAR == FUNCT)\n");
		  exit(-1);
		}
	      
	      switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
	      case STRING_VALUE:
		return eq_var_funct_str;
		break;
	      case CHARACTER_VALUE: 
		return eq_var_funct_chr;
		break;
	      case LONG_VALUE:
		return eq_var_funct_long;
		break;
	      }
	      
	    }
	  /* HASH_VAR == FUNCT */
	  else
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"arg_funct: ERROR: type mismatch in condition (HASH_VAR == FUNCT)\n");
		  exit(-1);
		}
	      
	      return eq_hashvar_funct;
	    }
	}
      /* SEGNAPOSTO == FUNCT */
      else
	{
	  if ( l->arg->content.handler->type_label != funct_ret_type(r->funct) )
	    {
	      fprintf(stderr,"arg_funct: ERROR: type mismatch in condition (LABEL == FUNCT)\n");
	      exit(-1);
	    }
	  
	  
	  switch ( l->arg->content.handler->type_label ) {
	  case STRING_VALUE:
	    return eq_label_funct_str;
	    break;
	  case CHARACTER_VALUE: 
	    return eq_label_funct_chr;
	    break;
	  case LONG_VALUE:
	    return eq_label_funct_long;
	    break;
	  }
	  
	  
	}
    }
  
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"arg_funct: ERROR\n");
  exit(-1);
}


cond_funct_ptr
funct_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if (r->arg->type_content == SIMPLE_VALUE)
    {
      /* FUNCT == SIMPLE_VALUE */
      
      if ( funct_ret_type(l->funct) != r->arg->type_value )
	{
	  fprintf(stderr,"funct_arg: ERROR: type mismatch in condition (FUNCT == SIMPLE_VALUE)\n");
	  exit(-1);
	}

      
      switch (r->arg->type_value) {
      case STRING_VALUE:
	return eq_funct_str;
	break;
      case CHARACTER_VALUE: 
	return eq_funct_chr;
	break;
      case LONG_VALUE:
	return eq_funct_long;
	break;
      }
    }
  else /* (r->arg->type_content == HANDLER) */ 
    {
      if (r->arg->content.handler->var_name) 
	{
	  /* FUNCT == VAR */
	  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"funct_arg: ERROR: type mismatch in condition (FUNCT == VAR)\n");
		  exit(-1);
		}
	      
	      
	      switch (get_type_values(r->arg->content.handler->var_name, num_macro)) {
	      case STRING_VALUE:
		return eq_funct_var_str;
		break;
	      case CHARACTER_VALUE: 
		return eq_funct_var_chr;
		break;
	      case LONG_VALUE:
		return eq_funct_var_long;
		break;
	      }
	      
	    }
	  /* FUNCT == HASH_VAR */
	  else
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"funct_arg: ERROR: type mismatch in condition (FUNCT == HASH_VAR)\n");
		  exit(-1);
		}
	      
	      
	      return eq_funct_hashvar;
	    }
	}
      /* FUNCT == SEGNAPOSTO */
      else
	{
	  if ( funct_ret_type(l->funct) != r->arg->content.handler->type_label )
	    {
	      fprintf(stderr,"funct_arg: ERROR: type mismatch in condition (FUNCT == SEGNAPOSTO)\n");
	      exit(-1);
	    }
	  
	  
	  switch ( r->arg->content.handler->type_label ) {
	  case STRING_VALUE:
	    return eq_funct_label_str;
	    break;
	  case CHARACTER_VALUE: 
	    return eq_funct_label_chr;
	    break;
	  case LONG_VALUE:
	    return eq_funct_label_long;
	    break;
	  }
	  
	  
	}
    }
  
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"funct_arg: ERROR\n");
  exit(-1); 
}


cond_funct_ptr 
funct_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if ( funct_ret_type(l->funct) != funct_ret_type(r->funct) )
    {
      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT == FUNCT)\n");
      exit(-1);
    }
  
  switch (funct_ret_type(l->funct)) {
  case STRING_VALUE:
    return eq_funct_funct_str;
    break;
  case CHARACTER_VALUE:
    return eq_funct_funct_chr;
    break;
  case LONG_VALUE:
    return eq_funct_funct_long;
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in funct_funct\n");
  exit(-1); 
}


/*************************************************************************
 ****************** Funzioni chiamate a run-time *************************
 *************************************************************************/

int
eq_str_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  /* cambiare funzione se non e' efficiente */
  return (strcmp(l->arg->content.value, r->arg->content.value) == 0);
}

int
eq_str_var(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
      
  return (strcmp(l->arg->content.value,
		 ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) == 0);
}

int
eq_var_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return (strcmp(r->arg->content.value,
		 ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) == 0);
}

int
eq_str_hashvar(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return (g_hash_table_lookup( (GHashTable *)((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values,
			       l->arg->content.value) != NULL);
}

int
eq_hashvar_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return (g_hash_table_lookup( (GHashTable *)((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			       r->arg->content.value) != NULL);
}

int
eq_str_funct(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;
  
  char *ret_value = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;
  
  res = (strcmp(l->arg->content.value, ret_value) == 0);
  free(ret_value);
  
  return res;
}

int
eq_funct_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  char *ret_value = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res = (strcmp(r->arg->content.value, ret_value) == 0);
  free(ret_value);
  
  return res;
}

int 
eq_chr_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  return ( *(char *)(l->arg->content.value) == *(char *)(r->arg->content.value) );
}

int
eq_chr_var(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return ( *(char *)(l->arg->content.value) == 
	   *(char *)(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) );
}

int
eq_var_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return ( *(char *)(r->arg->content.value) == 
	   *(char *)(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) );
}

int
eq_chr_hashvar(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return (g_hash_table_lookup( (GHashTable *)((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values,
			       l->arg->content.value) != NULL);
}

int
eq_hashvar_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  return (g_hash_table_lookup( (GHashTable *)((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			       r->arg->content.value) != NULL);
}

int 
eq_chr_funct(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  char *ret_value = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res = ( *(char *)(l->arg->content.value) == *(char *)(ret_value) );
  free(ret_value);

  return res;
}

int 
eq_funct_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;
  
  char *ret_value = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res =  ( *(char *)(r->arg->content.value) == *(char *)(ret_value) );
  free(ret_value);

  return res;
}

int 
eq_long_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  return ( *(long *)(l->arg->content.value) == *(long *)(r->arg->content.value) );
}

int
eq_long_var(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return ( *(long *)(l->arg->content.value) == 
	   *(long *)(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) );
}

int
eq_var_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return ( *(long *)(r->arg->content.value) == 
	   *(long *)(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) );
}

int
eq_long_hashvar(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return (g_hash_table_lookup( (GHashTable *)((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values,
			       l->arg->content.value) != NULL);
}

int
eq_hashvar_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return (g_hash_table_lookup( (GHashTable *)((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			       r->arg->content.value) != NULL);
}

int 
eq_long_funct(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;
  
  long *ret_value = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;


  res = ( *(long *)(l->arg->content.value) == *(long *)(ret_value) );
  free(ret_value);

  return res;
}

int 
eq_funct_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  long *ret_value = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res = ( *(long *)(r->arg->content.value) == *(long *)(ret_value) );
  free(ret_value);

  return res;
}

int
eq_var_var_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  return (strcmp(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
		 ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) == 0);
}

int
eq_var_var_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  return ( *(char *)(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) == 
	   *(char *)(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) );
}

int
eq_var_var_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  return ( *(long *)(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) == 
	   *(long *)(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) );
}


int
eq_var_hashvar(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  return ( g_hash_table_lookup( (GHashTable *)((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values,
				((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) != NULL );
}

int
eq_hashvar_var(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
    
  return ( g_hash_table_lookup((GHashTable *)((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values, 
			       ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) != NULL );
}

int
eq_hashvar_hashvar(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *l_ht, *r_ht;
  
  if ( 
      g_hash_table_size(l_ht = (GHashTable *)((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values)
      !=
      g_hash_table_size(r_ht = (GHashTable *)((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values)
      ) 
    return 0;
  
  /* ATTENZIONE: e' inefficente perche' mi scorre tutti gli elementi nella hash table anche se il primo e' diverso */ 
  g_hash_table_foreach (l_ht, confront_ht, &r_ht);
  
  return (r_ht != NULL);
}

void
confront_ht (gpointer key, gpointer value, gpointer user_data)
{
  if ( *((GHashTable **)user_data) != NULL)
    if ( g_hash_table_lookup( *((GHashTable **)user_data), key ) == NULL )  
                                                        *((GHashTable **)user_data) = NULL;
}

int
eq_var_funct_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  char *ret_value = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res =  (strcmp(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
		     ret_value) == 0 );
  free(ret_value);
  
  return res;
}

int
eq_var_funct_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  char *ret_value = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res =  ( *(char *)(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) == 
	   *(char *) ret_value );
  free(ret_value);
  
  return res;
}

int
eq_var_funct_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  long *ret_value = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res =  ( *(long *)(((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values) == 
	       *(long *) ret_value );
  free(ret_value);
  
  return res;
}

int
eq_funct_var_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  char *ret_value = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res =  (strcmp(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values, 
		 (char *) ret_value) == 0 );
  free(ret_value);
  
  return res;
}

int
eq_funct_var_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  char *ret_value = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res = ( *(char *)(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) == 
	  *(char *) ret_value );
  free(ret_value);
  
  return res;
}

int
eq_funct_var_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  long *ret_value = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res = ( *(long *)(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values) == 
	  *(long *) ret_value );
  free(ret_value);
  
  return res;
}


int
eq_hashvar_funct(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;


  void *ret_value = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res = ( g_hash_table_lookup((GHashTable *)((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
				  ret_value) != NULL );
  free(ret_value);
  
  return res;
}

int
eq_funct_hashvar(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  int res;

  void *ret_value = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (ret_value == NULL))
    return NOEV;

  res = ( g_hash_table_lookup((GHashTable *)((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values,
				  ret_value) != NULL );
  free(ret_value);
  
  return res;
}


/*************************************************************
 ******** Funzioni chiamate a run-time con SEGNAPOSTO ********
 *************************************************************/

/*
int upeek (int a, long b, long *c)
{
}
*/

/* struct sysent sysent2[10]; */

void *
take_arg(int num_arg, struct tcb *tcp) /* num_arg == 0 indica il primo argomento */
{
  if (num_arg == RVAL) /* se num_arg == rval allora restituire valore di ritorno */
    {
      tcp->arg_state[num_arg] = TAKEN; /* ATTENZIONE: perche' ? */
      return &tcp->u_rval; /* ATTENZIONE: potrebbe essere necessario fare una clone_long(&tcp->u_rval) */
    }
  else /* num_arg e' >= 0 dunque e' un arg della syscall */
    {
      switch ( tcp->arg_state[num_arg] )
	{  
	case EMPTY: 
	  if (upeek(tcp->pid, (num_arg) * 4, &tcp->u_arg[num_arg]) < 0)
	    {
	      fprintf(stderr, "take_arg: upeek failed in case EMPTY\n");
	      return NULL;
	    }
	  tcp->arg_state[num_arg] = UPEEK;
	case UPEEK:
	  if ( (*sysent[tcp->scno].sys_func)(tcp, num_arg) < 0 )
	    {
#ifdef DEBUG_WIZ
	      fprintf(stderr, "take_arg: argument %d of syscall %ld not catched\n", num_arg+1, tcp->scno);
#endif
	      return NULL;
	    }
	  break;
	case TAKEN:
	  if ( entering(tcp) || !is_both(tcp->scno, num_arg + 1)) 
	    break;
	  
	  free((void *)tcp->u_arg[num_arg]);
	  if (upeek(tcp->pid, (num_arg) * 4, &tcp->u_arg[num_arg]) < 0)
	    {
	      fprintf(stderr, "take_arg: upeek failed in case TAKEN\n");
	      return NULL;
	    }
	  tcp->arg_state[num_arg] = UPEEK;
	  if ( (*sysent[tcp->scno].sys_func)(tcp, num_arg) < 0 )
	    {
	      fprintf(stderr, "take_arg: argument %d of syscall %ld not catched - case TAKEN\n", num_arg+1, tcp->scno);
	      return NULL;
	    }
	}
      
      return (void *) tcp->u_arg[num_arg];
    } 
}


int
eq_str_label(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;

  char *arg;

  //printf("arg_state: %d n_arg: %ld\n",tcp->arg_state[r->arg->content.handler->label_ref.n_arg],(r->arg->content.handler->label_ref.n_arg + 1));

  
  arg = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (l->arg->content.value == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno)); 
      return 0;
    }
  
  return (strcmp(l->arg->content.value, arg) == 0);
}

int
eq_label_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  char * arg;

  //printf("arg_state: %d n_arg: %ld\n",tcp->arg_state[l->arg->content.handler->label_ref.n_arg],(l->arg->content.handler->label_ref.n_arg + 1));

  arg = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (r->arg->content.value == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return (strcmp(r->arg->content.value, arg) == 0);
}

int
eq_chr_label(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  char *arg = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (l->arg->content.value == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }

  return ( *(char *)(l->arg->content.value) == *(char *)(arg) );
}

int
eq_label_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  char *arg = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);

  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (r->arg->content.value == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }  

  return ( *(char *)(r->arg->content.value) == *(char *)(arg) );
}

int
eq_long_label(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  long *arg = (long *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (l->arg->content.value == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return ( *(long *)(l->arg->content.value) == *(long *)(arg) );  
}

int
eq_label_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  long *arg = (long *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (r->arg->content.value == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return ( *(long *)(r->arg->content.value) == *(long *)(arg) );  
}



int
eq_label_var_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  char *arg = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  char *var = ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values;
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (var == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return (strcmp(arg, var) == 0);
}

int
eq_label_var_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  char *arg = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  char *var = ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values;

  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (var == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return ( *(char *)(arg) == *(char *)(var) );
}

int
eq_label_var_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  long *arg = (long *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  long *var = ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values;
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (var == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return ( *(long *)(arg) == *(long *)(var) );
}


int
eq_var_label_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  char *arg = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *var = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values;
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (var == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return (strcmp(arg, var) == 0);
}

int
eq_var_label_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  char *arg = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *var = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (var == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return ( *(char *)(arg) == *(char *)(var) );
}

int
eq_var_label_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  long *arg = (long *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  long *var = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values;
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (var == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return ( *(long *)(arg) == *(long *)(var) );
}


int
eq_label_hashvar(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  void *arg = take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  void *hashvar = g_hash_table_lookup( 
	   (GHashTable *)((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values, arg);
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (hashvar == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return (hashvar != NULL);
}

int
eq_hashvar_label(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  void *arg = take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  void *hashvar = g_hash_table_lookup( 
	   (GHashTable *)((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values, arg);
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (hashvar == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return (hashvar != NULL);
}


int
eq_label_label_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  char *arg_l = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);  
  char *arg_r = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
	{
	  if (arg_l == arg_r) return 1;
	}
      else 
	{
	  fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
	  return 0;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  return (strcmp(arg_l, arg_r) == 0);
}

int
eq_label_label_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  char *arg_l = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);  
  char *arg_r = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
	{
	  if (arg_l == arg_r) return 1;
	}
      else 
	{
	  fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
	  return 0;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }

  return ( *(arg_l) == *(arg_r) );
}

int
eq_label_label_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  long *arg_l = (long *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);  
  long *arg_r = (long *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
	{
	  if (arg_l == arg_r) return 1;
	}
      else 
	{
	  fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
	  return 0;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }

  return ( *(arg_l) == *(arg_r) );
}


int
eq_label_funct_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  int res;
  char *arg = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  char *retvalue_ptr = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (retvalue_ptr == NULL))
    return NOEV;

  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (retvalue_ptr == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }

  res = (strcmp((char *)arg, retvalue_ptr) == 0);
  free(retvalue_ptr);
  return res;
}

int
eq_label_funct_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  int res;
  char *arg = (char *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  char *retvalue_ptr = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (retvalue_ptr == NULL))
    return NOEV;
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (retvalue_ptr == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }

  res = ( *(char *)(arg) == *(char *)(retvalue_ptr) );
  free(retvalue_ptr);
  return res;
}

int
eq_label_funct_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  int res;
  long *arg = (long *) take_arg(l->arg->content.handler->label_ref.n_arg, tcp);
  long *retvalue_ptr = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval || (retvalue_ptr == NULL))
    return NOEV;
  
  if (tcp->arg_state[l->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (retvalue_ptr == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", l->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  res = ( *(long *)(arg) == *(long *)(retvalue_ptr) );
  free(retvalue_ptr);
  return res;
}


int
eq_funct_label_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  int res;
  char *arg = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *retvalue_ptr = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (retvalue_ptr == NULL))
    return NOEV;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (retvalue_ptr == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }

  res = (strcmp((char *)arg, retvalue_ptr) == 0);
  free(retvalue_ptr);
  return res;
}

int
eq_funct_label_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  int res;
  char *arg = (char *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *retvalue_ptr = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (retvalue_ptr == NULL))
    return NOEV;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (retvalue_ptr == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }

  res = ( *(char *)(arg) == *(char *)(retvalue_ptr) );
  free(retvalue_ptr);
  return res;
}

int
eq_funct_label_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  //arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  
  int res;
  long *arg = (long *) take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  long *retvalue_ptr = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval || (retvalue_ptr == NULL))
    return NOEV;
  
  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (retvalue_ptr == arg) return 1;
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno));
      return 0;
    }
  
  res = ( *(long *)(arg) == *(long *)(retvalue_ptr) );
  free(retvalue_ptr);
  return res;
}


/******************************************************************
 ******************************************************************/



int
eq_funct_funct_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  void *l_res;
  void *r_res;
  int res;
  
  l_res = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval)
    return NOEV;
  r_res = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return NOEV;
  
  res = (strcmp(l_res, r_res) == 0);
  free(l_res);
  free(r_res);
  return res;
}

int
eq_funct_funct_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  void *l_res;
  void *r_res;
  int res;

  l_res = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval)
    return NOEV;
  r_res = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return NOEV;

  res = ( *(char *)(l_res) == *(char *)(r_res) );
  free(l_res);
  free(r_res);
  return res;
}

int
eq_funct_funct_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  void *l_res;
  void *r_res;
  int res;

  l_res = eval_funct(g_node_first_child(funct_tree), ht, tcp);
  if (func_not_eval)
    return NOEV;
  r_res = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return NOEV;
 

  res = ( *(long *)(l_res) == *(long *)(r_res) );
  free(l_res);
  free(r_res);
  return res;
}


/***********************************************************************************************************
 ********************************************* eval_funct **************************************************
 ***********************************************************************************************************/


/* mi restituisce il valore di ritorno della funzione in   funct_tree->data->funct 
   dopo averla eseguita a run-time */
void *
eval_funct(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  void * args[10];
  void * funct_args[10] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  int i = 0; /* contatore per gli argomenti della funzione */
  int j = 0; /* conta quante funzioni ci sono come argomenti della funzione */
  void * ret_value;
  arg_node_t *a;
  arg_node_t *f = funct_tree->data;
  GNode *act = NULL;
  
  /* int num_args = g_node_n_children(funct_tree);
     args = (void **) calloc(num_args, sizeof(void *)); */
  
  func_not_eval = 0;

  act = g_node_first_child(funct_tree);
  while (act != NULL)
    {
      a = act->data;
      if (a->arg) /* e' un valore */
	{
	  if (a->arg->type_content == SIMPLE_VALUE)
	    {
	      args[i] = a->arg->content.value;
	    }
	  else if (a->arg->type_content == HANDLER)
	    {
	      if  (a->arg->content.handler->var_name) /* var o hashvar */
		{
		  args[i] = ((var_entry *)g_hash_table_lookup(ht, a->arg->content.handler->var_name))->values;
		}
	      else  /* SEGNAPOSTO */
		{
		  args[i] = take_arg(a->arg->content.handler->label_ref.n_arg, tcp);
		  
		  if (tcp->arg_state[a->arg->content.handler->label_ref.n_arg] != TAKEN)
		    {
#ifdef DEBUG_WIZ
		      fprintf(stderr,"eval_funct: arg %ld of syscall %s wasn't taken: function not evaluated\n", 
			      (a->arg->content.handler->label_ref.n_arg + 1) ,get_sysc_name(tcp->scno));
#endif
		      
		      func_not_eval = 1;
		      //return (void *)-1;
		      return NULL;
		    }
		}
	    }
	}
      else /* e' una funzione */
	{
	  funct_args[j] = args[i] = eval_funct(act, ht, tcp);

	  if (func_not_eval)
	    {
	      for (j = 0; funct_args[j] != NULL; j++) /* in caso di funzione non valutata, dealloco gli eventuali valori
							 restituiti da altre funzioni */
		free(funct_args[j]);
	      
	      return NULL;
	    }
	  
	  j++;
	}
      
      i++;
      act = g_node_next_sibling(act);
    }
  
  switch (i) {
  case 0:
    ret_value = (*f->funct)();
    break;
  case 1:
    ret_value = (*f->funct)(args[0]);
    break;
  case 2: 
    ret_value = (*f->funct)(args[0], args[1]);
    break;
  case 3:
    ret_value = (*f->funct)(args[0], args[1], args[2]);
    break;
  case 4:
    ret_value = (*f->funct)(args[0], args[1], args[2], args[3]);
    break;
  case 5:
    ret_value = (*f->funct)(args[0], args[1], args[2], args[3], args[4]);
    break;
  case 6:
    ret_value = (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5]);
    break;
  case 7:
    ret_value = (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6]);
    break;
  case 8:
    ret_value = (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7]);
    break;
  case 9:
    ret_value = (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8]);
    break;
  case 10:
    ret_value = (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8], args[9]);
    break;
  }
  
  /* dealloco gli eventuali valori restituiti da funzioni */
  for (j = 0; funct_args[j] != NULL; j++)
    free(funct_args[j]);
  
  return ret_value;
}




/******************************************************************************************************************
 **************************************** Funzioni per gli statements *********************************************
 ******************************************************************************************************************/


act_funct_ptr
assign(arg_node_t *left_op, arg_node_t *right_op, int num_macro)
{
  /* VAR = ... */
  if (get_type_var(left_op->arg->content.handler->var_name, num_macro) == scalar)
    return (*assign_to_var)(left_op, right_op, num_macro);
  
  else /* HASHVAR = ... */
    return (*assign_to_hashvar)(left_op, right_op, num_macro);

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in assign\n");
  exit(-1);
}


act_funct_ptr 
assign_to_var(arg_node_t *l, arg_node_t *r, int num_macro)
{
  if (r->arg) /* VAR = value */
    {
      if (r->arg->type_content == SIMPLE_VALUE)
	{                                       
	  /* controllo che var e value abbiano il solito tipo */
	  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in assignment (VAR = value)\n");
	      exit(-1);
	    }
	  
	  switch (r->arg->type_value) {
	  case STRING_VALUE:
	    return as_var_str;
	    break;
	  case CHARACTER_VALUE: 
	    return as_var_chr;
	    break;
	  case LONG_VALUE:
	    return as_var_long;
	    break;
	  }
	}
      
      else if (r->arg->type_content == HASH_VALUE)
	{
	  fprintf(stderr,"\nERROR: cannot assign more values to a simple_var (VAR = hash_value)\n");
	  exit(-1);
	}
      
      else if (r->arg->type_content == HANDLER)
	{
	  if (r->arg->content.handler->var_name) 
	    {
	      /* VAR = VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* controllo le var abbiano il solito tipo */
		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in assignment (VAR = VAR)\n");
		      exit(-1);
		    }
		  
		  switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
		  case STRING_VALUE:
		    return as_var_var_str;
		    break;
		  case CHARACTER_VALUE: 
		    return as_var_var_chr;
		    break;
		  case LONG_VALUE:
		    return as_var_var_long;
		    break;
		  }
		}
	      /* VAR = HASH_VAR */
	      else
		{
		  fprintf(stderr,"\nERROR: cannot assign an hash_var to simple_var (VAR = HASH_VAR)\n");
		  exit(-1);
		}
	      
	    }
	  /* VAR = SEGNAPOSTO */
	  else
	    {
	      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		  r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"\nERROR: type mismatch in assignment (VAR = LABEL)\n");
		  exit(-1);
		}
	      
	      switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
	      case STRING_VALUE:
		return as_var_label_str;
		break;
	      case CHARACTER_VALUE: 
		return as_var_label_chr;
		break;
	      case LONG_VALUE:
		return as_var_label_long;
		break;
	      }
	    }
	}
      
    }
  else /* VAR = funct */
    {
      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != 
	   funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"\nERROR: type mismatch in assignment (VAR = funct)\n");
	  exit(-1);
	}

            
      switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
      case STRING_VALUE:
	return as_var_funct_str;
	break;
      case CHARACTER_VALUE: 
	return as_var_funct_chr;
	break;
      case LONG_VALUE:
	return as_var_funct_long;
	break;
      }
      
    }

  /* non ci dovrebbe mai arrivare */
  fprintf(stderr,"ERRORE in assign_to_var\n");
  exit(-1);
}


act_funct_ptr 
assign_to_hashvar(arg_node_t *l, arg_node_t *r, int num_macro)
{
  if (r->arg) /* HASHVAR = value */
    {
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  
	  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in assignment (HASHVAR = value)\n");
	      exit(-1);
	    }
	  
	  switch (r->arg->type_value) {
	  case STRING_VALUE:
	    return as_hashvar_str;
	    break;
	  case CHARACTER_VALUE: 
	    return as_hashvar_chr;
	    break;
	  case LONG_VALUE:
	    return as_hashvar_long;
	    break;
	  }
	}
      
      else if (r->arg->type_content == HASH_VALUE)
	{
	  
	  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in assignment (HASHVAR = hash_value)\n");
	      exit(-1);
	    }
	  
	  switch (r->arg->type_value) {
	  case STRING_VALUE:
	    return as_hashvar_hstr;
	    break;
	  case CHARACTER_VALUE: 
	    return as_hashvar_hchr;
	    break;
	  case LONG_VALUE:
	    return as_hashvar_hlong;
	    break;
	  }
	}
      
      else if (r->arg->type_content == HANDLER)
	{
	  if (r->arg->content.handler->var_name) 
	    {
	      /* HASHVAR = VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* controllo le var abbiano il solito tipo */
		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in assignment (HASHVAR = VAR)\n");
		      exit(-1);
		    }
		  
		  switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
		  case STRING_VALUE:
		    return as_hashvar_var_str;
		    break;
		  case CHARACTER_VALUE: 
		    return as_hashvar_var_chr;
		    break;
		  case LONG_VALUE:
		    return as_hashvar_var_long;
		    break;
		  }
		  
		  
		}
	      /* HASHVAR = HASHVAR */
	      else
		{
		  /* controllo le var abbiano il solito tipo */
		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in assignment (HASHVAR = HASHVAR)\n");
		      exit(-1);
		    }
		  
		  switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
		  case STRING_VALUE:
		    return as_hashvar_hashvar_str;
		    break;
		  case CHARACTER_VALUE: 
		    return as_hashvar_hashvar_chr;
		    break;
		  case LONG_VALUE:
		    return as_hashvar_hashvar_long;
		    break;
		  }
		  
		}
	      
	    }
	  /* HASHVAR = SEGNAPOSTO */
	  else
	    {
	      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		  r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"\nERROR: type mismatch in assignment (HASHVAR = LABEL)\n");
		  exit(-1);
		}
	      
	      switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
	      case STRING_VALUE:
		return as_hashvar_label_str;
		break;
	      case CHARACTER_VALUE: 
		return as_hashvar_label_chr;
		break;
	      case LONG_VALUE:
		return as_hashvar_label_long;
		break;
	      }
	      
	      
	    }
	}
      
    }
  else /* HASHVAR = funct */
    {
      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != 
	   funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"\nERROR: type mismatch in assignment (HASHVAR = funct)\n");
	  exit(-1);
	}
      
            
      switch (get_type_values(l->arg->content.handler->var_name, num_macro)) {
      case STRING_VALUE:
	return as_hashvar_funct_str;
	break;
      case CHARACTER_VALUE: 
	return as_hashvar_funct_chr;
	break;
      case LONG_VALUE:
	return as_hashvar_funct_long;
	break;
      }

      
    }
  
  /* non ci dovrebbe mai arrivare */
  fprintf(stderr,"ERRORE in assign_to_hashvar\n");
  exit(-1);
}


/*******************************************************************
 ******** Funzioni chiamate a run-time per gli assegnamenti ********
 *******************************************************************/


/* utilizzata per eseguire funzioni messe come azioni */
void
eval_stmt(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  void * args[10];
  void * funct_args[10] = {NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL};
  int i = 0; /* contatore per gli argomenti della funzione */
  int j = 0; /* conta quante funzioni ci sono come argomenti della funzione */
  arg_node_t *a;
  arg_node_t *f = g_node_first_child(funct_tree)->data;
  GNode *act = NULL;
  
  /* int num_args = g_node_n_children(g_node_first_child(funct_tree));
     args = (void **) calloc(num_args, sizeof(void *)); */
  
  act = g_node_first_child(g_node_first_child(funct_tree));
  while (act != NULL)
    {
      a = act->data;
      if (a->arg) /* e' un valore */
	{
	  if (a->arg->type_content == SIMPLE_VALUE)
	    {
	      args[i] = a->arg->content.value;
	    }
	  else if (a->arg->type_content == HANDLER)
	    {
	      if  (a->arg->content.handler->var_name) /* var o hashvar */
		{
		  args[i] = ((var_entry *)g_hash_table_lookup(ht, a->arg->content.handler->var_name))->values;
		}
	      else  /* SEGNAPOSTO */
		{
		  args[i] = take_arg(a->arg->content.handler->label_ref.n_arg, tcp);
		  
		  if (tcp->arg_state[a->arg->content.handler->label_ref.n_arg] != TAKEN)
		    {
#ifdef DEBUG_WIZ
		      fprintf(stderr,"eval_stmt: arg %ld of syscall %s wasn't taken: statement not evaluated\n", 
			      (a->arg->content.handler->label_ref.n_arg + 1) ,get_sysc_name(tcp->scno));fflush(NULL);
#endif
		      
		      return;
		    }
		}
	    }
	}
      else /* e' una funzione */
	{
	  funct_args[j] = args[i] = eval_funct(act, ht, tcp);
	  
	  if (func_not_eval)
	    {
	      fprintf(stderr,"eval_stmt: func_not_eval in eval_stmt\n");fflush(NULL);
	      
	      for (j = 0; funct_args[j] != NULL; j++) /* in caso di funzione non valutata, dealloco gli eventuali valori
							 restituiti da altre funzioni */
		free(funct_args[j]);
	      
	      return;
	    }
	  
	  j++;
	}
      
      i++;
      act = g_node_next_sibling(act);
    }
  
  switch (i) {
  case 0:
    (*f->funct)();
    break;
  case 1:
    (*f->funct)(args[0]);
    break;
  case 2: 
    (*f->funct)(args[0], args[1]);
    break;
  case 3:
    (*f->funct)(args[0], args[1], args[2]);
    break;
  case 4:
    (*f->funct)(args[0], args[1], args[2], args[3]);
    break;
  case 5:
    (*f->funct)(args[0], args[1], args[2], args[3], args[4]);
    break;
  case 6:
    (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5]);
    break;
  case 7:
    (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6]);
    break;
  case 8:
    (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7]);
    break;
  case 9:
    (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8]);
    break;
  case 10:
    (*f->funct)(args[0], args[1], args[2], args[3], args[4], args[5], args[6], args[7], args[8], args[9]);
    break;
  }
  
  /* dealloco gli eventuali valori restituiti da funzioni */
  for (j = 0; funct_args[j] != NULL; j++)
    free(funct_args[j]);
  
  return;
}  



void
as_var_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono il valore da assegnare alla var */
  char *dup = clone_string(r->arg->content.value);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}

void
as_var_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono il valore da assegnare alla var */
  char *dup = clone_char(r->arg->content.value);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}

void
as_var_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
    
  /* clono il valore da assegnare alla var */
  long *dup = clone_long(r->arg->content.value);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}


void
as_var_var_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
    
  /* clono il valore da assegnare alla var */
  char *dup = clone_string(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}

void
as_var_var_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
    
  /* clono il valore da assegnare alla var */
  char *dup = clone_char(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}

void
as_var_var_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
    
  /* clono il valore da assegnare alla var */
  long *dup = clone_long(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}


void
as_var_label_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  char *arg = take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *dup;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (arg == NULL) 
	{
	  /* dealloco il precedente valore della var */
	  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
	  /* assegno il nuovo valore alla var */
	  var_ptr->values = NULL;
	  return;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken:\n Assignment not executed\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno)); 
      return;
    }
  
  /* clono il valore da assegnare alla var */
  dup = clone_string(arg);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}

void
as_var_label_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  char *arg = take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *dup;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (arg == NULL) 
	{
	  /* dealloco il precedente valore della var */
	  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
	  /* assegno il nuovo valore alla var */
	  var_ptr->values = NULL;
	  return;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken:\n Assignment not executed\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno)); 
      return;
    }
  
  /* clono il valore da assegnare alla var */
  dup = clone_char(arg);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}

void
as_var_label_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  long *arg = take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  long *dup;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (arg == NULL) 
	{
	  /* dealloco il precedente valore della var */
	  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
	  /* assegno il nuovo valore alla var */
	  var_ptr->values = NULL;
	  return;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken:\n Assignment not executed\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno)); 
      return;
    }
  
  /* clono il valore da assegnare alla var */
  dup = clone_long(arg);
  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  /* assegno il nuovo valore alla var */
  var_ptr->values = dup;
}


void
as_var_funct_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  char *tmp;
    
  /* clono il valore da assegnare alla var */
  //char *dup = clone_string( eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp) );
  
  /* assegno il nuovo valore alla var */
  tmp = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return;

  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  //var_ptr->values = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  var_ptr->values = tmp;
}

void
as_var_funct_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  char *tmp;
    
  /* clono il valore da assegnare alla var */
  //char *dup = clone_char( eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp) );
  
  /* assegno il nuovo valore alla var */
  tmp = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return;

  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );
  
  //var_ptr->values = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  var_ptr->values = tmp;
}

void
as_var_funct_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  long *tmp;
    
  /* clono il valore da assegnare alla var */
  //long *dup = clone_long( eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp) );

  /* assegno il nuovo valore alla var */
  tmp = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return;

  /* dealloco il precedente valore della var */
  free( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values );  
  
  //var_ptr->values = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  var_ptr->values = tmp;
}




void
as_hashvar_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  char *dup = clone_string(r->arg->content.value);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			      r->arg->content.value ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  char *dup = clone_char(r->arg->content.value);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			      r->arg->content.value ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  long *dup = clone_long(r->arg->content.value);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			      r->arg->content.value ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}


void
as_hashvar_hstr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono la hashtable da assegnare alla hashvar */
  GHashTable *dup = clone_hash_table(r->arg->content.hash_value, STRING_VALUE);
  /* dealloco la precedente hashtable assegnata a hashvar */
  if ( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values )
    free_hash_table( var_ptr->values );
  /* assegno il nuovo valore alla hashvar */
  var_ptr->values = dup;
}

void
as_hashvar_hchr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono la hashtable da assegnare alla hashvar */
  GHashTable *dup = clone_hash_table(r->arg->content.hash_value, CHARACTER_VALUE);
  /* dealloco la precedente hashtable assegnata a hashvar */
  if ( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values )
    free_hash_table( var_ptr->values );
  /* assegno il nuovo valore alla hashvar */
  var_ptr->values = dup;
}

void
as_hashvar_hlong(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono la hashtable da assegnare alla hashvar */
  GHashTable *dup = clone_hash_table(r->arg->content.hash_value, LONG_VALUE);
  /* dealloco la precedente hashtable assegnata a hashvar */
  if ( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values )
    free_hash_table( var_ptr->values );
  /* assegno il nuovo valore alla hashvar */
  var_ptr->values = dup;
}


void
as_hashvar_var_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  char *dup = clone_string(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_var_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  char *dup = clone_char(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_var_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  long *dup = clone_long(((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}


void
as_hashvar_hashvar_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono la hashtable da assegnare alla hashvar */
  GHashTable *old = ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values;
  GHashTable *dup = clone_hash_table(old, STRING_VALUE);
  /* dealloco la precedente hashtable assegnata a hashvar */
  if ( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values )
    free_hash_table( var_ptr->values );
  /* assegno il nuovo valore alla hashvar */
  var_ptr->values = dup;
}

void
as_hashvar_hashvar_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono la hashtable da assegnare alla hashvar */
  GHashTable *old = ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values;
  GHashTable *dup = clone_hash_table(old, CHARACTER_VALUE);
  /* dealloco la precedente hashtable assegnata a hashvar */
  if ( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values )
    free_hash_table( var_ptr->values );
  /* assegno il nuovo valore alla hashvar */
  var_ptr->values = dup;
}

void
as_hashvar_hashvar_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  var_entry *var_ptr;
  
  /* clono la hashtable da assegnare alla hashvar */
  GHashTable *old = ((var_entry *)g_hash_table_lookup(ht, r->arg->content.handler->var_name))->values;
  GHashTable *dup = clone_hash_table(old, LONG_VALUE);
  /* dealloco la precedente hashtable assegnata a hashvar */
  if ( (var_ptr = (var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values )
    free_hash_table( var_ptr->values );
  /* assegno il nuovo valore alla hashvar */
  var_ptr->values = dup;
}


void
as_hashvar_label_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  char *arg = take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *dup;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (arg == NULL) 
	{
	  return;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken:\n Assignment not executed\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno)); 
      return;
    }
  
  /* clono il valore da assegnare alla hashvar */
  dup = clone_string(arg);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values,
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_label_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  char *arg = take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  char *dup;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (arg == NULL) 
	{
	  return;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken:\n Assignment not executed\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno)); 
      return;
    }

  /* clono il valore da assegnare alla hashvar */
  dup = clone_char(arg);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values, 
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_label_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  long *arg = take_arg(r->arg->content.handler->label_ref.n_arg, tcp);
  long *dup;

  if (tcp->arg_state[r->arg->content.handler->label_ref.n_arg] == TAKEN)
    {
      if (arg == NULL) 
	{
	  return;
	}
    }
  else
    {
      fprintf(stderr,"arg %ld of syscall %s wasn't taken:\n Assignment not executed\n", r->arg->content.handler->label_ref.n_arg + 1,get_sysc_name(tcp->scno)); 
      return;
    }
  
  /* clono il valore da assegnare alla hashvar */
  dup = clone_long(arg);
  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values, 
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}


void
as_hashvar_funct_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  //char *dup = clone_string( eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp) );
  char *dup = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return;

  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values, 
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_funct_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  //char *dup = clone_char( eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp) );
  char *dup = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return;

  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values, 
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}

void
as_hashvar_funct_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  arg_node_t *l = g_node_first_child(funct_tree)->data;
  //arg_node_t *r = g_node_next_sibling(g_node_first_child(funct_tree))->data;
  GHashTable *ht_ptr;
  
  /* clono il valore da assegnare alla hashvar */
  //long *dup = clone_long( eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp) );
  long *dup = eval_funct(g_node_next_sibling(g_node_first_child(funct_tree)), ht, tcp);
  if (func_not_eval)
    return;

  /* dealloco il valore se lo trovo in hashvar */
  free( g_hash_table_lookup ( ht_ptr = ((var_entry *)g_hash_table_lookup(ht, l->arg->content.handler->var_name))->values, 
			      dup ) );
  /* assegno il nuovo valore alla hashvar */
  g_hash_table_insert(ht_ptr, dup, dup);
}


/********************************************************************************************************************
 ********************************************************************************************************************
 ********************************************************************************************************************/


cond_funct_ptr 
ne(arg_node_t *left_op, arg_node_t *right_op, int num_macro)
{
  if (left_op->arg)
    {
      if (right_op->arg) return (*ne_arg_arg)(left_op, right_op, num_macro);
      else               return (*ne_arg_funct)(left_op, right_op, num_macro);
    }
  
  else
    {
      if (right_op->arg) return (*ne_funct_arg)(left_op, right_op, num_macro);
      else               return (*ne_funct_funct)(left_op, right_op, num_macro);
    }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ne\n");
  exit(-1);
}

cond_funct_ptr 
ne_arg_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type = 0;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE == SIMPLE_VALUE (chi la utilizza una cond. del genere ?) */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->type_value != r->arg->type_value) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE != SIMPLE_VALUE)\n");
	      exit(-1);
	    }
	  type = l->arg->type_value;
	  
	}
      
      /* SIMPLE_VALUE == HANDLER */
      if (r->arg->type_content == HANDLER)
	{
	  
	  if (r->arg->content.handler->var_name) 
	    {
	      /* SIMPLE_VALUE == VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro))
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE != VAR)\n");
		      exit(-1);
		    }
		  
		  type = l->arg->type_value;
		  
		}
	      /* SIMPLE_VALUE == HASH_VAR */
	      else
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE != HASH_VAR)\n");
		      exit(-1);
		    }
		  
		  return ne_v_ht;
		  
		}

	    }
	  /* SIMPLE_VALUE == SEGNAPOSTO */
	  else
	    {
	      /* il valore del segnaposto viene preso a run-time ed in handler_t non ho l'effettivo valore */
	      if (l->arg->type_value !=
		  r->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE != LABEL)\n");
		  exit(-1);
		} 
	      
	      type = l->arg->type_value;
	      
	    }
	}
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      /* HANDLER == SIMPLE_VALUE */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->content.handler->var_name) 
	    {
	      /* VAR == SIMPLE_VALUE */
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR != SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  type=r->arg->type_value;
		  
		}
	      /* HASH_VAR == SIMPLE_VALUE */
	      else
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR != SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  return ne_ht_v;
		  
		}
	    }
	  /* SEGNAPOSTO == SIMPLE_VALUE */
	  else
	    {
	      if (r->arg->type_value !=
		  l->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL != SIMPLE_VALUE)\n");
		  exit(-1);
		}
	      
	      type=r->arg->type_value;
	      
	    }
	}
      else /* HANDLER == HANDLER */
	{
	  if ((l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{ 
		  /* VAR == VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR != VAR)\n");
			  exit(-1);
			}
		      
		      type=get_type_values(l->arg->content.handler->var_name, num_macro);
		      
		      
		    }
		  /* VAR == HASHVAR */ 
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR != HASHVAR)\n");
			  exit(-1);
			}
		      return ne_v_ht;
		    }
		}
	      else 
		{
		  /* HASHVAR == VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar) 
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR != VAR)\n");
			  exit(-1);
			}
		      return ne_ht_v;
		    }
		    /* HASHVAR == HASHVAR */
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR != HASHVAR)\n");
			  exit(-1);
			}
		      return ne_ht_ht;
		    }
		}
	      
	    }
	  else if ( !(l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* SEGNAPOSTO == VAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      /* printf("tipo var = %d\n", get_type_values(r->arg->content.handler->var_name, num_macro));
			 printf("tipo sysc = %d\n", l->arg->content.handler->type_label);
			 printf("num arg = %ld\n", l->arg->content.handler->label_ref.n_arg + 1); */
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL != VAR)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(r->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* SEGNAPOSTO == HASHVAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL != HASHVAR)\n");
		      exit(-1);
		    }
		  
		  return ne_v_ht;
		}
	    }
	  else if ( (l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* VAR == SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR != LABEL)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(l->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* HASHVAR == SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR != LABEL)\n");
		      exit(-1);
		    }
		  
		  return ne_ht_v;
		}
	    }
	  else if ( !(l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      /* SEGNAPOSTO == SEGNAPOSTO */

	      if ( l->arg->content.handler->type_label != 
		   r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL != LABEL)\n");
		  exit(-1);
		}
	      
	      type= l->arg->content.handler->type_label;
	      
	      
	    }
	}
      
    }
  
  switch (type) {
  case STRING_VALUE:
    return ne_str_str;
    break;
  case CHARACTER_VALUE: 
    return ne_chr_chr;
    break;
  case LONG_VALUE:
    return ne_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in ne_arg_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ne_arg_arg (caso non considerato)\n");
  exit(-1); 
}

/****************************************************************************/

cond_funct_ptr 
ne_arg_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE == FUNCT */
      if ( l->arg->type_value != funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE != FUNCT)\n");
	  exit(-1);
	}
      
            
      type=l->arg->type_value;
      
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      if (l->arg->content.handler->var_name) 
	{
	  /* VAR == FUNCT */
	  
	  if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (VAR != FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(l->arg->content.handler->var_name, num_macro);
	      	      
	    }
	  /* HASH_VAR == FUNCT */
	  else
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR != FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      return ne_ht_v;
	    }
	}
      /* SEGNAPOSTO == FUNCT */
      else
	{
	  if ( l->arg->content.handler->type_label != funct_ret_type(r->funct) )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL != FUNCT)\n");
	      exit(-1);
	    }
	  
	  
	  type= l->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return ne_str_str;
    break;
  case CHARACTER_VALUE: 
    return ne_chr_chr;
    break;
  case LONG_VALUE:
    return ne_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in ne_arg_funct (caso non considerato)\n");
    exit(-1);
    break;
  }
  
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ne_arg_funct\n");
  exit(-1);
}

cond_funct_ptr
ne_funct_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
 
  if (r->arg->type_content == SIMPLE_VALUE)
    {
      /* FUNCT == SIMPLE_VALUE */
      
      if ( funct_ret_type(l->funct) != r->arg->type_value )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT != SIMPLE_VALUE)\n");
	  exit(-1);
	}

      
      type=r->arg->type_value;
      
    }
  else /* (r->arg->type_content == HANDLER) */ 
    {
      if (r->arg->content.handler->var_name) 
	{
	  /* FUNCT == VAR */
	  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT != VAR)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(r->arg->content.handler->var_name, num_macro);
	      
	      
	    }
	  /* FUNCT == HASH_VAR */
	  else
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT != HASH_VAR)\n");
		  exit(-1);
		}
	      
	      
	      return ne_v_ht;
	    }
	}
      /* FUNCT == SEGNAPOSTO */
      else
	{
	  if ( funct_ret_type(l->funct) != r->arg->content.handler->type_label )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT != SEGNAPOSTO)\n");
	      exit(-1);
	    }
	  
	  
	  type= r->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return ne_str_str;
    break;
  case CHARACTER_VALUE: 
    return ne_chr_chr;
    break;
  case LONG_VALUE:
    return ne_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in ne_funct_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ne_funct_arg\n");
  exit(-1); 
}


cond_funct_ptr 
ne_funct_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if ( funct_ret_type(l->funct) != funct_ret_type(r->funct) )
    {
      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT != FUNCT)\n");
      exit(-1);
    }
  
  switch (funct_ret_type(l->funct)) {
  case STRING_VALUE:
    return ne_str_str;
    break;
  case CHARACTER_VALUE:
    return ne_chr_chr;
    break;
  case LONG_VALUE:
    return ne_long_long;
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ne_funct_funct\n");
  exit(-1); 
}

/***************************** Funzioni chiamate a runtime per != *****************************************/

/* queste funzioni ci sono per la compatibilita' con le funzioni di == */ 

int ne_str_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ne_s;
  res =  (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)())ne_str_str;
  return res;
}

int ne_chr_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ne_c;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct =(void *(*)()) ne_chr_chr;
  return res;
}

int ne_long_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ne_l;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ne_long_long;
  return res;
}

int ne_ht_v(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ne_hash_v;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ne_ht_v;
  return res;
}

int ne_v_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ne_v_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ne_v_ht;
  return res;
}

int ne_ht_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ne_hash_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ne_ht_ht;
  return res;
}

/* FINE queste funzioni ci sono per la compatibilita' con le funzioni di == */ 


int
ne_s(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != b);
  
  return (strcmp(a,b) != 0);
}

int
ne_c(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != b);
  
  return ( *(a) != *(b) );
}

int
ne_l(long *a, long *b)
{
  if (a == NULL || b == NULL) return (a != b);
  
  return ( *(a) != *(b) );
}

int
ne_hash_v(GHashTable *ht, void *b)
{
  if (ht == NULL || b == NULL) return (ht != b);
  
  return ( g_hash_table_lookup(ht, b) == NULL );
}

int
ne_v_hash(void *a, GHashTable *ht)
{
  if (a == NULL || ht == NULL) return (a != ht);
  
  return ( g_hash_table_lookup(ht, a) == NULL );
}

int
ne_hash_hash(GHashTable *a, GHashTable *b)
{
  GHashTable *temp = b;
  
  if (a == NULL || b == NULL) return (a != b);
  
  if ( g_hash_table_size(a) != g_hash_table_size(b) ) return 1;
  
  /* ATTENZIONE: e' inefficente perche' mi scorre tutti gli elementi nella hash table anche se il primo e' diverso */ 
  g_hash_table_foreach (a, confront_ht, &temp);
  
  return (temp == NULL);
}



/********************************************************************************************************************
 ********************************************************************************************************************
 ********************************************************************************************************************/


cond_funct_ptr 
gt(arg_node_t *left_op, arg_node_t *right_op, int num_macro)
{
  if (left_op->arg)
    {
      if (right_op->arg) return (*gt_arg_arg)(left_op, right_op, num_macro);
      else               return (*gt_arg_funct)(left_op, right_op, num_macro);
    }
  
  else
    {
      if (right_op->arg) return (*gt_funct_arg)(left_op, right_op, num_macro);
      else               return (*gt_funct_funct)(left_op, right_op, num_macro);
    }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in gt\n");
  exit(-1);
}

cond_funct_ptr 
gt_arg_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type=0;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE > SIMPLE_VALUE (chi la utilizza una cond. del genere ?) */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->type_value != r->arg->type_value) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE > SIMPLE_VALUE)\n");
	      exit(-1);
	    }
	  type = l->arg->type_value;
	  
	}
      
      /* SIMPLE_VALUE > HANDLER */
      if (r->arg->type_content == HANDLER)
	{
	  
	  if (r->arg->content.handler->var_name) 
	    {
	      /* SIMPLE_VALUE > VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro))
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE > VAR)\n");
		      exit(-1);
		    }
		  
		  type = l->arg->type_value;
		  
		}
	      /* SIMPLE_VALUE > HASH_VAR */
	      else
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE > HASH_VAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (SIMPLE_VALUE > HASH_VAR) not implemented\n");
		  exit(-1);
		  return gt_v_ht;
		  
		}

	    }
	  /* SIMPLE_VALUE > SEGNAPOSTO */
	  else
	    {
	      /* il valore del segnaposto viene preso a run-time ed in handler_t non ho l'effettivo valore */
	      if (l->arg->type_value !=
		  r->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE > LABEL)\n");
		  exit(-1);
		} 
	      
	      type = l->arg->type_value;
	      
	    }
	}
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      /* HANDLER > SIMPLE_VALUE */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->content.handler->var_name) 
	    {
	      /* VAR > SIMPLE_VALUE */
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR > SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  type=r->arg->type_value;
		  
		}
	      /* HASH_VAR > SIMPLE_VALUE */
	      else
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR > SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASH_VAR > SIMPLE_VALUE) not implemented\n");
		  exit(-1);
		  return gt_ht_v;
		  
		}
	    }
	  /* SEGNAPOSTO > SIMPLE_VALUE */
	  else
	    {
	      if (r->arg->type_value !=
		  l->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL > SIMPLE_VALUE)\n");
		  exit(-1);
		}
	      
	      type=r->arg->type_value;
	      
	    }
	}
      else /* HANDLER > HANDLER */
	{
	  if ((l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{ 
		  /* VAR > VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR > VAR)\n");
			  exit(-1);
			}
		      
		      type=get_type_values(l->arg->content.handler->var_name, num_macro);
		      
		      
		    }
		  /* VAR > HASHVAR */ 
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR > HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (VAR > HASHVAR) not implemented\n");
		      exit(-1);
		      return gt_v_ht;
		    }
		}
	      else 
		{
		  /* HASHVAR > VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar) 
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR > VAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR > VAR) not implemented\n");
		      exit(-1);
		      return gt_ht_v;
		    }
		    /* HASHVAR > HASHVAR */
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR > HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR > HASHVAR) not implemented\n");
		      exit(-1);
		      return gt_ht_ht;
		    }
		}
	      
	    }
	  else if ( !(l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* SEGNAPOSTO > VAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      /* printf("tipo var = %d\n", get_type_values(r->arg->content.handler->var_name, num_macro));
			 printf("tipo sysc = %d\n", l->arg->content.handler->type_label);
			 printf("num arg = %ld\n", l->arg->content.handler->label_ref.n_arg + 1); */
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL > VAR)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(r->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* SEGNAPOSTO > HASHVAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL > HASHVAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (LABEL > HASHVAR) not implemented\n");
		  exit(-1);
		  return gt_v_ht;
		}
	    }
	  else if ( (l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* VAR > SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR > LABEL)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(l->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* HASHVAR > SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR > LABEL)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASHVAR > LABEL) not implemented\n");
		  exit(-1);
		  return gt_ht_v;
		}
	    }
	  else if ( !(l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      /* SEGNAPOSTO > SEGNAPOSTO */

	      if ( l->arg->content.handler->type_label != 
		   r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL > LABEL)\n");
		  exit(-1);
		}
	      
	      type= l->arg->content.handler->type_label;
	      
	      
	    }
	}
      
    }
  
  switch (type) {
  case STRING_VALUE:
    return gt_str_str;
    break;
  case CHARACTER_VALUE: 
    return gt_chr_chr;
    break;
  case LONG_VALUE:
    return gt_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in gt_arg_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in gt_arg_arg (caso non considerato)\n");
  exit(-1); 
}

/****************************************************************************/

cond_funct_ptr 
gt_arg_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE > FUNCT */
      if ( l->arg->type_value != funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE > FUNCT)\n");
	  exit(-1);
	}
      
            
      type=l->arg->type_value;
      
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      if (l->arg->content.handler->var_name) 
	{
	  /* VAR > FUNCT */
	  
	  if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (VAR > FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(l->arg->content.handler->var_name, num_macro);
	      	      
	    }
	  /* HASH_VAR > FUNCT */
	  else
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR > FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (HASH_VAR > FUNCT) not implemented\n");
	      exit(-1);
	      return gt_ht_v;
	    }
	}
      /* SEGNAPOSTO > FUNCT */
      else
	{
	  if ( l->arg->content.handler->type_label != funct_ret_type(r->funct) )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL > FUNCT)\n");
	      exit(-1);
	    }
	  
	  
	  type= l->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return gt_str_str;
    break;
  case CHARACTER_VALUE: 
    return gt_chr_chr;
    break;
  case LONG_VALUE:
    return gt_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in gt_arg_funct (caso non considerato)\n");
    exit(-1);
    break;
  }
  
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in gt_arg_funct\n");
  exit(-1);
}

cond_funct_ptr
gt_funct_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
 
  if (r->arg->type_content == SIMPLE_VALUE)
    {
      /* FUNCT > SIMPLE_VALUE */
      
      if ( funct_ret_type(l->funct) != r->arg->type_value )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT > SIMPLE_VALUE)\n");
	  exit(-1);
	}

      
      type=r->arg->type_value;
      
    }
  else /* (r->arg->type_content == HANDLER) */ 
    {
      if (r->arg->content.handler->var_name) 
	{
	  /* FUNCT > VAR */
	  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT > VAR)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(r->arg->content.handler->var_name, num_macro);
	      
	      
	    }
	  /* FUNCT > HASH_VAR */
	  else
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT > HASH_VAR)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (FUNCT > HASH_VAR) not implemented\n");
	      exit(-1);
	      return gt_v_ht;
	    }
	}
      /* FUNCT > SEGNAPOSTO */
      else
	{
	  if ( funct_ret_type(l->funct) != r->arg->content.handler->type_label )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT > SEGNAPOSTO)\n");
	      exit(-1);
	    }
	  
	  
	  type= r->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return gt_str_str;
    break;
  case CHARACTER_VALUE: 
    return gt_chr_chr;
    break;
  case LONG_VALUE:
    return gt_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in gt_funct_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in gt_funct_arg\n");
  exit(-1); 
}


cond_funct_ptr 
gt_funct_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if ( funct_ret_type(l->funct) != funct_ret_type(r->funct) )
    {
      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT > FUNCT)\n");
      exit(-1);
    }
  
  switch (funct_ret_type(l->funct)) {
  case STRING_VALUE:
    return gt_str_str;
    break;
  case CHARACTER_VALUE:
    return gt_chr_chr;
    break;
  case LONG_VALUE:
    return gt_long_long;
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in gt_funct_funct\n");
  exit(-1); 
}

/***************************** Funzioni chiamate a runtime per > *****************************************/

/* queste funzioni ci sono per la compatibilita' con le funzioni di == */ 

int gt_str_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) gt_s;
  res =  (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)())gt_str_str;
  return res;
}

int gt_chr_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) gt_c;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct =(void *(*)()) gt_chr_chr;
  return res;
}

int gt_long_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) gt_l;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) gt_long_long;
  return res;
}

int gt_ht_v(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) gt_hash_v;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) gt_ht_v;
  return res;
}

int gt_v_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) gt_v_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) gt_v_ht;
  return res;
}

int gt_ht_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) gt_hash_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) gt_ht_ht;
  return res;
}

/* FINE queste funzioni ci sono per la compatibilita' con le funzioni di == */ 


int
gt_s(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return (strcmp(a,b) > 0);
}

int
gt_c(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) > *(b) );
}

int
gt_l(long *a, long *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) > *(b) );
}

int
gt_hash_v(GHashTable *ht, void *b)
{
  if (ht == NULL || b == NULL) return (ht != NULL);
  /* tutti gli elementi di ht devono essere maggiori di b */
  return ( g_hash_table_lookup(ht, b) == NULL );
}

int
gt_v_hash(void *a, GHashTable *ht)
{
  if (a == NULL || ht == NULL) return (a != ht);
  /* a deve essere maggiore di tutti gli elementi di ht */
  return ( g_hash_table_lookup(ht, a) == NULL );
}

int
gt_hash_hash(GHashTable *a, GHashTable *b)
{
  GHashTable *temp = b;
  
  if (a == NULL || b == NULL) return (a != b);
  /* tutti gli elementi di a devono essere maggiori di b */
  if ( g_hash_table_size(a) != g_hash_table_size(b) ) return 1;
  
  /* ATTENZIONE: e' inefficente perche' mi scorre tutti gli elementi nella hash table anche se il primo e' diverso */ 
  g_hash_table_foreach (a, confront_ht, &temp);
  
  return (temp == NULL);
}




/********************************************************************************************************************
 ********************************************************************************************************************
 ********************************************************************************************************************/


cond_funct_ptr 
ge(arg_node_t *left_op, arg_node_t *right_op, int num_macro)
{
  if (left_op->arg)
    {
      if (right_op->arg) return (*ge_arg_arg)(left_op, right_op, num_macro);
      else               return (*ge_arg_funct)(left_op, right_op, num_macro);
    }
  
  else
    {
      if (right_op->arg) return (*ge_funct_arg)(left_op, right_op, num_macro);
      else               return (*ge_funct_funct)(left_op, right_op, num_macro);
    }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ge\n");
  exit(-1);
}

cond_funct_ptr 
ge_arg_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type=0;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE >= SIMPLE_VALUE (chi la utilizza una cond. del genere ?) */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->type_value != r->arg->type_value) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE >= SIMPLE_VALUE)\n");
	      exit(-1);
	    }
	  type = l->arg->type_value;
	  
	}
      
      /* SIMPLE_VALUE >= HANDLER */
      if (r->arg->type_content == HANDLER)
	{
	  
	  if (r->arg->content.handler->var_name) 
	    {
	      /* SIMPLE_VALUE >= VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro))
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE >= VAR)\n");
		      exit(-1);
		    }
		  
		  type = l->arg->type_value;
		  
		}
	      /* SIMPLE_VALUE >= HASH_VAR */
	      else
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE >= HASH_VAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (SIMPLE_VALUE >= HASH_VAR) not implemented\n");
		  exit(-1);
		  return ge_v_ht;
		  
		}

	    }
	  /* SIMPLE_VALUE >= SEGNAPOSTO */
	  else
	    {
	      /* il valore del segnaposto viene preso a run-time ed in handler_t non ho l'effettivo valore */
	      if (l->arg->type_value !=
		  r->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE >= LABEL)\n");
		  exit(-1);
		} 
	      
	      type = l->arg->type_value;
	      
	    }
	}
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      /* HANDLER >= SIMPLE_VALUE */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->content.handler->var_name) 
	    {
	      /* VAR >= SIMPLE_VALUE */
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR >= SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  type=r->arg->type_value;
		  
		}
	      /* HASH_VAR >= SIMPLE_VALUE */
	      else
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR >= SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASH_VAR >= SIMPLE_VALUE) not implemented\n");
		  exit(-1);
		  return ge_ht_v;
		  
		}
	    }
	  /* SEGNAPOSTO >= SIMPLE_VALUE */
	  else
	    {
	      if (r->arg->type_value !=
		  l->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL >= SIMPLE_VALUE)\n");
		  exit(-1);
		}
	      
	      type=r->arg->type_value;
	      
	    }
	}
      else /* HANDLER >= HANDLER */
	{
	  if ((l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{ 
		  /* VAR >= VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR >= VAR)\n");
			  exit(-1);
			}
		      
		      type=get_type_values(l->arg->content.handler->var_name, num_macro);
		      
		      
		    }
		  /* VAR >= HASHVAR */ 
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR >= HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (VAR >= HASHVAR) not implemented\n");
		      exit(-1);
		      return ge_v_ht;
		    }
		}
	      else 
		{
		  /* HASHVAR >= VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar) 
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR >= VAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR >= VAR) not implemented\n");
		      exit(-1);
		      return ge_ht_v;
		    }
		    /* HASHVAR >= HASHVAR */
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR >= HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR >= HASHVAR) not implemented\n");
		      exit(-1);
		      return ge_ht_ht;
		    }
		}
	      
	    }
	  else if ( !(l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* SEGNAPOSTO >= VAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      /* printf("tipo var = %d\n", get_type_values(r->arg->content.handler->var_name, num_macro));
			 printf("tipo sysc = %d\n", l->arg->content.handler->type_label);
			 printf("num arg = %ld\n", l->arg->content.handler->label_ref.n_arg + 1); */
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL >= VAR)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(r->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* SEGNAPOSTO >= HASHVAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL >= HASHVAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (LABEL >= HASHVAR) not implemented\n");
		  exit(-1);
		  return ge_v_ht;
		}
	    }
	  else if ( (l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* VAR >= SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR >= LABEL)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(l->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* HASHVAR >= SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR >= LABEL)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASHVAR >= LABEL) not implemented\n");
		  exit(-1);
		  return ge_ht_v;
		}
	    }
	  else if ( !(l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      /* SEGNAPOSTO >= SEGNAPOSTO */

	      if ( l->arg->content.handler->type_label != 
		   r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL >= LABEL)\n");
		  exit(-1);
		}
	      
	      type= l->arg->content.handler->type_label;
	      
	      
	    }
	}
      
    }
  
  switch (type) {
  case STRING_VALUE:
    return ge_str_str;
    break;
  case CHARACTER_VALUE: 
    return ge_chr_chr;
    break;
  case LONG_VALUE:
    return ge_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in ge_arg_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ge_arg_arg (caso non considerato)\n");
  exit(-1); 
}

/****************************************************************************/

cond_funct_ptr 
ge_arg_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE >= FUNCT */
      if ( l->arg->type_value != funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE >= FUNCT)\n");
	  exit(-1);
	}
      
            
      type=l->arg->type_value;
      
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      if (l->arg->content.handler->var_name) 
	{
	  /* VAR >= FUNCT */
	  
	  if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (VAR >= FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(l->arg->content.handler->var_name, num_macro);
	      	      
	    }
	  /* HASH_VAR >= FUNCT */
	  else
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR >= FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (HASH_VAR >= FUNCT) not implemented\n");
	      exit(-1);
	      return ge_ht_v;
	    }
	}
      /* SEGNAPOSTO >= FUNCT */
      else
	{
	  if ( l->arg->content.handler->type_label != funct_ret_type(r->funct) )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL >= FUNCT)\n");
	      exit(-1);
	    }
	  
	  
	  type= l->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return ge_str_str;
    break;
  case CHARACTER_VALUE: 
    return ge_chr_chr;
    break;
  case LONG_VALUE:
    return ge_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in ge_arg_funct (caso non considerato)\n");
    exit(-1);
    break;
  }
  
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ge_arg_funct\n");
  exit(-1);
}

cond_funct_ptr
ge_funct_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
 
  if (r->arg->type_content == SIMPLE_VALUE)
    {
      /* FUNCT >= SIMPLE_VALUE */
      
      if ( funct_ret_type(l->funct) != r->arg->type_value )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT >= SIMPLE_VALUE)\n");
	  exit(-1);
	}

      
      type=r->arg->type_value;
      
    }
  else /* (r->arg->type_content == HANDLER) */ 
    {
      if (r->arg->content.handler->var_name) 
	{
	  /* FUNCT >= VAR */
	  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT >= VAR)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(r->arg->content.handler->var_name, num_macro);
	      
	      
	    }
	  /* FUNCT >= HASH_VAR */
	  else
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT >= HASH_VAR)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (FUNCT >= HASH_VAR) not implemented\n");
	      exit(-1);
	      return ge_v_ht;
	    }
	}
      /* FUNCT >= SEGNAPOSTO */
      else
	{
	  if ( funct_ret_type(l->funct) != r->arg->content.handler->type_label )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT >= SEGNAPOSTO)\n");
	      exit(-1);
	    }
	  
	  
	  type= r->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return ge_str_str;
    break;
  case CHARACTER_VALUE: 
    return ge_chr_chr;
    break;
  case LONG_VALUE:
    return ge_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in ge_funct_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ge_funct_arg\n");
  exit(-1); 
}


cond_funct_ptr 
ge_funct_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if ( funct_ret_type(l->funct) != funct_ret_type(r->funct) )
    {
      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT >= FUNCT)\n");
      exit(-1);
    }
  
  switch (funct_ret_type(l->funct)) {
  case STRING_VALUE:
    return ge_str_str;
    break;
  case CHARACTER_VALUE:
    return ge_chr_chr;
    break;
  case LONG_VALUE:
    return ge_long_long;
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in ge_funct_funct\n");
  exit(-1); 
}

/***************************** Funzioni chiamate a runtime per >= *****************************************/

/* queste funzioni ci sono per la compatibilita' con le funzioni di == */ 

int ge_str_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ge_s;
  res =  (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)())ge_str_str;
  return res;
}

int ge_chr_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ge_c;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct =(void *(*)()) ge_chr_chr;
  return res;
}

int ge_long_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ge_l;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ge_long_long;
  return res;
}

int ge_ht_v(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ge_hash_v;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ge_ht_v;
  return res;
}

int ge_v_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ge_v_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ge_v_ht;
  return res;
}

int ge_ht_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) ge_hash_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) ge_ht_ht;
  return res;
}

/* FINE queste funzioni ci sono per la compatibilita' con le funzioni di == */ 


int
ge_s(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return (strcmp(a,b) >= 0);
}

int
ge_c(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) >= *(b) );
}

int
ge_l(long *a, long *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) >= *(b) );
}

int
ge_hash_v(GHashTable *ht, void *b)
{
  if (ht == NULL || b == NULL) return (ht != NULL);
  /* tutti gli elementi di ht devono essere maggiori di b */
  return ( g_hash_table_lookup(ht, b) == NULL );
}

int
ge_v_hash(void *a, GHashTable *ht)
{
  if (a == NULL || ht == NULL) return (a != ht);
  /* a deve essere maggiore di tutti gli elementi di ht */
  return ( g_hash_table_lookup(ht, a) == NULL );
}

int
ge_hash_hash(GHashTable *a, GHashTable *b)
{
  GHashTable *temp = b;
  
  if (a == NULL || b == NULL) return (a != b);
  /* tutti gli elementi di a devono essere maggiori di b */
  if ( g_hash_table_size(a) != g_hash_table_size(b) ) return 1;
  
  /* ATTENZIONE: e' inefficente perche' mi scorre tutti gli elementi nella hash table anche se il primo e' diverso */ 
  g_hash_table_foreach (a, confront_ht, &temp);
  
  return (temp == NULL);
}




/********************************************************************************************************************
 ********************************************************************************************************************
 ********************************************************************************************************************/


cond_funct_ptr 
lt(arg_node_t *left_op, arg_node_t *right_op, int num_macro)
{
  if (left_op->arg)
    {
      if (right_op->arg) return (*lt_arg_arg)(left_op, right_op, num_macro);
      else               return (*lt_arg_funct)(left_op, right_op, num_macro);
    }
  
  else
    {
      if (right_op->arg) return (*lt_funct_arg)(left_op, right_op, num_macro);
      else               return (*lt_funct_funct)(left_op, right_op, num_macro);
    }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in lt\n");
  exit(-1);
}

cond_funct_ptr 
lt_arg_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type=0;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE < SIMPLE_VALUE (chi la utilizza una cond. del genere ?) */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->type_value != r->arg->type_value) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE < SIMPLE_VALUE)\n");
	      exit(-1);
	    }
	  type = l->arg->type_value;
	  
	}
      
      /* SIMPLE_VALUE < HANDLER */
      if (r->arg->type_content == HANDLER)
	{
	  
	  if (r->arg->content.handler->var_name) 
	    {
	      /* SIMPLE_VALUE < VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro))
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE < VAR)\n");
		      exit(-1);
		    }
		  
		  type = l->arg->type_value;
		  
		}
	      /* SIMPLE_VALUE < HASH_VAR */
	      else
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE < HASH_VAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (SIMPLE_VALUE < HASH_VAR) not implemented\n");
		  exit(-1);
		  return lt_v_ht;
		  
		}

	    }
	  /* SIMPLE_VALUE < SEGNAPOSTO */
	  else
	    {
	      /* il valore del segnaposto viene preso a run-time ed in handler_t non ho l'effettivo valore */
	      if (l->arg->type_value !=
		  r->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE < LABEL)\n");
		  exit(-1);
		} 
	      
	      type = l->arg->type_value;
	      
	    }
	}
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      /* HANDLER < SIMPLE_VALUE */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->content.handler->var_name) 
	    {
	      /* VAR < SIMPLE_VALUE */
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR < SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  type=r->arg->type_value;
		  
		}
	      /* HASH_VAR < SIMPLE_VALUE */
	      else
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR < SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASH_VAR < SIMPLE_VALUE) not implemented\n");
		  exit(-1);
		  return lt_ht_v;
		  
		}
	    }
	  /* SEGNAPOSTO < SIMPLE_VALUE */
	  else
	    {
	      if (r->arg->type_value !=
		  l->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL < SIMPLE_VALUE)\n");
		  exit(-1);
		}
	      
	      type=r->arg->type_value;
	      
	    }
	}
      else /* HANDLER < HANDLER */
	{
	  if ((l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{ 
		  /* VAR < VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR < VAR)\n");
			  exit(-1);
			}
		      
		      type=get_type_values(l->arg->content.handler->var_name, num_macro);
		      
		      
		    }
		  /* VAR < HASHVAR */ 
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR < HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (VAR < HASHVAR) not implemented\n");
		      exit(-1);
		      return lt_v_ht;
		    }
		}
	      else 
		{
		  /* HASHVAR < VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar) 
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR < VAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR < VAR) not implemented\n");
		      exit(-1);
		      return lt_ht_v;
		    }
		    /* HASHVAR < HASHVAR */
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR < HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR < HASHVAR) not implemented\n");
		      exit(-1);
		      return lt_ht_ht;
		    }
		}
	      
	    }
	  else if ( !(l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* SEGNAPOSTO < VAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      /* printf("tipo var = %d\n", get_type_values(r->arg->content.handler->var_name, num_macro));
			 printf("tipo sysc = %d\n", l->arg->content.handler->type_label);
			 printf("num arg = %ld\n", l->arg->content.handler->label_ref.n_arg + 1); */
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL < VAR)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(r->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* SEGNAPOSTO < HASHVAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL < HASHVAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (LABEL < HASHVAR) not implemented\n");
		  exit(-1);
		  return lt_v_ht;
		}
	    }
	  else if ( (l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* VAR < SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR < LABEL)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(l->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* HASHVAR < SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR < LABEL)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASHVAR < LABEL) not implemented\n");
		  exit(-1);
		  return lt_ht_v;
		}
	    }
	  else if ( !(l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      /* SEGNAPOSTO < SEGNAPOSTO */

	      if ( l->arg->content.handler->type_label != 
		   r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL < LABEL)\n");
		  exit(-1);
		}
	      
	      type= l->arg->content.handler->type_label;
	      
	      
	    }
	}
      
    }
  
  switch (type) {
  case STRING_VALUE:
    return lt_str_str;
    break;
  case CHARACTER_VALUE: 
    return lt_chr_chr;
    break;
  case LONG_VALUE:
    return lt_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in lt_arg_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in lt_arg_arg (caso non considerato)\n");
  exit(-1); 
}

/****************************************************************************/

cond_funct_ptr 
lt_arg_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE < FUNCT */
      if ( l->arg->type_value != funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE < FUNCT)\n");
	  exit(-1);
	}
      
            
      type=l->arg->type_value;
      
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      if (l->arg->content.handler->var_name) 
	{
	  /* VAR < FUNCT */
	  
	  if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (VAR < FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(l->arg->content.handler->var_name, num_macro);
	      	      
	    }
	  /* HASH_VAR < FUNCT */
	  else
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR < FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (HASH_VAR < FUNCT) not implemented\n");
	      exit(-1);
	      return lt_ht_v;
	    }
	}
      /* SEGNAPOSTO < FUNCT */
      else
	{
	  if ( l->arg->content.handler->type_label != funct_ret_type(r->funct) )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL < FUNCT)\n");
	      exit(-1);
	    }
	  
	  
	  type= l->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return lt_str_str;
    break;
  case CHARACTER_VALUE: 
    return lt_chr_chr;
    break;
  case LONG_VALUE:
    return lt_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in lt_arg_funct (caso non considerato)\n");
    exit(-1);
    break;
  }
  
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in lt_arg_funct\n");
  exit(-1);
}

cond_funct_ptr
lt_funct_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
 
  if (r->arg->type_content == SIMPLE_VALUE)
    {
      /* FUNCT < SIMPLE_VALUE */
      
      if ( funct_ret_type(l->funct) != r->arg->type_value )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT < SIMPLE_VALUE)\n");
	  exit(-1);
	}

      
      type=r->arg->type_value;
      
    }
  else /* (r->arg->type_content == HANDLER) */ 
    {
      if (r->arg->content.handler->var_name) 
	{
	  /* FUNCT < VAR */
	  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT < VAR)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(r->arg->content.handler->var_name, num_macro);
	      
	      
	    }
	  /* FUNCT < HASH_VAR */
	  else
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT < HASH_VAR)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (FUNCT < HASH_VAR) not implemented\n");
	      exit(-1);
	      return lt_v_ht;
	    }
	}
      /* FUNCT < SEGNAPOSTO */
      else
	{
	  if ( funct_ret_type(l->funct) != r->arg->content.handler->type_label )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT < SEGNAPOSTO)\n");
	      exit(-1);
	    }
	  
	  
	  type= r->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return lt_str_str;
    break;
  case CHARACTER_VALUE: 
    return lt_chr_chr;
    break;
  case LONG_VALUE:
    return lt_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in lt_funct_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in lt_funct_arg\n");
  exit(-1); 
}


cond_funct_ptr 
lt_funct_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if ( funct_ret_type(l->funct) != funct_ret_type(r->funct) )
    {
      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT < FUNCT)\n");
      exit(-1);
    }
  
  switch (funct_ret_type(l->funct)) {
  case STRING_VALUE:
    return lt_str_str;
    break;
  case CHARACTER_VALUE:
    return lt_chr_chr;
    break;
  case LONG_VALUE:
    return lt_long_long;
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in lt_funct_funct\n");
  exit(-1); 
}

/***************************** Funzioni chiamate a runtime per < *****************************************/

/* queste funzioni ci sono per la compatibilita' con le funzioni di == */ 

int lt_str_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) lt_s;
  res =  (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)())lt_str_str;
  return res;
}

int lt_chr_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) lt_c;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct =(void *(*)()) lt_chr_chr;
  return res;
}

int lt_long_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) lt_l;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) lt_long_long;
  return res;
}

int lt_ht_v(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) lt_hash_v;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) lt_ht_v;
  return res;
}

int lt_v_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) lt_v_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) lt_v_ht;
  return res;
}

int lt_ht_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) lt_hash_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) lt_ht_ht;
  return res;
}

/* FINE queste funzioni ci sono per la compatibilita' con le funzioni di == */ 


int
lt_s(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return (strcmp(a,b) < 0);
}

int
lt_c(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) < *(b) );
}

int
lt_l(long *a, long *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) < *(b) );
}

int
lt_hash_v(GHashTable *ht, void *b)
{
  if (ht == NULL || b == NULL) return (ht != NULL);
  /* tutti gli elementi di ht devono essere maggiori di b */
  return ( g_hash_table_lookup(ht, b) == NULL );
}

int
lt_v_hash(void *a, GHashTable *ht)
{
  if (a == NULL || ht == NULL) return (a != ht);
  /* a deve essere maggiore di tutti gli elementi di ht */
  return ( g_hash_table_lookup(ht, a) == NULL );
}

int
lt_hash_hash(GHashTable *a, GHashTable *b)
{
  GHashTable *temp = b;
  
  if (a == NULL || b == NULL) return (a != b);
  /* tutti gli elementi di a devono essere maggiori di b */
  if ( g_hash_table_size(a) != g_hash_table_size(b) ) return 1;
  
  /* ATTENZIONE: e' inefficente perche' mi scorre tutti gli elementi nella hash table anche se il primo e' diverso */ 
  g_hash_table_foreach (a, confront_ht, &temp);
  
  return (temp == NULL);
}




/********************************************************************************************************************
 ********************************************************************************************************************
 ********************************************************************************************************************/


cond_funct_ptr 
le(arg_node_t *left_op, arg_node_t *right_op, int num_macro)
{
  if (left_op->arg)
    {
      if (right_op->arg) return (*le_arg_arg)(left_op, right_op, num_macro);
      else               return (*le_arg_funct)(left_op, right_op, num_macro);
    }
  
  else
    {
      if (right_op->arg) return (*le_funct_arg)(left_op, right_op, num_macro);
      else               return (*le_funct_funct)(left_op, right_op, num_macro);
    }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in le\n");
  exit(-1);
}

cond_funct_ptr 
le_arg_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type=0;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE <= SIMPLE_VALUE (chi la utilizza una cond. del genere ?) */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->type_value != r->arg->type_value) 
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE <= SIMPLE_VALUE)\n");
	      exit(-1);
	    }
	  type = l->arg->type_value;
	  
	}
      
      /* SIMPLE_VALUE <= HANDLER */
      if (r->arg->type_content == HANDLER)
	{
	  
	  if (r->arg->content.handler->var_name) 
	    {
	      /* SIMPLE_VALUE <= VAR */
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro))
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE <= VAR)\n");
		      exit(-1);
		    }
		  
		  type = l->arg->type_value;
		  
		}
	      /* SIMPLE_VALUE <= HASH_VAR */
	      else
		{
		  if (l->arg->type_value != get_type_values(r->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE <= HASH_VAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (SIMPLE_VALUE <= HASH_VAR) not implemented\n");
		  exit(-1);
		  return le_v_ht;
		  
		}

	    }
	  /* SIMPLE_VALUE <= SEGNAPOSTO */
	  else
	    {
	      /* il valore del segnaposto viene preso a run-time ed in handler_t non ho l'effettivo valore */
	      if (l->arg->type_value !=
		  r->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE <= LABEL)\n");
		  exit(-1);
		} 
	      
	      type = l->arg->type_value;
	      
	    }
	}
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      /* HANDLER <= SIMPLE_VALUE */
      if (r->arg->type_content == SIMPLE_VALUE)
	{
	  if (l->arg->content.handler->var_name) 
	    {
	      /* VAR <= SIMPLE_VALUE */
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR <= SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  
		  type=r->arg->type_value;
		  
		}
	      /* HASH_VAR <= SIMPLE_VALUE */
	      else
		{
		  if (r->arg->type_value != get_type_values(l->arg->content.handler->var_name, num_macro)) 
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR <= SIMPLE_VALUE)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASH_VAR <= SIMPLE_VALUE) not implemented\n");
		  exit(-1);
		  return le_ht_v;
		  
		}
	    }
	  /* SEGNAPOSTO <= SIMPLE_VALUE */
	  else
	    {
	      if (r->arg->type_value !=
		  l->arg->content.handler->type_label)
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL <= SIMPLE_VALUE)\n");
		  exit(-1);
		}
	      
	      type=r->arg->type_value;
	      
	    }
	}
      else /* HANDLER <= HANDLER */
	{
	  if ((l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{ 
		  /* VAR <= VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR <= VAR)\n");
			  exit(-1);
			}
		      
		      type=get_type_values(l->arg->content.handler->var_name, num_macro);
		      
		      
		    }
		  /* VAR <= HASHVAR */ 
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (VAR <= HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (VAR <= HASHVAR) not implemented\n");
		      exit(-1);
		      return le_v_ht;
		    }
		}
	      else 
		{
		  /* HASHVAR <= VAR */
		  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar) 
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR <= VAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR <= VAR) not implemented\n");
		      exit(-1);
		      return le_ht_v;
		    }
		    /* HASHVAR <= HASHVAR */
		  else
		    {
		      if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
			  get_type_values(r->arg->content.handler->var_name, num_macro)) 
			{
			  fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR <= HASHVAR)\n");
			  exit(-1);
			}
		      fprintf(stderr,"\nERROR: (HASHVAR <= HASHVAR) not implemented\n");
		      exit(-1);
		      return le_ht_ht;
		    }
		}
	      
	    }
	  else if ( !(l->arg->content.handler->var_name) && (r->arg->content.handler->var_name))
	    {
	      if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* SEGNAPOSTO <= VAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      /* fprintf(stderr,"tipo var = %d\n", get_type_values(r->arg->content.handler->var_name, num_macro));
			 printf("tipo sysc = %d\n", l->arg->content.handler->type_label);
			 printf("num arg = %ld\n", l->arg->content.handler->label_ref.n_arg + 1); */
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL <= VAR)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(r->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* SEGNAPOSTO <= HASHVAR */

		  if (get_type_values(r->arg->content.handler->var_name, num_macro) != 
		      l->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL <= HASHVAR)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (LABEL <= HASHVAR) not implemented\n");
		  exit(-1);
		  return le_v_ht;
		}
	    }
	  else if ( (l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
		{
		  /* VAR <= SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (VAR <= LABEL)\n");
		      exit(-1);
		    }
		  
		  type=get_type_values(l->arg->content.handler->var_name, num_macro);
		  
		  
		}
	      else
		{
		  /* HASHVAR <= SEGNAPOSTO */

		  if (get_type_values(l->arg->content.handler->var_name, num_macro) != 
		      r->arg->content.handler->type_label)
		    {
		      fprintf(stderr,"\nERROR: type mismatch in condition (HASHVAR <= LABEL)\n");
		      exit(-1);
		    }
		  fprintf(stderr,"\nERROR: (HASHVAR <= LABEL) not implemented\n");
		  exit(-1);
		  return le_ht_v;
		}
	    }
	  else if ( !(l->arg->content.handler->var_name) && !(r->arg->content.handler->var_name) )
	    {
	      /* SEGNAPOSTO <= SEGNAPOSTO */

	      if ( l->arg->content.handler->type_label != 
		   r->arg->content.handler->type_label )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (LABEL <= LABEL)\n");
		  exit(-1);
		}
	      
	      type= l->arg->content.handler->type_label;
	      
	      
	    }
	}
      
    }
  
  switch (type) {
  case STRING_VALUE:
    return le_str_str;
    break;
  case CHARACTER_VALUE: 
    return le_chr_chr;
    break;
  case LONG_VALUE:
    return le_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in le_arg_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in le_arg_arg (caso non considerato)\n");
  exit(-1); 
}

/****************************************************************************/

cond_funct_ptr 
le_arg_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
  
  if (l->arg->type_content == SIMPLE_VALUE)
    {
      /* SIMPLE_VALUE <= FUNCT */
      if ( l->arg->type_value != funct_ret_type(r->funct) )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (SIMPLE_VALUE <= FUNCT)\n");
	  exit(-1);
	}
      
            
      type=l->arg->type_value;
      
    }
  else /* (l->arg->type_content == HANDLER) */ 
    {
      if (l->arg->content.handler->var_name) 
	{
	  /* VAR <= FUNCT */
	  
	  if (get_type_var(l->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (VAR <= FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(l->arg->content.handler->var_name, num_macro);
	      	      
	    }
	  /* HASH_VAR <= FUNCT */
	  else
	    {
	      if ( get_type_values(l->arg->content.handler->var_name, num_macro) != funct_ret_type(r->funct) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (HASH_VAR <= FUNCT)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (HASH_VAR <= FUNCT) not implemented\n");
	      exit(-1);
	      return le_ht_v;
	    }
	}
      /* SEGNAPOSTO <= FUNCT */
      else
	{
	  if ( l->arg->content.handler->type_label != funct_ret_type(r->funct) )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (LABEL <= FUNCT)\n");
	      exit(-1);
	    }
	  
	  
	  type= l->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return le_str_str;
    break;
  case CHARACTER_VALUE: 
    return le_chr_chr;
    break;
  case LONG_VALUE:
    return le_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in le_arg_funct (caso non considerato)\n");
    exit(-1);
    break;
  }
  
  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in le_arg_funct\n");
  exit(-1);
}

cond_funct_ptr
le_funct_arg(arg_node_t *l, arg_node_t *r, int num_macro)
{
  int type;
 
  if (r->arg->type_content == SIMPLE_VALUE)
    {
      /* FUNCT <= SIMPLE_VALUE */
      
      if ( funct_ret_type(l->funct) != r->arg->type_value )
	{
	  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT <= SIMPLE_VALUE)\n");
	  exit(-1);
	}

      
      type=r->arg->type_value;
      
    }
  else /* (r->arg->type_content == HANDLER) */ 
    {
      if (r->arg->content.handler->var_name) 
	{
	  /* FUNCT <= VAR */
	  if (get_type_var(r->arg->content.handler->var_name, num_macro) == scalar)
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT <= VAR)\n");
		  exit(-1);
		}
	      
	      
	      type=get_type_values(r->arg->content.handler->var_name, num_macro);
	      
	      
	    }
	  /* FUNCT <= HASH_VAR */
	  else
	    {
	      if ( funct_ret_type(l->funct) != get_type_values(r->arg->content.handler->var_name, num_macro) )
		{
		  fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT <= HASH_VAR)\n");
		  exit(-1);
		}
	      
	      
	      fprintf(stderr,"\nERROR: (FUNCT <= HASH_VAR) not implemented\n");
	      exit(-1);
	      return le_v_ht;
	    }
	}
      /* FUNCT <= SEGNAPOSTO */
      else
	{
	  if ( funct_ret_type(l->funct) != r->arg->content.handler->type_label )
	    {
	      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT <= SEGNAPOSTO)\n");
	      exit(-1);
	    }
	  
	  
	  type= r->arg->content.handler->type_label;
	  	  
	}
    }
  
  switch (type) {
  case STRING_VALUE:
    return le_str_str;
    break;
  case CHARACTER_VALUE: 
    return le_chr_chr;
    break;
  case LONG_VALUE:
    return le_long_long;
    break;
  default:
    fprintf(stderr,"ERRORE in le_funct_arg (caso non considerato)\n");
    exit(-1);
    break;
  }

  
  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in le_funct_arg\n");
  exit(-1); 
}


cond_funct_ptr 
le_funct_funct(arg_node_t *l, arg_node_t *r, int num_macro)
{
  
  if ( funct_ret_type(l->funct) != funct_ret_type(r->funct) )
    {
      fprintf(stderr,"\nERROR: type mismatch in condition (FUNCT <= FUNCT)\n");
      exit(-1);
    }
  
  switch (funct_ret_type(l->funct)) {
  case STRING_VALUE:
    return le_str_str;
    break;
  case CHARACTER_VALUE:
    return le_chr_chr;
    break;
  case LONG_VALUE:
    return le_long_long;
    break;
  }

  /* non ci dovrebbe mai arrivare */ 
  fprintf(stderr,"ERRORE in le_funct_funct\n");
  exit(-1); 
}

/***************************** Funzioni chiamate a runtime per <= *****************************************/

/* queste funzioni ci sono per la compatibilita' con le funzioni di == */ 

int le_str_str(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) le_s;
  res =  (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)())le_str_str;
  return res;
}

int le_chr_chr(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) le_c;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct =(void *(*)()) le_chr_chr;
  return res;
}

int le_long_long(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) le_l;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) le_long_long;
  return res;
}

int le_ht_v(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) le_hash_v;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) le_ht_v;
  return res;
}

int le_v_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) le_v_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) le_v_ht;
  return res;
}

int le_ht_ht(GNode *funct_tree, GHashTable *ht, struct tcb *tcp)
{
  int res;  
  arg_node_t *f = funct_tree->data;
  
  f->funct = (void *(*)()) le_hash_hash;
  res = (int) eval_funct(funct_tree, ht, tcp);
  if (func_not_eval)
    return NOEV;

  f->funct = (void *(*)()) le_ht_ht;
  return res;
}

/* FINE queste funzioni ci sono per la compatibilita' con le funzioni di == */ 


int
le_s(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return (strcmp(a,b) <= 0);
}

int
le_c(char *a, char *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) <= *(b) );
}

int
le_l(long *a, long *b)
{
  if (a == NULL || b == NULL) return (a != NULL);
  
  return ( *(a) <= *(b) );
}

int
le_hash_v(GHashTable *ht, void *b)
{
  if (ht == NULL || b == NULL) return (ht != NULL);
  /* tutti gli elementi di ht devono essere maggiori di b */
  return ( g_hash_table_lookup(ht, b) == NULL );
}

int
le_v_hash(void *a, GHashTable *ht)
{
  if (a == NULL || ht == NULL) return (a != ht);
  /* a deve essere maggiore di tutti gli elementi di ht */
  return ( g_hash_table_lookup(ht, a) == NULL );
}

int
le_hash_hash(GHashTable *a, GHashTable *b)
{
  GHashTable *temp = b;
  
  if (a == NULL || b == NULL) return (a != b);
  /* tutti gli elementi di a devono essere maggiori di b */
  if ( g_hash_table_size(a) != g_hash_table_size(b) ) return 1;
  
  /* ATTENZIONE: e' inefficente perche' mi scorre tutti gli elementi nella hash table anche se il primo e' diverso */ 
  g_hash_table_foreach (a, confront_ht, &temp);
  
  return (temp == NULL);
}
