/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: actions.h,v 1.3 2003/02/17 15:32:23 frassi Exp $
 */

#ifndef ACTIONS_H
#define ACTIONS_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include "parser.h"

#include <sys/socket.h>
//#include <net/netinet.h>
#include <netinet/in.h>
#include <arpa/inet.h>

/*************************
 * FUNCTIONS FOR ACTIONS *
 *************************/

void freeze();
void de_freeze(int *pid);
void kill_actual();
int *get_pid();
char *get_ip_addr(struct sockaddr_in *sa);
void log(int *level,char *msg);

void firewalla (char *ipaddr);
void printfile(char *filename, char *str);
void printconsole(char *str);
int printint(int *ve);
void printchar(char *ve);




long * find(char *text,char *str2,char *sep);  // trova nel buffer text (primo argomento) le parole in str2 (secondo arg) distinte da un separatore (terzo arg). Ritorna 1 se trovate altrimenti 0.


long * find_begin(char *text,char *str2,char *sep);//E' simile alla funzione find, soltanto che in piu' fa altri 2 controlli.
// Come prima cosa, davanti alla prima parola da cercare non ci deve essere nessun carattere.
// Come seconda, si considera due situazioni distinte: Se le parole del buffer text provengono da un browser, limito la ricerca sulla query string. Nel caso in cui invece provengono da una richiesta telnet, si cerca le parole sull'intero buffer.


long *two_find(char *text,char *str2,char *sep,int *seq); //analoga alla find, soltanto che in piu' utilizza una variabile globale che da o meno l'accesso a questa funzione(se questa var=0 la tale funzione restituisce sempre 0). 
// Puo' essere utilizzata solo in coppia con una one_"qualcosa" che precedentemente ha messo a 1 la variabile globale.
//Questo tipo di funzioni viene utilizzato quando voglio che una certa syscall si verifichi solo se e' arrivata un'altra syscall.


long *one_find_begin(char *text,char *str2,char *sep,int *seq);
 //analoga alla find_begin, soltanto che in piu' utilizza una variabile globale che da o meno l'accesso a questa funzione(se questa var=0 la tale funzione restituisce sempre 0). 
// Puo' essere utilizzata solo in coppia con una one_"qualcosa" che precedentemente ha messo a 1 la variabile globale.
//Questo tipo di funzioni viene utilizzato quando voglio che una certa syscall si verifichi solo se e' arrivata un'altra syscall.

int find_2_without_3(char *text,char *str2,char *par_in,char *sep); 
// Ritorna 1 se trova nel primo arg. str2 (la stringa da cercare) ma non trova par_in (che equivale a  "/"<sep><path1>) e cioe' la parte che non si deve trovare davanti.
 // Funzione richiamata dalla "find_begin_nobar"
                                                                  

long * find_begin_nobar(char *text,char *str2,char *sep); 
// Ritorna 1 se trova i pattern della stringa str2 divisi dal separatore sep, e se davanti al primo di questi non c'e' "/"

                                                           
long *one_find_begin_nobar(char *text,char *str2,char *sep,int *seq);
//analoga alla find_begin_nobar, soltanto che in piu' utilizza una variabile globale che da o meno l'accesso a questa funzione(se questa var=0 la tale funzione restituisce sempre 0). 
// Puo' essere utilizzata solo in coppia con una one_"qualcosa" che precedentemente ha messo a 1 la variabile globale.
//Questo tipo di funzioni viene utilizzato quando voglio che una certa syscall si verifichi solo se e' arrivata un'altra syscall.

char *read_str_pwd(char *file);  
// Legge e restituisce la prima riga del file che riceve per parametro (per es. /etc/passwd).

int  search_words_or (char *text,char **words); 
// Cerco le words nel buffer della chiamata di sistema, appena ne trova una ritorna 1,  altrimenti ritorna 0.
// Tale funzione viene richiamata dalla "find_or".


long * find_or(char *text,char *str2,char *sep);
// Trova nel buffer (primo arg.) le words (secondo arg.) distinte da un separatore (terzo arg.). Ritorna 1 se trovate altrimenti 0


long *two_find_or(char *text,char *str2,char *sep,int *seq);
 //analoga alla find_or, soltanto che in piu' utilizza una variabile globale che da o meno l'accesso a questa funzione(se questa var=0 la tale funzione restituisce sempre 0). 
// Puo' essere utilizzata solo in coppia con una one_"qualcosa" che precedentemente ha messo a 1 la variabile globale.
//Questo tipo di funzioni viene utilizzato quando voglio che una certa syscall si verifichi solo se e' arrivata un'altra syscall.


int checklenght (char *text,int maxpar);
// controlla che la lunghezza del primo parametro sia inferiore all'arg. "maxpar".
// Ritorna 1 se la lunghezza e' inferiore, altrimenti 0.
// Tale funzione viene chiamata dalla "checksize".


long * checksize (char *p1,char *p2[],int *maxpar1,int *maxpar2);
// controlla che la lunghezza di tutti e due i  parametri sia inferiore a quella dell' arg. "maxpar1" e "maxpar2".
// Ritorna 1 se lunghezza e' minore, altrimenti 0.
                                                     

int  search_cgi_e_dir (GHashTable *tab_cgi_insecure,char **words); 
// cerco le words nella tabella. Appena ne trovo una, ritorna 1,  altrimenti se nessuna e' presente ritorna 0
// Tale funzione e' richiamata dalla find_cgi_e_dir


int find_cgi_e_dir(GHashTable *tab_cgi_insecure,char *path_cgi,char *text); 
// Cerca i cgi pericolosi presenti nel file  di configurazione. 
// Come parametri riceve per primo la hash table,nella quale  ricerca il cgi, per secondo  il path nel quale sono installati gli script.  L'ultimo parametro e' il buffer della execve. Nel caso in cui questo indica che si sta per  eseguire di un cgi, ne viene filtrato la parte dell'estensione (.pl o .cgi) e le eventuali query presenti. Infine sia lo script, sia le varie directory nelle quali si trova, verranno cercate separatamente tramite la funzione search_cgi.
// Se per esempio ho "path_di_install/cgi-bin/path1/path2/script.cgi" cerco separatamente "path1" "path2" "cgi-bin" "script". 
// Puo' essere utile nel caso in cui lo script che ho installato e' stato magari rinominato, a differenza delle directory nelle quali si trova.In questo caso avendo un livello di intercettazione alto posso avere un alto numero di falsi allarmi. 

int  search_cgi (GHashTable *tab_cgi_insecure,char *string_cgi); 
// Analoga alla "search_cgi_e_dir", solo che invece di cercare piu' words ne cerca solo una. Tale funzione e' richiamata dalla "find_cgi_con_path" oppure dalla "find_cgi".
                                                             
int find_cgi_con_path(GHashTable *tab_cgi_insecure,char *path_cgi,char *text) ; // Analoga alla "find_cgi_e_dir", con la differenza che ricerca una stringa formata da dall'unione delle directory e dello script. Ad  es.se ho "path_di_install/cgi-bin/path1/path2/script.cgi" ricerca tutto "/path1/path2/script.cgi" come una stringa unica. Il vantaggio e' che usando tale funzione diminuiscono i falsi allarmi ma diminuisce anche la sicurezza, se infatti i cgi sono posizionati in un'altra dir non vengono intercettati.   

long * find_cgi(GHashTable *tab_cgi_insecure,char *path_cgi,char *text); // Analoga alla "find_cgi_e_dir", con la differenza che ricerca soltanto il nome del cgi e non le varie dir. Per esempio se ho "path_di_install/cgi-bin/path1/path2/script.cgi" ricerco solo "script".Tale funzione e' una via di mezzo tra "find_cgi_e_dir" e "find_cgi_con_path" ed e' la piu' conveniente.


char *union_str(char *parte1,char *parte2); 
// Date due stringhe che riceve per argomento restituisce la concatenazione delle due.
                                            
long * find_execve(char *text,char *argv[],char *str1,char *str2,char *sep,char *sep_arr);
// Come la funzione "multi_find_begin_nobar" si controlla che cio' che sta per eseguire l'execve sia o meno lecito.
// Con questa funzione si controlla sia il primo (opzionale) che il secondo argomento della execve.
// Se la stringa str1 e' vuota, passa direttamente a controllare l'array (argv).
// Se invece non e' vuota e non e' contenuta in 'text' ritorna 0.
// Se infine non e' vuota ed e' contenuta in argv[0] prosegue e  controlla argv.
// In particolare si controlla che il secondo argomento di una execve, che e' un array, sia quello che mi aspetto. Piu' precisamente da str1 si ricava un array. Ogni posizione di questo array deve essere contenuto nella rispettiva posizione dell'array argv, l'argomento 2 dell'execve.
// Si cerca in ogni posizione dell'array tramite la funzione find_begin_nobar le corrispondenti parole di str2 separate da un separatore sep. Le varie posizioni dell'array sono delimitate dal separatore sep_arr.Ritorna uno se ha trovato le parole.


long * find_or_array(char *text,char *argv[],char *str1,char *str2,char *sep,int *num_arr);
// Si controlla se nel secondo argomento dell'execve (argv) sia presente almeno uno dei caratteri particolari di str2 separati da sep.
// Con questa funzione si controlla sia il primo (opzionale) che il secondo argomento della execve.

// Se la stringa str1 e' vuota, passa direttamente a controllare l'array (argv).
// Se invece non e' vuota e non e' contenuta in 'text' ritorna 0.
// Se infine non e' vuota ed e' contenuta in argv[0] prosegue e  controlla argv.
// Ritorna 1 se trova almeno una delle parole di str2 divise da un separatore sep, nella posizione dell'array 'argv' specificata da num_arr.


long *multi_find_begin_nobar(char *testo,char *str2,char *sep,char *sep_control);
// Controlla che tutto cio' che tenta di eseguire l'execve sia lecito.
// In particolare si controlla che il primo argomento della execve, contenga almeno una delle chiamate previste che si trovano in str2.
//_________
// Per ogni sequenza delimitata dal separatore "sep_control" viene richiamata la funzione find_begin_nobar che riceve oltre alla seq. il primo argomento (testo) nel quale effettuare la ricerca. Se almento una di queste chiamate ha esito positivo ritorna 1 altrimenti 0.




long *
search_path (GHashTable *tab_cgi_insecure,char *string_cgi);

int *
get_sa_family(struct sockaddr *sa);
#endif /* !ACTIONS_H */
