/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: automa.c,v 1.3 2003/02/20 16:13:47 frassi Exp $
 */


#include <assert.h>
#include "runtime_util.h"
#include "automa.h"
#include "y.tab.h"

int max_event_num = 0;	 


/**************************************************
 * these functions erase all temporary structures *
 *  used for the construction of the automa       *
 **************************************************/

/*
  free_ev_set_rec

  input:  a pointer to a condition tree node
  output: nothing

  it deletes recursively all event sets starting from the
  given node, excluding those contained in the leaves of
  the condition tree
*/
static void
free_ev_set_rec (ct_node)
     struct condition_tree_node *ct_node;
{
  assert (ct_node);

  ////////////////////////if (!(G_NODE_IS_LEAF (ct_node)))
  ////////////////////////  {
      if (ct_node->ev_set)
	{
	  free_event_set (ct_node->ev_set);
	  ct_node->ev_set = NULL;
	}

      if (ct_node->cond_pos) free_ev_set_rec (ct_node->cond_pos);
      if (ct_node->cond_neg) free_ev_set_rec (ct_node->cond_neg);
  ////////////////////////  }

  return;
}


/*
  free_ev_sets_of_automa

  input:  an automaton
  output: nothing

  deletes all event sets associated with the states of the automa
*/
static void
free_ev_sets_of_automa (atm)
     automa atm;
{
  int i;			//,num_ev_el_removed;

  automa_node *temp_atm_node_ptr;

  for (i = 0; i < atm->len; i++)
    {
      temp_atm_node_ptr = &g_array_index (atm, automa_node, i);

      free_event_set (temp_atm_node_ptr->ev_set_ptr);

      //temp_atm_node_ptr->dim_ev_set=0;
      //temp_atm_node_ptr->hash_ev_set=0;
      temp_atm_node_ptr->ev_set_ptr = NULL;


      //g_hash_table_foreach(temp_atm_node_ptr->edg_set,free_ev_sets_of_edge,NULL);
    }
}

/*
  free_follow_array
  
  input:  the following events array
  output: nothing
  
  deletes the given following events array
*/
static void
free_follow_array (arf)
     follow_array_t arf;
{
  int i;
  //guint num_ev_el_removed;

  for (i = 1; i <= arf.dim_ar; i++)
    {
      if (arf.follow_array[i].ev_set_follow_ptr != NULL)
	free_event_set (arf.follow_array[i].ev_set_follow_ptr);
    }

  free ((arf.follow_array) + 1);
}

/*******
 * end *
 *******/

/*
  empty_expr
  
  input:  a syntax tree node
  
  output: TRUE if the expression is empty (e.g. A* is empty)
          FALSE otherwise
*/
static gboolean
empty_expr (st)
     syntax_tree st;
{
  if ((NODE_OP (st) == repetition)
      || (G_NODE_IS_LEAF (st) && IS_END_MARKER (st)))
    return TRUE;
  else
    {
      if ((G_NODE_IS_LEAF (st)) && (NODE_OP (st) == none))
	return FALSE;
      else
	{
	  if (NODE_OP (st) == sequencing)
	    return ((empty_expr (g_node_first_child (st)))
		    && (empty_expr (g_node_nth_child (st, 1))));
	  else if (NODE_OP (st) == alternation)
	    return ((empty_expr (g_node_first_child (st)))
		    || (empty_expr (g_node_nth_child (st, 1))));
	  else
	    /* 
	       this branch of code should never be executed, I had to write it only
	       because the compiler gave me a warning
	     */
	    {
	      fprintf (stderr, "empty_expr: ERROR - undefined case\n");
	      return FALSE;
	    }
	}
    }
}

/*
  first_ev_numbers_of_expr

  input:  a syntax tree node 
          a event set pointer
	  
	  both structures have to be initialized before passing them to this 
	  function

  output: nothing
  
  given an expression (==syntax tree node) it puts in the event set the
  number of events corresponding to "entry points" (or first events) of 
  the expression
*/
static void
first_ev_numbers_of_expr (st, ev_set_p)
     syntax_tree st;
     event_set_t *ev_set_p;
{

  assert (st != NULL);
  assert (ev_set_p != NULL);

  if ((G_NODE_IS_LEAF (st)) && (NODE_OP (st) == none))
    add_event (ev_set_p, NODE_NUM_EVENT (st));
  else
    {
      if ((NODE_OP (st) == alternation)
	  || ((NODE_OP (st) == sequencing)
	      && (empty_expr (g_node_first_child (st)))))
	{
	  first_ev_numbers_of_expr (g_node_first_child (st), ev_set_p);
	  first_ev_numbers_of_expr (g_node_nth_child (st, 1), ev_set_p);
	}
      else
	{
	  if ((NODE_OP (st) == sequencing) || (NODE_OP (st) == repetition))
	    first_ev_numbers_of_expr (g_node_first_child (st), ev_set_p);
	}
    }
}

/*
  following_ev_numbers_of_expr

  input:  a syntax tree node corresponding to an event number of which
          we want to determine the following event numbers

          a event set pointer which will contain the following event
	  numbers

	  both structures have to be initialized before passing them to this 
	  function

	  a follow_array, which will contain the event set

  output: nothing

  called by syntax_tree_visit
*/
static void
following_ev_numbers_of_expr (st, ev_set_p, arf)
     syntax_tree st;
     event_set_t *ev_set_p;
     follow_array_t arf;
{
  syntax_tree p, s, begin_subex, next_subex, rightmost;
  gboolean rm, op_notset;
  gboolean follow_not_defined = FALSE;
  event_operator op;

  assert ((st != NULL) && (ev_set_p != NULL));

  if (arf.tree_t == COMPLETE_TREE)
    follow_not_defined = (G_NODE_IS_LEAF (st) && IS_END_MARKER (st));
  else
    {
      rightmost = g_node_get_root (st);
      while (!(G_NODE_IS_LEAF (rightmost)))
	rightmost = g_node_last_child (rightmost);
      follow_not_defined = (st == rightmost);
    }

  /* 
     follow not defined is TRUE if I'm trying to determine the following event numbers
     of an event which hasn't any: the last event of a subtree or an end_marker event
   */

  if (follow_not_defined)
    fprintf (stderr,
	     "following_ev_numbers_of_expr: ERROR - follow not defined\n");
  else
    {
      p = st;
      rm = TRUE;
      op_notset = TRUE;

      /* 
         I have to compute the beginning of the subexpression containing
         the syntax tree node which was passed to this function....
       */

      while (!((NODE_OP (p) == sequencing) || (NODE_OP (p) == alternation))
	     || ((ONLY_CHILD (p)) && (rm)))
	{
	  begin_subex = p;
	  p = p->parent;
	  if ((rm) && (begin_subex == g_node_last_child (p)))
	    rm = TRUE;
	  else
	    {
	      rm = FALSE;
	      if (op_notset)
		{
		  op = NODE_OP (p);
		  op_notset = FALSE;
		}
	    }
	}

      if (op_notset)
	{
	  s = begin_subex;
	  while (s == g_node_last_child (p))
	    {
	      s = p;
	      p = p->parent;
	      if (NODE_OP (s) == repetition)
		begin_subex = s;
	    }

	  op = NODE_OP (p);

	  next_subex = g_node_last_child (p);
	}
      else
	next_subex = g_node_next_sibling (begin_subex);


      if (NODE_OP (begin_subex) == repetition)
	first_ev_numbers_of_expr (begin_subex, ev_set_p);	/* first follow rule */

      if (op == sequencing)
	first_ev_numbers_of_expr (next_subex, ev_set_p);	/* second follow rule */

      if ((op == alternation)
	  || ((op == sequencing) && empty_expr (next_subex)
	      && (!(G_NODE_IS_LEAF (next_subex) && IS_END_MARKER (next_subex)))))	/* third follow rule */
	{
	  rightmost = next_subex;
	  while (!(G_NODE_IS_LEAF (rightmost)))
	    rightmost = g_node_last_child (rightmost);
	  following_ev_numbers_of_expr (rightmost, ev_set_p, arf);
	}
    }
}


/*
  syntax_tree_visit

  input: a leaf of the syntax tree 
         a pointer to the follow array structure

  output: FALSE

  this function gets called on every leaf of the syntax tree.
  if it is called on the leaf corresponding to event x, it computes 
  the event set of following event numbers of x

  it returns always FALSE so that all leaves are visited 
  (see Glib documentation about tree visits )
*/
static gboolean
syntax_tree_visit (nd, arf)
     GNode *nd;
     gpointer arf;
{
  int ev_set_dim;
  event_set_t *ev_set_ptr;
  //unsigned int hashv;
  follow_array_t *localarf = ((follow_array_t *) arf);

  //hashv=0;

  assert ((nd != NULL) && (arf != NULL));

  localarf->follow_array[localarf->counter].evid.num_event =
    NODE_NUM_EVENT (nd);
  localarf->follow_array[localarf->counter].evid.context = NODE_CONTEXT (nd);

  localarf->follow_array[localarf->counter].ev_nod_ptr =
    ((event_node_t *) nd->data);

  if (IS_END_MARKER (nd))
    {
      localarf->follow_array[localarf->counter].is_endmarker = TRUE;
      //localarf->follow_array[localarf->counter].hash_ev_set_follow=0;
    }
  else
    {
      ev_set_ptr = new_event_set (max_event_num, NULL);

      if (!(IS_EMPTY_FOLLOW (nd)))
	following_ev_numbers_of_expr (nd, ev_set_ptr, *localarf);

      ev_set_dim = events_in_set (ev_set_ptr);
      //localarf->follow_array[localarf->counter].dim_ev_set_follow=ev_set_dim;
      localarf->follow_array[localarf->counter].ev_set_follow_ptr =
	ev_set_ptr;

      //if (!(IS_EMPTY_FOLLOW(nd)))
      //hashv=create_hash_ev_set(localarf->follow_array[localarf->counter].ev_set_follow_ptr);

      //localarf->follow_array[localarf->counter].hash_ev_set_follow=hashv;
    }

  localarf->counter++;

  return FALSE;
}


/*
  create_follow_array_of_tree

  input: the root of a syntax tree
         an int which says if it is a COMPLETE_TREE or a SUBTREE

  output: a follow_array_t structure
*/
follow_array_t create_follow_array_of_tree (root, tree_t)
     syntax_tree
       root;
     int
       tree_t;
{
  int i;
  guint leaves_num;
  follow_array_t arf;

  assert (root != NULL);

  leaves_num = g_node_n_nodes (root, G_TRAVERSE_LEAFS);

  max_event_num = leaves_num;

  arf.follow_array =
    ((follow_array_element_t *)
     calloc (leaves_num, sizeof (follow_array_element_t)));
  --(arf.follow_array);

  /* 
     this is a trick so that the array arf.follow_array begins at 1.
     You should NEVER try to access arf.follow_array[0].
     To free the array: free((arf.follow_array) + 1)
   */

  arf.counter = 1;
  arf.tree_t = tree_t;
  arf.dim_ar = leaves_num;

  for (i = 1; i <= leaves_num; i++)
    {
      arf.follow_array[i].is_endmarker = FALSE;
      /*INT_MIN=-2147483647 defined in limits.h */
      arf.follow_array[i].evid.num_event = INT_MIN;
      arf.follow_array[i].ev_set_follow_ptr = NULL;
      arf.follow_array[i].ev_nod_ptr = NULL;
      //arf.follow_array[i].hash_ev_set_follow=0;
      //arf.follow_array[i].dim_ev_set_follow=0;
    }

  g_node_traverse (root, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, syntax_tree_visit,
		   &arf);

  return arf;
}



/*
  build_root_cond_tree

  input:  -an event number
          -a pointer to an event_node_t struct
	  -a edges set (==hashtable containing all information about edges of a node)

  output: nothing
  

  it checks if the syscall X (==an edge marked with this syscall) is already present 
  in edg_set.
  if it isn't, it adds a new element to the edg_set concerning X
  it adds the event number (corresponding to syscall X) in the opportune event set
*/
static void
build_root_cond_tree (ev_num, ev_node_ptr, edg_set)
     int ev_num;
     event_node_t *ev_node_ptr;
     edges_set_t edg_set;
{
  int nev;			/* event_number */
  edge_key_t *key_ptr = ((edge_key_t *) malloc (sizeof (edge_key_t)));
  edge_t *new_edge_ptr;

  if (ev_node_ptr->simple_event->n_sysc != END_MARKER)
    {
      /* I can build condition trees only for real events */

      key_ptr->syscall_num = ev_node_ptr->simple_event->n_sysc;
      key_ptr->context = ev_node_ptr->simple_event->context;

      nev = ev_node_ptr->simple_event->num_event;

      assert (nev == ev_num);

      new_edge_ptr =
	((edge_t *) g_hash_table_lookup (((edges_set_t) edg_set), key_ptr));

      if (new_edge_ptr == NULL)	/*in edge_set there isn't key_ptr */
	{
	  new_edge_ptr = ((edge_t *) malloc (sizeof (edge_t)));
	  new_edge_ptr->key.syscall_num = key_ptr->syscall_num;
	  new_edge_ptr->key.context = key_ptr->context;
	  new_edge_ptr->ct_root =
	    ((struct condition_tree_node *)
	     malloc (sizeof (struct condition_tree_node)));

	  new_edge_ptr->ct_root->cond_ptr = NULL;
	  new_edge_ptr->ct_root->cond_pos = NULL;
	  new_edge_ptr->ct_root->cond_neg = NULL;
	  new_edge_ptr->ct_root->trans_l = NULL;
	  new_edge_ptr->ct_root->ev_set = new_event_set (max_event_num, NULL);

	  assert (new_edge_ptr->ct_root->ev_set != NULL);

	  g_hash_table_insert (((edges_set_t) edg_set), &(new_edge_ptr->key),
			       new_edge_ptr);
	}

      add_event (new_edge_ptr->ct_root->ev_set, ev_num);
      //assert(new_edge_ptr->ct_root!=NULL);
    }

  free (key_ptr);
}


/************************************************
 * functions used to determine if there is a    *
 * conflict between conditions we want to check *
 ************************************************/

/********************************************
 * struct and functions used for            *
 * comparing 2 variables of type HASH_VALUE *
 ********************************************/

struct generic_equal_each_hash_val_user_data
{
  GHashTable *ht;
  int ret;
};


static void
generic_equal_each_hash_val (key, value, user_data)
     gpointer key;
     gpointer value;
     gpointer user_data;
{
  gpointer value2;
  struct generic_equal_each_hash_val_user_data *ud = user_data;

  if (ud->ret)
    {
      value2 = g_hash_table_lookup (ud->ht, key);
      if (value2 == NULL)
	ud->ret = 0;		/* second condition should never happen */
    }

  return;
}


static int
equal_ht_values (a, b)
     GHashTable *a;
     GHashTable *b;
{
  struct generic_equal_each_hash_val_user_data user_data;
  user_data.ht = b;
  user_data.ret = 1;
  g_hash_table_foreach (a, generic_equal_each_hash_val, &user_data);

  return user_data.ret;
}


static int
equal_array (a, b)
     GList *a;
     GList *b;
{
  gboolean result = TRUE;

  for (; a != NULL; a = g_list_next (a), b = g_list_next (b))
    {
      container *cont1 = a->data;
      container *cont2 = b->data;

      if (cont1->type_content != cont2->type_content)
	result = FALSE;
      else
	{
	  switch (cont1->type_content)
	    {

	    case SIMPLE_VALUE:
	      if (cont1->type_value != cont2->type_value)
		result = FALSE;
	      else
		switch (cont1->type_value)
		  {
		  case LONG_VALUE:
		    result =
		      (*((long *) cont1->content.value) ==
		       *((long *) cont2->content.value));
		    break;
		  case CHARACTER_VALUE:
		    result =
		      (*((char *) cont1->content.value) ==
		       *((char *) cont2->content.value));
		    break;
		  case STRING_VALUE:
		    result =
		      (strcmp
		       (((char *) cont1->content.value),
			((char *) cont2->content.value)) == 0);
		    break;
		  }
	      break;

	    case HANDLER:
	      if (cont1->content.handler->var_name != NULL
		  && cont2->content.handler->var_name != NULL)
		{
		  if (cont1->content.handler->var_name !=
		      cont2->content.handler->var_name)
		    /* different variable names */
		    result = FALSE;
		}
	      else
		if (cont1->content.handler->var_name == NULL
		    && cont2->content.handler->var_name == NULL)
		{
		  if (cont1->content.handler->label_ref.n_arg !=
		      cont2->content.handler->label_ref.n_arg)
		    /* different placeholder names */
		    result = FALSE;
		}
	      else
		result = FALSE;	/* a placeholder and a variable */
	      break;
	    default:
	      ;
	    }
	}


      if (result == FALSE)
	return FALSE;
    }

  return TRUE;
}

/*******
 * end * 
 *******/

/*
  equal_args 

  input:  pointers to the container structures containing the 
          arguments we want to compare

  output: TRUE if equal
          FALSE otherwise

  compares 2 arguments used in an assignment
*/
static gboolean
equal_args (cont1, cont2)
     container *cont1;
     container *cont2;
{
  gboolean result = TRUE;
  guint dim_ht1, dim_ht2;
  //events_sorted_set_t* ev_sorted_set;

  /*cont1 == sm_asg_to_check   cont2 == sm_asgs_list_el */
  /*
     we have conflict if:
     1) arguments have different types (e.g. a handler and a simple 
     value or a simple value and a hashtable)
     2) arguments have equal types but different values
     3) arguments are of the variable names or placeholder types
     and they are different
   */
  if (cont1->type_content != cont2->type_content)	/*1) */
    result = FALSE;
  else				/* same type */
    {
      switch (cont1->type_content)
	{

	case SIMPLE_VALUE:
	  if (cont1->type_value != cont2->type_value)
	    result = FALSE;
	  else
	    switch (cont1->type_value)
	      {
	      case LONG_VALUE:
		result =
		  (*((long *) cont1->content.value) ==
		   *((long *) cont2->content.value));
		break;
	      case CHARACTER_VALUE:
		result =
		  (*((char *) cont1->content.value) ==
		   *((char *) cont2->content.value));
		break;
	      case STRING_VALUE:
		result =
		  (strcmp
		   (((char *) cont1->content.value),
		    ((char *) cont2->content.value)) == 0);
		break;
	      }
	  break;

	case HANDLER:
	  if (cont1->content.handler->var_name != NULL
	      && cont2->content.handler->var_name != NULL)
	    {
	      if (cont1->content.handler->var_name !=
		  cont2->content.handler->var_name)
		/* different variable names */
		result = FALSE;
	    }
	  else
	    if (cont1->content.handler->var_name == NULL
		&& cont2->content.handler->var_name == NULL)
	    {
	      if (cont1->content.handler->label_ref.n_arg !=
		  cont2->content.handler->label_ref.n_arg)
		/* different placeholder names */
		result = FALSE;
	    }
	  else
	    result = FALSE;	/* a placeholder and a variable */
	  break;

	case HASH_VALUE:
	  if (cont1->type_value != cont2->type_value)
	    result = FALSE;
	  else
	    {
	      dim_ht1 = g_hash_table_size (cont1->content.hash_value);
	      dim_ht2 = g_hash_table_size (cont2->content.hash_value);
	      if (dim_ht1 != dim_ht2)
		result = FALSE;
	      else
		{
		  result =
		    equal_ht_values (cont1->content.hash_value,
				     cont2->content.hash_value);
		  /* TODO RISCRIVERE!!!! */
		  //ev_sorted_set=copy_events_to_sorted_set(cont1->content.hash_value);
		  /*da controllare: cosa contengono le 2 hash table? funzionera' la copia insieme di numerini? */
		  //result=equal_event_set(ev_sorted_set,cont2->content.hash_value);

		  //free_events_sorted_set(ev_sorted_set);
		}
	    }
	  break;

	case ARRAY:
	  if (cont1->type_value != cont2->type_value)
	    result = FALSE;
	  else
	    {
	      int dim_arr1 = g_list_length (cont1->content.value);
	      int dim_arr2 = g_list_length (cont2->content.value);
	      if (dim_arr1 != dim_arr2)
		result = FALSE;
	      else
		{
		  result =
		    equal_array (cont1->content.value, cont2->content.value);
		}
	    }
	  break;
	}

    }
  return result;
}



/*
  equal_funct

  input:  2 nodes containing pointers to the functions we want compare
  output  TRUE if functions are equal (==no conflict in assignments)
	  FALSE otherwise

  compares two functions used in assignments
*/
static gboolean
equal_funct (root1, root2)
     GNode *root1;
     GNode *root2;
{
  int i;
  gboolean result = TRUE;
  guint n_figli1, n_figli2;
  GNode *f1, *f2;

  /*root1 == sm_asg_to_check   root2 == sm_asgs_list_el */
  /*
     -we have conflict if:
     1)same functions but different arguments (pay attention
     to the fact that functions can have a variable number of 
     arguments, therefore f(a,b,c) and f(a,b) causa a conflict.
     Moreover functions can have other functions as arguments).
     2)different functions
   */
  if ((root1 != NULL) && (root2 != NULL))
    {
      if (((arg_node_t *) root1->data)->funct !=
	  ((arg_node_t *) root2->data)->funct)
	/*different functions */
	result = FALSE;
      else
	{
	  n_figli1 = g_node_n_children (root1);
	  n_figli2 = g_node_n_children (root2);

	  if ((n_figli1 != n_figli2)
	      || (g_node_n_nodes (root1, G_TRAVERSE_ALL) !=
		  g_node_n_nodes (root2, G_TRAVERSE_ALL)))
	    /*same functions but different arguments */
	    result = FALSE;
	  else
	    {
	      /*same function and same number of arguments ==> I have to compare arguments... */
	      for (i = 0; i < n_figli1; i++)
		{
		  f1 = g_node_nth_child (root1, i);
		  f2 = g_node_nth_child (root2, i);
		  assert (f1 != NULL && f2 != NULL);
		  if (((arg_node_t *) f1->data)->funct != NULL
		      && ((arg_node_t *) f2->data)->funct != NULL)
		    result = result && equal_funct (f1, f2);
		  else
		    if (((arg_node_t *) f1->data)->arg != NULL
			&& ((arg_node_t *) f2->data)->arg != NULL)
		    result = result
		      && equal_args (((arg_node_t *) f1->data)->arg,
				     ((arg_node_t *) f2->data)->arg);
		  else
		    result = FALSE;
		}
	    }
	}
    }
  else if ((root1 == NULL) && (root2 == NULL))
    result = TRUE;
  else				/* root1 is NULL and root2 isn't or viceversa ===> not equal */
    result = FALSE;

  return result;
}


/*
  asgs_conflict

  input: -a list of assignments asgs_list
         -a pointer to an assignment asg_to_check
  output: TRUE if there is conflict between asg_to_check and one of the
          assignments contained in the list
	  FALSE otherwise
	  
  there is a conflict if two different assignments try to give the same variable
  two different value (ie. a=5 and a=8 causes a conflict).
  Action do not cause any conflict.
*/
static gboolean
asgs_conflict (asgs_list, asg_to_check)
     GList *asgs_list;
     GList *asg_to_check;
{
  /*
     pm=first member
     sm=second member
   */

  gboolean pm_conflict = FALSE;
  gboolean sm_conflict = FALSE;
  char *var_name_to_check, *var_name_to_compare;
  GNode *pm_asg_to_check, *pm_asgs_list_el, *sm_asg_to_check,
    *sm_asgs_list_el;
  GList *placeholder;

  if ((asgs_list != NULL) && (asg_to_check != NULL))
    {

      if (((arg_node_t *) ((GNode *) asg_to_check->data)->data)->funct !=
	  (void *(*)()) eval_stmt)
	{

	  pm_asg_to_check = g_node_first_child ((GNode *) asg_to_check->data);
	  assert (pm_asg_to_check != NULL);
	  var_name_to_check =
	    ((arg_node_t *) pm_asg_to_check->data)->arg->content.handler->
	    var_name;
	  sm_asg_to_check =
	    g_node_nth_child ((GNode *) asg_to_check->data, 1);
	  assert (sm_asg_to_check != NULL);

	  placeholder = asgs_list;

	  while (!(pm_conflict && sm_conflict) && (placeholder != NULL))
	    {

	      if (((arg_node_t *) ((GNode *) placeholder->data)->data)->
		  funct != (void *(*)()) eval_stmt)
		{

		  pm_asgs_list_el =
		    g_node_first_child ((GNode *) placeholder->data);
		  var_name_to_compare =
		    ((arg_node_t *) pm_asgs_list_el->data)->arg->content.
		    handler->var_name;

		  if (strcmp (var_name_to_check, var_name_to_compare) == 0)
		    {
		      pm_conflict = TRUE;
		      /*
		         I have to check second member: could be a function or a simple 
		         value

		         1)if one is a function and the other a simple value  ==> conflict
		       */
		      sm_asgs_list_el =
			g_node_nth_child ((GNode *) placeholder->data, 1);


		      if (
			  (((arg_node_t *) sm_asg_to_check->data)->arg !=
			   NULL)
			  && (((arg_node_t *) sm_asgs_list_el->data)->arg !=
			      NULL))
			sm_conflict =
			  !(equal_args
			    (((arg_node_t *) sm_asg_to_check->data)->arg,
			     ((arg_node_t *) sm_asgs_list_el->data)->arg));
		      else
			if (
			    (((arg_node_t *) sm_asg_to_check->data)->funct !=
			     NULL)
			    && (((arg_node_t *) sm_asgs_list_el->data)->
				funct != NULL))
			sm_conflict =
			  !(equal_funct (sm_asg_to_check, sm_asgs_list_el));
		      else
			sm_conflict = TRUE;	/*1) */

		    }
		}
	      placeholder = g_list_next (placeholder);
	    }
	}
    }
  return (pm_conflict && sm_conflict);
}

/*******
 * end *
 *******/


/*
  search_automa_state

  input: an event set
         an automaton

  output: -1 if the automaton doesn't contain any state with
             the same event set
	  or
	  the number of the automa state which has the same 
	  event set

  given an event set and an automaton it seeks a state of the
  automaton with the same event set
*/
static int
search_automa_state (atm, ev_set)
     automa atm;
     event_set_t *ev_set;
{
  automa_node *nd;
  int state_num, found_state, set_dim;

  found_state = 0;
  state_num = 0;
  set_dim = events_in_set (ev_set);

  while (state_num < atm->len && !found_state)
    {
      nd = &g_array_index (atm, automa_node, state_num);

      //if (nd->dim_ev_set == set_dim)
      found_state = equal_event_set (ev_set, nd->ev_set_ptr);

      state_num++;
    }

  if (found_state)
    return --state_num;

  return -1;
}

/*
  insert_new_state_in_automa

  input: an event set
         an automaton

  output: the number of the new automa state 

  given an event set and an automaton it creates a new 
  state of the automaton with the given event set and
  inserts it in the automaton
*/
static int
insert_new_state_in_automa (atm, ev_set)
     automa atm;
     event_set_t *ev_set;
{
  automa_node *new_node;
  int position;

  AUTOMA_NODE_INIT (new_node);

  new_node->ev_set_ptr = ev_set;
  //new_node->dim_ev_set=events_in_set(ev_set);
  //new_node->hash_ev_set=create_hash_ev_set(ev_set);

  g_array_append_val (atm, *new_node);
  position = (atm->len) - 1;

  if ((position % 50) == 0)
    printf ("created automa state %d\n", position);

  return position;
}

/*
  create_asgs_list

  input:  an event number
          2 event sets
	  a pointer to a list of assignments
	  the following events array

  output: -1 if no conflict occured during the creation of the
             assignment list
	  or
	     the number of the event which caused the conflict

   it puts together the lists of assignments associated to the
   event numbers contained in input_ev_set in a single list
   called asgs_l. It checks if any conflict arises during the
   creation of the list.
 */
static int
create_asgs_list (event, input_ev_set, output_ev_set, asgs_l, arf)
     int event;
     event_set_t *input_ev_set;
     event_set_t *output_ev_set;
     GList **asgs_l;
     follow_array_t arf;
{
  int conflict = 0;
  GList *asgs_l_temp, *placeholder;

  while (event != -1)
    {
      asgs_l_temp = NULL;
      placeholder = get_action_ptr (arf, event);

      while (placeholder
	     && !(conflict = asgs_conflict (*asgs_l, placeholder)))
	{
	  asgs_l_temp = g_list_append (asgs_l_temp, placeholder->data);
	  placeholder = g_list_next (placeholder);
	}

      if (conflict)
	{
	  g_list_free (asgs_l_temp);
	  return event;
	}
      else
	{
	  *asgs_l = g_list_concat (*asgs_l, asgs_l_temp);
	  add_event (output_ev_set, event);
	  first_event_after (input_ev_set, event, event);
	  //input_ev_set->iterator=event; 
	  //event = each_event(input_ev_set);
	}
    }
  return -1;
}


/* 
   compute_follow_of_ev_set

   input:  the following events array
           an event set

   output: an event set

   it computes the following event numbers of those
   given in the input event set and puts them in another
   event set
 */
static event_set_t *
compute_follow_of_ev_set (arf, ev_set)
     follow_array_t arf;
     event_set_t *ev_set;
{
  event_set_t *temp_set, *follow_ev_set;
  int event;

  follow_ev_set = new_event_set (max_event_num, NULL);

  first_event (ev_set, event);

  while (event != -1)
    {
      temp_set = arf.follow_array[event].ev_set_follow_ptr;
      follow_ev_set = join_event_set (follow_ev_set, temp_set);
      event = each_event (ev_set);
    }

  return follow_ev_set;
}



#define condition_tree_node_init(node,ev_set_to_join) \
{ \
   node = ((struct condition_tree_node*) malloc(sizeof(struct condition_tree_node))); \
   node->ev_set = new_event_set(max_event_num,NULL); \
   node->ev_set = join_event_set(node->ev_set,ev_set_to_join); \
   node->cond_ptr = NULL; \
   node->cond_pos = NULL; \
   node->cond_neg = NULL; \
   node->trans_l = NULL; \
}


/*
  bct_rec

  input:  an event number
          the following events array
	  an event set
	  a node of the condition tree
	  an automaton

  output: NULL

          builds recursively the condition tree
*/
static void
bct_rec (event, arf, ev_set, root, atm)
     int event;
     follow_array_t arf;
     event_set_t *ev_set;
     cond_tree_root root;
     automa atm;
{
  event_set_t *temp_ev_set;
  struct condition_tree_node *condtemp1, *condtemp2;
  int ev_num_to_erase, conflict;
  GList *asgs_l = NULL;
  event_set_t *new_candidate_ev_set_ptr;
  int number_of_node_to_create_epsilon_trans_with;
  single_transition *new_trans_ptr;

  if (event != -1)		/* there are still conditions to check */
    {
      /* update condition tree node */

      root->negated_cond = get_negated_condition (arf, event);
      root->cond_ptr = get_cond_ptr (arf, event);

      ev_num_to_erase = event;

      /* find next condition to check */
      first_event_after (ev_set, event, event);
      while ((event != -1) && (get_cond_ptr (arf, event) == NULL))
	event = each_event (ev_set);

      /* create new condition tree node */

      /*
         ev_set associated to this node is correct, but could be more efficient (e.g. smaller)
         if we could determine on the basis of the conditions to check if there are impossible
         combinations (i.e. (a && !a) is always false, so why check?)
       */

      condition_tree_node_init (condtemp1, root->ev_set);

      root->cond_pos = condtemp1;

      /*
         before creating !C branch, I check that cardinality of ev_set is > 1
         because it could happen something like this:
         {6}--C2-->{6} ok
         {6}--!C2-->{?} <-- empty ev_set, where do I go from here?
       */

      if (events_in_set (root->ev_set) > 1)
	{
	  condition_tree_node_init (condtemp2, root->ev_set);
	  del_event (condtemp2->ev_set, ev_num_to_erase);

	  root->cond_neg = condtemp2;
	}

      bct_rec (event, arf, ev_set, root->cond_pos, atm);

      if (root->cond_neg)
	bct_rec (event, arf, ev_set, root->cond_neg, atm);
    }
  else				/* checked all conditions */
    {
      /* negative_cond not interesting, in this case */
      root->negated_cond = -1;
      conflict = -1;

      first_event (root->ev_set, event);

      while (event != -1)
	{
	  temp_ev_set = new_event_set (max_event_num, NULL);
	  asgs_l = NULL;

	  conflict =
	    create_asgs_list (event, root->ev_set, temp_ev_set, &asgs_l, arf);

	  if (conflict > 0)
	    {
	      int decision;
	      printf
		("\n***WARNING: A conflict between assignments has been detected***\n");
/* la definisco se sono sicuro che compilero' file delle specifiche senza conflitti tra assegnamenti */
#ifdef NO_CONFLICT

	      exit(1);
#endif
	      printf
		("\nDo you want to (c)ontinue or (q)uit and modify the configuration file?");

	      do
		{
		  decision = getchar ();
		  if (decision == (int) 'q')
		    exit (1);
		}
	      while (decision != (int) 'c');
	    }

	  new_candidate_ev_set_ptr =
	    compute_follow_of_ev_set (arf, temp_ev_set);

	  free_event_set (temp_ev_set);

	  number_of_node_to_create_epsilon_trans_with =
	    search_automa_state (atm, new_candidate_ev_set_ptr);

	  if (number_of_node_to_create_epsilon_trans_with == -1)
	    number_of_node_to_create_epsilon_trans_with =
	      insert_new_state_in_automa (atm, new_candidate_ev_set_ptr);
	  else
	    free_event_set (new_candidate_ev_set_ptr);

	  /*create  epsilon transition toward automa node whose number is number_of_state.... */
	  new_trans_ptr =
	    ((single_transition *) malloc (sizeof (single_transition)));
	  new_trans_ptr->automa_node_number =
	    number_of_node_to_create_epsilon_trans_with;
	  new_trans_ptr->asgs_list = asgs_l;	/*this list contains both assignments and actions */

	  root->trans_l = g_list_append (root->trans_l, new_trans_ptr);
	  event = conflict;
	}
    }
}


/*
  init_bct 

  input:  a syscall number 
          a edge_t structure pointer
	  a pointer to user data

  output: NULL
  
  it is used to build the condition tree. It gets called on the root of
  the condition tree and subsequently it calls for the first time bct_rec
*/
static void
init_bct (key_ptr, edge_ptr, ud)
     gpointer key_ptr;
     gpointer edge_ptr;
     gpointer ud;
{
  struct userd *localud = ((struct userd *) ud);
  int event;

  localud->ev_set_ptr = (((edge_t *) edge_ptr)->ct_root->ev_set);

  first_event (localud->ev_set_ptr, event);

  while ((event != -1) && (get_cond_ptr (localud->arf, event) == NULL))
    event = each_event (localud->ev_set_ptr);

  bct_rec (event, localud->arf, localud->ev_set_ptr,
	   ((edge_t *) edge_ptr)->ct_root, localud->atm);

  // l'ho tolta per debug, ma serve:  
  free_ev_set_rec (((edge_t *) edge_ptr)->ct_root);
}


/*
  update_final_states

  input:  an automa
          the following events array

  output: NULL

  it marks the final states of the automa
*/
static void
update_final_states (atm, arf)
     automa atm;
     follow_array_t arf;
{
  int i = 1;
  int j;
  automa_node *temp_atm_node_ptr;

  while (i <= arf.dim_ar)
    {
      if (arf.follow_array[i].is_endmarker)
	{
	  j = 0;
	  while (j < atm->len)
	    {
	      temp_atm_node_ptr = &g_array_index (atm, automa_node, j);
	      if (lookup_event (temp_atm_node_ptr->ev_set_ptr, i))
		temp_atm_node_ptr->is_final = 1;
	      j++;
	    }
	}
      i++;
    }
}




/*
  AQE_gen

  input:  the root of a syntax tree
  output: an automaton 

  it builds an automaton corresponding to the given syntax tree
*/
automa AQE_gen (root)
     syntax_tree
       root;
{
  event_set_t *temp_ev_set_ptr;
  automa_node *temp_atm_node_ptr, *atm_node_index_ptr;
  int i, event;

  struct userd *ud = ((struct userd *) malloc (sizeof (struct userd)));
  follow_array_t arf;

  automa atm = AUTOMA_INIT (FALSE, FALSE);

  /* computes the following events array */
  arf = create_follow_array_of_tree (root, COMPLETE_TREE);

  /* creates the first state of the automa and inserts it in the automa */
  AUTOMA_NODE_INIT (temp_atm_node_ptr);
  temp_ev_set_ptr = new_event_set (max_event_num, NULL);

  first_ev_numbers_of_expr (root, temp_ev_set_ptr);

  temp_atm_node_ptr->ev_set_ptr = temp_ev_set_ptr;
  //temp_atm_node_ptr->dim_ev_set=events_in_set(temp_ev_set_ptr);
  //temp_atm_node_ptr->hash_ev_set=create_hash_ev_set(temp_atm_node_ptr->ev_set_ptr);

  g_array_append_val (atm, *temp_atm_node_ptr);

  i = 0;
  ud->arf = arf;
  ud->atm = atm;

  /* for each state of the automa... */
  while (i < atm->len)
    {
      atm_node_index_ptr = &g_array_index (atm, automa_node, i);

      /* builds the edges set, containing all transitions of the automa starting from state i */

      first_event (atm_node_index_ptr->ev_set_ptr, event);

      while (event != -1)
	{
	  build_root_cond_tree (event, arf.follow_array[event].ev_nod_ptr,
				atm_node_index_ptr->edg_set);
	  event = each_event (atm_node_index_ptr->ev_set_ptr);
	}

      ud->ev_set_ptr = NULL;

      /* for each element of edges_set (==a transition) it creates a condition tree */
      g_hash_table_foreach (atm_node_index_ptr->edg_set, init_bct, ud);

      i++;
    }

  update_final_states (atm, arf);

  //serve!! eliminata per debug
  free_ev_sets_of_automa (atm);

  free_follow_array (arf);

  free (ud);

  return atm;
}


/* 
  g_edge_hash
  
  hash function for the edge_set type (which is a GHashtable) 
*/
guint g_edge_hash (k1)
     gconstpointer
       k1;
{
  //  guint hash_val;
  gint ncs1;
  context_t cnt1;

  ncs1 = ((edge_key_t *) k1)->syscall_num;
  cnt1 = ((edge_key_t *) k1)->context;
  if (cnt1 != intro)
    ncs1 = ncs1 + MAX_NUM_SYSCALLS;

  //hash_val=g_int_hash(&ncs1);

  //return hash_val;
  return ncs1;
}

/* 
  g_edge_equal
  
  input:  2 edge_set elements
  output: TRUE if they are equal
          FALSE otherwise
*/
gint g_edge_equal (k1, k2)
     gconstpointer
       k1;
     gconstpointer
       k2;
{
  if (
      ((((edge_key_t *) k1)->syscall_num) ==
       (((edge_key_t *) k2)->syscall_num))
      && ((((edge_key_t *) k1)->context) == (((edge_key_t *) k2)->context)))
    return TRUE;
  else
    return FALSE;
}

/***********************************
 * funzioni used to determine      *
 * subsequences in the syntax tree *
 ***********************************/


/*
  sort_ev_num_func
  
  input:  2 event numbers pointers
  
  output: 0  if equal
          -1 if ev1<ev2
	  1  if ev1>ev2

  it is used to sort event numbers contained in a list
*/
gint sort_ev_num_func (list_el1, list_el2)
     gconstpointer
       list_el1;
     gconstpointer
       list_el2;
{
  int ev_num1, ev_num2;

  ev_num1 = *((int *) list_el1);
  ev_num2 = *((int *) list_el2);

  if (ev_num1 == ev_num2)
    return 0;
  else
    {
      if (ev_num1 <= ev_num2)
	return -1;
      else
	return 1;
    }
}



/*
  subex_is_single_call

  input:  2 nodes of the syntax tree

  output: 1 if the nodes refer to a single system call 
            (one is entering point of the call, the other is the exiting point of the call)
	  0 otherwise

  i.e. a_intro;a_exit is a "single call"
       a_exit;b_intro is not a "single call"
 */
static int
subex_is_single_call (first_child, second_child)
     GNode *first_child;
     GNode *second_child;
{
  if ((G_NODE_IS_LEAF (first_child)) && (G_NODE_IS_LEAF (second_child)))
    {
      if ((NODE_OP (first_child) == none) && (NODE_OP (second_child) == none))
	{
	  if (NODE_NUM_OF_SYSCALL (first_child) ==
	      NODE_NUM_OF_SYSCALL (second_child))
	    {
	      if ((NODE_CONTEXT (first_child) == intro)
		  && (NODE_CONTEXT (second_child) == exiting))
		{
		  return 1;
		}
	    }
	}
    }

  return 0;
}


/*
  compute_subrange_of_a_sequence

  input:  a node of the syntax tree
          a list of subranges

  output: a list of subranges

  it gets called on subexpr of the syntax tree
*/
static GList *
compute_subrange_of_a_sequence (node, lst)
     syntax_tree node;
     GList *lst;
{
  int *first_ev_num;
  syntax_tree temp_node;
  GNode *first_child, *second_child;

  if (NODE_OP (node) != sequencing)
    {
      if (!(empty_expr (node)))
	{
	  /*compute first ev_num of subexpr */
	  temp_node = node;

	  while (!(G_NODE_IS_LEAF (temp_node)))
	    temp_node = g_node_first_child (temp_node);

	  first_ev_num = &NODE_NUM_EVENT (temp_node);

	  /*add ev_num to the list of subranges */
	  lst = g_list_append (lst, first_ev_num);
	}
    }
  else
    {
      first_child = g_node_first_child (node);
      second_child = g_node_nth_child (node, 1);

      if (subex_is_single_call (first_child, second_child))
	lst = compute_subrange_of_a_sequence (first_child, lst);
      else
	{
	  lst = compute_subrange_of_a_sequence (first_child, lst);
	  lst = compute_subrange_of_a_sequence (second_child, lst);
	}
    }

  return lst;
}



/*
  compute_subranges_rec

  input:  root of the syntax tree
          a list of subranges

  output: a list of subranges

  it get called on the root of the syntax tree
  calculates the list of subranges
*/
begin_of_subseqs_ll compute_subranges_rec (node, ll)
     syntax_tree
       node;
     begin_of_subseqs_ll
       ll;
{
  GList *temp_list = NULL;
  syntax_tree temp_node;
  int *last_ev_num, *first_ev_num;

  assert (node != NULL);

  if (NODE_OP (node) == sequencing)
    {
      temp_node = node;

      /* compute the first and the last ev num of subexpr */
      assert (IS_END_MARKER (g_node_nth_child (temp_node, 1)));
      last_ev_num = &NODE_NUM_EVENT (g_node_nth_child (temp_node, 1));

      while (!(G_NODE_IS_LEAF (temp_node)))
	temp_node = g_node_first_child (temp_node);

      first_ev_num = &NODE_NUM_EVENT (temp_node);

      temp_list = compute_subrange_of_a_sequence (node, temp_list);

      if (temp_list)
	temp_list = g_list_sort (temp_list, sort_ev_num_func);

      temp_list = g_list_prepend (temp_list, last_ev_num);
      temp_list = g_list_prepend (temp_list, first_ev_num);

      ll = g_list_append (ll, temp_list);
    }
  else
    {
      if (NODE_OP (node) == alternation)
	{
	  ll = compute_subranges_rec (g_node_first_child (node), ll);
	  ll = compute_subranges_rec (g_node_nth_child (node, 1), ll);
	}
    }

  return ll;
}

/*******
 * end *
 *******/



/*************************************************************
 * DEBUG FUNCTIONS 1: functions used for printing the automa *
 *************************************************************/


void
print_ev_set (event_set_t * ev_set)
{
  int i;

  printf ("ev_set: {");
  if (events_in_set (ev_set) > 0)
    {
      for (i = 1; i <= ev_set->max_events; i++)
	if (lookup_event (ev_set, i))
	  printf ("%d,", i);
    }
  printf ("}\n");
}


void
print_follow_array (follow_array_t arf)
{
  int dim = arf.dim_ar;
  int i;

  printf ("dim=%d\ttree type=", dim);
  if (arf.tree_t == COMPLETE_TREE)
    printf ("COMPLETE TREE\n\n");
  else
    printf ("SUBTREE\n\n");

  for (i = 1; i <= dim; i++)
    {
      printf ("event: %d", arf.follow_array[i].evid.num_event);
      if (arf.tree_t == SUBTREE)
	{
	  if (arf.follow_array[i].evid.context == intro)
	    printf ("-intro\t");
	  else
	    printf ("-exiting\t");
	}
      else
	printf ("\t");
      if (arf.follow_array[i].is_endmarker)
	printf ("endmarker\t");
      //printf("hash htn: %d\n",arf.follow_array[i].hash_ev_set_follow);
      //printf("dim htn: %d\nhtn=",arf.follow_array[i].dim_ev_set_follow);
      if (arf.follow_array[i].ev_set_follow_ptr != NULL)
	print_ev_set (arf.follow_array[i].ev_set_follow_ptr);
      else
	printf ("no ev_set\n");
      printf ("---------------------------------\n");
    }
  printf ("\n\n");
}

void
print_asgs_list (assignments_and_actions lasgs)
{
  printf ("\n-----------------\n");
  printf ("assignments list\n");
  printf ("------------------\n");
  print_condition_or_action (lasgs);
  printf ("\n-----------------------\n");
  printf ("end assignments list\n");
  printf ("-----------------------\n\n");

}

void
print_single_transition (gpointer data, gpointer ud)
{
  single_transition *fp = ((single_transition *) data);

  printf ("$$$$$$$$begin single_transition$$$$$$$$\n");
  printf ("destination automa node= %d\n", fp->automa_node_number);
  if (fp->asgs_list != NULL)
    print_asgs_list (fp->asgs_list);
  else
    printf ("no assignments\n");
  printf ("$$$$$$$$end single_transition$$$$$$$$\n");
}


void
print_transitions_list (transitions_list tl)
{
  printf ("\n----------------------\n");
  printf ("begin transitions_list\n");
  printf ("----------------------\n");
  g_list_foreach (tl, print_single_transition, NULL);
  printf ("--------------------\n");
  printf ("end transitions_list\n");
  printf ("--------------------\n\n");

}


void
print_cond_tree_rec (cond_tree_root ct_node, int level, int branch)
{
  /*branch=0 ==> tree root
     branch=1 ==> C
     branch=2 ==> !C */

  if (ct_node != NULL)
    {
      printf ("-----------node type---------------\n");
      switch (branch)
	{
	case 0:
	  printf ("root node\n");
	  break;
	case 1:
	  printf ("level of father= %d -- branch C\n", level);
	  break;
	case 2:
	  printf ("level of father= %d -- branch !C\n", level);
	  break;
	}

      level++;

      printf ("node level = %d\n", level);

      printf ("-----------end node type----------\n");

      if (ct_node->cond_ptr != NULL)
	print_condition_or_action (ct_node->cond_ptr);
      else
	printf ("no conditions\n");
      printf ("negated cond = %d\n", ct_node->negated_cond);
      if (ct_node->ev_set != NULL)
	print_ev_set (ct_node->ev_set);
      else
	printf ("no ev_set\n");

      if (ct_node->trans_l != NULL)
	print_transitions_list (ct_node->trans_l);
      else
	printf ("no transitions\n");
      print_cond_tree_rec (ct_node->cond_pos, level, 1);
      print_cond_tree_rec (ct_node->cond_neg, level, 2);
    }
}


void
print_edge (gpointer edg_key, gpointer edg_ptr, gpointer ud)
{
  printf ("\n-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-\n");
  printf ("syscall num.= %d\n", ((edge_t *) edg_ptr)->key.syscall_num);
  printf ("context: ");
  if (((edge_t *) edg_ptr)->key.context == intro)
    printf ("intro\n");
  else
    printf ("exiting\n");
  printf ("\n-------------------------\n");
  printf ("associated condition tree\n");
  printf ("-------------------------\n");
  print_cond_tree_rec (((edge_t *) edg_ptr)->ct_root, -1, 0);
  printf ("-----------------------------\n");
  printf ("end associated condition tree\n");
  printf ("-----------------------------\n\n");

}

void
print_edges_set (edges_set_t edg_set)
{
  g_hash_table_foreach (edg_set, print_edge, NULL);
}



void
print_automa_node (automa_node * nd_ptr)
{
  if (nd_ptr->is_final)
    printf ("\nnodo finale\n");
  printf ("\n*****************************\n");
  printf ("**ev_set associated to node**\n");
  printf ("*****************************\n");
  if (nd_ptr->ev_set_ptr)
    print_ev_set (nd_ptr->ev_set_ptr);
  else
    printf ("empty\n");
  printf ("*********************\n");
  printf ("**edge set of node **\n");
  printf ("*********************\n");
  print_edges_set (nd_ptr->edg_set);
}

void
print_automa (automa atm)
{
  //int i;
  //automa_node* current_node_ptr;

  printf ("\nautoma contains %d elements\n\n", atm->len);

  //to print the entire automa uncomment the following lines
/*    for (i=0;i<atm->len;i++) */
/*      { */
/*        current_node_ptr=&g_array_index(atm,automa_node,i); */
/*        printf("\n\n-------------------------------------------------------\n"); */
/*        printf("--------------           node %-3d            ----------\n",i); */
/*        printf("-------------------------------------------------------\n"); */
/*        print_automa_node(current_node_ptr); */
/*      } */
}

/**************************************************************************
 * DEBUG FUNCTIONS 2: functions used for printing the subsequences ranges *
 **************************************************************************/

void
print_subseqs_ll (begin_of_subseqs_ll ll)
{
  GList *temp_list = g_list_first (ll);
  GList *temp_el;
  int i = 1;

  while (temp_list != NULL)
    {
      printf ("list number %d:\t", i);
      temp_el = g_list_first (((GList *) temp_list->data));

      while (temp_el != NULL)
	{
	  printf ("%d ", *((int *) temp_el->data));
	  temp_el = g_list_next (temp_el);
	}

      printf ("\n");
      i++;
      temp_list = g_list_next (temp_list);
    }
}


/*
int 
create_hash_ev_set (ev_set_p)
event_set_t* ev_set_p;
{
  int i,last_num,hashv;

  hashv=0;
  if (ev_set_p)
    {
      last_num=ev_set_p->max_events;
      for ( i=1 ; i<=last_num ; i++ )
	if ( lookup_event(ev_set_p,i) )
	  hashv+=i;
    }
  return hashv;
}
*/




/*
  equal_ev_sets

  input:  2 event sets
          
  output: TRUE if equal 
          FALSE otherwise

  compares 2 event sets
*/
/*int 
equal_ev_sets(set1, set2)
event_set_t* set1;
event_set_t* set2;
{
  int event;
  int result=(events_in_set(set1)==events_in_set(set2));

  first_event(set1,event);

  while (result && (event != -1 ))
    {
      result=lookup_event(set2,event);
      event=each_event(set1);
    }

  return result;
}
*/
