#ifndef AUTOMA_H
#define AUTOMA_H

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: automa.h,v 1.2 2003/02/20 16:19:39 frassi Exp $
 */


#include "parser.h"
#include "event_set.h"

#define NO_CONFLICT  1 /* defined if I'm sure that spec file doesn't contain conflicts */

/*------------------
definizioni generali
--------------------*/

//#define END_MARKER -10 /*valore momentaneo, da stabilire successivamente*/
#define MAX_NUM_SYSCALLS 500	/*serve per calcolare l'hash di un edge_t */
#define IS_END_MARKER(node) ((((event_node_t*) node->data)->simple_event)->n_sysc==-3)
#define NODE_NUM_OF_SYSCALL(node) (((event_node_t*) node->data)->simple_event)->n_sysc
#define IS_EMPTY_FOLLOW(node) ((((event_node_t*) node->data)->simple_event)->empty_follow)
#define NODE_OP(node) ((event_node_t*) node->data)->op
#define NODE_NUM_EVENT(node) ((((event_node_t*) node->data)->simple_event)->num_event)
#define ONLY_CHILD(p) ((p==g_node_first_sibling(p)) && (p==g_node_last_sibling(p)))
#define NODE_CONTEXT(node) (((event_node_t*) node->data)->simple_event)->context


/*----------------------------------
  hash table contenente i numerini
----------------------------------*/
/*typedef struct {
  int event_number;
  event_node_t* event_node_ptr;
} event_t;
 
typedef GHashTable* events_set_t;

#define EV_SET_INIT() g_hash_table_new(g_int_hash,g_int_equal)*/
/*es.: events_set_t ev_s=EV_SET_INIT();
  restituisce una struttura di tipo events_set_t*/

//#define EV_SET_INSERT(ev_set,ev) g_hash_table_insert(ev_set,&(ev->event_number),ev)
/*inserisce l'indirizzo di num nella tabella hash
  NON MODIFICARE L'AREA DI MEMORIA DOPO L'INSERIMENTO
  
  input: una struttura di tipo events_set_t,una struttura di tipo event_t
  output:void*/

/* ^^^ non usato*/

//#define EV_SET_DIMENSION(ev_set) g_hash_table_size(ev_set)
/* ^^^ non usato */

//void ev_set_copy (gpointer key,gpointer value,gpointer destination_ev_set);

/*-------------------
  events_sorted_set_t
---------------------*/
/*
typedef struct {
  event_t* event_t_array;
  int position;
  int array_dim;
} events_sorted_set_t;

void free_events_sorted_set(events_sorted_set_t* ev_sort_set_ptr);
void copy_event (gpointer key,gpointer value,gpointer copy);
events_sorted_set_t* copy_events_to_sorted_set(events_set_t ev_set);
*/
/*---------------------
  assegnamenti - azioni
-----------------------*/

typedef GList *assignments_and_actions;


/*-----------
  transizioni
-------------*/
typedef struct
{
  assignments_and_actions asgs_list;
  //events_set_t ev_set;
  int automa_node_number;	/*numero_nodo_automa_in_cui_mi_sposto_dopo_assegnamenti_e_azioni */
}
single_transition;

typedef GList *transitions_list;


/*------------------------------------------------
    i nodi  dell'albero binario delle condizioni
-------------------------------------------------*/

typedef GList *conditions_list;

struct condition_tree_node
{
  conditions_list cond_ptr;	/*condizione legata al nodo */
  int negated_cond;		/* se e' == 1 ==> bisogna considerare l'intera lista di condizioni puntata da cond */
  /* ha senso solo se cond e' !=NULL, viceversa puo' avere un valore qualunque */
  event_set_t *ev_set;
  struct condition_tree_node *cond_pos;	/*  C  */
  struct condition_tree_node *cond_neg;	/* !C  */
  transitions_list trans_l;	/*e' null se il matching non e' stato completato.
				   puo' essere costituita da un'unica single_transition, se le azioni
				   e gli assegnamenti non confliggono. Se invece c'e'
				   conflitto, e' necessario duplicare (anche + volte) lo 
				   stato dell'automa e seguire diversi percorsi
				   contemporaneamente. In quest'ultimo caso c'e' una vera
				   e propria lista di transizioni. */
};


typedef struct condition_tree_node *cond_tree_root;



/*---------------------------------------------
  tabella hash contenente le syscalls che 
  etichettano gli archi uscenti dai nodi dell'automa
----------------------------------------------*/
typedef struct
{
  int syscall_num;
  context_t context;
}
edge_key_t;


typedef struct
{
  edge_key_t key;
  cond_tree_root ct_root;
}
edge_t;

typedef GHashTable *edges_set_t;

guint g_edge_hash (gconstpointer k1);
gint g_edge_equal (gconstpointer k1, gconstpointer k2);

/*------------------ 
  i nodi dell'automa
  ------------------*/

typedef struct
{
  edges_set_t edg_set;
  event_set_t *ev_set_ptr;
  //int dim_ev_set;
  //unsigned int hash_ev_set;
  int is_final;
}
automa_node;

#define AUTOMA_NODE_INIT(an) \
{ \
    an=(automa_node*) malloc(sizeof(automa_node)); \
    an->edg_set=g_hash_table_new(g_edge_hash,g_edge_equal); \
    an->is_final=0; \
}
// an->dim_ev_set=0;}
/*alloca la memoria per una struttura di tipo automa_node.
  Viene allocata la memoria anche per l'hash table degli archi.
  s e' un puntatore a una struttura di tipo automa_node
  
  es.:   automa_node* pippo;
         AUTOMA_NODE_INIT(pippo);
*/


/*-------- 
  l'automa
  --------*/

typedef GArray *automa;

#define AUTOMA_INIT(z,c) g_array_new(z,c,sizeof(automa_node))
/*z(=zero_terminated) e c(=clear) sono dei gboolean
  se z=TRUE ==> alla fine dell'array viene aggiunto un elemento extra
                che e' messo a 0
  se c=TRUE ==> gli elementi dell'array vengono messi a 0 non appena
  sono allocati

  restituisce un automa.
  
  es.:   automa pippo=AUTOMA_INIT(FALSE,FALSE);
*/

/*g_array_append_val(a,s)*/
/*inserisce un nodo in coda all'automa
  a=l'automa al quale si vuole aggiungere un nodo
  s=il nodo da aggiungere

  restituisce un automa.
  
  es.:   automa pippo;
         automa_node pluto;
	 pippo=g_array_append_val(pippo,pluto);
*/

/*g_array_index(a,automa_node,pos)*/
/*a e' una variabile di tipo automa
  pos e' la posizione dell'automa cui si vuole accedere

  restituisce la struttura di tipo nodo contenuta nella posizione
  pos dell'automa.
  
  es.:  automa pippo;
        automa_node pluto;
	automa_node *minni;
	pluto=g_array_index(pippo,automa_node,3); --pluto prende il terzo elemento
					      dell'automa--
        minni=&g_array_index(pippo,automa_node,5) --minni punta al quinto elemento
	                                      dell'automa--
*/

/*#define AUTOMA_DISTRUGGI(a,fs) g_array_free(a,fs)*/
/*a e' la variabile di tipo automa che si vuole deallocare
  fs e' un gboolean
  se fs=TRUE ==> viene cancellato anche l'elemento contenuto in testa
                 all'array(?)

  non restituisce niente.
*/

/*----------
array follow
------------*/

#define COMPLETE_TREE 1
#define SUBTREE 2

typedef struct
{
  int num_event;
  context_t context;
}
event_id;

typedef struct
{				/*elemento dell'array di follow */
  event_id evid;
  event_node_t *ev_nod_ptr;
  event_set_t *ev_set_follow_ptr;
  //int dim_ev_set_follow;
  //unsigned int hash_ev_set_follow;
  gboolean is_endmarker;
}
follow_array_element_t;

typedef struct
{
  int tree_t;			/*puo' essere COMPLETE_TREE oppure SUBTREE */
  int counter;
  int dim_ar;
  follow_array_element_t *follow_array;
}
follow_array_t;


#define get_negated_condition(arf,event) arf.follow_array[event].ev_nod_ptr->simple_event->fl.negated_condition
#define get_cond_ptr(arf,event)          arf.follow_array[event].ev_nod_ptr->simple_event->condition
#define get_action_ptr(arf,event)        arf.follow_array[event].ev_nod_ptr->simple_event->action


/*--------------------
user data per AQE_gen
----------------------*/

struct userd
{
  follow_array_t arf;
  event_set_t *ev_set_ptr;
  automa atm;
};



/*----------------------------------------
  struttura dati per calcolo sottosequenze
  ----------------------------------------*/

typedef GList *begin_of_subseqs_ll;
/*ll=list of lists*/

#define get_sublist(subseq_ll,num_l) ((GList*) g_list_nth_data(subseq_ll,num_l))


void print_subseqs_ll (begin_of_subseqs_ll ll);
begin_of_subseqs_ll compute_subranges_rec (syntax_tree node,
					   begin_of_subseqs_ll ll);


/*-------------------------------------
  prototipi funzioni costruzione automa
  -------------------------------------*/

automa AQE_gen (syntax_tree root);
follow_array_t create_follow_array_of_tree (syntax_tree root, int tree_t);


/* le funzioni indicate con ^ vengono invocate solo all'interno di automa.c */

//gboolean empty_expr(syntax_tree st); /*^*/
//void first_ev_numbers_of_expr(syntax_tree st,event_set_t* ev_set_p); /*^*/
//void following_ev_numbers_of_expr(syntax_tree st,event_set_t* ev_set_p,follow_array_t arf); /*^*/
//int create_hash_ev_set (event_set_t* ev_set_p); /*^*/
//gboolean syntax_tree_visit(GNode *nd,gpointer arf); /*^*/
//void build_root_cond_tree(int ev_num, event_node_t* ev_node_ptr, edges_set_t edg_set); /*^*/
//gboolean asgs_conflict(GList* asgs_list,GList* asg_to_check); /*^*/
//gboolean erase_ev_set_el(gpointer ev_num_ptr,gpointer ev_node_ptr,gpointer ud); /*^*/
//void free_ev_set_rec(struct condition_tree_node* ct_node); /*^*/
//void bct_rec (int event, follow_array_t arf, event_set_t* ev_set, cond_tree_root root, automa atm); /*^*/
//void init_bct (gpointer key_ptr,gpointer edge_ptr,gpointer ud); /*^*/
//void update_final_states(automa atm,follow_array_t arf); /*^*/


/*------------------------------
prototipi funzioni stampa automa
--------------------------------*/

void print_follow_array (follow_array_t arf);	/* TODO:forse doppione di print_all_follow di Davide, eliminare */
void print_ev_set (event_set_t * ev_set);
void print_automa (automa atm);

/* ^ = usati solo all'interno di funzioni_print_automa.c */
//void print_ev_set_el(gpointer key_ptr,gpointer ev_node_ptr,gpointer ud); /*^*/
void print_asgs_list (assignments_and_actions lasgs);	/*^ */
void print_single_transition (gpointer data, gpointer ud);	/*^ */
void print_transitions_list (transitions_list tl);	/*^ */
void print_cond_tree_rec (cond_tree_root ct_node, int level, int branch);	/*^ */
void print_edge (gpointer edg_key, gpointer edg_ptr, gpointer ud);	/*^ */
void print_edges_set (edges_set_t edg_set);	/*^ */
void print_automa_node (automa_node * nd_ptr);	/*^ */

#endif /* !AUTOMA_H */
