#ifndef GESTIONE_AUTOMA_H
#define GESTIONE_AUTOMA_H 

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: automa_management.h,v 1.3 2003/02/17 15:41:54 frassi Exp $
 */


#include "defs.h"
#include "automa.h"


/*********************
 * TIPO AUTOMA STATE *
 *********************/

typedef GList* automa_state; 


/***********************
 * Trace Control Block *
 ***********************/

typedef enum { EMPTY = 0, UPEEK = 1, TAKEN = 2 } arg_state_t; 

struct tcb 
{
  short flags;		/* See below for TCB_ values */
  int pid;		/* Process Id of this entry */
  long scno;		/* System call number */
  int u_nargs;		/* System call arguments number */
  long u_arg[MAX_ARGS];	/* syscall arguments */
  arg_state_t arg_state[MAX_ARGS + 1]; /* return value (arg_state[rval]) and  arguments state */
  int u_error;		/* Error code */
  long u_rval;		/* (first) return value */
  FILE *outf;		/* Output file for this process */
  char *auxstr;		/* Auxiliary info from syscall (see RVAL_STR) */
  
  //automa atm;
  automa_state process_state;   /* stato del processo all'interno dell'automa */

			/* Support for tracing forked processes */
  struct tcb *parent;	/* Parent of this process */
  int nchildren;	/* # of traced children */
  int waitpid;		/* pid(s) this process is waiting for */
			/* (1st arg of wait4()) */
  long baddr;		/* `Breakpoint' address */
  long inst[2];		/* Instructions on above */

  //event_set_t *ev_set;
  //int send_anyway;
};

/* TCB flags for flags field */
#define TCB_STARTUP	00001	/* We have just begun ptracing this process */
/* #define TCB_INUSE	00002 */	/* This table entry is in use */
#define TCB_INSYSCALL	00004	/* A system call is in progress */
#define TCB_ATTACHED	00010	/* Process is not our own child */
#define TCB_EXITING	00020	/* As far as we know, this process is exiting */
#define TCB_SUSPENDED	00040	/* Process has done a wait(4), that can
				   not be allowed to complete just now */
#define TCB_BPTSET	00100	/* "Breakpoint" set after fork(2) */
#define TCB_SIGTRAPPED	00200	/* Process wanted to block SIGTRAP */
#define TCB_FOLLOWFORK	00400	/* Process should have forks followed */
/* #define TCB_REPRINT	01000 */	/* We should reprint this syscall on exit */

#define TCB_STOPPED     02000
#define TCB_KILLING     04000



#define entering(tcp)	(!((tcp)->flags & TCB_INSYSCALL))
#define exiting(tcp)	((tcp)->flags & TCB_INSYSCALL)
#define syserror(tcp)	((tcp)->u_error != 0)
#define verbose(tcp)	1 // stampa tutti i parametri struct
#define abbrev(tcp)	1 // abbrevia le struct lunghe
#define waiting_parent(tcp) \
		(tcp->parent && \
		(tcp->parent->flags & TCB_SUSPENDED) && \
		(tcp->parent->waitpid <= 0 || tcp->parent->waitpid == tcp->pid))

extern struct tcb *alloctcb(int);
extern void droptcb(struct tcb *);

//extern struct tcb tcbtab[];
extern GHashTable *tcbtab;
extern int nprocs;
extern struct tcb *tcp_last;




/****************
 * AUTOMA STATE *
 ****************/


typedef struct {
  int node_number;
  var_table_t tab_var;
} automa_state_el;


// le strutture di tipo automa_state_el vengono inserite in una GList.
// questa lista contiene normalmente un unico stato_automa, solo nel
// caso di conflitti, la lista ne conterra' + di uno

automa_state init_automa_state(automa_state csl,var_table_t tab_var);
GList* update_state(automa atm,struct tcb* tcb1,automa_state current_states_list);

automa_state *clone_automa_state(automa_state as);
void free_atm_state(automa_state process_state);

/* usate solo all'interno di gestione_automa.c */
void free_state_el(automa_state_el* s); /*^*/
var_table_t update_tab_var(GList* asg_list,var_table_t tabv,struct tcb* tcb1); /*^*/
GList* search_state(automa_state_el* s,automa_state cs_list); /*^*/


/******funzioni stampa**********/
void print_cs_list(automa_state csl);

/* usata solo all'interno di gestione_automa.c */
void print_current_state(gpointer acs,gpointer ud);



/***********************
 * PROCESS STATE TABLE *
 ***********************/

typedef GHashTable * process_state_table_t;

#define new_process_state_table()     (g_hash_table_new(g_int_hash, g_int_equal))
/* crea una tabella per gli stati relativi a ciascun processo */

int insert_process_state(process_state_table_t process_state_table, int pid, automa_state process_state);
/* inserisce un stato dell'automa nella tabella degli stati relativi a ciascun processo, 
   e restituisce 0 se a successo, -1 altrimenti */

#define get_process_state(process_state_table, pid)        \
           ((automa_state *) g_hash_table_lookup((process_state_table), (&(pid))))
/* restituisce un puntatore allo stato dell'automa per il processo identificato da pid */

//void free_process_state(process_state_table_t process_state_table, int pid);
/* libera lo stato dell'automa per il processo identificato da pid */				 



#endif /* !GESTIONE_AUTOMA_H */
