/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: bjm.c,v 1.2 2003/02/17 15:30:02 frassi Exp $
 */

#include "gets.h" /* get_error, get_str, get_path */
#include "util.h" /* arg_alloc, set_copy, RVAL_HEX */
#include "runtime_util.h" /* take_arg */
#include "pars_funct.h" /* get_nargs */
 
/* #include <linux/module.h> /\* QM_INFO, QM_MODULES,... *\/ */


/* static struct xlat which[] = { */
/*   { 0,		"0"		}, */
/*   { QM_MODULES,	"QM_MODULES"	}, */
/*   { QM_DEPS,	"QM_DEPS"	}, */
/*   { QM_REFS,	"QM_REFS"	}, */
/*   { QM_SYMBOLS,	"QM_SYMBOLS"	}, */
/*   { QM_INFO,	"QM_INFO"	}, */
/*   { 0,		NULL		}, */
/* }; */



/* static struct xlat modflags[] = { */
/*   { MOD_UNINITIALIZED,	"MOD_UNINITIALIZED"	}, */
/*   { MOD_RUNNING,	"MOD_RUNNING"		}, */
/*   { MOD_DELETED,	"MOD_DELETED"		}, */
/*   { MOD_AUTOCLEAN,	"MOD_AUTOCLEAN"		}, */
/*   { MOD_VISITED,	"MOD_VISITED"		}, */
/*   { MOD_USED_ONCE,	"MOD_USED_ONCE"		}, */
/*   { MOD_JUST_FREED,	"MOD_JUST_FREED"	}, */
/*   { 0,			NULL			}, */
/* }; */



#include <fcntl.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/utsname.h>
#include <sys/user.h>
#include <sys/syscall.h>
#include <signal.h>

/* Bits of module.flags.  */

#define MOD_UNINITIALIZED	0
#define MOD_RUNNING		1
#define MOD_DELETED		2
#define MOD_AUTOCLEAN		4
#define MOD_VISITED  		8
#define MOD_USED_ONCE		16
#define MOD_JUST_FREED		32
#define MOD_INITIALIZING	64

/* Values for query_module's which.  */

#define QM_MODULES	1
#define QM_DEPS		2
#define QM_REFS		3
#define QM_SYMBOLS	4
#define QM_INFO		5

struct module_symbol
{
	unsigned long value;
	const char *name;
};

struct module_info
{
	unsigned long addr;
	unsigned long size;
	unsigned long flags;
	long usecount;
};

static struct xlat which[] = {
	{ 0,		"0"		},
	{ QM_MODULES,	"QM_MODULES"	},
	{ QM_DEPS,	"QM_DEPS"	},
	{ QM_REFS,	"QM_REFS"	},
	{ QM_SYMBOLS,	"QM_SYMBOLS"	},
	{ QM_INFO,	"QM_INFO"	},
	{ 0,		NULL		},
};

static struct xlat modflags[] = {
	{ MOD_UNINITIALIZED,	"MOD_UNINITIALIZED"	},
	{ MOD_RUNNING,		"MOD_RUNNING"		},
	{ MOD_DELETED,		"MOD_DELETED"		},
	{ MOD_AUTOCLEAN,	"MOD_AUTOCLEAN"		},
	{ MOD_VISITED,		"MOD_VISITED"		},
	{ MOD_USED_ONCE,	"MOD_USED_ONCE"		},
	{ MOD_JUST_FREED,	"MOD_JUST_FREED"	},
	{ 0,			NULL			},
};


int 
print_query_module(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("query_module(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    realprintstr((char*) info->u_arg[0],-1);
	    break;
	  case 1:	
	    printxval(which, *((int*)info->u_arg[1]), "QM_???");
	    break;
	  case 2:
	    if (info->u_arg[1]==QM_INFO) 
	      {
		struct module_info* mi=((struct module_info*) info->u_arg[2]);
		if (mi)
		  {
		    tprintf("{address=%#lx, size=%lu, flags=", mi->addr, mi->size);
		    printflags(modflags, mi->flags);
		    tprintf(", usecount=%lu}", mi->usecount);
		  }
		else
		  tprintf("{NULL}");
	      } 
	    else 
	      if ((info->u_arg[1]==QM_MODULES) || (info->u_arg[1]==QM_DEPS) || (info->u_arg[1]==QM_REFS)) 
		{
		  size_t ret = *((size_t*)info->u_arg[4]);
		  char*	data = (char*)info->u_arg[2];
		  char*	mod = data;
		  size_t idx;
		  
		  tprintf("{");
		  
		  for (idx=0; idx<ret; idx++) 
		    {
		      if (idx!=0)
			tprintf(",");
		      tprintf(mod);
		      mod+=strlen(mod)+1;
		    }
		  tprintf("}, %d", ret);
		} 
	      else
		if (info->u_arg[1]==QM_SYMBOLS) 
		  {
		    size_t ret= *((size_t*)info->u_arg[4]);
		    char*  data	= (char*)info->u_arg[2];
		    struct module_symbol* sym = (struct module_symbol*)data;
		    size_t idx;
		    
		    tprintf("{");
		    
		    for (idx=0; idx<ret; idx++) 
		      {
			tprintf("{name=%s, value=%lu} ", data+(long)sym->name, sym->value);
			sym++;
		      }
		    
		    tprintf("}, %d", ret);
		  } 
		else
		  realprintstr((char*)info->u_arg[2],*((unsigned int*) info->u_arg[3]));
	    break;
	case 3:
	    tprintf("%lu", *((unsigned long*)info->u_arg[3]));
	    break;
	  case 4:
	    tprintf("%d", *((size_t*)info->u_arg[4]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}





int
sys_query_module(tcp,n_arg) /* CONTROLLARE */
struct tcb *tcp;
int n_arg;
{
  int which;
  size_t *retp;
  size_t bufsize;
  char *buf;
  struct module_info *mi;    

  switch(n_arg) {
  case 0: /* const char* */
    ((char *) tcp->u_arg[0]) = get_str(tcp, tcp->u_arg[0], -1);
    if ( get_error )
      {
	fprintf(stderr, "sys_query_module: get_str\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1: /* int */
    arg_alloc(tcp,n_arg);
    break;
  case 2: /* void* */
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_query_module: syscall failed\n");
	    return -1;
	  }
	
	if (tcp->u_rval==0) 
	  {
	    int *ptr;
	    if ((ptr =  (int *) take_arg(1, tcp)) == NULL )
	      {
		fprintf(stderr, "sys_query_module: take_arg(1, tcp)\n");
		return -1;
	      }
	    which = *ptr;
	    
	    if (((size_t *) take_arg(4, tcp)) == NULL)
	      {
		fprintf(stderr, "sys_query_module: take_arg(4, tcp)\n");
		return -1;
	      }

	    if (which == QM_INFO) 
	      {
		if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&mi, sizeof(struct module_info), "query_module"))
		  {
		    fprintf(stderr, "sys_query_module: set_copy\n");
		    return -1;
		  }
	      } 
	    else 
	      {
		size_t *ptr;
	      
		if ((ptr = (size_t *) take_arg(3, tcp)) == NULL)
		  {
		    fprintf(stderr, "sys_query_module: take_arg(3, tcp)\n");
		    return -1;
		  }
		bufsize = *ptr;

		if ((which==QM_MODULES) || (which==QM_DEPS) || (which==QM_REFS) || (which==QM_SYMBOLS)) 
		  {
		    if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&buf, bufsize, "query_module"))
		      {
			fprintf(stderr, "sys_query_module: set_copy\n");
			return -1;
		      }
		  } 
		else 
		  {
		    ((char *) tcp->u_arg[2]) = get_str(tcp, tcp->u_arg[2], bufsize);			
		    if ( get_error ) 
		      {
			fprintf(stderr, "sys_query_module: get_str\n");
			tcp->arg_state[2] = EMPTY;
			return -1;
		      }
		    tcp->arg_state[2]= TAKEN;
		  }
	      }
	  }
      }
    break;
    
  case 3: /* size_t */
    //if (exiting(tcp)) 
      arg_alloc(tcp,n_arg);
    break;
    
  case 4: /* size_t* */
    if (exiting(tcp)) 
      {
	int *ptr;
	
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_query_module: syscall failed\n");
	    return -1;
	  }
	
	if ((ptr = (int *) take_arg(1, tcp)) == NULL)
	  {
	    fprintf(stderr, "sys_query_module: take_arg(1, tcp)\n");
	    return -1;
	  }
	which = *ptr;

	if ((tcp->u_rval!=0) || (which==QM_INFO) || (which==QM_MODULES) || (which==QM_DEPS) || (which==QM_REFS) || (which==QM_SYMBOLS))
	  {
	    if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&retp, sizeof(size_t), "query_module"))
	      {
		fprintf(stderr, "sys_query_module: set_copy\n");
		return -1;
	      }
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  
  return 0;
}




int 
print_create_module(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("create_module(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1: 
	    tprintf("%lu", *((unsigned long*) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return RVAL_HEX;
}



int
sys_create_module(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch(n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_create_module: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN ;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}





int 
print_init_module(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("init_module(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1: 
	    tprintf("%#lx", *((long*) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_init_module(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch(n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error ) 
      {
	fprintf(stderr, "sys_init_module: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1: /* in realta' il secondo argomento e' moooooolto + complicato */
    arg_alloc(tcp,n_arg);
    break;
  }

  return 0;
}
