/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: desc.c,v 1.2 2003/02/17 15:30:18 frassi Exp $
 */

#include "gets.h" /* get_error */
#include "util.h" /* umove, arg_alloc */
#include "runtime_util.h" /* take_arg */
#include "pars_funct.h" /* get_nargs */

/* static struct xlat fcntlcmds[] = { */
/*   { F_DUPFD,	"F_DUPFD"	}, */
/*   { F_GETFD,	"F_GETFD"	}, */
/*   { F_SETFD,	"F_SETFD"	}, */
/*   { F_GETFL,	"F_GETFL"	}, */
/*   { F_SETFL,	"F_SETFL"	}, */
/*   { F_GETLK,	"F_GETLK"	}, */
/*   { F_SETLK,	"F_SETLK"	}, */
/*   { F_SETLKW,	"F_SETLKW"	}, */
/*   { F_GETOWN,	"F_GETOWN"	}, */
/*   { F_SETOWN,	"F_SETOWN"	}, */
/* #ifdef F_RSETLK */
/*   { F_RSETLK,	"F_RSETLK"	}, */
/* #endif */
/* #ifdef F_RSETLKW */
/*   { F_RSETLKW,	"F_RSETLKW"	}, */
/* #endif */
/* #ifdef F_RGETLK */
/*   { F_RGETLK,	"F_RGETLK"	}, */
/* #endif */
/* #ifdef F_CNVT */
/*   { F_CNVT,	"F_CNVT"	}, */
/* #endif */
/*   { 0,		NULL		}, */
/* }; */


/* static struct xlat fdflags[] = { */
/* #ifdef FD_CLOEXEC */
/*   { FD_CLOEXEC,	"FD_CLOEXEC"	}, */
/* #endif */
/*   { 0,		NULL		}, */
/* }; */



/* #ifdef LOCK_SH */

/* static struct xlat flockcmds[] = { */
/*   { LOCK_SH,	"LOCK_SH"	}, */
/*   { LOCK_EX,	"LOCK_EX"	}, */
/*   { LOCK_NB,	"LOCK_NB"	}, */
/*   { LOCK_UN,	"LOCK_UN"	}, */
/*   { 0,		NULL		}, */
/* }; */

/* #endif /\* LOCK_SH *\/ */


/* static struct xlat lockfcmds[] = { */
/*   { F_RDLCK,	"F_RDLCK"	}, */
/*   { F_WRLCK,	"F_WRLCK"	}, */
/*   { F_UNLCK,	"F_UNLCK"	}, */
/* #ifdef F_EXLCK */
/*   { F_EXLCK,	"F_EXLCK"	}, */
/* #endif */
/* #ifdef F_SHLCK */
/*   { F_SHLCK,	"F_SHLCK"	}, */
/* #endif */
/*   { 0,		NULL		}, */
/* }; */



/* static struct xlat whence[] = { */
/*   { SEEK_SET,	"SEEK_SET"	}, */
/*   { SEEK_CUR,	"SEEK_CUR"	}, */
/*   { SEEK_END,	"SEEK_END"	}, */
/*   { 0,		NULL		}, */
/* }; */


#include <fcntl.h>
#include <sys/file.h>



static struct xlat fcntlcmds[] = {
	{ F_DUPFD,	"F_DUPFD"	},
	{ F_GETFD,	"F_GETFD"	},
	{ F_SETFD,	"F_SETFD"	},
	{ F_GETFL,	"F_GETFL"	},
	{ F_SETFL,	"F_SETFL"	},
	{ F_GETLK,	"F_GETLK"	},
	{ F_SETLK,	"F_SETLK"	},
	{ F_SETLKW,	"F_SETLKW"	},
	{ F_GETOWN,	"F_GETOWN"	},
	{ F_SETOWN,	"F_SETOWN"	},
#ifdef F_RSETLK
	{ F_RSETLK,	"F_RSETLK"	},
#endif
#ifdef F_RSETLKW
	{ F_RSETLKW,	"F_RSETLKW"	},
#endif
#ifdef F_RGETLK
	{ F_RGETLK,	"F_RGETLK"	},
#endif
#ifdef F_CNVT
	{ F_CNVT,	"F_CNVT"	},
#endif
#ifdef F_SETSIG
	{ F_SETSIG,	"F_SETSIG"	},
#endif
#ifdef F_GETSIG
	{ F_GETSIG,	"F_GETSIG"	},
#endif
#ifdef F_CHKFL
	{ F_CHKFL,	"F_CHKFL"	},
#endif
#ifdef F_DUP2FD
	{ F_DUP2FD,	"F_DUP2FD"	},
#endif
#ifdef F_ALLOCSP
	{ F_ALLOCSP,	"F_ALLOCSP"	},
#endif
#ifdef F_ISSTREAM
	{ F_ISSTREAM,	"F_ISSTREAM"	},
#endif
#ifdef F_PRIV
	{ F_PRIV,	"F_PRIV"	},
#endif
#ifdef F_NPRIV
	{ F_NPRIV,	"F_NPRIV"	},
#endif
#ifdef F_QUOTACL
	{ F_QUOTACL,	"F_QUOTACL"	},
#endif
#ifdef F_BLOCKS
	{ F_BLOCKS,	"F_BLOCKS"	},
#endif
#ifdef F_BLKSIZE
	{ F_BLKSIZE,	"F_BLKSIZE"	},
#endif
#ifdef F_GETOWN
	{ F_GETOWN,	"F_GETOWN"	},
#endif
#ifdef F_SETOWN
	{ F_SETOWN,	"F_SETOWN"	},
#endif
#ifdef F_REVOKE
	{ F_REVOKE,	"F_REVOKE"	},
#endif
#ifdef F_SETLK
	{ F_SETLK,	"F_SETLK"	},
#endif
#ifdef F_SETLKW
	{ F_SETLKW,	"F_SETLKW"	},
#endif
#ifdef F_FREESP
	{ F_FREESP,	"F_FREESP"	},
#endif
#ifdef F_GETLK
	{ F_GETLK,	"F_GETLK"	},
#endif
#ifdef F_SETLK64
	{ F_SETLK64,	"F_SETLK64"	},
#endif
#ifdef F_SETLKW64
	{ F_SETLKW64,	"F_SETLKW64"	},
#endif
#ifdef F_FREESP64
	{ F_FREESP64,	"F_FREESP64"	},
#endif
#ifdef F_GETLK64
	{ F_GETLK64,	"F_GETLK64"	},
#endif
#ifdef F_SHARE
	{ F_SHARE,	"F_SHARE"	},
#endif
#ifdef F_UNSHARE
	{ F_UNSHARE,	"F_UNSHARE"	},
#endif
	{ 0,		NULL		},
};

static struct xlat fdflags[] = {
#ifdef FD_CLOEXEC
	{ FD_CLOEXEC,	"FD_CLOEXEC"	},
#endif
	{ 0,		NULL		},
};

#ifdef LOCK_SH

static struct xlat flockcmds[] = {
	{ LOCK_SH,	"LOCK_SH"	},
	{ LOCK_EX,	"LOCK_EX"	},
	{ LOCK_NB,	"LOCK_NB"	},
	{ LOCK_UN,	"LOCK_UN"	},
	{ 0,		NULL		},
};

#endif /* LOCK_SH */

static struct xlat lockfcmds[] = {
	{ F_RDLCK,	"F_RDLCK"	},
	{ F_WRLCK,	"F_WRLCK"	},
	{ F_UNLCK,	"F_UNLCK"	},
#ifdef F_EXLCK
	{ F_EXLCK,	"F_EXLCK"	},
#endif
#ifdef F_SHLCK
	{ F_SHLCK,	"F_SHLCK"	},
#endif
	{ 0,		NULL		},
};

static struct xlat whence[] = {
	{ SEEK_SET,	"SEEK_SET"	},
	{ SEEK_CUR,	"SEEK_CUR"	},
	{ SEEK_END,	"SEEK_END"	},
	{ 0,		NULL		},
};


/* fcntl/lockf */
static void
realprintflock(fl)
struct flock * fl;
{
  if (fl == NULL)
    {
      tprintf("{NULL}");
      return;
    }

  tprintf("{type=");
  printxval(lockfcmds, fl->l_type, "F_???");
  tprintf(", whence=");
  printxval(whence, fl->l_whence, "SEEK_???");
  tprintf(", start=%ld, len=%ld", fl->l_start, fl->l_len);
  tprintf(", pid=%lu}", (unsigned long) fl->l_pid);
}



#if _LFS64_LARGEFILE || HAVE_LONG_LONG_OFF_T

static void
realprintflock64(fl)
struct flock64 *fl;
{
	
  if (fl == NULL)
    {
      tprintf("{NULL}");
      return;
    }

  tprintf("{type=");
  printxval(lockfcmds, fl->l_type, "F_???");
  tprintf(", whence=");
  printxval(whence, fl->l_whence, "SEEK_???");
  tprintf(", start=%lld, len=%lld", (long long) fl->l_start, (long long) fl->l_len);
  tprintf(", pid=%lu}", (unsigned long) fl->l_pid);
  
}

struct flock64*
get_flock64(tcp, addr)
struct tcb *tcp;
long addr;
{
  struct flock64* fl;

  get_error = 0;

  if (!addr)
    return NULL;

  fl=((struct flock64*) malloc(sizeof(struct flock64)));

  if (!fl)
    {
      fprintf(stderr, "get_flock: malloc flock failed\n");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, fl) < 0) 
    {
      free(fl);
      fprintf(stderr, "get_flock: umove flock failed\n");
      get_error = 1;
      return NULL;
    }

  return fl;
}

#endif


/* fcntl/lockf */
struct flock*
get_flock(tcp, addr)
struct tcb *tcp;
long addr;
{
  struct flock* fl;

  get_error = 0;

  if (!addr)
    return NULL;

  fl=((struct flock*) malloc(sizeof(struct flock)));

  if (!fl)
    {
      fprintf(stderr, "get_flock: malloc flock failed\n");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, fl) < 0) 
    {
      free(fl);
      fprintf(stderr, "get_flock: umove flock failed\n");
      get_error = 1;
      return NULL;
    }

  return fl;
}




char *
sprintflags(xlat, flags)
struct xlat *xlat;
int flags;
{
  char *outstr=malloc(1024*sizeof(char));
  char *sep;
  
  if (outstr == NULL)
    return NULL;

  strcpy(outstr, "flags ");
  sep = "";
  for (; xlat->str; xlat++) 
    {
      if ((flags & xlat->val) == xlat->val) 
	{
	  sprintf(outstr + strlen(outstr), "%s%s", sep, xlat->str);
	  sep = "|";
	  flags &= ~xlat->val;
	}
    }
  
  if (flags)
    sprintf(outstr + strlen(outstr), "%s%#x", sep, flags);
  return outstr;
}



int 
print_fcntl(info)
msg_building_info *info;
{
  extern struct xlat openmodes[];

  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("fcntl(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*) info->u_arg[0])); 
	    break;
	  case 1:
	    printxval(fcntlcmds, *((int*) info->u_arg[1]), "F_???");
	    break;
	  case 2:
	    switch (info->u_arg[1]) {
	    case F_DUPFD:
	    case F_SETFD: case F_SETFL:
	    case F_SETLK: case F_SETLKW:
	    case F_SETOWN: case F_GETOWN:
	      break;
	    case F_GETFD:
	      if (info->u_rval == 0)
		return 0;
	      info->auxstr = sprintflags(fdflags, info->u_rval);
	      return RVAL_HEX|RVAL_STR;
	    case F_GETFL:
	      info->auxstr = sprintflags(openmodes, info->u_rval + 1);
	      return RVAL_HEX|RVAL_STR;
	    case F_GETLK:
	      realprintflock(((struct flock *) info->u_arg[2]));
	      break;
	      /**/
#if _LFS64_LARGEFILE
#if defined(F_GETLK64) && F_GETLK64+0!=F_GETLK
	    case F_GETLK64:
#endif
	      tprintf(", ");
	      realprintflock64(((struct flock64 *) info->u_arg[2]));
	      break;
#endif
	      /**/
	    default:
	      tprintf("_");
	      break;
	    }
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_fcntl(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  int flags, *ptr;

  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;
    
  case 2:
    /*struct flock* o long a seconda dei casi*/
    if ((ptr = (int *) take_arg(1, tcp)) == NULL)
      {
	fprintf(stderr, "sys_fcntl: take_arg(1, tcp)\n");
	return -1;
      }
    flags = *ptr;

    if (entering(tcp))
      {
	switch (flags) {
	case F_SETFD:
	case F_SETOWN: 
	case F_DUPFD: /*valore significativo*/
	case F_SETFL: /*valore significativo*/
	  arg_alloc(tcp,n_arg);
	  break;
	case F_SETLK: 
	case F_SETLKW: 
	  ((struct flock*) tcp->u_arg[2]) = get_flock(tcp, tcp->u_arg[2]);
	  if ( get_error )
	    {
	      fprintf(stderr, "sys_fcntl: get_flock\n");
	      tcp->arg_state[n_arg] = EMPTY;
	      return -1;
	    }
	  tcp->arg_state[2] = TAKEN;
	  break;
	  /**/
#if _LFS64_LARGEFILE
	  
#if defined(F_SETLK64) && F_SETLK64+0!=F_SETLK
	case F_SETLK64:
#endif
#if defined(F_SETLKW64) && F_SETLKW64+0!=F_SETLKW
	case F_SETLKW64:
#endif
	  ((struct flock64*) tcp->u_arg[2]) = get_flock64(tcp, tcp->u_arg[2]);
	  if ( get_error )
	    {
	      fprintf(stderr, "sys_fcntl: get_flock64\n");
	      tcp->arg_state[n_arg] = EMPTY;
	      return -1;
	    }
	  tcp->arg_state[2] = TAKEN;
	  
	  break;
#endif
	  /**/
	  
	}    
      }
    else /* exiting(tcp) */ 
      {
	if (flags == F_GETLK)
	  {
	    ((struct flock*) tcp->u_arg[2]) = get_flock(tcp, tcp->u_arg[2]);
	    if ( get_error )
	      {
		fprintf(stderr, "sys_fcntl: get_flock\n");
		tcp->arg_state[n_arg] = EMPTY;
		return -1;
	      }
	    tcp->arg_state[2] = TAKEN;
	  }
	/**/
#if _LFS64_LARGEFILE
	else if (flags == F_GETLK64)
	  {
	    ((struct flock64*) tcp->u_arg[2]) = get_flock64(tcp, tcp->u_arg[2]);
	    if ( get_error )
	      {
		fprintf(stderr, "sys_fcntl: get_flock64\n");
		tcp->arg_state[n_arg] = EMPTY;
		return -1;
	      }
	    tcp->arg_state[2] = TAKEN;
	  }
#endif
	/**/

      }
    break;
  }
  return 0;
}


int 
print_flock(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("flock(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    if (!printflags(flockcmds, *((int*)info->u_arg[1])))
	      tprintf("LOCK_???");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}



int
sys_flock(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_close(info)
msg_building_info *info;
{
  tprintf("close(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%ld", *((long*)info->u_arg[0]));
  else
    tprintf("_");
  
  tprintf(") ");
  
  return 0;
}


int
sys_close(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_dup(info)
msg_building_info *info;
{
  tprintf("dup(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%ld", *((long*)info->u_arg[0]));
  else
    tprintf("_");
  
  tprintf(") ");
  
  return 0;
}


int
sys_dup(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_dup2(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("dup2(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	tprintf("%ld", *((long*)info->u_arg[i]));
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}


int
sys_dup2(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_oldselect(info)
msg_building_info *info;
{
  int i,j,nfds;
  int nargs=get_nargs(info->scno);
  char *sep;
  fd_set* fds;
  
  tprintf("oldselect(");
  
  if ((nfds = info->u_rval) == 0)
    {
      tprintf(")");
      info->auxstr = strdup("Timeout");
      return RVAL_STR;
    }
  else
    for (i=0;i<nargs;i++)
      {
	if (info->arg_state[i]==TAKEN)
	  {
	    switch (i) {
	    case 0:
	      tprintf("%ld", *((long*) info->u_arg[i]));
	      break;
	    case 1:
	    case 2:
	    case 3:
	      if (((fd_set*) info->u_arg[i]) == NULL)
		tprintf("NULL");
	      else
		{
		  fds=(fd_set*) info->u_arg[i];
		  nfds=*((long*) info->u_arg[0]);
		  tprintf("[");
		  for (j = 0, sep = ""; j < nfds; j++) 
		    {
		      if (FD_ISSET(j, fds)) 
			{
			  tprintf("%s%u", sep, j);
			  sep = " ";
			}
		    }
		  tprintf("]");
		}
	      break;
	    case 4:
	      if (((struct timeval*) info->u_arg[4]) == NULL)
		tprintf("NULL");
	      else
		{
		  struct timeval *tv=((struct timeval*) info->u_arg[4]);
		  tprintf("{%lu, %lu}", (long) tv->tv_sec, (long) tv->tv_usec);
		}
	      break;
	    }
	  }
	else
	  tprintf("_");
	
	if (i!=nargs-1)
	  tprintf(", ");    
      }
  
  
  tprintf(") ");  
  return 0;
}




/* uguale a print_oldselect */
int 
print_select(info)
msg_building_info *info;
{
  int i,j,nfds;
  int nargs=get_nargs(info->scno);
  char *sep;
  fd_set* fds;
  
  tprintf("select(");
  
  if ((nfds = info->u_rval) == 0)
    { 
      tprintf(")");
      info->auxstr = strdup("Timeout");
      return RVAL_STR;
    }
  else
    for (i=0;i<nargs;i++)
      {
	if (info->arg_state[i]==TAKEN)
	  {
	    switch (i) {
	    case 0:
	      tprintf("%ld", *((long*) info->u_arg[i]));
	      break;
	    case 1:
	    case 2:
	    case 3:
	      if (((fd_set*) info->u_arg[i]) == NULL)
		tprintf("NULL");
	      else
		{
		  nfds=*((long*) info->u_arg[0]);
		  fds=(fd_set*) info->u_arg[i];
		  tprintf("[");
		  for (j = 0, sep = ""; j < nfds; j++) 
		    {
		      if (FD_ISSET(j, fds)) 
			{
			  tprintf("%s%u", sep, j);
			  sep = " ";
			}
		    }
		  tprintf("]");
		}
	      break;
	    case 4:
	      if (((struct timeval*) info->u_arg[4]) == NULL)
		tprintf("NULL");
	      else
		{
		  struct timeval *tv=((struct timeval*) info->u_arg[4]);
		  tprintf("{%lu, %lu}", (long) tv->tv_sec, (long) tv->tv_usec);
		}
	      break;
	    }
	  }
	else
	  tprintf("_");
	
	if (i!=nargs-1)
	  tprintf(", ");    
      }
  tprintf(") ");  
  return 0;
}


int
sys_select(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  fd_set *fds;
  struct timeval *tv;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
  case 2:
  case 3:
    if (tcp->u_arg[n_arg] == 0)
      { 
	tcp->arg_state[n_arg]=TAKEN;
	return 0;
      }

    fds=(fd_set *) malloc(sizeof(fd_set));

    if (fds == NULL)
      {
	fprintf(stderr,"sys_select: malloc fds failed");  
	return -1;
      }

    if (umove(tcp, tcp->u_arg[n_arg], fds) < 0) 
      {
	fprintf(stderr,"sys_select: umove arg %d failed", n_arg);
	free(fds);
	return -1;
      }

    (fd_set*) tcp->u_arg[n_arg]=fds;
    tcp->arg_state[n_arg]=TAKEN;
    break;
  case 4:
    if (!(tcp->u_arg[4]))
      tcp->arg_state[4] = TAKEN;
    else
      {
	tv=(struct timeval *) malloc(sizeof(struct timeval));
	
	if (tv==NULL)
	  {
	    fprintf(stderr,"sys_select: malloc tv failed");  
	    return -1;
	  }
	
	if (umove(tcp, tcp->u_arg[4], tv) < 0)
	  {
	    fprintf(stderr,"sys_select: umove tv failed");
	    free(tv);
	    return -1;
	  }
	
	(struct timeval *) tcp->u_arg[4]=tv;
	tcp->arg_state[4]=TAKEN;
      }
    break;
  }
  return 0; 
}
