/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: file.c,v 1.3 2003/02/17 15:27:47 frassi Exp $
 */


#include <asm/stat.h>

#include "gets.h"  /* get_error */
#include "util.h" /* umove, set_copy */
#include "xlat.h"
#include "runtime_util.h"
#include "pars_funct.h" /* get_nargs */


#include <dirent.h>
#define dirent kernel_dirent
#define dirent64 kernel_dirent64
#include <linux/types.h>
#include <linux/dirent.h>
#undef dirent




#    undef dev_t
#    undef ino_t
#    undef mode_t
#    undef nlink_t
#    undef uid_t
#    undef gid_t
#    undef off_t
#    undef loff_t

#    define dev_t __kernel_dev_t
#    define ino_t __kernel_ino_t
#    define mode_t __kernel_mode_t
#    define nlink_t __kernel_nlink_t
#    define uid_t __kernel_uid_t
#    define gid_t __kernel_gid_t
#    define off_t __kernel_off_t
#    define loff_t __kernel_loff_t

#    include <asm/stat.h>

#    undef dev_t
#    undef ino_t
#    undef mode_t
#    undef nlink_t
#    undef uid_t
#    undef gid_t
#    undef off_t
#    undef loff_t

#    define dev_t dev_t
#    define ino_t ino_t
#    define mode_t mode_t
#    define nlink_t nlink_t
#    define uid_t uid_t
#    define gid_t gid_t
#    define off_t off_t
#    define loff_t loff_t


#  define stat libc_stat
#  define stat64 libc_stat64
#  include <sys/stat.h>
#  undef stat
#  undef stat64



#include <fcntl.h>


#ifdef HAVE_SYS_VFS_H
#include <sys/vfs.h>
#endif



#ifdef MAJOR_IN_SYSMACROS
#include <sys/sysmacros.h>
#endif

#ifdef MAJOR_IN_MKDEV
#include <sys/mkdev.h>
#endif

#ifdef HAVE_SYS_ASYNCH_H
#include <sys/asynch.h>
#endif

/**********************/
#include <time.h>
#include <utime.h> /* per la struttura utimbuf, l'ho aggiunto io */
/**********************/

/*
 * This is a really dirty trick but it should always work.  Traditional
 * Unix says r/w/rw are 0/1/2, so we make them true flags 1/2/3 by
 * adding 1.  Just remember to add 1 to any arg decoded with openmodes.
 */
struct xlat openmodes[] = {
	{ O_RDWR+1,	"O_RDWR"	},
	{ O_RDONLY+1,	"O_RDONLY"	},
	{ O_WRONLY+1,	"O_WRONLY"	},
	{ O_NONBLOCK,	"O_NONBLOCK"	},
	{ O_APPEND,	"O_APPEND"	},
	{ O_CREAT,	"O_CREAT"	},
	{ O_TRUNC,	"O_TRUNC"	},
	{ O_EXCL,	"O_EXCL"	},
	{ O_NOCTTY,	"O_NOCTTY"	},
#ifdef O_SYNC
	{ O_SYNC,	"O_SYNC"	},
#endif
#ifdef O_ASYNC
	{ O_ASYNC,	"O_ASYNC"	},
#endif
#ifdef O_DSYNC
	{ O_DSYNC,	"O_DSYNC"	},
#endif
#ifdef O_RSYNC
	{ O_RSYNC,	"O_RSYNC"	},
#endif
#ifdef O_NDELAY
	{ O_NDELAY,	"O_NDELAY"	},
#endif
#ifdef O_PRIV
	{ O_PRIV,	"O_PRIV"	},
#endif
#ifdef O_DIRECT
	{ O_DIRECT,	"O_DIRECT"	},
#endif
#ifdef O_LARGEFILE
	{ O_LARGEFILE,	"O_LARGEFILE"   },
#endif
#ifdef O_DIRECTORY
	{ O_DIRECTORY,	"O_DIRECTORY"   },
#endif
#ifdef O_NOFOLLOW
	{ O_NOFOLLOW, 	"O_NOFOLLOW"	},
#endif

#ifdef FNDELAY
	{ FNDELAY,	"FNDELAY"	},
#endif
#ifdef FAPPEND
	{ FAPPEND,	"FAPPEND"	},
#endif
#ifdef FMARK
	{ FMARK,	"FMARK"		},
#endif
#ifdef FDEFER
	{ FDEFER,	"FDEFER"	},
#endif
#ifdef FASYNC
	{ FASYNC,	"FASYNC"	},
#endif
#ifdef FSHLOCK
	{ FSHLOCK,	"FSHLOCK"	},
#endif
#ifdef FEXLOCK
	{ FEXLOCK,	"FEXLOCK"	},
#endif
#ifdef FCREAT
	{ FCREAT,	"FCREAT"	},
#endif
#ifdef FTRUNC
	{ FTRUNC,	"FTRUNC"	},
#endif
#ifdef FEXCL
	{ FEXCL,	"FEXCL"		},
#endif
#ifdef FNBIO
	{ FNBIO,	"FNBIO"		},
#endif
#ifdef FSYNC
	{ FSYNC,	"FSYNC"		},
#endif
#ifdef FNOCTTY
	{ FNOCTTY,	"FNOCTTY"	},
#endif	
#ifdef O_SHLOCK
	{ O_SHLOCK,	"O_SHLOCK"	},
#endif
#ifdef O_EXLOCK
	{ O_EXLOCK,	"O_EXLOCK"	},
#endif
	{ 0,		NULL		},
};



int 
print_open(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("open(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    printflags(openmodes, *((int*)info->u_arg[1]) + 1);
	    break;
	  case 2:
	    if (((info->arg_state[1] == TAKEN) && (*((int*) info->u_arg[1]) & O_CREAT)) || ((info->arg_state[1]==UPEEK) && (info->u_arg[1] & O_CREAT)))
	      tprintf("%#lo", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_open(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{

  int flags, *ptr;

  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    
    if ( get_error )
      {
	fprintf(stderr, "sys_open:  get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;  
  case 2:
    if ( (ptr = (int *) take_arg(1, tcp)) == NULL )
      {
	fprintf(stderr, "sys_open: take_arg(1, tcp)\n");
	return -1;
      }
    flags = *ptr;
    
    if (flags & O_CREAT) 
      arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}


int 
print_creat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("creat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    tprintf("%#lo", *((unsigned long*)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}


int
sys_creat(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_creat:  get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}


static struct xlat access_flags[] = {
  { F_OK,       "F_OK"		},
  { R_OK,     	"R_OK"		},
  { W_OK,      	"W_OK"		},
  { X_OK,	"X_OK"		},
#ifdef EFF_ONLY_OK
  { EFF_ONLY_OK,"EFF_ONLY_OK"	},
#endif
#ifdef EX_OK
  { EX_OK,	"EX_OK"		},
#endif
  { 0,		NULL		},
};


int 
print_access(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("access(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    printflags(access_flags, *((int*) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}


int
sys_access(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_access:  get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}


int 
print_umask(info)
msg_building_info *info;
{
  tprintf("umask(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%#lo", *((unsigned long*)info->u_arg[0]));
  else
    tprintf("_");
  
  tprintf(") ");
  
  return RVAL_OCTAL;
}


int
sys_umask(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


static struct xlat whence[] = {
  { SEEK_SET,	"SEEK_SET"	},
  { SEEK_CUR,	"SEEK_CUR"	},
  { SEEK_END,	"SEEK_END"	},
  { 0,		NULL		},
};


int 
print_lseek(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("lseek(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long *)info->u_arg[0]));
	    break;
	  case 1:
	    if (((info->arg_state[2]==TAKEN) && (*((int*) info->u_arg[2])) == SEEK_SET) || ((info->arg_state[2]==UPEEK) && (info->u_arg[2] == SEEK_SET)))
	      tprintf("%lu", *((off_t*)info->u_arg[1]));
	    else
	      tprintf("%ld", *((off_t*)info->u_arg[1]));
	    break;
	  case 2:
	    printxval(whence, *((int*) info->u_arg[2]), "SEEK_???");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return RVAL_UDECIMAL;
}


int
sys_lseek(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


int 
print_llseek(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("llseek(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    continue;
	  case 2:
	    if (((info->arg_state[4]==TAKEN) && (*((int*) info->u_arg[4])) == SEEK_SET) || ((info->arg_state[4]==UPEEK) && (info->u_arg[4] == SEEK_SET)))
	      tprintf("%llu", (*((long long int*) info->u_arg[1]) << 32 | *((unsigned long long*) info->u_arg[2])));
	    else
	      tprintf("%lld",(*((long long int*) info->u_arg[1]) << 32 | *((unsigned long long*) info->u_arg[2])));
	    break;
	  case 3:
	    tprintf("[%llu], ", *((loff_t*)info->u_arg[3]));
	    break;
	  case 4:
	    printxval(whence, *((int*)info->u_arg[4]), "SEEK_???");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}



int
sys_llseek (tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  long long int *off;

  switch(n_arg) {
  case 0:
  case 4:
    arg_alloc(tcp,n_arg);
    break;
  case 1: 
    /* caso particolare: la print_llseek per stampare questo argomento richiede anche la 
       presenza di quello successivo, quindi devo prenderli tutti e due */
    arg_alloc(tcp,n_arg);
    if (take_arg(4, tcp) == NULL)   /* CONTROLLARE non credo serva */
      {
	fprintf(stderr, "sys_llseek:  take_arg(4, tcp)\n");
	return -1;
      }
    if (take_arg(2, tcp) == NULL)  /* CONTROLLARE */
      {
	fprintf(stderr, "sys_llseek:  take_arg(2, tcp)\n");
	return -1;
      }
    break;
  case 2:
    /* caso particolare: la print_llseek per stampare questo argomento richiede anche la 
       presenza di quello precedente, quindi devo prenderli tutti e due */
    arg_alloc(tcp,n_arg);
    if (take_arg(4, tcp) == NULL)  /* CONTROLLARE */
      {
	fprintf(stderr, "sys_llseek:  take_arg(4, tcp)\n");
	return -1;
      }
    if (take_arg(1, tcp) == NULL)  /* CONTROLLARE */
      {
	fprintf(stderr, "sys_llseek:  take_arg(1, tcp)\n");
	return -1;
      }
    break;
  case 3:
    if (exiting(tcp))
      {
	if (!(syserror(tcp))) 
	  {
	    if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&off, sizeof(long long int), "llseek"))
	      {
		fprintf(stderr, "sys_llseek:  set_copy\n");
		return -1;
	      }
	  }
	else
	  {
	    fprintf(stderr, "sys_llseek:  syscall failed\n");
	    return -1;
	  }
      }
    break;
  }

  return 0;
}


int 
print_truncate(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("truncate(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    tprintf("%lu", *((off_t*)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_truncate(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_truncate: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}



int 
print_ftruncate(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("ftruncate(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%lu", *((off_t*)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_ftruncate(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



/* several stats */

static struct xlat modetypes[] = {
  { S_IFREG,	"S_IFREG"	},
  { S_IFSOCK,	"S_IFSOCK"	},
  { S_IFIFO,	"S_IFIFO"	},
  { S_IFLNK,	"S_IFLNK"	},
  { S_IFDIR,	"S_IFDIR"	},
  { S_IFBLK,	"S_IFBLK"	},
  { S_IFCHR,	"S_IFCHR"	},
  { 0,		NULL		},
};


static char *
sprintmode(mode)
int mode;
{
  static char buf[64];
  char *s;
  
  if ((mode & S_IFMT) == 0)
    s = "";
  else 
    if ((s = get_xlat_str(modetypes, mode & S_IFMT)) == NULL) 
      {
	sprintf(buf, "%#o", mode);
	return buf;
      }
  sprintf(buf, "%s%s%s%s", s,
	  (mode & S_ISUID) ? "|S_ISUID" : "",
	  (mode & S_ISGID) ? "|S_ISGID" : "",
	  (mode & S_ISVTX) ? "|S_ISVTX" : "");
  mode &= ~(S_IFMT|S_ISUID|S_ISGID|S_ISVTX);
  if (mode)
    sprintf(buf + strlen(buf), "|%#o", mode);
  s = (*buf == '|') ? buf + 1 : buf;
  return *s ? s : "0";
}

static char *
sprinttime(t)
time_t t;
{
  struct tm *tmp;
  static char buf[32];
  
  if (t == 0) 
    {
      sprintf(buf, "0");
      return buf;
    }
  tmp = localtime(&t);
  sprintf(buf, "%02d/%02d/%02d-%02d:%02d:%02d",
	  tmp->tm_year + 1900, tmp->tm_mon + 1, tmp->tm_mday,
	  tmp->tm_hour, tmp->tm_min, tmp->tm_sec);
  return buf;
}


static void
realprintstat(statbuf)
struct stat *statbuf;
{
  if (statbuf == NULL)
    {
      tprintf("{NULL}");
      return;
    }
  
  tprintf("{st_dev=makedev(%lu, %lu), st_ino=%lu, st_mode=%s, ",
	  (unsigned long) major(statbuf->st_dev),
	  (unsigned long) minor(statbuf->st_dev),
	  (unsigned long) statbuf->st_ino, 
	  sprintmode(statbuf->st_mode));
  tprintf("st_nlink=%lu, st_uid=%lu, st_gid=%lu, ",
	  (unsigned long) statbuf->st_nlink,
	  (unsigned long) statbuf->st_uid,
	  (unsigned long) statbuf->st_gid);
#ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
  tprintf("st_blksize=%lu, ", (unsigned long) statbuf->st_blksize);
#endif /* HAVE_STRUCT_STAT_ST_BLKSIZE */
#ifdef HAVE_STRUCT_STAT_ST_BLOCKS
  tprintf("st_blocks=%lu, ", (unsigned long) statbuf->st_blocks);
#endif /* HAVE_STRUCT_STAT_ST_BLOCKS */

  switch (statbuf->st_mode & S_IFMT) {
  case S_IFCHR: case S_IFBLK:
#ifdef HAVE_STRUCT_STAT_ST_RDEV
    tprintf("st_rdev=makedev(%lu, %lu), ",
	    (unsigned long) major(statbuf->st_rdev),
	    (unsigned long) minor(statbuf->st_rdev));
#else /* !HAVE_STRUCT_STAT_ST_RDEV */
    tprintf("st_size=makedev(%lu, %lu), ",
	    (unsigned long) major(statbuf->st_size),
	    (unsigned long) minor(statbuf->st_size));
#endif /* !HAVE_STRUCT_STAT_ST_RDEV */
    break;
  default:
    tprintf("st_size=%lu, ", statbuf->st_size);
    break;
  }
  
  tprintf("st_atime=%s, ", sprinttime(statbuf->st_atime));
  tprintf("st_mtime=%s, ", sprinttime(statbuf->st_mtime));
  tprintf("st_ctime=%s", sprinttime(statbuf->st_ctime));

#if HAVE_STRUCT_STAT_ST_ACLCNT
  tprintf(", st_aclcnt=%d", statbuf->st_aclcnt);
#endif
#if HAVE_STRUCT_STAT_ST_LEVEL
  tprintf(", st_level=%ld", statbuf->st_level);
#endif
#if HAVE_STRUCT_STAT_ST_FSTYPE
  tprintf(", st_fstype=%.*s",
	  (int) sizeof statbuf->st_fstype, statbuf->st_fstype);
#endif
#if HAVE_STRUCT_STAT_ST_GEN
  tprintf(", st_gen=%u", statbuf->st_gen);
#endif
  tprintf("}");

}


struct stat *
get_stat(tcp, addr)
struct tcb *tcp;
int addr;
{
  struct stat* statbuf;

  get_error = 0;

  if (!addr)
    return NULL;
  /*
  if ((exiting(tcp)) && (syserror(tcp))) 
    {
      get_error = 1;
      return NULL;
    }
  */
  statbuf=((struct stat*) malloc(sizeof(struct stat)));

  if (statbuf == NULL)
    {
      tprintf("get_stat: unable to alloc statbuf");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, statbuf) < 0) 
    {
      fprintf(stderr, "get_stat: umove statbuf failed\n");
      get_error = 1;
      free(statbuf);
      return NULL;
    }

  return statbuf;
}



static void
realprintstat64(statbuf)
struct stat64* statbuf;
{
  if (statbuf == NULL)
    {
      tprintf("{NULL}");
      return;
    }

#ifdef HAVE_LONG_LONG
  tprintf("{st_dev=makedev(%lu, %lu), st_ino=%llu, st_mode=%s, ",
#else
	  tprintf("{st_dev=makedev(%lu, %lu), st_ino=%lu, st_mode=%s, ",
#endif
		  (unsigned long) major(statbuf->st_dev),
		  (unsigned long) minor(statbuf->st_dev),
#ifdef HAVE_LONG_LONG
		  (unsigned long long) statbuf->st_ino,
#else
		  (unsigned long) statbuf->st_ino,
#endif
		  sprintmode(statbuf->st_mode));
	  tprintf("st_nlink=%lu, st_uid=%lu, st_gid=%lu, ",
		  (unsigned long) statbuf->st_nlink,
		  (unsigned long) statbuf->st_uid,
		  (unsigned long) statbuf->st_gid);
#ifdef HAVE_STRUCT_STAT_ST_BLKSIZE
	  tprintf("st_blksize=%lu, ",
		  (unsigned long) statbuf->st_blksize);
#endif /* HAVE_STRUCT_STAT_ST_BLKSIZE */
#ifdef HAVE_STRUCT_STAT_ST_BLOCKS
	  tprintf("st_blocks=%lu, ", (unsigned long) statbuf->st_blocks);
#endif /* HAVE_STRUCT_STAT_ST_BLOCKS */
	  
	  switch (statbuf->st_mode & S_IFMT) {
	  case S_IFCHR: case S_IFBLK:
#ifdef HAVE_STRUCT_STAT_ST_RDEV
	    tprintf("st_rdev=makedev(%lu, %lu), ",
		    (unsigned long) major(statbuf->st_rdev),
		    (unsigned long) minor(statbuf->st_rdev));
#else /* !HAVE_STRUCT_STAT_ST_RDEV */
	    tprintf("st_size=makedev(%lu, %lu), ",
		    (unsigned long) major(statbuf->st_size),
		    (unsigned long) minor(statbuf->st_size));
#endif /* !HAVE_STRUCT_STAT_ST_RDEV */
	    break;
	  default:
	    tprintf("st_size=%llu, ", statbuf->st_size);
	    break;
	  }
	  tprintf("st_atime=%s, ", sprinttime(statbuf->st_atime));
	  tprintf("st_mtime=%s, ", sprinttime(statbuf->st_mtime));
	  tprintf("st_ctime=%s", sprinttime(statbuf->st_ctime));
	  
#if HAVE_STRUCT_STAT_ST_ACLCNT
	  tprintf(", st_aclcnt=%d", statbuf->st_aclcnt);
#endif
#if HAVE_STRUCT_STAT_ST_LEVEL
	  tprintf(", st_level=%ld", statbuf->st_level);
#endif
#if HAVE_STRUCT_STAT_ST_FSTYPE
	  tprintf(", st_fstype=%.*s",
		  (int) sizeof statbuf->st_fstype, statbuf->st_fstype);
#endif
#if HAVE_STRUCT_STAT_ST_GEN
	  tprintf(", st_gen=%u", statbuf->st_gen);
#endif
	  tprintf("}");
}


struct stat64 *
get_stat64(tcp, addr) 
struct tcb *tcp;
int addr;
{
  struct stat64 *statbuf;
  
  get_error = 0;

  if (!addr) 
    return NULL;
  /*
  if ((exiting(tcp)) && (syserror(tcp))) 
    {
      get_error = 1;
      return NULL;
    }
  */
  statbuf=((struct stat64*) malloc(sizeof(struct stat64)));
  
  if (statbuf == NULL)
    {
      tprintf("get_stat64: unable to alloc statbuf");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, statbuf) < 0) 
    {
      fprintf(stderr, "get_stat64: umove statbuf failed\n");
      get_error = 1;
      free(statbuf);
      return NULL;
    } 
  
  return statbuf;
}


static void
convertoldstat(oldbuf, newbuf)
const struct __old_kernel_stat *oldbuf;
struct stat *newbuf;
{
    newbuf->st_dev=oldbuf->st_dev;
    newbuf->st_ino=oldbuf->st_ino;
    newbuf->st_mode=oldbuf->st_mode;
    newbuf->st_nlink=oldbuf->st_nlink;
    newbuf->st_uid=oldbuf->st_uid;
    newbuf->st_gid=oldbuf->st_gid;
    newbuf->st_rdev=oldbuf->st_rdev;
    newbuf->st_size=oldbuf->st_size;
    newbuf->st_atime=oldbuf->st_atime;
    newbuf->st_mtime=oldbuf->st_mtime;
    newbuf->st_ctime=oldbuf->st_ctime;
    newbuf->st_blksize=0;	/* not supported in old_stat */
    newbuf->st_blocks=0;		/* not supported in old_stat */
}



static void
realprintoldstat(statbuf)
struct __old_kernel_stat* statbuf; 
{
	struct stat newstatbuf;
	
	if (statbuf == NULL)
	  {
	    tprintf("{NULL}");
	    return;
	  }
	
	convertoldstat(statbuf, &newstatbuf);
	realprintstat(&newstatbuf);
}


struct __old_kernel_stat *
get_oldstat(tcp, addr)
struct tcb *tcp;
int addr;
{
  struct __old_kernel_stat *statbuf;

  get_error = 0;
  
  if (!addr) 
    return NULL;
  /*
  if ((exiting(tcp)) && (syserror(tcp))) 
    {
      get_error = 1;
      return NULL;
    }
  */
  statbuf=((struct __old_kernel_stat*) malloc(sizeof(struct __old_kernel_stat)));

  if (statbuf == NULL)
    {
      tprintf("get_oldstat: unable to alloc statbuf");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, statbuf) < 0) 
    {
      get_error = 1;
      fprintf(stderr, "get_oldstat: umove statbuf failed\n");
      free(statbuf);
      return NULL;
    }

  return statbuf;
}



int 
print_stat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("stat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    realprintstat(((struct stat *) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_stat(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_stat: get_path\n");
	tcp->arg_state[n_arg] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_stat: syscall failed");
	    return -1;
	  }
	
	((struct stat *) tcp->u_arg[1]) = get_stat(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_stat: get_stat\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  
  return 0;
}



int 
print_stat64(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("stat64(");

  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    realprintstat64(((struct stat64 *) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_stat64(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_stat64: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_stat64: syscall failed");
	    return -1;
	  }
	
	((struct stat*) tcp->u_arg[1]) = get_stat64(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_stat64: get_stat64\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }

  return 0;
}



int 
print_oldstat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("oldstat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    realprintoldstat(((struct __old_kernel_stat *) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}




int
sys_oldstat(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_oldstat: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_oldstat: syscall failed");
	    return -1;
	  }
	
	((struct __old_kernel_stat *) tcp->u_arg[1]) = get_oldstat(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_oldstat:  get_oldstat\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  
  return 0;
}


int 
print_fstat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("fstat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstat(((struct stat *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_fstat(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_fstat: syscall failed");
	    return -1;
	  }
	
	((struct stat*) tcp->u_arg[1]) = get_stat(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_fstat: get_stat\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  
  return 0;
}


int 
print_fstat64(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("fstat64(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstat64(((struct stat64 *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_fstat64(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_fstat64: syscall failed");
	    return -1;
	  }
	
	((struct stat64*) tcp->u_arg[1]) = get_stat64(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_fstat64: get_stat64\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }

  return 0;
}



int 
print_oldfstat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("oldfstat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld, ", *((long*) info->u_arg[0]));
	    break;
	  case 1:
	    realprintoldstat(((struct __old_kernel_stat *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_oldfstat(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_oldfstat: syscall failed");
	    return -1;
	  }
	
	((struct __old_kernel_stat *) tcp->u_arg[1]) = get_oldstat(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_oldfstat: get_stat64\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  
  return 0;
}


int 
print_lstat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("lstat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    realprintstat(((struct stat *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_lstat(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, ":sys_lstat: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_lstat: syscall failed");
	    return -1;
	  }
	
	((struct stat*) tcp->u_arg[1]) = get_stat(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_lstat: get_stat\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }

  return 0;
}



int 
print_lstat64(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("lstat64(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    realprintstat64(((struct stat64 *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_lstat64(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_lstat64: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_lstat64: syscall failed");
	    return -1;
	  }
	
	((struct stat*) tcp->u_arg[1]) = get_stat64(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_lstat64: get_stat64\n");
	    tcp->arg_state[0] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }

  return 0;
}


int 
print_oldlstat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("oldlstat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    realprintoldstat(((struct __old_kernel_stat *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_oldlstat(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_oldlstat: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_oldlstat: syscall failed");
	    return -1;
	  }
	
	((struct __old_kernel_stat *) tcp->u_arg[1]) = get_oldstat(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_oldlstat: get_oldstat\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  
  return 0;
}


static struct xlat fsmagic[] = {
	{ 0x73757245,	"CODA_SUPER_MAGIC"	},
	{ 0x012ff7b7,	"COH_SUPER_MAGIC"	},
	{ 0x1373,	"DEVFS_SUPER_MAGIC"	},
	{ 0x1cd1,	"DEVPTS_SUPER_MAGIC"	},
	{ 0x414A53,	"EFS_SUPER_MAGIC"	},
	{ 0xef51,	"EXT2_OLD_SUPER_MAGIC"	},
	{ 0xef53,	"EXT2_SUPER_MAGIC"	},
	{ 0x137d,	"EXT_SUPER_MAGIC"	},
	{ 0xf995e849,	"HPFS_SUPER_MAGIC"	},
	{ 0x9660,	"ISOFS_SUPER_MAGIC"	},
	{ 0x137f,	"MINIX_SUPER_MAGIC"	},
	{ 0x138f,	"MINIX_SUPER_MAGIC2"	},
	{ 0x2468,	"MINIX2_SUPER_MAGIC"	},
	{ 0x2478,	"MINIX2_SUPER_MAGIC2"	},
	{ 0x4d44,	"MSDOS_SUPER_MAGIC"	},
	{ 0x564c,	"NCP_SUPER_MAGIC"	},
	{ 0x6969,	"NFS_SUPER_MAGIC"	},
	{ 0x9fa0,	"PROC_SUPER_MAGIC"	},
	{ 0x002f,	"QNX4_SUPER_MAGIC"	},
	{ 0x52654973,	"REISERFS_SUPER_MAGIC"	},
	{ 0x02011994,	"SHMFS_SUPER_MAGIC"	},
	{ 0x517b,	"SMB_SUPER_MAGIC"	},
	{ 0x012ff7b6,	"SYSV2_SUPER_MAGIC"	},
	{ 0x012ff7b5,	"SYSV4_SUPER_MAGIC"	},
	{ 0x00011954,	"UFS_MAGIC"		},
	{ 0x54190100,	"UFS_CIGAM"		},
	{ 0x012ff7b4,	"XENIX_SUPER_MAGIC"	},
	{ 0x012fd16d,	"XIAFS_SUPER_MAGIC"	},
	{ 0,		NULL			},
};





static char *
sprintfstype(magic)
int magic;
{
  static char buf[32];
  char *s;
  
  s = get_xlat_str(fsmagic, magic);
  if (s) 
    {
      sprintf(buf, "\"%s\"", s);
      return buf;
    }
  sprintf(buf, "%#x", magic);
  return buf;
}


static void
realprintstatfs(statbuf)
struct statfs* statbuf;
{
  if (statbuf == NULL)
    {
      tprintf("{NULL}");
      return;
    }

  tprintf("{f_type=%s, f_bsize=%lu, f_blocks=%lu, f_bfree=%lu, ",
	  sprintfstype(statbuf->f_type),
	  (unsigned long)statbuf->f_bsize,
	  (unsigned long)statbuf->f_blocks,
	  (unsigned long)statbuf->f_bfree);
  tprintf("f_files=%lu, f_ffree=%lu",
	  (unsigned long)statbuf->f_files,
	  (unsigned long)statbuf->f_ffree);
  tprintf(", f_namelen=%lu", (unsigned long)statbuf->f_namelen);
  tprintf("}");
}


struct statfs *
get_statfs(tcp, addr)
struct tcb *tcp;
int addr;
{
  struct statfs* statbuf;

  get_error = 0;

  if (!addr)
    return NULL;
  /*
  if ((exiting(tcp)) && (syserror(tcp))) 
    {
      get_error = 1;
      return NULL;
    }
  */
  statbuf=((struct statfs*) malloc(sizeof(struct statfs)));

  if (statbuf == NULL)
    {
      tprintf("get_statfs: unable to alloc statbuf");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, statbuf) < 0) {
    get_error = 1;
    fprintf(stderr, "get_statfs: umove statbuf failed\n");
    free(statbuf);
    return NULL;
  }

  return statbuf;
}


int 
print_statfs(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("statfs(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    realprintstatfs(((struct statfs *) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_statfs(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_statfs: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_statfs: syscall failed");
	    return -1;
	  }

	((struct statfs*) tcp->u_arg[1]) = get_statfs(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_statfs: get_statfs\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  
  return 0;
}


int 
print_fstatfs(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("fstatfs(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstatfs(((struct statfs *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_fstatfs(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_fstatfs: syscall failed");
	    return -1;
	  }
	
	((struct stat*) tcp->u_arg[1]) = get_statfs(tcp, tcp->u_arg[1]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_fstatfs: get_statfs\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  
  return 0;
}

/*
int 
print_pivotroot(info)
msg_building_info *info;
{

}
*/

int
sys_pivotroot(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char*) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_pivotroot: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    ((char*) tcp->u_arg[1]) = get_path(tcp, tcp->u_arg[1]);
    if ( get_error )
      {
	fprintf(stderr, "sys_pivotroot: get_path\n");
	tcp->arg_state[1] = EMPTY;
	return -1;
      }
    tcp->arg_state[1] = TAKEN;
    break;
  }
  
  return 0;
}


/* directory */
int 
print_chdir(info)
msg_building_info *info;
{
  tprintf("chdir(");
  
  if (info->arg_state[0]==TAKEN)
    string_quote((char *) info->u_arg[0]);
  else
    tprintf("_");
  
  tprintf(") ");
  
  return 0;
}


int
sys_chdir(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  ((char*) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_chdir:  get_path\n");
      tcp->arg_state[n_arg] = EMPTY;
      return  -1;
    }
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}



int 
print_mkdir(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("mkdir(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    tprintf("%#lo", *((unsigned long *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_mkdir(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_mkdir:  get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}



int 
print_rmdir(info)
msg_building_info *info;
{
  tprintf("rmdir(");
  
  if (info->arg_state[0]==TAKEN)
    {
      /* stampa il path */
      string_quote((char *) info->u_arg[0]);
    }
  else
    tprintf("_");
  
  tprintf(") ");

  return 0;
}


int
sys_rmdir(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_rmdir:  get_path\n");
      tcp->arg_state[n_arg] = EMPTY;
      return -1;
    }
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}


int 
print_fchdir(info)
msg_building_info *info;
{
  tprintf("fchdir(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%ld", *((long*) info->u_arg[0]));
  else
    tprintf("_");
  
  tprintf(") ");
  
  return 0;
}


int
sys_fchdir(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


int 
print_chroot(info)
msg_building_info *info;
{
  tprintf("chroot(");
  
  if (info->arg_state[0]==TAKEN)
    {
      /* stampa il path */
      string_quote((char *) info->u_arg[0]);
    }
  else
    tprintf("_");
  
  tprintf(") ");

  return 0;
}


int
sys_chroot(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_chroot:  get_path\n");
      tcp->arg_state[n_arg] = EMPTY;
      return -1;
    }
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}



int 
print_link(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("link(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
	  /* stampa il path */
	  string_quote((char *) info->u_arg[i]);
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_link(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_link:  get_path\n");
      tcp->arg_state[n_arg] = EMPTY;
      return -1;
      }
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}


int 
print_unlink(info)
msg_building_info *info;
{
  tprintf("unlink(");
  
  if (info->arg_state[0]==TAKEN)
    {
      /* stampa il path */
      string_quote((char *) info->u_arg[0]);
    }
  else
    tprintf("_");
  
  tprintf(") ");

  return 0;
}


int
sys_unlink(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_unlink:  get_path\n");
      tcp->arg_state[n_arg] = EMPTY;
      return -1;
    }
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}



int 
print_symlink(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("symlink(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
	    /* stampa il path */
	    string_quote((char *) info->u_arg[i]);
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_symlink(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_symlink:  get_path\n");
      tcp->arg_state[n_arg] = EMPTY; 
      return -1;
    }
  tcp->arg_state[n_arg] = TAKEN;
    
  return 0;
}


int 
print_readlink(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("readlink(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	  case 1:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[i]);
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_readlink(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
    if ( get_error )
      {
	fprintf(stderr, "sys_readlink:  get_path\n");
	tcp->arg_state[n_arg] = EMPTY;
	return -1;
      }
    tcp->arg_state[n_arg] = TAKEN;
    break;
  case 1:
    if (exiting(tcp))
      {
	if (!syserror(tcp))
	  {
	    ((char *) tcp->u_arg[1]) = get_pathn(tcp, tcp->u_arg[1], tcp->u_rval);
	    if ( get_error )
	      {
		fprintf(stderr, "sys_readlink:  get_pathn\n");
		tcp->arg_state[n_arg] = EMPTY;
		return -1;
	      }
	    tcp->arg_state[1] = TAKEN;
	  }
	else
	  {
	    fprintf(stderr, "sys_readlink: syscall fallita\n");
	    return -1;
	  }
      }
    break;
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}




int 
print_rename(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("rename(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
	    /* stampa il path */
	    string_quote((char *) info->u_arg[i]);
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_rename(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_rename: get_path\n");
      tcp->arg_state[n_arg] = EMPTY;
      return -1;
    }
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}


int 
print_chown(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("chown(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[i]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_chown(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[n_arg]) = get_path(tcp, tcp->u_arg[n_arg]);
    if ( get_error )
      {
	fprintf(stderr, "sys_chown: get_path\n");
	tcp->arg_state[n_arg] = EMPTY;
	return -1;
      }
    tcp->arg_state[n_arg] = TAKEN;
    break;
  case 1:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}



int 
print_fchown(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("fchown(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[i]));
	    break;
	  case 1:
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[i]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_fchown(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


int 
print_chmod(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("chmod(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    tprintf("%#lo", *((unsigned long*)info->u_arg[i]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_chmod(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_chmod: get_path\n");
	tcp->arg_state[0] = EMPTY; 
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}


int 
print_fchmod(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("fchmod(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[i]));
	    break;
	  case 1:
	    tprintf("%#lo", *((unsigned long*)info->u_arg[i]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_fchmod(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


int 
print_utime(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("utime(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    if (!info->u_arg[1])
	      tprintf("NULL");
	    else {
	      tprintf("[%s,", sprinttime(((struct utimbuf*) info->u_arg[1])->actime));
	      tprintf(" %s]", sprinttime(((struct utimbuf*) info->u_arg[1])->modtime));
	    }
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int 
sys_utime(struct tcb *tcp,int n_arg) 
{
  struct utimbuf* buf;
  
  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_utime: get_path\n");
	tcp->arg_state[0] = EMPTY;
        return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    if (!tcp->u_arg[1])
      tcp->arg_state[1] = TAKEN; /* TODO: ATTENZIONE questo e' un caso particolare */ 
    else 
      {
	if (set_copy(tcp, n_arg,tcp->u_arg[n_arg], (long*)&buf, sizeof(struct utimbuf), "utime"))
	  {
	    fprintf(stderr, "sys_utime: set_copy\n");
	    return -1;
	  }
      }
    break; 
  }
  
  return 0;
}



int 
print_mknod(info)
msg_building_info *info;
{
  int i,mode;
  int nargs=get_nargs(info->scno);
  
  tprintf("mknod(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    tprintf("%s", sprintmode(*((int*)info->u_arg[1])));
	    break;
	  case 2:
	    mode = *((int*)info->u_arg[1]);
	    
	    switch (mode & S_IFMT) {
	    case S_IFCHR: case S_IFBLK:
	      tprintf("makedev(%lu, %lu)",
		      (unsigned long) major(*((long*)info->u_arg[2])),
		      (unsigned long) minor(*((long*)info->u_arg[2])));
	      break;
	    default:
	      tprintf("_");
	      break;
	    } 
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_mknod(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  int mode, *ptr;

  switch (n_arg) {
  case 0:
    ((char *) tcp->u_arg[0]) = get_path(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_mknod: get_path\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  case 2: /* CONTROLLARE */
    if ((ptr = (int *) take_arg(1, tcp)) == NULL)
      {
	fprintf(stderr, "sys_mknod: take_arg(1, tcp)\n");
	return -1;
      }
    mode = *ptr;
    
    switch (mode & S_IFMT) 
      {
      case S_IFCHR: 
      case S_IFBLK:
	arg_alloc(tcp,n_arg);
	break;
      }
    break;
  }
  return 0;
}





int 
print_fsync(info)
msg_building_info *info;
{  
  tprintf("fsync(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%ld", *((long*)info->u_arg[0]));
  else
    tprintf("_");
  
  tprintf(") ");
  
  return 0;
}


int
sys_fsync(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



static void
realprintdir(d)
struct dirent *d;
{
    if (d == NULL)
      {
	tprintf("{NULL}");
	return;
      }

    tprintf("{d_ino=%ld, ", (unsigned long) d->d_ino);
    tprintf("d_name=");
    if (d->d_name == NULL)
      tprintf("NULL");
    else
      string_quote(d->d_name);
    tprintf("}");
}



struct dirent *
get_dir(tcp, addr)
struct tcb *tcp;
long addr;
{
  struct dirent *d;
  char* t;

  get_error = 0;

  if (!addr)
    return NULL;

  d = ((struct dirent *) malloc(sizeof(struct dirent)));

  if (d == NULL)
    {
      fprintf(stderr,"get_dir: unable to alloc dirent\n");
      get_error = 1;
      return NULL;
    }
  
  if (umove(tcp, addr, d) < 0) 
    {
      fprintf(stderr,"get_dir: umove dirent failed");
      get_error = 1;
      free(d);
      return NULL;
    }

  t = get_pathn(tcp, (long) ((struct dirent *) addr)->d_name, d->d_reclen);

  if (t)
    {
      strcpy(d->d_name,t);
      free(t);
    }
  else
    strcpy(d->d_name,"NULL");
  
  return d;
}


int 
print_readdir(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("readdir(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintdir(((struct dirent*)info->u_arg[1]));
	    break;
	  case 2:
	    /*Not much point in printing this out, it is always 1. */
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}


int
sys_readdir(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp))
      {
	if (!(syserror(tcp) || tcp->u_rval == 0))
	  {
	    ((struct dirent *) tcp->u_arg[1]) = get_dir(tcp, tcp->u_arg[1]);
	    if ( get_error )
	      {
		fprintf(stderr,"get_readdir: get_dir(tcp, tcp->u_arg[1])");
		tcp->arg_state[n_arg] = EMPTY;
		return -1;
	      }
	    tcp->arg_state[1] = TAKEN;
	  }
	else
	  {
	    fprintf(stderr,"get_readdir: syscall failed");
	    return -1;
	  }
      }
    break;
  case 2:
    /* Not much point in printing this out, it is always 1. */
    if (exiting(tcp))
      arg_alloc(tcp,n_arg);
    break;
  }

  return 0;
}


int 
print_getdents(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  int j,len,dents=0;
  char *buf;
  
  tprintf("getdents(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    buf = (char *) info->u_arg[1];
	    len = info->u_rval;

	    tprintf("{");
	    for (j = 0; j < len;) 
	      {
		struct kernel_dirent *d = (struct kernel_dirent *) &buf[j];
		tprintf("%s{d_ino=%lu, d_off=%lu, ", j ? " " : "", d->d_ino, d->d_off);
		tprintf("d_reclen=%u, d_name=\"%s\"}", d->d_reclen, d->d_name);
		j += d->d_reclen;
		dents++;
	      } 
	    tprintf("}");
	    tprintf(" /* %u entries */", dents);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}




int
sys_getdents(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  int len;
  char *buf;
  
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_getdents: syscall failed");
	    return -1;
	  }
	
	len = tcp->u_rval;

	if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&buf, len, "getdents"))
	  {
	    fprintf(stderr,"sys_getdents: set_copy");
	    return -1;
	  }
      }
    break;
  case 2:
    //if (exiting(tcp)) /* CONTROLLARE */
      arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}


int
sys_getdents64(tcp,n_arg) /* CONTROLLARE */
struct tcb *tcp;
int n_arg;
{
  int len;
  char *buf;
  
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_getdents64: syscall failed");
	    return -1;
	  }
	
	len = tcp->u_rval;

	if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&buf, len, "getdents64"))
	  {
	    fprintf(stderr,"sys_getdents64: set_copy");
	    return -1;
	  }
      }
    break;
  case 2:
    //if (exiting(tcp)) /* CONTROLLARE */
      arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}




int 
print_getcwd(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("getcwd(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* stampa il path */
	    string_quote((char *) info->u_arg[0]);
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_getcwd(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_getcwd: syscall failed");
	    return -1;
	  }
	
	((char *) tcp->u_arg[0]) = get_pathn(tcp, tcp->u_arg[0], tcp->u_rval - 1);
	if ( get_error ) 
	  {
	    fprintf(stderr,"sys_getcwd: get_pathn");
	    tcp->arg_state[0] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[0] = TAKEN;
      }
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}
