/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: follow.c,v 1.1.1.1 2001/12/17 08:58:19 frassi Exp $
 */

#include "follow.h"
#include <assert.h>
#include <stdio.h>

gboolean 
is_descendant(follow_array_t afs, int origin, context_t orig_context, int num_event, context_t context)
{
  int ind;

  for (ind = 1; ind <= afs.dim_ar && afs.follow_array[ind].evid.num_event != origin; ind++);
  
  assert( afs.follow_array[ind].evid.num_event == origin );
  
  /* assumo che gli eventi in entrata precedano di una posizione nell'array gli eventi 
     in uscita */ 
  if ( afs.follow_array[ind].evid.context != orig_context)
    ind += afs.follow_array[ind].evid.context == intro;
    
  /* a questo punto ind e' l'indice dell'evento che cercavo */
  if ( afs.follow_array[ind].ev_set_follow_ptr )
    {
      int event;
      
      first_event(afs.follow_array[ind].ev_set_follow_ptr, event);

      while ( event != -1 )
	{
	  context_t ancestral_context;
	  assert(afs.follow_array[event].ev_nod_ptr->op == none);
	  ancestral_context = afs.follow_array[event].ev_nod_ptr->simple_event->context;
	  if ( event == num_event && ancestral_context == context )
	    return TRUE;
	  event=each_event(afs.follow_array[ind].ev_set_follow_ptr);
	}
      
      first_event(afs.follow_array[ind].ev_set_follow_ptr, event);

      while ( event != -1 )
	{
	  context_t ancestral_context;
	  ancestral_context = afs.follow_array[event].ev_nod_ptr->simple_event->context;
	  if ( is_descendant( afs, event, ancestral_context, num_event, context ) )
	    return TRUE;
	  event=each_event(afs.follow_array[ind].ev_set_follow_ptr);
	}
      
    }

  return FALSE;
}


void 
print_all_follow(follow_array_t afs)
{
  int i, event;
  
  for (i = 1; i <= afs.dim_ar; i++)
    {
      printf("\nfollow(%d, ", afs.follow_array[i].evid.num_event);

      if ( !afs.follow_array[i].evid.context )
	printf("intro");
      else
	printf("exiting");

      printf(") = {");

      if ( afs.follow_array[i].ev_set_follow_ptr )
	{
	  first_event(afs.follow_array[i].ev_set_follow_ptr, event);
	  
	  while ( event != -1 )
	    {
	      printf(" (%d, ", event );

	      if (!(afs.follow_array[event].ev_nod_ptr->simple_event->context))
		printf("intro");
	      else
		printf("exiting");
	      printf(")");

	      printf(",");

	      event=each_event(afs.follow_array[i].ev_set_follow_ptr);
	    }

	}
      printf(" }");
    }
}


