#ifndef GETS_H
#define GETS_H

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: gets.h,v 1.2 2003/02/17 15:50:20 frassi Exp $
 */

#include "automa_management.h"

//#include <fcntl.h>
//#include <sys/file.h>  
//#include <asm/stat.h>
#include <sys/resource.h>
#include <sys/time.h>
#include <dirent.h>
#include <sys/uio.h>
#include <sys/socket.h>
#include <sys/sem.h>
#include <sys/shm.h>
#include <signal.h>


extern int get_error; /* ATTENZIONE, non dovrebbe essere dichiarata in un .c e qui solo extern ???? SI 
			 e' stata dichiarata in util.c */

/* variabile globale utilizzata dalle seguenti funzioni 
   per indicare il verificarsi di un errore per cui non si deve considerare il valore restituito */

char *get_str(struct tcb *, long, int);
/* restituisce una copia della stringa contenuta a partire dall'indirizzo di memoria 
   passatogli come secondo argomento, nello spazio del processo tracciato; il terzo 
   argomento indica la lunghezza della stringa; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

long *get_num(struct tcb *, long);
/* restituisce l'indirizzo ad una copia della parola di memoria contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */ 

char *get_path(struct tcb *, long);
/* restituisce una copia della stringa che rappresenta un path contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct stat *get_stat(struct tcb *, int);
/* restituisce il puntatore ad una copia della struttura stat contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato;
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct rusage *get_rusage(struct tcb *, long); 
/* restituisce il puntatore ad una copia della struttura rusage contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct flock *get_flock(struct tcb *, long);
/* restituisce il puntatore ad una copia della struttura flock contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct flock64 *get_flock64(struct tcb *, long);

//sockaddrbuf *get_sock(struct tcb *, long, int);       sockaddrbuf e' definito in net.c

int move_arg(struct tcb *, int); /* non utilizza get_error     TODO: togliere */

struct timeval *get_tv(struct tcb *, long);
/* restituisce il puntatore ad una copia della struttura timeval contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct timezone *get_tz(struct tcb *, long);
/* restituisce il puntatore ad una copia della struttura timezone contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct stat64 *get_stat64(struct tcb *, int);
/* restituisce il puntatore ad una copia della struttura stat64 contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

char *get_pathn(struct tcb *, long, int);
/* restituisce una copia della stringa che rappresenta un path contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; il terzo argomento 
   indica la lunghezza della stringa; se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

char **get_argv(struct tcb *, long);
/* restituisce una copia della array di stringhe che rappresentano gli argomenti passati ad una execve, 
   contenuto a partire dall'indirizzo indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL; se si verifica un errore durante la copia di una delle 
   stringhe setta get_error a 1, ma restituisce comunque un array di argomenti */



struct __old_kernel_stat *get_oldstat(struct tcb *, int);
/* restituisce il puntatore ad una copia della struttura oldstat contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct statfs *get_statfs(struct tcb *, int);
/* restituisce il puntatore ad una copia della struttura statfs contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL;
   se si verifica un errore setta get_error a 1 */

struct dirent *get_dir(struct tcb *, long);
/* restituisce il puntatore ad una copia della struttura statfs contenuta a partire dall'indirizzo 
   indicato nel secondo argomento, nello spazio del processo tracciato; 
   se il secondo argomento e' NULL restituisce NULL; se si verifica un errore durante la copia 
   della struttura o di uno dei suoi campi setta get_error a 1 */

struct iovec *get_iovec(struct tcb *,struct iovec *,long);

struct msghdr *get_msghdr(struct tcb *tcp,long addr);            

struct timex *get_timex(struct tcb *tcp,long addr);

struct itimerval *get_itv(struct tcb *tcp,long addr);

union semun *get_union_semun(struct tcb *tcp,long addr,int cmd);

struct shmid_ds *get_shmid_ds(struct tcb *tcp,int n,long addr);

/* TODO: completare con le funzioni definite in modo sparso tra le sys_* */

#endif /* !GETS_H */
