/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: internal_syscall.c,v 1.3 2003/02/17 15:51:06 frassi Exp $
 */


#include "util.h"
#include "runtime_util.h"

#include <fcntl.h> // Aggiunto dopo averlo commentato in defs.h
#include <sys/stat.h>
#include <sys/wait.h>
#include <signal.h>
#include <sys/syscall.h>

#ifdef HAVE_SYS_REG_H
# include <sys/reg.h>
#ifndef PTRACE_PEEKUSR
# define PTRACE_PEEKUSR PTRACE_PEEKUSER
#endif
#ifndef PTRACE_POKEUSR
# define PTRACE_POKEUSR PTRACE_POKEUSER
#endif
#endif

extern automa atm;

extern void long_to_sigset(long, sigset_t *);

extern char * sprintsigmask(char *, sigset_t *, int);

/* /\* */
/*  * Check process TCP for the disposition of signal SIG. */
/*  * Return 1 if the process would somehow manage to  survive signal SIG, */
/*  * else return 0.  This routine will never be called with SIGKILL. */
/*  *\/ */
/* int */
/* sigishandled(tcp, sig) */
/* struct tcb *tcp; */
/* int sig; */
/* { */
/* 	int sfd; */
/* 	char sname[32]; */
/* 	char buf[1024]; */
/* 	char *s; */
/* 	int i; */
/* 	unsigned int signalled, blocked, ignored, caught; */

/* 	/\* This is incredibly costly but it's worth it. *\/ */

/* 	sprintf(sname, "/proc/%d/stat", tcp->pid); */
/* 	if ((sfd = open(sname, O_RDONLY)) == -1) { */
/* 		perror(sname); */
/* 		return 1; */
/* 	} */
/* 	i = read(sfd, buf, 1024); */
/* 	buf[i] = '\0'; */
/* 	close(sfd); */
/* 	/\* */
/* 	 * Skip the extraneous fields. This loses if the */
/* 	 * command name has any spaces in it.  So be it. */
/* 	 *\/ */
/* 	for (i = 0, s = buf; i < 30; i++) { */
/* 	  while (*++s != ' ') { */
/* 	    if (!*s) */
/* 	      break; */
/* 	  } */
/* 	} */
/* 	if (sscanf(s, "%u%u%u%u", &signalled, &blocked, &ignored, &caught) != 4) { */
/* 	  fprintf(stderr, "/proc/pid/stat format error\n"); */
/* 	  return 1; */
/* 	} */
/* #ifdef DEBUG */
/* 	fprintf(stderr, "sigs: %08x %08x %08x %08x\n", */
/* 		signalled, blocked, ignored, caught); */
/* #endif */
/* 	if ((ignored & sigmask(sig)) || (caught & sigmask(sig))) */
/* 		return 1; */

/* 	switch (sig) { */
/* 	case SIGCONT: */
/* 	case SIGSTOP: */
/* 	case SIGTSTP: */
/* 	case SIGTTIN: */
/* 	case SIGTTOU: */
/* 	case SIGCHLD: */
/* 	case SIGIO: */
/* #if defined(SIGURG) && SIGURG != SIGIO */
/* 	case SIGURG: */
/* #endif */
/* 	case SIGWINCH: */
/* 		/\* Gloria Gaynor says ... *\/ */
/* 		return 1; */
/* 	default: */
/* 		break; */
/* 	} */
/* 	return 0; */
/* } */


static void
parse_sigset_t (const char *str, sigset_t *set) /* era in signal.c */
{
	unsigned int digit;
	int i;

	sigemptyset(set);

	for (i = _NSIG - 4; i >= 0; i -= 4, ++str) {
		if (*str >= 'a')
			digit = (*str - 'a') + 10;
		else
			digit = *str - '0';
		if (digit & 1)
			sigaddset(set, i + 1);
		if (digit & 2)
			sigaddset(set, i + 2);
		if (digit & 4)
			sigaddset(set, i + 3);
		if (digit & 8)
			sigaddset(set, i + 4);
	}
}


/*
 * Check process TCP for the disposition of signal SIG.
 * Return 1 if the process would somehow manage to  survive signal SIG,
 * else return 0.  This routine will never be called with SIGKILL.
 */
int
sigishandled(tcp, sig) /* era in signal.c */
struct tcb *tcp;
int sig;
{
#ifdef LINUX
	int sfd;
	char sname[32];
	char buf[2048];
	char *s;
	int i;
	sigset_t ignored, caught;
#endif
#ifdef SVR4
	/*
	 * Since procfs doesn't interfere with wait I think it is safe
	 * to punt on this question.  If not, the information is there.
	 */
	return 1;
#else /* !SVR4 */
	switch (sig) {
	case SIGCONT:
	case SIGSTOP:
	case SIGTSTP:
	case SIGTTIN:
	case SIGTTOU:
	case SIGCHLD:
	case SIGIO:
#if defined(SIGURG) && SIGURG != SIGIO
	case SIGURG:
#endif
	case SIGWINCH:
		/* Gloria Gaynor says ... */
		return 1;
	default:
		break;
	}
#endif /* !SVR4 */
#ifdef LINUX

	/* This is incredibly costly but it's worth it. */
	/* NOTE: LinuxThreads internally uses SIGRTMIN, SIGRTMIN + 1 and
	   SIGRTMIN + 2, so we can't use the obsolete /proc/%d/stat which
	   doesn't handle real-time signals). */
	sprintf(sname, "/proc/%d/status", tcp->pid);
	if ((sfd = open(sname, O_RDONLY)) == -1) {
		perror(sname);
		return 1;
	}
	i = read(sfd, buf, sizeof(buf));
	buf[i] = '\0';
	close(sfd);
	/*
	 * Skip the extraneous fields. We need to skip
	 * command name has any spaces in it.  So be it.
	 */
	s = strstr(buf, "SigIgn:\t");
	if (!s)
	{
		fprintf(stderr, "/proc/pid/status format error\n");
		return 1;
	}
	parse_sigset_t(s + 8, &ignored);

	s = strstr(buf, "SigCgt:\t");
	if (!s)
	{
		fprintf(stderr, "/proc/pid/status format error\n");
		return 1;
	}
	parse_sigset_t(s + 8, &caught);

#ifdef DEBUG
	fprintf(stderr, "sigs: %016qx %016qx (sig=%d)\n",
		*(long long *) &ignored, *(long long *) &caught, sig);
#endif
	if (sigismember(&ignored, sig) || sigismember(&caught, sig))
		return 1;
#endif /* LINUX */

#ifdef SUNOS4
	void (*u_signal)();

	if (upeek(tcp->pid, uoff(u_signal[0]) + sig*sizeof(u_signal),
	    (long *) &u_signal) < 0) {
		return 0;
	}
	if (u_signal != SIG_DFL)
		return 1;
#endif /* SUNOS4 */

	return 0;
}





/*
clearbpt

input: una struttura di tipo tcp

output:0 se tutto ok, -1 in caso di errore

dovrebbe eliminare un breakpoint
*/
int
clearbpt(tcp)
struct tcb *tcp;
{

	long eip;

       	if (!(tcp->flags & TCB_BPTSET)) {
		fprintf(stderr, "PANIC: TCB not set in pid %u\n", tcp->pid);
		return -1;
	}
	errno = 0;
	ptrace(PTRACE_POKETEXT, tcp->pid, (char *) tcp->baddr, tcp->inst[0]);
	if (errno) {
	  perror("clearbtp: ptrace(PTRACE_POKETEXT, ...)");
	  return -1;
	}
	tcp->flags &= ~TCB_BPTSET;
	

	if (upeek(tcp->pid, 4*EIP, &eip) < 0)
		return -1;
	if (eip != tcp->baddr) {
	  /* The breakpoint has not been reached yet.  */
	  //if (debug)
	  fprintf(stderr, "NOTE: PC not at bpt (pc %#lx baddr %#lx)\n", eip, tcp->baddr);
	  return 0;
	}

	return 0;
}



/*
setbpt

input: una struttura di tipo tcp

output:0 se tutto ok, -1 in caso di errore

dovrebbe settare un breakpoint???
*/

int
setbpt(tcp)
struct tcb *tcp;
{

#define LOOP	0x0000feeb


	if (tcp->flags & TCB_BPTSET) {
		fprintf(stderr, "PANIC: bpt already set in pid %u\n", tcp->pid);
		return -1;
	}

	if (upeek(tcp->pid, 4*EIP, &tcp->baddr) < 0)
		return -1;
	
	/* memorizzo la prossima istruzione che avrebbe eseguito il processo padre, 
	   per mezzo dell'Instruction Pointer (registro eip) */
     	tcp->inst[0] = ptrace(PTRACE_PEEKTEXT, tcp->pid, (char *) tcp->baddr, 0);
	if (errno) {
		perror("setbpt: ptrace(PTRACE_PEEKTEXT, ...)");
		return -1;
	}

	/*
	 * XXX - BRUTAL MODE ON
	 * We cannot set a real BPT in the child, since it will not be
	 * traced at the moment it will reach the trap and would probably
	 * die with a core dump.
	 * Thus, we are force our way in by taking out two instructions
	 * and insert an eternal loop instead, in expectance of the SIGSTOP
	 * generated by out PTRACE_ATTACH.
	 * Of cause, if we evaporate ourselves in the middle of all this...
	 */

	ptrace(PTRACE_POKETEXT, tcp->pid, (char *) tcp->baddr, LOOP);
	if (errno) {
		perror("setbpt: ptrace(PTRACE_POKETEXT, ...)");
		return -1;
	}

	tcp->flags |= TCB_BPTSET;

	return 0;
}


int internal_exit(struct tcb *tcp) {
  if (entering(tcp))
    tcp->flags |= TCB_EXITING;
  
  return 0;
}




/* defines copied from linux/sched.h since we can't include that
 * ourselves (it conflicts with *lots* of libc includes)
 */
#define CSIGNAL         0x000000ff      /* signal mask to be sent at exit */
#define CLONE_VM        0x00000100      /* set if VM shared between processes */
#define CLONE_FS        0x00000200      /* set if fs info shared between processes */
#define CLONE_FILES     0x00000400      /* set if open files shared between processes */
#define CLONE_SIGHAND   0x00000800      /* set if signal handlers shared */
#define CLONE_PID       0x00001000      /* set if pid shared */
#define CLONE_PTRACE    0x00002000      /* set if we want to let tracing continue on the child too */
#define CLONE_VFORK     0x00004000      /* set if the parent wants the child to wake it up on mm_release */
#define CLONE_PARENT    0x00008000      /* set if we want to have the same parent as the cloner */

struct xlat clone_flags[] = {
    { CLONE_VM,		"CLONE_VM"	},
    { CLONE_FS,		"CLONE_FS"	},
    { CLONE_FILES,	"CLONE_FILES"	},
    { CLONE_SIGHAND,	"CLONE_SIGHAND"	},
    { CLONE_PID,	"CLONE_PID"	},
    { CLONE_PTRACE,	"CLONE_PTRACE"	},
    { CLONE_VFORK,	"CLONE_VFORK"	},
    { CLONE_PARENT,	"CLONE_PARENT"	},
    { 0,		NULL		},
};


/* #ifdef SYS_clone */
/* int internal_clone(struct tcb *tcp)  */
/* { */
/*   struct tcb *tcpchild; */
/*   int pid; */
/*   int dont_follow = 0; */

/*   if (entering(tcp))  */
/*     { */
/*     /\* For now we don't follow clone yet.. we're just preparing the code *\/ */
/*     dont_follow = 1; */
    
/*     if (dont_follow) */
/*       return 0; */
/*     if (nprocs == MAX_PROCS)  */
/*       { */
/* 	tcp->flags &= ~TCB_FOLLOWFORK; */
/* 	fprintf(stderr, "sys_fork: tcb table full\n"); */
/* 	return 0; */
/*       } */
/*     tcp->flags |= TCB_FOLLOWFORK; */
    
/*     /\* XXX  */
/*      * We will take the simple approach and add CLONE_PTRACE to the clone */
/*      * options. This only works on Linux 2.2.x and later. This means that */
/*      * we break all programs using clone on older kernels.. */
/*      * We should try to fallback to the bpt-trick if this fails, but right */
/*      * now we don't. */
/*      *\/ */
    
/*     /\* TODO: actually change the flags *\/ */
/*   } else { */
/*     if (!(tcp->flags & TCB_FOLLOWFORK)) */
/*       return 0; */

/*     if (syserror(tcp)) */
/*       return 0; */
    
/*     pid = tcp->u_rval; */
/*     if ((tcpchild = alloctcb(pid)) == NULL) { */
/*       fprintf(stderr, " [tcb table full]\n"); */
/*       kill(pid, SIGKILL); /\* XXX *\/ */
/*       return 0; */
/*     } */
    
/*     /\* For fork we need to re-attach, but thanks to CLONE_PTRACE we're */
/*      * already attached. */
/*      *\/ */
/*     tcpchild->flags |= TCB_ATTACHED; */
/*     tcp->nchildren++;                                       		        */
/*     //fprintf(stderr, "Process %d attached\n", pid); */
/*   } */
/*   return 0; */
/* } */
/* #endif */


#ifdef SYS_clone
int
internal_clone(tcp) /* CONTROLLARE: AGGIUNGERE LA PARTE DI CLONING DELL'AUTOMA */
struct tcb *tcp;
{
	struct tcb *tcpchild;
	int pid;
	if (entering(tcp)) {
	  //if (!followfork)
	  //	return 0;
		if (nprocs == MAX_PROCS) {
			tcp->flags &= ~TCB_FOLLOWFORK;
			fprintf(stderr, "sys_fork: tcb table full\n");
			return 0;
		}
		tcp->flags |= TCB_FOLLOWFORK;


		if (setbpt(tcp) < 0)
			return 0;
	} else {
		int bpt = tcp->flags & TCB_BPTSET;

		if (!(tcp->flags & TCB_FOLLOWFORK))
			return 0;

		if (syserror(tcp)) {
			if (bpt)
				clearbpt(tcp);
			return 0;
		}

		pid = tcp->u_rval;
		if ((tcpchild = alloctcb(pid)) == NULL) {
			if (bpt)
				clearbpt(tcp);
			fprintf(stderr, " [tcb table full]\n");
			kill(pid, SIGKILL); /* XXX */
			return 0;
		}

		/* Attach to the new child */
		if (ptrace(PTRACE_ATTACH, pid, (char *) 1, 0) < 0) {
			if (bpt)
				clearbpt(tcp);
			perror("PTRACE_ATTACH");
			fprintf(stderr, "Too late?\n");
			droptcb(tcpchild);
			return 0;
		}

		if (bpt)
			clearbpt(tcp);

		tcpchild->flags |= TCB_ATTACHED;
		if (bpt) {
			tcpchild->flags |= TCB_BPTSET;
			tcpchild->baddr = tcp->baddr;
			memcpy(tcpchild->inst, tcp->inst,
				sizeof tcpchild->inst);
		}
		//newoutf(tcpchild);
		tcpchild->parent = tcp;
		tcp->nchildren++;
		
		fprintf(stderr, "Process %d attached\n", pid);
	}
	return 0;
}
#endif


int
change_syscall(tcp, new)
struct tcb *tcp;
int new;
{
	/* Attempt to make vfork into fork, which we can follow. */
	if (ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(ORIG_EAX * 4), new) < 0) 
		return -1;
	
	return 0;
}


int
internal_fork(tcp)
struct tcb *tcp;
{
	struct tcb *tcpchild;
	int pid;
	int dont_follow = 0;

	if (tcp->scno == SYS_vfork) {
		/* Attempt to make vfork into fork, which we can follow. */
		if (change_syscall(tcp, SYS_fork) < 0)
			dont_follow = 1;
	}

	if (entering(tcp)) 
	  {
	    if (dont_follow)
	      return 0;
	    
	    if (nprocs == MAX_PROCS) 
	      {
		tcp->flags &= ~TCB_FOLLOWFORK;
		fprintf(stderr, "sys_fork: tcb table full\n");
		return 0;
	      }

	    tcp->flags |= TCB_FOLLOWFORK;
	    if (setbpt(tcp) < 0)
	      return 0;
	  }
	else 
	  {
	    int bpt = tcp->flags & TCB_BPTSET;
	    automa_state *child_state_ptr;
	    
	    if (!(tcp->flags & TCB_FOLLOWFORK))
	      return 0;
	    
	    if (bpt)
	      clearbpt(tcp);
	    
	    if (syserror(tcp))
	      return 0;
	    
	    pid = tcp->u_rval;
	    if ((tcpchild = alloctcb(pid)) == NULL) 
	      {
		fprintf(stderr, " [tcb table full]\n");
		kill(pid, SIGKILL); /* XXX */
		return 0;
	      }
	    
	    /*****************************************************************************
	     ******************** AGGIORNAMENTO AUTOMA ***********************************
	     *****************************************************************************/
	    if ((child_state_ptr = clone_automa_state(tcp->process_state)) == NULL)
	      {
		fprintf(stderr,"internal fork: cloning process_state impossible");    
		kill(pid, SIGKILL); /* XXX */
		droptcb(tcpchild);
		return 0;
	      }
	    
	    tcpchild->process_state = *child_state_ptr;
	    /* non sono in grado di intercettare la fork, in uscita del processo figlio percio' la simulo */
	    tcpchild->flags |= TCB_INSYSCALL;
	    tcpchild->scno = 2;        /* fork */
	    tcpchild->u_rval = 0;
	    
	    tcpchild->process_state = update_state(atm, tcpchild, tcpchild->process_state);
	    
	    tcpchild->flags &= ~TCB_INSYSCALL;
	    /******************************************************************************/

	    if (ptrace(PTRACE_ATTACH, pid, (char *) 1, 0) < 0) 
	      {
		perror("PTRACE_ATTACH");
		fprintf(stderr, "Too late?\n");
		droptcb(tcpchild);
		return 0;
	      }

	    tcpchild->flags |= TCB_ATTACHED;
	    /* Child has BPT too, must be removed on first occasion */
	    if (bpt) 
	      {
		tcpchild->flags |= TCB_BPTSET;
		tcpchild->baddr = tcp->baddr;
		memcpy(tcpchild->inst, tcp->inst, sizeof tcpchild->inst);
	      }

	    tcpchild->parent = tcp;
	    tcp->nchildren++;
	    fprintf(stderr, "Process %d attached\n", pid);
	  }
	return 0;
}


int
internal_exec(tcp) /* Che ci sta a fare ???? */
struct tcb *tcp;
{
  return 0;
}


int
internal_wait(tcp)
struct tcb *tcp;
{
  if (entering(tcp)) 
    {
      if (upeek(tcp->pid, 8, &tcp->u_arg[2]) < 0)
	{
	  fprintf(stderr, "internal_wait: upeek non riuscita\n");
	  return -1;
	}
      tcp->arg_state[2] = UPEEK;
      
      /* WTA: fix bug with hanging children */
      if (!(tcp->u_arg[2] & WNOHANG) && tcp->nchildren > 0)
	{
	  /* There are traced children */
	  if (upeek(tcp->pid, 0, &tcp->u_arg[0]) < 0)
	    {
	      fprintf(stderr, "internal_wait: upeek non riuscita\n");
	      return -1;
	    }
	  tcp->arg_state[0] = UPEEK;
	  
	  tcp->flags |= TCB_SUSPENDED;
	  tcp->waitpid = tcp->u_arg[0];
	}
    }
  return 0;
}



int
internal_sigsetmask(tcp)
struct tcb *tcp;
{
  int i;
  
  for (i = 0; i < tcp->u_nargs; i++)
    {
      if (tcp->arg_state[i] == EMPTY)
	{
	  if (upeek(tcp->pid, i*4, &tcp->u_arg[i]) < 0)
	    {
	      fprintf(stderr, "internal_sigsetmask: upeek\n");
	      return -1;
	    }
	  tcp->arg_state[i] = UPEEK;
	}
    }
  
  if (entering(tcp)) 
    {
      if ((tcp->u_arg[0] & sigmask(SIGTRAP))) 
	{  
	  /* Mark attempt to block SIGTRAP */
	  tcp->flags |= TCB_SIGTRAPPED;
	  /* Send unblockable signal */
	  kill(tcp->pid, SIGSTOP);
	}
    }
  return 0;
}


int
internal_oldselect(tcp) 
struct tcb *tcp;
{
  int i;

  if (exiting(tcp))
    {
      for (i = 0; i < tcp->u_nargs; i++) 
	if (tcp->arg_state[i] == TAKEN)
	  free( (int *) tcp->u_arg[i]);
    }

  if (upeek(tcp->pid, 0, &tcp->u_arg[0]) < 0)
    {
      fprintf(stderr,"internal_oldselect: sys_oldselect -  upeek failed\n");
      return -1;
    }

  if (umoven(tcp, tcp->u_arg[0], 5 * sizeof(long), (char *) (tcp->u_arg)) < 0) 
    {
      fprintf(stderr,"internal_oldselect: sys_oldselect - umove failed\n");
      return -1;
    }

  for ( i=0 ; i<5 ; i++ )
    tcp->arg_state[i] = UPEEK;

 return 0;
}


int
internal_sigreturn(tcp) /* CONTROLLARE */
struct tcb *tcp;
{
  long esp;
  struct sigcontext_struct sc;
  
  if (entering(tcp)) 
    {
      tcp->u_arg[0] = 0;
      if (upeek(tcp->pid, 4*UESP, &esp) < 0)
	return 0;
      if (umove(tcp, esp, &sc) < 0)
	return 0;
      tcp->u_arg[0] = 1;
      tcp->u_arg[1] = sc.oldmask;
    }
  else { // AGGIUNTO
    tcp->u_rval = tcp->u_error = 0;
  }

  return 0;
}



#ifdef HAVE_SIGACTION

struct old_sigaction {
  __sighandler_t __sa_handler;
  unsigned long sa_mask;
  unsigned long sa_flags;
  void (*sa_restorer)(void);
};
#define SA_HANDLER __sa_handler


#ifndef SA_HANDLER                                                           
#define SA_HANDLER sa_handler                                                
#endif

int 
internal_sigaction(tcp)
struct tcb *tcp;
{
  struct old_sigaction *sa_ptr;

  if (take_arg(0, tcp) == NULL)
    {
      fprintf(stderr, "internal_sigaction: take_arg(0, tcp) failed\n");
      return 0;
    }
  
  if (entering(tcp)) 
    {
      if ((sa_ptr = (struct old_sigaction *) take_arg(1, tcp)) == NULL)
	{
	  fprintf(stderr, "internal_sigaction: take_arg(1, tcp) failed\n");
	  return 0;
	}
    }
  else
    {
      if ((sa_ptr = (struct old_sigaction *) take_arg(2, tcp)) == NULL)
	{	
	  fprintf(stderr, "internal_sigaction: take_arg(2, tcp) failed\n");
	  return 0;
	}
    }

  switch ((long) (sa_ptr->SA_HANDLER)) {
  case (long) SIG_ERR:
  case (long) SIG_DFL:
    break;
  case (long) SIG_IGN:
    if (*((int*)tcp->u_arg[0]) == SIGTRAP) 
      {
	tcp->flags |= TCB_SIGTRAPPED;
	kill(tcp->pid, SIGSTOP);
      }
    break;
  default:
    if (*((int*)tcp->u_arg[0]) == SIGTRAP) 
      {
	tcp->flags |= TCB_SIGTRAPPED;
	kill(tcp->pid, SIGSTOP);
      }
  }
  return 0;
}

#endif


int
internal_signal(tcp)
struct tcb *tcp;
{
  int i;
  /* CONTROLLARE questo ciclo */
  for (i = 0; i < tcp->u_nargs; i++)
    {
      if (tcp->arg_state[i] == EMPTY)
	{
	  if (upeek(tcp->pid, i*4, &tcp->u_arg[i]) < 0)
	    {
	      fprintf(stderr, "internal_signal: upeek failed\n");
	      return -1;
	    }
	  tcp->arg_state[i] = UPEEK;
	}
    }
  
  if (entering(tcp)) 
    {
      switch (tcp->u_arg[1]) {
      case (int) SIG_ERR:
      case (int) SIG_DFL:
	break;
      case (int) SIG_IGN:      
	if (tcp->u_arg[0] == SIGTRAP) 
	  {
	    tcp->flags |= TCB_SIGTRAPPED;
	    kill(tcp->pid, SIGSTOP);
	  }
	break;
      default:
	if (tcp->u_arg[0] == SIGTRAP) 
	  {
	    tcp->flags |= TCB_SIGTRAPPED;
	    kill(tcp->pid, SIGSTOP);
	  }
      }
    }
  return 0;
}

/* Non viene mai eseguita */
int 
internal_none(tcp)
struct tcb *tcp;
{
  return 0;
}
