/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: io.c,v 1.2 2003/02/17 15:26:23 frassi Exp $
 */

#include "gets.h" /* get_error */
#include "util.h" /* arg_alloc */
#include "runtime_util.h"
#include "pars_funct.h" /* get_nargs */


#include <fcntl.h>
#include <sys/uio.h>
#include <asm/ioctl.h>

#ifdef HAVE_TERMIO_H
#include <termio.h>
#endif /* HAVE_TERMIO_H */

#include <termios.h>

#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif

/* #include <fcntl.h> */
/* #if HAVE_SYS_UIO_H */
/* #include <sys/uio.h> */
/* #endif */

/* #ifdef HAVE_LONG_LONG_OFF_T */
/* /\* */
/*  * Hacks for systems that have a long long off_t */
/*  *\/ */

/* #define sys_pread64	sys_pread */
/* #define sys_pwrite64	sys_pwrite */
/* #endif */



int 
print_read(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("read(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], info->u_rval);
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_read(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  int read_bytes;

  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp))
      {
	if (tcp->u_rval >= 0)
	  read_bytes = tcp->u_rval;
	else
	  {
	    //int *ptr;
	    /*
	      if (syserror(tcp)) // l'errore che da qui e' il 512 - ERESTARTSYS cioe' la syscall dovrebbe essere riavviata 
	      fprintf(stderr, "err%d -- rval%d", (tcp)->u_error,  (tcp)->u_rval);fflush(NULL);
	    */

#ifdef DEBUG_WIZ
	    fprintf(stderr, "sys_read: syscall failed\n");
#endif

	    /*la read e' fallita, leggo il buffer finche' posso...*/   /* CONTROLLARE */
	    /*if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	      {
		fprintf(stderr, "sys_read: take_arg(2, tcp)\n");
		return -1;
	      }
	      read_bytes = *ptr;*/

	    return -1;
	  }
      

	((char*) tcp->u_arg[1]) = get_str(tcp, tcp->u_arg[1], read_bytes);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_read: get_str\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }	
    break;
  }

  return 0;
}



int 
print_write(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("write(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], *((unsigned int *)info->u_arg[2]));
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int
sys_write(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  int read_bytes, *ptr;

  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
      {
	fprintf(stderr, "sys_write: take_arg(2, tcp)\n");
	return -1;
      }
    read_bytes = *ptr;

    ((char*) tcp->u_arg[1]) = get_str(tcp, tcp->u_arg[1], read_bytes);
    if ( get_error )
      {
	fprintf(stderr, "sys_write: get_str\n");
	tcp->arg_state[1] = EMPTY;
	return -1;
      }
    tcp->arg_state[1] = TAKEN;
    break;
  }

  return 0;
}





void 
free_writev(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i,j,count=0;

  if (arg_state[2] == TAKEN)
    count=*((int*) u_arg[2]);

  /*se non e' stato preso il secondo argomento, non puo' essere stato preso il primo*/

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 2:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 1:
	if (arg_state[1] == TAKEN)
	  if (u_arg[1] && (count>=0))
	    {
	      struct iovec *iov = (struct iovec *) u_arg[1];
	      for (j=0;j<count;j++)
		free(iov[j].iov_base);
	      free(iov);
	    }
	arg_state[1] = EMPTY;
	break;
      }
    }
  return;
}


void 
free_readv(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i,j,count=0;

  if (arg_state[2] == TAKEN)
    count=*((int*)u_arg[2]);

  /*se non e' stato preso il secondo argomento, non puo' essere stato preso il primo*/

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 2:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 1:
	if (arg_state[1] == TAKEN)
	  if (u_arg[1] && (count>=0))
	    {
	      struct iovec *iov = (struct iovec *) u_arg[1];
	      for (j=0;j<count;j++)
		free(iov[j].iov_base);
	      free(iov);
	    }
	arg_state[1] = EMPTY;
	break;
      }
    }
  return;
}  




int 
print_readv(info)
msg_building_info *info;
{
  int i, j, len;
  int nargs=get_nargs(info->scno);
  struct iovec *iov;
  
  tprintf("readv(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    iov=(struct iovec *) info->u_arg[1];
	    
	    if (iov)
	      {
		len = *((int*) info->u_arg[2]);
		
		tprintf("[");
		for (j = 0; j < len; j++) 
		  {
		    if (j)
		      tprintf(", ");
		    tprintf("{");
		    realprintstr((char*) iov[j].iov_base, ((unsigned long)iov[j].iov_len));
		    tprintf(", %lu}", ((unsigned long)iov[j].iov_len));
		  }
		tprintf("]");
	      }
	    else
	      tprintf("[NULL]");
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) descrittore del file da cui leggere (int)
  -) array di buffer destinato a contenere i dati letti (const struct iovec *)
  -) numero di buffer nell'array (int)
*/
int 
sys_readv(tcp, n_arg)
struct tcb *tcp;
int n_arg; 
{
  int len, *ptr;
  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_readv: syscall failed\n");
	    return -1;
	  }
	
	if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	  {
	    fprintf(stderr, "sys_readv: take_arg(2, tcp)\n");
	    return -1;
	  }
	len = *ptr;

	tcp->u_arg[n_arg] = (long) get_iovec(tcp, (struct iovec *)tcp->u_arg[n_arg], len);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_readv: get_iovec\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[1] = TAKEN;
      }
    break;
  }
  return 0;
}




int 
print_writev(info)
msg_building_info *info;
{
  int i, j, len;
  int nargs=get_nargs(info->scno);
  struct iovec *iov;
  
  tprintf("writev(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    iov=(struct iovec *) info->u_arg[1];

	    if (iov)
	      {
		len = *((int*) info->u_arg[2]);

		tprintf("[");
		for (j = 0; j < len; j++) 
		  {
		    if (j)
		      tprintf(", ");
		    tprintf("{");
		    realprintstr((char*) iov[j].iov_base, ((unsigned long)iov[j].iov_len));
		    tprintf(", %lu}", ((unsigned long)iov[j].iov_len));
		  }
		tprintf("]");
	      }
	    else
	      tprintf("[NULL]");
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_writev(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  int len, *ptr;

  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
      {
	fprintf(stderr, "sys_writev: take_arg(2, tcp)\n");
	return -1;
      }
    len = *ptr;
    
    tcp->u_arg[n_arg] = (long) get_iovec(tcp, (struct iovec *)tcp->u_arg[n_arg], len);
    if ( get_error )
      {
	fprintf(stderr, "sys_writev: get_iovec\n");
	tcp->arg_state[1] = EMPTY;
	return -1;
      }
    tcp->arg_state[1] = TAKEN;
    break;
  }
  return 0;
}




int 
print_pread(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("pread(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], info->u_rval);
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  case 3:
	    tprintf("%llu", *(unsigned long long *)info->u_arg[3]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 4
  -) descrittore del file da cui leggere (int)
  -) buffer destinato a contenere i dati letti (void *)
  -) byte da leggere (size_t == unsigned int)
  -) offset dall'inizio del file, a partire da dove iniziare a leggere (off_t == long)
*/

int 
sys_pread(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int len;
  switch (n_arg) {
  case 0:
  case 2:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_pread: syscall failed\n");
	    return -1;
	  }
      
	len = tcp->u_rval;
	
	tcp->u_arg[n_arg] = (long) get_str(tcp,tcp->u_arg[n_arg],len);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_pread: get_str\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  return 0;
}



int 
print_pwrite(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("pwrite(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], *((unsigned int*)info->u_arg[2]));
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  case 3:
	    tprintf("%llu", *(unsigned long long *)info->u_arg[3]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 4
  -) descrittore del file da cui leggere (int)
  -) buffer contenente i dati da scrivere (void *)
  -) byte da scrivere (size_t == unsigned int)
  -) offset dall'inizio del file, a partire da dove iniziare a scrivere (off_t == long)
*/
int 
sys_pwrite(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int len, *ptr;
  switch (n_arg) {
  case 0:
  case 2:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    
//   if (entering(tcp))  CONTROLLARE (se volessi controllare questo arg solo in uscita? ) 
    //  {
	if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	  {
	    fprintf(stderr, "sys_pwrite: take_arg(2, tcp)\n");
	    return -1;
	  }
	len = *ptr;
	
	tcp->u_arg[n_arg] = (long) get_str(tcp,tcp->u_arg[n_arg],len);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_pwrite: get_str\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
    //  }
    break;
  }
  return 0;
}



int 
print_sendfile(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("sendfile(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	  case 1:
	    tprintf("%ld", *((long*)info->u_arg[i]));
	    break;
	  case 2:
	    if (!info->u_arg[2])
	      tprintf("NULL");
	    else
	      tprintf("[%lu]", *((off_t*)info->u_arg[2]));
	    break;
	  case 3:
	    tprintf("%lu", *(unsigned long *)info->u_arg[3]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 4
  -) descrittore del file in scrittura (int)
  -) descrittore del file in lettura (int)
  -) puntatore ad una variabile con il valore dell'offset da dove iniziare a leggere (off_t * == long *)
  -) numero di byte da copiare (size_t == unsigned int)
*/
int 
sys_sendfile(tcp, n_arg)
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
  case 1:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  case 2:
  //if (entering(tcp)) 
  //  {
	tcp->u_arg[n_arg] = (long) get_num(tcp,tcp->u_arg[n_arg]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_sendfile: get_num\n");
	    tcp->arg_state[n_arg] = TAKEN;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
   //  }
    break;
  }
  return 0;
}



int 
print_ioctl(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  char* symbol;
  
  tprintf("ioctl(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    symbol = ioctl_lookup(*((long*)info->u_arg[1]));
	    if (symbol)
	      tprintf("%s", symbol);
	    else
	      tprintf("NULL");
	    break;
	  case 2:
	    if (ioctl_decode(info, ((long*)info->u_arg[1]), ((long*)info->u_arg[2])) == 0)
	      tprintf("%#lx", info->u_arg[2]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) descrittore di file (int)
  -) valore che codifica se si tratta di un parametro in ingresso o in uscita e la dimensione del
     terzo argomento (int)
  -) argomento (char *)
*/
int 
sys_ioctl(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct termios *tios;
  struct termio *tio;
  struct winsize *ws;
#ifdef TIOCGSIZE
  struct ttysize *ts;
#endif

  long addr;
  
  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp, n_arg);
    break;
  case 2:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_ioctl: syscall failed\n");
	    return -1;
	  }
	
	if (take_arg(1, tcp) == NULL )
	  {
	    fprintf(stderr, "sys_ioctl: take_arg(1, tcp)\n");
	    return -1;
	  }
	  
	
	switch(*((int*)tcp->u_arg[1])) { /* CONTROLLARE 	switch ((code >> 8) & 0xff) {...... */
#ifdef TCGETS
	case TCGETS:
	  if (syserror(tcp))
	    return 0;
	case TCSETS:
	case TCSETSW:
	case TCSETSF:

	  tios=(struct termios*) malloc(sizeof(struct termios));

	  if (!tios)
	    {
	      fprintf(stderr,"sys_ioctl: unable to alloc termios\n");
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  if (umove(tcp, tcp->u_arg[2], tios) < 0)
	    {
	      fprintf(stderr,"sys_ioctl: unable to umove termios\n");
	      free(tios);
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  ((struct termios *) tcp->u_arg[n_arg]) = tios;
	  tcp->arg_state[n_arg] = TAKEN;
	  break;
#endif /* TCGETS */

#ifdef TCGETA
	case TCGETA:
	  if (syserror(tcp))
	    return 0;
	case TCSETA:
	case TCSETAW:
	case TCSETAF:

	  tio=(struct termio*) malloc(sizeof(struct termio));
	  
	  if (!tio)
	    {
	      fprintf(stderr,"sys_ioctl: unable to alloc termio\n");
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  if (umove(tcp, tcp->u_arg[2], tio) < 0)
	    {
	      fprintf(stderr,"sys_ioctl: unable to umove termio\n");
	      free(tio);
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  ((struct termio *) tcp->u_arg[n_arg]) = tio;
	  tcp->arg_state[n_arg] = TAKEN;
	  break;
#endif /* TCGETA */
	  
	  /* ioctls with winsize or ttysize args */
#ifdef TIOCGWINSZ
	case TIOCGWINSZ:
	  if (syserror(tcp))
	    return 0;
	case TIOCSWINSZ:

	  ws=(struct winsize*) malloc(sizeof(struct winsize));
	  
	  if (!ws)
	    {
	      fprintf(stderr,"sys_ioctl: unable to alloc winsize\n");
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  if (umove(tcp, tcp->u_arg[2], ws) < 0)
	    {
	      fprintf(stderr,"sys_ioctl: unable to umove winsize\n");
	      free(ws);
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  ((struct winsize *) tcp->u_arg[n_arg]) = ws;
	  tcp->arg_state[n_arg] = TAKEN;
	  break;
#endif /* TIOCGWINSZ */

#ifdef TIOCGSIZE
	case TIOCGSIZE:
	  if (syserror(tcp))
	    return 0;
	case TIOCSSIZE:
	  ts=(struct ttysize*) malloc(sizeof(struct ttysize));
	  
	  if (!ts)
	    {
	      fprintf(stderr,"sys_ioctl: unable to alloc ttysize\n");
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  if (umove(tcp, tcp->u_arg[2], ts) < 0)
	    {
	      fprintf(stderr,"sys_ioctl: unable to umove ttysize\n");
	      free(ts);
	      tcp->arg_state[n_arg]=EMPTY;
	      return -1;
	    }

	  ((struct ttysize *) tcp->u_arg[n_arg]) = ts;
	  tcp->arg_state[n_arg] = TAKEN;
	  break;
#endif
	/* ioctls with a direct decodable arg */

	case TCXONC:
	case TCFLSH:
	  arg_alloc(tcp,n_arg);
	  break;

	/* ioctls with an indirect parameter displayed as modem flags */

#ifdef TIOCMGET
	case TIOCMGET:
	case TIOCMBIS:
	case TIOCMBIC:
	case TIOCMSET:
	  addr=tcp->u_arg[n_arg];
	  (long*) tcp->u_arg[n_arg] = (long*) malloc(sizeof(long));
	  if (tcp->u_arg[n_arg] == 0) 
	    { 
	      fprintf(stderr, "sys_ioctl: no memory\n"); 
	      return -1; 
	    } 
	  if (umove(tcp, addr, (long*) tcp->u_arg[2]) < 0)
	    return 0;
	  tcp->arg_state[n_arg] = TAKEN;
	  break;
#endif /* TIOCMGET */

	/* ioctls with an indirect parameter displayed in decimal */

	case TIOCSPGRP:
	case TIOCGPGRP:
#ifdef TIOCGETPGRP
	case TIOCGETPGRP:
#endif
#ifdef TIOCSETPGRP
	case TIOCSETPGRP:
#endif
#ifdef FIONREAD
	case FIONREAD:
#endif
	case TIOCOUTQ:
#ifdef FIONBIO
	case FIONBIO:
#endif
#ifdef FIOASYNC
	case FIOASYNC:
#endif
#ifdef FIOGETOWN
	case FIOGETOWN:
#endif
#ifdef FIOSETOWN
	case FIOSETOWN:
#endif
#ifdef TIOCGETD
	case TIOCGETD:
#endif
#ifdef TIOCSETD
	case TIOCSETD:
#endif
#ifdef TIOCPKT
	case TIOCPKT:
#endif
#ifdef TIOCREMOTE
	case TIOCREMOTE:
#endif
#ifdef TIOCUCNTL
	case TIOCUCNTL:
#endif
#ifdef TIOCTCNTL
	case TIOCTCNTL:
#endif
#ifdef TIOCSIGNAL
	case TIOCSIGNAL:
#endif
#ifdef TIOCSSOFTCAR
	case TIOCSSOFTCAR:
#endif
#ifdef TIOCGSOFTCAR
	case TIOCGSOFTCAR:
#endif
#ifdef TIOCISPACE
	case TIOCISPACE:
#endif
#ifdef TIOCISIZE
	case TIOCISIZE:
#endif
#ifdef TIOCSINTR
	case TIOCSINTR:
#endif
#ifdef TIOCSPTLCK
	case TIOCSPTLCK:
#endif
#ifdef TIOCGPTN
	case TIOCGPTN:
#endif
	  ((long*) tcp->u_arg[n_arg]) = get_num(tcp, tcp->u_arg[n_arg]);
	  if ( get_error ) 
	    {
	      fprintf(stderr, "sys_ioctl: get_num\n");
	      tcp->arg_state[n_arg] = EMPTY;
	      return -1;
	    }
	  tcp->arg_state[n_arg] = TAKEN;
	  break;
	  
	/* ioctls with an indirect parameter displayed as a char */

#ifdef TIOCSTI
	case TIOCSTI:
#endif

	  ((char*)tcp->u_arg[n_arg]) = get_str(tcp, tcp->u_arg[n_arg], 1);
	  if ( get_error )
	    {
	      fprintf(stderr, "sys_ioctl: get_str\n");
	      tcp->arg_state[n_arg] = EMPTY;
	      return -1;
	    }
	  tcp->arg_state[n_arg] = TAKEN;
	  break;
	  
	/* ioctls with no parameters */

#ifdef TIOCSCTTY
	case TIOCSCTTY:
#endif
#ifdef TIOCNOTTY
	case TIOCNOTTY:
#endif
#ifdef FIOCLEX
	case FIOCLEX:
#endif
#ifdef FIONCLEX
	case FIONCLEX:
#endif
#ifdef TIOCCONS
	case TIOCCONS:
#endif
		return 1;

	/* ioctls which are unknown */

	default:
	  return 0;
	} // switch	  
      } // if exiting
    break;
    
  } // switch esterno
  return 0;
}

