/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1993 Ulrich Pegelow <pegelow@moorea.uni-muenster.de>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: ipc.c,v 1.2 2003/02/17 15:30:33 frassi Exp $
 */

#include "gets.h" /* get_error */
#include "util.h" /* arg_alloc set_copy umove */
#include "runtime_util.h"
#include "pars_funct.h" /* get_nargs */

#include <sys/ipc.h>
#include <sys/sem.h>
#include <sys/msg.h>
#include <sys/shm.h>

/***********Aggiunti per compilare.... se inserisco nel Makefile -D_GNU_SOURCE allora non da problemi.. BOOOOO **************/
#ifndef IPC_INFO
#define IPC_INFO 3
#endif
#ifndef MSG_EXCEPT
#define MSG_EXCEPT  020000
#endif
/******************************/
#ifndef MSG_STAT
#define MSG_STAT 11
#endif
#ifndef MSG_INFO
#define MSG_INFO 12
#endif
#ifndef SHM_STAT
#define SHM_STAT 13
#endif
#ifndef SHM_INFO
#define SHM_INFO 14
#endif
#ifndef SEM_STAT
#define SEM_STAT 18
#endif
#ifndef SEM_INFO
#define SEM_INFO 19
#endif

static struct xlat msgctl_flags[] = {  /* utilizzati in sys_msgctl */ 
  { IPC_RMID,	"IPC_RMID"	},
  { IPC_SET,	"IPC_SET"	},
  { IPC_STAT,	"IPC_STAT"	},
  { IPC_INFO,	"IPC_INFO"	},
  { MSG_STAT,	"MSG_STAT"	},
  { MSG_INFO,	"MSG_INFO"	},
  { 0,		NULL		},
};

static struct xlat semctl_flags[] = {  /* utilizzati in sys_semctl */
  { IPC_RMID,	"IPC_RMID"	},
  { IPC_SET,	"IPC_SET"	},
  { IPC_STAT,	"IPC_STAT"	},
  { IPC_INFO,	"IPC_INFO"	},
  { SEM_STAT,	"SEM_STAT"	},
  { SEM_INFO,	"SEM_INFO"	},
  { GETPID,	"GETPID"	},
  { GETVAL,	"GETVAL"	},
  { GETALL,	"GETALL"	},
  { GETNCNT,	"GETNCNT"	},
  { GETZCNT,	"GETZCNT"	},
  { SETVAL,	"SETVAL"	},
  { SETALL,	"SETALL"	},
  { 0,		NULL		},
};

static struct xlat shmctl_flags[] = {  /* utilizzati in sys_shmctl */
  { IPC_RMID,	"IPC_RMID"	},
  { IPC_SET,	"IPC_SET"	},
  { IPC_STAT,	"IPC_STAT"	},
  { IPC_INFO,	"IPC_INFO"	},
  { SHM_STAT,	"SHM_STAT"	},
  { SHM_INFO,	"SHM_INFO"	},
  { SHM_LOCK,	"SHM_LOCK"	},
  { SHM_UNLOCK,	"SHM_UNLOCK"	},
  { 0,		NULL		},
};

static struct xlat resource_flags[] = {  /* utilizzati in sys_msgget, sys_semget, sys_shmget */   
  { IPC_CREAT,	"IPC_CREAT"	},
  { IPC_EXCL,	"IPC_EXCL"	},
  { IPC_NOWAIT,	"IPC_NOWAIT"	},
  { 0,		NULL		},
};

static struct xlat shm_flags[] = {  /* utilizzati in sys_shmat */
  { SHM_REMAP,	"SHM_REMAP"	},
  { SHM_RDONLY,	"SHM_RDONLY"	},
  { SHM_RND,	"SHM_RND"	},
  { 0,		NULL		},
};

static struct xlat msg_flags[] = {  /* utilizzati in sys_msgsnd e sys_msgrcv */
  { MSG_NOERROR,"MSG_NOERROR"	},
  { MSG_EXCEPT,	"MSG_EXCEPT"	},
  { IPC_NOWAIT,	"IPC_NOWAIT"	},
  { 0,		NULL		},
};


/* sono definite le sys_... estratte da sys_ipc (syscall n. 117) */


// potrebbe anche essere tolta
int 
print_ipc(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("ipc(");
  
  for (i=0;i<nargs;i++)
    {
      /* 
	 attenzione, funziona solo perche' il 5o argomento
	 che e' un void* non viene mai preso dalla sys_ipc
      */
      if (info->arg_state[i]==TAKEN)
	{
	  if (info->u_arg[i])
	    tprintf("%d", *((int*)info->u_arg[i]));
	  else 
	    tprintf("0");
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}

/*
int 
sys_ipc(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  if (n_arg!=4)
    {
      arg_alloc(tcp,n_arg);
      return 0;
    }
  else
    if(((void*) tcp->u_arg[n_arg])==NULL)
      tcp->arg_state[n_arg]=TAKEN;

  return -1;
}
*/



int 
print_msgget(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("msgget(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    if (*((unsigned long*)info->u_arg[0]) != 0)
	      tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    else
	      tprintf("IPC_PRIVATE");
	    break;
	  case 1:
	    if (printflags(resource_flags, *((int*)info->u_arg[1])) != 0)
	      tprintf("|");
	    tprintf("%#lo", *((unsigned long*)info->u_arg[1]) & 0666);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 2
  -) chiave (key_t == int)
  -) flags (int)
*/
int 
sys_msgget(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_msgctl(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  char *cmd;
  
  tprintf("msgctl(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    cmd = get_xlat_str(msgctl_flags, *((int*)info->u_arg[1]));
	    tprintf("%s", cmd == NULL ? "MSG_???" : cmd);
	    break;
	  case 2:
	    tprintf("struct msqid_ctl");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) identificatore della coda di messaggi (int)
  -) comando (int)
  -) struttura dati con info sulla coda messaggi (struct msqid_ds *)
*/
int 
sys_msgctl(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct msqid_ds *mq;

  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  case 2:
    if (upeek(tcp->pid, 3 * 4, &tcp->u_arg[3]) < 0)
      {   
	fprintf(stderr, "sys_msgctl: upeek failed\n");
	return -1;
      }
    tcp->arg_state[3] = UPEEK;

    if (tcp->u_arg[3])
      if (set_copy(tcp, n_arg, tcp->u_arg[3], (long *)&mq, sizeof(struct msqid_ds), "msgctl"))
	{
	  fprintf(stderr, "sys_msgctl: set_copy\n");
	  return -1;
	}
    tcp->arg_state[n_arg] = TAKEN; /* CONTROLLARE */
    break;
  }
  return 0;
}



struct mymsgbuf {
  long mtype;
  char *mtext;
};


void 
free_msgsnd(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i;

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 2:
      case 3:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 1:
	if (arg_state[1] == TAKEN)
	  if (u_arg[1])
	    {
	      struct mymsgbuf *buf = (struct mymsgbuf *) u_arg[1];
	      if (buf->mtext)
		free (buf->mtext);
	      free(buf);
	    }
	arg_state[1] = EMPTY;
	break;
      }
    }
  return;
}  



int 
print_msgsnd(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  struct mymsgbuf *buf;
  
  tprintf("msgsnd(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[i]));
	    break;
	  case 1:
	    buf=((struct mymsgbuf *) info->u_arg[1]);
	    tprintf("{%lu, ", buf->mtype);
	    realprintstr(buf->mtext, *((unsigned int*)info->u_arg[2]));
	    tprintf("}");
	    break;
	  case 3:
	    if (printflags(msg_flags, *((int*)info->u_arg[3])) == 0)
	      tprintf("0");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 4
  -) identificatore della coda di messaggi (int)
  -) struttura dati contenente il messaggio da accodare (struct msgbuf *)
  -) dimensione dell'array di dati interno alla struttura (int)
  -) flag per specificare il comportamento della syscall (int)

  secondo la pagina del man e' cosi', in realta' nel codice originale dell'strace gli
  argomenti figurano in ordine diverso:

  secondo il           man                    strace

  arg[0]               int msgid              int msgid
  arg[1]               struct msgbuf *msgp    int msgsz
  arg[2]               int msgsz              int msgflg
  arg[3]               int msgflg             struct msgbuf *msgp

  in questo caso il codice e' fatto in maniera tale da conformarsi alla
  pagina del man
*/
int 
sys_msgsnd(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct mymsgbuf* buf;

  long tmp;
  int i, sz, flg;

  switch (n_arg) {
  case 0: /* int msgid */
    arg_alloc(tcp,n_arg);
    break;
  case 1: /* struct msgp */
  case 2: /* int msgsz */
  case 3: /* msgflag */
    
    for (i=1;i<=3;i++)
      {
	if (tcp->arg_state[i] == EMPTY)
	  {
	    if (upeek(tcp->pid, (i) * 4, &tcp->u_arg[i]) < 0)
	      {   
		fprintf(stderr, "sys_msgsnd: upeek\n");
		return -1;
	      }
	    tcp->arg_state[i] = UPEEK;
	  }
      }

    buf=((struct mymsgbuf *) malloc(sizeof(struct mymsgbuf)));

    if (!buf)
      {
	fprintf(stderr,"sys_msgsnd: malloc buf failed\n");
	return -1;
      }
    
    sz = tcp->u_arg[1]; 
    flg = tcp->u_arg[2];

    if (umove(tcp, tcp->u_arg[3], &tmp) < 0)
      {
	fprintf(stderr,"sys_msgsnd: umove arg 1 failed\n");
	free(buf);
	return -1;
      }
    else
      {
	buf->mtype=tmp;
	buf->mtext=get_str(tcp, tcp->u_arg[3] + sizeof(long),sz);
	
	if ( get_error )
	  {
	    fprintf(stderr, "sys_msgsnd: get_str\n");
	    tcp->arg_state[1] = EMPTY;
	    return -1;
	  }
	
	((struct mymsgbuf *) tcp->u_arg[1])=buf;
      }

    tcp->arg_state[1] = TAKEN;

    /* prendo anche il secondo argomento */
    (long*) tcp->u_arg[2] = (long*) malloc(sizeof(long)); 
    if (tcp->u_arg[2] == 0) 
      { 
	fprintf(stderr, "sys_msgsnd: no memory\n"); 
	return -1; 
      } 
    *((long*) tcp->u_arg[2])=sz; 
    tcp->arg_state[2] = TAKEN; 

    /* prendo anche il terzo argomento */
    (long*) tcp->u_arg[3] = (long*) malloc(sizeof(long)); 
    if (tcp->u_arg[3] == 0) 
      { 
	fprintf(stderr, "sys_msgsnd: no memory\n"); 
	return -1; 
      } 
    *((long*) tcp->u_arg[3])=flg;
    tcp->arg_state[3] = TAKEN;

    break;
  }
  return 0;
}


struct ipc_wrapper {
  struct mymsgbuf *msgp;
  long msgtyp;
};


void 
free_msgrcv(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i;

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 2:
      case 3:
      case 4:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 1:
	if (arg_state[1] == TAKEN)
	  if (u_arg[1])
	    {
	      struct mymsgbuf *buf = (struct mymsgbuf *) u_arg[1];
	      if (buf->mtext)
		free (buf->mtext);
	      free(buf);
	    }
	arg_state[1] = EMPTY;
	break;
      }
    }
  return;
}  


int 
print_msgrcv(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  struct mymsgbuf *buf;
  
  tprintf("msgrcv(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0: /* int msgid */
	  case 2: /* int msgsz */
	    tprintf("%lu", *((unsigned long*)info->u_arg[i]));
	    break;
	  case 1: /* struct msgbuf * */
	    buf=((struct mymsgbuf *) info->u_arg[1]);
	    tprintf("{%lu, ", buf->mtype);
	    realprintstr(buf->mtext, *((unsigned int*)info->u_arg[2]));
	    tprintf("}");
	    break;
	  case 3: /* long msgtyp */
	    tprintf("%ld", *((long*)info->u_arg[3]));
	    break;
	  case 4: /* int msgflg */
	    if (printflags(msg_flags, *((int*)info->u_arg[4])) == 0)
	      tprintf("0");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 5
  -) identificatore della coda di messaggi (int)
  -) struttura dati che conterra' il messaggio accodato (struct msgbuf *)
  -) dimensione dell'array di dati interno alla struttura (int)
  -) tipo di messaggio cercato (long)
  -) flag per specificare il compostamento della syscall (int)

  secondo la pagina del man e' cosi', in realta' nel codice originale dell'strace gli
  argomenti figurano in ordine diverso:

  secondo il           man                    strace

  arg[0]               int msgid              int msgid
  arg[1]               struct msgbuf *msgp    int msgsz
  arg[2]               int msgsz              int msgflg
  arg[3]               long msgtyp            struct msgbuf *msgp
  arg[4]               int msgflg             non viene preso 

  in questo caso il codice e' fatto in maniera tale da conformarsi alla
  pagina del man
*/
int 
sys_msgrcv(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct ipc_wrapper *tmp;
  long mtype;
  int i, sz, flg;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1: /* struct msgp */
  case 2: /* int msgsz */
  case 3: /* long msgtyp */
  case 4: /* int msgflg */

    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {	
	    fprintf(stderr, "sys_msgrcv: syscall failed\n");
	    return -1;
	  }
	
	for (i=1;i<=3;i++)
	  {
	    if (tcp->arg_state[i] == EMPTY)
	      {
		if (upeek(tcp->pid, (i) * 4, &tcp->u_arg[i]) < 0)
		  {
		    fprintf(stderr, "sys_msgrcv: upeek\n"); 
		    return -1;
		  }
		tcp->arg_state[i] = UPEEK;
	      }
	  }
	
	tmp=((struct ipc_wrapper *) malloc(sizeof(struct ipc_wrapper)));

	if (!tmp)
	  {
	    fprintf(stderr,"sys_msgrcv: malloc ipc_wrapper failed\n");
	    return -1;
	  }

	sz = tcp->u_arg[1]; 
	flg = tcp->u_arg[2];
    
	if (umove(tcp, tcp->u_arg[3], &tmp) < 0)
	  {
	    fprintf(stderr,"sys_msgrcv: umove arg 1 failed\n");
	    free(tmp);
	    return -1;
	  }
	else
	  {
	    umove(tcp, ((long) tmp->msgp), &mtype);		
	    tmp->msgp=((struct mymsgbuf *) malloc(sizeof(struct mymsgbuf)));
	    tmp->msgp->mtype=mtype;
	    tmp->msgp->mtext=get_str(tcp, (long) (tmp->msgp) + sizeof(long),sz);

	    if ( get_error )
	      {
		fprintf(stderr, "sys_msgsnd: get_str\n"); 
		tcp->arg_state[1] = EMPTY;
		return -1;
	      }

	    ((struct mymsgbuf *) tcp->u_arg[1])=tmp->msgp;
	    tcp->arg_state[1] = TAKEN;

	    (long*) tcp->u_arg[3] = (long*) malloc(sizeof(long)); 
	    if (tcp->u_arg[3] == 0) 
	      { 
		fprintf(stderr, "arg_alloc: no memory\n"); 
		return -1; 
	      } 
	    *((long*) tcp->u_arg[3])=tmp->msgtyp; 
	    tcp->arg_state[3] = TAKEN;

	    free(tmp);

	    (long*) tcp->u_arg[2] = (long*) malloc(sizeof(long)); 
	    if (tcp->u_arg[2] == 0) 
	      { 
		fprintf(stderr, "arg_alloc: no memory\n"); 
		return -1; 
	      } 
	    *((long*) tcp->u_arg[2])=sz; 
	    tcp->arg_state[2] = TAKEN; 

	    (long*) tcp->u_arg[4] = (long*) malloc(sizeof(long)); 
	    if (tcp->u_arg[4] == 0) 
	      { 
		fprintf(stderr, "arg_alloc: no memory\n"); 
		return -1; 
	      } 
	    *((long*) tcp->u_arg[4])=flg; 
	    tcp->arg_state[4] = TAKEN; 
	  }
      }
    break;
  }
  return 0;
}




int 
print_semop(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("semop(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[i]));
	    break;
	  case 1:
	    tprintf("struct semop");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) identificatore del semaforo (int)
  -) array struttura dati contenenti le operazioni da fare sul semafoto (struct sembuf *)
  -) numero di strutture nell'array (unsigned int)

  stesso discorso visto sopra, vale per tutte le ipc_subcall
*/

int 
sys_semop(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct sembuf *sb;
  long temp_l;
  int i;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
  case 2:
    //if (entering(tcp)) 
      {
	for (i=1;i<=3;i++)
	  {
	    if (tcp->arg_state[i] == EMPTY)
	      {
		if (upeek(tcp->pid, (i) * 4, &tcp->u_arg[i]) < 0)
		  {
		    fprintf(stderr, "arg_alloc: no upeek\n"); 
		    return -1;
		  }
		tcp->arg_state[i] = UPEEK;
	      }
	  }

	temp_l=tcp->u_arg[1];
	
	if (tcp->u_arg[3])
	  {
	    if (set_copy(tcp, 1, tcp->u_arg[3], (long *)&sb, sizeof(struct sembuf), "semop"))
	      {
		fprintf(stderr, "arg_alloc: set_copy\n"); 
		return -1;
	      }
	  }
	tcp->arg_state[1] = TAKEN;

	(long*) tcp->u_arg[2] = (long*) malloc(sizeof(long)); 
	if (tcp->u_arg[2] == 0) 
	  { 
	    fprintf(stderr, "arg_alloc: no memory\n"); 
	    return -1; 
	  } 
	*((long*) tcp->u_arg[2])=temp_l; 
	tcp->arg_state[2] = TAKEN; 
      }
    break;
  }
  return 0;
}



int 
print_semget(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("semget(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    if (*((unsigned long*)info->u_arg[0]) != 0)
	      tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    else
	      tprintf("IPC_PRIVATE");
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  case 2:
	    if (printflags(resource_flags, *((int*)info->u_arg[2])) != 0)
	      tprintf("|");
	    tprintf("%#lo", *((unsigned long*)info->u_arg[2]) & 0666);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) chiave identificativa dell'insieme di semafori (key_t == int)
  -) numero di semafori da creare nell'insieme (int)
  -) flags per la creazione (int)
*/
int 
sys_semget(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


#if defined(__GNU_LIBRARY__) && !defined(_SEM_SEMUN_UNDEFINED)
/* union semun is defined by including <sys/sem.h> */
#else
/* according to X/OPEN we have to define it ourselves */
union semun {
  int val;                    /* value for SETVAL */
  struct semid_ds *buf;       /* buffer for IPC_STAT, IPC_SET */
  unsigned short int *array;  /* array for GETALL, SETALL */
  struct seminfo *__buf;      /* buffer for IPC_INFO */
};
#endif

union semun *
get_union_semun(tcp, addr, cmd) 
struct tcb *tcp;
long addr;
int cmd;
{
  union semun *ue;
  
  get_error = 0;
  
  if ( (void *) addr == NULL)
    return NULL;

  if ((ue=malloc(sizeof(union semun))) == NULL) {
    fprintf(stderr,"get_union_enum: no memory\n");
    get_error = 1;
    return NULL;
  }
  switch (cmd) {
  case IPC_STAT:
  case IPC_SET:
    if ((ue->buf=malloc(sizeof(struct semid_ds))) == NULL) {
      fprintf(stderr,"get_union_enum: no memory\n");
      get_error = 1;
      return NULL;
    }
    break;
  case SETVAL:
    break;
  case GETALL:
  case SETALL:
    if ((ue->array=malloc(sizeof(unsigned short int))) == NULL) {
      fprintf(stderr,"get_union_enum: no memory\n");
      get_error = 1;
      return NULL;
    }
    break;
  case IPC_INFO:
 if ((ue->__buf=malloc(sizeof(struct seminfo))) == NULL) {
      fprintf(stderr,"get_union_enum: no memory\n");
      get_error = 1;
      return NULL;
    }
    break;
  default:
    free(ue);
    ue=NULL;
    break;
  }

  if (ue)
    if (umove(tcp,addr,ue) == -1) {
      fprintf(stderr,"get_union_enum: umove failed\n");
      get_error = 1;
      free(ue);
      return NULL;
    }
  
  return ue;
}


void 
free_semctl(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i,cmd=0;

  if (arg_state[2] == TAKEN)
    cmd=*((int*)u_arg[2]);

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 1:
      case 2:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 3:
	if (arg_state[3] == TAKEN)
	  if (u_arg[3])
	    {
	      union semun * us= (union semun *) u_arg[3];
	      switch (cmd) {
	      case IPC_STAT:
	      case IPC_SET:
		free(us->buf);
		break;
	      case GETALL:
	      case SETALL:
		free(us->array);
		break;
	      case IPC_INFO:
		free(us->__buf);
		break;
	      }
	      free(us);
	    }
	arg_state[3] = EMPTY;
	break;
      }
    }
  return;
}  


int 
print_semctl(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("semctl(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[i]));
	    break;
	  case 2:
	    printxval(semctl_flags, *((int*)info->u_arg[2]), "SEM_???");
	    break;
	  case 3:
	    tprintf("union semun");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}




/*
  Numero argomenti: 4
  -) identificatore dell'insieme di semafori (int)
  -) identificatore del semaforo nell'insieme (int)
  -) comando da eseguire (int)
  -) struttura dati contenente vari possibili valori (union semun)
*/
int 
sys_semctl(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int cmd, *ptr;

  switch (n_arg) {
  case 0:
  case 1:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 3:
    //if (entering(tcp)) {
      if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	{
	  fprintf(stderr,"sys_semctl: take_arg(2, tcp)\n");
	  return -1;
	}
      cmd = *ptr;
 
      tcp->u_arg[n_arg] = (long) get_union_semun(tcp, (long) tcp->u_arg[n_arg], cmd);
      if ( get_error )
	{
	  fprintf(stderr,"sys_semctl: get_union_semun\n");
	  tcp->arg_state[n_arg] = EMPTY;
	  return -1;
	}
      tcp->arg_state[n_arg] = TAKEN;
      //}
    break;
  }
  return 0;
}



int 
print_shmget(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("shmget(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    if (*((unsigned long*)info->u_arg[0]) != 0)
	      tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    else
	      tprintf("IPC_PRIVATE");
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  case 2:
	    if (printflags(resource_flags, *((int*)info->u_arg[2])) != 0)
	      tprintf("|");
	    tprintf("%#lo", *((unsigned long*)info->u_arg[2]) & 0666);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) chiave identificativa del segmento di memoria condivisa (key_t == int)
  -) dimensione del probabile nuovo segmento di memoria (int)
  -) flags (int)
*/
int 
sys_shmget(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



#include <sys/shm.h>

struct local_shmid_ds 
{
  struct    ipc_perm shm_perm;  /* operation perms */
  int  shm_segsz;               /* size of segment (bytes) */
  time_t    shm_atime;          /* last attach time */
  time_t    shm_dtime;          /* last detach time */
  time_t    shm_ctime;          /* last change time */
  unsigned short shm_cpid;      /* pid of creator */
  unsigned short shm_lpid;      /* pid of last operator */
  short     shm_nattch;         /* no. of current attaches */
                                /* the following were private */
  unsigned short   shm_npages;  /* size of segment (pages) */
  unsigned long   *shm_pages;
  struct shm_desc *attaches;    /* descriptors for attaches */
};


struct shmid_ds *
get_shmid_ds(tcp, n, addr) 
struct tcb *tcp;
int n;
long addr;
{

  struct local_shmid_ds *sd;
  struct local_shmid_ds *my = (struct local_shmid_ds *)tcp->u_arg[n];;

  get_error = 0;
  
  if ( (void *) addr == NULL)
    return NULL;
    
  if ((sd=malloc(sizeof(struct shmid_ds))) == NULL) {
    fprintf(stderr,"get_shmid_ds: no memory\n");
    get_error = 1;
    return NULL;
  }
  //si alloca memoria per contenere "shm_npages" indirizzi di pagine (unsigned long)
  if ((sd->shm_pages=malloc(sizeof(my->shm_pages))) == NULL) {
    fprintf(stderr,"get_shmid_ds: no memory\n");
    get_error = 1;
    free(sd);
    return NULL;
  }
  //si alloca memoria per contenere "sd->shm_nattch" strutture di descrittori (struct shm_desc)
  if ((sd->attaches=malloc(sizeof(my->attaches))) == NULL) {
    fprintf(stderr,"get_shmid_ds: no memory\n");
    get_error = 1;
    free(sd->shm_pages);
    free(sd);
    return NULL;
  }
  //si copia tutta la memoria necessaria
  if (umove(tcp,addr,sd) == -1) {
    fprintf(stderr,"get_shmid_ds: umove failed\n");
    get_error = 1;
    free(sd->shm_pages);
    free(sd->attaches);
    free(sd);
    return NULL;
  }
  return (struct shmid_ds*) sd;
}




void 
free_shmctl(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i;

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 1:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 2:
	if (arg_state[2] == TAKEN)
	  if (u_arg[2])
	    {
	      struct local_shmid_ds *sd=(struct local_shmid_ds *) u_arg[2];
	      free(sd->shm_pages);
	      free(sd->attaches);
	      free(sd);
	    }
	arg_state[2] = EMPTY;
	break;
      }
    }
  return;
}  




int 
print_shmctl(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("shmctl(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    printxval(shmctl_flags, *((int*)info->u_arg[1]), "SHM_???");
	    break;
	  case 2:
	    tprintf("struct shmid_ds");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) identificatore del segmento di memoria condivisa (int)
  -) comando da eseguire (int)
  -) struttura dati con le info sul segmento di memoria (struct shmid_ds *)
*/
int 
sys_shmctl(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  case 2:

    if (upeek(tcp->pid, (3) * 4, &tcp->u_arg[3]) < 0)
      {
	fprintf(stderr,"sys_semctl: upeek\n");
	return -1;
      }
    tcp->arg_state[3] = UPEEK;

    tcp->u_arg[n_arg] = (long) get_shmid_ds(tcp,n_arg,tcp->u_arg[3]);
    if ( get_error )
      {
	fprintf(stderr,"sys_semctl: get_shmid_ds\n");
	tcp->arg_state[n_arg] = EMPTY;
	return -1;
      }
    tcp->arg_state[n_arg] = TAKEN;
    break;  
  }
  return 0;
}


int 
print_shmat(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("shmat(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    if ((void *)info->u_arg[1] != NULL)  tprintf("%#lx", *(unsigned long*)info->u_arg[1]);
	    else tprintf("NULL");
	    break;
	  case 2:
	    if (printflags(shm_flags, *((int*)info->u_arg[2])) == 0)
	      tprintf("0");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  if (info->arg_state[3] == UPEEK)
    {
      info->u_rval = info->u_arg[3];
      return RVAL_HEX;
    }

  return RVAL_NONE;
}



/*
  Numero argomenti: 3
  -) identificatore del segmento di memoria condivisa (int)
  -) indirizzo del segmento di memoria da attaccare (const void *)
  -) flags di attachment (int)
*/
int 
sys_shmat(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int i;
  unsigned long *ind;
  long temp_l;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
  case 2:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {	
	    fprintf(stderr, "sys_shmat: syscall failed\n");
	    return -1;
	  }

	for (i=1;i<=3;i++)
	  {
	    if (tcp->arg_state[i] == EMPTY)
	      {
		if (upeek(tcp->pid, (i) * 4, &tcp->u_arg[i]) < 0)
		  {		 
		    fprintf(stderr, "sys_shmat: upeek\n");
		    return -1;
		  }
		tcp->arg_state[i] = UPEEK;
	      }
	  }

	temp_l=tcp->u_arg[1];      
	
	if (set_copy(tcp, 1, tcp->u_arg[3], (long *)&ind, sizeof(unsigned int), "shmat"))
	  {
	    fprintf(stderr, "sys_shmat:set_copy\n");
	    return -1;
	  }

	if (umove(tcp, tcp->u_arg[2], &(tcp->u_arg[3])) < 0)
	  tcp->arg_state[3] = EMPTY;
	/* ^^ serve per stampare il valore di ritorno*/
	
	(long*) tcp->u_arg[2] = (long*) malloc(sizeof(long)); 
	if (tcp->u_arg[2] == 0) 
	  { 
	    fprintf(stderr, "sys_shmat: no memory\n"); 
	    return -1; 
	  } 
	*((long*) tcp->u_arg[2])=temp_l; 
	tcp->arg_state[2] = TAKEN; 
      }
    break;
  }
  return 0;
}





int 
print_shmdt(info)
msg_building_info *info;
{
  tprintf("shmdt(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%#lx", *((unsigned long*)info->u_arg[0]));
  else
    tprintf("_");
  
  tprintf(") ");
  
  return 0;
}



/*
  Numero argomenti: 1
  -) indirizzo del segmento di memoria da staccare (const void *)
*/
int 
sys_shmdt(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  unsigned long *ind;
  
  if (entering(tcp)) 
    {
      if (set_copy(tcp, 0, tcp->u_arg[3], (long *)&ind, sizeof(unsigned int), "shmdt"))
	{
	  fprintf(stderr, "sys_shmdt: set_copy\n");
	  return -1;
	}
    }
  return 0;
}
