/*
 * Copyright (c) 2000, 2001 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <rete_nera@yahoogroups.com>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995 Rick Sladkey <jrs@world.std.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: syscall.h,v 1.2 2003/02/17 16:12:14 frassi Exp $
 */

#include "dummy.h"

/**********************************
 * FUNCTIONS TO TAKE SYSCALL ARGS *
 **********************************/

/* primary syscalls */

int sys_setup(), sys_exit(), sys_fork(), sys_read(), sys_write();
int sys_open(), sys_close(), sys_waitpid(), sys_creat(), sys_link();
int sys_unlink(), sys_execve(), sys_chdir(), sys_time(), sys_mknod();
int sys_chmod(), sys_chown(), sys_break(), sys_oldstat();
int sys_lseek(), sys_getpid(), sys_mount(), sys_umount(), sys_umount2();
int sys_setuid(), sys_getuid(), sys_stime(), sys_ptrace();
int sys_alarm(), sys_oldfstat(), sys_pause(), sys_utime();
int sys_stty(), sys_gtty(), sys_access(), sys_nice(), sys_ftime();
int sys_sync(), sys_kill(), sys_rename(), sys_mkdir(), sys_rmdir();
int sys_dup(), sys_pipe(), sys_times(), sys_prof(), sys_brk();
int sys_setgid(), sys_getgid(), sys_signal(), sys_geteuid();
int sys_getegid(), sys_acct(), sys_phys(), sys_lock(), sys_ioctl();
int sys_fcntl(), sys_mpx(), sys_setpgid(), sys_ulimit();
int sys_olduname(), sys_umask(), sys_chroot(), sys_ustat();
int sys_dup2(), sys_getppid(), sys_getpgrp(), sys_setsid();
int sys_sigaction(), sys_siggetmask(), sys_sigsetmask();
int sys_setreuid(), sys_setregid(), sys_sigsuspend();
int sys_sigpending(), sys_sethostname(), sys_setrlimit();
int sys_getrlimit(), sys_getrusage(), sys_gettimeofday();
int sys_settimeofday(), sys_getgroups(), sys_setgroups();
int sys_oldselect(), sys_symlink(), sys_oldlstat(), sys_readlink();
int sys_uselib(), sys_swapon(), sys_reboot(), sys_readdir();
int sys_mmap(), sys_munmap(), sys_truncate(), sys_ftruncate();
int sys_fchmod(), sys_fchown(), sys_getpriority();
int sys_setpriority(), sys_profil(), sys_statfs(), sys_fstatfs();
int sys_ioperm(), sys_socketcall(), sys_syslog(), sys_setitimer();
int sys_getitimer(), sys_stat(), sys_lstat(), sys_fstat();
int sys_uname(), sys_iopl(), sys_vhangup(), sys_idle(), sys_vm86();
int sys_wait4(), sys_swapoff(), sys_ipc(), sys_sigreturn();
int sys_fsync(), sys_clone(), sys_setdomainname(), sys_sysinfo();
int sys_modify_ldt(), sys_adjtimex(), sys_mprotect();
int sys_sigprocmask(), sys_create_module(), sys_init_module();
int sys_delete_module(), sys_get_kernel_syms(), sys_quotactl();
int sys_getpgid(), sys_fchdir(), sys_bdflush();
int sys_sysfs(), sys_personality(), sys_afs_syscall();
int sys_setfsuid(), sys_setfsgid(), sys_llseek();
int sys_getdents(), sys_flock(), sys_msync();
int sys_readv(), sys_writev(), sys_select();
int sys_getsid(), sys_fdatasync(), sys_sysctl();
int sys_mlock(), sys_munlock(), sys_mlockall(), sys_munlockall(), sys_madvise();
int sys_sched_setparam(), sys_sched_getparam();
int sys_sched_setscheduler(), sys_sched_getscheduler(), sys_sched_yield();
int sys_sched_get_priority_max(), sys_sched_get_priority_min();
int sys_sched_rr_get_interval(), sys_nanosleep(), sys_mremap();
int sys_sendmsg(), sys_recvmsg(), sys_setresuid(), sys_setresgid();
int sys_getresuid(), sys_getresgid(), sys_pread(), sys_pwrite(), sys_getcwd();
int sys_sigaltstack(), sys_rt_sigprocmask(), sys_rt_sigaction();
int sys_rt_sigpending(), sys_rt_sigsuspend(), sys_rt_sigqueueinfo();
int sys_rt_sigtimedwait(), sys_prctl(), sys_poll(), sys_vfork();
int sys_sendfile(), sys_old_mmap(), sys_stat64(), sys_lstat64(), sys_fstat64();
int sys_truncate64(), sys_ftruncate64(), sys_pivotroot();
int sys_getdents64(), sys_clone2();


/* sys_socketcall subcalls */

int sys_socket(), sys_bind(), sys_connect(), sys_listen();
int sys_accept(), sys_getsockname(), sys_getpeername(), sys_socketpair();
int sys_send(), sys_recv(), sys_sendto(), sys_recvfrom();
int sys_shutdown(), sys_setsockopt(), sys_getsockopt();

/* new ones */
int sys_query_module();
//int sys_poll();


/* architecture-specific calls */

#define SYS_socket_subcall	300

#define SYS_socket		(SYS_socket_subcall + 1)
#define SYS_bind		(SYS_socket_subcall + 2)
#define SYS_connect		(SYS_socket_subcall + 3)
#define SYS_listen		(SYS_socket_subcall + 4)
#define SYS_accept		(SYS_socket_subcall + 5)
#define SYS_getsockname		(SYS_socket_subcall + 6)
#define SYS_getpeername		(SYS_socket_subcall + 7)
#define SYS_socketpair		(SYS_socket_subcall + 8)
#define SYS_send		(SYS_socket_subcall + 9)
#define SYS_recv		(SYS_socket_subcall + 10)
#define SYS_sendto		(SYS_socket_subcall + 11)
#define SYS_recvfrom		(SYS_socket_subcall + 12)
#define SYS_shutdown		(SYS_socket_subcall + 13)
#define SYS_setsockopt		(SYS_socket_subcall + 14)
#define SYS_getsockopt		(SYS_socket_subcall + 15)
#define SYS_sendmsg		(SYS_socket_subcall + 16)
#define SYS_recvmsg		(SYS_socket_subcall + 17)

#define SYS_socket_nsubcalls	18



/* sys_ipc subcalls */

int sys_semget(), sys_semctl(), sys_semop();
int sys_msgsnd(), sys_msgrcv(), sys_msgget(), sys_msgctl();
int sys_shmat(), sys_shmdt(), sys_shmget(), sys_shmctl();


#define SYS_ipc_subcall		((SYS_socket_subcall)+(SYS_socket_nsubcalls))
#define SYS_semop		(SYS_ipc_subcall + 1)
#define SYS_semget		(SYS_ipc_subcall + 2)
#define SYS_semctl		(SYS_ipc_subcall + 3)
#define SYS_msgsnd		(SYS_ipc_subcall + 11)
#define SYS_msgrcv		(SYS_ipc_subcall + 12)
#define SYS_msgget		(SYS_ipc_subcall + 13)
#define SYS_msgctl		(SYS_ipc_subcall + 14)
#define SYS_shmat		(SYS_ipc_subcall + 21)
#define SYS_shmdt		(SYS_ipc_subcall + 22)
#define SYS_shmget		(SYS_ipc_subcall + 23)
#define SYS_shmctl		(SYS_ipc_subcall + 24)

#define SYS_ipc_nsubcalls	25


int /*sys_setpgrp(),*/ sys_gethostname(), sys_getdtablesize(), sys_utimes();
int sys_capget(), sys_capset();



/***********************************
 * FUNCTIONS TO PRINT SYSCALL ARGS *
 ***********************************/


/* primary syscalls */

int print_setup(), print_exit(), print_fork(), print_read(), print_write();
int print_open(), print_close(), print_waitpid(), print_creat(), print_link();
int print_unlink(), print_execve(), print_chdir(), print_time(), print_mknod();
int print_chmod(), print_chown(), print_break(), print_oldstat();
int print_lseek(), print_getpid(), print_mount(), print_umount(), print_umount2();
int print_setuid(), print_getuid(), print_stime(), print_ptrace();
int print_alarm(), print_oldfstat(), print_pause(), print_utime();
int print_stty(), print_gtty(), print_access(), print_nice(), print_ftime();
int print_sync(), print_kill(), print_rename(), print_mkdir(), print_rmdir();
int print_dup(), print_pipe(), print_times(), print_prof(), print_brk();
int print_setgid(), print_getgid(), print_signal(), print_geteuid();
int print_getegid(), print_acct(), print_phys(), print_lock(), print_ioctl();
int print_fcntl(), print_mpx(), print_setpgid(), print_ulimit();
int print_olduname(), print_umask(), print_chroot(), print_ustat();
int print_dup2(), print_getppid(), print_getpgrp(), print_setsid();
int print_sigaction(), print_siggetmask(), print_sigsetmask();
int print_setreuid(), print_setregid(), print_sigsuspend();
int print_sigpending(), print_sethostname(), print_setrlimit();
int print_getrlimit(), print_getrusage(), print_gettimeofday();
int print_settimeofday(), print_getgroups(), print_setgroups();
int print_oldselect(), print_symlink(), print_oldlstat(), print_readlink();
int print_uselib(), print_swapon(), print_reboot(), print_readdir();
int print_mmap(), print_munmap(), print_truncate(), print_ftruncate();
int print_fchmod(), print_fchown(), print_getpriority();
int print_setpriority(), print_profil(), print_statfs(), print_fstatfs();
int print_ioperm(), print_socketcall(), print_syslog(), print_setitimer();
int print_getitimer(), print_stat(), print_lstat(), print_fstat();
int print_uname(), print_iopl(), print_vhangup(), print_idle(), print_vm86();
int print_wait4(), print_swapoff(), print_ipc(), print_sigreturn();
int print_fsync(), print_clone(), print_setdomainname(), print_sysinfo();
int print_modify_ldt(), print_adjtimex(), print_mprotect();
int print_sigprocmask(), print_create_module(), print_init_module();
int print_delete_module(), print_get_kernel_syms(), print_quotactl();
int print_getpgid(), print_fchdir(), print_bdflush();
int print_sysfs(), print_personality(), print_afs_syscall();
int print_setfsuid(), print_setfsgid(), print_llseek();
int print_getdents(), print_flock(), print_msync();
int print_readv(), print_writev(), print_select();
int print_getsid(), print_fdatasync(), print_sysctl();
int print_mlock(), print_munlock(), print_mlockall(), print_munlockall();
int print_sched_setparam(), print_sched_getparam();
int print_sched_setscheduler(), print_sched_getscheduler(), print_sched_yield();
int print_sched_get_priority_max(), print_sched_get_priority_min();
int print_sched_rr_get_interval(), print_nanosleep(), print_mremap();
int print_sendmsg(), print_recvmsg(), print_setresuid(), print_setresgid();
int print_getresuid(), print_getresgid(), print_pread(), print_pwrite(), print_getcwd();
int print_sigaltstack(), print_rt_sigprocmask(), print_rt_sigaction();
int print_rt_sigpending(), print_rt_sigsuspend(), print_rt_sigqueueinfo();
int print_rt_sigtimedwait(), print_vfork();
int print_sendfile(), print_old_mmap(), print_stat64(), print_lstat64(), print_fstat64();
int print_truncate64(), print_ftruncate64();


/* print_socketcall subcalls */

int print_socket(), print_bind(), print_connect(), print_listen();
int print_accept(), print_getsockname(), print_getpeername(), print_socketpair();
int print_send(), print_recv(), print_sendto(), print_recvfrom();
int print_shutdown(), print_setsockopt(), print_getsockopt();

/* new ones */
int print_query_module();
int print_poll();


/* sys_ipc subcalls */

int print_semget(), print_semctl(), print_semop();
int print_msgsnd(), print_msgrcv(), print_msgget(), print_msgctl();
int print_shmat(), print_shmdt(), print_shmget(), print_shmctl();




int print_capget(), print_capset();


/******************************************
 * functions to free syscall arguments    *
 ******************************************/

void  free_syscall(); /* generic function */

void  free_execve();
void  free_readv(), free_writev();
void  free_sendmsg(), free_recvmsg();
void  free_msgsnd(), free_msgrcv();
void  free_semctl(), free_shmctl();
void  free_sysctl();

/*********************
 * INTERNAL FUNCTION *
 *********************/

int internal_none();  /* trivial function */

int internal_exit(), internal_clone(), internal_fork();
int internal_exec(), internal_wait(), internal_shmat();
int internal_sigreturn(), internal_sigaction(), internal_signal();
int internal_sigsetmask(), internal_siggetmask();
int internal_oldselect();
