/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: mem.c,v 1.2 2003/02/17 15:30:46 frassi Exp $
 */

#include "util.h"
#include "runtime_util.h"
#include "pars_funct.h" /* get_nargs */

#include <linux/mman.h>
#include <sys/mman.h>
#include <asm/ldt.h>




int 
print_brk(info)
msg_building_info *info;
{
  tprintf("brk(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%#lx", *((unsigned long*)info->u_arg[0]));
  else
    tprintf("_");
  
  tprintf(") ");
  
  return RVAL_HEX;
}



int
sys_brk(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}

/* flag di protezione utilizzati da sys_old_mmap, sys_mmap e sys_mprotect */
static struct xlat mmap_prot[] = {  
  { PROT_NONE,	"PROT_NONE",	},
  { PROT_READ,	"PROT_READ"	},
  { PROT_WRITE,	"PROT_WRITE"	},
  { PROT_EXEC,	"PROT_EXEC"	},
  { 0,		NULL		},
};

/* flag utilizzati da sys_old_mmap e sys_mmap */
static struct xlat mmap_flags[] = {
  { MAP_SHARED,	"MAP_SHARED"	},
  { MAP_PRIVATE,"MAP_PRIVATE"	},
  { MAP_FIXED,	"MAP_FIXED"	},
#ifdef MAP_ANONYMOUS
  { MAP_ANONYMOUS,"MAP_ANONYMOUS"	},
#endif
#ifdef MAP_RENAME
  { MAP_RENAME,	"MAP_RENAME"	},
#endif
#ifdef MAP_NORESERVE
  { MAP_NORESERVE,"MAP_NORESERVE"	},
#endif
#ifdef _MAP_NEW
  { _MAP_NEW,	"_MAP_NEW"	},
#endif
#ifdef MAP_GROWSDOWN
  { MAP_GROWSDOWN,"MAP_GROWSDOWN"	},
#endif
#ifdef MAP_DENYWRITE
  { MAP_DENYWRITE,"MAP_DENYWRITE"	},
#endif
#ifdef MAP_EXECUTABLE
  { MAP_EXECUTABLE,"MAP_EXECUTABLE"},
#endif
#ifdef MAP_INHERIT
  { MAP_INHERIT,"MAP_INHERIT"	},
#endif
#ifdef MAP_FILE
  { MAP_FILE,"MAP_FILE"},
#endif
#ifdef MAP_LOCKED
  { MAP_LOCKED,"MAP_LOCKED"},
#endif
	/* FreeBSD ones */
#ifdef MAP_ANON
	{ MAP_ANON,		"MAP_ANON"	},
#endif
#ifdef MAP_HASSEMAPHORE
	{ MAP_HASSEMAPHORE,	"MAP_HASSEMAPHORE"	},
#endif
#ifdef MAP_STACK
	{ MAP_STACK,		"MAP_STACK"	},
#endif
#ifdef MAP_NOSYNC
	{ MAP_NOSYNC,		"MAP_NOSYNC"	},
#endif
#ifdef MAP_NOCORE
	{ MAP_NOCORE,		"MAP_NOCORE"	},
#endif
  { 0,		NULL		},
};


static void
realprintmmap(info)   
msg_building_info *info;
{
  int i;

  for (i=0;i<6;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    /* addr */
	    if (info->u_arg[0])
	      { 
		if (!*((long*)info->u_arg[0]))
		  tprintf("NULL");
		else
		  tprintf("%#lx", *((unsigned long*)info->u_arg[0]));
	      }
	    else
	      tprintf("NULL");
	    break;
	  case 1:
	    /* len */
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  case 2:
	    /* prot */
	    printflags(mmap_prot, *((int*)info->u_arg[2]));
	    break;
	  case 3:
	    /* flags */
	    printxval(mmap_flags, *((int*)info->u_arg[3]) & MAP_TYPE, "MAP_???");
	    addflags(mmap_flags, *((int*)info->u_arg[3]) & ~MAP_TYPE);
	    break;
	  case 4:
	    /* fd */
	    tprintf("%ld", *((long*)info->u_arg[4]));
	    break;
	  case 5:
	    /* offset */
	    tprintf("%#lx", *((unsigned long*)info->u_arg[5]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=5)
	tprintf(", ");    
    }
  return;
}



int 
print_old_mmap(info)
msg_building_info *info;
{
  tprintf("old_mmap(");
  realprintmmap(info);
  tprintf(") ");
  return RVAL_HEX;
}



int
sys_old_mmap(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  long *u_arg;
  int j; 

  if (tcp->arg_state[0]==TAKEN)
    return 0;

  if ((u_arg = ((long*) calloc(6,sizeof(long))))==NULL)
    {
      fprintf(stderr,"sys_old_mmap: calloc failed\n");
      return -1;
    }
  
  if (upeek(tcp->pid, 0, &tcp->u_arg[0]) < 0)
    {  
      fprintf(stderr,"sys_old_mmap: upeek failed\n");
      return -1;
    }
  tcp->arg_state[0] = UPEEK;

  if (umoven(tcp, tcp->u_arg[0], 6 * sizeof(long), (char *) u_arg) == -1) 
    {
      fprintf(stderr,"sys_old_mmap: umove args failed\n");
      free(u_arg);
      return -1;
    }
 
  /* visto che la umove ha spostato tutto, 
     tanto vale fare tutti gli assegnamenti */

  for(j=0;j<6;j++)
    {
      tcp->u_arg[j] = u_arg[j];
      tcp->arg_state[j] = UPEEK;
      arg_alloc(tcp,j);
    }

  free(u_arg);

  return 0;
}



int 
print_mmap(info)
msg_building_info *info;
{
  tprintf("mmap(");
  realprintmmap(info);
  tprintf(") ");
  return RVAL_HEX;
}




int
sys_mmap(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  switch(n_arg) {
  case 0:  /* addr */
    if (tcp->u_arg[0])
      arg_alloc(tcp,n_arg);
    
    tcp->arg_state[n_arg] = TAKEN;
    break;
  case 1:   /* len */
  case 2:   /* prot */
  case 3:   /* flags */ 
  case 4:   /* fd */
  case 5:   /* offset */
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}



int 
print_munmap(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("munmap(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%#lx", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}



int 
sys_munmap(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}




int 
print_mprotect(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("mprotect(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%#lx", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  case 2:
	    if (!printflags(mmap_prot, *((int*)info->u_arg[2])))
	      tprintf("PROT_???");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_mprotect(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



static struct xlat mremap_flags[] = {   // vedi funz. sotto
  { MREMAP_MAYMOVE, "MREMAP_MAYMOVE" },
};



int 
print_mremap(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("mremap(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%#lx", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[i]));
	    break;
	  case 3:
	    printflags(mremap_flags, *((int*)info->u_arg[3]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return RVAL_HEX;
}



/*
  Numero argomenti: 4
  -) vecchio indirizzo del blocco di memoria virtuale da espandere (void *)
  -) dimensione del vecchio blocco (size_t == unsigned int)
  -) dimensione richiesta per il nuovo blocco (size_t == unsigned int)
  -) flags (unsigned long)
*/
int 
sys_mremap(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  long *mem_ind;

  switch (n_arg) {
  case 0:
    //if (entering(tcp))
    if ( set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*) &mem_ind, sizeof(long), "mremap") )
      {
	fprintf(stderr,"sys_mremap: set_copy\n");
	return -1;
      }
    break;
  case 1:
  case 2:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  }
  return 0;
}



static struct xlat madvise_flags[] = {
#ifdef MADV_NORMAL
	{ MADV_NORMAL,		"MADV_NORMAL" },
#endif
#ifdef MADZV_RANDOM
	{ MADV_RANDOM,		"MADV_RANDOM" },
#endif
#ifdef MADV_SEQUENTIAL
	{ MADV_SEQUENTIAL,	"MADV_SEQUENTIAL" },
#endif
#ifdef MADV_WILLNEED
	{ MADV_WILLNEED,	"MADV_WILLNEED" },
#endif
#ifdef MADV_DONTNED
	{ MADV_DONTNEED,	"MADV_DONTNEED" },
#endif
	{ 0,			NULL },
};


int
sys_madvise(tcp, n_arg) /* CONTROLLARE e fare la print */
struct tcb *tcp;
int n_arg;
{
  /*
	if (entering(tcp)) {
		tprintf("%#lx, %lu, ", tcp->u_arg[0], tcp->u_arg[1]);
		printflags(madvise_flags, tcp->u_arg[2]);
	}
	return 0;
  */
  long *mem_ind;

  switch (n_arg) {
  case 0:
    //if (entering(tcp))
    if ( set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*) &mem_ind, sizeof(long), "mremap") )
      {
	fprintf(stderr,"sys_madvise: set_copy\n");
	return -1;
      }
    break;
  case 1:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  }
  return 0;
}



static struct xlat mlockall_flags[] = {
#ifdef MCL_CURRENT
	{ MCL_CURRENT,	"MCL_CURRENT" },
#endif
#ifdef MCL_FUTURE
	{ MCL_FUTURE,	"MCL_FUTURE" },
#endif
	{ 0,		NULL}
};

int
sys_mlockall(tcp,n_arg) /* CONTROLLARE e fare la print */
struct tcb *tcp;
int n_arg;
{
  /*
    if (entering(tcp)) {
		printflags(mlockall_flags, tcp->u_arg[0]);
		}
	return 0;
  */
  arg_alloc(tcp,n_arg);
  return 0;
}



static struct xlat mctl_sync[] = {  // vedi funz. sotto
	{ MS_ASYNC,	"MS_ASYNC"	},
	{ MS_INVALIDATE,"MS_INVALIDATE"	},
	{ MS_SYNC,	"MS_SYNC"	},
	{ 0,		NULL		},
};


int 
print_msync(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("msync(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%#lx", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  case 2:
	    if (!printflags(mctl_sync, *((int*)info->u_arg[2])))
	      tprintf("MS_???");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}




/*
  Numero argomenti: 3
  -) indirizzo di partenza dell'area di memoria contenente il file da sincronizzare (const void *)
  -) dimensione dell'area di memoria (size_t == unsigned int)
  -) flags (int)
*/
int 
sys_msync(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  long *mem_ind;
  
  switch (n_arg) {
  case 0:
    //if (entering(tcp))
      if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&mem_ind, sizeof(long), "msync"))
	{
	  fprintf(stderr,"sys_msync: set_copy\n");
	  return -1;
	}
    break;
  case 1:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}




int 
print_modify_ldt(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  struct modify_ldt_ldt_s *copy;
  
  tprintf("modify_ldt(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    if (((long *)info->u_arg[1])==NULL)
	      {
		tprintf("NULL");
		break;
	      }

	    if ((*((long*)info->u_arg[1]) == 0) || *((unsigned long*)info->u_arg[2]) != sizeof (struct modify_ldt_ldt_s))
	      tprintf("%lx",*((unsigned long*)info->u_arg[1]));
	    else {
	      copy=(struct modify_ldt_ldt_s *) info->u_arg[1];
	      tprintf(", {entry_number:%d, ", copy->entry_number);
	      tprintf("base_addr:%#08lx, "
		      "limit:%d, "
		      "seg_32bit:%d, "
		      "contents:%d, "
		      "read_exec_only:%d, "
		      "limit_in_pages:%d, "
		      "seg_not_present:%d, "
		      "useable:%d}",
		      copy->base_addr,
		      copy->limit,
		      copy->seg_32bit,
		      copy->contents,
		      copy->read_exec_only,
		      copy->limit_in_pages,
		      copy->seg_not_present,
		      copy->useable);  
	    }
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}




/*
  Numero argomenti: 3
  -) booleano che determina l'azione da compiere (int)
  -) puntatore ad un'area di memoria o ad una struttura (void *)
  -) dimensione della struttura o dell'area di memoria (unsigned long)
*/
int 
sys_modify_ldt(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct modify_ldt_ldt_s *copy;

  switch (n_arg) 
    {
    case 0:
    case 2:
      arg_alloc(tcp,n_arg);
      break;
    case 1:
      //if (entering(tcp)) 
	{
	  int *ptr,len;
	  
	  if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	    {
	      fprintf(stderr,"sys_modify_ldt: take_arg(2, tcp)\n");
	      return -1;
	    }
	  len = *ptr;

	  if (tcp->u_arg[1] == 0 || len != sizeof (struct modify_ldt_ldt_s))
	    break;
	  else
	    {
	      if ( set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *) &copy, sizeof(struct modify_ldt_ldt_s), "modify_ldt") )
		{
		  fprintf(stderr,"sys_modify_ldt: set_copy\n");
		  return -1;
		}
	    }
	}
      break;
    } 

  return 0;
}
