/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: net.c,v 1.2 2003/02/17 15:20:55 frassi Exp $
 */

#include "gets.h" /* getiovec_struct */
#include "util.h"
#include "runtime_util.h"
#include "pars_funct.h" /* get_nargs */


/*
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <asm/types.h>
// sul nostro sistema __GLIBC__ = 2 e __GLIBC_MINOR__ = 1
#if defined(__GLIBC__) && (__GLIBC__ >= 2) && (__GLIBC__ + __GLIBC_MINOR__ >= 3)
#  include <netipx/ipx.h>
#else
#  include <linux/ipx.h>
#endif

#if defined (__GLIBC__) && ((__GLIBC__ < 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 1))
#if defined(HAVE_LINUX_IN6_H)
#include <linux/in6.h>
#endif
#endif

#if defined(HAVE_SYS_UIO_H)
#include <sys/uio.h>
#endif

#if defined(HAVE_LINUX_NETLINK_H)
#include <linux/netlink.h>
#endif

#if defined(HAVE_LINUX_IF_PACKET_H)
#include <linux/if_packet.h>
#endif


#ifndef PF_UNSPEC
#define PF_UNSPEC AF_UNSPEC
#endif
*/



#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/un.h>


#include <netinet/in.h>
#ifdef HAVE_NETINET_TCP_H
#include <netinet/tcp.h>
#endif
#ifdef HAVE_NETINET_UDP_H
#include <netinet/udp.h>
#endif
#include <arpa/inet.h>
#include <net/if.h>
#include <asm/types.h>
#if defined(__GLIBC__) && (__GLIBC__ >= 2) && (__GLIBC__ + __GLIBC_MINOR__ >= 3)
#  include <netipx/ipx.h>
#else
#  include <linux/ipx.h>
#endif

#if defined (__GLIBC__) && (((__GLIBC__ < 2) || (__GLIBC__ == 2 && __GLIBC_MINOR__ < 1)) || defined(HAVE_SIN6_SCOPE_ID_LINUX))
#if defined(HAVE_LINUX_IN6_H)
#if defined(HAVE_SIN6_SCOPE_ID_LINUX)
#undef in6_addr
#undef ipv6_mreq
#undef sockaddr_in6
#define in6_addr in6_addr_kernel
#define ipv6_mreq ipv6_mreq_kernel
#define sockaddr_in6 sockaddr_in6_kernel
#endif
#include <linux/in6.h>
#if defined(HAVE_SIN6_SCOPE_ID_LINUX)
#undef in6_addr
#undef ipv6_mreq
#undef sockaddr_in6
#define in6_addr in6_addr_libc
#define ipv6_mreq ipv6_mreq_libc
#define sockaddr_in6 sockaddr_in6_kernel
#endif
#endif
#endif

#if defined(HAVE_SYS_UIO_H)
#include <sys/uio.h>
#endif

#if defined(HAVE_LINUX_NETLINK_H)
#include <linux/netlink.h>
#endif

#if defined(HAVE_LINUX_IF_PACKET_H)
#include <linux/if_packet.h>
#endif

#if defined(HAVE_LINUX_ICMP_H)
#include <linux/icmp.h>
#endif

#ifndef PF_UNSPEC
#define PF_UNSPEC AF_UNSPEC
#endif

#if UNIXWARE >= 7
#define HAVE_SENDMSG		1		/* HACK - *FIXME* */
#endif



/* Under Linux these are enums so we can't test for them with ifdef. */
#define IPPROTO_EGP  IPPROTO_EGP
#define IPPROTO_PUP  IPPROTO_PUP
#define IPPROTO_IDP  IPPROTO_IDP
#define IPPROTO_IGMP IPPROTO_IGMP
#define IPPROTO_RAW  IPPROTO_RAW
#define IPPROTO_MAX  IPPROTO_MAX


static struct xlat domains[] = {  /* dominio di comunicazione */
	{ PF_UNSPEC,	"PF_UNSPEC"	},
	{ PF_UNIX,	"PF_UNIX"	},
	{ PF_INET,	"PF_INET"	},
#ifdef PF_NETLINK
	{ PF_NETLINK,	"PF_NETLINK"	},
#endif
#ifdef PF_PACKET
	{ PF_PACKET,	"PF_PACKET"	},
#endif
#ifdef PF_INET6
	{ PF_INET6,	"PF_INET6"	},
#endif
#ifdef PF_ATMSVC
	{ PF_ATMSVC,	"PF_INET6"	},
#endif
#ifdef PF_INET6
	{ PF_INET6,	"PF_INET6"	},
#endif
#ifdef PF_LOCAL
	{ PF_LOCAL,	"PS_LOCAL"	},
#endif
#ifdef PF_ISO
	{ PF_ISO,	"PF_ISO"	},
#endif
#ifdef PF_AX25
	{ PF_AX25,	"PF_AX25"	},
#endif
#ifdef PF_IPX
	{ PF_IPX,	"PF_IPX"	},
#endif
#ifdef PF_APPLETALK
	{ PF_APPLETALK,	"PF_APPLETALK"	},
#endif
#ifdef PF_NETROM
	{ PF_NETROM,	"PF_NETROM"	},
#endif
#ifdef PF_BRIDGE
	{ PF_BRIDGE,	"PF_BRIDGE"	},
#endif
#ifdef PF_AAL5
	{ PF_AAL5,	"PF_AAL5"	},
#endif
#ifdef PF_X25
	{ PF_X25,	"PF_X25"	},
#endif
#ifdef PF_ROSE
	{ PF_ROSE,	"PF_ROSE"	},
#endif
#ifdef PF_DECNET
	{ PF_DECNET,	"PF_DECNET"	},
#endif
#ifdef PF_NETBEUI
	{ PF_NETBEUI,	"PF_NETBEUI"	},
#endif
#ifdef PF_IMPLINK
	{ PF_IMPLINK,	"PF_IMPLINK"	},
#endif
	{ 0,		NULL		},
};


static struct xlat addrfams[] = {  /* famiglia di indirizzi */
	{ AF_UNSPEC,	"AF_UNSPEC"	},
	{ AF_UNIX,	"AF_UNIX"	},
	{ AF_INET,	"AF_INET"	},
#ifdef AF_INET6
	{ AF_INET6,	"AF_INET6"	},
#endif
	{ AF_DECnet,	"AF_DECnet"	},
#ifdef PF_ATMSVC
	{ AF_ATMSVC,	"AF_ATMSVC"	},
#endif
#ifdef AF_PACKET
	{ AF_PACKET,	"AF_PACKET"	},
#endif
#ifdef AF_NETLINK
	{ AF_NETLINK,	"AF_NETLINK"	},
#endif
#ifdef AF_ISO
	{ AF_ISO,	"AF_ISO"	},
#endif
#ifdef AF_IMPLINK
	{ AF_IMPLINK,	"AF_IMPLINK"	},
#endif
	{ 0,		NULL		},
};


static struct xlat socktypes[] = {   /* tipo di socket */
	{ SOCK_STREAM,	"SOCK_STREAM"	},
	{ SOCK_DGRAM,	"SOCK_DGRAM"	},
#ifdef SOCK_RAW
	{ SOCK_RAW,	"SOCK_RAW"	},
#endif
#ifdef SOCK_SEQPACKET
	{ SOCK_SEQPACKET,"SOCK_SEQPACKET"},
#endif
#ifdef SOCK_RDM
	{ SOCK_RDM,	"SOCK_RDM"	},
#endif
#ifdef SOCK_PACKET
	{ SOCK_PACKET,	"SOCK_PACKET"	},
#endif
	{ 0,		NULL		},
};

/*******************/
/* nuova struttura */
/*******************/
static struct xlat socketlayers[] = {
#if defined(SOL_IP)
	{ SOL_IP,	"SOL_IP"	},
#endif
#if defined(SOL_ICMP)
	{ SOL_ICMP,	"SOL_ICMP"	},
#endif
#if defined(SOL_TCP)
	{ SOL_TCP,	"SOL_TCP"	},
#endif
#if defined(SOL_UDP)
	{ SOL_UDP,	"SOL_UDP"	},
#endif
#if defined(SOL_IPV6)
	{ SOL_IPV6,	"SOL_IPV6"	},
#endif
#if defined(SOL_ICMPV6)
	{ SOL_ICMPV6,	"SOL_ICMPV6"	},
#endif
#if defined(SOL_RAW)
	{ SOL_RAW,	"SOL_RAW"	},
#endif
#if defined(SOL_IPX)
	{ SOL_IPX,	"SOL_IPX"	},
#endif
#if defined(SOL_IPX)
	{ SOL_IPX,	"SOL_IPX"	},
#endif
#if defined(SOL_AX25)
	{ SOL_AX25,	"SOL_AX25"	},
#endif
#if defined(SOL_ATALK)
	{ SOL_ATALK,	"SOL_ATALK"	},
#endif
#if defined(SOL_NETROM)
	{ SOL_NETROM,	"SOL_NETROM"	},
#endif
#if defined(SOL_ROSE)
	{ SOL_ROSE,	"SOL_ROSE"	},
#endif
#if defined(SOL_DECNET)
	{ SOL_DECNET,	"SOL_DECNET"	},
#endif
#if defined(SOL_X25)
	{ SOL_X25,	"SOL_X25"	},
#endif
#if defined(SOL_PACKET)
	{ SOL_PACKET,	"SOL_PACKET"	},
#endif
#if defined(SOL_ATM)
	{ SOL_ATM,	"SOL_ATM"	},
#endif
#if defined(SOL_AAL)
	{ SOL_AAL,	"SOL_AAL"	},
#endif
#if defined(SOL_IRDA)
	{ SOL_IRDA,	"SOL_IRDA"	},
#endif
	{ SOL_SOCKET,	"SOL_SOCKET"	},	/* Never used! */
};


/*** WARNING: DANGER WILL ROBINSON: NOTE "socketlayers" array above
     falls into "protocols" array below!!!!   This is intended!!! ***/
static struct xlat protocols[] = {  /* protocolli */
	{ IPPROTO_IP,	"IPPROTO_IP"	},
	{ IPPROTO_ICMP,	"IPPROTO_ICMP"	},
	{ IPPROTO_TCP,	"IPPROTO_TCP"	},
	{ IPPROTO_UDP,	"IPPROTO_UDP"	},
#ifdef IPPROTO_GGP
	{ IPPROTO_GGP,	"IPPROTO_GGP"	},
#endif
#ifdef IPPROTO_EGP
	{ IPPROTO_EGP,	"IPPROTO_EGP"	},
#endif
#ifdef IPPROTO_PUP
	{ IPPROTO_PUP,	"IPPROTO_PUP"	},
#endif
#ifdef IPPROTO_IDP
	{ IPPROTO_IDP,	"IPPROTO_IDP"	},
#endif
#ifdef IPPROTO_IPV6
	{ IPPROTO_IPV6,	"IPPROTO_IPV6"	},
#endif
#ifdef IPPROTO_ICMPV6
	{ IPPROTO_ICMPV6,"IPPROTO_ICMPV6"},
#endif
#ifdef IPPROTO_IGMP
	{ IPPROTO_IGMP,	"IPPROTO_IGMP"	},
#endif
#ifdef IPPROTO_HELLO
	{ IPPROTO_HELLO,"IPPROTO_HELLO"	},
#endif
#ifdef IPPROTO_ND
	{ IPPROTO_ND,	"IPPROTO_ND"	},
#endif
#ifdef IPPROTO_RAW
	{ IPPROTO_RAW,	"IPPROTO_RAW"	},
#endif
#ifdef IPPROTO_MAX
	{ IPPROTO_MAX,	"IPPROTO_MAX"	},
#endif
#ifdef IPPROTO_IPIP
	{ IPPROTO_IPIP,	"IPPROTO_IPIP"	},
#endif
	{ 0,		NULL		},
};

static struct xlat msg_flags[] = {  /* flag utilizzati nelle send.. e nelle recv.. */
	{ MSG_OOB,	"MSG_OOB"	},
#ifdef MSG_DONTROUTE
	{ MSG_DONTROUTE,"MSG_DONTROUTE"	},
#endif
#ifdef MSG_PEEK
	{ MSG_PEEK,	"MSG_PEEK"	},
#endif
#ifdef MSG_CTRUNC
	{ MSG_CTRUNC,	"MSG_CTRUNC"	},
#endif
#ifdef MSG_PROXY
	{ MSG_PROXY,	"MSG_PROXY"	},
#endif
#ifdef MSG_EOR
	{ MSG_EOR,	"MSG_EOR"	},
#endif
#ifdef MSG_WAITALL
	{ MSG_WAITALL,	"MSG_WAITALL"	},
#endif
#ifdef MSG_TRUNC
	{ MSG_TRUNC,	"MSG_TRUNC"	},
#endif
#ifdef MSG_CTRUNC
	{ MSG_CTRUNC,	"MSG_CTRUNC"	},
#endif
#ifdef MSG_ERRQUEUE
	{ MSG_ERRQUEUE,	"MSG_ERRQUEUE"	},
#endif
#ifdef MSG_DONTWAIT
	{ MSG_DONTWAIT,	"MSG_DONTWAIT"	},
#endif
#ifdef MSG_CONFIRM
	{ MSG_CONFIRM,	"MSG_CONFIRM"	},
#endif
#ifdef MSG_PROBE
	{ MSG_PROBE,	"MSG_PROBE"	},
#endif
	{ 0,		NULL		},
};

static struct xlat sockoptions[] = { /* opzioni dei socket in getsockopt e setsockopt */
#ifdef SO_PEERCRED
	{ SO_PEERCRED,	"SO_PEERCRED"	},
#endif
#ifdef SO_PASSCRED
	{ SO_PASSCRED,	"SO_PASSCRED"	},
#endif
#ifdef SO_DEBUG
	{ SO_DEBUG,	"SO_DEBUG"	},
#endif
#ifdef SO_REUSEADDR
	{ SO_REUSEADDR,	"SO_REUSEADDR"	},
#endif
#ifdef SO_KEEPALIVE
	{ SO_KEEPALIVE,	"SO_KEEPALIVE"	},
#endif
#ifdef SO_DONTROUTE
	{ SO_DONTROUTE,	"SO_DONTROUTE"	},
#endif
#ifdef SO_BROADCAST
	{ SO_BROADCAST,	"SO_BROADCAST"	},
#endif
#ifdef SO_LINGER
	{ SO_LINGER,	"SO_LINGER"	},
#endif
#ifdef SO_OOBINLINE
	{ SO_OOBINLINE,	"SO_OOBINLINE"	},
#endif
#ifdef SO_TYPE
	{ SO_TYPE,	"SO_TYPE"	},
#endif
#ifdef SO_ERROR
	{ SO_ERROR,	"SO_ERROR"	},
#endif
#ifdef SO_SNDBUF
	{ SO_SNDBUF,	"SO_SNDBUF"	},
#endif
#ifdef SO_RCVBUF
	{ SO_RCVBUF,	"SO_RCVBUF"	},
#endif
#ifdef SO_NO_CHECK
	{ SO_NO_CHECK,	"SO_NO_CHECK"	},
#endif
#ifdef SO_PRIORITY
	{ SO_PRIORITY,	"SO_PRIORITY"	},
#endif
#ifdef SO_ACCEPTCONN
	{ SO_ACCEPTCONN,"SO_ACCEPTCONN"	},
#endif
#ifdef SO_USELOOPBACK
	{ SO_USELOOPBACK,"SO_USELOOPBACK"},
#endif
#ifdef SO_SNDLOWAT
	{ SO_SNDLOWAT,	"SO_SNDLOWAT"	},
#endif
#ifdef SO_RCVLOWAT
	{ SO_RCVLOWAT,	"SO_RCVLOWAT"	},
#endif
#ifdef SO_SNDTIMEO
	{ SO_SNDTIMEO,	"SO_SNDTIMEO"	},
#endif
#ifdef SO_RCVTIMEO
	{ SO_RCVTIMEO,	"SO_RCVTIMEO"	},
#endif
#ifdef SO_BSDCOMPAT
	{ SO_BSDCOMPAT,	"SO_BSDCOMPAT"	},
#endif
#ifdef SO_REUSEPORT
	{ SO_REUSEPORT,	"SO_REUSEPORT"	},
#endif
#ifdef SO_ORDREL
	{ SO_ORDREL,	"SO_ORDREL"	},
#endif
#ifdef SO_IMASOCKET
	{ SO_IMASOCKET,	"SO_IMASOCKET"	},
#endif
#ifdef SO_MGMT
	{ SO_MGMT,	"SO_MGMT"	},
#endif
#ifdef SO_LISTENING
	{ SO_LISTENING,	"SO_LISTENING"	},
#endif
#ifdef SO_RDWR
	{ SO_RDWR,	"SO_RDWR"	},
#endif
#ifdef SO_SEMA
	{ SO_SEMA,	"SO_SEMA"	},
#endif
#ifdef SO_PARALLELSVR
	{ SO_PARALLELSVR,"SO_PARALLELSVR"},
#endif
#ifdef SO_PROTOTYPE
	{ SO_PROTOTYPE,	"SO_PROTOTYPE"	},
#endif
#ifdef SO_ALLRAW
	{ SO_ALLRAW,	"SO_ALLRAW"	},
#endif
#ifdef SO_ICS
	{ SO_ICS,	"SO_ICS"	},
#endif
	{ 0,		NULL		},
};


#if !defined (SOL_IP) && defined (IPPROTO_IP)
#define SOL_IP IPPROTO_IP
#endif

#ifdef SOL_IP
static struct xlat sockipoptions[] = {   /* opzioni dei socket in getsockopt e setsockopt a livello IP */
#ifdef IP_TOS
	{ IP_TOS,		"IP_TOS"		},
#endif
#ifdef IP_TTL
	{ IP_TTL,		"IP_TTL"		},
#endif
#ifdef IP_HDRINCL
	{ IP_HDRINCL,		"IP_HDRINCL"		},
#endif
#ifdef IP_OPTIONS
	{ IP_OPTIONS,		"IP_OPTIONS"		},
#endif
#ifdef IP_ROUTER_ALERT
	{ IP_ROUTER_ALERT,	"IP_ROUTER_ALERT"	},
#endif
#ifdef IP_RECVOPTIONS
	{ IP_RECVOPTIONS,	"IP_RECVOPTIONS"	},
#endif
#ifdef IP_RECVOPTS
	{ IP_RECVOPTS,		"IP_RECVOPTS"		},
#endif
#ifdef IP_RECVRETOPTS
	{ IP_RECVRETOPTS,	"IP_RECVRETOPTS"	},
#endif
#ifdef IP_RECVDSTADDR
	{ IP_RECVDSTADDR,	"IP_RECVDSTADDR"	},
#endif
#ifdef IP_RETOPTS
	{ IP_RETOPTS,		"IP_RETOPTS"		},
#endif
#ifdef IP_PKTINFO
	{ IP_PKTINFO,		"IP_PKTINFO"		},
#endif
#ifdef IP_PKTOPTIONS
	{ IP_PKTOPTIONS,	"IP_PKTOPTIONS"		},
#endif
#ifdef IP_MTU_DISCOVER
	{ IP_MTU_DISCOVER,	"IP_MTU_DISCOVER"	},
#endif
#ifdef IP_RECVERR
	{ IP_RECVERR,		"IP_RECVERR"		},
#endif
#ifdef IP_RECVTTL
	{ IP_RECVTTL,		"IP_RECRECVTTL"		},
#endif
#ifdef IP_RECVTOS
	{ IP_RECVTOS,		"IP_RECRECVTOS"		},
#endif
#ifdef IP_MTU
	{ IP_MTU,		"IP_MTU"		},
#endif
#ifdef IP_MULTICAST_IF
	{ IP_MULTICAST_IF,	"IP_MULTICAST_IF"	},
#endif
#ifdef IP_MULTICAST_TTL
	{ IP_MULTICAST_TTL,	"IP_MULTICAST_TTL"	},
#endif
#ifdef IP_MULTICAST_LOOP
	{ IP_MULTICAST_LOOP,	"IP_MULTICAST_LOOP"	},
#endif
#ifdef IP_ADD_MEMBERSHIP
	{ IP_ADD_MEMBERSHIP,	"IP_ADD_MEMBERSHIP"	},
#endif
#ifdef IP_DROP_MEMBERSHIP
	{ IP_DROP_MEMBERSHIP,	"IP_DROP_MEMBERSHIP"	},
#endif
#ifdef IP_BROADCAST_IF
	{ IP_BROADCAST_IF,	"IP_BROADCAST_IF"	},
#endif
#ifdef IP_RECVIFINDEX
	{ IP_RECVIFINDEX,	"IP_RECVIFINDEX"	},
#endif
	{ 0,			NULL			},
};
#endif /* SOL_IP */


#ifdef SOL_IPX
static struct xlat sockipxoptions[] = {   /* opzioni dei socket in getsockopt e setsockopt a livello IPX */
  { IPX_TYPE,     "IPX_TYPE"      },
  { 0,            NULL            },
};
#endif /* SOL_IPX */

/*******************/
/* nuova struttura */
/*******************/
#ifdef SOL_RAW
static struct xlat sockrawoptions[] = {
#if defined(ICMP_FILTER)
	{ ICMP_FILTER,		"ICMP_FILTER"	},
#endif
	{ 0,			NULL		},
};
#endif /* SOL_RAW */


/*******************/
/* nuova struttura */
/*******************/
#ifdef SOL_PACKET
static struct xlat sockpacketoptions[] = {
	{ PACKET_ADD_MEMBERSHIP,	"PACKET_ADD_MEMBERSHIP"	},
	{ PACKET_DROP_MEMBERSHIP,	"PACKET_DROP_MEMBERSHIP"},
#if defined(PACKET_RECV_OUTPUT)
	{ PACKET_RECV_OUTPUT,		"PACKET_RECV_OUTPUT"	},
#endif
#if defined(PACKET_RX_RING)
	{ PACKET_RX_RING,		"PACKET_RX_RING"	},
#endif
#if defined(PACKET_STATISTICS)
	{ PACKET_STATISTICS,		"PACKET_STATISTICS"	},
#endif
	{ 0,				NULL			},
};
#endif /* SOL_PACKET */


#if  !defined (SOL_TCP) && defined (IPPROTO_TCP)
#define SOL_TCP IPPROTO_TCP
#endif

#ifdef SOL_TCP
static struct xlat socktcpoptions[] = {   /* opzioni dei socket in getsockopt e setsockopt a livello TCP */
	{ TCP_NODELAY,		"TCP_NODELAY"	},
	{ TCP_MAXSEG,		"TCP_MAXSEG"	},
#if defined(TCP_CORK)
	{ TCP_CORK,		"TCP_CORK"	},
#endif
#if defined(TCP_KEEPIDLE)
	{ TCP_KEEPIDLE,		"TCP_KEEPIDLE" },
#endif
#if defined(TCP_KEEPINTVL)
	{ TCP_KEEPINTVL,	"TCP_KEEPINTVL" },
#endif
#if defined(TCP_KEEPCNT)
	{ TCP_KEEPCNT,		"TCP_KEEPCNT" },
#endif
#if defined(TCP_NKEEP)
	{ TCP_NKEEP,		"TCP_NKEEP"	},
#endif
#if defined(TCP_SYNCNT)
	{ TCP_SYNCNT,		"TCP_SYNCNT" },
#endif
#if defined(TCP_LINGER2)
	{ TCP_LINGER2,		"TCP_LINGER2" },
#endif
#if defined(TCP_DEFER_ACCEPT)
	{ TCP_DEFER_ACCEPT,	"TCP_DEFER_ACCEPT" },
#endif
#if defined(TCP_WINDOW_CLAMP)
	{ TCP_WINDOW_CLAMP,	"TCP_WINDOW_CLAMP" },
#endif
#if defined(TCP_INFO)
	{ TCP_INFO,		"TCP_INFO" },
#endif
#if defined(TCP_QUICKACK)
	{ TCP_QUICKACK,		"TCP_QUICKACK" },
#endif
	{ 0,			NULL		},
};
#endif  /* SOL_TCP */


/*******************/
/* nuova struttura */
/*******************/
#ifdef SOL_RAW
static struct xlat icmpfilterflags[] = {
#if defined(ICMP_ECHOREPLY)
	{ (1<<ICMP_ECHOREPLY),		"ICMP_ECHOREPLY"	},
#endif
#if defined(ICMP_DEST_UNREACH)
	{ (1<<ICMP_DEST_UNREACH),	"ICMP_DEST_UNREACH"	},
#endif
#if defined(ICMP_SOURCE_QUENCH)
	{ (1<<ICMP_SOURCE_QUENCH),	"ICMP_SOURCE_QUENCH"	},
#endif
#if defined(ICMP_REDIRECT)
	{ (1<<ICMP_REDIRECT),		"ICMP_REDIRECT"		},
#endif
#if defined(ICMP_ECHO)
	{ (1<<ICMP_ECHO),		"ICMP_ECHO"		},
#endif
#if defined(ICMP_TIME_EXCEEDED)
	{ (1<<ICMP_TIME_EXCEEDED),	"ICMP_TIME_EXCEEDED"	},
#endif
#if defined(ICMP_PARAMETERPROB)
	{ (1<<ICMP_PARAMETERPROB),	"ICMP_PARAMETERPROB"	},
#endif
#if defined(ICMP_TIMESTAMP)
	{ (1<<ICMP_TIMESTAMP),		"ICMP_TIMESTAMP"	},
#endif
#if defined(ICMP_TIMESTAMPREPLY)
	{ (1<<ICMP_TIMESTAMPREPLY),	"ICMP_TIMESTAMPREPLY"	},
#endif
#if defined(ICMP_INFO_REQUEST)
	{ (1<<ICMP_INFO_REQUEST),	"ICMP_INFO_REQUEST"	},
#endif
#if defined(ICMP_INFO_REPLY)
	{ (1<<ICMP_INFO_REPLY),		"ICMP_INFO_REPLY"	},
#endif
#if defined(ICMP_ADDRESS)
	{ (1<<ICMP_ADDRESS),		"ICMP_ADDRESS"		},
#endif
#if defined(ICMP_ADDRESSREPLY)
	{ (1<<ICMP_ADDRESSREPLY),	"ICMP_ADDRESSREPLY"	},
#endif
	{ 0,				NULL			},
};
#endif /* SOL_RAW */

/*******************/
/* nuova struttura */
/*******************/
#if defined(AF_PACKET) /* from e.g. linux/if_packet.h */
static struct xlat af_packet_types[] = {
#if defined(PACKET_HOST)
	{ PACKET_HOST,			"PACKET_HOST"		},
#endif
#if defined(PACKET_BROADCAST)
	{ PACKET_BROADCAST,		"PACKET_BROADCAST"	},
#endif
#if defined(PACKET_MULTICAST)
	{ PACKET_MULTICAST,		"PACKET_MULTICAST"	},
#endif
#if defined(PACKET_OTHERHOST)
	{ PACKET_OTHERHOST,		"PACKET_OTHERHOST"	},
#endif
#if defined(PACKET_OUTGOING)
	{ PACKET_OUTGOING,		"PACKET_OUTGOING"	},
#endif
#if defined(PACKET_LOOPBACK)
	{ PACKET_LOOPBACK,		"PACKET_LOOPBACK"	},
#endif
#if defined(PACKET_FASTROUTE)
	{ PACKET_FASTROUTE,		"PACKET_FASTROUTE"	},
#endif
	{ 0,				NULL			},
};
#endif /* defined(AF_PACKET) */

/****************************************************************************************************************************
 ****************************************************************************************************************************/


typedef union various_sockaddr {
  char pad[128];
  struct sockaddr sa;
  struct sockaddr_in sin;
  struct sockaddr_un sau;
#ifdef HAVE_INET_NTOP
  struct sockaddr_in6 sa6;
#endif
#if defined(AF_IPX)
  struct sockaddr_ipx sipx;
#endif
#ifdef AF_PACKET
  struct sockaddr_ll ll;
#endif
#ifdef AF_NETLINK
  struct sockaddr_nl nl;
#endif
} sockaddrbuf;




/* utilizzata da printmsghdr, sys_bind, sys_accept, 
sys_sendto, sys_recvfrom */
void
realprintsock(info, addrbuf, addrlen)      
msg_building_info *info;
sockaddrbuf *addrbuf;
int addrlen;
{
  char string_addr[100];

  if (!addrbuf) 
    {
      tprintf("NULL");
      return;
    }

  tprintf("{sa_family=");
  printxval(addrfams, addrbuf->sa.sa_family, "AF_???");
  tprintf(", ");


  /*
  switch (addrbuf->sa.sa_family) {
  
  case AF_UNIX:
    if (addrlen==2) 
      {
	tprintf("<nil>");
      } 
    else 
      {
	if (addrbuf->sau.sun_path[0]) 
	  tprintf("path=\"%*.*s\"", addrlen-2, addrlen-2, addrbuf->sau.sun_path);
	else 
	  tprintf("path=@%*.*s", addrlen-3, addrlen-3, addrbuf->sau.sun_path+1);
      }
    break;

  case AF_INET:
    tprintf("sin_port=htons(%u), sin_addr=inet_addr(\"%s\")}", ntohs(addrbuf->sin.sin_port), inet_ntoa(addrbuf->sin.sin_addr));
    break;

#ifdef HAVE_INET_NTOP
  case AF_INET6:
    inet_ntop(AF_INET6, &(addrbuf->sa6.sin6_addr), string_addr, sizeof(string_addr));
    tprintf("sin6_port=htons(%u), inet_pton(AF_INET6, \"%s\", &sin6_addr), sin6_flowinfo=htonl(%u)}", ntohs(addrbuf->sa6.sin6_port), string_addr, ntohl(addrbuf->sa6.sin6_flowinfo));
    break;	
#endif

#if defined(AF_IPX) 
  case AF_IPX:
    {
      int i;
      tprintf("{sipx_port=htons(%u), ", ntohs(addrbuf->sipx.sipx_port));*/
      /* otherwise the IPX
       * addresses just look monstrous...
       * Anyways, feel free if you don't like
       * this way.. :) 
       */
  /*
      tprintf("%08lx:", (unsigned long)ntohl(addrbuf->sipx.sipx_network));
      for (i = 0; i<IPX_NODE_LEN; i++)
	tprintf("%02x", addrbuf->sipx.sipx_node[i]);
      tprintf("/[%02x]", addrbuf->sipx.sipx_type);
    }
    break;
    #endif*/ /* AF_IPX */
  /*
#ifdef AF_PACKET
  case AF_PACKET:
    {
      int i;
      tprintf("proto=%#04x, if%d, pkttype=%d, addr(%d)={%d, ",
	      ntohs(addrbuf->ll.sll_protocol),
	      addrbuf->ll.sll_ifindex,
	      addrbuf->ll.sll_pkttype,
	      addrbuf->ll.sll_halen,
	      addrbuf->ll.sll_hatype);
      for (i=0; i<addrbuf->ll.sll_addr[i]; i++) 
	tprintf("%02x", addrbuf->ll.sll_addr[i]);
    }
    break;
    #endif*/ /* AF_APACKET */
  /*
#ifdef AF_NETLINLK
  case AF_NETLINK:
    tprintf("pid=%d, groups=%08x", addrbuf->nl.nl_pid, addrbuf->nl.nl_groups);
    break;
    #endif*/ /* AF_NETLINK */

    /* AF_AX25 AF_APPLETALK AF_NETROM AF_BRIDGE AF_AAL5
       AF_X25 AF_ROSE etc. still need to be done */
  /*
  default:
    tprintf("{sa_family=%u, sa_data=", addrbuf->sa.sa_family);
    break;
    }*/

  	switch (addrbuf->sa.sa_family) {
	case AF_UNIX:
		if (addrlen==2) {
			tprintf("<nil>");
		} else if (addrbuf->sau.sun_path[0]) {
			tprintf("path=\"%-.*s\"", addrlen-2, addrbuf->sau.sun_path);
		} else {
			tprintf("path=@%-.*s", addrlen-3, addrbuf->sau.sun_path+1);
		}
		break;
	case AF_INET:
		tprintf("sin_port=htons(%u), sin_addr=inet_addr(\"%s\")",
			ntohs(addrbuf->sin.sin_port), inet_ntoa(addrbuf->sin.sin_addr));
		break;
#ifdef HAVE_INET_NTOP
	case AF_INET6:
		inet_ntop(AF_INET6, &addrbuf->sa6.sin6_addr, string_addr, sizeof(string_addr));
		tprintf("sin6_port=htons(%u), inet_pton(AF_INET6, \"%s\", &sin6_addr), sin6_flowinfo=%u",
				ntohs(addrbuf->sa6.sin6_port), string_addr,
				addrbuf->sa6.sin6_flowinfo);
#ifdef HAVE_STRUCT_SOCKADDR_IN6_SIN6_SCOPE_ID
		{
#if defined(HAVE_IF_INDEXTONAME) && defined(IN6_IS_ADDR_LINKLOCAL) && defined(IN6_IS_ADDR_MC_LINKLOCAL)
		    int numericscope = 0;
		    if (IN6_IS_ADDR_LINKLOCAL (&addrbuf->sa6.sin6_addr)
			    || IN6_IS_ADDR_MC_LINKLOCAL (&addrbuf->sa6.sin6_addr)) {
			char scopebuf[IFNAMSIZ + 1];
			
			if (if_indextoname (addrbuf->sa6.sin6_scope_id, scopebuf) == NULL)
			    numericscope++;
			else
			    tprintf(", sin6_scope_id=if_nametoindex(\"%s\")", scopebuf);
		    } else
			numericscope++;
		    
		    if (numericscope)
#endif
			tprintf(", sin6_scope_id=%u", addrbuf->sa6.sin6_scope_id);
		}
#endif
		    break;
#endif
#if defined(AF_IPX) && defined(linux)
	case AF_IPX:
		{
			int i;
			tprintf("sipx_port=htons(%u), ",
					ntohs(addrbuf->sipx.sipx_port));
			/* Yes, I know, this does not look too
			 * strace-ish, but otherwise the IPX
			 * addresses just look monstrous...
			 * Anyways, feel free if you don't like
			 * this way.. :) 
			 */
			tprintf("%08lx:", (unsigned long)ntohl(addrbuf->sipx.sipx_network));
			for (i = 0; i<IPX_NODE_LEN; i++)
				tprintf("%02x", addrbuf->sipx.sipx_node[i]);
			tprintf("/[%02x]", addrbuf->sipx.sipx_type);
		}
		break;
#endif /* AF_IPX && linux */
#ifdef AF_PACKET
	case AF_PACKET:
		{
			int i;
			tprintf("proto=%#04x, if%d, pkttype=",
					ntohs(addrbuf->ll.sll_protocol),
					addrbuf->ll.sll_ifindex);
			printxval(af_packet_types, addrbuf->ll.sll_pkttype, "?");
			tprintf(", addr(%d)={%d, ",
					addrbuf->ll.sll_halen,
					addrbuf->ll.sll_hatype);
			for (i=0; i<addrbuf->ll.sll_halen; i++) 
				tprintf("%02x", addrbuf->ll.sll_addr[i]);
		}
		break;

#endif /* AF_APACKET */
#ifdef AF_NETLINLK
	case AF_NETLINK:
		tprintf("pid=%d, groups=%08x", addrbuf->nl.nl_pid, addrbuf->nl.nl_groups);
		break;
#endif /* AF_NETLINK */
	/* AF_AX25 AF_APPLETALK AF_NETROM AF_BRIDGE AF_AAL5
	AF_X25 AF_ROSE etc. still need to be done */

	default:
	  //tprintf("sa_data=");
	  //printstr(tcp, (long) &((struct sockaddr *) addr)->sa_data,
	  //	sizeof addrbuf->sa.sa_data);
	  //break;
	  tprintf("{sa_family=%u, sa_data=", addrbuf->sa.sa_family);
	  break;
	}



  tprintf("}");
}



/* utilizzata da printmsghdr, sys_bind, sys_accept, 
sys_sendto, sys_recvfrom e dalla get_msghdr */
sockaddrbuf *
get_sock(tcp, addr, addrlen)      
struct tcb *tcp;
long addr;
int addrlen;
{
  sockaddrbuf *sab=((sockaddrbuf*) malloc(sizeof(sockaddrbuf)));

  get_error = 0;
  
  if (!sab)
    {
      get_error=1;
      fprintf(stderr,"get_sock: malloc failed\n");
      return NULL;
    }
    
  if (addr == 0) 
    {
      free(sab);
      fprintf(stderr,"get_sock: addr == 0\n");
      return NULL;
    }

  if ((addrlen<2) || (addrlen>sizeof(sockaddrbuf)))
    addrlen=sizeof(sockaddrbuf);

  if (umoven(tcp, addr, addrlen, (char*)sab) < 0) 
    {
      fprintf(stderr, "get_sock: umove socket failed\n");
      get_error = 1;
      free(sab);
      return NULL;
    }
 
  return sab;
}


/* utilizzata da printmsghdr */
static void
realprintiovec(info, iov, len)  
msg_building_info *info;
struct iovec *iov;
long   len;
{
  int i;
  
  if (iov == NULL) 
    {
      tprintf("NULL");
      return;
    }

  tprintf("[");
  for (i = 0; i < len; i++) 
    {
      if (i)
	tprintf(", ");
      tprintf("{");
      realprintstr((char *)iov[i].iov_base, (unsigned int)iov[i].iov_len);
      /*string length = iov[i].iov_len */
      tprintf("_, %lu}", (unsigned long)iov[i].iov_len);
    }
  tprintf("]");
}


/* utilizzata da printmsghdr e da get_msghdr */
struct iovec *
get_iovec(tcp, iovec, len) 
struct tcb *tcp;
struct iovec *iovec;
long len;
{
  int i;
  struct iovec *iov = (struct iovec *) malloc(len * sizeof *iov);
 
  get_error = 0;

  if (iov == NULL) 
    {
      fprintf(stderr, "get_iovec: No memory\n");
      get_error = 1;
      return NULL;
    }

  if (umoven(tcp, (long)iovec,len * sizeof *iov, (char *) iov) < 0) 
    {
      free(iov);
      get_error = 1;
      return NULL;
    }
  else 
    {
      for (i = 0; i < len; i++)
	iov[i].iov_base= (void *) get_str(tcp, (long) iov[i].iov_base,iov[i].iov_len);
      
      return iov;
    }

  return NULL;
}



/* utilizzata da sys_sendmsg e sys_recvmsg*/
static void
realprintmsghdr(info, msg) 
msg_building_info *info;
struct msghdr *msg;
{
  if (!msg) 
    {
      tprintf("NULL");
      return;
    }

  tprintf("{msg_name(%d)=", msg->msg_namelen);
  
  if (msg->msg_name)
    realprintsock(info, (long)msg->msg_name, msg->msg_namelen);
  else
    tprintf("NULL");

  tprintf(", msg_iov(%lu)=", (unsigned long)msg->msg_iovlen);

  if (msg->msg_iov)
    realprintiovec(info, msg->msg_iov, msg->msg_iovlen);
  else
   tprintf("NULL"); 
  
#ifdef HAVE_STRUCT_MSGHDR_MSG_CONTROL
  tprintf(", msg_controllen=%lu", (unsigned long) msg->msg_controllen);
  if (msg->msg_controllen) 
    tprintf(", msg_control=%#lx, ", (unsigned long) msg->msg_control);
  tprintf(", msg_flags=");
  if (printflags(msg_flags, msg->msg_flags)==0)
    tprintf("0");
#else /* !HAVE_STRUCT_MSGHDR_MSG_CONTROL */
  /*
    Criptika affermazione trovata in linux/socket.h:
    As we do 4.4BSD message passing we use a 4.4BSD message passing
    system, not 4.3. Thus msg_accrights(len) are now missing. They
    belong in an obscure libc emulation or the bin.    
  */
  tprintf("msg_accrights=%#lx, msg_accrightslen=%u", (unsigned long) msg->msg_accrights, msg->msg_accrightslen);
#endif /* !HAVE_STRUCT_MSGHDR_MSG_CONTROL */
  tprintf("}");
}



/*
  get_msghdr
  input: una struttura tcb, un indirizzo di memoria (struct msghdr *)
  output: un puntatore ad una struct msghdr
  derivata dalla printmsghdr
  utilizzata da sys_sendmsg e sys_recvmsg
*/
struct msghdr *
get_msghdr(tcp, addr)
struct tcb *tcp;
long addr; 
{
  struct msghdr *msg;
  
  get_error = 0;
  
  if ((void *) addr == NULL)
    return NULL;
  
  msg =(struct msghdr *)malloc(sizeof(struct msghdr));

  if (umove(tcp, addr, msg) < 0) 
    {
      get_error = 1;
      free(msg);
      fprintf(stderr, "get_msghdr: umove error\n");
      return NULL;
    }

  ((sockaddrbuf *)msg->msg_name) = get_sock(tcp,(long)msg->msg_name,msg->msg_namelen);

  msg->msg_iov = get_iovec(tcp,msg->msg_iov,msg->msg_iovlen);

  if (get_error) 
    get_error = 0;

  return msg;
}



int 
print_socket(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("socket(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(domains, *((int*)info->u_arg[0]), "PF_???");
	    break;
	  case 1:
	    printxval(socktypes, *((int*)info->u_arg[1]), "SOCK_???");
	    break;
	  case 2:
	    switch (*((int*)info->u_arg[0])) {
	    case PF_INET:
	      printxval(protocols, *((int*)info->u_arg[2]), "IPPROTO_???");
	      break;
#ifdef PF_IPX
	    case PF_IPX:  
	      /* BTW: I don't believe this.. */
	      tprintf("[");
	      printxval(domains, *((int*)info->u_arg[2]), "PF_???");
	      tprintf("]");
	      break;
#endif /* PF_IPX */
	    default:
	      tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	      break;
	    }
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) dominio (int)
  -) tipo di socket (int)
  -) protocollo (int)
*/
int
sys_socket(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;  
  case 2:
    if ( take_arg(0, tcp) == NULL )
      {
	fprintf(stderr, "sys_socket: take_arg(0, tcp) == NULL\n");
	return -1;
      }
    arg_alloc(tcp,n_arg);
    break;
  } 
  return 0;
}

  

int 
print_bind(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("bind(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintsock(info, ((sockaddrbuf*)info->u_arg[1]), *((int*)info->u_arg[2]));
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) descrittore del socket (int)
  -) struttura contenente l'indirizzo locale della macchina (struct sockaddr *)
  -) dimensione della struttura contenente l'indirizzo locale (socklen_t == unsigned long)
*/
int 
sys_bind(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int len, *ptr;

  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
//  if (entering(tcp)) {  CONTROLLARE perche' escludo il caso exiting

      if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	{
	  fprintf(stderr, "sys_bind: take_arg(2, tcp)\n");
	  return -1;
	}
      len = *ptr;

      ((sockaddrbuf *) tcp->u_arg[n_arg]) = get_sock(tcp, tcp->u_arg[n_arg],len);
      if ( get_error )
	{
	  fprintf(stderr, "sys_bind:  get_sock\n");
	  tcp->arg_state[n_arg] = EMPTY;
	  return -1;
	}
      tcp->arg_state[n_arg] = TAKEN;
//  }
    break;
  }
  return 0;
}


int 
print_connect(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("connect(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintsock(info, ((sockaddrbuf*)info->u_arg[1]), *((int*)info->u_arg[2]));
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int 
sys_connect(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  return sys_bind(tcp,n_arg);
}


int 
print_listen(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("listen(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int 
sys_listen(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_accept(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("accept(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    if (!info->u_arg[2] || (info->u_arg[1] == 0) || (syserror(info)))
		tprintf("%#lx", *((unsigned long *)info->u_arg[1]));
	    else 
	      realprintsock(info, ((sockaddrbuf*)info->u_arg[1]), *((int*)info->u_arg[2]));
	    break;
	  case 2:
	    if (!info->u_arg[2])
	      tprintf("NULL");
	    else
	      tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_accept(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  int len, *ptr;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_accept: syscall failed\n");
	    return -1;
	  }

	if (tcp->u_arg[1] == 0)
	  {
	    fprintf(stderr, "sys_accept: tcp->u_arg[1] == 0\n");
	    tcp->arg_state[1] = TAKEN; /* CONTROLLARE */
	  }
	else
	  {
	    if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	      {
		fprintf(stderr, "sys_accept: take_arg(2, tcp)\n");
		return -1;
	      }
	    len = *ptr;

	    ((sockaddrbuf *) tcp->u_arg[1]) = get_sock(tcp, tcp->u_arg[1],len);
	    if ( get_error )
	      {
		fprintf(stderr, "sys_accept: get_sock\n");
		tcp->arg_state[1] = EMPTY;
		return -1;
	      }
	    tcp->arg_state[1] = TAKEN;
	  }
      }
    break;
  case 2:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_accept: syscall failed\n");
	    return -1;
	  }

	((long*) tcp->u_arg[2]) = get_num(tcp, tcp->u_arg[2]);
	if ( get_error ) 
	  {
	    fprintf(stderr, "sys_accept: get_num\n");
	    tcp->arg_state[2] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[2] = TAKEN;
      }
    break;
  }
  return 0;
}




int 
print_send(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("send(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], *((unsigned int *)info->u_arg[2]));
	    break;
	  case 2:
	    /*lunghezza stringa argomento 1*/
	    tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  case 3:
	    if (printflags(msg_flags, *((int*)info->u_arg[3])) == 0)
	      tprintf("0");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero di argomenti: 4
  -) descrittore del socket (int)
  -) messaggio (const void *)
  -) lunghezza del messaggio (size_t)
  -) flags per come spedire il messaggio (int)
*/
int 
sys_send(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int bytes, *ptr;
  switch (n_arg) {
  case 0:
  case 2:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
//  if (entering(tcp)) {  CONTROLLARE (se lo volessi controllare in uscita l'arg ?) 
      
      if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	{
	  fprintf(stderr, "sys_send: take_arg(2, tcp)\n");
	  return -1;
	}
      bytes = *ptr;

      ((char*)tcp->u_arg[n_arg]) = get_str(tcp, tcp->u_arg[n_arg], bytes);
      if ( get_error )
	{
	  fprintf(stderr, "sys_send:  get_str\n");
	  tcp->arg_state[n_arg] = EMPTY;
	  return -1;
	}
      tcp->arg_state[n_arg] = TAKEN;
//  }
    break;
  }
  return 0;
}



int 
print_sendto(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("sendto(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], *((unsigned int *)info->u_arg[2]));
	    break;
	  case 2:
	    /*lunghezza stringa argomento 1*/
	    tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  case 3:
	    if (printflags(msg_flags, *((int*)info->u_arg[3])) == 0)
	      tprintf("0");
	    break;
	  case 4:
	    realprintsock(info, ((sockaddrbuf*)info->u_arg[4]), *((int*)info->u_arg[5]));
    	    break;
	  case 5:
	    tprintf("%lu", *((unsigned long*)info->u_arg[5]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero di argomenti: 6
  -) descrittore del socket (int)
  -) messaggio (const void *)
  -) lunghezza del messaggio (size_t = unsigned int)
  -) flags per come spedire il messaggio (int)
  -) indirizzo del destinatario (const struct sockaddr *)
  -) dimensione del parametro precedente (socklen_t = unsigned int);
*/
int 
sys_sendto(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int bytes, len, *ptr;

  switch (n_arg) {
  case 0:
  case 2:
  case 3:
  case 5:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
//  if (entering(tcp)) {  CONTROLLARE 

      if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	{
	  fprintf(stderr,"sys_sendto: take_arg(2, tcp)\n");
	  return -1;
	}
      bytes = *ptr;
           
      ((char*)tcp->u_arg[n_arg]) = get_str(tcp, tcp->u_arg[n_arg], bytes);
      if ( get_error )
	{
	  fprintf(stderr,"sys_sendto: get_str\n");
	  tcp->arg_state[n_arg] = EMPTY;
	  return -1;
	}
      tcp->arg_state[n_arg] = TAKEN;
//  }
    break;
  case 4:
 // if (entering(tcp)) {  CONTROLLARE 
      
      if ( (ptr = (int *) take_arg(5, tcp)) == NULL )
	{
	  fprintf(stderr,"sys_sendto: take_arg(5, tcp)\n");
	  return -1;
	}
      len = *ptr;

      ((sockaddrbuf *)tcp->u_arg[n_arg]) = get_sock(tcp, tcp->u_arg[n_arg],len);
      if ( get_error )
	{
	  fprintf(stderr,"sys_sendto: get_sock\n");
	  tcp->arg_state[n_arg] = EMPTY;
	  return -1;
	}
      tcp->arg_state[n_arg] = TAKEN;
//  }
    break;
  }
  return 0;
}


void 
free_sendmsg(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i,j,count;

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 2:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 1:
	if (arg_state[1] == TAKEN)
	  if (u_arg[1])
	    {
	      struct msghdr *msg = (struct msghdr *) u_arg[1];
	      if (msg->msg_name)
		free (msg->msg_name);
	      if (msg->msg_iov)
		{
		  struct iovec *iov=msg->msg_iov;
		  count=msg->msg_iovlen;
		  for (j=0;j<count;j++)
		    free(iov[j].iov_base);
		  free(msg->msg_iov);
		}
	      free(msg);
	    }
	arg_state[1] = EMPTY;
	break;
      }
    }
  return;
}  


void 
free_recvmsg(u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[]; 
int u_nargs;
{
  int i,j,count;

  for (i = 0; i < u_nargs; i++) 
    {
      switch (i) {
      case 0:
      case 2:
	if (arg_state[i] == TAKEN)
	  free( (int *) u_arg[i]); 
	arg_state[i] = EMPTY;
	break;
      case 1:
	if (arg_state[1] == TAKEN)
	  if (u_arg[1])
	    {
	      struct msghdr *msg = (struct msghdr *) u_arg[1];
	      if (msg->msg_name)
		free (msg->msg_name);
	      if (msg->msg_iov)
		{
		  struct iovec *iov=msg->msg_iov;
		  count=msg->msg_iovlen;
		  for (j=0;j<count;j++)
		    free(iov[j].iov_base);
		  free(msg->msg_iov);
		}
	      free(msg);
	    }
	arg_state[1] = EMPTY;
	break;
      }
    }
  return;
}  


int 
print_sendmsg(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("sendmsg(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintmsghdr(info,((struct msghdr *)info->u_arg[1]));
	    break;
	  case 2:
	    if (printflags(msg_flags, *((int*)info->u_arg[2])) == 0)
	      tprintf("0");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero di argomenti: 3
  -) descrittore del socket (int)
  -) una struttura utilizzata per minimizzare il numero di parametri forniti (man recv)
     E' definita in sys/socket.h (const struct msghdr *)
     struct msghdr {
       void         *msg_name;       //destination address 
                                     //serve solo se il socket e' sconnesso
                                     //puo' essere NULL se non sono richiesti o desiderati dei nomi
       socklen_t     msg_namelen;    //size of address
                                     //serve solo se c'e' msg_name
       struct iovec *msg_iov;        //scatter/gather array
                                     //per descrivere locazioni scatter-gather
       size_t        msg_iovlen;     //numero di elementi in msg_iov
       void         *msg_control;    //dati ausiliari:
                                     //punta ad un buffer di messaggi di controllo di altri protocolli
				     //oppure di dati ausiliari eterogenei.
       socklen_t     msg_controllen; //lunghezza del buffer di dati ausiliari
       int           msg_flags;      //flags on received message
     };
  -) flags per come spedire il messaggio (int)
*/
int 
sys_sendmsg(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
  //if (entering(tcp)) {  CONTROLLARE 
      ((struct msghdr *) tcp->u_arg[n_arg]) = get_msghdr(tcp,(long)tcp->u_arg[n_arg]);
      if ( get_error )
	{
	  fprintf(stderr, "sys_sendmsg: get_msghdr\n");
	  tcp->arg_state[n_arg] = EMPTY;
	  return -1;
	}
      tcp->arg_state[n_arg] = TAKEN;
//  }
    break;
  }
  return 0;
}





int 
print_recv(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("recv(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], info->u_rval);
	    break;
	  case 2:
	    /*lunghezza stringa argomento 1*/
	    tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  case 3:
	    if (printflags(msg_flags, *((int*)info->u_arg[3])) == 0)
	      tprintf("0");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 4
  -) descrittore del socket (int)
  -) buffer contenente il messaggio (void *)
  -) lunghezza del buffer (size_t == int)
  -) flags (int)
*/
int 
sys_recv(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int bytes, *ptr;

  switch (n_arg) {
  case 0:
  case 2:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (!(syserror(tcp)))
	  {
	    if ( (ptr = (int *) take_arg(2, tcp)) == NULL )
	      {
		fprintf(stderr, "sys_recv: take_arg(2, tcp)\n");
		return -1;
	      }
	    bytes = *ptr;
	    
	    ((char*)tcp->u_arg[n_arg]) = get_str(tcp, tcp->u_arg[n_arg], bytes);
	    if ( get_error )
	      {
		fprintf(stderr, "sys_recv: get_str\n");
		tcp->arg_state[n_arg] = EMPTY;
		return -1;
	      }
	    tcp->arg_state[n_arg] = TAKEN;
	  }
	else
	  {
	    fprintf(stderr, "sys_recv: syscall failed\n");
	    return -1;
	  }
      }
    break;
  }
  return 0;
}




int 
print_recvfrom(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("recvfrom(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintstr((char*)info->u_arg[1], info->u_rval);
	    /*lunghezza stringa: info->u_rval*/
	    break;
	  case 2:
	    tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  case 3:
	    if (printflags(msg_flags, *((int*)info->u_arg[3])) == 0)
	      tprintf("0");
	    break;
	  case 4:
	    if (info->u_arg[4] == 0)
	      tprintf("NULL");
	    else
	      realprintsock(info, ((sockaddrbuf*)info->u_arg[4]), *((int*)info->u_arg[5]));
    	    break;
	  case 5:
	    if (info->u_arg[5] == 0)
	      tprintf("NULL");
	    else
	      tprintf("%lu", *((unsigned long*)info->u_arg[5]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
 Numero argomenti: 6
 -) descrittore del socket (int)
 -) buffer che contiene il messaggio ricevuto (void *)
 -) lunghezza del messaggio (size_t == unsigned int)
 -) flags di ricezione messi in OR (int)
 -) struttura contenente i dati del mittente, se il socket non e' connection-oriented (struct sockaddr *)
 -) dimensione del buffer con i dati del mittente (socklen_t == unsigned long)
*/

int 
sys_recvfrom(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int len, *ptr;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_recvfrom: syscall failed\n");
	    return -1;
	  }

	((char *)tcp->u_arg[n_arg]) = get_str(tcp, tcp->u_arg[n_arg], tcp->u_rval);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_recvfrom: get_str\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  case 2:
  case 3:
    if (exiting(tcp)) /* CONTROLLARE */
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_recvfrom: syscall failed\n");
	    return -1;
	  }
	
	arg_alloc(tcp,n_arg);
      }
    break;
  case 4:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_recvfrom: syscall failed\n");
	    return -1;
	  }

	if (tcp->u_arg[n_arg]) 
	  {	    
	    if ( (ptr = (int *) take_arg(5, tcp)) == NULL )
	      {
		fprintf(stderr, "sys_recvfrom: take_arg(5, tcp)\n");
		return -1;
	      }
	    len = *ptr;

	    ((sockaddrbuf *)tcp->u_arg[n_arg]) = get_sock(tcp, tcp->u_arg[n_arg],len);
	    if ( get_error )
	      {
		fprintf(stderr, "sys_recvfrom: get_sock\n");
		tcp->arg_state[n_arg] = EMPTY;
		return -1;
	      }
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  case 5:
    if (exiting(tcp)) /* CONTROLLARE */
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_recvfrom: syscall failed\n");
	    return -1;
	  }
	
	((long*) tcp->u_arg[n_arg]) = get_num(tcp, tcp->u_arg[n_arg]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_recvfrom: get_num\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  return 0;
}


int 
print_recvmsg(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("recvmsg(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintmsghdr(info,((struct msghdr *)info->u_arg[1]));
	    break;
	  case 2:
	    if (printflags(msg_flags, *((int*)info->u_arg[2])) == 0)
	      tprintf("0");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -)descrittore del socket (int)
  -) una struttura utilizzata per minimizzare il numero di parametri forniti (man recv)
     definita in sys/socket.h (struct msghdr *)
  -) dei flags di ricezione (int)
*/
int 
sys_recvmsg(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (!(syserror(tcp)))
	  {
	    ((struct msghdr *)tcp->u_arg[n_arg]) = get_msghdr(tcp,(long)tcp->u_arg[n_arg]);
	    if ( get_error )
	      {
		fprintf(stderr, "sys_recvmsg: get_msghdr\n");
		tcp->arg_state[n_arg] = EMPTY;
		return -1;
	      }
	    tcp->arg_state[n_arg] = TAKEN;
	  }
	else
	  {
	    fprintf(stderr, "sys_recvmsg: syscall failed\n");
	    return -1;
	  }
      }
    break;
  }
  return 0;
}



int 
print_shutdown(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("shutdown(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%ld", *((long*)info->u_arg[1]));
	    switch (*((int*)info->u_arg[1])) {
	    case 0:
	      tprintf("%s", " / receive /");
	      break;
	    case 1:
	      tprintf("%s", " / send /");
	      break;
	    case 2:
	      tprintf("%s", " / send and receive /");
	      break;
	    }
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");
  
  return 0;
}



int
sys_shutdown(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_getsockname(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("getsockname(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    if (!info->u_arg[2] || (info->u_arg[1] == 0) || (syserror(info)))
		tprintf("%#lx", *((unsigned long *)info->u_arg[1]));
	    else 
	      realprintsock(info, ((sockaddrbuf*)info->u_arg[1]), *((int*)info->u_arg[2]));
	    break;
	  case 2:
	    if (!info->u_arg[2])
	      tprintf("NULL");
	    else
	      tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_getsockname(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  return sys_accept(tcp,n_arg);
}



int 
print_getpeername(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("getpeername(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    if (!info->u_arg[2] || (info->u_arg[1] == 0) || (syserror(info)))
		tprintf("%#lx", *((unsigned long *)info->u_arg[1]));
	    else 
	      realprintsock(info, ((sockaddrbuf*)info->u_arg[1]), *((int*)info->u_arg[2]));
	    break;
	  case 2:
	    if (!info->u_arg[2])
	      tprintf("NULL");
	    else
	      tprintf("%lu", *((unsigned long *)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int 
sys_getpeername(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  return sys_accept(tcp,n_arg);
}



int 
print_pipe(info)
msg_building_info *info;
{
  int *fds;
  
  tprintf("pipe(");
  
  if (info->arg_state[0]==TAKEN)
    {
      fds=(int*)info->u_arg[0];
      tprintf("[%u, %u]", fds[0], fds[1]);
    }
  else
    tprintf("_");
  
  tprintf(") ");

  return 0;
}



int 
sys_pipe(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int *fds;

  if (exiting(tcp)) 
    {
      if (syserror(tcp))
	{
	  fprintf(stderr, "sys_pipe: syscall failed\n");
	  return -1;
	}

      if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&fds, 2*sizeof(int), "pipe")) /* CONTROLLARE */
	return -1;
    }
  return 0;
}



int 
print_socketpair(info)
msg_building_info *info;
{
  int i, *fds;
  int nargs=get_nargs(info->scno);
  
  tprintf("socketpair(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(domains, *((int*)info->u_arg[0]), "PF_???");
	    break;
	  case 1:
	    printxval(socktypes, *((int*)info->u_arg[1]), "SOCK_???");
	    break;
	  case 2:
	    switch (*((int*)info->u_arg[0])) {
	    case PF_INET:
	      printxval(protocols, *((int*)info->u_arg[2]), "IPPROTO_???");
	      break;
#ifdef PF_IPX
	    case PF_IPX:
	      /* BTW: I don't believe this.. */
	      tprintf("[");
	      printxval(domains, *((int*)info->u_arg[2]), "PF_???");
	      tprintf("]");
	      break;
#endif
	      /* PF_IPX */
	    default:	
	      tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	      break;
	    }	    
	    break;
	  case 3:
	    fds = (int*)info->u_arg[3];
	    tprintf("[%u, %u]", fds[0], fds[1]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}




/*
  Numero argomenti: 4
  -) dominio della coppia di socket (int)
  -) tipo della connessione (int)
  -) protocollo della coppia di socket (int)
  -) descrittori della coppia di socket (int [])
*/
int
sys_socketpair(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int *fds;

  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  case 2:
    if ( take_arg(0, tcp) == NULL )
      {
	fprintf(stderr,"sys_socketpair: take_arg(0, tcp)\n");
	return -1;
      }
    arg_alloc(tcp,n_arg);
    break;
  case 3:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_socketpair: syscall failed\n");
	    return -1;
	  }
	
	if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&fds, 2*sizeof(int), "socketpair")) /* CONTROLLARE */
	  return -1;
	
	((int *)tcp->u_arg[n_arg]) = fds;
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  return 0;
}




int 
print_getsockopt(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("getsockopt(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    printxval(socketlayers, *((int*)info->u_arg[1]), "SOL_???");
	    break;
/* 	  case 1: */
/* 	    switch (*((int*)info->u_arg[1])) { */
/* 	    case SOL_SOCKET: */
/* 	      tprintf("SOL_SOCKET"); */
/* 	      break; */
/* #ifdef SOL_IP */
/* 	    case SOL_IP: */
/* 	      tprintf("SOL_IP"); */
/* 	      break; */
/* #endif */
/* #ifdef SOL_IPX */
/* 	    case SOL_IPX: */
/* 	      tprintf("SOL_IPX"); */
/* 	      break; */
/* #endif */
/* #ifdef SOL_TCP */
/* 	    case SOL_TCP: */
/* 	      tprintf("SOL_TCP"); */
/* 	      break; */
/* #endif */
/* 	      /\* SOL_AX25 SOL_ROSE SOL_ATALK SOL_NETROM SOL_UDP SOL_DECNET SOL_X25 */
/* 		 etc. still need work *\/ */
/* 	    default:  */
/* 	      /\* XXX - should know socket family here *\/ */
/* 	      printxval(protocols, *((int*)info->u_arg[1]), "IPPROTO_???"); */
/* 	      break; */
/* 	    } */
/* 	    break; */
	  case 2:
	    switch (*((int*)info->u_arg[1])) {
	    case SOL_SOCKET:
	      printxval(sockoptions, *((int*)info->u_arg[2]), "SO_???");
	      break;
#ifdef SOL_IP
	    case SOL_IP:
	      printxval(sockipoptions, *((int*)info->u_arg[2]), "IP_???");
	      break;
#endif
#ifdef SOL_IPX
	    case SOL_IPX:
	      printxval(sockipxoptions, *((int*)info->u_arg[2]), "IPX_???");
	      break;
#endif
#ifdef SOL_PACKET
	    case SOL_PACKET:
	      printxval(sockpacketoptions, *((int*)info->u_arg[2]), "PACKET_???");
	      break;
#endif
#ifdef SOL_TCP
	    case SOL_TCP:
	      printxval(socktcpoptions, *((int*)info->u_arg[2]), "TCP_???");
	      break;
#endif
	      /* SOL_AX25 SOL_ROSE SOL_ATALK SOL_NETROM SOL_UDP SOL_DECNET SOL_X25
		 etc. still need work */
	    default: 
	      /* XXX - should know socket family here */
	      tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	      break;
	    }
	    break;
	  case 3:  /* CONTROLLARE non  detto che sia solo un intero (il case 4  OK) */

/*	    switch (tcp->u_arg[1]) {
	    case SOL_SOCKET:
	      switch (tcp->u_arg[2]) {
#ifdef SO_LINGER
	      case SO_LINGER:
		if (len == sizeof (struct linger)) {
		  struct linger linger;
		  if (umove (tcp,
			     tcp->u_arg[3],
			     &linger) < 0)
		    break;
		  tprintf(", {onoff=%d, linger=%d}, "
			  "[%ld]",
			  linger.l_onoff,
			  linger.l_linger,
			  len);
		  return 0;
		}
		break;
#endif
	      }
	      break;
	    }*/
	  case 4:
	    realprintnum(((int*)info->u_arg[i]), "%ld");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 5
  -) dominio del socket (int)
  -) livello delle opzioni: IP, TCP, HTTP, ... (int)
  -) nome dell'opzione richiesta: specifico per ogni protocollo (int)
  -) buffer nel quale andranno le opzioni richieste (void *)
  -) all'entrata contiene la dimensione del buffer precedente,
     all'uscita la dimensione effettiva del valore restituito.
     Puo' essere NULL. (socklen_t * == unsigned long *)
*/
int 
sys_getsockopt(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  case 2:
    if ( take_arg(1, tcp) == NULL )
      {
	fprintf(stderr, "sys_getsockopt: take_arg(1, tcp)\n");
	return -1;
      }

    arg_alloc(tcp,n_arg);
    break;
  case 3:  /* CONTROLLARE non  detto che sia solo un intero (il case 4  OK) */
  case 4:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_getsockopt: syscall failed\n");
	    return -1;
	  }
	
	((long*) tcp->u_arg[n_arg]) = get_num(tcp, tcp->u_arg[n_arg]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_getsockopt: get_num\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  return 0;
}


/* #if defined(ICMP_FILTER) */
/* static void printicmpfilter(tcp, addr) */
/* //struct tcb *tcp; */
/* msg_building_info *tcp; */
/* long addr; */
/* { */
/* 	struct icmp_filter	filter; */

/* 	if (!addr) { */
/* 		tprintf("NULL"); */
/* 		return; */
/* 	} */
/* 	if (syserror(tcp) || !verbose(tcp)) { */
/* 		tprintf("%#lx", addr); */
/* 		return; */
/* 	} */
/* 	if (umove(tcp, addr, &filter) < 0) { */
/* 		tprintf("{...}"); */
/* 		return; */
/* 	} */

/* 	tprintf("~("); */
/* 	if (printflags(icmpfilterflags, ~filter.data) == 0) */
/* 		tprintf("0"); */
/* 	tprintf(")"); */
/* } */
/* #endif /\* ICMP_FILTER *\/ */



int 
print_setsockopt(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("setsockopt(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    printxval(socketlayers, *((int*)info->u_arg[1]), "SOL_???");
	    break;
/* 	  case 1: */
/* 	    switch (*((int*)info->u_arg[1])) { */
/* 	    case SOL_SOCKET: */
/* 	      tprintf("SOL_SOCKET"); */
/* 	      break; */
/* #ifdef SOL_IP */
/* 	    case SOL_IP: */
/* 	      tprintf("SOL_IP"); */
/* 	      break; */
/* #endif */
/* #ifdef SOL_IPX */
/* 	    case SOL_IPX: */
/* 	      tprintf("SOL_IPX"); */
/* 	      break; */
/* #endif */
/* #ifdef SOL_TCP */
/* 	    case SOL_TCP: */
/* 	      tprintf("SOL_TCP"); */
/* 	      break; */
/* #endif */
/* 	      /\* SOL_AX25 SOL_ROSE SOL_ATALK SOL_NETROM SOL_UDP SOL_DECNET SOL_X25 */
/* 		 etc. still need work *\/ */
/* 	    default:  */
/* 	      /\* XXX - should know socket family here *\/ */
/* 	      printxval(protocols, *((int*)info->u_arg[1]), "IPPROTO_???"); */
/* 	      break; */
/* 	    } */
/* 	    break; */
	  case 2:
	    switch (*((int*)info->u_arg[1])) {
	    case SOL_SOCKET:
	      printxval(sockoptions, *((int*)info->u_arg[2]), "SO_???");
/*	      switch (name) {
#if defined(SO_LINGER)       CONTROLLARE
	      case SO_LINGER:
		if (len == sizeof (struct linger)) {
		  struct linger linger;
		  if (umove (tcp, addr, &linger) < 0)
		    break;
		  tprintf(", {onoff=%d, linger=%d}",
			  linger.l_onoff,
			  linger.l_linger);
		  return 0;
		}
		break;
#endif
}*/
	      break;
#ifdef SOL_IP
	    case SOL_IP:
	      printxval(sockipoptions, *((int*)info->u_arg[2]), "IP_???");
	      break;
#endif
#ifdef SOL_IPX
	    case SOL_IPX:
	      printxval(sockipxoptions, *((int*)info->u_arg[2]), "IPX_???");
	      break;
#endif
#ifdef SOL_PACKET
	    case SOL_PACKET:
	      printxval(sockpacketoptions, *((int*)info->u_arg[2]), "PACKET_???");
	      /* TODO: decode packate_mreq for PACKET_*_MEMBERSHIP */
	      break;
#endif
#ifdef SOL_TCP
	    case SOL_TCP:
	      printxval(socktcpoptions, *((int*)info->u_arg[2]), "TCP_???");
	      break;
#endif
#ifdef SOL_RAW
	    case SOL_RAW:
	      printxval(sockrawoptions, *((int*)info->u_arg[2]), "RAW_???");
/*	      switch (name) { CONTROLLARE
#if defined(ICMP_FILTER)
	      case ICMP_FILTER:
		tprintf(", ");
		printicmpfilter(info, info->u_arg[3]);
		return 0;
#endif
} */
	      break;
#endif
	      /* SOL_AX25 SOL_ROSE SOL_ATALK SOL_NETROM SOL_UDP SOL_DECNET SOL_X25
		 etc. still need work */
	    default: 
	      /* XXX - should know socket family here */
	      tprintf("%lu", *((unsigned long*)info->u_arg[2]));
	      break;
	    }
	    break;
	  case 3: /* CONTROLLARE: non  detto che sia un intero */
	    realprintnum(((int*)info->u_arg[i]), "%ld");
	    break;
	  case 4:
	    tprintf("%lu", *((unsigned long*)info->u_arg[4]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}




/*
  int setsockopt(int s, int level, int optname, const void *optval, socklen_t optlen);
  Numero argomenti: 5
  -) dominio del socket (int)
  -) livello delle opzioni: IP, TCP, HTTP, ... (int)
  -) nome dell'opzione richiesta: specifico per ogni protocollo (int)
  -) array di booleani per indicare se un certo parametro deve essere settato o meno (void *)
  -) dimensione dell'array precedente (socklen_t == unsigned long)
*/
int 
sys_setsockopt(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch(n_arg) {
  case 0:
  case 1:
  case 4:
    arg_alloc(tcp, n_arg);
    break;
  case 2:
    if ( take_arg(1, tcp) == NULL )
      {
	fprintf(stderr,"sys_setsockopt: take_arg(1, tcp)\n");
	return -1;
      }
    
    arg_alloc(tcp,n_arg);
    break;
  case 3: /* CONTROLLARE non  detto che sia un intero */
    ((long*) tcp->u_arg[n_arg]) = get_num(tcp, tcp->u_arg[n_arg]);
    if ( get_error )
      {
	fprintf(stderr,"sys_setsockopt: get_num\n");
	tcp->arg_state[n_arg] = EMPTY;
	return -1;
      }
    tcp->arg_state[n_arg] = TAKEN;
    break;
  }
  
  return 0;
}

