/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: pars_funct.c,v 1.2 2003/02/17 15:55:15 frassi Exp $
 */

/* funzione che dato il nome di una syscall restituisce 
   il corrispondente numero. Utilizza la struttura sysent
   allocata in syscall.c  
  
   struct sysent sysent0[] = {
     #include "syscallent.h"
   };
   struct sysent *sysent;
   sysent = sysent0;
*/

/* definita in defs.h (in questo file ci vorra'  #include "defs.h") */
//#include "y.tab.h"

#include <assert.h>

#include "type_sys_args.h"
#include "pars_funct.h"


/*************************
 * TABELLA DELLE SYSCALL *
 *************************/			 

/*
static struct sysent sysent[] = {
#include "linux/syscallent.h"
};
*/

extern int nsyscalls0; /* dichiarata in syscall.c *** il suo valore attuale e' 343. Per un fatto di prestazioni, sarebbe meglio 
			  definirla tramite define... */


int 
find_sysc_number(char *sysc_name)
     /* dato il nome di una syscall restituisce il numero corrispondente */
{
  int i;
  
  for (i = 0; i < nsyscalls0; i++) {
    if (strcmp(sysc_name, sysent[i].sys_name) == 0) 
      return i;
  }
  
  return -1; /* sysc_name inesistente in sysent */
}


char *
get_sysc_name(int num_sysc)
     /* dato il numero di una syscall ne restituisce il nome */
{
  
  if  (num_sysc == -1) return "ANY";
  if  (num_sysc == -2) return "DELTA";
  if  (num_sysc == -3) return "#";
    
  if (num_sysc < nsyscalls0) 
    return sysent[num_sysc].sys_name;
  
  return NULL;
}


int
get_nargs(int num_sysc)
     /* dato il numero di una syscall ne restistuisce il numero di argomenti */
{
  if (num_sysc < nsyscalls0)
    return sysent[num_sysc].nargs;
  else 
    return -1;
}


/*****************************************
 * TABELLA DEGLI ARGOMENTI DELLE SYSCALL *
 *****************************************/

/* entry della tabella */
struct sysargs {
  int	a;          /* tipo del primo argomento */
  int	b;          /* tipo del secondo argomento */
  int   c;          /* ... */
  int	d;
  int   e;
  int   f;
};


static struct sysargs sysargs[] = {
#include "linux/syscallargsent.h"
};


sys_args_type_t
find_type (int sysc_n, int arg_n) /* Utilizzata in wids.y (PARSER) */
     /* dato il numero della syscall e il numero dell'argomento 
	restituisce il tipo dell'argomento */
{
  switch (arg_n) {
  case   1:
    return sysargs[sysc_n].a;
    break; 
  case   2:
    return sysargs[sysc_n].b;
    break; 
  case   3:
    return sysargs[sysc_n].c;
    break; 
  case   4:
    return sysargs[sysc_n].d;
    break; 
  case   5:
    return sysargs[sysc_n].e;
    break; 
  case   6:
    return sysargs[sysc_n].f;
    break;
  case -1:
    return LONG;
    break;
  }
  return 0; /* non dovrebbe essere mai eseguito */
}


/*********************************************************
 * TABELLA SULLA VALIDITA' DEGLI ARGOMENTI DELLA SYSCALL *
 *********************************************************/

#define EUK -1  /* exist but unknown */ 
#define A   0   /* absent */
#define B   1   /* both: argomento valido sia in entrata che in uscita */
#define I   2   /* intro: argomento valido in entrata */
#define O   3   /* output: argomento in uscita */
#define IG  4   /* argomento esistente nel prototipo ma, in realta', ignorato */

struct args_io {
  int io0; /*indica se il parametro in questione e' valido in input o in output*/
  int io1; /*indica se il parametro in questione e' valido in input o in output*/
  int io2; /*indica se il parametro in questione e' valido in input o in output*/
  int io3; /*indica se il parametro in questione e' valido in input o in output*/
  int io4; /*indica se il parametro in questione e' valido in input o in output*/
  int io5; /*indica se il parametro in questione e' valido in input o in output*/
};

struct args_io args_io[] = {
#include "linux/syscallinout.h"
};


int 
is_input(int sysc_num, int arg_number)
{
  int io;
  
  arg_number--;
  assert( sysc_num < nsyscalls0 && arg_number >= 0 && arg_number < get_nargs(sysc_num));
  
  io = *(((int *) (args_io + sysc_num)) + arg_number);
  switch (io)
    {
    case I:
    case B:
      return 1;
    default:
      ;
    }
  return 0;
}


int 
is_output(int sysc_num, int arg_number)
{
  int io;
  
  arg_number--;
  assert( sysc_num < nsyscalls0 && arg_number >= 0 && arg_number < get_nargs(sysc_num));

  io = *(((int *) (args_io + sysc_num)) + arg_number);
  switch (io)
    {
    case O:
    case B:
      return 1;
    default:
      ;
    }
  return 0;
}


int 
is_both(int sysc_num, int arg_number)
{
  int io;

  arg_number--;
  assert( sysc_num < nsyscalls0 && arg_number >= 0 && arg_number < get_nargs(sysc_num));

  io = *(((int *) (args_io + sysc_num)) + arg_number);
  if (io == B)
    return 1;
  
  return 0;
}


/**************************************************
 * TABELLA DEI TIPI DEGLI ARGOMENTI DELLE SYSCALL *
 **************************************************

 Questa tabella riunisce tutti i tipi degli argomenti
 delle syscall del Kernel Linux.

*/


struct xlat types_ent[] = {
#include "linux/type_sys_args_ent.h"
};


sys_args_type_t
get_type(char *type_name) /* Utilizzata wids.lex quando siamo nella dichiarazione di un prototipo di funzione */
{
  return get_xlat_val(types_ent, type_name);
}
