#ifndef PARS_FUNCT_H
#define PARS_FUNCT_H

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: pars_funct.h,v 1.1.1.1 2001/12/17 08:58:29 frassi Exp $
 */


#include "xlat.h"

/*************************
 * TABELLA DELLE SYSCALL *
 *************************/			 

int find_sysc_number(char *sysc_name);
/* dato il nome di una syscall restituisce il numero corrispondente */

char *get_sysc_name(int num_sysc);
/* dato il numero di una syscall ne restituisce il nome */

int get_nargs(int num_sysc);
/* dato il numero di una syscall ne restistuisce il numero di argomenti */



/*****************************************
 * TABELLA DEGLI ARGOMENTI DELLE SYSCALL *
 *****************************************/

int find_type(int sysc_n, int arg_n);
/* dato il numero della syscall e il numero dell'argomento 
   restituisce il tipo dell'argomento */



/*********************************************************
 * TABELLA SULLA VALIDITA' DEGLI ARGOMENTI DELLA SYSCALL *
 *********************************************************/

int is_input(int sysc_num, int arg_number);
/* (*) argomenti:
   sysc_num       - numero della chiamata di sistema
   arg_number     - numero dell'argomento compreso in {1,..,6} 
   (*) valore di ritorno:
   1   - se il parametro arg_number e' un valore necessario in entrata 
         alla chiamata di sistema sysc_num, 
   0   - altrimenti 
*/

int is_output(int sysc_num, int arg_number);
/* (*) argomenti:
   sysc_num       - numero della chiamata di sistema
   arg_number     - numero dell'argomento compreso in {1,..,6} 
   (*) valore di ritorno:
   1    - se il paramentro arg_number e' un valore restituito dalla
          chiamata di sistema sysc_num, 
   0    - altrimenti 
*/

int is_both(int sysc_num, int arg_number);
/* (*) argomenti:
   sysc_num       - numero della chiamata di sistema
   arg_number     - numero dell'argomento compreso in {1,..,6} 
   (*) valore di ritorno:
   1 - se il parametro arg_number e' valido in entrata alla chiamata di sistema sysc_num, 
       e in uscita puo' essere sovrascritto dalla stessa con un nuovo valore
   0 - altrimenti
*/


/**************************************************
 * TABELLA DEI TIPI DEGLI ARGOMENTI DELLE SYSCALL *
 **************************************************

 Questa tabella riunisce tutti i tipi degli argomenti
 delle syscall del kernel linux.

*/

int get_type(char *type_name);
/*   dato il nome del tipo di un argomento di una syscall ne restituisce 
     l'identificatore intero associato; se non esiste restituisce NOT_FOUND */


#endif /* !PARS_FUNCT_H */



