/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: parser.c,v 1.1.1.1 2001/12/17 08:58:30 frassi Exp $
 */

#include <assert.h>

#include "parser.h"
#include "runtime_util.h"

#ifndef Y_TAB_H
#define Y_TAB_H

/* file header per wids.lex */
#include "y.tab.h"

#endif /* !Y_TAB_H */

extern char *get_sysc_name(int num_sysc);

/**************************************************************************************
 *************** DEFINIZIONI DI FUNZIONI PER LE TABELLE DEI SIMBOLI *******************
 **************************************************************************************/



/***************************
 * TABELLA DELLE VARIABILI *
 ***************************/


var_entry *
clone_var_entry(var_entry *ve)
{
  var_entry *new_ve = (var_entry *) malloc(sizeof(var_entry));
  new_ve->name = clone_string(ve->name);
  new_ve->type_var = ve->type_var;
  new_ve->type_values = ve->type_values;
  
  if (ve->type_var == scalar)
    {
      if (ve->values != NULL)
	switch (ve->type_values) 
	  {
	  case STRING_VALUE: 
	    new_ve->values = clone_string((char *) ve->values);
	    break;
	  case CHARACTER_VALUE: 
	    new_ve->values = clone_char((char *) ve->values);
	    break;
	  case LONG_VALUE:
	    new_ve->values = clone_long((long *) ve->values);
	    break;
	  default:
	    assert(ve->type_values != STRING_VALUE && ve->type_values != CHARACTER_VALUE && ve->type_values != LONG_VALUE);
	  }
      else new_ve->values = NULL;
    }
  else
    new_ve->values = clone_hash_table((GHashTable *) ve->values, ve->type_values);
  
  return new_ve;
}


void 
free_var_entry(var_entry *ve)
{
  if ( ve->type_var == scalar )
    free(ve->values);
  else
    free_hash_table((GHashTable *) ve->values);
  free(ve->name);
  free(ve);
}


static void 
var_table_insert(gpointer key, gpointer value, gpointer user_data)
{
  var_entry *value_clone;
  //char *key_clone = clone_string((char *) key);

  value_clone = clone_var_entry((var_entry *) value);
  g_hash_table_insert((GHashTable *) user_data, value_clone->name, value_clone);
}


var_table_t
clone_var_table(var_table_t var_table)
{
  GHashTable *new_var_table = g_hash_table_new(g_str_hash, g_str_equal);
  g_hash_table_foreach(var_table, var_table_insert, new_var_table);
  return new_var_table;
}


static void  
free_var_el(gpointer key, gpointer value, gpointer user_data)
{
  /* free(key);  key e value->name puntano al solito spazio allocato (ci pensa free_var_entry a deallocarlo) */
  free_var_entry((var_entry *) value);
}


void 
free_var_table(var_table_t var_table)
{
  g_hash_table_foreach(var_table, free_var_el, NULL);
  g_hash_table_destroy(var_table);
}


/* chiave per tabelle di tipo macro_var_table_t */

struct macro_var_key {
  char *name;
  int macro_id;
};

/* * */

void 
putmacrovar(macro_var_table_t var_table, char *var_name, int macro_id, var_entry *var_attr)
{ 
  struct macro_var_key *a_macro_var_key = (struct macro_var_key *) malloc(sizeof(struct macro_var_key)); 
  a_macro_var_key->name = var_name; 
  a_macro_var_key->macro_id = macro_id; 
  g_hash_table_insert(var_table, a_macro_var_key, var_attr);
}


var_entry *
getmacrovar(macro_var_table_t var_table, char *var_name, int macro_id)  
{ 
  struct macro_var_key a_macro_var_key; 
  a_macro_var_key.name = var_name; 
  a_macro_var_key.macro_id = macro_id; 
  return g_hash_table_lookup(var_table, &a_macro_var_key); 
}


/* struttura di supporto per la funzione inmacrovarlist() */

struct inmacrovarlist_data_t {
  int macro_id;
  GList *list;
};

/* * */

static void 
inmacrovarlist(gpointer key, gpointer value, gpointer user_data)
{
  if ( ((struct macro_var_key *) key)->macro_id == ((struct inmacrovarlist_data_t *) user_data)->macro_id )
    ((struct inmacrovarlist_data_t *) user_data)->list = g_list_prepend( ((struct inmacrovarlist_data_t *) user_data)->list, value);
} 


GList *
getmacrovarlist(macro_var_table_t var_table, int macro_id)
{
  struct inmacrovarlist_data_t data; 
  data.macro_id = macro_id;
  data.list = NULL;
  g_hash_table_foreach(var_table, inmacrovarlist, &data);
  return data.list;
}


static void
print_var_el(gpointer key, gpointer value, gpointer user_data)
{
  printf("%s\n", ((var_entry *) value)->name);
}


void
print_var_table(var_table_t var_table)
{
  printf("\nVAR TABLE\n");
  g_hash_table_foreach(var_table, print_var_el, NULL);
}


static void
free_macro_var_el(gpointer key, gpointer value, gpointer user_data)
{
  free(((struct macro_var_key *) key)->name);
  free(key);
  free_var_entry((var_entry *) value);
}


void 
free_macro_var_table(macro_var_table_t macro_var_table)
{
  g_hash_table_foreach(var_table, free_macro_var_el, NULL);
  g_hash_table_destroy(var_table);
}


static void 
cmp_var_table_value(gpointer key, gpointer value, gpointer user_data)
{
  
  if ( * (GHashTable **) user_data != NULL )
    {
      var_entry *second_value = (var_entry *) g_hash_table_lookup( * (GHashTable **) user_data, (char *) key );
      
      assert( second_value != NULL );
      assert( ((var_entry *) value)->type_var == second_value->type_var && ((var_entry *) value)->type_values == second_value->type_values );

      if (second_value->type_var == scalar)
	{
	  if ( !(second_value->values == NULL && ((var_entry *) value)->values == NULL) )
	    {
	      assert( !(second_value->values == NULL || ((var_entry *) value)->values == NULL) );
	      switch (second_value->type_values) 
	      {
	      case STRING_VALUE:
		if ( 0 != strcmp( (char *) ((var_entry *) value)->values, (char *) second_value->values )) 
		  * (GHashTable **) user_data = NULL;
		break;
	      case CHARACTER_VALUE:
		if ( *((char *) ((var_entry *) value)->values) != *((char *) second_value->values) )
		  * (GHashTable **) user_data = NULL;
		break;
	      case LONG_VALUE:
		if ( *((int *) ((var_entry *) value)->values) != *((int *) second_value->values) )
		  * (GHashTable **) user_data = NULL;
		break;
	      }
	    }
	}
      else
	if ( !(second_value->values == NULL && ((var_entry *) value)->values == NULL) )
	  {
	    assert( !(second_value->values == NULL || ((var_entry *) value)->values == NULL) );
	    
	    if ( g_hash_table_size( (GHashTable *) ((var_entry *) value)->values) != g_hash_table_size( (GHashTable *) second_value->values) )
	      * (GHashTable **) user_data = NULL;
	    else
	      if (cmp_hash_tables((GHashTable *) ((var_entry *) value)->values, (GHashTable *) second_value->values) == 0)
		* (GHashTable **) user_data = NULL;
	      
	  }
    }
}


int 
cmp_var_table_values(var_table_t first, var_table_t second)
{
  gpointer data = second;

  assert( g_hash_table_size(first) == g_hash_table_size(second) );
  g_hash_table_foreach(first, cmp_var_table_value, &data);
  
  return data == NULL ? 0 : 1;
}


/* le seguenti 2 funzioni servono per creare la GHashTable relativa */

guint 
macro_var_hash(gconstpointer key)
{
  return g_str_hash(((struct macro_var_key *) key)->name) * ((struct macro_var_key *) key)->macro_id;
}


gboolean 
macro_var_equal(gconstpointer a, gconstpointer b)
{
  return strcmp(((struct macro_var_key *) a)->name, ((struct macro_var_key *) b)->name) == 0 
    && ((struct macro_var_key *) a)->macro_id == ((struct macro_var_key *) b)->macro_id;  
}





/***************************************
 * TABELLA VISIBILITA' DELLE VARIABILI *
 ***************************************
 
 La tabella delle visibilita' e' una tabella hash di coppie (key, val).
 key contiene il nome della variabile e il numero della sequenza (o regola) in cui 
 la variabile e' definita. val contiene una lista (ordinata) di elementi scope_el_t.
 
 La tabella utilizzata tenendo conto della seguente convenzione:              
 - il numero di sequenza (num_seq) e' negativo per le definizioni di macro                   
 - num_seq e' uguale a zero per le variabili definite globalmente 
 - num_seq e' positivo per le sequenze normali

*/



struct var_scope_key_t {
  char *name;            /* nome variabile */
  int num_seq;           /* numero sequenza */
};

struct var_scope_val_t {
  GList *first_scope;    /* punta al primo elemento di una lista di scope_el_t */
  GList *last_scope;     /* punta all'ultimo elemento della lista di scope_el_t */
};

 
static GHashTable *var_scope_table = NULL;   


static GList **first_scope_el = (GList **) NULL;
/* 
   punta ad un'area di memoria che conserva l'indirizzo del primo elemento
   della lista contenente l'ultima struttura scope_el_t restituita da una
   qualunque delle funzioni definite di seguito
*/

static GList **last_scope_el  = (GList **) NULL;
/*
   punta ad un'area di memoria che conserva l'indirizzo dell'ultimo elemento
   della lista contenente l'ultima struttura scope_el_t restituita da una
   qualunque delle funzioni definite di seguito
*/

static GList *cur_scope_el = NULL;
/* 
   elemento corrente contenente l'ultima struttura scope_el_t restituita da 
   una qualunque delle funzioni definite di seguito 
*/

/* * */


static guint 
var_scope_key_hash(gconstpointer key)
{
  return abs(g_str_hash(((struct var_scope_key_t *) key)->name) + ((struct var_scope_key_t *) key)->num_seq);
}


static gint        
var_scope_key_equal(gconstpointer a, gconstpointer b)
{
  return ((struct var_scope_key_t *) a)->num_seq == ((struct var_scope_key_t *) b)->num_seq 
    && strcmp(((struct var_scope_key_t *) a)->name, ((struct var_scope_key_t *) b)->name) == 0; 
}


void 
init_var_scope_table()
{
  var_scope_table = g_hash_table_new(var_scope_key_hash, var_scope_key_equal);
}


scope_el_t *
putscope(char *name, int num_seq)
{
  scope_el_t *scope_element;
  struct var_scope_val_t *val;
  struct var_scope_key_t *key = (struct var_scope_key_t *) malloc(sizeof(struct var_scope_key_t));
 
 
  key->name = name;
  key->num_seq = num_seq;
  
  if ( (val = g_hash_table_lookup(var_scope_table, key)) == NULL )
    {
      val = (struct var_scope_val_t *) malloc(sizeof(struct var_scope_val_t));
      val->first_scope = NULL;
      val->last_scope = NULL;
      g_hash_table_insert(var_scope_table, key, val);
    }
  
  scope_element = (scope_el_t *) malloc(sizeof(scope_el_t));
  scope_element->num_event = -1;
  scope_element->context = intro;
  scope_element->num_statement = 0;

  val->last_scope = g_list_append(val->last_scope, scope_element);
  if ( val->first_scope == NULL)
    val->first_scope = val->last_scope;
  else
    val->last_scope = g_list_next(val->last_scope);
  
  first_scope_el = &(val->last_scope);
  last_scope_el  = &(val->first_scope);
  cur_scope_el = val->last_scope;
  
  return scope_element;
}


scope_el_t *
getscope(char *name, int num_seq, unsigned int n)
{
  struct var_scope_val_t *val;
  struct var_scope_key_t key;
  GList *ptr;

  key.name = name;
  key.num_seq = num_seq;
  
  assert(n > 0);
    
  if ( (val = g_hash_table_lookup(var_scope_table, &key)) != NULL )
    {
      ptr = g_list_nth(val->first_scope, n - 1);
      if (ptr)
	{
	  first_scope_el = &(val->first_scope);
	  last_scope_el = &(val->last_scope);
	  cur_scope_el = ptr;
	  return (scope_el_t *) ptr->data;
	}
    }
  
  first_scope_el = (GList **) NULL;
  last_scope_el = (GList **) NULL;
  cur_scope_el = NULL;
  return NULL;
}


scope_el_t *
lastscope(char *name, int num_seq)
{
  struct var_scope_val_t *val;
  struct var_scope_key_t key;
  
  key.name = name;
  key.num_seq = num_seq;
  
  if ( (val = g_hash_table_lookup(var_scope_table, &key)) != NULL )
    {
      first_scope_el = &(val->first_scope);
      last_scope_el = &(val->last_scope);
      cur_scope_el = val->last_scope;
      return (scope_el_t *) val->last_scope->data;
    }
  
  first_scope_el = (GList **) NULL;
  last_scope_el = (GList **) NULL;
  cur_scope_el = NULL;
  return NULL;
}


static gint 
scope_el_cmp(scope_el_t *ael, scope_el_t * bel)
     /*
       Restituisce un intero 
	    < 0   se ael < bel, 
	   == 0   se ael == bel, 
	    > 0   se ael > bel.
     */
{
  if (ael->num_event == bel->num_event)
    {
      if (ael->context == ael->context)
	return ael->num_statement - bel->num_statement;
      else
	return ael->context - bel->context;
    }
  
  return abs(ael->num_event) - abs(bel->num_event);
}



scope_el_t *
sort_current_scope(void)
     /*
       Sposta l'elemento contenente l'ultima struttura scope_el_t restituita da una delle
       precedenti funzioni, in maniera che la lista delle strutture scope_el_t sia mantenuta
       ordinata per numeri assoluti crescenti. Restituisce l'elemento corrente.
     */
{
  GList *ptr, *tmp;
  
  int changed = 1;
  ptr = cur_scope_el;

  if ( cur_scope_el )
    {
      while( changed )
	{
	  changed = 0;
	  tmp = g_list_previous(ptr);
	  if ( tmp != NULL && scope_el_cmp((scope_el_t *) tmp->data, (scope_el_t *) ptr->data) > 0)
	    {
	      tmp->next = ptr->next;
	      ptr->prev = tmp->prev;
	      tmp->prev = ptr;
	      ptr->next = tmp;
	      
	      if (ptr->prev == NULL)
		*first_scope_el = ptr;
	      else
		ptr->prev->next = ptr;
	      
	      if (tmp->next == NULL)
		*last_scope_el = tmp;
	      else
		tmp->next->prev = tmp;
	      
	      changed = 1;
	    }
	  else
	    {
	      tmp = g_list_next(ptr);
	      if ( tmp != NULL && scope_el_cmp((scope_el_t *) ptr->data, (scope_el_t *) tmp->data) > 0)
		{
		  ptr->next = tmp->next; 
		  tmp->prev = ptr->prev;
		  tmp->next = ptr;
		  ptr->prev = tmp;
		  
		  if (tmp->prev == NULL)
		    *first_scope_el = tmp;
		  else
		    tmp->prev->next = tmp;
		  
		  if (ptr->next == NULL)
		    *last_scope_el = ptr;
		  else
		    ptr->next->prev = ptr;
		  
		  changed = 1;
		}
	    }
	}

      return (scope_el_t *) cur_scope_el->data;
    }
  
  return NULL;
}


scope_el_t *
nextscope(void)
     /*
       Restituisce il puntatore alla struttura scope_el_t successiva all'ultima
       struttura restituita da una delle seguenti funzioni:
       putscope(), getscope(), firstscope(), lastscope(), nextscope().
     */
{
  if (cur_scope_el)
    {
      cur_scope_el = g_list_next(cur_scope_el);
      if (cur_scope_el)
	return (scope_el_t *) cur_scope_el->data;
    }

  return NULL;
}


scope_el_t *
prevscope(void)
     /*
       Restituisce il puntatore alla struttura scope_el_t precedente all'ultima
       struttura restituita da una delle seguenti funzioni:
       putscope(), getscope(), firstscope(), lastscope(), nextscope().
     */
{
  if (cur_scope_el)
    {
      cur_scope_el = g_list_previous(cur_scope_el);
      if (cur_scope_el)
	return (scope_el_t *) cur_scope_el->data;
    }
  
  return NULL;
}



/**************************
 * TABELLE DEI SEGNAPOSTI *
 **************************/


label_value_t *
put_label(label_table_t table, char *label_name, int num_event)
{
  label_key_t *key = (label_key_t *) malloc(sizeof(label_key_t));
  label_value_t *value = (label_value_t *) malloc(sizeof(label_value_t));

  key->name = strdup(label_name);
  key->num_event = num_event;
  value->key = key;
  value->type_label = 0;
  value->n_arg = -1;

  g_hash_table_insert(table, key, value); 

  return value;
}


label_value_t *
get_label(label_table_t table, char *label_name, int num_event)
{
  label_key_t key;

  key.name = label_name;
  key.num_event = num_event;

  return (label_value_t *) g_hash_table_lookup(table, &key);
}


static gboolean 
free_label_table_element(gpointer key, gpointer value, gpointer user_data)
{
  free( ((label_key_t *) key)->name );
  free(key);
  free(value);

  return TRUE;
}


void
free_label_table(label_table_t table)
{
  g_hash_table_foreach_remove(table, free_label_table_element, NULL);
  g_hash_table_destroy(table);
}


/* le seguenti 2 funzioni servono per creare la GHashTable realtiva */

guint 
label_hash(gconstpointer key)
{
  return g_str_hash(((label_key_t *) key)->name) + ((label_key_t *) key)->num_event;
}


gboolean
label_equal(gconstpointer a, gconstpointer b)
{
  return strcmp(((label_key_t *) a)->name, ((label_key_t *) b)->name) == 0 && ((label_key_t *) a)->num_event == ((label_key_t *) b)->num_event;
}



/***********************
 * TABELLA DELLE MACRO *
 ***********************/

macro_entry * 
putmacro(char *macro_name, int macro_id, macro_entry **table)
{
  macro_entry *ptr;
  
  ptr = (macro_entry *) malloc(sizeof(macro_entry));
  assert(ptr != NULL);
  ptr->name = macro_name;
  ptr->macro_id = macro_id;
  ptr->macro_val = NULL;
  ptr->next = (struct macro_entry *) (*table);
  (*table) = ptr;

  return ptr;
}


macro_entry * 
getmacro_by_name(char *macro_name, macro_entry **table)
{
  macro_entry *ptr;

  for (ptr = (*table); ptr != (macro_entry *) NULL; ptr = (macro_entry *) ptr->next )
    {
      if (strcmp (ptr->name, macro_name) == 0) 
	return ptr;
    }

  return 0;
} 


void 
print_macro_entries(macro_entry *table) 
{
  int i;
  
  printf("\nmacros in symbol table:\n");
  for ( i = 1; table != NULL; table = table->next, i++) 
    printf("macro n. %d: %s\n", i, table->name);
}


void
free_macro_table(macro_entry *macro_table)
{
  macro_entry *ptr, *next;
  
  ptr = macro_table;
  while (ptr != NULL)
    {
      next = (macro_entry *)ptr->next;
      free(ptr->name);
      free_parse_tree(ptr->macro_val);
      free(ptr);
      ptr = next;
    }
}



/**************************
 * TABELLA DELLE FUNZIONI *
 **************************/
#include "type_sys_args.h"
#include "actions.h"


static funct_entry funct_predefined[] = {
#include "protofunct.h"
  { NULL, NULL, 0, 0, {0, 0, 0, 0, 0, 0, 0, 0, 0, 0}, NULL },
};


void 
init_funct_table(funct_table_t *funct_table)
     /* crea una nuova tabella delle funzioni inizializzandola con quelle predefinite */
{
  funct_entry *ptr;
  *funct_table = g_hash_table_new(g_str_hash, g_str_equal);
  
  /* aggiungo le funzioni predefinite */
  for (ptr = funct_predefined; ptr->name != NULL; ptr++)
    g_hash_table_insert(*funct_table, ptr->name, ptr);
}



int 
funct_ret_type ( void *(*funct_ptr)() )
{
  funct_entry *temp;
  
  for (temp = funct_predefined; temp->name != NULL; temp++)
    {
      if (temp->ptr == funct_ptr) return temp->ret_type; 
    }
  
  return -1; /* non ci dovrebbe mai arrivare */
}


/************************************************************************************************
 ***************** FUNZIONI PER GLI ELEMENTI SINTATTICI DEL PARSE TREE **************************
 ************************************************************************************************/


/***********
 * HANDLER *
 ***********/


handler_t *
clone_handler(handler_t *handler)
{
  handler_t *new_handler = (handler_t *) malloc(sizeof(handler_t));
  
  if (handler->var_name != NULL) 
    new_handler->var_name = clone_string(handler->var_name);
  else 
    new_handler->var_name = NULL;
  new_handler->type_label = handler->type_label;
  new_handler->label_ref = handler->label_ref; 
  
  return new_handler;
}


void
print_handler(handler_t *a)
{
  assert(a != NULL);
  if (a->var_name == NULL)
    {
      printf("segnaposto-> tipo segnaposto %ld  ", a->type_label);
      printf("num. arg %ld\n", a->label_ref.n_arg);
    }
  else
    printf("var name %s  ", a->var_name);
}




/*************
 * CONTAINER *
 *************/


container *
clone_container(container *a)
{
  container *new = (container *) malloc (sizeof(container));
  
  assert(a != NULL);
  new->type_content = a->type_content;
  new->type_value = a->type_value;
  if (a->type_content == HANDLER) 
    new->content.handler = clone_handler(a->content.handler);
  
  
  else if (a->type_content == ARRAY) new->content.value = (GList *) clone_array(a->content.value);
  else if (a->type_content == HASH_VALUE) 
    {
      if (a->content.hash_value == NULL) new->content.hash_value = NULL;
      else new->content.hash_value = clone_hash_table(a->content.hash_value, a->type_value);
    }  
  else if (a->type_content == SIMPLE_VALUE) 
    {
      if (a->type_value == STRING_VALUE) new->content.value = (char *) clone_string(a->content.value);
      else if (a->type_value == CHARACTER_VALUE) new->content.value = (char *) clone_char(a->content.value);
      else if (a->type_value == LONG_VALUE) new->content.value = (long *) clone_long(a->content.value);
      else if (a->type_value == FLAG) new->content.value = (char *) clone_string(a->content.value);
    }
  
  return new;
}


void
print_container(container *a)
{
  assert(a != NULL);
  switch (a->type_content) {
  case HANDLER:
    printf("HANDLER->");
    print_handler(a->content.handler);
    break;
  case SIMPLE_VALUE:
    printf("SIMPLE_VALUE->"); 
    if (a->type_value == STRING_VALUE) {printf("STRING->");            printf("%s  ", (char *)a->content.value);}
    else if (a->type_value == CHARACTER_VALUE) {printf("CHARACTER->"); printf("%c  ", *((char *)a->content.value));}
    else if (a->type_value == LONG_VALUE) {printf("LONG->");           printf("%ld  ", *((long *)a->content.value));}
    else if (a->type_value == FLAG) {printf("FLAG->");           printf("%s  ", (char *)a->content.value);}
    else printf("type_value not defined  ");
    break;          
  case HASH_VALUE:
    printf("HASH_VALUE->");
    if (a->type_value == STRING_VALUE) printf("STRING->"); 
    else if (a->type_value == CHARACTER_VALUE) printf("CHARACTER->");
    else if (a->type_value == LONG_VALUE) printf("LONG->");
    else if (a->type_value == FLAG) printf("FLAG->");
    else printf("type_value not defined  ");
    if (a->content.hash_value == NULL) printf("empty hash table  ");
    else print_hash_table(a->content.hash_value, a->type_value);
    break;   
  case ARRAY:
    printf("ARRAY->");
    if (a->type_value == STRING_VALUE) printf("STRING->"); 
    else if (a->type_value == CHARACTER_VALUE) printf("CHARACTER->");
    else if (a->type_value == LONG_VALUE) printf("LONG->");
    else if (a->type_value == FLAG) printf("FLAG->");
    else printf("type_value not defined  ");
    print_array(a->content.value);
    break;
  }
}


void
free_container(container *a)
{
  assert(a != NULL);
  
  switch (a->type_content)
    {
    case HANDLER: 
      free(a->content.handler); /* le var_entry vengono deallocate dalla free_macro_var_table */
      break;
    case ARRAY: 
      free_array(a->content.value);
      break;
    case HASH_VALUE: 
      if (a->content.hash_value != NULL) free_hash_table(a->content.hash_value);
      break;
    case SIMPLE_VALUE: 
      free(a->content.value);
      break;
    }
  free(a);
}




/* funzioni per il contenuto del container */

char *
clone_string(char *a)
{
  char *new;

  assert(a != NULL);
  new = (char *) malloc (strlen(a)+1);
  memcpy((char *) new, a, strlen(a)+1);

  return new;
}


char *
clone_char(char *a)
{
  char *new = (char *) malloc (sizeof(char));
  assert(a != NULL);
  *new = (char) *a;

  return new;
}


long *
clone_long(long *a)
{
  long *new = (long *) malloc(sizeof(long));
  assert(a != NULL);
  *new = *a;
  
  return new;
}


GList * 
clone_array(GList *a)
{
  GList *new_list = NULL, *now;
  
  assert(a != NULL);
  now = g_list_first(a);
  while (now != NULL) {
    new_list = g_list_append (new_list, (container *) clone_container((container *)now->data) );
    now = g_list_next(now);
  }
  
  return new_list;
}


void
print_array(GList *a)
{
  GList *now;
  
  assert(a != NULL);
  printf("[");
  now = g_list_first(a);
  while (now != NULL) {
    print_container((container *)now->data);
    now = g_list_next(now);
  }
  
  printf("]");
}


void
free_array(GList *a)
{
  GList *now;
  
  assert(a != NULL);
  now = g_list_first(a);
  while (now != NULL) {
    free_container((container *)now->data);
    now = g_list_next(now);
  }
  
  g_list_free(a);
}



/* struttura utilizzata da clone_hash_table e insert_in_hash_table */

typedef struct {     
  GHashTable *new;
  int type_value;
} hasht_and_type;   

/* * */


static void 
insert_in_hash_table(gpointer key, gpointer value, gpointer temp)
{
  hasht_and_type *aaa = temp;
 
  if (aaa->type_value == STRING_VALUE)
    {
      char *str;
      str = clone_string( (char *) value);
      g_hash_table_insert(aaa->new, str, str);
    }
  else if (aaa->type_value == CHARACTER_VALUE)
    {
      char *ch;
      ch = clone_char( (char *) value);
      g_hash_table_insert(aaa->new, ch, ch);
    }
  else if (aaa->type_value == LONG_VALUE)
    {
      long *lo;
      lo = clone_long( (long *) value);
      g_hash_table_insert(aaa->new, lo, lo);
    }
  else printf("\nERRORE in insert_in_hash_t\n");
  
}


GHashTable *
clone_hash_table(GHashTable *a, int type_value)
{
  GHashTable *new;
  hasht_and_type   temp;
  
  assert(a != NULL);
  switch (type_value) {
  case STRING_VALUE:
    new = g_hash_table_new(g_str_hash, g_str_equal);
    break;
  case CHARACTER_VALUE: case LONG_VALUE:
    new = g_hash_table_new(g_int_hash, g_int_equal);
    break;
  } 
  
  temp.new = new;    temp.type_value = type_value;
  g_hash_table_foreach (a, insert_in_hash_table, (hasht_and_type *) &temp);
  
  return new;
}


static void 
print_row_in_hash_table(gpointer key, gpointer value, gpointer temp)
{
  int *type_value = temp;
 
  if (*type_value == STRING_VALUE)           printf("\"%s\" ", (char *) value);
  else if (*type_value == CHARACTER_VALUE)   printf("\'%c\' ", * ((char *) value));
  else if (*type_value == LONG_VALUE)        printf("%ld ", * ((long *) value));
  else printf("\ntipo non definito\n");
}


void
print_hash_table(GHashTable *a, int type_value)
{
  assert(a != NULL);
  g_hash_table_foreach (a, print_row_in_hash_table, (int *) &type_value);
}


static void 
free_hash_element (gpointer key, gpointer value, gpointer temp)
{
  free(key);
  if (key != value) free(value);
}


void
free_hash_table(GHashTable *a)
{
  assert(a != NULL);
  g_hash_table_foreach (a, free_hash_element, NULL);
  g_hash_table_destroy(a);
}


static void
cmp_hash_table_value(gpointer key, gpointer value, gpointer user_data)
{
  if ( *((GHashTable **)user_data) != NULL)
    if ( g_hash_table_lookup( *((GHashTable **)user_data), key ) == NULL ) *((GHashTable **)user_data) = NULL;
}


int 
cmp_hash_tables(GHashTable *a, GHashTable *b)
{
  gpointer data = b;

  g_hash_table_foreach(a, cmp_hash_table_value, &data);

  return data == NULL ? 0 : 1;
}

/* * */



/************
 * arg_node *
 ************/


arg_node_t *
clone_arg_node(arg_node_t *a)
{
  arg_node_t *new = (arg_node_t *) malloc (sizeof(arg_node_t));
  
  assert(a != NULL);
  new->funct = a->funct;
  if (a->funct == NULL)   new->arg = clone_container(a->arg);
  else new->arg = a->arg;
  
  return new;
}


GNode *
clone_tree_with_arg_node(GNode *a)
{
  GNode *new, *now;
  
  assert(a != NULL);
  new = g_node_new ( (arg_node_t *) clone_arg_node(a->data) );
  now = a->children;
  while (now != NULL)
    {
      g_node_append(new, clone_tree_with_arg_node(now) );
      now = now->next;
    }
  
  return new;
}


static int
free_arg_node (GNode *node, gpointer data)
{
  arg_node_t *a = node->data;
  assert(a != NULL);
  
  if (a->arg != NULL) free_container(a->arg);
  free(a);

  return 0;
}


void
free_tree_with_arg_node(GNode *a)
{
  assert(a != NULL);
  g_node_traverse (a, G_IN_ORDER, G_TRAVERSE_ALL, -1, free_arg_node, NULL);
  
  g_node_destroy(a);  
}


static int
print_arg_node(GNode *node, gpointer data)
{
  arg_node_t *a = node->data;
  assert(a != NULL);
  
  if (a->funct != NULL)   printf("FUNCT_PTR: ");
  else print_container(a->arg);
  
  return 0;
}


void
print_tree_with_arg_node(GNode *a)
{
  assert(a != NULL);
  g_node_traverse (a, G_PRE_ORDER, G_TRAVERSE_ALL, -1, print_arg_node, NULL);
}




/****************
 * simple_event *
 ****************/


simple_event_t *
clone_simple_event(simple_event_t *a)
{
  simple_event_t *new = (simple_event_t *) malloc (sizeof(simple_event_t));
  
  assert(a != NULL);
  new->n_sysc = a->n_sysc;
  new->num_event = a->num_event;
  new->context = a->context;
  new->fl = a->fl;
  new->empty_follow = a->empty_follow;
  if (a->condition != NULL) new->condition = clone_condition_or_action(a->condition);
  else new->condition = NULL;
  if (a->action != NULL)    new->action = clone_condition_or_action(a->action);
  else new->action = NULL;
  
  return new;
}


void
print_simple_event(simple_event_t *a)
{
  assert(a != NULL);
  if (a->n_sysc == END_MARKER)  printf("#  ");
  else {
    if (a->context == intro)    printf("%s  ", get_sysc_name(a->n_sysc) );
    if (a->context == exiting)  printf("%s_exit  ", get_sysc_name(a->n_sysc) );
  }
  printf("num event %d  ", a->num_event);
  
  if (a->fl.negated_condition) printf("\nNEGATED  cond  ");
  if (a->empty_follow) printf("EMPTY follow");
  if (a->condition != NULL) {printf("\ncondition:  "); print_condition_or_action(a->condition);}
  else printf("no condition  ");
  if (a->action != NULL)    {printf("\naction:  "); print_condition_or_action(a->action);}
  else printf("no action  ");
  printf("\n");
}


void
free_simple_event(simple_event_t *a)
{
  assert(a != NULL);
  if (a->condition != NULL) free_condition_or_action(a->condition);
  if (a->action != NULL) free_condition_or_action(a->action);
  free(a);
}


GList * 
clone_condition_or_action(GList *a)
{
  GList *new_list = NULL, *now;
  
  assert(a != NULL);
  now = g_list_first(a);
  while (now != NULL) {
    new_list = g_list_append (new_list, (GNode *) clone_tree_with_arg_node((GNode *) now->data) );
    now = g_list_next(now);
  }
  
  return new_list;
}


void
print_condition_or_action(GList *a)
{
  GList *now;
  
  assert(a != NULL);
  now = g_list_first(a);
  while (now != NULL) {
    print_tree_with_arg_node((GNode *) now->data);
    now = g_list_next(now);
  }
}


void
free_condition_or_action(GList *a)
{
  GList *now;
  
  if (a != NULL)
    {
      now = g_list_first(a);
      while (now != NULL) {
	free_tree_with_arg_node((GNode *)now->data);
	now = g_list_next(now);
      }
      
      g_list_free(a);
    }
}


/* * */




/**************
 * event_node *
 **************/


event_node_t *
clone_event_node(event_node_t *a)
{
  event_node_t *new = (event_node_t *) malloc (sizeof(event_node_t));
  
  assert(a != NULL);
  new->op = a->op;
  if (a->op == none)   new->simple_event = clone_simple_event(a->simple_event);
  else new->simple_event = a->simple_event;
  
  return new;
}



void
print_event_node(event_node_t *a)
{
  assert(a != NULL);
  if (a->op == none)   print_simple_event(a->simple_event);
  else 
    {
      switch (a->op) {
      case alternation:
	printf("\n|| ");
	break;
      case non_occurence:  /* ATTENZIONE */
	printf("!\n");
	break;
      case repetition:  
	printf("*\n");
	break;
      case sequencing:
	printf("\n; ");
	break;
      default :
	printf("\n %d ",a->op);
	printf("  ******ERRORE \n");
	break;
      }   
    }
}


GNode *
clone_tree_with_event_node(GNode *a)
{
  GNode *new, *now;
  
  assert(a != NULL);
  new = g_node_new ( (event_node_t *) clone_event_node(a->data) );
  now = a->children;
  while (now != NULL)
    {
      g_node_append(new, clone_tree_with_event_node(now) );
      now = now->next;
    }
  
  return new;
}


static int
print_each_event_node(GNode *node, gpointer data)
{
  event_node_t *a = node->data;
  assert(a != NULL);
  
  print_event_node(a);
  return 0;
}


void
print_tree_with_event_node(GNode *a)
{
  assert(a != NULL);
  g_node_traverse (a, G_IN_ORDER, G_TRAVERSE_ALL, -1, print_each_event_node, NULL);
}


static int
free_event_node(GNode *node, gpointer data)
{
  event_node_t *a = node->data;
  assert(a != NULL);
  
  if (a->op == none) free_simple_event(a->simple_event);
  free(a);

  return 0;
}


void
free_tree_with_event_node(GNode *a)
{
  assert(a != NULL);
  g_node_traverse(a, G_IN_ORDER, G_TRAVERSE_ALL, -1, free_event_node, NULL);

  g_node_destroy(a);
}

/* end */
