/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: resource.c,v 1.2 2003/02/17 15:31:08 frassi Exp $
 */

#include "gets.h"
#include "util.h"
#include "pars_funct.h" /* get_nargs */

#include <sys/resource.h>
#include <sys/times.h>
#include <linux/kernel.h>
#include <sys/quota.h>


static struct xlat resources[] = {
#ifdef RLIMIT_CPU
	{ RLIMIT_CPU,	"RLIMIT_CPU"	},
#endif
#ifdef RLIMIT_FSIZE
	{ RLIMIT_FSIZE,	"RLIMIT_FSIZE"	},
#endif
#ifdef RLIMIT_DATA
	{ RLIMIT_DATA,	"RLIMIT_DATA"	},
#endif
#ifdef RLIMIT_STACK
	{ RLIMIT_STACK,	"RLIMIT_STACK"	},
#endif
#ifdef RLIMIT_CORE
	{ RLIMIT_CORE,	"RLIMIT_CORE"	},
#endif
#ifdef RLIMIT_RSS
	{ RLIMIT_RSS,	"RLIMIT_RSS"	},
#endif
#ifdef RLIMIT_NPROC
	{ RLIMIT_NPROC,"RLIMIT_NPROC"	},
#endif
#ifdef RLIMIT_NOFILE
	{ RLIMIT_NOFILE,"RLIMIT_NOFILE"	},
#endif
#ifdef RLIMIT_MEMLOCK
	{ RLIMIT_MEMLOCK,	"RLIMIT_MEMLOCK"	},
#endif
#ifdef RLIMIT_VMEM
	{ RLIMIT_VMEM,	"RLIMIT_VMEM"	},
#endif
#ifdef RLIMIT_AS
	{ RLIMIT_AS,	"RLIMIT_AS"	},
#endif
	{ 0,		NULL		},
};

/* #include <sys/times.h> */
/* #include <linux/kernel.h> */


/* static struct xlat resources[] = { */
/* #ifdef RLIMIT_CPU */
/*   { RLIMIT_CPU,	"RLIMIT_CPU"	}, */
/* #endif */
/* #ifdef RLIMIT_FSIZE */
/*   { RLIMIT_FSIZE,	"RLIMIT_FSIZE"	}, */
/* #endif */
/* #ifdef RLIMIT_DATA */
/*   { RLIMIT_DATA,	"RLIMIT_DATA"	}, */
/* #endif */
/* #ifdef RLIMIT_STACK */
/*   { RLIMIT_STACK,	"RLIMIT_STACK"	}, */
/* #endif */
/* #ifdef RLIMIT_CORE */
/*   { RLIMIT_CORE,	"RLIMIT_CORE"	}, */
/* #endif */
/* #ifdef RLIMIT_RSS */
/*   { RLIMIT_RSS,	"RLIMIT_RSS"	}, */
/* #endif */
/* #ifdef RLIMIT_NOFILE */
/*   { RLIMIT_NOFILE,"RLIMIT_NOFILE"	}, */
/* #endif */
/* #ifdef RLIMIT_VMEM */
/*   { RLIMIT_VMEM,	"RLIMIT_VMEM"	}, */
/* #endif */
/* #ifdef RLIMIT_AS */
/*   { RLIMIT_AS,	"RLIMIT_AS"	}, */
/* #endif */
/*   { 0,		NULL		}, */
/* }; */


static char *
sprintrlim(lim)
long lim;
{
  static char buf[32];
  
  if (lim == RLIM_INFINITY)
    sprintf(buf, "RLIM_INFINITY");
  else if (lim > 1024 && lim%1024 == 0)
    sprintf(buf, "%ld*1024", lim/1024);
  else
    sprintf(buf, "%ld", lim);
  return buf;
}



int 
print_getrlimit(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  struct rlimit *rlim;
  
  tprintf("getrlimit(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(resources, *((int*)info->u_arg[0]), "RLIMIT_???");
	    break;
	  case 1:
	    rlim=((struct rlimit *) info->u_arg[1]);
	    if (rlim)
	      {
		tprintf("{rlim_cur=%s,", sprintrlim(rlim->rlim_cur));
		tprintf(" rlim_max=%s}", sprintrlim(rlim->rlim_max));
	      }
	    else
	      tprintf("{NULL}");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 2
  -) numero della risorsa (int)
  -) struttura contenente i limiti (struct rlimit *)
*/
int 
sys_getrlimit(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct rlimit *rlim;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr,"sys_getrlimit: syscall failed\n");
	    return -1;
	  }
	if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&rlim, sizeof(struct rlimit), "getrlimit"))
	  {
	    fprintf(stderr,"sys_getrlimit: set_copy \n");
	    return -1;
	  }
      }
    break;
  }
  return 0;
}




int 
print_setrlimit(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  struct rlimit *rlim;
  
  tprintf("setrlimit(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(resources, *((int*)info->u_arg[0]), "RLIMIT_???");
	    break;
	  case 1:
	    rlim=((struct rlimit *) info->u_arg[1]);
	    if (rlim)
	      {
		tprintf("{rlim_cur=%s,", sprintrlim(rlim->rlim_cur));
		tprintf(" rlim_max=%s}", sprintrlim(rlim->rlim_max));
	      }
	    else
	      tprintf("{NULL}");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 2
  -) numero della risorsa (int)
  -) struttura contenente i limiti (const struct rlimit *)
*/
int 
sys_setrlimit(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct rlimit *rlim;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    //if (entering(tcp))
      if ( set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *) &rlim, sizeof(struct rlimit), "setrlimit"))
	{
	  fprintf(stderr,"sys_setrlimit: set_copy\n");
	  return -1;
	}
    break;
  }
  return 0;
}



static struct xlat usagewho[] = {
	{ RUSAGE_SELF,		"RUSAGE_SELF"		},
	{ RUSAGE_CHILDREN,	"RUSAGE_CHILDREN"	},
#ifdef RUSAGE_BOTH
	{ RUSAGE_BOTH,		"RUSAGE_BOTH"		},
#endif
	{ 0,			NULL			},
};


void
realprintrusage(info, rup)
msg_building_info *info;
struct rusage *rup;
{
  if (!rup)
    tprintf("{NULL}");
  else 
    {
      tprintf("{ru_utime={%lu, %lu}, ru_stime={%lu, %lu}, ",
	      ((long) rup->ru_utime.tv_sec), ((long) rup->ru_utime.tv_usec),
	      ((long) rup->ru_stime.tv_sec), ((long) rup->ru_stime.tv_usec));
      tprintf("ru_maxrss=%lu, ru_ixrss=%lu, ", rup->ru_maxrss, rup->ru_ixrss);
      tprintf("ru_idrss=%lu, ru_isrss=%lu, ", rup->ru_idrss, rup->ru_isrss);
      tprintf("ru_minflt=%lu, ru_majflt=%lu, ru_nswap=%lu, ", rup->ru_minflt, rup->ru_majflt, rup->ru_nswap);
      tprintf("ru_inblock=%lu, ru_oublock=%lu, ", rup->ru_inblock, rup->ru_oublock);
      tprintf("ru_msgsnd=%lu, ru_msgrcv=%lu, ", rup->ru_msgsnd, rup->ru_msgrcv);
      tprintf("ru_nsignals=%lu, ru_nvcsw=%lu, ru_nivcsw=%lu}", rup->ru_nsignals, rup->ru_nvcsw, rup->ru_nivcsw);
    }
}



struct rusage *
get_rusage(tcp, addr) 
struct tcb *tcp;
long addr;
{
  struct rusage *ru;
  
  get_error = 0;
  
  if (!addr)
    return NULL;
  /*
  if ((exiting(tcp)) && (syserror(tcp)))
    {
      get_error = 1;
      return NULL;
    }
  */
  ru = ((struct rusage *) malloc(sizeof(struct rusage)));

  if (ru == NULL)
    {
      tprintf("get_rusage: unable to alloc rusage");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, ru) < 0) 
    {
      fprintf(stderr, "get_rusage: umove failed\n");
      get_error = 1;
      return NULL;
    }

  return ru;
}


int 
print_getrusage(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("getrusage(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(usagewho, *((int*)info->u_arg[0]), "RUSAGE_???");
	    break;
	  case 1:
	    realprintrusage(info, ((struct rusage *) info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}




/*
  Numero argomenti: 2
  -) processo del quale si richiede l'uso della risorsa (int)
  -) struttura contenente i dati sull'uso (struct rusage *)
*/
int 
sys_getrusage(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_getrusage: syscall failed\n");
	    return -1;
	  }
	
	((struct rusage*) tcp->u_arg[n_arg]) = get_rusage(tcp, tcp->u_arg[n_arg]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_getrusage: get_rusage\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  return 0;
}



int 
print_sysinfo(info)
msg_building_info *info;
{
  struct sysinfo *si;

  tprintf("sysinfo(");
  
  if (info->arg_state[0]==TAKEN)
    {
      si=((struct sysinfo*) info->u_arg[0]);
      if (si)
	{
	  tprintf("{uptime=%lu, loads=[%lu, %lu, %lu] ",si->uptime, si->loads[0], si->loads[1],si->loads[2]);
	  tprintf("totalram=%lu, freeram=%lu, ",si->totalram, si->freeram);
	  tprintf("sharedram=%lu, bufferram=%lu} ",si->sharedram, si->bufferram);
	  tprintf("totalswap=%lu, freeswap=%lu, procs=%hu}",si->totalswap, si->freeswap, si->procs);
	}
      else
	tprintf("{NULL}");
    }
  else
    tprintf("_)");

  return 0;
}



/*
  Numero argomenti: 1
  -) struttura contenente le info del sistema (struct sysinfo *)
*/
int 
sys_sysinfo(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct sysinfo *si;

  if (exiting(tcp)) 
    {
      if (syserror(tcp))
	{
	  fprintf(stderr, "sys_sysinfo: syscall failed\n");
	  return -1;
	}
      
      if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&si, sizeof(struct sysinfo), "sysinfo"))
	{
	  fprintf(stderr, "sys_sysinfo: set_copy\n");
	  return -1;
	}
    }
  return 0;
}



static struct xlat priorities[] = {
	{ PRIO_PROCESS,	"PRIO_PROCESS"	},
	{ PRIO_PGRP,	"PRIO_PGRP"	},
	{ PRIO_USER,	"PRIO_USER"	},
	{ 0,		NULL		},
};




int 
print_getpriority(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("getpriority(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(priorities, *((int*)info->u_arg[0]), "PRIO_???");
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 2
  -) indicatore del destinatario: process, process group, user (int)
  -) indicatore di quale destinatario esattamente (int)
*/
int 
sys_getpriority(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


int 
print_setpriority(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("setpriority(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(priorities, *((int*)info->u_arg[0]), "PRIO_???");
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long*)info->u_arg[1]));
	    break;
	  case 2:
	    tprintf("%ld", *((long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) indicatore del destinatario: process, process group, user (int)
  -) indicatore di quale destinatario esattamente (int)
  -) livello di priorita' (int)
*/
int 
sys_setpriority(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


int 
print_nice(info)
msg_building_info *info;
{
  tprintf("nice(");
  
  if (info->arg_state[0]==TAKEN)
    tprintf("%ld)", *((long*) info->u_arg[0]));
  else
    tprintf("_)");

  return 0;
}


/*
  Numero argomenti: 1
  -) incremento di priorita' (int)
*/
int 
sys_nice(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}



int 
print_times(info)
msg_building_info *info;
{
  struct tms *tbuf;

  tprintf("times(");
  
  if (info->arg_state[0]==TAKEN)
    {
      tbuf=((struct tms *) info->u_arg[0]);

      if (tbuf)
	{
	  tprintf("{tms_utime=%lu, tms_stime=%lu, ",tbuf->tms_utime, tbuf->tms_stime);
	  tprintf("tms_cutime=%lu, tms_cstime=%lu})",tbuf->tms_cutime, tbuf->tms_cstime);
	}
      else
	tprintf("{NULL}");
    }
  else
    tprintf("_)");

  return 0;
}


/*
  Numero argomenti: 1
  -) contenitore dei tempi dei processi attuali (struct tms *)
*/
int 
sys_times(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct tms *tbuf;

  if (exiting(tcp)) 
    {
      if (syserror(tcp))
	{
	  fprintf(stderr, "sys_times: syscall failed\n");
	  return -1;
	}

      if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&tbuf, sizeof(struct tms), "times"))
	{
	  fprintf(stderr, "sys_times: set_copy\n");
	  return -1;
	}
    }
  return 0;
}



static struct xlat quotacmds[] = {
	{ Q_QUOTAON,	"Q_QUOTAON"	},
	{ Q_QUOTAOFF,	"Q_QUOTAOFF"	},
	{ Q_GETQUOTA,	"Q_GETQUOTA"	},
	{ Q_SETQUOTA,	"Q_SETQUOTA"	},
	{ Q_SETUSE,	"Q_SETUSE"	},
	{ Q_SYNC,	"Q_SYNC"	},
	{ Q_SETQLIM,	"Q_SETQLIM"	},
	{ Q_GETSTATS,	"Q_GETSTATS"	},
	{ Q_RSQUASH,	"Q_RSQUASH"	},
	{ 0,		NULL		},
};

static struct xlat quotatypes[] = {
	{ USRQUOTA,	"USRQUOTA"	},
	{ GRPQUOTA,	"GRPQUOTA"	},
	{ 0,		NULL		},
};

int
sys_quotactl(tcp, n_arg) /* ATTENZIONE: manca la print_ */
struct tcb *tcp;
int n_arg;
{
  /*
	if (entering(tcp)) {
		printxval(quotacmds, tcp->u_arg[0] >> SUBCMDSHIFT, "Q_???");
		tprintf("|");
		printxval(quotatypes, tcp->u_arg[0] & SUBCMDMASK, "???QUOTA");
		tprintf(", ");
		printstr(tcp, tcp->u_arg[1], -1);
		tprintf(", %lu, ", tcp->u_arg[2]);
	}
	else {
		struct dqblk dq;

		if (!tcp->u_arg[3])
			tprintf("NULL");
		else if (!verbose(tcp))
			tprintf("%#lx", tcp->u_arg[3]);
                else if (umoven(tcp, tcp->u_arg[3], sizeof(struct dqblk),   
                    (char *) &dq) < 0)
                        tprintf("???");
		else {
                        tprintf("{");
			tprintf("%u, ", dq.dqb_bhardlimit);
			tprintf("%u, ", dq.dqb_bsoftlimit);
			tprintf("%u, ", dq.dqb_curblocks);
			tprintf("%u, ", dq.dqb_ihardlimit);
			tprintf("%u, ", dq.dqb_isoftlimit);
			tprintf("%u, ", dq.dqb_curinodes);
			tprintf("%lu, ", dq.dqb_btime);
			tprintf("%lu", dq.dqb_itime);
                        tprintf("}");
		}

	}
	return 0;
  */
  struct dqblk *dq;

  switch (n_arg) {
  case 0:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
        
    ((char *) tcp->u_arg[1]) = get_str(tcp, tcp->u_arg[1], -1);
    if ( get_error )
      {
	fprintf(stderr, "sys_quotactl: get_str\n");
	tcp->arg_state[1] = EMPTY;
	return -1;
      }
    tcp->arg_state[1] = TAKEN;
    
    break;
  case 3:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_quotactl: syscall failed\n");
	    return -1;
	  }
	
        if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&dq, sizeof(struct dqblk), "sys_quotactl"))
	{
	  fprintf(stderr, "sys_times: set_copy\n");
	  return -1;
	}
      }
    break;
  }

  return 0;
}
