/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * Copyright (c) 1999 IBM Deutschland Entwicklung GmbH, IBM Corporation
 *                     Linux for s390 port by D.J. Barrow
 *                    <barrow_dj@mail.yahoo.com,djbarrow@de.ibm.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: signal.c,v 1.2 2003/02/17 15:31:23 frassi Exp $
 */

#include "gets.h"
#include "util.h"
#include "runtime_util.h"
#include "pars_funct.h" /* get_nargs */

#include <signal.h>
#include <sys/user.h>
#include <fcntl.h>

#ifdef HAVE_SYS_REG_H
# include <sys/reg.h>
#ifndef PTRACE_PEEKUSR
# define PTRACE_PEEKUSR PTRACE_PEEKUSER
#endif
#ifndef PTRACE_POKEUSR
# define PTRACE_POKEUSR PTRACE_POKEUSER
#endif
#elif defined(HAVE_LINUX_PTRACE_H)
#undef PTRACE_SYSCALL
#include <linux/ptrace.h>
#endif


#if HAVE_ASM_REG_H
#ifdef SPARC
#  define fpq kernel_fpq
#  define fq kernel_fq
#  define fpu kernel_fpu
#endif
#include <asm/reg.h>
#ifdef SPARC
#  undef fpq
#  undef fq
#  undef fpu 
#endif

#endif /* HAVE_ASM_REG_H */
#ifdef HAVE_ASM_SIGCONTEXT_H
#ifdef SPARC
typedef struct {
	struct regs		si_regs;
	int			si_mask;
} m_siginfo_t;
#elif !defined(IA64)
#include <asm/sigcontext.h>
#endif /* SPARC */
#else /* !HAVE_ASM_SIGCONTEXT_H */

struct sigcontext_struct {
	unsigned short gs, __gsh;
	unsigned short fs, __fsh;
	unsigned short es, __esh;
	unsigned short ds, __dsh;
	unsigned long edi;
	unsigned long esi;
	unsigned long ebp;
	unsigned long esp;
	unsigned long ebx;
	unsigned long edx;
	unsigned long ecx;
	unsigned long eax;
	unsigned long trapno;
	unsigned long err;
	unsigned long eip;
	unsigned short cs, __csh;
	unsigned long eflags;
	unsigned long esp_at_signal;
	unsigned short ss, __ssh;
	unsigned long i387;
	unsigned long oldmask;
	unsigned long cr2;
};

#endif /* !HAVE_ASM_SIGCONTEXT_H */
#ifndef NSIG
#define NSIG 32
#endif
#ifdef ARM
#undef NSIG
#define NSIG 32
#endif


char *signalent0[] = {
#include "signalent.h"
};
int nsignals0 = sizeof signalent0 / sizeof signalent0[0];

char **signalent;
int nsignals;


#ifdef HAVE_SIGACTION

static struct xlat sigact_flags[] = {
#ifdef SA_RESTORER
	{ SA_RESTORER,	"SA_RESTORER"	},
#endif
#ifdef SA_STACK
	{ SA_STACK,	"SA_STACK"	},
#endif
#ifdef SA_RESTART
	{ SA_RESTART,	"SA_RESTART"	},
#endif
#ifdef SA_INTERRUPT
	{ SA_INTERRUPT,	"SA_INTERRUPT"	},
#endif
#ifdef SA_NOMASK
	{ SA_NOMASK,	"SA_NOMASK"	},
#endif
#ifdef SA_ONESHOT
	{ SA_ONESHOT,	"SA_ONESHOT"	},
#endif
#ifdef SA_SIGINFO
	{ SA_SIGINFO,	"SA_SIGINFO"	},
#endif
#ifdef SA_RESETHAND
	{ SA_RESETHAND,	"SA_RESETHAND"	},
#endif
#ifdef SA_ONSTACK
	{ SA_ONSTACK,	"SA_ONSTACK"	},
#endif
#ifdef SA_NODEFER
	{ SA_NODEFER,	"SA_NODEFER"	},
#endif
#ifdef SA_NOCLDSTOP
	{ SA_NOCLDSTOP,	"SA_NOCLDSTOP"	},
#endif
#ifdef SA_NOCLDWAIT
	{ SA_NOCLDWAIT,	"SA_NOCLDWAIT"	},
#endif
#ifdef _SA_BSDCALL
	{ _SA_BSDCALL,	"_SA_BSDCALL"	},
#endif
	{ 0,		NULL		},
};

static struct xlat sigprocmaskcmds[] = {
	{ SIG_BLOCK,	"SIG_BLOCK"	},
	{ SIG_UNBLOCK,	"SIG_UNBLOCK"	},
	{ SIG_SETMASK,	"SIG_SETMASK"	},
#ifdef SIG_SETMASK32
	{ SIG_SETMASK32,"SIG_SETMASK32"	},
#endif
	{ 0,		NULL		},
};

#endif /* HAVE_SIGACTION */

/* Anonymous realtime signals. */
/* Under glibc 2.1, SIGRTMIN et al are functions, but __SIGRTMIN is a
   constant.  This is what we want.  Otherwise, just use SIGRTMIN. */
#ifdef SIGRTMIN
#ifndef __SIGRTMIN
#define __SIGRTMIN SIGRTMIN
#define __SIGRTMAX SIGRTMAX /* likewise */
#endif
#endif




char *
signame(sig)
int sig;
{
  static char buf[30];
  if (sig < nsignals) 
    {
      return signalent[sig];
#ifdef SIGRTMIN
    } 
  else 
    if (sig >= __SIGRTMIN && sig <= __SIGRTMAX) 
      {
	sprintf(buf, "SIGRT_%ld", (long)(sig - __SIGRTMIN));
	return buf;
#endif /* SIGRTMIN */
      } 
    else 
      {
	sprintf(buf, "%d", sig);
	return buf;
      }
}


void
long_to_sigset(l, s)
long l;
sigset_t *s;
{
  sigemptyset(s);
  *(long *)s = l;
}


/* int */
/* get_sigset_len(tcp, addr, s, len) */
/* struct tcb *tcp; */
/* int addr; */
/* sigset_t *s; */
/* int len; */
/* { */
/*   if (len > sizeof(*s)) */
/*     len = sizeof(*s); */
  
/*   sigemptyset(s); */
  
/*   if (umoven(tcp, addr, len, (char *)s) < 0) */
/*     return -1; */
  
/*   return 0; */
/* } */


/* #define get_sigset(tcp, addr, s) get_sigset_len(tcp, addr, s, sizeof(long)) */


static int
copy_sigset_len(tcp, addr, s, len)
struct tcb *tcp;
long addr;
sigset_t *s;
int len;
{
	if (len > sizeof(*s))
		len = sizeof(*s);
	sigemptyset(s);
	if (umoven(tcp, addr, len, (char *)s) < 0)
		return -1;
	return 0;
}


/* Original sigset is unsigned long */
#define copy_sigset(tcp, addr, s) copy_sigset_len(tcp, addr, s, sizeof(long))


/* char * */
/* sprintsigmask(s, mask, rt) */
/* char *s; */
/* sigset_t *mask; */
/* int rt; /\* set might include realtime sigs *\/ */
/* { */
/*   int i, nsigs; */
/*   int maxsigs; */
/*   char *format; */
/*   char *outstr=malloc(256*sizeof(char)); */
  
/*   strcpy(outstr, s); */
/*   s = outstr + strlen(outstr); */
/*   nsigs = 0; */
/*   maxsigs = nsignals; */
/* #ifdef __SIGRTMAX */
/*   if (rt) */
/*     maxsigs = __SIGRTMAX; /\* instead *\/ */
/* #endif */
/*   for (i = 1; i < maxsigs; i++)  */
/*     { */
/*       if (sigismember(mask, i) == 1) */
/* 	nsigs++; */
/*     } */
/*   if (nsigs >= nsignals * 2 / 3)  */
/*     { */
/*       *s++ = '~'; */
/*       for (i = 1; i < maxsigs; i++)  */
/* 	{ */
/* 	  switch (sigismember(mask, i)) { */
/* 	  case 1: */
/* 	    sigdelset(mask, i); */
/* 	    break; */
/* 	  case 0: */
/* 	    sigaddset(mask, i); */
/* 	    break; */
/* 	  } */
/* 	} */
/*     } */
/*   format = "%s"; */
/*   *s++ = '['; */
/*   for (i = 1; i < maxsigs; i++)  */
/*     { */
/*       if (sigismember(mask, i) == 1)  */
/* 	{ */
/* 	  sprintf(s, format, signame(i) + 3); s += strlen(s); */
/* 	  format = " %s"; */
/* 	} */
/*     } */
/*   *s++ = ']'; */
/*   *s = '\0'; */
/*   return outstr; */
/* } */


static char *
sprintsigmask(s, mask, rt)
char *s;
sigset_t *mask;
int rt; /* set might include realtime sigs */
{
	int i, nsigs;
	int maxsigs;
	char *format;
	static char outstr[256];

	strcpy(outstr, s);
	s = outstr + strlen(outstr);
	nsigs = 0;
	maxsigs = nsignals;
#ifdef __SIGRTMAX
	if (rt)
		maxsigs = __SIGRTMAX; /* instead */
#endif
	for (i = 1; i < maxsigs; i++) {
		if (sigismember(mask, i) == 1)
			nsigs++;
	}
	if (nsigs >= nsignals * 2 / 3) {
		*s++ = '~';
		for (i = 1; i < maxsigs; i++) {
			switch (sigismember(mask, i)) {
			case 1:
				sigdelset(mask, i);
				break;
			case 0:
				sigaddset(mask, i);
				break;
			}
		}
	}
	format = "%s";
	*s++ = '[';
	for (i = 1; i < maxsigs; i++) {
		if (sigismember(mask, i) == 1) {
			/* real-time signals on solaris don't have
			 * signalent entries
			 */
			if (i < nsignals) {
				sprintf(s, format, signalent[i] + 3);
			}
			else {
				char tsig[32];
				sprintf(tsig, "%u", i);
				sprintf(s, format, tsig);
			}
			s += strlen(s);
			format = " %s";
		}
	}
	*s++ = ']';
	*s = '\0';
	return outstr;
}


/* static void */
/* printsigmask(mask, rt) */
/* sigset_t *mask; */
/* int rt; */
/* { */
/*   char *tmpmask=sprintsigmask("", mask, rt); */
/*   tprintf("%s", tmpmask); */
/*   free(tmpmask); */
/* } */


static void
printsigmask(mask, rt)
sigset_t *mask;
int rt;
{
	tprintf("%s", sprintsigmask("", mask, rt));
}


void
realprintsignal(nr)
int *nr;
{
  if (nr == NULL)
    tprintf("NULL");
  else
    tprintf(signame(*nr));
}



#ifdef LINUX

#ifndef ILL_ILLOPC
#define ILL_ILLOPC      1       /* illegal opcode */
#define ILL_ILLOPN      2       /* illegal operand */
#define ILL_ILLADR      3       /* illegal addressing mode */
#define ILL_ILLTRP      4       /* illegal trap */
#define ILL_PRVOPC      5       /* privileged opcode */
#define ILL_PRVREG      6       /* privileged register */
#define ILL_COPROC      7       /* coprocessor error */
#define ILL_BADSTK      8       /* internal stack error */
#define FPE_INTDIV      1       /* integer divide by zero */
#define FPE_INTOVF      2       /* integer overflow */
#define FPE_FLTDIV      3       /* floating point divide by zero */
#define FPE_FLTOVF      4       /* floating point overflow */
#define FPE_FLTUND      5       /* floating point underflow */
#define FPE_FLTRES      6       /* floating point inexact result */
#define FPE_FLTINV      7       /* floating point invalid operation */
#define FPE_FLTSUB      8       /* subscript out of range */
#define SEGV_MAPERR     1       /* address not mapped to object */
#define SEGV_ACCERR     2       /* invalid permissions for mapped object */
#define BUS_ADRALN      1       /* invalid address alignment */
#define BUS_ADRERR      2       /* non-existant physical address */
#define BUS_OBJERR      3       /* object specific hardware error */
#define TRAP_BRKPT      1       /* process breakpoint */
#define TRAP_TRACE      2       /* process trace trap */
#define CLD_EXITED      1       /* child has exited */
#define CLD_KILLED      2       /* child was killed */
#define CLD_DUMPED      3       /* child terminated abnormally */
#define CLD_TRAPPED     4       /* traced child has trapped */
#define CLD_STOPPED     5       /* child has stopped */
#define CLD_CONTINUED   6       /* stopped child has continued */
#define POLL_IN         1       /* data input available */
#define POLL_OUT        2       /* output buffers available */
#define POLL_MSG        3       /* input message available */
#define POLL_ERR        4       /* i/o error */
#define POLL_PRI        5       /* high priority input available */
#define POLL_HUP        6       /* device disconnected */
#define SI_USER         0       /* sent by kill, sigsend, raise */
#define SI_QUEUE        -1      /* sent by sigqueue */
#define SI_TIMER        -2      /* sent by timer expiration */
#define SI_MESGQ        -3      /* sent by real time mesq state change */
#define SI_ASYNCIO      -4      /* sent by AIO completion */
#endif

#if __GLIBC_MINOR__ < 1
/* Type for data associated with a signal.  */
typedef union sigval
{
	int sival_int;
	void *sival_ptr;
} sigval_t;

# define __SI_MAX_SIZE     128
# define __SI_PAD_SIZE     ((__SI_MAX_SIZE / sizeof (int)) - 3)

typedef struct siginfo
{
	int si_signo;               /* Signal number.  */
	int si_errno;               /* If non-zero, an errno value associated with
								   this signal, as defined in <errno.h>.  */
	int si_code;                /* Signal code.  */

	union
	{
		int _pad[__SI_PAD_SIZE];

		/* kill().  */
		struct
		{
			__pid_t si_pid;     /* Sending process ID.  */
			__uid_t si_uid;     /* Real user ID of sending process.  */
		} _kill;

		/* POSIX.1b timers.  */
		struct
		{
			unsigned int _timer1;
			unsigned int _timer2;
		} _timer;

		/* POSIX.1b signals.  */
		struct
		{
			__pid_t si_pid;     /* Sending process ID.  */
			__uid_t si_uid;     /* Real user ID of sending process.  */
			sigval_t si_sigval; /* Signal value.  */
		} _rt;

		/* SIGCHLD.  */
		struct
		{
			__pid_t si_pid;     /* Which child.  */
			int si_status;      /* Exit value or signal.  */
			__clock_t si_utime;
			__clock_t si_stime;
		} _sigchld;

		/* SIGILL, SIGFPE, SIGSEGV, SIGBUS.  */
		struct
		{
			void *si_addr;      /* Faulting insn/memory ref.  */
		} _sigfault;

		/* SIGPOLL.  */
		struct
		{
			int si_band;        /* Band event for SIGPOLL.  */
			int si_fd;
		} _sigpoll;
	} _sifields;
} siginfo_t;

#define si_pid		_sifields._kill.si_pid
#define si_uid		_sifields._kill.si_uid
#define si_status	_sifields._sigchld.si_status
#define si_utime	_sifields._sigchld.si_utime
#define si_stime	_sifields._sigchld.si_stime
#define si_value	_sifields._rt.si_sigval
#define si_int		_sifields._rt.si_sigval.sival_int
#define si_ptr		_sifields._rt.si_sigval.sival_ptr
#define si_addr		_sifields._sigfault.si_addr
#define si_band		_sifields._sigpoll.si_band
#define si_fd		_sifields._sigpoll.si_fd

#endif

#endif



static struct xlat siginfo_codes[] = {
#ifdef SI_NOINFO
	{ SI_NOINFO,	"SI_NOINFO"	},
#endif
#ifdef SI_USER
	{ SI_USER,	"SI_USER"	},
#endif
#ifdef SI_LWP
	{ SI_LWP,	"SI_LWP"	},
#endif
#ifdef SI_QUEUE
	{ SI_QUEUE,	"SI_QUEUE"	},
#endif
#ifdef SI_TIMER
	{ SI_TIMER,	"SI_TIMER"	},
#endif
#ifdef SI_ASYNCIO
	{ SI_ASYNCIO,	"SI_ASYNCIO"	},
#endif
#ifdef SI_MESGQ
	{ SI_MESGQ,	"SI_MESGQ"	},
#endif
	{ 0,		NULL		},
};

static struct xlat sigill_codes[] = {
	{ ILL_ILLOPC,	"ILL_ILLOPC"	},
	{ ILL_ILLOPN,	"ILL_ILLOPN"	},
	{ ILL_ILLADR,	"ILL_ILLADR"	},
	{ ILL_ILLTRP,	"ILL_ILLTRP"	},
	{ ILL_PRVOPC,	"ILL_PRVOPC"	},
	{ ILL_PRVREG,	"ILL_PRVREG"	},
	{ ILL_COPROC,	"ILL_COPROC"	},
	{ ILL_BADSTK,	"ILL_BADSTK"	},
	{ 0,		NULL		},
};

static struct xlat sigfpe_codes[] = {
	{ FPE_INTDIV,	"FPE_INTDIV"	},
	{ FPE_INTOVF,	"FPE_INTOVF"	},
	{ FPE_FLTDIV,	"FPE_FLTDIV"	},
	{ FPE_FLTOVF,	"FPE_FLTOVF"	},
	{ FPE_FLTUND,	"FPE_FLTUND"	},
	{ FPE_FLTRES,	"FPE_FLTRES"	},
	{ FPE_FLTINV,	"FPE_FLTINV"	},
	{ FPE_FLTSUB,	"FPE_FLTSUB"	},
	{ 0,		NULL		},
};

static struct xlat sigtrap_codes[] = {
	{ TRAP_BRKPT,	"TRAP_BRKPT"	},
	{ TRAP_TRACE,	"TRAP_TRACE"	},
	{ 0,		NULL		},
};

static struct xlat sigchld_codes[] = {
	{ CLD_EXITED,	"CLD_EXITED"	},
	{ CLD_KILLED,	"CLD_KILLED"	},
	{ CLD_DUMPED,	"CLD_DUMPED"	},
	{ CLD_TRAPPED,	"CLD_TRAPPED"	},
	{ CLD_STOPPED,	"CLD_STOPPED"	},
	{ CLD_CONTINUED,"CLD_CONTINUED"	},
	{ 0,		NULL		},
};

static struct xlat sigpoll_codes[] = {
	{ POLL_IN,	"POLL_IN"	},
	{ POLL_OUT,	"POLL_OUT"	},
	{ POLL_MSG,	"POLL_MSG"	},
	{ POLL_ERR,	"POLL_ERR"	},
	{ POLL_PRI,	"POLL_PRI"	},
	{ POLL_HUP,	"POLL_HUP"	},
	{ 0,		NULL		},
};

static struct xlat sigprof_codes[] = {
#ifdef PROF_SIG
	{ PROF_SIG,	"PROF_SIG"	},
#endif
	{ 0,		NULL		},
};

#ifdef SIGEMT
static struct xlat sigemt_codes[] = {
#ifdef EMT_TAGOVF
	{ EMT_TAGOVF,	"EMT_TAGOVF"	},
#endif
	{ 0,		NULL		},
};
#endif

static struct xlat sigsegv_codes[] = {
	{ SEGV_MAPERR,	"SEGV_MAPERR"	},
	{ SEGV_ACCERR,	"SEGV_ACCERR"	},
	{ 0,		NULL		},
};

static struct xlat sigbus_codes[] = {
	{ BUS_ADRALN,	"BUS_ADRALN"	},
	{ BUS_ADRERR,	"BUS_ADRERR"	},
	{ BUS_OBJERR,	"BUS_OBJERR"	},
	{ 0,		NULL		},
};







int 
print_sigsetmask(info)
msg_building_info *info;
{
  tprintf("sigsetmask(");
  
  if (info->arg_state[0]==TAKEN)
    { 
      sigset_t sigm;
      long_to_sigset(*((long*)info->u_arg[0]), &sigm);
      printsigmask(&sigm, 0);
      tprintf(")");
    }
  else
    tprintf("_)");

  if (!syserror(info)) 
    {
      sigset_t sigm;
      long_to_sigset(info->u_rval, &sigm);
      info->auxstr = sprintsigmask("old mask ", &sigm, 0);
      return RVAL_HEX | RVAL_STR;
    }  
  return 0;
}



/*
  Numero argomenti: 1
  -) maskera (int)
*/
int 
sys_sigsetmask(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


#ifdef HAVE_SIGACTION


struct old_sigaction {
  __sighandler_t __sa_handler;
  unsigned long sa_mask;
  unsigned long sa_flags;
  void (*sa_restorer)(void);
};
#define SA_HANDLER __sa_handler


#ifndef SA_HANDLER                                                           
#define SA_HANDLER sa_handler                                                
#endif



int 
print_sigaction(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  struct old_sigaction *sap;
  sigset_t sigset;
  
  tprintf("sigaction(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    realprintsignal((int*)info->u_arg[0]);
	    break;
	  case 1:
	  case 2:
	    sap=(struct old_sigaction*) info->u_arg[i];
	    if (sap)
	      {
		switch ((long) sap->SA_HANDLER) {
		case (long) SIG_ERR:
		  tprintf("{SIG_ERR}");
		  break;
		case (long) SIG_DFL:
		  tprintf("{SIG_DFL}");
		  break;
		case (long) SIG_IGN:
		  tprintf("{SIG_IGN}");
		  break;
		default:
		  tprintf("{%#lx, ", (long) sap->SA_HANDLER);
		  long_to_sigset(sap->sa_mask, &sigset);
		  printsigmask(&sigset, 0);
		  tprintf(", ");
		  if (!printflags(sigact_flags, sap->sa_flags))
		      tprintf("0");
		  tprintf("}");
		}
		tprintf(", %#lx", (unsigned long) sap->sa_restorer);
	      }
	    else
	      tprintf("NULL");
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) numero del segnale per il quale cambiare il gestore (int)
  -) struttura che dovrebbe contenere la nuova azione (const struct sigaction *)
  -) struttura che dovrebbe contenere la vecchia azione (struct sigaction *)
*/
int 
sys_sigaction(tcp, n_arg) /* internal_sigaction in internal_syscall.c */
struct tcb *tcp;
int n_arg;
{
  struct old_sigaction *sas;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    /* dovrebbe essere valido sia in entrata che in uscita */
    
    if (tcp->u_arg[1] == 0)
      (void*) tcp->u_arg[1] = NULL;
    else
      {
	if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&sas, sizeof(struct old_sigaction), "sigaction"))
	  {	
	    fprintf(stderr, "sys_sigaction: set_copy\n");
	    return -1;
	  }
      }
    tcp->arg_state[1] = TAKEN;
    break;
  case 2:
    if (exiting(tcp))
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_sigaction: syscall failed\n");
	    return -1;
	  }
	
	if (tcp->u_arg[2] == 0)
	  (void*) tcp->u_arg[2] = NULL;
	else
	  {
	    if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&sas, sizeof(struct old_sigaction), "sigaction"))
	      {
		fprintf(stderr, "sys_sigaction: set_copy\n");
		return -1;
	      }
	  }
	tcp->arg_state[2] = TAKEN;
      }
    break;
  }
  return 0;
}



int 
print_signal(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("signal(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    realprintsignal((int*)info->u_arg[0]);
	    break;
	  case 1:
	    switch (*((int*)info->u_arg[1])) {
	    case (int) SIG_ERR:
	      tprintf("{SIG_ERR}");
	      break;
	    case (int) SIG_DFL:
	      tprintf("{SIG_DFL}");
	      break;
	    case (int) SIG_IGN:
	      tprintf("{SIG_IGN}");
	      break;
	    default:
	      tprintf("%#lx", *((unsigned long*) info->u_arg[1]));
	      break;
	    }
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 2
  -) numero del segnale per il quale cambiare il gestore (int)
  -) puntatore alla nuova funzione gestore (sighandler_t)
*/
int 
sys_signal(tcp, n_arg) /* internal_signal in internal_syscall.c */
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp, n_arg);
  return 0;
}

#endif /* HAVE_SIGACTION */




int 
print_sigreturn(info)
msg_building_info *info;
{
  sigset_t sigm;
      
  tprintf("sigreturn()");

  if (info->arg_state[0] == TAKEN)
    {
      long_to_sigset(*((long*)info->u_arg[0]), &sigm);
      info->u_rval = info->u_error = 0;
      info->auxstr = sprintsigmask("mask now ", &sigm, 0);
      return RVAL_NONE | RVAL_STR;
    }

  return 0;
}



/*
  Numero argomenti: 1
  -) valore il cui uso e significato varia in base all'architettura (unsigned long)
*/
int 
sys_sigreturn(tcp, n_arg) /* CONTROLLARE */
struct tcb *tcp;
int n_arg;
{
  long temp_l; 

  if (entering(tcp) && tcp->u_arg[0])
    {
      temp_l = tcp->u_arg[1]; 
      (long*) tcp->u_arg[0] = (long*) malloc(sizeof(long)); 
      if (tcp->u_arg[0] == 0) 
	{ 
	  fprintf(stderr, "sys_sigreturn: arg_alloc - no memory\n"); 
	  return -1; 
	} 
      *((long*) tcp->u_arg[0]) = temp_l;
      tcp->arg_state[0] = TAKEN; 
    }
  
  return 0;
}




int 
print_siggetmask(info)
msg_building_info *info;
{
  sigset_t sigm;

  tprintf("siggetmask()");

  long_to_sigset(info->u_rval, &sigm);
  info->auxstr = sprintsigmask("mask ", &sigm, 0);

  return RVAL_HEX | RVAL_STR;
}


/*
  Numero argomenti: 0
*/
int 
sys_siggetmask(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  return 0;
}


int 
print_sigsuspend(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  sigset_t sigm;
  
  tprintf("sigsuspend(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%d", *((int*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%#x", *((int*)info->u_arg[1]));
	    break;
	  case 2:
	    long_to_sigset(*((long*)info->u_arg[2]), &sigm);
	    printsigmask(&sigm, 0);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: da 1 a 3
  ?) valore intero (int)  |--->     sono un
  ?) valore intero (int)  |--->   anacronismo
  -) nuovo valore della maschera dei segnali del processo (const sigset_t * == const unsigned long *)
*/
int 
sys_sigsuspend(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  case 2:
    tcp->u_arg[n_arg] = (long) get_num(tcp, (long) tcp->u_arg[n_arg]); /* CONTROLLARE, mi sembra ci vada arg_alloc */
    if ( get_error )
      {
	fprintf(stderr, "sys_sigsuspend: get_num\n");
	tcp->arg_state[n_arg] = EMPTY;
        return -1;
      }
    tcp->arg_state[n_arg] = TAKEN;
    break;
  }
  return 0;
}



#if defined LINUX && !defined SS_ONSTACK
#define SS_ONSTACK      1
#define SS_DISABLE      2
#if __GLIBC_MINOR__ == 0
typedef struct
{
	__ptr_t ss_sp;
	int ss_flags;
	size_t ss_size;
} stack_t;
#endif
#endif

static struct xlat sigaltstack_flags[] = {
	{ SS_ONSTACK,	"SS_ONSTACK"	},
	{ SS_DISABLE,	"SS_DISABLE"	},
	{ 0,		NULL		},
};


void
realprintstack_t(tcp, ss)
struct tcb *tcp;
stack_t *ss;
{
  if (ss == NULL)
    tprintf("{NULL}");
  else
    {
      tprintf("{ss_sp=%#lx, ss_flags=", (unsigned long) ss->ss_sp);
      if (!printflags(sigaltstack_flags, ss->ss_flags))
	tprintf("0");
      tprintf(", ss_size=%lu}", (unsigned long) ss->ss_size);
    }

  return;
}


int 
print_sigaltstack(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("sigaltstack(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	  realprintstack_t((stack_t *) info->u_arg[i]);
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 2
  -) puntatore ad una struttura contenente il contenuto dello stack dopo la chiamata (const stack_t *)
  -) se non e' NULL, punta ad una struttura con il contenuto dello stack prima della chiamata (stack_t *)
*/
int 
sys_sigaltstack(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  stack_t *ss;

  if ( ( (n_arg == 0) /*&& (entering(tcp) )*/ ) || ((n_arg == 1) && (exiting(tcp) && (!(syserror(tcp)))) ) )
    if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&ss, sizeof(stack_t), "sigaltstack"))
      {
	fprintf(stderr, "sys_sigaltstack: set_copy\n");
	return -1;
      }

  return 0;
}


#ifdef HAVE_SIGACTION

int 
print_sigprocmask(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("sigprocmask(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(sigprocmaskcmds, *((int*)info->u_arg[0]), "SIG_???");
	    break;
	  case 1:
	  case 2:
	    if ((sigset_t *) info->u_arg[i] == NULL)
	      tprintf("NULL");
	    else
	      printsigmask(((sigset_t*)info->u_arg[i]), 0);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 3
  -) valore che determina il comportamento della funzione (int)
  -) il nuovo valore della maschera (const sigset_t * == const unsigned long *)
  -) un campo che conterra' il valore del vecchio valore della maschera (sigset_t * == unsigned long *)
*/
int 
sys_sigprocmask(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  sigset_t *sigset;

  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
  case 2:
    if ( (void *) tcp->u_arg[n_arg] == NULL )
      {
	tcp->arg_state[n_arg] = TAKEN;
	return 0;
      }
    
    if ( ((n_arg==1) /*&& (entering(tcp))*/) || ((n_arg==2) && (exiting(tcp) && (!(syserror(tcp))))) ) 
      {
	if ((sigset = malloc(sizeof(sigset_t))) == NULL) 
	  {
	    fprintf(stderr,"sys_sigprocmask: no memory\n");
	    return -1;
	  }
	
	if (copy_sigset(tcp, tcp->u_arg[n_arg], sigset) < 0) 
	  {
	    fprintf(stderr,"sys_sigprocmask: copy_sigset failed\n");
	    free(sigset);
	    return -1;
	  }
	tcp->u_arg[n_arg] = (long) sigset;
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }

  return 0;
}

#endif /* HAVE_SIGACTION */


int 
print_kill(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("kill(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%ld", *((long*)info->u_arg[0]));
	    break;
	  case 1:
	    tprintf("%s", signame(*((int*)info->u_arg[1])));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_kill(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  arg_alloc(tcp,n_arg);
  return 0;
}


int 
print_sigpending(info)
msg_building_info *info;
{
  tprintf("sigpending(");
  
  if (info->arg_state[0]==TAKEN)    
    {
      if ((void*) info->u_arg[0] == NULL)
	tprintf("NULL");
      else
	printsigmask(*((sigset_t*) info->u_arg[0]), 0);
    }
  else
    tprintf("_");
      
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 1
  -) maschera dei segnali pendenti da analizzare (sigset_t * == unsigned long *)
*/
int 
sys_sigpending(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  sigset_t *sigset;
  
  if ( (void *) tcp->u_arg[n_arg] == NULL)
    {
      tcp->arg_state[n_arg] = TAKEN;
      return 0;
    }
  
  if (exiting(tcp))
    {
      if (syserror(tcp))
	{
	  fprintf(stderr, "sys_sigpending: syscall failed\n");
	  return -1;
	}     

      if ((sigset = malloc(sizeof(sigset_t))) == NULL) 
	{
	  fprintf(stderr,"sys_sigpending: no memory\n");
	  return -1;
	}
      
      if (copy_sigset(tcp, tcp->u_arg[n_arg], sigset) < 0) 
	{
	  fprintf(stderr,"sys_sigpending: copy_sigset failed\n");
	  free(sigset);
	  return -1;
	}
      
      tcp->u_arg[n_arg] = (long) sigset;
      tcp->arg_state[n_arg] = TAKEN;
    }
  return 0;
}



int 
print_rt_sigprocmask(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("rt_sigprocmask(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(sigprocmaskcmds, *((int*)info->u_arg[0]), "SIG_???");
	    break;
	  case 1:
	  case 2:
	    if ((sigset_t *) info->u_arg[i] == NULL)
	      tprintf("NULL");
	    else
	      printsigmask(((sigset_t*)info->u_arg[i]), 1);
	    break;
	  case 3:
	    tprintf("%lu", *((unsigned long *)info->u_arg[3]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


int
sys_rt_sigprocmask(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  sigset_t* sigset;
  int len, *ptr;

  switch (n_arg) {
  case 0:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    if (!tcp->u_arg[1])
      {	
	tcp->arg_state[1] = TAKEN;
	return 0;
      }
    
    if ((ptr = (int *) take_arg(3, tcp)) == NULL )
      {
	fprintf(stderr, "sys_rt_sigprocmask: take_arg(3, tcp)\n");
	return -1;
      }
    len = *ptr;
    
    sigset = ((sigset_t*) malloc(sizeof(sigset_t)));
    if ( !sigset )
      {
	fprintf(stderr,"sys_rt_sigprocmask: no memory\n");
	return -1;
      }
    
    if (copy_sigset_len(tcp, tcp->u_arg[1], sigset, len) < 0)
      {
	free(sigset);
	fprintf(stderr,"sys_rt_sigprocmask: copy_sigset_len failed\n");
	return -1;
      }
    
    ((sigset_t*) tcp->u_arg[1]) = sigset;
    tcp->arg_state[1] = TAKEN;
    break;
  case 2:
    if (exiting(tcp))
      {
	if(syserror(tcp))
	  {
	    fprintf(stderr,"sys_rt_sigprocmask: syscall failed\n");
	    return -1;
	  }

	if (!tcp->u_arg[2])
	  {	
	    tcp->arg_state[2] = TAKEN;
	    return 0;
	  }
    
	if ((ptr = (int *) take_arg(3, tcp)) == NULL )
	  {
	    fprintf(stderr, "sys_rt_sigprocmask: take_arg(3, tcp)\n");
	    return -1;
	  }
	len = *ptr;
    
	sigset = ((sigset_t*) malloc(sizeof(sigset_t)));
	if ( !sigset )
	  {
	    fprintf(stderr,"sys_rt_sigprocmask: no memory");
	    return -1;
	  }

	if (copy_sigset_len(tcp, tcp->u_arg[2], sigset, len) < 0)
	  {
	    free(sigset);
	    fprintf(stderr,"sys_rt_sigprocmask: copy_sigset_len failed\n");
	    return -1;
	  }
    
	((sigset_t*) tcp->u_arg[2]) = sigset;
	tcp->arg_state[2] = TAKEN;
      }
    break;
  }

  return 0;
}



/* Structure describing the action to be taken when a signal arrives.  */
struct new_sigaction
{
  union
  {
    __sighandler_t __sa_handler;
    void (*__sa_sigaction) (int, siginfo_t *, void *);
  }
  __sigaction_handler;
  unsigned long sa_flags;
  void (*sa_restorer) (void);
  unsigned long int sa_mask[2];
};


int 
print_rt_sigaction(info)
msg_building_info *info;
{
  int i;
  sigset_t sigset;
  int nargs=get_nargs(info->scno);
  struct new_sigaction *sap;
  
  tprintf("rt_sigaction(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    realprintsignal((int*)info->u_arg[0]);
	    break;
	  case 1:
	  case 2:
	    sap=(struct new_sigaction*) info->u_arg[i];
	    if (sap)
	      {
		switch ((long) sap->__sigaction_handler.__sa_handler) {
		case (long) SIG_ERR:
		  tprintf("{SIG_ERR}");
		  break;
		case (long) SIG_DFL:
		  tprintf("{SIG_DFL}");
		  break;
		case (long) SIG_IGN:
		  tprintf("{SIG_IGN}");
		  break;
		default:
		  tprintf("{%#lx, ", (long) sap->__sigaction_handler.__sa_handler);

		  sigemptyset(&sigset);

		  if (info->arg_state[3] == TAKEN)
		    {
		      if (* (unsigned long *) info->u_arg[3] <= sizeof(sigset))
			memcpy(&sigset, &(sap->sa_mask), * (unsigned long *) info->u_arg[3]);
		      else
			memcpy(&sigset, &(sap->sa_mask), sizeof(sigset));
		      
		      printsigmask(&sigset, 1);
		      tprintf(", ");
		      
		      if (!printflags(sigact_flags, sap->sa_flags))
			tprintf("0");
		      tprintf("}");
		    }
		}
	      }
	    else
	      tprintf("NULL");
	    break;
	  case 3:
	    tprintf("%lu", *((unsigned long*)info->u_arg[3]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int 
sys_rt_sigaction(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct new_sigaction *sa;
 
  switch (n_arg) {
  case 0:
  case 3:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    /* dovrebbe essere valido sia in entrata che in uscita */
    //take_arg(3,tcp); /* CONTROLLARE */
    
    if (tcp->u_arg[1])
      {
	if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *) &sa, sizeof(struct new_sigaction), "rt_sigaction"))
	  {
	    fprintf(stderr,"sys_rt_sigaction: set_copy");
	    return -1;
	  }
      }
    tcp->arg_state[1] = TAKEN;
    break;
  case 2:
    if (exiting(tcp))
      {
	//take_arg(3,tcp); /* CONTROLLARE */
	if (syserror(tcp))
	{
	  fprintf(stderr, "sys_rt_sigaction: syscall failed\n");
	  return -1;
	}
	
	if (tcp->u_arg[2])
	  {
	    if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*) &sa, sizeof(struct new_sigaction), "rt_sigaction"))
	      {
		fprintf(stderr,"sys_rt_sigaction: set_copy");
		return -1;
	      }
	  }
        tcp->arg_state[2] = TAKEN; 
      }
    break;
  }

  return 0;  
}


int 
print_rt_sigpending(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("rt_sigpending(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printsigmask(((sigset_t*) info->u_arg[0]), 1);
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 2
  -) (sigset_t * == unsigned long *)
  -) (size_t == unsigned int)
*/
int 
sys_rt_sigpending(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  sigset_t *sigset;
  int len, *ptr;

  switch (n_arg) {
  case 0:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_rt_sigpending: syscall failed\n");
	    return -1;
	  }

	if ((ptr = (int *) take_arg(1, tcp)) == NULL )
	  {
	    fprintf(stderr, "sys_rt_sigpending:  take_arg(1, tcp)\n");
	    return -1;
	  }
	len = *ptr;
	
	if ((sigset = malloc(sizeof(sigset_t))) == NULL) 
	  {
	    fprintf(stderr,"sys_rt_sigpending: no memory\n");
	    return -1;
	  }
	
	if (copy_sigset_len(tcp, tcp->u_arg[n_arg], sigset, len) < 0) 
	  {
	    fprintf(stderr,"sys_rt_sigpending: copy_sigset failed\n");
	    free(sigset);
	    return -1;
	  }
	
	tcp->u_arg[n_arg] = (long) sigset;
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}




int 
print_rt_sigsuspend(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);

  tprintf("rt_sigsuspend(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printsigmask(((sigset_t*) info->u_arg[0]), 1);
	    break;
	  case 1:
	    tprintf("%lu", *((unsigned long *)info->u_arg[1]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 2
  -) (sigset_t * == unsigned long *)
  -) (size_t == unsigned int)
*/
int 
sys_rt_sigsuspend(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  sigset_t *sigset;
  int len, *ptr;

  switch (n_arg) {
  case 0:
    //if (entering(tcp)) 
      {
	if ((ptr = (int *) take_arg(1, tcp)) == NULL )
	  {
	    fprintf(stderr, "sys_rt_sigsuspend:  take_arg(1, tcp)\n");
	    return -1;
	  }
	len = *ptr;

	if ((sigset = malloc(sizeof(sigset_t))) == NULL) 
	  {
	    fprintf(stderr,"sys_rt_sigsuspend: no memory\n");
	    return -1;
	  }
	
	if (copy_sigset_len(tcp, tcp->u_arg[n_arg], sigset, len) < 0) 
	  {
	    fprintf(stderr,"sys_rt_sigsuspend: copy_sigset_len failed\n");
	    free(sigset);
	    return -1;
	  }
	
	tcp->u_arg[n_arg] = (long) sigset;
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  }
  return 0;
}



static void
realprintsiginfo(si)
siginfo_t *si;
{
  if (si==NULL)
    {
      tprintf("NULL");
      return;
    }

  tprintf("{si_signo=");
  realprintsignal(&(si->si_signo));
  tprintf(", si_errno=%d, si_code=", si->si_errno);
  switch(si->si_signo)
    {
    case SIGILL:
      if (!printflags(sigill_codes, si->si_code))
	tprintf("%d /* ILL_??? */", si->si_code);
      tprintf(", si_addr=%lx", (unsigned long) si->si_addr);
      break;
    case SIGFPE:
      if (!printflags(sigfpe_codes, si->si_code))
	tprintf("%d /* FPE_??? */", si->si_code);
      tprintf(", si_addr=%lx", (unsigned long) si->si_addr);
      break;
    case SIGSEGV:
      if (!printflags(sigsegv_codes, si->si_code))
	tprintf("%d /* SEGV_??? */", si->si_code);
      tprintf(", si_addr=%lx", (unsigned long) si->si_addr);
      break;
    case SIGBUS:
      if (!printflags(sigbus_codes, si->si_code))
	tprintf("%d /* BUS_??? */", si->si_code);
      tprintf(", si_addr=%lx", (unsigned long) si->si_addr);
      break;
    case SIGTRAP:
      if (!printflags(sigtrap_codes, si->si_code))
	tprintf("%d /* TRAP_??? */", si->si_code);
      break;
    case SIGCHLD:
      if (!printflags(sigchld_codes, si->si_code))
	tprintf("%d /* CLD_??? */", si->si_code);
      tprintf(", si_pid=%d, si_uid=%d, si_status=%d, si_utime=%lu, si_stime=%lu",
	      si->si_pid,
	      si->si_uid,
	      si->si_status,
	      si->si_utime,
	      si->si_stime);
      break;
    case SIGPOLL:
      if (!printflags(sigpoll_codes, si->si_code))
	tprintf("%d /+ POLL_??? +/", si->si_code);
      if (si->si_code == POLL_IN || si->si_code == POLL_OUT || si->si_code == POLL_MSG)
	tprintf(", si_bind=%lu, si_fd=%d", (unsigned long) si->si_band, si->si_fd);
      break;
    default:
      if (!printflags(siginfo_codes, si->si_code))
	tprintf("%d /+ SI_???  /", si->si_code);
      tprintf(", si_pid=%lu, si_uid=%lu, si_value={", (unsigned long) si->si_pid, (unsigned long) si->si_uid);
      tprintf("sival_int=%u, sival_ptr=%#lx", si->_sifields._rt.si_sigval.sival_int, (unsigned long) si->_sifields._rt.si_sigval.sival_ptr);
      tprintf("}");
      break;
    }
  tprintf("}");
}



int 
print_rt_sigqueueinfo(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("rt_sigqueueinfo(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    tprintf("%lu", *((unsigned long*)info->u_arg[0]));
	    break;
	  case 1:
	    realprintsignal(((int*)info->u_arg[1]));
	    break;
	  case 2:
	    realprintsiginfo((siginfo_t*)info->u_arg[2]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 3
  -) (int)
  -) (int)
  -) (siginfo_t *)
*/
int 
sys_rt_sigqueueinfo(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  siginfo_t *si;

  switch (n_arg) {
  case 0:
  case 1:
    arg_alloc(tcp,n_arg);
    break;
  case 2:
    //if (entering(tcp))
    if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&si, sizeof(siginfo_t), "rt_sigqueueinfo"))
      {
	fprintf(stderr, "sys_rt_sigqueueinfo: set_copy\n");
	return -1;
      }
    break;
  }
  
  return 0;
}




int 
print_rt_sigtimedwait(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("rt_sigtimedwait(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printsigmask(((sigset_t*) info->u_arg[0]), 1);
	    break;
	  case 1:
	    realprintsiginfo((siginfo_t*)info->u_arg[1]);
	    break;
	  case 2:
	    tprintf("%#lx", *((unsigned long*)info->u_arg[2]));
	    break;
	  case 3:
	    tprintf("%d", *((int*) info->u_arg[3]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}


/*
  Numero argomenti: 4
  -) (const sigset_t * == const unsigned long *)
  -) (siginfo_t *)
  -) (const struct timespec *)
  -) (size_t == unsigned int);
*/
int 
sys_rt_sigtimedwait(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  int len, *ptr;
  sigset_t *sigset;
  siginfo_t *si;
  struct timespec *ts;

  switch (n_arg) 
    {
    case 0:
      //if (entering(tcp)) 
      {
	  if ((ptr = (int *) take_arg(3, tcp)) == NULL)
	    {
	      fprintf(stderr, "sys_rt_sigtimedwait: take_arg(3, tcp)\n");
	      return -1;
	    }
	  len = *ptr;
	  
	  if ((sigset = malloc(sizeof(sigset_t))) == NULL) 
	    {
	      fprintf(stderr,"sys_rt_sigtimedwait: no memory\n");
	      return -1;
	    }
	  
	  if (copy_sigset_len(tcp, tcp->u_arg[n_arg], sigset, len) < 0) 
	    {
	      fprintf(stderr,"sys_rt_sigtimedwait: copy_sigset_len failed\n");
	      free(sigset);
	      return -1;
	    }
	  
	  tcp->u_arg[n_arg] = (long) sigset;
	  tcp->arg_state[n_arg] = TAKEN;
	}
      break;
    case 1:
      if (exiting(tcp)) 
	{
	  if (syserror(tcp))
	    {
	      fprintf(stderr, "sys_rt_sigtimedwait: syscall failed\n");
	      return -1;
	    }
	 
	  if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&si, sizeof(siginfo_t), "rt_sigtimedwait"))
	    {
	      fprintf(stderr, "sys_rt_sigtimedwait: set_copy\n");
	      return -1;
	    }
	}
      break;
    case 2:
      if (exiting(tcp)) 
	{
	  if (syserror(tcp))
	    {
	      fprintf(stderr, "sys_rt_sigtimedwait: syscall failed\n");
	      return -1;
	    }
	  
	  if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long *)&ts, sizeof(struct timespec), "rt_sigtimedwait"))
	    {
	      fprintf(stderr, "sys_rt_sigtimedwait: set_copy\n");
	      return -1;
	    }
	}
      break;
    case 3:
      arg_alloc(tcp,n_arg);
      break;
    }
  
  return 0;
}
