/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: stream.c,v 1.2 2003/02/17 15:31:38 frassi Exp $
 */

#include "util.h"
#include "runtime_util.h"
#include "pars_funct.h" /* get_nargs */

/* #include <sys/poll.h> */

/* #define RS_HIPRI 1 */
/* struct strbuf { */
/*   int     maxlen;                 /\* no. of bytes in buffer *\/ */
/*   int     len;                    /\* no. of bytes returned *\/ */
/*   char    *buf;                   /\* pointer to data *\/ */
/* }; */
/* #define MORECTL 1 */
/* #define MOREDATA 2 */


/* #ifdef HAVE_SYS_TIUSER_H */
/* #include <sys/tiuser.h> */
/* #include <sys/sockmod.h> */
/* #include <sys/timod.h> */
/* #endif /\* HAVE_SYS_TIUSER_H *\/ */




/* static struct xlat pollflags[] = { */
/*   { POLLIN,	"POLLIN"	}, */
/*   { POLLPRI,	"POLLPRI"	}, */
/*   { POLLOUT,	"POLLOUT"	}, */
/* #ifdef POLLRDNORM */
/*   { POLLRDNORM,	"POLLRDNORM"	}, */
/* #endif */
/* #ifdef POLLWRNORM */
/*   { POLLWRNORM,	"POLLWRNORM"	}, */
/* #endif */
/* #ifdef POLLRDBAND */
/*   { POLLRDBAND,	"POLLRDBAND"	}, */
/* #endif */
/* #ifdef POLLWRBAND */
/*   { POLLWRBAND,	"POLLWRBAND"	}, */
/* #endif */
/*   { POLLERR,	"POLLERR"	}, */
/*   { POLLHUP,	"POLLHUP"	}, */
/*   { POLLNVAL,	"POLLNVAL"	}, */
/*   { 0,		NULL		}, */
/* }; */


#ifdef HAVE_POLL_H
#include <poll.h>
#endif
#ifdef HAVE_SYS_POLL_H
#include <sys/poll.h>
#endif
#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif
#ifdef HAVE_SYS_CONF_H
#include <sys/conf.h>
#endif
#ifdef HAVE_SYS_STREAM_H
#include <sys/stream.h>
#endif
#ifdef HAVE_SYS_TIHDR_H
#include <sys/tihdr.h>
#endif



#ifndef HAVE_STROPTS_H
#define RS_HIPRI 1
struct strbuf {
        int     maxlen;                 /* no. of bytes in buffer */
        int     len;                    /* no. of bytes returned */
        char    *buf;                   /* pointer to data */
};
#define MORECTL 1
#define MOREDATA 2
#endif /* !HAVE_STROPTS_H */

#ifdef HAVE_SYS_TIUSER_H
#include <sys/tiuser.h>
#include <sys/sockmod.h>
#include <sys/timod.h>
#endif /* HAVE_SYS_TIUSER_H */







static struct xlat pollflags[] = {
#ifdef POLLIN
	{ POLLIN,	"POLLIN"	},
	{ POLLPRI,	"POLLPRI"	},
	{ POLLOUT,	"POLLOUT"	},
#ifdef POLLRDNORM
	{ POLLRDNORM,	"POLLRDNORM"	},
#endif
#ifdef POLLWRNORM
	{ POLLWRNORM,	"POLLWRNORM"	},
#endif
#ifdef POLLRDBAND
	{ POLLRDBAND,	"POLLRDBAND"	},
#endif
#ifdef POLLWRBAND
	{ POLLWRBAND,	"POLLWRBAND"	},
#endif
	{ POLLERR,	"POLLERR"	},
	{ POLLHUP,	"POLLHUP"	},
	{ POLLNVAL,	"POLLNVAL"	},
#endif
	{ 0,		NULL		},
};


int 
print_poll(info)
msg_building_info *info;
{
  int i,j,nfds;
  int nargs=get_nargs(info->scno);
  struct pollfd *pollp;
  
  tprintf("poll(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    pollp = (struct pollfd *) info->u_arg[0];
	    nfds=*((int*) info->u_arg[1]);

	    if (pollp)
	      {
		tprintf("[");
		for (j = 0; j < nfds; j++) 
		  {
		    if (j)
		      tprintf(", ");
		    if (pollp[j].fd < 0) 
		      {
			tprintf("{fd=%d}", pollp[j].fd);
			continue;
		      }
		    tprintf("{fd=%d, events=", pollp[j].fd);
		    if (!printflags(pollflags, pollp[j].events))
		      tprintf("0");
		    if (!syserror(info) && pollp[j].revents) 
		      {
			tprintf(", revents=");
			if (!printflags(pollflags, pollp[j].revents))
			  tprintf("0");
		      }
		    tprintf("}");
		  }
		tprintf("]");
	      }
	    break;
	  case 1:
	    tprintf("%d",*((int*)info->u_arg[1])); 
	    break;
	  case 2:
#ifdef INFTIM
	    if (*((long*)info->u_arg[2]) == INFTIM)
	      tprintf("INFTIM");
	    else
#endif
	      tprintf("%ld", *((long*)info->u_arg[2]));
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    
    }
  
  tprintf(") ");

  return 0;
}



int 
sys_poll(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  struct pollfd *pollp;
  int nfds, *ptr;

  switch(n_arg) {
  case 0:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_poll: syscall failed\n");
	    return -1;
	  }
	
	if ((ptr = (int *) take_arg(1, tcp)) == NULL)
	  {
	    fprintf(stderr, "sys_poll: take_arg(1, tcp)\n");
	    return -1;
	  }
	nfds = *ptr;
	
	if (nfds <= 0)  /* nfds = numero file descriptor, non puo' essere negativo, perche' non ha senso */
	  break;
	
	if (set_copy(tcp, n_arg, tcp->u_arg[n_arg], (long*)&pollp, nfds * sizeof(*pollp), "poll"))
	  {
	    fprintf(stderr, "sys_poll: set_copy\n");
	    return -1;
	  }
      }
    break;
  case 1:
  case 2:
    arg_alloc(tcp,n_arg);
    break;
  }
  
  return 0;
}
