/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * Copyright (c) 1999 IBM Deutschland Entwicklung GmbH, IBM Corporation
 *                     Linux for s390 port by D.J. Barrow
 *                    <barrow_dj@mail.yahoo.com,djbarrow@de.ibm.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: syscall.c,v 1.2 2003/02/17 15:29:06 frassi Exp $
 */


#include <signal.h>
#include <time.h>
#include <errno.h>
#include <sys/user.h>
#include <sys/syscall.h>
#include <sys/param.h>

#include "util.h"
/* #include "socket_utility.h" */
/* #include "threads_utility.h" // serve per il tipo "final_state_table_t" */
/* #include "queue.h" */
#include "internal_syscall.h"
#include "pars_funct.h"
#include "runtime_util.h"

/* GHashTable *htsyscall = NULL; */
/* FILE *filesyscall = NULL; */
/* void prova(void); */

#ifdef HAVE_SYS_REG_H
#include <sys/reg.h>
#ifndef PTRACE_PEEKUSR
# define PTRACE_PEEKUSR PTRACE_PEEKUSER
#endif
#elif defined(HAVE_LINUX_PTRACE_H)
#undef PTRACE_SYSCALL
#include <linux/ptrace.h>
#endif


#ifndef SYS_ERRLIST_DECLARED
extern int sys_nerr;
//extern char *sys_errlist[];
#endif /* SYS_ERRLIST_DECLARED */


#ifndef ERESTARTSYS
#define ERESTARTSYS	512
#endif
#ifndef ERESTARTNOINTR
#define ERESTARTNOINTR	513
#endif
#ifndef ERESTARTNOHAND
#define ERESTARTNOHAND	514	/* restart if no handler.. */
#endif
#ifndef ENOIOCTLCMD
#define ENOIOCTLCMD	515	/* No ioctl command */
#endif
#ifndef NSIG
#define NSIG 32
#endif


#include "syscall.h"

struct sysent sysent0[] = {
#include "syscallent.h"
};
int nsyscalls0 = sizeof sysent0 / sizeof sysent0[0];


struct sysent *sysent;
int nsyscalls;

char *errnoent0[] = {
#include "errnoent.h"
};
int nerrnos0 = sizeof errnoent0 / sizeof errnoent0[0];


char **errnoent;
int nerrnos;


int
set_personality (personality)
int personality;
{
  switch (personality)
    {
    case 0:
      errnoent = errnoent0;
      nerrnos = nerrnos0;
      sysent = sysent0;
      nsyscalls = nsyscalls0;
      ioctlent = ioctlent0;
      nioctlents = nioctlents0;
      signalent = signalent0;
      nsignals = nsignals0;
      break;

    default:
      return -1;
    }

  return 0;
}




enum subcall_style { shift_style, deref_style };


static void
decode_subcall (tcp, subcall, nsubcalls, style)
struct tcb *tcp;
int subcall;
int nsubcalls;
enum subcall_style style;
{
  int i, addr, arg;

  if (tcp->u_arg[0] < 0 || tcp->u_arg[0] >= nsubcalls)
    return;

  switch (style)
    {
    case shift_style:
      tcp->scno = subcall + tcp->u_arg[0];
      if (sysent[tcp->scno].nargs != -1)
	tcp->u_nargs = sysent[tcp->scno].nargs;
      else /* non dovrebbe mai essere eseguito */
	{
	  tcp->arg_state[tcp->u_nargs - 1] = EMPTY;
	  tcp->u_nargs--;
	}
      for (i = 0; i < tcp->u_nargs; i++)
	tcp->u_arg[i] = tcp->u_arg[i + 1];
      break;
    case deref_style:
      tcp->scno = subcall + tcp->u_arg[0];
      addr = tcp->u_arg[1];
      for (i = 0; i < sysent[tcp->scno].nargs; i++)
	{
	  if (umove (tcp, addr, &arg) < 0)
	    arg = 0;
	  tcp->u_arg[i] = arg;
	  tcp->arg_state[i] = UPEEK;
	  addr += sizeof (arg);
	}
      tcp->u_nargs = sysent[tcp->scno].nargs;
      break;
    }
}


struct tcb *tcp_last = NULL;

/**
 * La variabile 'canale_log' viene inizializzata nel main.
 * Essa contiene 'num_log' descrittori di file nei quali eseguire il
 * log di eventi o errori che si verificano.
 */
/* extern int num_log, *canale_log; */

/* extern final_state_table_t final_event; */
/* extern int debug_level; */

extern automa atm;


/* POTREBBE ESSERE UTILE.....
int
force_result(tcp, error, rval)
	struct tcb *tcp;
	int error;
	long rval;
{

	eax = error ? -error : rval;
	if (ptrace(PTRACE_POKEUSER, tcp->pid, (char*)(EAX * 4), eax) < 0)
		return -1;

		//All branches reach here on success (only).
	tcp->u_error = error;
	tcp->u_rval = rval;
	return 0;
}
*/

int
//trace_syscall (struct tcb *tcp, msg_queue * mqueue)
trace_syscall (struct tcb *tcp)
{
  //int i;
  int sys_res = 1;
  long scno = 0;
  long eax;
  //int fine_attacco = 0;
  int pid = tcp->pid;
  
  if (upeek (pid, 4 * ORIG_EAX, &scno) < 0)
    {
      fprintf(stderr, "trace_syscall: upeek (pid, 4 * ORIG_EAX, &scno) < 0\n");
      return -1;
    }
  if (upeek (pid, 4 * EAX, &eax) < 0)
    {
      fprintf(stderr, "trace_syscall: upeek (pid, 4 * EAX, &eax) < 0\n");
      return -1;
    }
  if (eax != -ENOSYS && !(tcp->flags & TCB_INSYSCALL)) /* se il proc e' fermo all'entrata di una syscall e eax != -ENOSYS */
    {
      //fprintf(stderr, "trace_syscall: stray syscall exit: eax = %ld\n", eax);
      return 0;
    }

  if (tcp->flags & TCB_INSYSCALL)
    {
      /***********************/
      /* Exiting system call */
      /***********************/
      long u_error;

      if (eax < 0 && -eax < nerrnos)
	{
	  tcp->u_rval = -1;
	  u_error = -eax;
	}
      else
	{
	  tcp->u_rval = eax;
	  u_error = 0;
	  tcp->arg_state[RVAL] = TAKEN;
	}

      tcp->u_error = u_error;

      /* per alcune chiamate di sistema eseguo delle istruzioni particolari
	 (fork, clone, exec, wait, exit, oldselect, sigaction, signal, sigsetmask, sigreturn) */
      if (sysent[tcp->scno].internal_func != internal_none)
	(*sysent[tcp->scno].internal_func) (tcp);
      /*******************************************/
      /* AGGIORNAMENTO STATO AUTOMA DEL PROCESSO */
      /*******************************************/
      tcp->process_state = update_state (atm, tcp, tcp->process_state); // e' all'interno di questa funzione che verranno valutate condizioni ed eseguite azioni

      //controlla il livello di debug.
      /*
      if (debug_level == 1)
	{
	  GList *list = (GList *) g_list_first (tcp->process_state);

	  while (list)
	    {
	      int node_number = ((automa_state_el *) list->data)->node_number;
	      automa_node *temp_atm_node_ptr =
		&g_array_index (atm, automa_node, node_number);
	      fine_attacco = temp_atm_node_ptr->is_final;
	      if (fine_attacco)
		break;
	      list = g_list_next (list);
	    }
	  //altrimenti non mi importa...
	}

      if (tcp->ev_set)
	{
	  if (((debug_level == 1) && (fine_attacco)) || (debug_level == 2))
	    {
	      //Eventualmente:
	      //1) prepara un mbi
	      //2) lo inserisce in coda
	      msg_building_info *mbi = make_msg (tcp);

	      if (mbi == NULL)
		{
		  log_event
		    ("trace_syscall: impossibile creare un messaggio\n",
		     canale_log, num_log);
		  return -1;
		}

	      pthread_mutex_lock (mqueue->mut);

	      if (mqueue->full)
		free_msg_element (&mbi);
	      else
		{
		  push_msg_queue (mqueue, mbi);
		  pthread_cond_signal (mqueue->notEmpty);
		}
	      pthread_mutex_unlock (mqueue->mut);
	    }
	}
      */

      /* resetto lo stato del valore di ritorno della syscall */
      tcp->arg_state[RVAL] = EMPTY;

      /****************************************************/ // ATTENZIONE: non e' una perdita di tempo controllare tutti i par. di ogni syscall...
      /* dealloco gli argomenti della chiamata di sistema */
      /****************************************************/
      /*
      if ( ((debug_level == 1) && (!fine_attacco)) || (debug_level == 0) || (tcp->ev_set == NULL) )
	(*sysent[tcp->scno].free_sys) (tcp->u_arg, tcp->arg_state, tcp->u_nargs);
      else
	for (i = 0; i < tcp->u_nargs; i++)
	  tcp->arg_state[i] = EMPTY;
      */
      (*sysent[tcp->scno].free_sys) (tcp->u_arg, tcp->arg_state, tcp->u_nargs);
      
      tcp->flags &= ~TCB_INSYSCALL;
      
      return 0;
    }

  /************************/
  /* Entering system call */
  /************************/
  tcp->scno = scno;
      
  {
    int i;
    /*
    if (tcp->scno >= 0 && tcp->scno < nsyscalls && sysent[tcp->scno].nargs != -1) CONTROLLO SUPERFICIALE ?
      tcp->u_nargs = sysent[tcp->scno].nargs;
    else
      tcp->u_nargs = MAX_ARGS;
    */
    tcp->u_nargs = sysent[tcp->scno].nargs;


    switch (tcp->scno)
      {
      case SYS_socketcall:
	for (i = 0; i < tcp->u_nargs; i++)
	  {
	    if (upeek (pid, i * 4, &tcp->u_arg[i]) < 0)
	      {
		fprintf(stderr, "trace_syscall: case SYS_socketcall - upeek\n");
		return -1;
	      }
	    tcp->arg_state[i] = UPEEK;
	  }
	decode_subcall (tcp, SYS_socket_subcall, SYS_socket_nsubcalls, deref_style);
	break;

      case SYS_ipc:
	for (i = 0; i < tcp->u_nargs; i++)
	  {
	    if (upeek (pid, i * 4, &tcp->u_arg[i]) < 0)
	      {
		fprintf(stderr, "trace_syscall: case SYS_ipc - upeek\n");
		return -1;
	      }
	    tcp->arg_state[i] = UPEEK;
	  }
	decode_subcall (tcp, SYS_ipc_subcall, SYS_ipc_nsubcalls, shift_style);
	break;
      }
  }

  ////////////////////////////////////////////////// AGGIUNTO
/*   { */
/*     char *tmp=NULL; */
    
/*     if (htsyscall == NULL)  */
/*       { */
/* 	htsyscall = g_hash_table_new (g_str_hash, g_str_equal); */
/* 	atexit(prova); */
/*       } */
/*     tmp = get_sysc_name(tcp->scno); */
    
/*     if ( g_hash_table_lookup(htsyscall, tmp ) == NULL) */
/*       { */
/* 	g_hash_table_insert (htsyscall, tmp, tmp); */
	
/* 	if (filesyscall == NULL) */
/* 	  filesyscall = fopen ("syscall.txt", "w"); */
	
/* 	fprintf(filesyscall, "%s() ||\n", tmp); */
/*       } */
/*   } */
  /////////////////////////////////////////////////

  /* per alcune chiamate di sistema eseguo delle istruzioni particolari
     (fork, clone, exec, wait, exit, oldselect, sigaction, signal, sigsetmask, sigreturn) */
  if (sysent[tcp->scno].internal_func != internal_none)
    (*sysent[tcp->scno].internal_func) (tcp);
  /*******************************************/
  /* AGGIORNAMENTO STATO AUTOMA DEL PROCESSO */
  /*******************************************/
  tcp->process_state = update_state (atm, tcp, tcp->process_state); // e' all'interno di questa funzione che verranno valutate condizioni ed eseguite azioni

  tcp_last = tcp;
  
  tcp->flags |= TCB_INSYSCALL;
  
  return sys_res;
}


int
printargs (info)
msg_building_info *info;
{

  int i;
  int nargs = get_nargs (info->scno);

  tprintf ("%s(", get_sysc_name (info->scno));

  for (i = 0; i < nargs; i++)
    {
      if (info->arg_state[i] == TAKEN)
	tprintf ("%#lx", *((long *) info->u_arg[i]));
      else
	tprintf ("_");

      if (i != nargs - 1)
	tprintf (", ");

    }

  tprintf (") ");

  return 0;
}


int
move_arg (tcp, n_arg)
struct tcb *tcp;
int n_arg;
{
  if (n_arg < tcp->u_nargs)
    arg_alloc (tcp, n_arg);
  
  return 0;
}


void
free_syscall (u_arg, arg_state, u_nargs)
long u_arg[];
arg_state_t arg_state[];
int u_nargs;
{
  while (--u_nargs >= 0)
    {
      if (arg_state[u_nargs] == TAKEN)
	free ((int *) u_arg[u_nargs]);
      
      arg_state[u_nargs] = EMPTY; /* lo devo fare sempre */
    }
}


int
print_retval (format_ret, u_rval, u_error, auxstr)
int format_ret;
long u_rval;
long u_error;
char *auxstr;
{

  if (!(format_ret & RVAL_NONE) && u_error)
    {
      switch (u_error)
	{
	case ERESTARTSYS:
	  tprintf ("= ? ERESTARTSYS (To be restarted)");
	  break;
	case ERESTARTNOINTR:
	  tprintf ("= ? ERESTARTNOINTR (To be restarted)");
	  break;
	case ERESTARTNOHAND:
	  tprintf ("= ? ERESTARTNOHAND (To be restarted)");
	  break;
	default:

	  tprintf ("= -1 ");
	  if (u_error < nerrnos && u_error < sys_nerr)
	    tprintf ("%s (%s)", errnoent[u_error], sys_errlist[u_error]);
	  else if (u_error < nerrnos)
	    tprintf ("%s (errno %ld)", errnoent[u_error], u_error);
	  else if (u_error < sys_nerr)
	    tprintf ("ERRNO_%ld (%s)", u_error, sys_errlist[u_error]);
	  else
	    tprintf ("E??? (errno %ld)", u_error);
	  break;
	}
    }
  else
    {
      if (format_ret & RVAL_NONE)
	tprintf ("= ?");
      else
	{
	  switch (format_ret & RVAL_MASK)
	    {
	    case RVAL_HEX:
	      tprintf ("= %#lx", u_rval);
	      break;
	    case RVAL_OCTAL:
	      tprintf ("= %#lo", u_rval);
	      break;
	    case RVAL_UDECIMAL:
	      tprintf ("= %lu", u_rval);
	      break;
	    case RVAL_DECIMAL:
	      tprintf ("= %ld", u_rval);
	      break;
	    default:
	      fprintf (stderr, "invalid rval format\n");
	      break;
	    }
	}

      if ((format_ret & RVAL_STR) && auxstr)
	tprintf (" (%s)", auxstr);
    }

  return 0;
}


/* void prova(void) */
/* { */
  
/*   //g_hash_table_foreach (GHashTable *hash_table, GHFunc func, gpointer user_data); */
/*   int i; */
/*   FILE *file; */
  
/*   file = fopen ("syscall_neg.txt", "w"); */
    
/*   for (i = 0; i < nsyscalls0; i++) { */
    
/*     if ( g_hash_table_lookup(htsyscall, sysent[i].sys_name) == NULL) */
/*       { */
/* 	fprintf(file, "%s() <exit:  printf(\"%s not permit\\n\") > .\n", sysent[i].sys_name, sysent[i].sys_name); */

/* 	//open() <exit:	  printf("open not permit\n") >. */
/*       } */
    
    
/*   } */

/*   fclose(file); */
/* } */
