/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: term.c,v 1.1.1.1 2001/12/17 08:58:58 frassi Exp $
 */

#include "xlat.h"
#include "util.h"

#ifdef HAVE_TERMIO_H /* definita in config.h */
#include <termio.h>
#endif /* HAVE_TERMIO_H */

#include <termios.h>

#ifdef HAVE_SYS_FILIO_H /*undefined in config.h,penso sia parte superflua*/
#include <sys/filio.h> /*file non esistente*/
#endif

static struct xlat tcxonc_options[] = 
{
  { TCOOFF,	"TCOOFF"	},
  { TCOON,	"TCOON"		},
  { TCIOFF,	"TCIOFF"	},
  { TCION,	"TCION"		},
  { 0,		NULL		},
};

static struct xlat tcflsh_options[] = 
{
  { TCIFLUSH,	"TCIFLUSH"	},
  { TCOFLUSH,	"TCOFLUSH"	},
  { TCIOFLUSH,	"TCIOFLUSH"	},
  { 0,		NULL		},
};

/*
static struct xlat baud_options[] = 
{
  { B0,		"B0"		},
  { B50,	"B50"		},
  { B75,	"B75"		},
  { B110,	"B110"		},
  { B134,	"B134"		},
  { B150,	"B150"		},
  { B200,	"B200"		},
  { B300,	"B300"		},
  { B600,	"B600"		},
  { B1200,	"B1200"		},
  { B1800,	"B1800"		},
  { B2400,	"B2400"		},
  { B4800,	"B4800"		},
  { B9600,	"B9600"		},
#ifdef B19200 presente in /usr/include/bits/termios.h
  { B19200,	"B19200"	},
#endif
#ifdef B38400 presente in /usr/include/bits/termios.h
  { B38400,	"B38400"	},
#endif
#ifdef EXTA presente in /usr/include/bits/termios.h ma non sono sicuro se e' definita o meno, vedi anche /usr/include/features.h
  { EXTA,       "EXTA"		},
#endif
#ifdef EXTB presente in /usr/include/bits/termios.h ma non sono sicuro se e' definita o meno, vedi anche /usr/include/features.h 
  { EXTB,       "EXTB"		},
#endif
  { 0,		NULL		},
};
*/

/*i modem_flags sono tutti definiti in /usr/include/bits/ioctl-types.h*/
static struct xlat modem_flags[] = 
{
#ifdef TIOCM_LE
  { TIOCM_LE,	"TIOCM_LE",	},
#endif
#ifdef TIOCM_DTR
  { TIOCM_DTR,	"TIOCM_DTR",	},
#endif
#ifdef TIOCM_RTS
  { TIOCM_RTS,	"TIOCM_RTS",	},
#endif
#ifdef TIOCM_ST
  { TIOCM_ST,	"TIOCM_ST",	},
#endif
#ifdef TIOCM_SR
  { TIOCM_SR,	"TIOCM_SR",	},
#endif
#ifdef TIOCM_CTS
  { TIOCM_CTS,	"TIOCM_CTS",	},
#endif
#ifdef TIOCM_CAR
  { TIOCM_CAR,	"TIOCM_CAR",	},
#endif
#ifdef TIOCM_CD
  { TIOCM_CD,	"TIOCM_CD",	},
#endif
#ifdef TIOCM_RNG
  { TIOCM_RNG,	"TIOCM_RNG",	},
#endif
#ifdef TIOCM_RI
  { TIOCM_RI,	"TIOCM_RI",	},
#endif
#ifdef TIOCM_DSR
  { TIOCM_DSR,	"TIOCM_DSR",	},
#endif
  { 0,		NULL,		},
};

int term_ioctl(msg_building_info *info, long *code, long *arg) {
  struct termios *tios;
  struct termio *tio;
  struct winsize *ws;
#ifdef TIOCGSIZE /*non esiste*/
  struct  ttysize *ts;
#endif
  int i;

  if (arg!=NULL)
    {
      tios=(struct termios*) arg;
      tio=(struct termio*) arg;
      ws=(struct winsize*) arg;
#ifdef TIOCGSIZE /*non esiste*/
      ts=(struct ttysize*) arg;
#endif
    }
  else
    {
      return 0;
    }

  switch ((*(code))) {
    
    /* ioctls with termios or termio args */
    
#ifdef TCGETS /*trovata in asm/ioctls.h*/
  case TCGETS:
    if (syserror(info))
      return 0;
  case TCSETS:
  case TCSETSW:
  case TCSETSF:
    tprintf(", {c_iflags=%#lx, c_oflags=%#lx, ",((long) tios->c_iflag), ((long) tios->c_oflag));
    tprintf("c_cflags=%#lx, c_lflags=%#lx, ",((long) tios->c_cflag), ((long) tios->c_lflag));
    tprintf("c_line=%u, ", tios->c_line);
    if (!(tios->c_lflag & ICANON))
      tprintf("c_cc[VMIN]=%d, c_cc[VTIME]=%d, ",tios->c_cc[VMIN], tios->c_cc[VTIME]);
    tprintf("c_cc=\"");
    for (i = 0; i < NCCS; i++)
      tprintf("\\x%02x", tios->c_cc[i]);
    tprintf("\"}");
    return 1;
#endif /* TCGETS */
    
#ifdef TCGETA /*trovata in asm/ioctls.h*/
  case TCGETA:
    if (syserror(info))
      return 0;
  case TCSETA:
  case TCSETAW:
  case TCSETAF:
    tprintf(", {c_iflags=%#lx, c_oflags=%#lx, ",((long) tio->c_iflag), ((long) tio->c_oflag));
    tprintf("c_cflags=%#lx, c_lflags=%#lx, ",((long) tio->c_cflag), ((long) tio->c_lflag));
    tprintf("c_line=%u, ", tio->c_line);
#ifdef _VMIN /*non trovata*/
    if (!(tio->c_lflag & ICANON))
      tprintf("c_cc[_VMIN]=%d, c_cc[_VTIME]=%d, ",tio->c_cc[_VMIN], tio->c_cc[_VTIME]);
#else /* !_VMIN */
    if (!(tio->c_lflag & ICANON))
      tprintf("c_cc[VMIN]=%d, c_cc[VTIME]=%d, ",tio->c_cc[VMIN], tio->c_cc[VTIME]);
#endif /* !_VMIN */
    tprintf("c_cc=\"");
    for (i = 0; i < NCC; i++)
      tprintf("\\x%02x", tio->c_cc[i]);
    tprintf("\"}");
    return 1;
#endif /* TCGETA */
    
    /* ioctls with winsize or ttysize args */
    
#ifdef TIOCGWINSZ /*definita in /usr/include/asm/ioctls.h*/
  case TIOCGWINSZ:
    if (syserror(info))
      return 0;
  case TIOCSWINSZ:
    tprintf(", {ws_row=%d, ws_col=%d, ws_xpixel=%d, ws_ypixel=%d}",ws->ws_row, ws->ws_col, ws->ws_xpixel, ws->ws_ypixel);
    return 1;
#endif /* TIOCGWINSZ */
    
#ifdef TIOCGSIZE /*non la trovo*/
  case TIOCGSIZE:
    if (syserror(info))
      return 0;
  case TIOCSSIZE:
    tprintf(", {ts_lines=%d, ts_cols=%d}",ts->ts_lines, ts->ts_cols);
    return 1;
#endif
    
    /* ioctls with a direct decodable arg */
    
  case TCXONC:
    tprintf(", ");
    printxval(tcxonc_options, *((int*)arg), "TC???");
    return 1;
  case TCFLSH:
    tprintf(", ");
    printxval(tcflsh_options, *((int*)arg), "TC???");
    return 1;
    
    /* ioctls with an indirect parameter displayed as modem flags */
    
#ifdef TIOCMGET /*definita in /usr/include/asm/ioctls.h*/
  case TIOCMGET:
  case TIOCMBIS:
  case TIOCMBIC:
  case TIOCMSET:
    tprintf(", [");
    if (!printflags(modem_flags, *((int*)arg)))
      tprintf("0");
    tprintf("]");
    return 1;
#endif /* TIOCMGET */
    
    /* ioctls with an indirect parameter displayed in decimal */
    
  case TIOCSPGRP:
  case TIOCGPGRP:
#ifdef TIOCGETPGRP /*non trovata*/
  case TIOCGETPGRP:
#endif
#ifdef TIOCSETPGRP /*non trovata*/
  case TIOCSETPGRP:
#endif
#ifdef FIONREAD /*definita in /usr/include/asm/ioctls.h*/
  case FIONREAD:
#endif
  case TIOCOUTQ:
#ifdef FIONBIO /*definita in /usr/include/asm/ioctls.h*/
  case FIONBIO:
#endif
#ifdef FIOASYNC /*definita in /usr/include/asm/ioctls.h*/
  case FIOASYNC:
#endif
#ifdef FIOGETOWN /*definita in asm/sockios.h*/
  case FIOGETOWN:
#endif
#ifdef FIOSETOWN /*definita in asm/sockios.h*/
  case FIOSETOWN:
#endif
#ifdef TIOCGETD /*definita in /usr/include/asm/ioctls.h*/
  case TIOCGETD:
#endif
#ifdef TIOCSETD /*definita in /usr/include/asm/ioctls.h*/
  case TIOCSETD:
#endif
#ifdef TIOCPKT /*definita in /usr/include/asm/ioctls.h*/
  case TIOCPKT:
#endif
#ifdef TIOCREMOTE /*non trovata*/
  case TIOCREMOTE:
#endif
#ifdef TIOCUCNTL /*non trovata*/
  case TIOCUCNTL:
#endif
#ifdef TIOCTCNTL /*non trovata*/
  case TIOCTCNTL:
#endif
#ifdef TIOCSIGNAL /*non trovata*/
  case TIOCSIGNAL:
#endif
#ifdef TIOCSSOFTCAR /*trovata in asm/ioctls.h*/
  case TIOCSSOFTCAR:
#endif
#ifdef TIOCGSOFTCAR /*trovata in asm/ioctls.h*/
  case TIOCGSOFTCAR:
#endif
#ifdef TIOCISPACE /*non trovata*/
  case TIOCISPACE:
#endif
#ifdef TIOCISIZE /*non trovata*/
  case TIOCISIZE:
#endif
#ifdef TIOCSINTR /*non trovata*/
  case TIOCSINTR:
#endif
#ifdef TIOCSPTLCK /*trovata in asm/ioctls.h*/
  case TIOCSPTLCK:
#endif
#ifdef TIOCGPTN /*trovata in asm/ioctls.h*/
  case TIOCGPTN:
#endif
    tprintf(", ");
    realprintnum((int*) arg, "%d");
    return 1;

    /* ioctls with an indirect parameter displayed as a char */
    
#ifdef TIOCSTI /*trovata in asm/ioctls.h*/
  case TIOCSTI:
#endif
    tprintf(", ");
    realprintstr(((char*)arg), 1);
    return 1;
    
    /* ioctls with no parameters */
    
#ifdef TIOCSCTTY /*trovata in asm/ioctls.h*/
  case TIOCSCTTY:
#endif
#ifdef TIOCNOTTY /*trovata in asm/ioctls.h*/
  case TIOCNOTTY:
#endif
#ifdef FIOCLEX /*trovata in asm/ioctls.h*/
  case FIOCLEX:
#endif
#ifdef FIONCLEX /*trovata in asm/ioctls.h*/ 
  case FIONCLEX:
#endif
#ifdef TIOCCONS /*trovata in asm/ioctls.h*/
  case TIOCCONS:
#endif
    return 1;
    /* ioctls which are unknown */   
  default:
    return 0;
  }
}







