/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: time.c,v 1.2 2003/02/17 15:28:48 frassi Exp $
 */

//#include "defs.h"
//#include "gestione_automa.h"
#include "gets.h"
#include "util.h"
#include "pars_funct.h" /* get_nargs */


#include <linux/version.h>
#include <sys/timex.h>


void
realprinttv(tv)
struct timeval *tv;
{
  if (tv == NULL)
    tprintf("{NULL}");
  else
    tprintf("{%lu, %lu}", (long) tv->tv_sec, (long) tv->tv_usec);
}


struct timeval *
get_tv(tcp,addr)
struct tcb *tcp;
long addr;
{
  struct timeval *tv;
  
  get_error = 0;
  
  if ((void *) addr == NULL)
    return NULL;

  if ((tv = ((struct timeval*) malloc(sizeof(struct timeval)))) == NULL)
    {
      get_error = 1;
      fprintf(stderr,"get_tv: no memory\n");
      return NULL;
    }
    
  if (umove(tcp, addr, tv) < 0) 
    {
      get_error = 1;
      fprintf(stderr, "get_tv: umove\n");
      free (tv);
      return NULL;
    }
  
  return tv;
}


void
realprinttz(tz)
struct timezone *tz;
{
  if (tz == NULL)
    tprintf("{NULL}");
  else
    tprintf("{%d, %d}", (int) tz->tz_minuteswest, (int) tz->tz_dsttime);
}



struct timezone *
get_tz(tcp,addr)
struct tcb *tcp;
long addr;
{
  struct timezone *tz;
  
  get_error = 0;
  
  if ((void *) addr == NULL)
    return NULL;
  
  if ((tz = ((struct timezone*) malloc(sizeof(struct timezone)))) == NULL)
    {
      get_error = 1;
      fprintf(stderr,"get_tz: no memory\n");
      return NULL;
    }

  if (umove(tcp, addr, tz) < 0) 
    {
      get_error = 1;
      fprintf(stderr, "get_tz: umove\n");
      free (tz);
      return NULL;
    }
  
  return tz;
}

struct itimerval *
get_itv(tcp, addr) 
struct tcb *tcp;
long addr;
{
  struct itimerval *itv;
  
  get_error = 0;

  if ((void *) addr == NULL)
    return NULL;

  if ((itv=((struct itimerval*) malloc(sizeof(struct itimerval)))) == NULL) 
    {
      get_error = 1;
      fprintf(stderr,"get_itv: no memory\n");
      return NULL;
    }
  
  if ((umove(tcp, addr, itv) < 0)) 
    {
      get_error = 1;
      fprintf(stderr, "get_itv: umove\n");
      free (itv);
      return NULL;
    }
  
  return itv;
}



struct timex *
get_timex(tcp, addr) 
struct tcb *tcp;
long addr;
{
  struct timex *tx;
  
  get_error = 0;
  
  if ((void *) addr == NULL)
    return NULL;

  if ((tx = ((struct timex*) malloc(sizeof(struct timex)))) == NULL)
    {
      get_error = 1;
      fprintf(stderr,"get_timex: no memory\n");
      return NULL;
    }
    
  if (umove(tcp, addr, tx) < 0) 
    {
      get_error = 1;
      fprintf(stderr, "get_timex: umove\n");
      free (tx);
      return NULL;
    }
  
  return tx;
}

int 
print_time(info)
msg_building_info *info;
{
  tprintf("time(");
  
  if (info->arg_state[0]==TAKEN)
    { 
      realprintnum((int*)info->u_arg[0], "%ld");
      tprintf(")");
    }
  else
    tprintf("_)");
  
  return 0;
}


int 
sys_time(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  if (exiting(tcp)) {

    if (syserror(tcp)) {
      fprintf(stderr, "sys_time: syscall failed\n");
      return -1;
    }

    ((long*) tcp->u_arg[n_arg]) = get_num(tcp, tcp->u_arg[n_arg]);
    if ( get_error )
      {
	fprintf(stderr, "sys_time: get_num\n");
	tcp->arg_state[n_arg] = EMPTY;
	return -1;
      }
    
    tcp->arg_state[n_arg] = TAKEN;
  }

  return 0;
}



int 
print_stime(info)
msg_building_info *info;
{
  tprintf("stime(");
  
  if (info->arg_state[0]==TAKEN)
    { 
      realprintnum((int*)info->u_arg[0], "%ld");
      tprintf(")");
    }
  else
    tprintf("_)");
  
  return 0;
}


/*
  Numero argomenti: 1
  -) struttura contenente i valori data e ora (time_t * == long int *)
*/
int 
sys_stime(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  if (exiting(tcp)) {

    if (syserror(tcp)) {
      fprintf(stderr, "sys_stime: syscall failed\n");
      return -1;
    }

    ((long*) tcp->u_arg[0]) = get_num(tcp, tcp->u_arg[0]);
    if ( get_error )
      {
	fprintf(stderr, "sys_stime: get_num\n");
	tcp->arg_state[0] = EMPTY;
	return -1;
      }
    tcp->arg_state[0] = TAKEN;
  }
  
  return 0;
}



int 
print_gettimeofday(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("gettimeofday(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    realprinttv((struct timeval*)info->u_arg[0]);
	    break;
	  case 1:
	    realprinttz((struct timezone*)info->u_arg[1]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    

    }
  
  tprintf(") ");

  return 0;
}



int
sys_gettimeofday(tcp,n_arg)
struct tcb *tcp;
int n_arg;
{
  if (exiting(tcp)) {

    if (syserror(tcp)) {
      fprintf(stderr, "sys_gettimeofday: syscall failed\n");
      return -1;
    }

    switch(n_arg) {
    case 0:
      ((struct timeval*) tcp->u_arg[0]) = get_tv(tcp, tcp->u_arg[0]);
      if ( get_error )
	{
	  fprintf(stderr, "sys_gettimeofday: get_tv\n");
	  tcp->arg_state[0] = EMPTY;
	  return -1;
	}
      tcp->arg_state[0] = TAKEN;
      break;
    case 1:
      /*The  use of the timezone struct is obsolete*/
      /*perhaps it would be better if it returned NULL?*/
      ((struct timezone *) tcp->u_arg[1]) = get_tz(tcp, tcp->u_arg[1]);
      if ( get_error )
	{
	  fprintf(stderr, "sys_gettimeofday: get_tz\n");
	  tcp->arg_state[1] = EMPTY;
	  return -1;
	}
      tcp->arg_state[1] = TAKEN;
      break;
    }
  }
  
  return 0;
}



int 
print_settimeofday(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("settimeofday(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    realprinttv((struct timeval*)info->u_arg[0]);
	    break;
	  case 1:
	    realprinttz((struct timezone*)info->u_arg[1]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    

    }
  
  tprintf(") ");

  return 0;
}




/*
  Numero argomenti: 2
  -) struttura contenente i valori data e ora (const struct timeval *)
  -) struttura contenente i valori relativi alla zona (const struct timezone *)
*/
int 
sys_settimeofday(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  //if (entering(tcp)) {
    switch (n_arg) {
    case 0:
      tcp->u_arg[0] = (long) get_tv(tcp,tcp->u_arg[0]);
      if ( get_error )
	{
	  fprintf(stderr, "sys_settimeofday: get_tv\n");
	  tcp->arg_state[0] = EMPTY;
	  return -1;
	}
      tcp->arg_state[0] = TAKEN;
      break;
    case 1:
      tcp->u_arg[1] = (long) get_tz(tcp,tcp->u_arg[1]);
      if ( get_error )
	{
	  fprintf(stderr, "sys_settimeofday: get_tz\n");
	  tcp->arg_state[1] = EMPTY;
	  return -1;
	}
      tcp->arg_state[1] = TAKEN;
      break;
    }
    // }
  return 0;
}



int 
print_adjtime(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("gettimeofday(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	realprinttv((struct timeval*)info->u_arg[i]);
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    

    }
  
  tprintf(") ");

  return 0;
}



/*
  Utilizzata per la sys_nanosleep
  -------------
  Numero argomenti: 2
  -) struttura contenente i valori di data e ora (const struct timeval *)
  -) struttura che conterra' i valori di data e ora (struct timeval *)
*/
int 
sys_adjtime(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    //if (entering(tcp)) 
    //{
	tcp->u_arg[n_arg] = (long) get_tv(tcp,tcp->u_arg[n_arg]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_adjtime: get_tv\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
	// }
    break;
  case 1:
    if (exiting(tcp)) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_adjtime: syscall failed\n");
	    return -1;
	  }
	
	tcp->u_arg[n_arg] = (long) get_tv(tcp,tcp->u_arg[n_arg]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_adjtime: get_tv\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  return 0;
}


static struct xlat which[] = {
	{ ITIMER_REAL,	"ITIMER_REAL"	},
	{ ITIMER_VIRTUAL,"ITIMER_VIRTUAL"},
	{ ITIMER_PROF,	"ITIMER_PROF"	},
	{ 0,		NULL		},
};



static void
realprintitv(itv)
struct itimerval *itv;
{
  if (itv == NULL)
    tprintf("{NULL}");
  else 
    {
      tprintf("{it_interval={%lu, %lu}, it_value={%lu, %lu}}",
	      (long) itv->it_interval.tv_sec, (long) itv->it_interval.tv_usec,
	      (long) itv->it_value.tv_sec, (long) itv->it_value.tv_usec);
    }
}


int 
print_getitimer(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("getitimer(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(which, *((int*)info->u_arg[0]), "ITIMER_???");
	    break;
	  case 1:
	    realprintitv((struct itimerval*)info->u_arg[1]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    

    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 2
  -) indicatore del timer da restituire (int)
  -) struttura che conterra' i valori del timer restituito (struct itimerval *)
*/
int 
sys_getitimer(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) 
    {
    case 0:
      arg_alloc(tcp,n_arg);
      break;
    case 1:
      if (exiting(tcp)) 
	{
	  if (syserror(tcp))
	    {
	      fprintf(stderr, "sys_getitimer: syscall failed\n");
	      return -1;
	    }

	  tcp->u_arg[n_arg] = (long) get_itv(tcp,tcp->u_arg[n_arg]);
	  if ( get_error )
	    {
	      fprintf(stderr, "sys_getitimer: get_itv\n");
	      tcp->arg_state[n_arg] = EMPTY;
	      return -1;
	    }
	  tcp->arg_state[n_arg] = TAKEN;
	}
      break;
    }
  return 0;
}




int 
print_setitimer(info)
msg_building_info *info;
{
  int i;
  int nargs=get_nargs(info->scno);
  
  tprintf("setitimer(");
  
  for (i=0;i<nargs;i++)
    {
      if (info->arg_state[i]==TAKEN)
	{
 	  switch (i) {
	  case 0:
	    printxval(which, *((int*)info->u_arg[0]), "ITIMER_???");
	    break;
	  case 1:
	  case 2:
	    realprintitv((struct itimerval*)info->u_arg[i]);
	    break;
	  }
	}
      else
	tprintf("_");
      
      if (i!=nargs-1)
	tprintf(", ");    

    }
  
  tprintf(") ");

  return 0;
}



/*
  Numero argomenti: 3
  -) indicatore del timer da restituire (int)
  -) struttura contenente i valori del timer (const struct itimerval *)
  -) struttura che conterra' i vecchi valori del timer (struct itimerval *)
*/
int 
sys_setitimer(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  switch (n_arg) {
  case 0:
    arg_alloc(tcp,n_arg);
    break;
  case 1:
    //if (entering(tcp)) {
      tcp->u_arg[n_arg] = (long) get_itv(tcp,tcp->u_arg[n_arg]);
      if ( get_error )
	{
	  fprintf(stderr, "sys_setitimer: get_itv\n");
	  tcp->arg_state[n_arg] = EMPTY;
	  return -1;
	}
      tcp->arg_state[n_arg] = TAKEN;
      //}
    break;
  case 2:
    if ((exiting(tcp)) && (tcp->u_arg[n_arg])) 
      {
	if (syserror(tcp))
	  {
	    fprintf(stderr, "sys_setitimer: syscall failed\n");
	    return -1;
	  }
	
	tcp->u_arg[n_arg] = (long) get_itv(tcp,tcp->u_arg[n_arg]);
	if ( get_error )
	  {
	    fprintf(stderr, "sys_setitimer: get_itv\n");
	    tcp->arg_state[n_arg] = EMPTY;
	    return -1;
	  }
	tcp->arg_state[n_arg] = TAKEN;
      }
    break;
  }
  return 0;
}



int 
print_adjtimex(info)
msg_building_info *info;
{
  struct timex *txc;

  tprintf("adjtimex(");
  
  if (info->arg_state[0]==TAKEN)
    { 
      txc=(struct timex *) info->u_arg[0];
      if (txc)
	{
#if LINUX_VERSION_CODE < 66332 /*non dovrebbe passare il controllo*/

	  tprintf("{mode=%d, offset=%ld, frequency=%ld, ",txc->mode, txc->offset, txc->frequency);
	  tprintf("maxerror=%ld, esterror=%lu, status=%u, ",txc->maxerror, txc->esterror, txc->status);
	  tprintf("time_constant=%ld, precision=%lu, ",txc->time_constant, txc->precision);
	  tprintf("tolerance=%ld, time={%lu, %lu}}",txc->tolerance, (long) txc->time.tv_sec,(long) txc->time.tv_usec);
#else 
	  tprintf("{modes=%d, offset=%ld, freq=%ld, ",txc->modes, txc->offset, txc->freq);
	  tprintf("maxerror=%ld, esterror=%lu, status=%u, ",txc->maxerror, txc->esterror, txc->status);
	  tprintf("constant=%ld, precision=%lu, ",txc->constant, txc->precision);
	  tprintf("tolerance=%ld, time={%lu, %lu}}",txc->tolerance, (long) txc->time.tv_sec,(long) txc->time.tv_usec);
#endif
	}
      else
	tprintf("{NULL}");
      tprintf(")");
    }
  else
    tprintf("_)");
  
  return 0;
}



/*
  Numero argomenti: 1
  -) struttura contenente i nuovi parametri del kernel (struct timex *);
*/
int 
sys_adjtimex(tcp, n_arg) 
struct tcb *tcp;
int n_arg;
{
  tcp->u_arg[n_arg]= (long) get_timex(tcp,tcp->u_arg[n_arg]);
  if ( get_error )
    {
      fprintf(stderr, "sys_adjtimex: get_timex\n");
      tcp->arg_state[n_arg] = EMPTY;
      return -1;
    }
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}
