#ifndef TYPE_SYS_ARGS_H
#define TYPE_SYS_ARGS_H

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: type_sys_args.h,v 1.1.1.1 2001/12/17 08:58:59 frassi Exp $
 */


#include <glib.h>
#include "parser.h"
#include "runtime_util.h"

#ifndef Y_TAB_H
#define Y_TAB_H

//file header per wids.lex 
#include "y.tab.h"

#endif  /* !Y_TAB_H */


typedef enum { 

  /* tipi per le tabelle hash */
  HASH_LONG   = 374,   /* e' il valore contenuto in y.tab.h di LONG_VALUE + 100 */
  HASH_CHAR   = 375,   /* e' il valore contenuto in y.tab.h di CHARACTER_VALUE + 100 */
  HASH_STRING = 376,   /* e' il valore contenuto in y.tab.h di STRING_VALUE + 100 */
  /* fine tipi per le tabelle hash */
  
  INT     =   LONG_VALUE,
  INT_PTR =   LONG_VALUE,
  UINT    =   LONG_VALUE,
  UNSIGNED_INT  =   LONG_VALUE,
  UNSIGNED_INT_PTR =  LONG_VALUE,
  LONG = LONG_VALUE,
  UNSIGNED_LONG = LONG_VALUE,
  UNSIGNED_LONG_PTR =  LONG_VALUE,
  VOID_PTR  =   -1,
  /* CONST_VOID_PTR = -1, */
  CHAR_PTR = STRING_VALUE,
  /* CONST_CHAR_PTR = STRING_VALUE, */
  CHAR_PTR_PTR = -1,
  SIZE_T = LONG_VALUE,
  SIZE_T_PTR = LONG_VALUE,
  PID_T = LONG_VALUE,
  DEV_T = LONG_VALUE,
  MODE_T = LONG_VALUE,
  UID_T = LONG_VALUE,
  UID_T_PTR = LONG_VALUE,
  GID_T = LONG_VALUE,
  GID_T_PTR = LONG_VALUE,
  OFF_T = LONG_VALUE,
  OFF_T_PTR = LONG_VALUE,
  LOFF_T_PTR = LONG_VALUE,
  LOFF_T = LONG_VALUE,
  SIGHANDLER_T = -1,
  OLD_SIGSET_T = -1,
  OLD_SIGSET_T_PTR = -1,
  SIGSET_T_PTR = -1,
  /* CONST_SIGSET_T_PTR = -1, */
  FD_SET_PTR = -1,
  STRUCT_OLD_KERNEL_STAT_PTR = -1,
  STRUCT_TIMEVAL_PTR = -1,
  STRUCT_UTIMBUF_PTR = -1,
  STRUCT_TIMEZONE_PTR = -1,
  STRUCT_TMS_PTR = -1,
  STRUCT_OLDOLD_UTSNAME_PTR = -1,
  STRUCT_USTAT_PTR = -1,
  STRUCT_STAT_PTR = -1,
  STRUCT_OLD_SIGACTION_PTR = -1,
  /* CONST_STRUCT_OLD_SIGACTION_PTR = -1, */
  /* CONST_STRUCT_SIGACTION_PTR = -1, */
  STRUCT_SIGACTION_PTR = -1,
  STRUCT_RLIMIT_PTR = -1,
  STRUCT_SEL_ARG_STRUCT_PTR = -1,
  STRUCT_MMAP_ARG_STRUCT_PTR = -1,
  STRUCT_RUSAGE_PTR = -1,
  STRUCT_OLD_UTSNAME_PTR = -1,
  STRUCT_NEW_UTSNAME_PTR = -1,
  STRUCT_STATFS_PTR = -1,
  STRUCT_ITIMERVAL_PTR = -1,
  STRUCT_VM86_STRUCT_PTR = -1,
  STRUCT_VM86PLUS_STRUCT_PTR = -1,
  STRUCT_SYSINFO_PTR = -1,
  STRUCT_PT_REGS = -1,
  STRUCT_TIMEX_PTR = -1,
  STRUCT_KERNEL_SYM_PTR = -1,
  STRUCT_MODULE_PTR = -1,
  CADDR_T = -1,
  STRUCT_SYSCTL_ARGS_PTR = -1,
  /* CONST_STRUCT_IOVEC_PTR = -1, */
  STRUCT_IOVEC_PTR = -1,
  STRUCT_SCHED_PARAM_PTR = -1,
  STRUCT_TIMESPEC_PTR = -1,
  /* CONST_STRUCT_TIMESPEC_PTR = -1, */
  STRUCT_POLLFD_PTR = -1,
  SIGINFO_T_PTR = -1,
  CAP_USER_HEADER_T = -1,
  CONST_STACK_T_PTR = -1,
  CAP_USER_DATA_T = -1,
  /* CONST_CAP_USER_DATA_T = -1, */
  STACK_T_PTR = -1,
  STRUCT_SOCKADDR_PTR = -1,
  SOCKLEN_T = -1,
  /* CONST_STRUCT_SOCKADDR_PTR = -1, */
  SOCKLEN_T_PTR = -1,
  /* ARRAY_OF_TWO_INT = -1, */
  /* CONST_STRUCT_MSGHDR_PTR = -1, */
  STRUCT_MSGHDR_PTR = -1,
  STRUCT_MSGHDT_PTR = -1,
  OLD_SIGSET_PTR = -1,
  U_SHORT_PTR = LONG_VALUE

} sys_args_type_t;

#endif /* !TYPE_SYS_ARGS_H */
