/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * Copyright (c) 1999 IBM Deutschland Entwicklung GmbH, IBM Corporation
 *                     Linux for s390 port by D.J. Barrow
 *                    <barrow_dj@mail.yahoo.com,djbarrow@de.ibm.com>            
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: util.c,v 1.2 2003/02/17 15:29:22 frassi Exp $
 */


#include "gets.h"
#include "util.h"

#include <sys/user.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/uio.h>


int get_error; /* prima era dichiarata in gets.h */


#ifdef HAVE_SYS_REG_H
#include <sys/reg.h>
# define PTRACE_PEEKUSR PTRACE_PEEKUSER
#elif defined(HAVE_LINUX_PTRACE_H)
#undef PTRACE_SYSCALL
#include <linux/ptrace.h>
#endif


/* macros */
#ifndef MAX
#define MAX(a,b)		(((a) > (b)) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b)		(((a) < (b)) ? (a) : (b))
#endif


/*
printxval

input: una struttura xlat, un valore intero

output: nessuno

stampa la stringa contenuta nella tabella delle struct xlat corrispondente al 
valore VAL. Se non trova nessuna stringa, stampa il valore stesso e un commento (dflt).

commento originale: Print entry in struct xlat table, if there.
*/

void
printxval(xlat, val, dflt)
struct xlat *xlat;
int val;
char *dflt;
{
  char *str = get_xlat_str(xlat, val);
  
  if (str)
    tprintf("%s", str);
  else
    tprintf("%#x /* %s */", val, dflt);
}

/*
addflags

input: una struttura xlat,un intero flags

output:il numero di flags stampati

stampa a video le stringhe corrispondenti ai flags contenuti nella variabile 
flags, estraendoli uno a uno. In piu' l'eventuale valore della variabile 
flags, se diverso da zero, alla fine di questo procedimento 

commento originale: * Interpret `xlat' as an array of flags
                    * print the entries whose bits are on in `flags'
                    * return # of flags printed.
*/

int
addflags(xlat, flags)
struct xlat *xlat;
int flags;
{
  int n;

  for (n = 0; xlat->str; xlat++) 
    {
      if (xlat->val && (flags & xlat->val) == xlat->val) 
	{
	  tprintf("|%s", xlat->str);
	  flags &= ~xlat->val;
	  n++;
	}
    }

  if (flags) 
    {
      tprintf("|%#x", flags);
      n++;
    }

  return n;
}

/*
printflags

input: una struttura xlat, un intero flags

output:il numero di flags stampati

simile alla addflags.
prevede il caso in cui l'xlat contiene un unico valore (lo zero).
Stampa le stringhe corrispondenti ai flags a video, separandole con un 
opportuno carattere di separazione
*/

int
printflags(xlat, flags)
struct xlat *xlat;
int flags;
{
  int n;
  char *sep;
  
  if (flags == 0 && xlat->val == 0) 
    {
      tprintf("%s", xlat->str);
      return 1;
    }

  sep = "";

  for (n = 0; xlat->str; xlat++) 
    {
      if (xlat->val && (flags & xlat->val) == xlat->val) 
	{
	  tprintf("%s%s", sep, xlat->str);
	  flags &= ~xlat->val;
	  sep = "|";
	  n++;
	}
    }
	
  if (flags) 
    {
      tprintf("%s%#x", sep, flags);
      n++;
    }

  return n;
}

/*
printnum

input:una struttura di tipo tcb, un indirizzo di memoria, una stringa di 
formattazione

output: nessuno

stampa la parola di memoria presente all'indirizzo addr, seguendo la stringa 
di formattazione fornitagli.
Se addr non e' un indirizzo stampa NULL ed esce
Se la lettura della parola di memoria fallisce, allora ne stampa l'indirizzo
*/


void
printnum(tcp, addr, fmt)
struct tcb *tcp;
long addr;
char *fmt;
{
  int num;
  
  if (!addr) 
    {
      tprintf("NULL");
      return;
    }
	
  if (umove(tcp, addr, &num) < 0) 
    {
      tprintf("%#lx", addr);
      return;
    }
  
  tprintf("[");
  tprintf(fmt, num);
  tprintf("]");
}

static char path[MAXPATHLEN + 1];

/*
string_quote

input: una stringa

output: nessuno

seghe mentali per stampare un path

procedura usata da printpath e printpathn
*/

void
string_quote(str)
char *str;
{
	char buf[2 * MAXPATHLEN + 1];
	char *s;

	if (!strpbrk(str, "\"\'\\")) {
		tprintf("\"%s\"", str);
		return;
	}
	for (s = buf; *str; str++) {
		switch (*str) {
		case '\"': case '\'': case '\\':
			*s++ = '\\'; *s++ = *str; break;
		default:
			*s++ = *str; break;
		}
	}
	*s = '\0';
	tprintf("\"%s\"", buf);
}

/*
printpath

input: una struttura di tipo tcb, un indirizzo

output: nessuno

stampa una stringa rappresentante un path contenuta nella struttura tcp 
all'indirizzo addr.
se la lettura della stringa fallisce, stampa l'indirizzo che ha cercato di 
leggere.
*/

void
printpath(tcp, addr)
struct tcb *tcp;
long addr;
{
  if (umovestr(tcp, addr, MAXPATHLEN, path) < 0)
    tprintf("%#lx", addr);
  else
    string_quote(path);
  return;
}

/*
printpathn

input: una struttura di tipo tcb, un indirizzo, la lunghezza n 

output: nessuno

uguale a printpath, con la differenza che posso specificare usando n, 
quanto deve essere lungo il path da leggere
*/


void
printpathn(tcp, addr, n)
struct tcb *tcp;
long addr;
int n;
{
  if (umovestr(tcp, addr, n, path) < 0)
    tprintf("%#lx", addr);
  else 
    {
      path[n] = '\0';
      string_quote(path);
    }
}

/*
printstr

input:una struttura di tipo tcb, un indirizzo di memoria, una lunghezza len

output: nessuno

stampa una stringa contenuta all'indirizzo addr.
*/


void
printstr(tcp, addr, len)
struct tcb *tcp;
long addr;
int len;
{
  static unsigned char *str = NULL;
  static char *outstr;
  int i, n, c, usehex;
  char *s, *outend;
  
  if (!addr) 
    {
      tprintf("NULL");
      return;
    }
	
  if (!str) 
    {
      if ((str = malloc(max_strlen)) == NULL || (outstr = malloc(2*max_strlen)) == NULL) 
	{
	  fprintf(stderr, "printstr: no memory\n");
	  tprintf("%#lx", addr);
	  return;
	}
    }
  outend = outstr + max_strlen * 2 - 10;
  if (len < 0) 
    {
      n = max_strlen;
      if (umovestr(tcp, addr, n, (char *) str) < 0) 
	{
	  tprintf("%#lx", addr);
	  return;
	}
    }
  else 
    {
      n = MIN(len, max_strlen);
      if (umoven(tcp, addr, n, (char *) str) < 0) 
	{
	  tprintf("%#lx", addr);
	  return;
	}
    }

  usehex = 0;
	
/* 
il for controlla se sono tutti caratteri ASCII altrimenti li stampa in hex
*/

	for (i = 0; i < n; i++) {
	  c = str[i];
	  if (len < 0 && c == '\0')
	    break;
	  if (!isprint(c) && !isspace(c)) {
	    usehex = 1;
	    break;
	  }
	}
	
        s = outstr;
	*s++ = '\"';
	
	if (usehex) {
		for (i = 0; i < n; i++) {
			c = str[i];
			if (len < 0 && c == '\0')
				break;
			sprintf(s, "\\x%02x", c);
			s += 4;
			if (s > outend)
				break;
		}
	}
	else {
		for (i = 0; i < n; i++) {
			c = str[i];
			if (len < 0 && c == '\0')
				break;
			switch (c) {
			case '\"': case '\'': case '\\':
				*s++ = '\\'; *s++ = c; break;
			case '\f':
				*s++ = '\\'; *s++ = 'f'; break;
			case '\n':
				*s++ = '\\'; *s++ = 'n'; break;
			case '\r':
				*s++ = '\\'; *s++ = 'r'; break;
			case '\t':
				*s++ = '\\'; *s++ = 't'; break;
			case '\v':
				*s++ = '\\'; *s++ = 'v'; break;
			default:
				if (isprint(c))
					*s++ = c;
				else if (i < n - 1 && isdigit(str[i + 1])) {
					sprintf(s, "\\%03o", c);
					s += 4;
				}
				else {
					sprintf(s, "\\%o", c);
					s += strlen(s);
				}
				break;
			}
			if (s > outend)
				break;
		}
	}

	*s++ = '\"';
	if (i < len || (len < 0 && (i == n || s > outend))) {
		*s++ = '.'; *s++ = '.'; *s++ = '.';
	}
	*s = '\0';
	tprintf("%s", outstr);
}



// #define PAGMASK	(~(PAGSIZ - 1))


/*
umoven == umove

input: una struttura di tipo tcp, un indirizzo di memoria (relativo al 
processo 'pid'), un intero (rappresentante la lunghezza di uno spazio di 
memoria), un indirizzo (locale) laddr espresso come stringa

output:0 oppure -1

sposta 'len' bytes di dati relativi al processo 'pid'  a partire 
dall'indirizzo 'addr' nel nostro spazio all'indirizzo 'laddr'

commento originale: * move `len' bytes of data from process `pid'
                    * at address `addr' to our space at `laddr'

#define umove(pid, addr, objp)       umoven((pid), (addr), sizeof *(objp), (char *) (objp))

*/

int
umoven(tcp, addr, len, laddr)
struct tcb *tcp;
long addr;
int len;
char *laddr;
{
	int pid = tcp->pid;
	int n, m;
	int started = 0;
	union {
		long val;
		char x[sizeof(long)];
	} u;

	if (addr & (sizeof(long) - 1)) {
		/* addr not a multiple of sizeof(long) */
		n = addr - (addr & -sizeof(long)); /* residue */
		addr &= -sizeof(long); /* residue */
		errno = 0;
		u.val = ptrace(PTRACE_PEEKDATA, pid, (char *) addr, 0);
		if (errno) {
			if (started && (errno==EPERM || errno==EIO)) {
				/* Ran into 'end of memory' - stupid "printpath" */
				return 0;
			}
			/* But if not started, we had a bogus address. */
			perror("ptrace: umoven");
			return -1;
		}
		started = 1;
		memcpy(laddr, &u.x[n], m = MIN(sizeof(long) - n, len));
		addr += sizeof(long), laddr += m, len -= m;
	}
	while (len) {
		errno = 0;
		u.val = ptrace(PTRACE_PEEKDATA, pid, (char *) addr, 0); /* che succede se addr e' 0 */
		if (errno) {
			if (started && (errno==EPERM || errno==EIO)) {
				/* Ran into 'end of memory' - stupid "printpath" */
				return 0;
			}
			perror("ptrace: umoven");
			return -1;
		}
		started = 1;
		memcpy(laddr, u.x, m = MIN(sizeof(long), len));
		addr += sizeof(long), laddr += m, len -= m;
	}

	return 0;
}

/*
umovestr

input: una struttura di tipo tcp, un indirizzo di memoria 
(relativo al processo 'pid'), un intero (rappresentante la lunghezza di uno 
spazio di memoria), un indirizzo (locale) laddr espresso come stringa

output:0 oppure -1
 
like umove but make the additional effort of looking for a terminating zero byte.
*/

int
umovestr(tcp, addr, len, laddr)
struct tcb *tcp;
long addr;
int len;
char *laddr;
{
       	int started = 0;
	int pid = tcp->pid;
	int i, n, m;
	union {
		long val;
		char x[sizeof(long)];
	} u;

	if (addr & (sizeof(long) - 1)) {
		/* addr not a multiple of sizeof(long) */
		n = addr - (addr & -sizeof(long)); /* residue */
		addr &= -sizeof(long); /* residue */
		errno = 0;
		u.val = ptrace(PTRACE_PEEKDATA, pid, (char *)addr, 0);
		if (errno) {
			if (started && (errno==EPERM || errno==EIO)) {
				/* Ran into 'end of memory' - stupid "printpath" */
				return 0;
			}
			perror("umovestr");
			return -1;
		}
		started = 1;
		memcpy(laddr, &u.x[n], m = MIN(sizeof(long)-n,len));
		while (n & (sizeof(long) - 1))
		  if (u.x[n++] == '\0')
		    return 1; /* valore originario era 0*/
		addr += sizeof(long), laddr += m, len -= m;
	}
	while (len) {
		errno = 0;
		u.val = ptrace(PTRACE_PEEKDATA, pid, (char *)addr, 0);
		if (errno) {
			if (started && (errno==EPERM || errno==EIO)) {
				/* Ran into 'end of memory' - stupid "printpath" */
				return 0;
			}
			perror("umovestr");
			return -1;
		}
		started = 1;
		memcpy(laddr, u.x, m = MIN(sizeof(long), len));
		for (i = 0; i < sizeof(long); i++)
		  if (u.x[i] == '\0')
		    return 1; /* valore originario era 0 */

		addr += sizeof(long), laddr += m, len -= m;
	}
	return 0;
}

#define PTRACE_WRITETEXT	101
#define PTRACE_WRITEDATA	102

/*
upeek

input:il pid di un processo,un offset off, un long RES usato per restituire 
un risultato

output:0 se tutto ok, -1 in caso di errore

legge una parola di memoria nella user area del figlio pid, a partire 
dall'offset off.
Mette la parola letta nella variabile res.
*/

int
upeek(pid, off, res)
int pid;
long off;
long *res;
{
	long val;

	errno = 0;
	val = ptrace(PTRACE_PEEKUSER, pid, (char *) off, 0);
	if (val == -1 && errno) {
		perror("upeek: ptrace(PTRACE_PEEKUSER, ... )");
		return -1;
	}
	*res = val;
	return 0;
}


/*
get_str

input:una struttura di tipo tcb, un indirizzo di memoria, una lunghezza len

output: un puntatore alla stringa

prende la stringa contenuta all'indirizzo addr, alloca la memoria per la stringa
e restituisce un puntatore a questa area di memoria.

deriva dalla printstr

*/


char *
get_str(tcp, addr, len)
struct tcb *tcp;
long addr;
int len;
{
  /* int standard_length = 4096;  TODO: sostituire questa variabile con una costante con un valore appropriato */
#define STANDARD_LENGTH       4096 // lunghezza massima di una stringa
        int got_all_str = 0;
	
	char *str = NULL;
	char *buf;
	int n,offset,result;

	get_error = 0;

	if (!addr) 
	  return NULL;

	if (len >= 0)
	  str = malloc(len+1);
	else
	  str = malloc(STANDARD_LENGTH);
	  
	if (str == NULL) /* controllo fallimento malloc */
	  {
	    fprintf(stderr, "get_str: no memory\n");
	    get_error = 1;
	    return NULL;
	  }

	if (len < 0)
	  {
	    n = 2;
	    offset = 0;
	    buf = str;
	    while(!(got_all_str))
	      {
		if ((result = umovestr(tcp, addr, STANDARD_LENGTH, (char *) buf)) < 0) 
		  {
		    fprintf(stderr, "get_str: umovestr failed\n");
		    free(str);
		    get_error = 1;
		    return NULL;
		  }

		if (result != 1) /* la stringa non e' ancora terminata */
		  {
		    addr += STANDARD_LENGTH;
		    offset += STANDARD_LENGTH;
		    str = realloc(str, n * STANDARD_LENGTH);
		    buf = &str[offset];
		    n++;
		  } 
		else
		  got_all_str = 1;

	      }
	  }
	else
	  {
	    if (umoven(tcp, addr, len, (char *) str) < 0) 
	      {
		fprintf(stderr, "getstr: umoven failed\n");
		get_error = 1;
		free(str);
		return NULL;
	      }
	    
	    /* forse non necessario, ma cautelarsi e' meglio */
	    str[len] = '\0';
	  }
	
	return str;
}






/*
realprintstr

input: un puntatore a stringa

output: nessuno

stampa una stringa, eventualmente in esadecimale
*/


void
realprintstr(str,len)
unsigned char * str;
unsigned int len;
{

	static char *outstr;
	int i, n, c, usehex;
	char *s, *outend;
	

	if (!str) 
	  {
	    tprintf("NULL");
	    return;
	  }

	if (len >= 0)
	  n = MIN(len, max_strlen);
	else
	  n = MIN(strlen(str), max_strlen);

	if ((outstr = malloc(2*n)) == NULL) 
	  {
	    fprintf(stderr, "realprintstr: no memory\n");
	    return;
	  }
	
	outend = outstr + n * 2 - 10;

	usehex = 0;
	
/* 
il for controlla se sono tutti caratteri ASCII altrimenti li stampa in hex
*/

	for (i = 0; i < n; i++) 
	  {
	    c = str[i];
	    if (len < 0 && c == '\0')
	      break;
	    if (!isprint(c) && !isspace(c)) 
	      {
		usehex = 1;
		break;
	      }
	  }
	
        s = outstr;
	*s++ = '\"';
	
	if (usehex) 
	  {
	    for (i = 0; i < n; i++) 
	      {
		c = str[i];
		if (len < 0 && c == '\0')
		  break;
		sprintf(s, "\\x%02x", c);
		s += 4;
		if (s > outend)
		  break;
	      }
	  }
	else 
	  {
	    for (i = 0; i < n; i++) 
	      {
		c = str[i];
		if (len < 0 && c == '\0')
		  break;
		switch (c) {
		case '\"': case '\'': case '\\':
		  *s++ = '\\'; *s++ = c; break;
		case '\f':
		  *s++ = '\\'; *s++ = 'f'; break;
		case '\n':
		  *s++ = '\\'; *s++ = 'n'; break;
		case '\r':
		  *s++ = '\\'; *s++ = 'r'; break;
		case '\t':
		  *s++ = '\\'; *s++ = 't'; break;
		case '\v':
		  *s++ = '\\'; *s++ = 'v'; break;
		default:
		  if (isprint(c))
		    *s++ = c;
		  else if (i < n - 1 && isdigit(str[i + 1])) 
		    {
		      sprintf(s, "\\%03o", c);
		      s += 4;
		    }
		  else 
		    {
		      sprintf(s, "\\%o", c);
		      s += strlen(s);
		    }
		  break;
		}
		if (s > outend)
		  break;
	      }
	  }

	*s++ = '\"';

	if (i < len || (len < 0 && (i == n || s > outend))) 
	 {
	   *s++ = '.'; *s++ = '.'; *s++ = '.';
         }
	
	*s = '\0';
	tprintf("%s", outstr);
	
	free(outstr);
}






void
realprintnum(num, fmt)
int *num;
char *fmt;
{
  if (!num) {
    tprintf("NULL");
    return;
  }
  
  tprintf("[");
  tprintf(fmt, *num);
  tprintf("]");
}




/*
get_num

input:una struttura di tipo tcb, un indirizzo di memoria

output: un puntatore al numero

prende la parola di memoria presente all'indirizzo addr.
Se addr non e' un indirizzo stampa NULL ed esce
Se la lettura della parola di memoria fallisce, allora ne stampa l'indirizzo

deriva dalla printnum
*/


long *
get_num(tcp, addr)
struct tcb *tcp;
long addr;
{
  long *num;
  
  get_error = 0;
  
  if (!addr) 
    return NULL;

  num = ((long*) malloc(sizeof(long)));
 
  if (num == NULL) /* controllo fallimento malloc */
    {
      fprintf(stderr, "get_num: no memory\n");
      get_error = 1;
      return NULL;
    }

  if (umove(tcp, addr, num) < 0) 
    {
      fprintf(stderr, "get_num: umove\n");
      free(num);
      get_error = 1;
      return NULL;
    }
  
  return num;
}


/*-------------------------------------*/


/*
get_path
deriva da printpath

input: una struttura di tipo tcb, un indirizzo

output: puntatore a path espresso come stringa

restituisce una stringa rappresentante un path contenuta nella struttura tcp all'indirizzo addr.
se la lettura della stringa fallisce, restituisce NULL.
*/

char *
get_path(tcp, addr)
struct tcb *tcp;
long addr;
{
  char *path; 

  get_error = 0;

  if (!addr)
    return NULL;
  
  path = ((char*) malloc (MAXPATHLEN + 1));

  if (path == NULL) /* controllo fallimento malloc */
    {
      fprintf(stderr, "get_path: no memory\n");
      get_error = 1;
      return NULL;
    }

  if (umovestr(tcp, addr, MAXPATHLEN, path) < 0)
    {
      fprintf(stderr, "get_path: umovestr failed\n");
      get_error = 1;
      free(path);
      return NULL;
    }

  return path;
}


/* deriva da prinpathn */
char *
get_pathn(tcp, addr, n)
struct tcb *tcp;
long addr;
int n;
{
  char *path;
  
  get_error = 0;

  if (!addr)
    return NULL;

  //path = ((char*) malloc (MAXPATHLEN + 1));
  path = ((char*) malloc (n + 1));

  if (path == NULL) /* controllo fallimento malloc */
    {
      fprintf(stderr, "get_pathn: no memory\n");
      get_error = 1;
      return NULL;
    }

  if (umovestr(tcp, addr, n, path) < 0)
    {
      fprintf(stderr, "get_pathn: umovestr path failed\n");
      get_error = 1;
      free(path);
      return NULL;
    }
  else 
    {
      path[n] = '\0';
      return path;
    }
}

/*
  VDS modification org.
  
  input: 1) una struttura tcb
         2) il numero dell'argomento da settare all'interno della struct tcb,
	 3) l'indirizzo sorgente
	 4) il puntatore all'indirizzo di una struttura detinazione appropriata
	 5) la dimensione dell'area di memoria da allocare
	 6) una stringa con il nome della funzione chiamante

  output: OK_COPY in caso di sucesso o se addr == NULL, ERR_COPY altrimenti

  Tenta di allocare la quantita' di memoria richiesta, assegnandola alla struttura indicata
  dal chiamante, per poi copiarvi il contenuto dell'area di memoria sorgente. Infine assegna
  l'indirizzo destinazione all'elemento "n_arg" della struttura tcb e ne setta il flag di
  stato a TAKEN.
  
  #define umove(pid, addr, objp)       umoven((pid), (addr), sizeof *(objp), (char *) (objp))

*/
int set_copy(struct tcb *tcp, int n_arg, long addr, long *tmp, int size, char *fun_name)
{
  if ((void *) addr == NULL)
    {
      tcp->arg_state[n_arg] = TAKEN;
      return 0;
    }

  if (size < 0) {
    fprintf(stderr,"set_copy - %s: tried to malloc negative size\n", fun_name);
    return 1;
  }

  if ((*tmp = (long)malloc(size)) == 0) {
    fprintf(stderr,"set_copy - %s: no memory\n", fun_name);
    return 1;
  }
  
  if (umoven(tcp, addr, size, (char *) *tmp) < 0) {
    fprintf(stderr,"set_copy - %s: umove failed\n", fun_name);
    free(tmp);
    tmp = (long)NULL; /* CONTROLLARE: non serve a niente... */
    return 1;
  }
  
  (long)tcp->u_arg[n_arg] = (long)*tmp;
  tcp->arg_state[n_arg] = TAKEN;
  
  return 0;
}

/*
  VDS modification org.
  
  input: 1) una struttura tcb
	 2) l'indirizzo sorgente
	 3) il puntatore all'indirizzo di una struttura detinazione appropriata
	 4) la dimensione dell'area di memoria da allocare
	 5) una stringa con il nome della funzione chiamante

  output: OK_COPY in caso di sucesso, ERR_COPY altrimenti

  Simile alla set_copy, non effettua pero' nessun assegnamento alla struct tcb.
*/

int set_element(struct tcb *tcp, long addr, long *tmp, int size, char *fun_name)
{
  if (size < 0) {
    fprintf(stderr,"%s: tried to malloc negative size\n",fun_name);
    return 1;
  }

  if ((*tmp=(long)malloc(size)) == 0) 
    {
      fprintf(stderr,"set_element - %s: no memory\n",fun_name);
      return 1;
    }
  
  if (umoven(tcp, addr, size, (char *) *tmp) < 0) {
    fprintf(stderr,"set_element - %s: umove failed\n",fun_name);
    free(tmp);
    tmp = (long)NULL; /* CONTROLLARE: non serve a niente... */
    return 1;
  }

  return 0;
}


extern FILE *outf/*, *pipeOut*/;

#include <stdarg.h>

void
tprintf(const char *fmt, ...)
{
  va_list args;
  
  va_start(args, fmt);
  if (outf)
    vfprintf(outf, fmt, args);
  va_end(args);
  
  // TODO: fare bene il lavoro col pipe
/*   va_start(args, fmt); */
/*   if (pipeOut) */
/*     vfprintf(pipeOut, fmt, args); */
/*   va_end(args); */
  
  return;
}


extern struct tcb *current;



void
print_syscall_realtime()
{
  int ret;
  msg_building_info mbi;
  struct tcb *tcp = current;
  
  mbi.scno = tcp->scno;
  mbi.u_error = tcp->u_error;
  mbi.u_rval = tcp->u_rval;
  
  if ( tcp->auxstr )
    mbi.auxstr = strdup(tcp->auxstr);
  else
    mbi.auxstr = NULL;

  memcpy(mbi.u_arg, tcp->u_arg, sizeof tcp->u_arg);
  memcpy(mbi.arg_state, tcp->arg_state, sizeof tcp->arg_state);
  //if (sysent[tcp->scno].print_sys == printargs)
  //{
      //tprintf("%s()", sysent[tcp->scno].sys_name);
      //ret = 0;
  //  ret = (*sysent[tcp->scno].print_sys)(&mbi);
  //}
  //else
  ret = (*sysent[tcp->scno].print_sys)(&mbi);

  print_retval(ret, mbi.u_rval, mbi.u_error, mbi.auxstr);
  tprintf("\n");

  free(mbi.auxstr);
  
  return;
}
