#ifndef UTIL_H
#define UTIL_H

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * Copyright (c) 1999 IBM Deutschland Entwicklung GmbH, IBM Corporation
 *                     Linux for s390 port by D.J. Barrow
 *                    <barrow_dj@mail.yahoo.com,djbarrow@de.ibm.com>            * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: util.h,v 1.2 2003/02/17 15:56:58 frassi Exp $
 */


#include "automa_management.h"
#include "xlat.h"
#include "comms.h" /* msg_building_info */
#include <sys/resource.h> /* per struct rusage */
/* resource.h serve per condividere le realprintusage e realprintsignal in signal.c, process.c, resource.c */


/******************
 * STRUCTS OFFSET *
 ******************/

#ifndef offsetof
#define offsetof(type, member)	(((char *) &(((type *) NULL)->member)) - \
				 ((char *) (type *) NULL))
#endif /* !offsetof */

/* get offset of member within a user struct */
#define uoff(member)	offsetof(struct user, member)

 



/************************************
 * MEMORY CLONING OF TRACED PROCESS *
 ************************************/

int upeek(int pid, long off, long *res);

int umoven(struct tcb *, long, int, char *);

#define umove(pid, addr, objp)	\
	umoven((pid), (addr), sizeof *(objp), (char *) (objp))

int umovestr(struct tcb *, long, int, char *);


int set_copy(struct tcb *,int,long,long *,int,char *);
/*  input: 1) una struttura tcb
           2) il numero dell'argomento da settare all'interno della struct tcb,
           3) l'indirizzo sorgente
	   4) il puntatore all'indirizzo di una struttura detinazione appropriata
	   5) la dimensione dell'area di memoria da allocare
	   6) una stringa con il nome della funzione chiamante

   output: OK_COPY in caso di sucesso, ERR_COPY altrimenti

   Tenta di allocare la quantita' di memoria richiesta, assegnandola alla struttura indicata
   dal chiamante, per poi copiarvi il contenuto dell'area di memoria sorgente. Infine assegna
   l'indirizzo destinazione all'elemento "n_arg" della struttura tcb e setta lo stato dell'argomento 
   a TAKEN */

int set_element(struct tcb *,long,long *,int,char *);
/*  input: 1) una struttura tcb
           2) il numero dell'argomento da settare all'interno della struct tcb,
           3) l'indirizzo sorgente
	   4) il puntatore all'indirizzo di una struttura detinazione appropriata
	   5) la dimensione dell'area di memoria da allocare
	   6) una stringa con il nome della funzione chiamante

   output: OK_COPY in caso di sucesso, ERR_COPY altrimenti

   Tenta di allocare la quantita' di memoria richiesta, assegnandola alla struttura indicata
   dal chiamante, per poi copiarvi il contenuto dell'area di memoria sorgente. Infine assegna
   l'indirizzo destinazione all'elemento "n_arg" della struttura tcb */


#define arg_alloc(tcp,n_arg) { \
                               long temp_l; \
                               temp_l = tcp->u_arg[n_arg]; \
                               (long*) tcp->u_arg[n_arg] = (long*) malloc(sizeof(long)); \
                               if (tcp->u_arg[n_arg] == 0) \
                                 { \
	                           fprintf(stderr, "arg_alloc: no memory\n"); \
	                           return -1; \
                                 } \
                               *((long*) tcp->u_arg[n_arg])=temp_l; \
                               tcp->arg_state[n_arg] = TAKEN; \
                             }
/* 
   Macro blocco 
   input: tcp      struct tcb,
          n_arg    numero argomento della syscall tcp->scno
   
   Clona il valore di tcp->u_arg[n_arg] in uno spazio di memoria proprio e ne salva l'indirizzo
   in tcp->u_arg[n_arg].
*/
       

/*******************
 * PRINT FUNCTIONS *
 *******************/

//extern int msg_buf_written_bytes;

#ifdef __STDC__
#ifdef __GNUC__
void tprintf(const char *fmt, ...)
	__attribute__ ((format (printf, 1, 2)));
#else
void tprintf(const char *fmt, ...);
#endif
#else
void tprintf();
#endif


void printxval(struct xlat *, int, char *); 

int addflags(struct xlat *, int); 

int printflags(struct xlat *, int); 

void string_quote(char *str); 

/*TODO: probabilmente va eliminata printstr, in quanto l'ho suddivisa nella get_str e nella realprintstr*/
/* void printstr(struct tcb *, long, int); NO */

void realprintrusage(msg_building_info * info, struct rusage * rup);

void realprintsignal(int *nr);

void realprintstr(unsigned char *,unsigned int);

void realprintnum(int*, char *);

/* void printnum(struct tcb *, long, char *); NO */

/* void printpath(struct tcb *, long); NO */

/* void printpathn(struct tcb *, long, int); NO */

/* void printmsgstr(long, int); NO */


/* void printleader(struct tcb *);  */

/* void printtrailer(struct tcb *);  */

int printargs(msg_building_info *info);

int print_retval(int format_ret, long u_rval, long u_error, char *auxstr);

void print_syscall_realtime();


/*void printsignal(int);*/

/* void fake_execve(struct tcb *tcp, char *program, char *argv[], char *envp[]); */



/***********************************
 * FORMAT OF SYSCALL RETURN VALUES *
 ***********************************/

#define RVAL_DECIMAL	000	/* decimal format */
#define RVAL_HEX	001	/* hex format */
#define RVAL_OCTAL	002	/* octal format */
#define RVAL_UDECIMAL	003	/* unsigned decimal format */
#define RVAL_MASK	007	/* mask for these values */

#define RVAL_STR	010	/* Print `auxstr' field after return val */
#define RVAL_NONE	020	/* Print nothing */


/***********************************************************
 * DECLARATIONS AND UTILITY FOR IO CONTROL DEVICE FUNCTION *
 ***********************************************************/

extern char *ioctl_lookup(long);
extern int ioctl_decode(msg_building_info *, long*, long*);
extern int term_ioctl(msg_building_info *, long*, long*);
extern int sock_ioctl(msg_building_info *, long*, long*);

struct ioctlent {
	char *doth;
	char *symbol;
	unsigned long code;
};

extern struct ioctlent *ioctlent; 

extern int nioctlents; 


extern struct ioctlent ioctlent0[];
extern int nioctlents0;


/********************************************************
 * DECLARATIONS AND UTILITY FOR SIGNAL TRACING FUNCTION *
 ********************************************************/

extern char *signalent0[];
extern int nsignals0;
extern char **signalent; 
extern int nsignals; 

char *signame(int);

#endif /* !UTIL_H */
