%{

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: wids.lex,v 1.2 2003/02/17 15:57:58 frassi Exp $
 */


#include <glib.h>
#include <assert.h>

#include "defs.h"

#include "parser.h"
#include "runtime_util.h"
#include "pars_funct.h"

#ifndef Y_TAB_H
#define Y_TAB_H

/* file header per wids.lex */
#include "y.tab.h"

#endif /* !Y_TAB_H */

  /* #define LEXDEBUG 1    */
 
  extern int yyerror(char *s, ...);
  void yyset(void);
  void stringset(void);
  
  static char *token; 
  int num_lines = 1;
  static int num_seq = 1;         /* mantiene il numero attuale della sequenza eventi */
  static int num_macro = 0;       /* mantiene il numero attuale della sequenza di macro */

  int num_macro_in_event = 0;  /* utilizzato per dare un numero alle variabili utilizzate dalle macro */
  int num_macro_in_macro = 0;
  
  /* indicazioni di contesto */
  int in_declarations = 1;    /* uguale a 1 se sono nella sezione dichiarazioni    */
  int in_macro        = 0;    /* uguale a 1 se sono in una macro                   */
  int sys_args        = 0;    /* uguale a 1 per gli argomenti di una syscall       */
  int sys_env         = 0;    /* uguale a 1 per gli argomenti, le condizioni e le
				 azioni relative ad una macro o di una syscall     */
  int comb_env        = 0;    /* uguale a 2 per le condizioni e le azioni relative
				 ad una combinazione di eventi                     */ 
  
  int counter = 0;            /* contatore incrementato ad ogni syscall, macro e combinazione di eventi */
  int counter_in_macro = 0;   /* contatore incrementato ad ogni syscall, macro e combinazione di eventi
				 all'interno di una definizione di macro */
  
  int once_var;   /* indica che la variabile corrente e' stata trovata per la prima volta
		     e quindi non si trovava in var_table */
  
  extern int in_flags;        /* uguale a 1 quando il parser esamina una lista di flag */ 

  /* TABELLE DEI SIMBOLI */
  extern var_table_t var_table;
  extern macro_var_table_t macro_var_table;
  
  extern label_table_t macro_label_table;  
  extern label_table_t label_table;
  
  extern macro_entry *macro_table;
  
%}

S               [ \t]*
COMMENT         \#[^\n]*
IDENTIFIER      [a-z_][a-zA-Z0-9_]*
CHARACTER       \'\\?[^\n\t]\'
/* REAL            ("+"|"-")?[0-9]+"."[0-9]*((e|E)[+-]?[0-9]+)? */
HEX             0(x|X)[0-9a-fA-F]+
OCT             0[0-7]+
DEC             ("+"|"-")?[0-9]+

%%


^include                   {
                           #ifdef LEXDEBUG
                           ECHO;
			   #endif
			   return(INCLUDE);
                           }

"string"                   {
	                   #ifdef LEXDEBUG
                           printf("string ");
			   #endif
                           return (TYPE_STRING);
                           }

"int"                      {
	                   #ifdef LEXDEBUG
                           printf("int ");
			   #endif
                           return (TYPE_INT);
                           }

"char"                     {
	                   #ifdef LEXDEBUG
                           printf("char ");
			   #endif
                           return (TYPE_CHAR);
                           }

"exit"                     {
	                   #ifdef LEXDEBUG
                           printf("exit ");
			   #endif
                           return (EXIT);
                           }

"intro"                    {
	                   #ifdef LEXDEBUG
                           printf("intro ");
			   #endif
                           return (INTRO);
                           }

"_"                        { 
                           #ifdef LEXDEBUG
                           printf(" _ ");
			   #endif
			   return (UNDEFINED);     
                           }

"begin"{S}/\(              {
	                   #ifdef LEXDEBUG
                           printf("begin ");
			   #endif
                           return (START);
                           }

{IDENTIFIER}/{S}\(         {
                           if (sys_env) 
                             {
			       /* mi aspetto una funzione */
			       /* controllare che la funzione sia presente nelle tabelle relative e restituire il puntatore */
			       #ifdef LEXDEBUG
			       printf("(funct:) '%s' ", yytext);
			       #endif
			       yyset();
			       return(FUNCTION);
			     }
			   else /* !sys_env */
			     {
			       /* mi aspetto una syscall o una macro */
			       int sysc_num = find_sysc_number(yytext);
			       sys_env = 1;
			       //sys_args = 1;
			       
			       if (sysc_num < 0)
				 { 
				   /* macro */ 
				   macro_entry *a_macro = getmacro_by_name(yytext, &macro_table);
				   if (a_macro == NULL)
				     {
				       if ((in_declarations) && (!in_macro))
					 {
					   char *macro_name = (char *) malloc(sizeof(char) * (strlen(yytext) + 1));
					   memcpy((char *) macro_name, yytext, strlen(yytext) + 1);
					   num_macro++;
					   a_macro = putmacro(macro_name, num_macro, &macro_table);
					 }
				       else /* !in_declarations */
					 {
					   char *error_message = (char *) malloc(sizeof(char) * 
						  (strlen("ERROR: 'X' macro undefined or syscall name wrong") + strlen(yytext) + 1));
					   sprintf(error_message, "ERROR: '%s' macro undefined or syscall name wrong", yytext);
					   yyerror(error_message);
					   free(error_message);
					   /* exit(1); */
					 }
				     }
				   
				   /* 
				      if (in_declarations) 
				      { 
				      if (in_macro)
				      {
				      counter_in_macro++; 
				      #ifdef LEXDEBUG 
				      printf(" (m_c:%d) ", counter_in_macro); 
				      #endif
				      }
				      } 
				      else 
				      { 
				      counter++; 
				      #ifdef LEXDEBUG
				      printf(" (c:%d) ", counter); 
				      #endif
				     }
				   */
				   
				   if (in_macro) num_macro_in_macro++;
				   if (!in_declarations) num_macro_in_event--;
				   
				   #ifdef LEXDEBUG
				   printf("(macro:) '%s' ", yytext);
				   #endif
				   yylval.macro_ptr = a_macro;
				   return(MACRO);				       
				 } 
			       else /* sysc_num > 0 */
				 {
				   /* syscall */ 
				   
				   sys_args = 1;
				   
				   #ifdef LEXDEBUG
				   printf("(sysc:) '%s' ", yytext);
				   #endif
				   
				   if (!in_declarations) { 
				     counter++; 
				     #ifdef LEXDEBUG 
				     printf(" (c:%d) ", counter); 
                                     #endif 
				   } 
				   else { 
				     counter_in_macro++; 
				     #ifdef LEXDEBUG
				     printf(" (m_c:%d) ", counter_in_macro); 
				     #endif
				   }
				   
				   yylval.integer = sysc_num;
				   return(SYSCALL);
				 }
			     }
			   
			   /*
			     se sys_env e' 1 allora mi devo aspettare una funzione 
			     se sys_env e' 0 allora mi devo aspettare una syscall 
			     oppure una macro e poi imposto sys_env = 1 
			   */
			   }  
			     

{IDENTIFIER}               { 
	                   label_value_t *v = NULL;
                           			   			   
                           #ifdef LEXDEBUG
                           printf("(id:) '%s' ", yytext);
			   #endif
			   			
			   if (in_declarations)
			     {
			       if (!in_macro)
				 {
				   /* dovrei essere nella dichiarazione di un prototipo di funzione */
				   int type_id = get_type(yytext);
				   
				   if (type_id == NOT_FOUND)
				     {
				       char *error_message = (char *) malloc(sizeof(char) * (strlen("ERROR: identifier 'X' unknown") + strlen(yytext) + 1));
				       sprintf(error_message, "ERROR: identifier '%s' unknown", yytext);
				       yyerror(error_message);
				       free(error_message);
				     }
				   
				   yylval.integer = type_id;
				   return(LABEL_TYPE);
				 }

			       if ((in_macro) && (counter_in_macro > 0))
				 {
				   if ((v = get_label(macro_label_table, yytext, -counter_in_macro)) == 0) 
				     /* se non entra nell'if allora e' una label gia' inser. in label_table */
				     { 
				       /* printf("\nDEBUG: sys_args: %d, comb_env %d", sys_args, comb_env); */
				       v = put_label(macro_label_table, yytext, -counter_in_macro);
				       if ( !sys_args && comb_env != 2 )  
					 {
					   char *error_message = (char *) malloc(sizeof(char) * 
						  (strlen("ERROR: label 'X' undefined") + strlen(yytext) + 1));
					   sprintf(error_message, "ERROR: label '%s' undefined", yytext);
					   yyerror(error_message);
					   free(error_message);
					   v->n_arg = -2; /* valore arbitrario che indica un errore */
					   /* exit(1); */
					 }
				     } 
				 }
			       
			     }
			   else if (counter > 0)
			     {
			       if ((v = get_label(label_table, yytext, counter)) == 0) 
				 /* se non entra nell'if allora e' una label gia' inser. in label_table */
				 { 
				   /* printf("\nDEBUG: sys_args: %d, comb_env %d", sys_args, comb_env); */
				   v = put_label(label_table, yytext, counter);
				   if ( !sys_args && comb_env != 2 )
				     {
				       char *error_message = (char *) malloc(sizeof(char) * 
						  (strlen("ERROR: label 'X' undefined") + strlen(yytext) + 1));
				       sprintf(error_message, "ERROR: label '%s' undefined", yytext);
				       yyerror(error_message);
				       free(error_message);
				       v->n_arg = -2; /* valore arbitrario che indica un errore */
				       /* exit(1); */
				     }
				 } 
			     }

			   yylval.label_ptr = v;
			   return (IDENTIFIER);
                           }

{CHARACTER}                          {
	                             char *c = (char *) malloc (sizeof(char));

                                     #ifdef LEXDEBUG
                                     ECHO;
				     #endif
				     
                                     if ((yytext[1] == '\\') && (yyleng == 4)) {
				       switch (yytext[2]) 
					 {
					 case 'a' : *c = '\a'; break;
					 case '\\': *c = '\\'; break;
					 case 'b' : *c = '\b'; break;
					 case 'r' : *c = '\r'; break;
					 case '"' : *c = '\"'; break;
					 case 'f' : *c = '\f'; break;
					 case 't' : *c = '\t'; break;
					 case 'n' : *c = '\n'; break;
					 case '0' : *c = '\0'; break;
					 case '\'': *c = '\''; break;
					 case 'v' : *c = '\v'; break;
					 default :
					   {
					     printf("\nERROR in character\n");
					     exit(1);
					     break;
					   }
					 }
                                     }
				     else *c = (char) yytext[1];
				     
				     yylval.character_ptr = c;
				     return (CHARACTER_VALUE);
                                     }

{HEX}                                {
	                             long *l = (long *) malloc(sizeof(long));

                                     #ifdef LEXDEBUG
                                     ECHO;
				     #endif
                                     *l = strtol(yytext,NULL,16);
				     
				     yylval.integer_ptr = l;
				     return (LONG_VALUE);
                                     }

{OCT}                                {
	                             long *l = (long *) malloc(sizeof(long));

                                     #ifdef LEXDEBUG
                                     ECHO;
				     #endif
                                     *l = strtol(yytext,NULL,8);

				     yylval.integer_ptr = l;
				     return (LONG_VALUE);
                                     }

{DEC}                                {
	                             long *l = (long *) malloc(sizeof(long));

                                     #ifdef LEXDEBUG
                                     ECHO;
				     #endif
                                     *l = atoi(yytext);

				     yylval.integer_ptr = l;
				     return (LONG_VALUE);
                                     }    

\"[^\n\"]*                  {   
                              if (yytext[yyleng - 1] == '\\') yymore();
			      else
				{
				  #ifdef LEXDEBUG
				  ECHO;
                                  #endif
				  
				  if (input() != '"') 
				    { 
				      char *error_message = (char *) malloc(sizeof(char) * 
						(strlen("ERROR: parse error in string 'X'") + strlen(yytext) + 1));
				      sprintf(error_message, "ERROR: parse error in string '%s'", yytext);
				      yyerror(error_message);
				      free(error_message);
				      /* exit(1); */
				    }
				  yytext++; /* non includo il " iniziale */
				  stringset();
				  return (STRING_VALUE); 
				}
                            }

[A-Z_]+                              {
	                             #ifdef LEXDEBUG
                                     ECHO;
                                     #endif
                                     yyset();
				     return (FLAG);
                                     }    

\${IDENTIFIER}             { 
	                   var_entry *v;
                           char *token;  /* ci allochero' lo spazio che contiene il nome della var. */

                           #ifdef LEXDEBUG
                           printf("(var:) %s ", yytext);
			   #endif
                           yytext++;
                           
                           once_var = 0;
			   if ((v = getvar(var_table, yytext)) == NULL)    /* se non entra nell'if allora e' una
									  var globale gia' inser. in var_table */
                             { 
			       if (in_macro) 
				 {
				   token = (char *) malloc( strlen(yytext)+1);
				   memcpy((char *) token, yytext, strlen(yytext)+1);
				   if ((v = getmacrovar(macro_var_table, token, num_macro)) == NULL) 
				     /* se non entra nell'if allora e' una var gia' inser. in macro_var_table */
				     { 
				       v = (var_entry *) malloc(sizeof(var_entry));
				       putmacrovar(macro_var_table, token, num_macro, v);
				       printf("\nDEBUG putmacrovar %s, %d", token, num_macro);
				       v->name = token;
				       v->type_var = scalar;
				       v->type_values = 0;
				       v->values = NULL;
				       once_var = 1;
				     } 
				   else free(token);
				 }
			       else
				 {
				   if (in_declarations) {
				     token = (char *) malloc( strlen(yytext)+1);
				     memcpy((char *) token, yytext, strlen(yytext)+1);
				   }
				   else {
				     char num[5]; /* non credo si arrivi a un numero di seq. con + di 4 cifre... */
				     
				     snprintf(num, 5, "%d", num_seq);
				     token = (char *) malloc( strlen(num) + strlen(yytext)+1);
				     memcpy((char *) token, num, strlen(num));
				     memcpy((char *) token+strlen(num), yytext, strlen(yytext)+1);
				   }
				   
				   if ((v = getvar(var_table, token)) == NULL)  { /* se == 0 => non e' nella sym. table
										   altr. e' una locale gia' ins. in var_table */
				     v = (var_entry *) malloc(sizeof(var_entry));
				     putvar(var_table, token, v);
				     v->name = token;
				     v->type_var = scalar;
				     v->type_values = 0;
				     v->values = NULL;
				     once_var = 1;
				   } else free(token);
				 }
			     }
			   
			   yylval.var_ptr = v;
                           return (VARIABLE);
                           }

\${IDENTIFIER}\{\}         { 
	                   var_entry *v;
                           char *token;  /* ci allochero' lo spazio che contiene il nome della var. */

                           #ifdef LEXDEBUG
                           printf("(hashVar:) %s ", yytext);
			   #endif
                           yytext++;
                           
			   once_var = 0;
			   if ((v = getvar(var_table, yytext)) == NULL) /* se non entra nell'if allora e' una
									  var globale gia' inser. in var_table */
                             { 
			       if (in_macro) 
				 {
				   token = (char *) malloc( strlen(yytext)+1);
				   memcpy((char *) token, yytext, strlen(yytext)+1);
				   if ((v = getmacrovar(macro_var_table, token, num_macro)) == NULL) 
				     /* se non entra nell'if allora e' una var gia' inser. in macro_var_table */
				     { 
				       v = (var_entry *) malloc(sizeof(var_entry));
				       putmacrovar(macro_var_table, token, num_macro, v);
				       printf("\nDEBUG putmacrovar %s, %d", token, num_macro);
				       v->name = token;
				       v->type_var = hash;
				       v->type_values = 0;
				       v->values = NULL;
				       once_var = 1;
				     } 
				   else free(token);
				 }
			       else
				 {
				   if (in_declarations) {
				     token = (char *) malloc( strlen(yytext)+1);
				     memcpy((char *) token, yytext, strlen(yytext)+1);
				   }
				   else {
				     char num[5]; /* non credo si arrivi a un numero di seq. con + di 4 cifre... */
				     
				     snprintf(num, 5, "%d", num_seq);
				     token = (char *) malloc( strlen(num) + strlen(yytext)+1);
				     memcpy((char *) token, num, strlen(num));
				     memcpy((char *) token+strlen(num), yytext, strlen(yytext)+1);
				   }
				   
				   if ((v = getvar(var_table, token)) == NULL)  { /* se == 0 => non e' nella sym. table
										  altr. e' una locale gia' ins. in var_table */
				     v = (var_entry *) malloc(sizeof(var_entry));
				     putvar(var_table, token, v);
				     v->name = token;
				     v->type_var = hash;
				     v->type_values = 0;
				     v->values = NULL;
				     once_var = 1;
				   } 
				   else free(token);
				 }
			     }
			   
			   yylval.var_ptr = v;
			   return (HASH_VAR);
                           }

"||"                       { 
                           #ifdef LEXDEBUG
                           ECHO;
                           #endif 
			   sys_env = 0;
			   comb_env = 0;
			   return (OR);     
                           }

";"                        {
                           #ifdef LEXDEBUG
                           ECHO;
                           #endif
                           sys_env = 0;
			   comb_env = 0;
                           return(';');
                           }

"."                        { 
                           #ifdef LEXDEBUG
                           ECHO;
                           #endif 
                           
			   sys_env = 0;
			   comb_env = 0;
			   ++num_seq;
			   
                           return ('.');     
                           }

":="                       { 
                           #ifdef LEXDEBUG
                           printf(" := ");
                           #endif 
                           
			   sys_env = 0;
			   in_macro = 1;
			   /* creo una symbol table per ogni macro che trovo nelle dichiarazioni */
			                              
			   return (MACRO_ASSIGN);     
                           }

"%%"                       {
                           #ifdef LEXDEBUG
                           ECHO;
                           #endif 
                           in_declarations = 0;
                           return (DECL_SEP);  
                           }


@                          {
	                   #ifdef LEXDEBUG
                           printf(" any ");
			   #endif
			   
			   sys_env = 1;
			   
			   if (!in_declarations) { 
			     counter++;  
			     #ifdef LEXDEBUG
			     printf(" (c:%d) ", counter);
			     #endif
			   }
			   else { 
			     counter_in_macro++; 
			     #ifdef LEXDEBUG
			     printf(" (m_c:%d) ", counter_in_macro);
			     #endif
			   }
			   
                           return ('@');
                           } 


"|"                        {
                           #ifdef LEXDEBUG
                           ECHO;
			   #endif
			   
			   if (!in_flags) /* entro nelle condizioni */
			     {
			       sys_args = 0;
			       comb_env += comb_env ? 1 : 0;
			       
			       /* printf("\nDEBUG: comb_env %d", comb_env); */
			       if (comb_env == 2) 
				 {
				   if (in_macro) 
				     { 
				       counter_in_macro++;
                                       #ifdef LEXDEBUG
				       printf(" (m_c:%d) ", counter_in_macro);
                                       #endif
				     }
				   else
				     {
				       counter++;
                                       #ifdef LEXDEBUG
				       printf(" (c:%d) ", counter);
				       #endif
				     }
				 }
			     }
			   
			   return(yytext[0]);
                           }

"}"                        { 
                           #ifdef LEXDEBUG
                           ECHO;
			   #endif

                           comb_env += !comb_env ? 1 : 0;
			   /* printf("\nDEBUG: comb_env %d", comb_env); */

			   return(yytext[0]);
                           }


"<"                    {
                           #ifdef LEXDEBUG
                           ECHO;
			   #endif

			   sys_args = 0;
			   comb_env += comb_env == 1 ? 1 : 0;
			   /* printf("\nDEBUG: comb_env %d", comb_env); */
			   if (comb_env == 2) 
			     {
			       if (in_macro) 
				 { 
				   counter_in_macro++;
				   #ifdef LEXDEBUG
				   printf(" (m_c:%d) ", counter_in_macro);
				   #endif
				 }
			       else
				 {
				   counter++;
				   #ifdef LEXDEBUG
				   printf(" (c:%d) ", counter);
				   #endif
				 }
			     }
			   
			   return(yytext[0]);
                         }




                        

")"                        {
                           #ifdef LEXDEBUG
                           ECHO;
                           #endif
			   /* sys_args = 0; */ 
			   return(yytext[0]);
                           }

"{"|":"|"?"|"-"|"="|","|"("|"+"|"["|"]"|"!"|"*"|"&"|">" {
                           #ifdef LEXDEBUG
                           ECHO;
                           #endif
                           return(yytext[0]);
                           }

{COMMENT}      ;

[ \t]          {
               #ifdef LEXDEBUG
               ECHO;
               #endif
               }

.              {
                 char *error_message = (char *) malloc(sizeof(char) * 
				 (strlen("ERROR: token 'X' unknown") + strlen(yytext) + 1));
		 sprintf(error_message, "ERROR: token '%s' unknown", yytext);
		 yyerror(error_message);
		 free(error_message);
               }     

\n             {
               #ifdef LEXDEBUG
               printf("\n\n");
               #endif
               ++num_lines;
	       }

%% 


void
yyset(void) 
{ 
  token = (char *) malloc( strlen(yytext)+1);
  memcpy((char *) token, yytext, strlen(yytext)+1);
  
  yylval.string = token;
}

void
stringset(void)
{ 
  int i;
  char *temp = (char *) malloc( strlen(yytext)+1);
  
  for (i = 0; i <= strlen(yytext); i++)
    {
      if (yytext[i] == '\"') { printf("\nERROR in string\n"); exit(1); }
      else if (yytext[i] != '\\') temp[i] = yytext[i];
      else
	{
	  yytext++;
	  
	  switch (yytext[i]) 
	    {
	    case 'a' : temp[i] = '\a'; break;
	    case '\\': temp[i] = '\\'; break;
	    case 'b' : temp[i] = '\b'; break;
	    case 'r' : temp[i] = '\r'; break;
	    case '"' : temp[i] = '\"'; break;
	    case 'f' : temp[i] = '\f'; break;
	    case 't' : temp[i] = '\t'; break;
	    case 'n' : temp[i] = '\n'; break;
	    case '0' : temp[i] = '\0'; break;
	    case '\'': temp[i] = '\''; break;
	    case 'v' : temp[i] = '\v'; break;
	    default :
	      {
		printf("\nERROR in string\n");
		exit(1);
		break;
	      }
	    }
	}
    }
  
  token = (char *) malloc( strlen(temp)+1);
  memcpy((char *) token, temp, strlen(temp)+1);
  free(temp);
  
  yylval.string = token;
}

/* end */
