%{

/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: wids.y,v 1.1.1.1 2001/12/17 08:58:56 frassi Exp $
 */


#include <glib.h>
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <ctype.h>
  
#include "defs.h"
  
#include "parser.h"
#include "follow.h"
#include "runtime_util.h"
#include "flags.h"

  extern int find_type(int sysc_n, int arg_n);  
  extern int get_type(char *type_name);

  static funct_entry *current_prototype; /* all'interno di una definizione di prototipo la struttura correntemente usata */
  static GList *templist;
  static GList *function_stack = NULL;   /* pila contenente in prima posizione il GNode corrispondente alla 
					    funzione di cui si sta effettuando il parsing */  

  static GList *unverified_var_stack = NULL; /* pila contenente gli handler delle variabili da verificare
                                                (contiene in ordine handler in exiting, un elemento vuoto, handler in intro) */
  
  static GList *any_sysc = NULL;         /* lista delle syscall che compaiono nel file delle specifiche */
  static GList *minus_sysc = NULL;       /* lista di syscall correntemente negate (event_operations: '!' event) */
  
  static int sysc_num;                   /* numero della syscall corrente */
  static int arg_number = 0;             /* numero dell'argomento correntemente esaminato della chiamata di sistema n. sysc_num */
  
  static funct_entry *fe_ptr[8];         /* array che contiene il puntatore alla funct_entry della function corrente 
					    e di tutte le sue annidate */
  static int func_arg_number[8];         /* array che contiene il numero dell'argomento correntemente esaminato della 
					    function corrente e di tutte le sue annidate */
  static int num_func = 0;               /* numero sequenziale per contare le function in azioni o condizioni es. se ho 
					    func1(..., func2(),...)   ==>  num_func == 2 */
  
  static int out_sysc;                   /* flag per sapere se c'e' un segnaposto al valore di ritorno di una syscall */
  
  static int num_event = 0;              /* memorizza quanti eventi semplici sono nel file */
  
  static int in_intro_statements = 0;    /* uguale a 1 nelle istruzioni da eseguire all'entrata di una syscall */
  static int in_exit_statements  = 0;    /* uguale a 1 nelle istruzioni da eseguire all'uscita di un syscall */
  static int num_statements = 0;         /* contatore per la sequenza di istruzioni in entrata o in uscita a una syscall */ 
  
  static int num_seq = 1;                /* contatore che conserva il numero della regola correntemente esaminata */
  static int num_macro = 0;              /* contatore che conserva il numero della definizione di macro correntemente esaminata */
		
  
  /* variabili dichiarate in wids.lex */
  extern int num_macro_in_event;         /* contatore che conserva il numero della macro correntmemente esaminata 
					    trovata nella sequenza di eventi */
  extern int num_macro_in_macro;         /* contatore che conserva il numero della macro correntemente esaminata 
					    trovata nella definizione di macro corrente */
  extern int num_lines;                  /* numero della linea corrente nel file delle specifiche */
  
  /* TABELLE DEI SIMBOLI */
  extern var_table_t var_table;          /* tabella contenente le variabili globali e le variabili locali 
					    alle sequenze nel file delle specifiche (escluse le variabili locali alle macro) */
  extern macro_var_table_t macro_var_table;   /* tabella contenente le variabili locali alle macro */

  extern label_table_t label_table;       /* tabella contenente i segnaposti definiti per ciascun evento 
					     o combinazione di eventi di una regola */
  extern label_table_t macro_label_table; /* tabella contenente i segnaposti definiti per ciascun evento o combinazione
					     di eventi appartenente ad una macro */

  extern macro_entry *macro_table;       /* tabella contenente le macro definite nella sezione delle dichiarazioni */
  
  
  /* variabili di contesto */
  extern int in_declarations;            /* indica se il parser sta esaminando la sezione delle dichiarazioni del file 
					    delle specifiche */
  extern int in_macro;                   /* indica se il parser sta esaminando una macro */
  extern int counter;                    /* conta gli eventi e le combinazioni di eventi nelle sequenze */
  extern int counter_in_macro;           /* conta gli eventi e le combinazioni di eventi nelle definizioni di macro nella sezione
					    delle dichiarazioni */
  extern int once_var;                   /* indica se la variabile compare per la prima volta nel file delle specifiche */
  extern int sys_env;                    /* uguale a 1 per gli argomenti, le condizioni e le azioni relative ad una macro o di una syscall */
  extern int comb_env;                   /* uguale a 2 per le condizioni e le azioni relative ad una combinazione di eventi */

  int in_flags = 0;                      /* uguale a 1 se il parser sta esaminando una lista di flag in OR */
  
  /* var dichiarata in main.c */ 
  extern syntax_tree events_tree;


  extern char *get_sysc_name(int num_sysc);

  extern int is_input(int sysc_num, int arg_number);

  extern int is_output(int sysc_num, int arg_nuber);

  extern int get_nargs(int num_sysc);

  /***************************************
   * funzioni di supporto per il parsing *
   ***************************************/
  
  void finalize_tree(GNode *tree);
  /* 1. crea nella var globale any_sysc la lista con tutti i tipi di syscall trovati in tree 
     2. sostituisce all'ANY l'OR delle syscall trovate nel file
     3. aggiorno in ogni simple_event il numero dell'evento */
  
  GNode *make_tree_for_arg_value(container *cont, int arg_number);
  /* crea una condizione per ciascun argomento di una chiamata di sistema */

  int min_num_event(GNode *tree);
  /* restituisce il numero evento minimo trovato in tree (tree e' l'evento di sinistra di   event OR event) */

  int max_num_event(GNode *tree);
  /* restituisce il numero evento massimo trovato in tree (tree e' l'evento di sinistra di   event OR event) */

  void check_vars_in_or_combination(GNode *tree, int min, int max);
  /* funzione utilizzata sull'evento di destra di   event OR event 
     per controllare che non vi compaiano variabili non definite */

  void modify_macro_num_events(GNode *macro);
  /* rinumera gli eventi contenuti nell'albero di una macro passatogli:
     se siamo nella dichiarazione di una macro li rinumera a partire da -(counter_in_macro + 1)
     e counter_in_macro assume il valore assoluto dell'ultimo evento, 
     altimenti li rinumera a partire da counter + 1, e counter assume il valore dell'ultimo evento */
  
  int check_event_for_not(GNode *node, gpointer data);
  /* controlla che il nodo correntemente esaminato faccia parte di un albero costituito da un OR di eventi */

  GNode *make_any_for_not(GList *l);
  /* crea una coppia {intro; exiting} per l'evento speciale any 
     e restituisce il puntatore alla radice dell'albero  */

  GNode *get_negated_event(GNode *tree);
  /* isola il primo nodo contenente una coppia di simple event {intro; exiting} dell'abero passatogli 
     e lo restituisce negandone le condizioni se esistono; per ottenere i nodi successivi riutilizzare
     la funzione get_negated_event() con argomento uguale a NULL */

  gboolean rename_var_handler(GNode *node, gpointer data);
  /* rinomina la variabile eventualmente contenuta nel nodo correntemente esaminato */

  
  /*********************************************
   * Gestione degli errori durante il parsing  *
   *********************************************/

  int num_error = 0;          /* numero di errori individuati */
  int error_local = 0;        /* flag che indica l'individuazione di un errore */

  int yyerror(char *s, ...);

  
%}

%union {
  
  /* tipi base */
  int integer;
  long *integer_ptr;
  char *character_ptr;
  char *string;
  cond_funct_ptr (*relational_func)();
  GHashTable *hash;

  /* elementi delle tabelle dei simboli */
  var_entry *var_ptr;
  macro_entry *macro_ptr;
  label_value_t *label_ptr;
  funct_entry *funct_ptr;

  /* elementi del parse tree */
  handler_t *handler_ptr;
  container *container_ptr;
  simple_event_t *simple_event_ptr;
  GList *list;
  GList *tlist[2];
  syntax_tree parse_tree;
}

%token                  INTRO    EXIT     UNDEFINED     OR     START   TYPE_STRING  TYPE_CHAR  TYPE_INT
%token                  BRK_LOCK   BY   FROM   TO       MACRO_ASSIGN   DECL_SEP     INCLUDE
%token <integer>        SYSCALL    LABEL_TYPE
%token <integer_ptr>    LONG_VALUE
%token <character_ptr>  CHARACTER_VALUE   
%token <string>         STRING_VALUE 
%token <string>         FLAG
%token <string>         FUNCTION
%token <var_ptr>        VARIABLE      HASH_VAR
%token <macro_ptr>      MACRO
%token <label_ptr>      IDENTIFIER


%type <integer>             name_syscall     type_var     return_value      flags
%type <container_ptr>       an_integer       array     value     variable2     right_hash_value
%type <handler_ptr>         variable
%type <simple_event_ptr>    syscall     syscall_args 
%type <list>                characters     strings     integers     args     conditions     func_args   
%type <list>                statement_sequence    intro_conditions  exit_conditions   intro_statements  exit_statements
%type <parse_tree>          arg_value     arg_label     arg     assignment     condition     function    
%type <parse_tree>          function_name     macro     func_arg     syscalls     simple_event
%type <parse_tree>          statement     event     event_operations     events     attack_sequences  
%type <parse_tree>          left_el       right_el  relational_operand
%type <tlist>               event_condition         statements
%type <relational_func>     relational_operator
%type <hash> 		    more_strings        more_integers      more_characters
%type <var_ptr>		    var_declaration
%type <macro_ptr>           macro_declaration   macro_name
%type <funct_ptr>           protoname

%left  ';'
%left  OR
%left  '*'
%right '!'
%left  '|'
%left  '<'
%left  ','   /* se lo tolgo mi da 8 shift/reduce in array */


%start specification_file


%%


specification_file  :    declarations  DECL_SEP  attack_sequences 
			 {
			   /* dealloco macro_label_table e label_table */			     
			   free_label_table(label_table);
			   free_label_table(macro_label_table);

			   /* dealloco macro_var_table e macro_table */
			   free_macro_table(macro_table);
			   free_var_table(macro_var_table);
			   
			   finalize_tree($3);
			   
			   //printf("\n\n*****************************************************************************\n");
			   //printf("\nnel file di conf. ci sono %d eventi (compresi gli END MARKER)\n\n",num_event);
			   //printf("===============PARSE TREE===============\n");
			   //print_parse_tree($3);
			   //print_var_table(var_table);
			 
			   events_tree = $3;
			 }
		    
                    |    DECL_SEP  attack_sequences
			 {
			   /* dealloco macro_label_table e label_table */			     
			   free_label_table(label_table);
			   free_label_table(macro_label_table);

			   /* dealloco macro_var_table e macro_table */
			   free_macro_table(macro_table);
			   free_var_table(macro_var_table);
			   
			   finalize_tree($2);
			   
			   //printf("\nnel file di conf. ci sono %d eventi (compresi gli END MARKER)\n\n",num_event);
			   //print_parse_tree($2);
			   //print_var_table(var_table);

			   events_tree = $2;
			 }

                    |    error DECL_SEP attack_sequences
                         {
			   yyerror("ERROR: parse error in declarations"); 

			   /* dealloco macro_label_table e label_table */			     
			   free_label_table(label_table);
			   free_label_table(macro_label_table);

			   /* dealloco macro_var_table e macro_table */
			   free_macro_table(macro_table);
			   free_var_table(macro_var_table);

			   free_parse_tree($3);
			 } 
		    ;

/***************************** DICHIARAZIONI ********************************/

declarations        :     prototype_declaration
                    |     var_declaration                     
                          { comb_env = 0; }
		    |     macro_declaration                   
                          { comb_env = 0; }
		    |     declarations  var_declaration       
                          { comb_env = 0; }
		    |     declarations  macro_declaration     
                          { comb_env = 0; }
                    |     declarations  prototype_declaration 
                          { comb_env = 0; }
                    |     declarations  error              
                          {
			    yyerror("ERROR: parse error in declarations");
			    comb_env = 0; 
			  }
                    ;


prototype_declaration :   INCLUDE STRING_VALUE protoname '(' args_types ')'
                          {
			    /* in STRING_VALUE il nome del file shared object (.so) 
			       che contiene la definizione della funzione indicata nel prototipo */
			    module_entry *me;
			    char *error;
			    
			    if (!module_table)
			      yyerror("ERROR: user defined function unsupported");
			    else
			      {
				if ((me = getmodule(module_table, $2)) == NULL)
				  {
				    /* carica modulo */
				    GModule *libhandle = g_module_open($2, 0);
				    if (libhandle == NULL || ((error = g_module_error()) != NULL))
				      {
					yyerror("ERROR: module not found");
				      }
				    else 
				      {
					me = (module_entry *) malloc(sizeof(module_entry));
					me->so_filename = $2;
					me->libhandle = libhandle;
				      }
				  }
			      
				
				if ( !(me->libhandle && g_module_symbol(me->libhandle, $3->name, (void *)&($3->ptr)) ) )
				  {
				    yyerror("ERROR: function name not found");
				  }
			      }
			    
			    current_prototype = NULL;
			  }
                      ;

protoname             :   type_var FUNCTION 
                          {
			    funct_entry *fe = (funct_entry *) malloc(sizeof(funct_entry));

			    fe->name = $2;
			    fe->so_filename = NULL;
			    fe->args_num = 0;
			    fe->ptr = NULL;
			    
			    fe->ret_type = $1;
			    /* 
			    if ( (fe->ret_type = get_type($1)) == NOT_FOUND)
			      {
				yyerror("ERROR: type unknown");
			      }
			    */
			    
			    if (getfunct(funct_table, fe->name) != NULL)
			      {
				yyerror("WARNING: function redefined");
			      }
			    else
			      putfunct(funct_table, fe->name, fe);
			    
			    $$ = current_prototype = fe;
                          }
                      ;

args_types            :   LABEL_TYPE
                          {
			    current_prototype->args_num++;
			    current_prototype->args_type[current_prototype->args_num - 1] = $1;
			  }

                      |   args_types ',' LABEL_TYPE 
                          {
			    current_prototype->args_num++;
			    current_prototype->args_type[current_prototype->args_num - 1] = $3;
                          }
                      ;


/* TODO: gestire in modo corretto i flag */

var_declaration     :     type_var  VARIABLE   '='  value
			  { 
			    scope_el_t *scope;
			    
			    if (($1 == 0) && ($2->type_values == 0))
			      {
			        yyerror("ERROR: var %s without type", $2->name);
				exit(-1);
			      }
			    else if (($1 != 0) && ($2->type_values != 0))
			      {
				yyerror("ERROR: var %s already declared", $2->name);
				exit(-1);
			      }
			    else if ($1 != 0)
			      $2->type_values = $1;
			    
			    if ($1 != $4->type_value) 
			      {
				yyerror("ERROR: type mismatch in var %s assignment", $2->name);
				YYERROR; 				     
			      }

			    /* dealloco l'eventuale valore della prec. dichiarazione */
			    if ($2->values != NULL) free($2->values);
			    
			    $2->values = $4->content.value;
			    free($4);
			    
			    /* inserisco la variabile nella tabella delle visibilita' */
			    if ( firstscope($2->name, 0) == NULL )
			      {
				/* la variabile non e' stata definita precedentemente */
				scope = putscope($2->name, 0);
				scope->num_event = 0;
			      }
			    
			    $$ = $2;
			  }
		    
                    |     type_var  HASH_VAR   '='   right_hash_value
			  {
			    scope_el_t *scope;
			    
			    if (($1 == 0) && ($2->type_values == 0))
			      {
			        yyerror("ERROR: var %s without type", $2->name);
				exit(-1);
			      }
			    else if (($1 != 0) && ($2->type_values != 0))
			      {
				yyerror("ERROR: var %s already declared", $2->name);
				exit(-1);
			      }
			    else if ($1 != 0) 
			      $2->type_values = $1; 
			    
			    if (($1 != $4->type_value ) && ($4->content.hash_value != NULL))  
			      {
				yyerror("ERROR: type mismatch in var %s assignment", $2->name);
				YYERROR; 				     
			      }
			    
			    /* se right_hash_value == { } alloco una hashtable vuota */
			    if ($4->content.hash_value == NULL)
			      {
				$4->type_value = $1;
				
				switch ($1) {
				case STRING_VALUE:
				  $4->content.hash_value = g_hash_table_new(g_str_hash, g_str_equal);
				  break;
				case CHARACTER_VALUE: case LONG_VALUE:
				  $4->content.hash_value = g_hash_table_new(g_int_hash, g_int_equal);
				  break;
				default:
				  yyerror("ERROR: unknown hash value in var %s assignment", $2->name);
				  $4->content.hash_value = g_hash_table_new(NULL, NULL);
				  break;
				}
			      }
			    
			    /* dealloco l'eventuale valore della prec. dichiarazione */
			    if ($2->values != NULL) free_hash_table((GHashTable *)$2->values);
			    
			    $2->values = $4->content.hash_value;
			    free($4);
			    
			    if ( firstscope($2->name, 0) == NULL )
			      {
				scope = putscope($2->name, 0);
				scope->num_event = 0;
			      }
			    
			    $$ = $2;
			  }
                   
                    |     error '=' value
                          {
			    yyerror("ERROR: parse error in lvalue of initialization");
                          }     

                    |     error '=' right_hash_value
                          {
			    yyerror("ERROR: parse error in hash lvalue of initialization");
			  }

                    |     type_var HASH_VAR '=' error
                          {
			    scope_el_t *scope;
			    
			    $2->type_var = hash;
			    $2->type_values = $1; 
			  			  			    
			    if ( firstscope($2->name, 0) == NULL )
			      {
				scope = putscope($2->name, 0);
				scope->num_event = 0;
			      }
			    
			    $2->values = NULL;
			    $$ = $2;

			    yyerror("ERROR: parser error in hash rvalue of var %s assignment", $2->name);
                          }

                    |     type_var VARIABLE '=' error 
                          {
			    scope_el_t *scope;

			    $2->type_var = scalar;
			    $2->type_values = $1;

			    $2->values = NULL;
			    
			    /* inserisco la variabile nella tabella delle visibilita' */
			    if ( firstscope($2->name, 0) == NULL )
			      {
				/* la variabile non e' stata definita precedentemente */
				scope = putscope($2->name, 0);
				scope->num_event = 0;
			      }
			    
			    $$ = $2;
			    
			    yyerror("ERROR: parser error in rvalue of var %s assignment", $2->name);
			  }
                    ;


type_var            :     TYPE_STRING
			  {
			    $$ = STRING_VALUE;
			  }
		    
                    |     TYPE_CHAR
			  {
			    $$ = CHARACTER_VALUE;
			  }
		    
                    |     TYPE_INT
			  {
			    $$ = LONG_VALUE;  /* INTEGER */
			  }

                    |     /* il tipo non e' specificato */
		          {
			    $$ = 0;
			  }
		    ;


value               :     STRING_VALUE                        
                          {
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = SIMPLE_VALUE;
			    $$->type_value = STRING_VALUE;
			    $$->content.value = (char *) $1;
			  }
		    
                    |     CHARACTER_VALUE			
                          {
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = SIMPLE_VALUE;
			    $$->type_value = CHARACTER_VALUE;
			    $$->content.value = (char *) $1;
			  }
		    
                    |     an_integer			
                          {
			    $$ = $1;
			  }
		    ; 


right_hash_value    :    '{' value '}'
			  { 
			    GHashTable *hash_val;
			    
			    switch ($2->type_value) {
			    case STRING_VALUE:
			      /* ATTENZIONE: CAMBIARE LA FUNZIONE G_STR_HASH CON UNA PIU' ADATTA */
			      hash_val = g_hash_table_new(g_str_hash, g_str_equal);
			      break;
			    case CHARACTER_VALUE: case LONG_VALUE:
			      hash_val = g_hash_table_new(g_int_hash, g_int_equal);
			      break;
			    default:
			      yyerror("ERROR: unknown hash value");
			      hash_val = g_hash_table_new(NULL, NULL);
			    }
			    
			    g_hash_table_insert(hash_val, $2->content.value, $2->content.value);
			    $2->type_content = HASH_VALUE;
			    $2->content.hash_value = hash_val;
			    $$ = $2;
			  }
		    
                    |     '{' more_strings '}'
			  {
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HASH_VALUE;
			    $$->type_value = STRING_VALUE;
			    $$->content.hash_value = $2;
			  }
		    
                    |     '{' more_characters '}'
			  {
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HASH_VALUE;
			    $$->type_value = CHARACTER_VALUE;
			    $$->content.hash_value = $2;
			  }
		    
                    |     '{' more_integers '}'
			  {
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HASH_VALUE;
			    $$->type_value = LONG_VALUE;
			    $$->content.hash_value = $2;
			  }
		    
                    |     '{' '}'
			  { 
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HASH_VALUE;
			    $$->type_value = 0;
			    $$->content.hash_value = NULL; 
			    /* equivale alla tabella hash vuota: non posso creare una tabella hash 
			       finche' non conosco il tipo dei valori che vi andro' a mettere */
			  }

                    |     '{' error '}'
                          {
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HASH_VALUE;
			    $$->content.hash_value = NULL; 
			    
			    yyerror("ERROR: parse error in hash rvalue");
			  }
		    ;     
		

more_strings        :     STRING_VALUE        ','  STRING_VALUE
			  {
			    $$ = g_hash_table_new(g_str_hash, g_str_equal);
			    g_hash_table_insert($$, $1, $1);
			    g_hash_table_insert($$, $3, $3);
			  }
		    
                    |     more_strings  ','  STRING_VALUE
			  {
			    g_hash_table_insert($1, $3, $3);
			  }

                    |     more_strings ',' error
                          {
			    yyerror("ERROR: parse error in a element of hash rvalue");
			  }   
		    ;


more_characters     :     CHARACTER_VALUE        ','  CHARACTER_VALUE
			  {
			    $$ = g_hash_table_new(g_int_hash, g_int_equal);
			    g_hash_table_insert($$, $1, $1);
			    g_hash_table_insert($$, $3, $3);
			  }
		    
                    |     more_characters  ','  CHARACTER_VALUE
			  {
			    g_hash_table_insert($1, $3, $3);
			  }

                    |     more_characters ',' error
                          {
			    yyerror("ERROR: parse error in a element of hash rvalue of initialization");
			  }   
		    ;


more_integers       :     an_integer     ','  an_integer
			  {
			    $$ = g_hash_table_new(g_int_hash, g_int_equal);
			    g_hash_table_insert($$, $1->content.value, $1->content.value);
			    g_hash_table_insert($$, $3->content.value, $3->content.value);
			    free($1);
			    free($3);
			  }
		    
                    |     more_integers  ','  an_integer
			  {
			    g_hash_table_insert($1, $3->content.value, $3->content.value);
			    free($3);
			  }

                    |     more_integers ',' error
		          {
			    yyerror("ERROR: parse error in a element of hash rvalue");
			  }
		    ;

/* quando si crea una macro occorre specificare le variabili globali o locali
   utilizzate e gli assegnamenti */

macro_name          :     MACRO '(' ')'
                          {
			    num_macro++;
			    $$ = $1;
			  }

/* NOTA sulle macro:
   per i controlli che vengono eseguiti una macro non pu contenere piu' di 10000 eventi 
   (contando le combinazioni e gli eventi al loro interno) (c.a.) e non e' possibile innestare 
   una macro dentro l'altra superando i 400000 (c.a.) innesti 
*/

macro_declaration   : 	  macro_name   MACRO_ASSIGN  '{'  event  '}'
			  {
			    if ($1 != NULL)
			      {
				/* cerco i macro nodi nell'albero e li espando 
				g_node_traverse($4, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, expand_macro_node, NULL);
				*/
				$1->macro_val = $4;
				$$ = $1;
			      }
			    else
			      error_local = 1;
			    
			    sys_env = 0;
			    
			    in_macro = 0;
			    num_macro_in_macro = 0;

			    if (error_local)
			      {
				error_local = 0;
				num_error++;
				YYERROR;
			      }
			  }

                    |     error '{' event '}'
                          {
			    yyerror("ERROR: parse error in macro assigment");
			    free_parse_tree($3);
                          }

                    |     macro_name  MACRO_ASSIGN  '{' error '}' 
                          {
                            yyerror("ERROR: parse error in event combination of macro assignment of '%s'", $1->name);
                          }
             /*     |     MACRO '(' parameters ')'    MACRO_ASSIGN  '{'  event  '}'     */
		    ;

/* i parametri non sono stati implementati

parameters 	    :     parameter
		    |     parameters ',' parameter
		    ;

parameter           :     VARIABLE
		    |     HASH_VAR
		    ;
*/

/**************************** FINE DICHIARAZIONI ****************************/

/***************************** AZIONI ***************************************/

intro_stms_counter  :     /* empty */ 
                          {
			    in_exit_statements = 0;
			    in_intro_statements = 1;
			    num_statements = 0;
			  }
                    ;


exit_stms_counter   :     /* empty */
                          {
			    in_intro_statements = 0;
			    in_exit_statements = 1;
			    num_statements = 0;
			  }
                    ;

statements          :     '<' intro_statements  exit_statements '>'
			  {
			    $$[0] = $2;
			    $$[1] = $3;
			  }

                    |     '<' intro_statements '>'
                          {
			    $$[0] = $2;
			    $$[1] = NULL;
			  }

		    |     '<' exit_statements '>'
                          {
			    $$[0] = NULL;
			    $$[1] = $2;
			  }

                    |     '<' error exit_statements '>'
                          {
			    yyerror("ERROR: parse error in intro statements");
			    $$[0] = NULL;
			    $$[1] = $3;
			  }

		    |     '<' intro_statements error '>'
                          {
			    yyerror("ERROR: parse error in exit statements");
			    $$[0] = $2;
			    $$[1] = NULL;
			  }

                    |     '<' error '>'
                          {
			    yyerror("ERROR: parse error in statements");
			    $$[0] = NULL;
			    $$[1] = NULL;
			  }
		    ;


intro_statements    :     INTRO intro_stms_counter ':' statement_sequence
                          {
			    $$ = $4;
			    in_intro_statements = 0;  
                          }

                    |     INTRO intro_stms_counter ':' error
                          {
			    yyerror("ERROR: parse error in intro statements");
			    $$ = NULL;
			    in_intro_statements = 0;
                          }   
		          
                    ;

exit_statements     :     EXIT exit_stms_counter  ':' statement_sequence
                          {
                            $$ = $4;
			    in_exit_statements = 0;
			  }

                    |     EXIT exit_stms_counter  ':' error
                          {
			    yyerror("ERROR: parse error in exit statements");
			    $$ = NULL;
			    in_exit_statements = 0;
			  }
                    ;

statement_sequence  :     statement
			  {
			    $$ = NULL;
			    if ($1 != NULL) 
			      $$ = g_list_append($$, $1);
		          }			    
		    
                    |     statement_sequence ',' statement
			  {
			    if ($3 != NULL)
			      $$ = g_list_append($$, $3);
			  }	

                    |     statement_sequence ',' error
                          {
			    char *context = in_intro_statements ? "intro" : "exit";
                            yyerror("ERROR: parse error in %s statement n %d", context, num_statements);
                          }    
		    ;


statement           :     function
			  {
			    arg_node_t *node_op;
			    
			    node_op = (arg_node_t *) malloc(sizeof(arg_node_t));
			    node_op->funct = (void *(*)()) eval_stmt;  /* e' la funzione che verra' chiamata a run-time 
									  per eseguire function */
			    node_op->arg = NULL;
			    
			    $$ = g_node_new(node_op);
			    
			    /* assegno a funct il puntatore vero e proprio della funzione */
			    //((arg_node_t *)($1->data))->funct = ((funct_entry *)((arg_node_t *)($1->data))->funct)->ptr;
			    
			    g_node_append($$, $1);
			    
			    /* $$ = $1; */
			  }
		    
                    |     assignment
			  {
			    $$ = $1;
			  }
                    ;

/* se le variabili sono gia' state dichiarate il loro valore viene modificato
   altrimenti vengono aggiunte alle symbol table (la variabile sara' locale
   per questa sequenza di eventi) */

assignment          :     left_el '=' right_el
                          {
			    arg_node_t *l = $1->data;  
			    arg_node_t *r = $3->data;

			    arg_node_t *assign_node = (arg_node_t *) malloc(sizeof(arg_node_t));
			    /* assign_node->funct = (void *(*)()) assign;  TODO: togliere perche' non serve */
			    assign_node->arg = NULL;
			    $$ = g_node_new(assign_node);

			    g_node_append($$, $1);
			    g_node_append($$, $3);

			    /* se right_el == { } alloco una hashtable vuota */
			    if (r->arg != NULL)
			      if (r->arg->type_content == HASH_VALUE)
				if (r->arg->content.hash_value == NULL)
				  {
				    var_entry *var = NULL;
				    
				    if (in_macro)
				      var = getmacrovar(macro_var_table, l->arg->content.handler->var_name, num_macro);
				    
				    if (var == NULL)
				      {
					/* la variabile  globale oppure  locale ad una regola */
					var = getvar(var_table, l->arg->content.handler->var_name);
				      }
						  
				    assert( var != NULL );
				    
				    r->arg->type_value = var->type_values;
				    
				    switch (var->type_values) {
				    case STRING_VALUE:
				      r->arg->content.hash_value = g_hash_table_new(g_str_hash, g_str_equal);
				      break;
				    case CHARACTER_VALUE: case LONG_VALUE:
				      r->arg->content.hash_value = g_hash_table_new(g_int_hash, g_int_equal);
				      break;
				    default:
				      yyerror("ERROR: unknown hash value");
				      r->arg->content.hash_value = g_hash_table_new(NULL, NULL);
				      break;
				    }
				  }
			    			    
			    /* assegno a assign_node->funct la funzione da eseguire a run-time per
			       effettuare l'assegnamento a seconda di left_el e right_el */
			    if (comb_env != 2)
			      assign_node->funct =  (void *(*)()) assign($1->data, $3->data, num_macro); 
			    
			  }

                    |     error '=' right_el
                          {
                            char *context = in_intro_statements ? "intro" : "exit";
                            yyerror("ERROR: parse error in lvalue of assignment (%s statement n. %d)", context, num_statements);
                            $$ = NULL;
			    free_tree_with_arg_node($3);
                          }

                    |     left_el '=' error
                          {
                            char *context = in_intro_statements ? "intro" : "exit";
			    
			    yyerror("ERROR: parse error in rvalue of assignment (%s statement n. %d)", context, num_statements);
                            $$ = NULL;
			    free_tree_with_arg_node($1);
                          }     
                    ;

left_el             :     type_var VARIABLE
                          {
			    arg_node_t *left_child;
			    int num;
			    
			    if (($1 == 0) && ($2->type_values == 0))
			      {
			        yyerror("ERROR: var %s without type", $2->name);
				exit(-1);
			      }
			    else if (($1 != 0) && ($2->type_values != 0))
			      {
				yyerror("ERROR: var %s already declared", $2->name);
				exit(-1);
			      }
			    else if ($1 != 0) 
			      $2->type_values = $1;
			    

			    /* numero statments */
			    num_statements++;

			    /* inserisco la variabile nella tabella della visibilita' */
			    if (in_macro)
			      num = -num_macro;
			    else
			      num = num_seq;
			    
			    if ( firstscope($2->name, 0) == NULL )
			      {
				/* VARIABLE non e' definita globalmente */
				scope_el_t *scope = putscope($2->name, num);
				scope->num_event = in_macro ? -counter_in_macro : counter;
				scope->context = in_intro_statements ? intro : exiting;
				scope->num_statement = num_statements; 
			      }
			    

			    left_child = (arg_node_t *) malloc(sizeof(arg_node_t));
			    left_child->funct = NULL;
			    left_child->arg = (container *) malloc(sizeof(container));
			    left_child->arg->type_content = HANDLER;
			    left_child->arg->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    left_child->arg->content.handler->var_name = clone_string($2->name);
			    left_child->arg->content.handler->type_label = 0;
			    $$ = g_node_new(left_child);
                          } 
  
                    |     type_var HASH_VAR
                          {
			    arg_node_t *left_child;
			    int num;
			    
			    if (($1 == 0) && ($2->type_values == 0))
			      {
			        yyerror("ERROR: var %s without type", $2->name);
				exit(-1);	     
			      }
			    else if (($1 != 0) && ($2->type_values != 0))
			      {
				yyerror("ERROR: var %s already declared", $2->name);
				exit(-1);
			      }
			    else if ($1 != 0) 
			      $2->type_values = $1;
			    
			    /* numero statments */
			    num_statements++;

			    /* inserisco la variabile nella tabella della visibilita' */
			    if (in_macro)
			      num = -num_macro;
			    else
			      num = num_seq;
			    
			    if ( firstscope($2->name, 0) == NULL )
			      {
				/* VARIABLE non e' definita globalmente */
				scope_el_t *scope = putscope($2->name, num);
				scope->num_event = in_macro ? -counter_in_macro : counter;
				scope->context = in_intro_statements ? intro : exiting;
				scope->num_statement = num_statements; 
			      }
			    

			    left_child = (arg_node_t *) malloc(sizeof(arg_node_t));
			    left_child->funct = NULL;
			    left_child->arg = (container *) malloc(sizeof(container));
			    left_child->arg->type_content = HANDLER;
			    left_child->arg->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    left_child->arg->content.handler->var_name = clone_string($2->name);
			    left_child->arg->content.handler->type_label = 0;
			    $$ = g_node_new(left_child);
                          }
                    ;

right_el            :     value
                          {
			    arg_node_t *right_child = (arg_node_t *) malloc(sizeof(arg_node_t));
			    
			    right_child->funct = NULL;
			    right_child->arg = $1;
			    $$ = g_node_new(right_child);
			  }

                    |     right_hash_value
                          {
			    arg_node_t *right_child = (arg_node_t *) malloc(sizeof(arg_node_t));
			    
			    right_child->funct = NULL;
			    right_child->arg = $1;
			    $$ = g_node_new(right_child);
			  }

                    |     function
                          {
			    $$ = $1;
			  }

                    |     IDENTIFIER
                          {
			    arg_node_t *right_child = (arg_node_t *) malloc(sizeof(arg_node_t));
			    
			    right_child->funct = NULL;
			    right_child->arg = (container *) malloc(sizeof(container));
			    right_child->arg->type_content = HANDLER;
			    right_child->arg->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    right_child->arg->content.handler->var_name = NULL;
			    
			    right_child->arg->content.handler->type_label = $1->type_label;
			    
			    if ($1->n_arg != -1)
			      right_child->arg->content.handler->label_ref.n_arg = $1->n_arg;
			    else 
			      right_child->arg->content.handler->label_ref.label_name = $1->key->name;

			    $$ = g_node_new(right_child);
			  }

                    |     VARIABLE
                          {
			    arg_node_t *right_child = (arg_node_t *) malloc(sizeof(arg_node_t));
			    
			    right_child->funct = NULL;
			    right_child->arg = (container *) malloc(sizeof(container));
			    right_child->arg->type_content = HANDLER;
			    right_child->arg->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    right_child->arg->content.handler->var_name = clone_string($1->name);
			    right_child->arg->content.handler->type_label = 0; 

			    if (once_var) 
			      {
                                yyerror("ERROR: var %s undefined", $1->name);						      
                                YYERROR;
			      }

			    $$ = g_node_new(right_child);
			  }

                    |     HASH_VAR
                          {
			    arg_node_t *right_child = (arg_node_t *) malloc(sizeof(arg_node_t));
			    
			    right_child->funct = NULL;
			    right_child->arg = (container *) malloc(sizeof(container));
			    right_child->arg->type_content = HANDLER;
			    right_child->arg->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    right_child->arg->content.handler->var_name = clone_string($1->name);
			    right_child->arg->content.handler->type_label = 0; 
			    
			    if (once_var) 
			      {
                                yyerror("ERROR: var %s undefined", $1->name);
                                YYERROR;
			      }

			    $$ = g_node_new(right_child);
			  }
                    ;

/**************************** FINE AZIONI **********************************/

/***************************** CONDIZIONI **********************************/

event_condition     :     '(' conditions ')'
			  {
			    $$[0] = $2;
			    $$[1] = NULL;

			    if (unverified_var_stack)
			      {
				/* aggiungo un elemento nullo per distinguere il contesto */
				unverified_var_stack = g_list_prepend(unverified_var_stack, NULL);
			      }
			  }

		    |     '(' intro_conditions  exit_conditions ')'      
			  {
			    $$[0] = $2;
			    $$[1] = $3;
	                  }

		    |     '(' exit_conditions ')'            
			  {
			    $$[0] = NULL;
			    $$[1] = $2;
		          }

		    |     '(' intro_conditions ')'            
			  {  
			    $$[0] = $2;
			    $$[1] = NULL;

			    if (unverified_var_stack)
			      {
				/* aggiungo un elemento nullo per distinguere il contesto */
				unverified_var_stack = g_list_prepend(unverified_var_stack, NULL);
			      }
	                  }

                    |     '(' error exit_conditions ')'
                          {
			    yyerror("ERROR: parse error in intro conditions");
			    $$[0] = NULL;
			    $$[1] = $3;
			    
			  }

                    |     '(' intro_conditions error ')'
                          {
			    yyerror("ERROR: parse error in exit conditions");
			    $$[0] = $2;
			    $$[1] = NULL;
			  }

                    |     '(' error ')'
                          {
			    yyerror("ERROR: parser error in conditions");
			    $$[0] = NULL;
			    $$[1] = NULL;
			  }
		    ;

intro_conditions    :     INTRO ':' conditions
                          {
			    $$ = $3;
			    
			    if (unverified_var_stack)
			      {
				/* aggiungo un elemento nullo per distinguere il contesto */
				unverified_var_stack = g_list_prepend(unverified_var_stack, NULL);
			      }
			  }

                    |     INTRO ':' error
                          {
			    yyerror("ERROR: parse error in intro conditions");
			    $$ = NULL;
			    if (unverified_var_stack)
			      {
				/* aggiungo un elemento nullo per distinguere il contesto */
				unverified_var_stack = g_list_prepend(unverified_var_stack, NULL);
			      }
			  }
                    ;

exit_conditions     :     EXIT ':' conditions
                          {
			    $$ = $3;
			  }

                    |     EXIT ':' error
                          {
			    yyerror("ERROR: parse error in exit conditions");
			    $$ = NULL;
			  }
                    ;

conditions          :     condition                               
                          {
			    $$ = NULL;
			    if ( $1 != NULL )
			      $$ = g_list_append($$, $1);   
			  }
		    
                    |     conditions ',' condition                
                          {  
			    if ( $3 != NULL )
			      $$ = g_list_append($$, $3);
			  }
		    ;

condition           :     relational_operand   relational_operator   relational_operand
                          {
			    arg_node_t *node_op = (arg_node_t *) malloc(sizeof(arg_node_t));
			    node_op->funct = (void *(*)()) $2;
			    node_op->arg = NULL;
			    
			    $$ = g_node_new(node_op);
			    
			    g_node_append($$, $1);
			    g_node_append($$, $3);
			    
			    /* assegno a node_op->funct la funzione da eseguire a run-time per valutare la
			       condizione a seconda del tipo dei 2 relational_operand */
			    if (comb_env != 2)
			      node_op->funct = (*node_op->funct)($1->data, $3->data, num_macro);
			  }
			    
                    |     function
                          {
			    /* creo una condizione del tipo funct != 0 */
			    arg_node_t *operand_right, *node_op;
			    
			    node_op = (arg_node_t *) malloc(sizeof(arg_node_t));
			    node_op->funct = (void *(*)()) ne;
			    node_op->arg = NULL;

			    operand_right = (arg_node_t *) malloc(sizeof(arg_node_t));
			    operand_right->funct = NULL;
			    operand_right->arg = (container *) malloc(sizeof(container));
			    
			    operand_right->arg->type_content = SIMPLE_VALUE;
			    operand_right->arg->type_value = LONG_VALUE;
			    operand_right->arg->content.value = (long *) malloc(sizeof(long));
			    *((long *)(operand_right->arg->content.value)) = 0;

			    $$ = g_node_new(node_op);
			    
			    g_node_append($$, $1);
			    g_node_append_data($$, operand_right);
			    
			    if (comb_env != 2)
			      node_op->funct = (*node_op->funct)($1->data, g_node_last_child($$)->data, num_macro);
			  }

                    |     '(' condition ')'
                          {
			    $$ = $2;
			  }

                    |     error relational_operator relational_operand
                          {
			    yyerror("ERROR: parse error in left operand of a condition");
			    $$ = NULL;
			    free_tree_with_arg_node($3);
			  }

                    |     relational_operand relational_operator error
                          {
			    yyerror("ERROR: parse error in right operand of a condition");
			    $$ = NULL;
			    free_tree_with_arg_node($1);
			  }

                    |     '(' error ')'
                          {
			    yyerror("ERROR: parse error in a condition");
			    $$ = NULL;
			  }
                    ;

relational_operand  :     function
                          {
			    $$ = $1;
			  }

                    |     value
                          {
			    arg_node_t *operand = (arg_node_t *) malloc(sizeof(arg_node_t));
			    operand->funct = NULL;
			    operand->arg = $1;
			    $$ = g_node_new(operand);
			  }

                    |     variable2
                          {
			    arg_node_t *operand = (arg_node_t *) malloc(sizeof(arg_node_t));
			    operand->funct = NULL;
			    operand->arg = $1;
			    $$ = g_node_new(operand);
			  }
                    ;

relational_operator       :     '<''='               { $$ = le; }
		          |     '>''='	             { $$ = ge; }
		          |     '<'		     { $$ = lt; }
		          |     '>'		     { $$ = gt; }
		          |     '!''='	             { $$ = ne; }
		          |     '=''='               { $$ = eq; }
		          ;

/**************************** FINE CONDIZIONI ******************************/

/****************** Funzioni utilizzate da Azioni e Condizioni *************/

function_name       :     FUNCTION  
                          { 
			    /* se la funzione non e' riconosciuta non inserirla nello stack */
			    arg_node_t *arg_node = (arg_node_t *) malloc(sizeof(arg_node_t));
			    funct_entry *fe = NULL;
			    
			    if ((fe = getfunct(funct_table, $1)) == NULL)
			      {
				yyerror("WARNING: function %s not defined", $1);
				exit(-1);
			      }
			    
			    if ((in_intro_statements || in_exit_statements) && (function_stack == NULL)) 
			      num_statements++;
			    
			    /* IMPOSTO IL PUNTATORE ALLA FUNZIONE CORRISP.*/
			    assert(fe->ptr != NULL);
			    //arg_node->funct = (void *(*)()) fe->ptr;  
			    arg_node->funct = (void *(*)()) fe;
			    arg_node->arg = NULL;
			    
			    $$ = g_node_new(arg_node);
			    function_stack = g_list_prepend(function_stack, $$);
			  
			    num_func++;
			    fe_ptr[num_func] = fe;
                          }
		    ;

function            :     function_name  '(' func_args ')'       
                          {
			    funct_entry *fe;
			    GList *first = g_list_first(function_stack);
			    function_stack = g_list_remove_link(function_stack, first);
			    $$ = (GNode *) first->data;
			    g_list_free_1(first);
			    
			    if (func_arg_number[num_func] < (fe = ((funct_entry *)((arg_node_t *)$$->data)->funct))->args_num)
			      {
				yyerror("\nERROR: too less args in %s\n", fe->name);
				exit(-1);
			      }
			    /*else if (func_arg_number[num_func] > fe->args_num)
			      {
				printf("\nERROR: too much args in %s\n", fe->name);
				exit(-1);
				}*/
			    
			    func_arg_number[num_func] = 0;
			    fe_ptr[num_func] = NULL;
			    num_func--;
			    /* ATTENZIONE: controllare argomenti passati a function */
			    
			    /* IMPOSTO IL PUNTATORE ALLA FUNZIONE CORRISP.*/
			    ((arg_node_t *)$$->data)->funct = fe->ptr;
			  }
		    
                    |     function_name  '(' ')'                 
                          {
			    funct_entry *fe;
			    GList *first = g_list_first(function_stack);
			    function_stack = g_list_remove_link(function_stack, first);
			    $$ = (GNode *) first->data;
			    g_list_free_1(first);

			    if ( (fe = (funct_entry *)((arg_node_t *)$$->data)->funct)->args_num != 0)
			      {
				yyerror("\nERROR: too less args in %s\n", fe->name);
				exit(-1);
			      }
			    
			    func_arg_number[num_func] = 0;
			    fe_ptr[num_func] = NULL;
			    num_func--;
			    /* ATTENZIONE: controllare argomenti passati a function */
			    
			    /* IMPOSTO IL PUNTATORE ALLA FUNZIONE CORRISP.*/
			    ((arg_node_t *)$$->data)->funct = fe->ptr;
			  }

                    |     error '(' func_args ')'
                          {
			    char *context = in_intro_statements ? "intro statements" : in_exit_statements ? "exit statements" : "conditions";
			    
			    yyerror("ERROR: parse error in function call of %s", context);
			    $$ = NULL;
			  }

		    |     error '(' ')'
                          {
			    char *context = in_intro_statements ? "intro statements" : in_exit_statements ? "exit statements" : "conditions";

			    yyerror("ERROR: parse error in function call of %s", context);
			    $$ = NULL;
			  }

		    |     function_name '(' error ')'
                          {
			    char *context = in_intro_statements ? "intro statements" : in_exit_statements ? "exit statements" : "conditions";

			    /* TODO: implementare una funzione che dato un puntatore restituisce il nome della funzione  */
			    yyerror("ERROR: parse error in the arguments of function call of %s", context);
			    $$ = NULL;
			  }  
		    ;

func_args           :     func_num_arg  func_arg                              
                          {
			    /* ci deve essere almeno un elemento in function_stack */
			    assert(function_stack != NULL);
			    
			    g_node_append((GNode *) function_stack->data, $2);
			    
			    $$ = function_stack;
			  }
		    
                    |     func_args ',' func_num_arg  func_arg                
                          {
			    /* ci deve essere almeno un elemento in function_stack */
			    assert(function_stack != NULL);
			    
			    g_node_append((GNode *) function_stack->data, $4);
			    
			    $$ = function_stack;
			  }

                    |     func_args ',' error
                          {
			    char *context = in_intro_statements ? "intro statements" : in_exit_statements ? "exit statements" : "conditions";
			    /* TODO: implementare una funzione che dato un puntatore restituisce il nome della funzione  */
			    yyerror("ERROR: parse error in an argument of function call of %s", context);
			  }
		    ;

func_num_arg        :    
                          {
			    func_arg_number[num_func]++;
			    
			    if (func_arg_number[num_func] > fe_ptr[num_func]->args_num)
			      {
				yyerror("ERROR: too much args in %s", fe_ptr[num_func]->name);
				exit(-1);
			      }
                          }
		    ;


func_arg            :     value            
                          {
			    arg_node_t *arg_node = (arg_node_t *) malloc(sizeof(arg_node_t));
			    arg_node->arg = $1;
			    arg_node->funct = NULL;
			    $$ = g_node_new(arg_node);
			    
			    /* controllo il tipo del valore passato come arg all'attuale function */
			    if ($1->type_value != fe_ptr[num_func]->args_type[func_arg_number[num_func]-1])
			      {
				yyerror("ERROR: type mismatch in arg %d of function %s", 
				       func_arg_number[num_func], fe_ptr[num_func]->name);
				exit(-1);
			      }
			  }
        	    
                    |     variable2        
                          {
			    var_entry *var = NULL;
			    arg_node_t *arg_node = (arg_node_t *) malloc(sizeof(arg_node_t));
			    arg_node->arg = $1;
			    arg_node->funct = NULL;
			    $$ = g_node_new(arg_node);
			    
			    /* controllo il tipo della var o label passata come arg all'attuale function */
			    if ($1->content.handler->var_name) /* e' VAR o HASHVAR */
			      {
				if (in_macro)
				  var = getmacrovar(macro_var_table, $1->content.handler->var_name, num_macro);
				
				if (var == NULL)
				  {
				    /* la variabile  globale oppure  locale ad una regola */
				    var = getvar(var_table, $1->content.handler->var_name);
				  }
				
				assert( var != NULL );
				
				if (var->type_var == scalar) /* VAR */
				  {
				    if (var->type_values != fe_ptr[num_func]->args_type[func_arg_number[num_func]-1])
				      {
					yyerror("ERROR: type mismatch in arg %d of function %s", 
					       func_arg_number[num_func], fe_ptr[num_func]->name);
					exit(-1);
				      }
				  }
				else /* HASHVAR */
				  {
				    /* ATTENZIONE: cambiare numero 100 con costante definita altrove */
				    
				    if ((var->type_values + 100) != fe_ptr[num_func]->args_type[func_arg_number[num_func]-1])
				      {
					yyerror("ERROR: type mismatch in arg %d of function %s", 
					       func_arg_number[num_func], fe_ptr[num_func]->name);
					exit(-1);
				      }
				  }
			      }
			    else /* e' un SEGNAPOSTO */
			      {
				if ($1->content.handler->type_label !=
				    fe_ptr[num_func]->args_type[func_arg_number[num_func]-1])
				  {
				    yyerror("ERROR: type mismatch in arg %d of function %s", 
					   func_arg_number[num_func], fe_ptr[num_func]->name);
				    exit(-1);
				  }
			      }
			    
			  }
		    
                    |     function         
                          {
			    $$ = $1;
			  
			    /* controllo il tipo della funzione passata come arg ad un'altra funzione */
			    if ( funct_ret_type( ((arg_node_t *)$1->data)->funct ) != 
				 fe_ptr[num_func]->args_type[func_arg_number[num_func]-1] )
			      {
				yyerror("ERROR: type mismatch in arg %d of function %s", 
				       func_arg_number[num_func], fe_ptr[num_func]->name);
				exit(-1);
			      }
			  }
		    ;

/* in questo punto si presuppone che le variabile siano dichiarate precedemente
   ed abbiano un valore */

variable2           :     VARIABLE      /* scalare    */
                          { 
			    
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HANDLER;
			    $$->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    $$->content.handler->var_name = clone_string($1->name);
			    $$->content.handler->type_label = 0;
			    
			    if (once_var)
			      {
				if (in_intro_statements || in_exit_statements)
				  {
				    yyerror("ERROR: variable %s undefined", $1->name);
				    YYERROR;
				  }
				else /* sono in una condizione */
				  unverified_var_stack = g_list_prepend(unverified_var_stack, $$->content.handler);
			      }
			  }
		    
                    |     HASH_VAR      /* hash       */    
                          { 
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HANDLER;
			    $$->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    $$->content.handler->var_name = clone_string($1->name);
			    $$->content.handler->type_label = 0;
			    
			    if (once_var)
			      {
				if (in_intro_statements || in_exit_statements)
                                  {
                                    yyerror("ERROR: variable %s undefined", $1->name);
				    YYERROR;
				  }
				else /* sono in una condizione */
				  unverified_var_stack = g_list_prepend(unverified_var_stack, $$->content.handler);
			      }
			  }

		    |     IDENTIFIER    /* segnaposto */    
                          {
			    /* potrebbe dare errori in casi con eventi composti 
			       if ($1->num_event == 0) 
			       { 
			       printf(" ERROR: \"%s\" ", $1->name);
			       yyerror("label non assegnata ad argomenti");
			       YYERROR; 
			       } 
			       else */ 
			    
			    $$ = (container *) malloc(sizeof(container));
			    $$->type_content = HANDLER;
			    $$->content.handler = (handler_t *) malloc(sizeof(handler_t));
			    $$->content.handler->var_name = NULL;

			    $$->content.handler->type_label = $1->type_label;
			    
			    if ( $1->n_arg != -1 )
			      $$->content.handler->label_ref.n_arg = $1->n_arg;
			    else
			      {
				assert(comb_env == 2);
				$$->content.handler->label_ref.label_name = $1->key->name;
			      }
			    
			    /*
			    if (comb_env == 2)
			      {
				label_value_t *fe;
				int ne;
				
				if (!in_macro) 
				  ne = counter;
				else   
				  ne = -counter_in_macro;
				
				
				if (in_macro)
				  {
				    assert( ne < 0 ); 
				    fe = get_label(macro_label_table, $$->content.handler->label_ref.label_name, ne);
				  }
				else
				  {
				    assert( ne > 0 );
				    fe = get_label(label_table, $$->content.handler->label_ref.label_name, ne);
				  }
				
				if (fe)
				  {
				    $$->content.handler->label_ref.n_arg = fe->n_arg;
				    $$->content.handler->type_label = fe->type_label;
				  }
				else
				  {
				    char *error_message;
				    printf("ERROR: label %s undefined",$1->key->name);
				  }
			      }
			    */
			    
			  }
		    ;

/***************************** SEQUENZE EVENTI *******************************/

attack_sequences   :     events '.'                          
			  {
			     event_node_t *father, *right_child; GNode *temp_tree;
			     
			     father = (event_node_t *) malloc(sizeof(event_node_t));
		             father->op = sequencing;
			     
			     /* creo l'evento END_MARKER */
			     right_child = (event_node_t *) malloc(sizeof(event_node_t));
			     right_child->op = none;
			     right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
		             (*right_child->simple_event).n_sysc = END_MARKER;
			     (*right_child->simple_event).num_event = 0;
			     (*right_child->simple_event).fl.negated_condition = 0;
			     (*right_child->simple_event).empty_follow = 0;
			     (*right_child->simple_event).condition = NULL;
		             (*right_child->simple_event).action = NULL;
   
			     temp_tree = g_node_new(father);
			     g_node_append(temp_tree, $1);
			     g_node_append_data(temp_tree, right_child); 
			     $$ = $1 = temp_tree;

			     num_seq++;
			  }
	            
                    |     attack_sequences  events '.'       
			  {
                             event_node_t *father, *big_father, *right_child; GNode *temp_tree;
			     
			     father = (event_node_t *) malloc(sizeof(event_node_t));
		             father->op = sequencing;
			     
			     /* creo l'evento END_MARKER */
			     right_child = (event_node_t *) malloc(sizeof(event_node_t));
			     right_child->op = none;
			     right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
		             (*right_child->simple_event).n_sysc = END_MARKER;
			     (*right_child->simple_event).num_event = 0;
			     (*right_child->simple_event).fl.negated_condition = 0;
			     (*right_child->simple_event).empty_follow = 0;
			     (*right_child->simple_event).condition = NULL;
		             (*right_child->simple_event).action = NULL;
			     
			     temp_tree = g_node_new(father);
			     g_node_append(temp_tree, $2);
			     g_node_append_data(temp_tree, right_child); 
			     $2 = temp_tree; 
			     
			     big_father = (event_node_t *) malloc(sizeof(event_node_t));
		             big_father->op = alternation;
			     
			     $$ = g_node_new(big_father);
			     g_node_append($$, $1);
			     g_node_append($$, $2);

			     num_seq++;
			  }

                    |     attack_sequences error '.'
                          {
			    
			  }
            	    ;


events              :     event
			  {
			     event_node_t *father, *left_child, *right_child, *father2, *father3;
			     GNode *temp_tree, *repet_tree;
			     
			     assert(!in_macro); 
			     
			     /* creo un' ANY* all'inizio delle sequenze che non contengono begin */	
			     father = (event_node_t *) malloc(sizeof(event_node_t));
			     father->op = sequencing;
			     
			     left_child = (event_node_t *) malloc(sizeof(event_node_t));
			     left_child->op = none;
			     left_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
			     (*left_child->simple_event).n_sysc = ANY;
			     (*left_child->simple_event).num_event = 0;
			     (*left_child->simple_event).context = intro;
			     (*left_child->simple_event).fl.minus_sysc = NULL;
			     (*left_child->simple_event).empty_follow = 0;
			     (*left_child->simple_event).condition = NULL;
			     (*left_child->simple_event).action = NULL;
			     
			     right_child = (event_node_t *) malloc(sizeof(event_node_t));
			     right_child->op = none;
			     right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
			     (*right_child->simple_event).n_sysc = ANY;
			     (*right_child->simple_event).num_event = 0;
			     (*right_child->simple_event).context = exiting;
			     (*right_child->simple_event).fl.minus_sysc = NULL;
			     (*right_child->simple_event).empty_follow = 0;
			     (*right_child->simple_event).condition = NULL;
			     (*right_child->simple_event).action = NULL; 
			     
			     temp_tree = g_node_new(father);
			     g_node_append_data(temp_tree, left_child);
			     g_node_append_data(temp_tree, right_child);
			     
			     /* creo * da mettere davanti ad ANY */
			     father3 = (event_node_t *) malloc(sizeof(event_node_t));
		             father3->op = repetition;
			     repet_tree = g_node_new(father3);
			     g_node_append(repet_tree, temp_tree);
			     
			     /* creo ANY* ; event */
			     father2 = (event_node_t *) malloc(sizeof(event_node_t));
			     father2->op = sequencing;
			     
			     $$ = g_node_new(father2);
			     g_node_append($$, repet_tree);
			     g_node_append($$, $1);
			  }
	            
                    |     event_begin ';' event
			  {
			     $$ = $3;
			  }

                    |     event_begin ';' error
                          {
			    
			  }
		    ;


event_begin         :     START '('')'     /* non creo nessun nodo per l'evento begin */
		    ;


event               :     simple_event          
			  {
			     $$ = $1;
			  }
                    
            /*      |     macro
			  { 
			     $$ = $1; 
			  } */
		    
                    |     event_operations
			  {
			     $$ = $1;
			  }
		    
                    |     '{' event_operations '}' '|' event_condition 
		          {
			    GList *an_intro_condition, *an_exit_condition;

			    GList *sel;  /* simple_event element */
			    
			    GList *simple_event_list = NULL;

			    follow_array_t afs = create_follow_array_of_tree($2,SUBTREE);
			    /* print_all_follow(afs); */
			    			    
			    g_node_traverse($2, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, find_simple_event, &simple_event_list);
			    
			    simple_event_list = g_list_reverse(simple_event_list); 
			    
			    for (sel = simple_event_list; sel; sel = g_list_next(sel)) 
			      {
				
				if ( ((simple_event_t *) sel->data)->context == intro )
				  for ( an_intro_condition = $5[0]; an_intro_condition; an_intro_condition = g_list_next(an_intro_condition) )
				    {
				      GList *prev, *next;
				      
				      /* isolo l'elemento della lista per passarlo alla append_group_action */
				      next = an_intro_condition->next;
				      prev = an_intro_condition->prev;
				      an_intro_condition->next = NULL;
				      an_intro_condition->prev = NULL;
				      
				      append_group_condition(an_intro_condition, (simple_event_t *) sel->data, afs);
				      
				      /* ripristino la lista */
				      an_intro_condition->next = next;
				      an_intro_condition->prev = prev;
				    }
				
				
				if ( ((simple_event_t *) sel->data)->context == exiting )
				  for ( an_exit_condition = $5[1]; an_exit_condition; an_exit_condition = g_list_next(an_exit_condition) )
				    {
				      GList *prev, *next;
				      
				      /* isolo l'elemento della lista per passarlo alla append_group_action */
				      next = an_exit_condition->next;
				      prev = an_exit_condition->prev;
				      an_exit_condition->next = NULL;
				      an_exit_condition->prev = NULL;
				      
				      append_group_condition(an_exit_condition, (simple_event_t *) sel->data, afs);
				      
					  /* ripristino la lista */
				      an_exit_condition->next = next;
				      an_exit_condition->prev = prev;
				    }
				
			      }

			    free_condition_or_action($5[0]);
			    free_condition_or_action($5[1]);
			    
			    g_list_free(simple_event_list);
			    
			    $$ = $2;
			  }
		    
		    |     '{' event_operations '}' '|' event_condition statements
		          {
			    GList *an_intro_action, *an_exit_action;
			    GList *an_intro_condition, *an_exit_condition;

			    GList *sel;  /* simple_event element */
			    
			    GList *simple_event_list = NULL;
			    
			    follow_array_t afs = create_follow_array_of_tree($2,SUBTREE);
			    /* print_all_follow(afs); */
			    			    
			    g_node_traverse($2, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, find_simple_event, &simple_event_list);
			    simple_event_list = g_list_reverse(simple_event_list);

			    for (sel = simple_event_list; sel; sel = g_list_next(sel)) 
			      {
				if ( ((simple_event_t *) sel->data)->context == intro )
				  {
				    int intro_action_ind = 0;
				    
				    for ( an_intro_condition = $5[0]; an_intro_condition; an_intro_condition = g_list_next(an_intro_condition) )
				      {
					GList *prev, *next;
					
					/* isolo l'elemento della lista per passarlo alla append_group_action */
					next = an_intro_condition->next;
					prev = an_intro_condition->prev;
					an_intro_condition->next = NULL;
					an_intro_condition->prev = NULL;
					
					append_group_condition(an_intro_condition, (simple_event_t *) sel->data, afs);
					
					/* ripristino la lista */
					an_intro_condition->next = next;
					an_intro_condition->prev = prev;
				      }
				    
				    for ( an_intro_action = $6[0]; an_intro_action; an_intro_action = g_list_next(an_intro_action)  )
				      {
					GList *prev, *next;
					
					intro_action_ind++;
					
					/* isolo l'elemento della lista per passarlo alla append_group_action */
					next = an_intro_action->next;
					prev = an_intro_action->prev;
					an_intro_action->next = NULL;
					an_intro_action->prev = NULL;
					
					append_group_action(an_intro_action, intro_action_ind, (simple_event_t *) sel->data, afs); 
					
					/* ripristino la lista */
					an_intro_action->next = next;
					an_intro_action->prev = prev;
				      }
				    
				  }
				    
				if ( ((simple_event_t *) sel->data)->context == exiting )
				  {
				    int exit_action_ind = 0;
				    
				    for ( an_exit_condition = $5[1]; an_exit_condition; an_exit_condition = g_list_next(an_exit_condition) )
				      {
					GList *prev, *next;
					
					/* isolo l'elemento della lista per passarlo alla append_group_action */
					next = an_exit_condition->next;
					prev = an_exit_condition->prev;
					an_exit_condition->next = NULL;
					an_exit_condition->prev = NULL;
					
					append_group_condition(an_exit_condition, (simple_event_t *) sel->data, afs);
					
					/* ripristino la lista */
					an_exit_condition->next = next;
					an_exit_condition->prev = prev;
				      }
				    
				    
				    for ( an_exit_action = $6[1]; an_exit_action; an_exit_action = g_list_next(an_exit_action) )
				      {
					GList *prev, *next;
					
					exit_action_ind++;
					
					/* isolo l'elemento della lista per passarlo alla append_group_action */
					next = an_exit_action->next;
					prev = an_exit_action->prev;
					an_exit_action->next = NULL;
					an_exit_action->prev = NULL;
					
					append_group_action(an_exit_action, exit_action_ind, (simple_event_t *) sel->data, afs); 
					
					/* ripristino la lista */
					an_exit_action->next = next;
					an_exit_action->prev = prev;
				      }
				  
				  }
			      }

			    free_condition_or_action($5[0]);
			    free_condition_or_action($5[1]);
			    free_condition_or_action($6[0]);
			    free_condition_or_action($6[1]);
			    
			    g_list_free(simple_event_list);
			    			    
			    $$ = $2;
			  }
		    
                    |     '{' event_operations '}' statements
		          {
			    GList *an_intro_action   = NULL;
			    GList *an_exit_action    = NULL;

			    GList *simple_event_list = NULL;
			     
			    GList *sel;               /* simple_event element */
			    
			    follow_array_t afs = create_follow_array_of_tree($2,SUBTREE);
			    /* print_all_follow(afs); */
			    			    
			    g_node_traverse($2, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, find_simple_event, &simple_event_list);
			    simple_event_list = g_list_reverse(simple_event_list);

			    for (sel = simple_event_list; sel; sel = g_list_next(sel)) 
			      {
				if ( ((simple_event_t *) sel->data)->context == intro )
				  {
				    int intro_action_ind = 0;
				    for ( an_intro_action = $4[0]; an_intro_action; an_intro_action = g_list_next(an_intro_action)  )
				      {
					GList *prev, *next;
					
					intro_action_ind++;
					
					/* isolo l'elemento della lista per passarlo alla append_group_action */
					next = an_intro_action->next;
					prev = an_intro_action->prev;
					an_intro_action->next = NULL;
					an_intro_action->prev = NULL;
					
					append_group_action(an_intro_action, intro_action_ind, (simple_event_t *) sel->data, afs); 
					
					/* ripristino la lista */
					an_intro_action->next = next;
					an_intro_action->prev = prev;
				      }
				  }
				
				if ( ((simple_event_t *) sel->data)->context == exiting )
				  {
				    int exit_action_ind = 0;
				    
				    for ( an_exit_action = $4[1]; an_exit_action; an_exit_action = g_list_next(an_exit_action) )
				      {
					GList *prev, *next;
					
					exit_action_ind++;
					
					/* isolo l'elemento della lista per passarlo alla append_group_action */
					next = an_exit_action->next;
					prev = an_exit_action->prev;
					an_exit_action->next = NULL;
					an_exit_action->prev = NULL;
					
					append_group_action(an_exit_action, exit_action_ind, (simple_event_t *) sel->data, afs); 
					/* ripristino la lista */
					an_exit_action->next = next;
					an_exit_action->prev = prev;
				      }
				  }
			      
			      }
			    
			    free_condition_or_action($4[0]);
			    free_condition_or_action($4[1]);
			    
			    g_list_free(simple_event_list);
			    
			    $$ = $2;
			    
			    /*
			      fflush(stdout);
			      printf("\n\n********************yacc*****************\n\n");
			      print_tree_with_event_node_t($$);
			      printf("\n\n********************yacc*****************\n\n");
			    */
			  }

		    |     '{' event '}'
			  {
			     //printf("\n'{' event '}'\n");
			     $$ = $2;
			  }
                    |     '{' error '}' '|' event_condition
                    |     '{' error '}' '|' event_condition statements
                    |     '{' error '}' '|' statements
                    |     '{' event_operations '}' '|' error
                    |     '{' event_operations '}' '|' error statements
                    |     '{' event_operations '}' error
                    |     '{' error '}'
		    ;


simple_event        :     syscalls      
			  { 
			     $$ = $1; 
			  }
		    
                    |	  syscalls '|' event_condition       
			  {
			    event_node_t *father, *left_child, *right_child;
			    GList *el;
			    
			    father = $1->data;
			    switch (father->op) 
			      {
			      case   sequencing:
				left_child = (g_node_first_child($1))->data;
				right_child = (g_node_last_child($1))->data;
				left_child->simple_event->condition 
				  = g_list_concat(left_child->simple_event->condition, $3[0]);
				right_child->simple_event->condition 
				  = g_list_concat(right_child->simple_event->condition, $3[1]);
				break;
			      default:
				;
			      }
			    
			    $$ = $1;

			    for ( el = unverified_var_stack; el; el = g_list_next(el) )
			      if ( el->data != NULL )
				{
				  yyerror("ERROR: variable %s undefined", ((handler_t *)el->data)->var_name);
				}
			    
			    if (unverified_var_stack) 
			      {
				g_list_free(unverified_var_stack);
				unverified_var_stack = NULL;
			      } 
			    			    			    
			  }
		    
                    |     syscalls '|' event_condition  statements 
                          {
			    event_node_t *father, *left_child, *right_child;
					 
			    father = $1->data;
			    switch (father->op) 
			      {
			      case   sequencing:
				left_child = (g_node_first_child($1))->data;
				right_child = (g_node_last_child($1))->data;
				left_child->simple_event->condition 
				  = g_list_concat(left_child->simple_event->condition, $3[0]);
				right_child->simple_event->condition 
				= g_list_concat(right_child->simple_event->condition, $3[1]);
				left_child->simple_event->action = $4[0];
				right_child->simple_event->action = $4[1];
				break;
			      default:
				;
			      }
			    
                             
			    $$ = $1;
			    
			    if (unverified_var_stack)
			      {
				GList *el;
				context_t context = exiting;
				int num;
				
				if (in_macro)
				  num = -num_macro;
				else
				  num = num_seq;
				
				for ( el = unverified_var_stack; el; el = g_list_next(el) )
				  {
				    scope_el_t *scope;
				    
				    if (el->data == NULL) /* l'elemento vuoto indica il cambiamento di contesto */
				      context = intro;
				    else
				      {
					scope = firstscope(((handler_t *) el->data)->var_name, num);
					if (!scope || (scope && scope->context >= context ))
					  {
					    yyerror("ERROR: variable %s undefined", ((handler_t *)el->data)->var_name);
					  }
				      }
				  }
			      }
			    
			    if (unverified_var_stack) 
			      {
				g_list_free(unverified_var_stack);
				unverified_var_stack = NULL;
			      } 
			  }
		    
                    |     syscalls   statements
			  {
			    event_node_t *father, *left_child, *right_child;
					 
		             father = $1->data;
		             switch (father->op) 
			       {
			       case   sequencing:
				 left_child = (g_node_first_child($1))->data;
				 right_child = (g_node_last_child($1))->data;
				 left_child->simple_event->action = $2[0];
				 right_child->simple_event->action = $2[1];
				 break;
			       default:
				 ;
			       }       
                             $$ = $1;
			     
			     
			     assert(unverified_var_stack == NULL);
			     
			  }
                    |     error '|' event_condition
                    |     error '|' event_condition statements
                    |     syscalls '|' error
                    |     syscalls '|' error statements
                    |     syscalls '|' event_condition error
                    |     syscalls error
		    ;


syscalls            :     '@'      
                          {
			    event_node_t *father, *left_child, *right_child;
			    
			    /* suddivido l'any in due eventi consecutivi per distinguere
			       eventuali condizioni e istruzioni in entrata da quelli in uscita */
			    father = (event_node_t *) malloc(sizeof(event_node_t));
			    father->op = sequencing;
			    
			    left_child = (event_node_t *) malloc(sizeof(event_node_t));
			    left_child->op = none;
			    left_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
			    (*left_child->simple_event).n_sysc = ANY;
			    (*left_child->simple_event).num_event = 0;
			    (*left_child->simple_event).context = intro;
			    (*left_child->simple_event).fl.minus_sysc = NULL;
			    (*left_child->simple_event).empty_follow = 0;
			    (*left_child->simple_event).condition = NULL;
			    (*left_child->simple_event).action = NULL;
			    
			    right_child = (event_node_t *) malloc(sizeof(event_node_t));
			    right_child->op = none;
			    right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
			    (*right_child->simple_event).n_sysc = ANY;
			    (*right_child->simple_event).num_event = 0;
			    (*right_child->simple_event).context = exiting;
			    (*right_child->simple_event).fl.minus_sysc = NULL;
			    (*right_child->simple_event).empty_follow = 0;
			    (*right_child->simple_event).condition = NULL;
			    (*right_child->simple_event).action = NULL; 
			    
			    $$ = g_node_new(father);
			    g_node_append_data($$, left_child);
			    g_node_append_data($$, right_child);   
			  }         
		    
                    |     syscall
			  { 
			    if ( $1 == NULL )
			      {
				$$ = NULL;
				YYERROR;
			      }
			    else
			      {
				event_node_t *father, *left_child, *right_child;
				GList *arg_exit_list, *now, *next;
				
				/* suddivido l'evento in entrata da quello in uscita */
				father = (event_node_t *) malloc(sizeof(event_node_t));
				father->op = sequencing;
				
				left_child = (event_node_t *) malloc(sizeof(event_node_t));
				left_child->op = none;
				left_child->simple_event = $1;
				(*left_child->simple_event).context = intro;
				
				right_child = (event_node_t *) malloc(sizeof(event_node_t));
				right_child->op = none;
				right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
				(*right_child->simple_event).n_sysc = (*left_child->simple_event).n_sysc;
				(*right_child->simple_event).num_event
				  = (*left_child->simple_event).num_event;
				(*right_child->simple_event).context = exiting;
				(*right_child->simple_event).fl.negated_condition
				  = (*left_child->simple_event).fl.negated_condition;
				(*right_child->simple_event).empty_follow
				  = (*left_child->simple_event).empty_follow;
								
				/* divido le condizioni INTRO da quelle EXIT */
				arg_exit_list = NULL;
				now = g_list_find($1->condition, NULL);
				next = g_list_next(now);
				while (next != NULL) {
				  arg_exit_list = g_list_append(arg_exit_list, next->data);
				  $1->condition = g_list_remove_link ($1->condition, now);
				  g_list_free_1(now);
				  now = next;
				  next = g_list_next(next);
				}
				
				/* elimino l'ultimo elemento nelle condizioni INTRO */
				$1->condition = g_list_remove_link ($1->condition, now);
				g_list_free_1(now);
				
				(*right_child->simple_event).condition = arg_exit_list;
				(*right_child->simple_event).action = NULL;             
				
				$$ = g_node_new(father);
				g_node_append_data($$, left_child);
				g_node_append_data($$, right_child);
			      }
			  }
                    ;


event_operations    :     event    OR    event
			  {
			     int min = 0; 
			     int max = 0; 
			     event_node_t *father;
			     
			     /* controllo la correttezza nell'uso di variabili in $3 */
			     min = min_num_event($1);
			     max = max_num_event($1);
			     if ((min != 0) && (max != 0)) check_vars_in_or_combination($3, min, max);
			     
			     father = (event_node_t *) malloc(sizeof(event_node_t));
		             father->op = alternation;
			     
			     $$ = g_node_new(father);
			     g_node_append($$, $1);
			     g_node_append($$, $3);
			  }
		    
                    |     event    ';'   event
			  {
			     event_node_t *father;
				
			     father = (event_node_t *) malloc(sizeof(event_node_t));
		             father->op = sequencing;
			     
			     $$ = g_node_new(father);
			     g_node_append($$, $1);
			     g_node_append($$, $3);
			  }

                    |     '!' event                 /*  evento != @ , broken_lock  */
			  {
			     event_node_t *father;
			     event_node_t *child = $2->data;
			     GNode *prev = NULL, *next, *tree;
			     
			     if (child->op == non_occurence)
			       {
				 yyerror("syntax error (more ! sequently not permitted)");
				 YYERROR;	     
			       }
			     			     
			     minus_sysc = NULL; /* var globale */
			     g_node_traverse ($2, G_IN_ORDER, G_TRAVERSE_ALL, -1, check_event_for_not, NULL);
			     assert(minus_sysc != NULL);
			     
			     prev = make_any_for_not(minus_sysc);
			     
			     next = get_negated_event($2);
			     do
			       {
				 father = (event_node_t *) malloc(sizeof(event_node_t)); 
				 father->op = alternation;
				 
				 tree = g_node_new(father);
				 g_node_append(tree, prev);
				 g_node_append(tree, next);
				 prev = tree;
			       }
			     while ((next = get_negated_event(NULL)) != NULL);
			     			     
			     if (prev != NULL) $$ = prev;
			  }   
		    
                    |     event '*'
			  {
			     event_node_t *father;
			     event_node_t *child = $1->data;
			    
			     if (child->op == repetition)
			       {
				 yyerror("syntax error (more * sequently not permitted)");
				 YYERROR; 				     
			       }
			     
			     father = (event_node_t *) malloc(sizeof(event_node_t));
		             father->op = repetition;
						      
			     $$ = g_node_new(father);
			     g_node_append($$, $1);
			  }
		    
                    |     '{' event_operations '}'
			  {
			    $$ = $2;
			  }

                    |     macro
                          {
			    $$ = $1;
			  }  

                    |     event OR error
                    |     error OR event
		    |     event ';' error
		    |     error ';' event
		    ;	  

/*
broken_lock         :     BRK_LOCK  '(' variable ')' BY '{'event'}' FROM '{'event'}' TO '{'event'}'
		    ;
*/

/***************************** FINE SEQUENZE EVENTI ***************************/

macro		    :     MACRO '(' ')'     
                          {
			    if ($1 != NULL)
			      {
				GList *list, *el;
				
				if ($1->macro_val == NULL)  
				  {
				    yyerror("ERROR: macro inside same macro not permitted");
				    YYERROR;
				  }
				
				/* rinomino le variabili locali alla macro e le inserisco nella 
				   tabella delle variabili cui appartiene l'elemento sintattico corrente */
				list = getmacrovarlist(macro_var_table, $1->macro_id);
								
				for ( el = list; el != NULL; el = g_list_next(el) )
				  {
				    char num[5];
				    char *name_var;
				    var_entry *ve = clone_var_entry((var_entry *) el->data);
				    
				    snprintf(num, 5, "%+d", !in_macro ? num_macro_in_event : num_macro_in_macro);
				    name_var = (char *) malloc( strlen(num) + strlen(ve->name)+1);
				    memcpy((char *) name_var, num, strlen(num));
				    memcpy((char *) name_var+strlen(num), ve->name, strlen(ve->name)+1);
				    
				    free(ve->name);
				    ve->name = name_var;

				    if (in_declarations) {
				      assert ( getmacrovar(macro_var_table, ve->name, num_macro) == NULL ); /* controllo */
				      putmacrovar(macro_var_table, ve->name, num_macro, ve);
				    }
				    else {
				      assert ( getvar(var_table, ve->name) == NULL ); /* controllo */
				      putvar(var_table, ve->name, ve);
				    }
				  }

				if (list) 
				  {
				    g_list_free(list);
				    list = NULL;
				  }
				
				
				/* duplico il syntax tree associato alla macro e rinomino le variabili in esso contenute */
				$$ = clone_tree_with_event_node((GNode *) $1->macro_val);
				
				g_node_traverse($$, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, find_simple_event, &list);
				
				for ( el = list; el != NULL; el = g_list_next(el) )
				  {
				    GList *one;
				    for ( one = ((simple_event_t *) el->data)->condition; one != NULL; one = g_list_next(one) )
				      g_node_traverse((GNode *) one->data, G_POST_ORDER, G_TRAVERSE_LEAFS, -1, 
						      rename_var_handler, NULL);
				    for ( one = ((simple_event_t *) el->data)->action; one != NULL; one = g_list_next(one) )
				      g_node_traverse((GNode *) one->data, G_POST_ORDER, G_TRAVERSE_LEAFS, -1, 
						      rename_var_handler, NULL);
				  }
				
				/* rinumero gli eventi contenuti nel syntax tree della macro */
			        modify_macro_num_events($$); 
				
			      }
			    else
			      {
				$$ = NULL;
				YYERROR;
			      }
			  }
;


syscall             :     syscall_args
                          {
			    $$ = $1;
			  }

                    |     syscall_args return_value
                          {
			    $$ = $1;
			  }

                    |     error return_value
                          {
			    yyerror("ERROR: parse error in a syscall");
			    $$ = NULL;
			  }
                    ;

syscall_args        :     name_syscall '(' ')'
                          {
			    simple_event_t *event_sysc;
			    
			    event_sysc = (simple_event_t *) malloc(sizeof(simple_event_t));
			    event_sysc->n_sysc = $1;
			    if (!in_macro) event_sysc->num_event = counter;
			    else   event_sysc->num_event = -counter_in_macro;
			    event_sysc->fl.negated_condition = 0;
			    event_sysc->empty_follow = 0;
			    event_sysc->condition = NULL;
			    event_sysc->action = NULL;
			    $$ = event_sysc;	
			  }

                    |     name_syscall '(' args ')'
                          {
			    simple_event_t *event_sysc;
			    
			    event_sysc = (simple_event_t *) malloc(sizeof(simple_event_t));
			    event_sysc->n_sysc = $1;
			    if (!in_macro) event_sysc->num_event = counter;
			    else   event_sysc->num_event = -counter_in_macro;
			    event_sysc->fl.negated_condition = 0;
			    event_sysc->empty_follow = 0;
			    event_sysc->condition = $3;
			    event_sysc->action = NULL;
			    $$ = event_sysc;	
                          }    

                    |     error '(' ')'
                          {
			    yyerror("ERROR: parse error in a syscall");
			    $$ = NULL;
			  }

                    |     error '(' args ')'
                          {
			    yyerror("ERROR: parse error in a syscall");
			    $$ = NULL;
			    free_condition_or_action($3);
			  }
 
                    |     name_syscall error
                          {
                            char *syscall_name;
			    
			    syscall_name = get_sysc_name($1);

			    yyerror("ERROR: parse error in %s call", syscall_name);
                            $$ = NULL;
			  }

                    |     name_syscall '(' error ')'
                          {
			    char *syscall_name;
			    syscall_name = get_sysc_name($1);
			    yyerror("ERROR: parse error in arguments %s call", syscall_name);
                            $$ = NULL;
			  }
                    ;


name_syscall        :     SYSCALL           
                          { 
			    $$ = sysc_num = $1;
			    arg_number = 0; 
			  }
		    ;


return_value        :     '=' {out_sysc = 1;} arg_label {out_sysc = 0;}
                    ;


args                :     num_arg arg                 
                          { 
			    $$ = NULL;
			    $$ = g_list_append($$, NULL); /* aggiungo separatore per condiz. 
							     che andranno in INTRO e in EXIT */
			    if ($2 != NULL) 
			      {
				if (is_input(sysc_num, arg_number)) $$ = g_list_prepend($$, $2);
				else if (is_output(sysc_num, arg_number)) $$ = g_list_append($$, $2);
				else yyerror("ERROR: arg ignored");
					 
			      }
			  }           
		    
                    |     args ',' num_arg arg        
                          { 
			    if ($4 != NULL)
			      {
				if (is_input(sysc_num, arg_number)) $$ = g_list_prepend($$, $4);
				else if (is_output(sysc_num, arg_number)) $$ = g_list_append($$, $4);
				else yyerror("ERROR: arg ignored");
			      }
			  }

                    |     args ',' num_arg error
                          {
                            char *syscall_name;

                            syscall_name = get_sysc_name(sysc_num);
			    yyerror("ERROR: parse error in argument n %d of %s call", arg_number, syscall_name);
			  }
		    ;


num_arg             :     /* niente */   
                          {
			    arg_number++;
			    if (get_nargs(sysc_num) < arg_number)
			      { 
				char *syscall_name;
				syscall_name = get_sysc_name(sysc_num);
                                yyerror( "ERROR: too much args in %s call", syscall_name ); 
				YYERROR; 
			      }
                           }
		    ;


arg                 :     arg_value         
                          { $$ = $1; } 
                    |     arg_label         
                          { $$ = $1; }   /* il valore assegnato a questo sara' usato solo
					                      nella eventuale condizione e/o eventuale azione */
	            ;

arg_value           :     UNDEFINED         
                          {
			    /* non creo niente */
			    $$ = NULL;
			  }
		    
                    |     STRING_VALUE            
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = STRING_VALUE;
			    temp_cont->content.value = $1;
			    
			    $$ = make_tree_for_arg_value(temp_cont, arg_number);		     
			  }
	            
                    |     CHARACTER_VALUE         
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = CHARACTER_VALUE;
			    temp_cont->content.value = $1;
			    
			    $$ = make_tree_for_arg_value(temp_cont, arg_number);
			  }
		    
                    |     an_integer 	    
                          {
			    /* in an_integer ho creato il container */
			    
			    $$ = make_tree_for_arg_value($1, arg_number);
			  }        
        	    
                    |     array             
                          {
			    /* in array ho creato il container */
			    if ($1 != NULL)
			      $$ = make_tree_for_arg_value($1, arg_number);
			    else
			      $$ = NULL;
			  }              
		    
                    |     variable          
                          { /* la variabile (normale o hash) deve essere stata dichiarata
			       precedentemente in un'azione o nelle dichiarazioni globali */
			    
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = HANDLER;
			    temp_cont->content.handler = $1;
			    
			    $$ = make_tree_for_arg_value(temp_cont, arg_number);
			  } 
		    ;


arg_label           :     IDENTIFIER        
                          {
			    if ($1->n_arg != -1) 
			      { 
				yyerror("ERROR: label already used for precedent arg");
				YYERROR; 
			      }
			    
			    if (!out_sysc)
			      {
				$1->n_arg = arg_number - 1;
				$1->type_label = find_type(sysc_num, arg_number);
			      }	    
			    else
			      {
				$1->n_arg = RVAL; /* il segnaposto e' messo al valore di ritorno della syscall */
				$1->type_label = find_type(sysc_num, -1);
			      }
			    
			    $$ = NULL;
			  }
		    ;


array               :     '[' characters ']'     
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = ARRAY;
			    temp_cont->type_value = CHARACTER_VALUE;
			    temp_cont->content.value = $2;
			    $$ = temp_cont;
			  }
		    
                    |     '[' strings ']'        
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = ARRAY;
			    temp_cont->type_value = STRING_VALUE;
			    temp_cont->content.value = $2;
			    $$ = temp_cont;
			  }     
		    
                    |     '[' integers ']'       
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = ARRAY;
			    temp_cont->type_value = LONG_VALUE;    /* fra gli elementi ci possono
								anche essere dei FLAG */
			    temp_cont->content.value = $2;
			    $$ = temp_cont;
			  }

                    |     '[' error ']'          
                          {
			    yyerror("ERROR: parse error in array");
			    $$ = NULL;
			  }
	            ;


characters          :     CHARACTER_VALUE                  
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = CHARACTER_VALUE;
			    temp_cont->content.value = $1;
			    
			    templist = NULL;
			    $$ = templist = g_list_append(templist, temp_cont);	
			  }
		    
                    |     characters ',' CHARACTER_VALUE   
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = CHARACTER_VALUE;
			    temp_cont->content.value = $3;
			    
			    $$ = templist = g_list_append(templist, temp_cont);
			  }

                    |     characters ',' UNDEFINED   
                          {
			    container *temp_cont;

                            temp_cont = (container *) malloc(sizeof(container));
                            temp_cont->type_content = UNDEFINED;

                            $$ = templist = g_list_append(templist,temp_cont);
                          }  

                    |     UNDEFINED ',' characters   
                          {
			    container *temp_cont;

                            temp_cont = (container *) malloc(sizeof(container));
                            temp_cont->type_content = UNDEFINED;

                            $$ = templist = g_list_prepend(templist,temp_cont);
                          }
		    
                    |     characters ',' variable    
                          { 					     
                            container *temp_cont;
                            var_entry *var = NULL;

                            if (in_macro)
			      var = getmacrovar(macro_var_table, $3->var_name, num_macro);
				
			    if (var == NULL)
			      {
				/* la variabile  globale oppure  locale ad una regola */
				var = getvar(var_table, $3->var_name);
			      }
						  
			    assert( var != NULL );

                            if ((var->type_values != CHARACTER_VALUE) && (var->type_values != 0)) 
			      { 
				yyerror("ERROR: var is not compatible");
				YYERROR; 
			      }

			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = HANDLER;
			    temp_cont->content.handler = $3;
			    
			    $$ = templist = g_list_append(templist,temp_cont);
			  }     
	            
                    |     variable ',' characters    
                          {          
			    container *temp_cont;
			    var_entry *var = NULL;

                            if (in_macro)
			      var = getmacrovar(macro_var_table, $1->var_name, num_macro);
				
			    if (var == NULL)
			      {
				/* la variabile  globale oppure  locale ad una regola */
				var = getvar(var_table, $1->var_name);
			      }

			    assert( var != NULL );

			    if ((var->type_values != CHARACTER_VALUE) && (var->type_values != 0)) 
			      { 
				yyerror("ERROR: variabile is not compatible");
				YYERROR; 
			      }
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = HANDLER;
			    temp_cont->content.handler = $1;
			    
			    $$ = templist = g_list_prepend(templist,temp_cont);
			  }
                    |     characters ',' error
                          {
			    yyerror("ERROR: parse error in an element of array of characters"); 
			  }

                    |     error ',' characters
                          {
			    yyerror("ERROR: parse error in an element of array of characters");
			  }
		    ;


strings             :     STRING_VALUE                     
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = STRING_VALUE;
			    temp_cont->content.value = $1;
			    
			    templist = NULL;
			    $$ = templist = g_list_append(templist, temp_cont);
			  } 
		    
                    |     strings ',' STRING_VALUE         
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = STRING_VALUE;
			    temp_cont->content.value = $3;
			    
			    $$ = templist = g_list_append(templist, temp_cont);
			  }
		    
                    |     strings ',' UNDEFINED      
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = UNDEFINED;
			    
			    $$ = templist = g_list_append(templist,temp_cont);
			  }
		    
                    |     UNDEFINED ',' strings      
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = UNDEFINED;
			    
			    $$ = templist = g_list_prepend(templist,temp_cont);
			  }
		    
                    |     strings ',' variable       
                          {					     
			    container *temp_cont;
			    var_entry *var = NULL;

                            if (in_macro)
			      var = getmacrovar(macro_var_table, $3->var_name, num_macro);
				
			    if (var == NULL)
			      {
				/* la variabile  globale oppure  locale ad una regola */
				var = getvar(var_table, $3->var_name);
			      }
				
			    assert(var != NULL);

			    if ((var->type_values != STRING_VALUE) && (var->type_values != 0)) 
			      { 
				yyerror("ERROR: variabile is not compatible");
				YYERROR; 
			      }

			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = HANDLER;
			    temp_cont->content.handler = $3;
			    
			    $$ = templist = g_list_append(templist,temp_cont);
			  }
		    
                    |     variable ',' strings       
                          {          
			    container *temp_cont;
			    var_entry *var = NULL;

                            if (in_macro)
			      var = getmacrovar(macro_var_table, $1->var_name, num_macro);
				
			    if (var == NULL)
			      {
				/* la variabile  globale oppure  locale ad una regola */
				var = getvar(var_table, $1->var_name);
			      }

			    assert(var != NULL);
				
			    if ((var->type_values != STRING_VALUE) && (var->type_values != 0)) 
			      { 
				yyerror("ERROR: variabile is not compatible");
				YYERROR; 
			      }
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = HANDLER;
			    temp_cont->content.handler = $1;
			    
			    $$ = templist = g_list_prepend(templist,temp_cont);
			  }

                    |     strings ',' error
                          {
			    yyerror("ERROR: parse error in an element of array of strings"); 
			  }
                    |     error ',' strings
                          {
			    yyerror("ERROR: parse error in an element of array of strings"); 
			  }
	            ;


integers            :     an_integer                 
                          {
			    /* in an_integer ho creato il container */
			    templist = NULL;
			    $$ = templist = g_list_append(templist, $1);
			  }  
		    
                    |     integers ',' an_integer    
                          {
			    /* in an_integer ho creato il container */
			    $$ = templist = g_list_append(templist, $3);
			  }
		    
                    |     integers ',' UNDEFINED     
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = UNDEFINED;
			    
			    $$ = templist = g_list_append(templist,temp_cont);
			  } 
		    
                    |     UNDEFINED ',' integers     
                          {
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = UNDEFINED;
			    
			    $$ = templist = g_list_prepend(templist,temp_cont);
			  }
		    
                    |     integers ',' variable      
                          {					     
			    container *temp_cont;
			    var_entry *var = NULL;

                            if (in_macro)
			      var = getmacrovar(macro_var_table, $3->var_name, num_macro);
				
			    if (var == NULL)
			      {
				/* la variabile  globale oppure  locale ad una regola */
				var = getvar(var_table, $3->var_name);
			      }
				

			    if ((var->type_values != LONG_VALUE) && (var->type_values != 0)) 
			      { 
				yyerror("ERROR: variable is not compatible");
				YYERROR; 
			      }
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = HANDLER;
			    temp_cont->content.handler = $3;
			    
			    $$ = templist = g_list_append(templist,temp_cont);
			  }
		    
                    |     variable ',' integers      
                          {          
			    container *temp_cont;
			    var_entry *var = NULL;

                            if (in_macro)
			      var = getmacrovar(macro_var_table, $1->var_name, num_macro);
				
			    if (var == NULL)
			      {
				/* la variabile  globale oppure  locale ad una regola */
				var = getvar(var_table, $1->var_name);
			      }
				
			    if ((var->type_values != LONG_VALUE) && (var->type_values != 0))
			      { 
				yyerror("ERROR: variable is not compatible");
				YYERROR; 
			      }
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = HANDLER;
			    temp_cont->content.handler = $1;
			    
			    $$ = templist = g_list_prepend(templist,temp_cont);
			  }

                    |     integers ',' error
                          {
			    yyerror("ERROR: parse error in an element of array of integers"); 
			  }

		    |     error ',' integers
                          {
			    yyerror("ERROR: parse error in an element of array of integers"); 
			  }
		    ;


an_integer          :     LONG_VALUE     
			  {
			    /* creo un container per un an_integer */
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = LONG_VALUE;
			    temp_cont->content.value = $1;
			    
			    $$ = temp_cont;
			  }
		    
                    |     {in_flags = 1;}  flags  {in_flags = 0;}
			  {
			    /* creo un container per un flag */
			    container *temp_cont;
			    
			    temp_cont = (container *) malloc(sizeof(container));
			    temp_cont->type_content = SIMPLE_VALUE;
			    temp_cont->type_value = LONG_VALUE;
			    temp_cont->content.value = (long *) malloc(sizeof(long));
			    *((long *)(temp_cont->content.value)) = $2;
			    
			    $$ = temp_cont;
			  }         
   		    ;


flags               :     FLAG
                          {
			    int flag_value = lookup_flag($1);
			    
			    if (flag_value == -100000) 
			      { 
				yyerror("ERROR: FLAG doesn't exist");
				YYERROR; 
			      }
			    free($1);
			    
			    $$ = flag_value;
			  }

                    |     flags '|' FLAG
                          {
			    int flag_value = lookup_flag($3);
			    
			    if (flag_value == -100000) 
			      { 
				yyerror("ERROR: FLAG doesn't exist");
				YYERROR; 
			      }
			    free($3);
			    
			    $$ = $$ | flag_value;
			  }
                    ;


variable            :     VARIABLE      /* scalare    */ 
			  {
			    /* controllo se variabile esiste in global_vars o in local_vars: 
			       se esiste creo un handler nel quale imposto il riferimento */
			    handler_t *temp_handler;
			    
			    if (once_var) { 
			      yyerror("ERROR: var not declared");
			      YYERROR; 
			    }
			    
			    temp_handler = (handler_t *) malloc (sizeof(handler_t));
			    temp_handler->var_name = clone_string($1->name);
			    $$ = temp_handler;
			  }					 
		    
                    |     HASH_VAR      /* hash       */	 
		    	  {
			    /* come per le variabili scalari */
			    handler_t *temp_handler;
			    
			    if (once_var) { 
			      yyerror("ERROR: var not declared");
			      YYERROR; 
			    }
			    
			    temp_handler = (handler_t *) malloc (sizeof(handler_t));
			    temp_handler->var_name = clone_string($1->name);
			    $$ = temp_handler;
			  }
                    ;

%%

/*********************************** finalize_tree *****************************************/
		   
/*  utilizza la var. globale any_sysc  */
extern GList *any_sysc;

static int 
insert_in_any_sysc(GNode *node, gpointer data) 
     /* inserisce nella lista cui punta la variabile globale any_sysc il numero della
	chiamata di sistema del nodo correntemente esaminato;
        NOTA: funzione utilizzata da make_any_list() */
{
  event_node_t *event; 
  int sysc_num; 
  
  event = node->data; 
  assert(event->op == none); 
  
  sysc_num = event->simple_event->n_sysc; 

  if ((sysc_num == ANY) || (sysc_num == END_MARKER)) return 0;
  
  if (!g_list_find(any_sysc, GINT_TO_POINTER(sysc_num)))  
    { 
      any_sysc = g_list_append(any_sysc, GINT_TO_POINTER(sysc_num));
    } 
  
  return 0; 
} 


static void
make_any_list(GNode *tree_with_event_node)
     /* crea una lista contenente i numeri delle chiamate di sistema contenute 
	nell'albero passatogli come primo argomento;
        NOTA: funzione utilizzata da finalize_tree() */
{ 
  g_node_traverse(tree_with_event_node, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, insert_in_any_sysc, NULL);
}


static int
visita_tree (GNode *node, gpointer data)
     /* stampa il nome della syscall contenuta nel nodo esaminato;
	NOTA: funzione utilizzata da finaliza_tree */
{
  event_node_t *father;
  
  father = node->data;
  switch (father->op) {
  case   none:
    printf("\n ---%s--- ", get_sysc_name(father->simple_event->n_sysc) );
    break; 
  default:
    ; 
  }

  return 0;
}


static int
exist_in_list(int num, GList *list)
     /* controlla che la syscall corrispondente a num compaia nella lista
        NOTA: funzione utilizzata da make_tree_for_any() */
{
  GList *now;
  
  now = g_list_first(list);
  
  while (now != NULL) {
    if (GPOINTER_TO_INT(now->data) == num) return 1;
    now = g_list_next(now);
  }
  
  return 0;
}


static GNode *
make_tree_for_any(GList *any_list, GList *condition[2], GList *action[2], GList *list)
     /* crea l'albero che mette in OR tutte le chiamate di sistema che compaiono nel file delle specifiche
	e l'evento DELTA; l'albero sostituir l'ANY;
        NOTA: funzione utilizzata da remove_any() */
{
  GList *now; 
  GNode *tree, *prev = NULL, *next = NULL;
  event_node_t *father, *father2, *right_child, *left_child;
  
  now = g_list_first(any_list);
  while (now != NULL) {
    
    /* la syscall negata non la metto nell'ANY */
    if ( exist_in_list( GPOINTER_TO_INT(now->data) , list ) ) 
      {
	now = g_list_next(now); 
	continue;
      }
    
    father = (event_node_t *) malloc(sizeof(event_node_t));
    father->op = sequencing;
    
    left_child = (event_node_t *) malloc(sizeof(event_node_t));
    left_child->op = none;
    left_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
    (*left_child->simple_event).n_sysc = GPOINTER_TO_INT(now->data);
    (*left_child->simple_event).num_event = 0;
    (*left_child->simple_event).context = intro;
    (*left_child->simple_event).fl.negated_condition = 0;
    (*left_child->simple_event).empty_follow = 0;
    /* ATTENZIONE: tutte le syscall in OR dell'ANY puntano al solito spazio allocato per condizioni e azioni */
    if (condition != NULL)  (*left_child->simple_event).condition = condition[0];
    else (*left_child->simple_event).condition = NULL;
    if (action != NULL)     (*left_child->simple_event).action = action[0];
    else (*left_child->simple_event).action = NULL;
    
    right_child = (event_node_t *) malloc(sizeof(event_node_t));
    right_child->op = none;
    right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
    (*right_child->simple_event).n_sysc = GPOINTER_TO_INT(now->data);
    (*right_child->simple_event).num_event = 0;
    (*right_child->simple_event).context = exiting;
    (*right_child->simple_event).fl.negated_condition = 0;
    (*right_child->simple_event).empty_follow = 0;
    /* ATTENZIONE: tutte le syscall in OR dell'ANY puntano al solito spazio allocato per condizioni e azioni */
    if (condition != NULL)  (*right_child->simple_event).condition = condition[1];
    else (*right_child->simple_event).condition = NULL;
    if (action != NULL)     (*right_child->simple_event).action = action[1];
    else (*right_child->simple_event).action = NULL;
    
    next = g_node_new(father);
    g_node_append_data(next, left_child);
    g_node_append_data(next, right_child);      
    
    if (prev == NULL) prev = next;
    else  
      {
	father2 = (event_node_t *) malloc(sizeof(event_node_t)); 
	father2->op = alternation;
	
	tree = g_node_new(father2);
	g_node_append(tree, prev);
	g_node_append(tree, next);
	prev = tree;
      }
    
    now = g_list_next(now); 
  }

  /* creazione dell'evento DELTA */

  father = (event_node_t *) malloc(sizeof(event_node_t));
  father->op = sequencing;
  
  left_child = (event_node_t *) malloc(sizeof(event_node_t));
  left_child->op = none;
  left_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
  (*left_child->simple_event).n_sysc = DELTA;
  (*left_child->simple_event).num_event = 0;
  (*left_child->simple_event).context = intro;
  (*left_child->simple_event).fl.negated_condition = 0;
  (*left_child->simple_event).empty_follow = 0;

  /* ATTENZIONE: tutte le syscall in OR dell'ANY puntano al solito spazio allocato per condizioni e azioni */
  if (condition != NULL)  (*left_child->simple_event).condition = condition[0];
  else (*left_child->simple_event).condition = NULL;
  if (action != NULL)     (*left_child->simple_event).action = action[0];
  else (*left_child->simple_event).action = NULL;
  
  right_child = (event_node_t *) malloc(sizeof(event_node_t));
  right_child->op = none;
  right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
  (*right_child->simple_event).n_sysc = DELTA;
  (*right_child->simple_event).num_event = 0;
  (*right_child->simple_event).context = exiting;
  (*right_child->simple_event).fl.negated_condition = 0;
  (*right_child->simple_event).empty_follow = 0;
  
  /* ATTENZIONE: tutte le syscall in OR dell'ANY puntano al solito spazio allocato per condizioni e azioni */
  if (condition != NULL)  (*right_child->simple_event).condition = condition[1];
  else (*right_child->simple_event).condition = NULL;
  if (action != NULL)     (*right_child->simple_event).action = action[1];
  else (*right_child->simple_event).action = NULL;
  
  next = g_node_new(father);
  g_node_append_data(next, left_child);
  g_node_append_data(next, right_child);      
  
  if (prev == NULL) prev = next;
  else  
    {
      father2 = (event_node_t *) malloc(sizeof(event_node_t)); 
      father2->op = alternation;
      
      tree = g_node_new(father2);
      g_node_append(tree, prev);
      g_node_append(tree, next);
      prev = tree;
    }
  
  return prev;
}


static int
remove_any(GNode *node, gpointer data)
     /* rimouove il simple_event contenente l'any e lo sostituisce con l'OR di tutte le chiamate di 
	sistema trovate nel file delle specifiche e dell'evento DELTA;
        NOTA: funzione utilizzata da finalize_tree() */
{
  event_node_t *intro_event, *exit_event;      
  int pos;
  GNode *tree_to_remove = (*node).parent;
  GNode *father_of_ttr = (*tree_to_remove).parent;  /* il padre del nodo da sostituire con il tree delle syscall ANY */  
  GNode *node_any_exit;
  GList *condition[2], *action[2], *minus_s;
  GNode *tree = NULL;
  
  intro_event = node->data;
  switch (intro_event->op) {
  case   none:
    if (intro_event->simple_event->n_sysc == ANY) 
      {
	assert((node_any_exit = g_node_next_sibling(node)) != NULL);
	condition[0] = intro_event->simple_event->condition;
	action[0]    = intro_event->simple_event->action;
	
	exit_event = node_any_exit->data;   
        assert(exit_event->simple_event->context == exiting);
	condition[1] = exit_event->simple_event->condition;
	action[1]    = exit_event->simple_event->action;

	assert (intro_event->simple_event->fl.minus_sysc == exit_event->simple_event->fl.minus_sysc);
	minus_s = intro_event->simple_event->fl.minus_sysc;
	
	
	free(intro_event->simple_event);
	free(exit_event->simple_event);
	free(intro_event);
	free(exit_event);
	pos = g_node_child_position (father_of_ttr, tree_to_remove);
	g_node_unlink (tree_to_remove);
	free(tree_to_remove->data);
	g_node_destroy (tree_to_remove);
	
	if ((tree = make_tree_for_any(any_sysc, condition, action, minus_s)) != NULL)  
	  g_node_insert (father_of_ttr, pos, tree);
	else 
	  { 
	    /* 
	       GNode *new = (*father_of_ttr).parent;
	       GNode *son = g_node_first_child(father_of_ttr);
	       assert(son != NULL);  se non ci sono syscall negate si blocca qui 
	       
	       g_node_unlink (son);
	       g_node_unlink (father_of_ttr);
	       free(father_of_ttr->data);
	       g_node_destroy (father_of_ttr);
	       
	       g_node_prepend (new, son);
	       potare l'albero a seconda dei casi */
	    printf("\nATTENZIONE: all'evento ANY non e' stato sostituito nessun albero di eventi\n");
	  }
	
	/* dealloco la lista delle syscall da togliere a questo ANY in quanto non serve piu' */
	g_list_free (minus_s);
      }
    break;
    default:
       ;  
  }
  
  return 0;
}


/*
static int
modify_event_in_handler (GNode *node, gpointer data)
      rinumera il numero dell'evento di un handler;
        NOTA: funzione utilizzata da modify_num_event() 
{
  arg_node_t *arg_node;  
  
  arg_node = node->data;
  if (arg_node->funct == NULL) 
    if (arg_node->arg->type_content == HANDLER)
      if (arg_node->arg->content.handler->var_name == NULL)
	{
	   //assert (arg_node->arg->content.handler->num_event == GPOINTER_TO_INT(data)); 
	   //arg_node->arg->content.handler->num_event = num_event;  
	} 
  
  return 0;
}
*/


/* variabili globali utilizzate da modify_num_event */
extern int num_event;

static int
modify_num_event(GNode *node, gpointer data)
     /* rinumera gli eventi contenuti nell'albero passatogli;
        NOTA: funzione utilizzata da finalize_tree() */
{
  event_node_t *event;
  /* int actual_event;
     GList *condition, *action, *now; */
  /* int num_event e' variabile globale */  
  
  event = node->data;
  switch (event->op) {
  case   none:
    
    num_event++;
    /* actual_event = event->simple_event->num_event;
       condition    = event->simple_event->condition;
       action       = event->simple_event->action; */
    
    /*
    now = g_list_first(condition);  entro nelle condizioni e modifico gli eventuali segnaposti 
    while (now != NULL) {
      g_node_traverse (now->data, G_IN_ORDER, G_TRAVERSE_ALL, -1, 
		       modify_event_in_handler, GINT_TO_POINTER(actual_event) );
      now = g_list_next(now);
    }

    now = g_list_first(action);  entro nelle azioni e modifico gli eventuali segnaposti 
    while (now != NULL) {
      g_node_traverse (now->data, G_IN_ORDER, G_TRAVERSE_ALL, -1, 
		       modify_event_in_handler, GINT_TO_POINTER(actual_event) );
      now = g_list_next(now);
    }
    */
    
    event->simple_event->num_event = num_event;
    break;
   default:
     ;  
  }
  
  return 0;
}


static int
print_event_label(GNode *node, gpointer data)
     /* stampa il numero dell'evento e l'argomento della syscall cui si riferisce l'handler
	eventualmente contenuto nel nodo esaminato;
	NOTA: funzione utilzzata da print_label
     */
{
  arg_node_t *arg_node;  
  
  arg_node = node->data;
  if (arg_node->funct == NULL) 
    if (arg_node->arg->type_content == HANDLER)
      if (arg_node->arg->content.handler->var_name == NULL)
	{
	  printf("\n---->type_label %ld  n_arg %ld\n",arg_node->arg->content.handler->type_label,
		 arg_node->arg->content.handler->label_ref.n_arg);
	} 
  
  return 0;
}


static int
print_event_labels(GNode *node, gpointer data)
     /* stampa le label relative alla syscall contenuta nel nodo esaminato;
	NOTA: funzione utilizzata da finalize_tree() */
{
  event_node_t *event;
  int actual_event;
  GList *condition, *action, *now;
   
  event = node->data;
  switch (event->op) {
  case   none:
    
    condition    = event->simple_event->condition;
    action       = event->simple_event->action;
    
    now = g_list_first(condition); /* entro nelle condizioni */
    while (now != NULL) {
      g_node_traverse (now->data, G_IN_ORDER, G_TRAVERSE_ALL, -1, 
		       print_event_label, GINT_TO_POINTER(actual_event) );
      now = g_list_next(now);
    }
    
    now = g_list_first(action); /* entro nelle azioni */
    while (now != NULL) {
      g_node_traverse (now->data, G_IN_ORDER, G_TRAVERSE_ALL, -1, 
		       print_event_label, GINT_TO_POINTER(actual_event) );
      now = g_list_next(now);
    }
    
    break;
   default:
     ;  
  }
  
  return 0;
}


void
finalize_tree(GNode *parse_tree)
{
  /* crea nella var globale any_sysc la lista con tutti i tipi di syscall trovati in tree */
  make_any_list(parse_tree);
  
  /* printf("\nSequenza syscall in tutto il file\n");
     g_node_traverse (tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, visita_tree, NULL); */
  
  /* sostituisco all'ANY l'OR delle syscall trovate nel file */ 			     
  g_node_traverse(parse_tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, remove_any, NULL);
  
  /* dealloco any_sysc */
  g_list_free(any_sysc);
  
  /* printf("\nSequenza syscall in tutto il file (senza ANY)\n");
     g_node_traverse (tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, visita_tree, NULL); */

  /* aggiorno in ogni simple_event il numero dell'evento */
  g_node_traverse(parse_tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, modify_num_event, NULL);
  
  /* printf("\nSEGNAPOSTI\n");
     g_node_traverse (tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, print_event_labels, NULL); */
}





/************************** make_tree_for_arg_value **************************************/


GNode *
make_tree_for_arg_value(container *cont, int arg_number)
     /* crea il sottoalbero che rappresenta la condizione di uguaglianza tra il valore contenuto nel container 
        e l'argomento "arg_number"-esimo della syscall correntemente esaminata */
{
  container *temp_cont;
  GNode *temp_tree;
  arg_node_t *root, *child1, *child2;
  handler_t *temp_handler; 
  
  /* creo un handler_t e vi inserisco i dati del segnaposto */
  temp_handler = (handler_t *) malloc (sizeof(handler_t));
  temp_handler->var_name = NULL;
  
  
  temp_handler->type_label = find_type(sysc_num, arg_number);
  temp_handler->label_ref.n_arg = arg_number - 1;
  
  /* creo il container contenente l'handler_t sopra creato */
  temp_cont = (container *) malloc(sizeof(container));
  temp_cont->type_content = HANDLER;
  temp_cont->content.handler = temp_handler;
  
  /* devo creare 3 arg_node_t : la funzione equal alla radice handler e container come figli */
  root = (arg_node_t *) malloc(sizeof(arg_node_t));
  root->funct = (void *(*)()) eq;
  root->arg = NULL;
  child1 = (arg_node_t *) malloc(sizeof(arg_node_t));
  child1->arg = temp_cont;
  child1->funct = NULL;
  child2 = (arg_node_t *) malloc(sizeof(arg_node_t));
  child2->arg = cont;              /* il container che mi arriva da arg_value */
  child2->funct = NULL;

  /* creo il mini albero */
  temp_tree = g_node_new(root);
  g_node_append_data(temp_tree, child1);
  g_node_append_data(temp_tree, child2);

  //printf("%x**\n", root->funct);
  if (root->funct == (void *(*)())eq) root->funct = (*root->funct)(g_node_first_child(temp_tree)->data, 
								   g_node_last_child(temp_tree)->data, num_macro);
  //printf("%x\n\n", root->funct);
  
  return temp_tree;
}




/*********************************** min_num_event ***************************************/



static int
take_min_event (GNode *node, gpointer ret_min)
     /* inserisce nell'area di memoria cui punta ret_min, il numero di evento minimo
	degli eventi contenuti nell'albero cui appartiene il nodo esaminato
        NOTA: funzione utilizzata da min_num_event() */
{
  event_node_t *father;
  
  father = node->data;
  assert(father->op == none);
  
  if (!in_declarations)
    {
      if (father->simple_event->num_event > 0) { 
	*((int *)ret_min) = father->simple_event->num_event;
	return 1;  /* la visita viene subito interrotta */
      }
    } 
  
  if (in_declarations)
    { 
      assert(father->simple_event->num_event < 0);
      if (father->simple_event->num_event > -10000) {
	*((int *)ret_min) = father->simple_event->num_event;
        return 1;  /* la visita viene subito interrotta */
      }
    }
  
  return 0; 
}



int
min_num_event(GNode *tree)
     /* restituisce il numero evento minimo trovato in tree (tree e' l'evento di sinistra di   event OR event) */
{
  int ret = 0;
  
  g_node_traverse (tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, take_min_event, &ret);  
  
  return ret;
}





/********************************* max_num_event *****************************************/

static int
take_max_event (GNode *node, gpointer ret_max)
     /* inserisce nell'area di memoria cui punta ret_min, il numero di evento massimo
	degli eventi contenuti nell'albero cui appartiene il nodo esaminato
        NOTA: funzione utilizzata da max_num_event() */
{
  event_node_t *father;
  
  father = node->data;
  assert(father->op == none);
  
  if (!in_declarations) {
    if (father->simple_event->num_event > *((int *)ret_max))
      *((int *)ret_max) = father->simple_event->num_event;
  }
  
  if (in_declarations) {
    assert(father->simple_event->num_event < 0); 
    if ((father->simple_event->num_event > -10000) && (father->simple_event->num_event < *((int *)ret_max)))
      *((int *)ret_max) = father->simple_event->num_event;
  }
  
  return 0; /* la visita continua fino all'ultimo nodo */
}


int
max_num_event(GNode *tree)
     /* restituisce il numero evento massimo trovato in tree (tree e' l'evento di sinistra di   event OR event) */
{
  int ret = 0;
  
  g_node_traverse (tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, take_max_event, &ret);  
  
  return ret;
}




/************************ check_vars_in_or_combination ***********************************/

static int
go_in_condition (GNode *node, gpointer data)
     /* controlla che l'eventuale variabile contenuta nel nodo correntemente esaminato sia
	correttamente definita;
	NOTA: funzione utilizzata da search_vars() */
{
  arg_node_t *arg_node;
  char *var_name;
  int *temp = (int *)data;
  int min = temp[0]; int max = temp[1]; int actual = temp[2];int actual_context = temp[3];
  /*printf("\nmin %d max %d  ora %d\n", min, max, actual);*/
  
  arg_node = node->data;
  
  if (arg_node->funct == NULL) 
    if (arg_node->arg->type_content == HANDLER)
      if ((var_name = arg_node->arg->content.handler->var_name) != NULL) /* ho trovato una var in una condizione */
	{
	  scope_el_t *first_decl, *last_decl, *prev_decl, *temp_scope;
	  int num;
	  int assign_in_actual = 0;
	  
	  if (min > 0)
	    {
	      num = num_seq;
	      first_decl = firstscope(var_name, num);
	      
	      printf("\nDEBUG 1 go_in_condition");
	      if ((var_name[0] == '+') || (var_name[0] == '-')) return 0;
	      
	      printf("\nDEBUG 2 go_in_condition: %s %d\n", var_name, num); fflush(stdout);
	      if (first_decl) 
		{
		  last_decl = lastscope(var_name, num);   
		  while ( (prev_decl = prevscope()) )
		    {
		      if ((prev_decl->num_event) < actual) break; 
		    }
		  if (prev_decl == NULL) prev_decl = last_decl;
		  
		  if ( (temp_scope = nextscope()) ) if (temp_scope->num_event == actual) assign_in_actual++; 
		  if (temp_scope == NULL) temp_scope = last_decl; 
		  
		  /*printf("\n\nfirst %d . last %d . prev %d . context %d.\n",
		    first_decl->num_event,last_decl->num_event,prev_decl->num_event,temp_scope->context);*/
		  if ( ((first_decl->num_event >= min) && (first_decl->num_event <= max)) 
		       && ( (last_decl->num_event <= max) 
			    || ((last_decl->num_event > actual) && (prev_decl->num_event <= max) && (!assign_in_actual)) 
			    || ((assign_in_actual) && (prev_decl->num_event <= max) 
				&& ((temp_scope->context != intro) || (actual_context != exiting))) ) )
		    {
		      yyerror("ERROR: var declaration in ORed events (events seq)");
		      printf("\n%s\n",var_name);
		      // YYERROR;
		      exit(-1);
		    }
		}
	      else 
		assert( firstscope(var_name, 0) != NULL ); /* la variabile e' globale */ 
	    }
	  
	  if (min < 0) /* OR interno ad una macro */
	    {
	      num = -num_macro;
	      first_decl = firstscope(var_name, num);
	      
	      printf("\nDEBUG 1 go_in_condition");
	      if ((var_name[0] == '+') || (var_name[0] == '-')) return 0;
	      
	      printf("\nDEBUG 2 go_in_condition: %s %d\n", var_name, num); fflush(stdout);
	      if (first_decl)
		{
		  last_decl = lastscope(var_name, num);
		  while ( (prev_decl = prevscope()) )
		    {
		      if ((prev_decl->num_event) < -actual) break; 
		    }
		  if (prev_decl == NULL) prev_decl = last_decl;
		  
		  if ( (temp_scope = nextscope()) ) if (temp_scope->num_event == -actual) assign_in_actual++;
		  if (temp_scope == NULL) temp_scope = last_decl;
		  
		  /*printf("\n\nfirst %d . last %d . prev %d .\n",
		    first_decl->num_event,last_decl->num_event,prev_decl->num_event);*/
		  if ( ((-first_decl->num_event <= min) && (-first_decl->num_event >= max)) 
		       && ( (-last_decl->num_event >= max) 
			    || ((-last_decl->num_event < actual) && (-prev_decl->num_event >= max) && (!assign_in_actual))
			    || ((assign_in_actual) && (-prev_decl->num_event >= max) 
				&& ((temp_scope->context != intro) || (actual_context != exiting))) ) )
		    {
		      yyerror("ERROR: var %s declaration in ORed events (macro)",var_name);
		      /* YYERROR; */
		      exit(-1);
		    }
		}
	      else 
		assert( firstscope(var_name, 0) != NULL ); /* la variabile e' globale */ 
	    }
	} 
  
  return 0;
}


static int
go_in_action (GNode *node, gpointer data)
     /* controlla che l'eventuale variabile contenuta nel nodo correntemente esaminato sia
	correttamente definita;
	NOTA: funzione utilizzata da search_vars() */
{
  arg_node_t *arg_node;
  char *var_name;
  int *temp = (int *)data;
  int min = temp[0]; int max = temp[1]; int actual = temp[2]; int actual_context = temp[3];int num_stat = temp[4];
  /*printf("\nmin %d max %d  ora %d\n", min, max, actual);*/
  
  arg_node = node->data;
  
  if (arg_node->funct == NULL) 
    if (arg_node->arg->type_content == HANDLER)
      if (((var_name = arg_node->arg->content.handler->var_name) != NULL) && 
	  ((g_node_child_position(node->parent, node) >= 1)  
	   ||  
	   /* (((arg_node_t *) node->parent->data)->funct != (void *(*)()) assign ))) */
	   ( g_node_depth(node) >= 3 )) )
	/* ho trovato una var a destra di un assegnamento o come parametro di una funzione */
	{
	  scope_el_t *first_decl, *last_decl, *prev_decl, *temp_scope;
	  int num;
	  int assign_in_actual = 0;
	  
	  if (min > 0)
	    {
	      num = num_seq;
	      first_decl = firstscope(var_name, num);

	      if ((var_name[0] == '+') || (var_name[0] == '-')) return 0;
	      if (first_decl) 
		{
		  last_decl = lastscope(var_name, num);
		  while ( (prev_decl = prevscope()) )
		    {
		      if ((prev_decl->num_event) < actual) break; 
		    }
		  if (prev_decl == NULL) prev_decl = last_decl; 
		  
		  if ( (temp_scope = nextscope()) ) if (temp_scope->num_event == actual) assign_in_actual++; 
		  if (temp_scope == NULL) temp_scope = last_decl; 

		  /*printf("\n\nfirst %d . last %d . prev %d .\n",
		    first_decl->num_event,last_decl->num_event,prev_decl->num_event);*/
		  if ( ((first_decl->num_event >= min) && (first_decl->num_event <= max)) 
		       && ( (last_decl->num_event <= max) 
			    || ((last_decl->num_event > actual) && (prev_decl->num_event <= max) && (!assign_in_actual))
			    || ((assign_in_actual) && (prev_decl->num_event <= max) 
				&& (temp_scope->num_statement > num_stat) && (temp_scope->context >= actual_context)) ) )
		    {
		      yyerror("ERROR: var %s declaration in ORed events ACTION (events seq)", var_name);
		      exit(-1);
		    }
		}
	      else 
		assert( firstscope(var_name, 0) != NULL ); /* la variabile e' globale */ 
	    }
	  
	  if (min < 0) /* OR interno ad una macro */
	    {
	      num = -num_macro;
	      first_decl = firstscope(var_name, num);
	      
	      if ((var_name[0] == '+') || (var_name[0] == '-')) return 0;
	      if (first_decl)
		{
		  last_decl = lastscope(var_name, num);
		  while ( (prev_decl = prevscope()) )
		    {
		      if ((prev_decl->num_event) < -actual) break; 
		    }
		  if (prev_decl == NULL) prev_decl = last_decl;
		  
		  if ( (temp_scope = nextscope()) ) if (temp_scope->num_event == -actual) assign_in_actual++;
		  if (temp_scope == NULL) temp_scope = last_decl;

		  /*printf("\n\nfirst %d . last %d . prev %d .\n",
		    first_decl->num_event,last_decl->num_event,prev_decl->num_event);*/
		  if ( ((-first_decl->num_event <= min) && (-first_decl->num_event >= max)) 
		       && ( (-last_decl->num_event >= max) 
			    || ((-last_decl->num_event < actual) && (-prev_decl->num_event >= max) && (!assign_in_actual))
			    || ((assign_in_actual) && (-prev_decl->num_event >= max) 
				&& (temp_scope->num_statement > num_stat) && (temp_scope->context >= actual_context)) ) )
		    {
		      yyerror("ERROR: var %s declaration in ORed events ACTION (macro)", var_name);
		      exit(-1);
		    }
		}
	      else 
		assert( firstscope(var_name, 0) != NULL ); /* la variabile e' globale */ 
	    }
	} 
  
  return 0;
}


static int
search_vars(GNode *node, gpointer min_max)
     /* cerca le variabili nelle condizioni e nelle azioni dell'evento contenuto
	nel nodo correntemente esaminato e controlla che esse siano correttamente 
	definite;
	NOTA: funzione utilizzata da check_vars_in_or_combination() */
{
  event_node_t *event;
  GList *condition, *action, *now;
  int *temp = (int *)min_max;
  int num_stat = 0;

  event = node->data;
  assert (event->op == none);
  
  condition = event->simple_event->condition;
  action = event->simple_event->action;
  temp[2] = event->simple_event->num_event;
  temp[3] = event->simple_event->context;

  //printf("\nDEBUG in search_vars: %d, %d", event->simple_event->n_sysc, event->simple_event->num_event);
  
  if ((!in_declarations) && (temp[2] < 0)) return 0; /* non controllo le var. nelle condizioni e azioni delle macro 
							(gia' controllate nelle dichiarazioni globali) */
  if ((in_declarations) && (temp[2] < -10000)) return 0; /* non controllo le var. nelle condizioni e azioni delle
							    macro in macro (gia' controllate prima della macro 'grande') */
  
  
  now = g_list_first(condition); /* entro nelle condizioni alla ricerca di var */
  while (now != NULL) {
    g_node_traverse (now->data, G_IN_ORDER, G_TRAVERSE_ALL, -1, go_in_condition, min_max); 
    now = g_list_next(now);
  }
  
  now = g_list_first(action);  /* entro nelle azioni alla ricerca di var a destra di assegnamenti o 
				  fra parametri di funzioni */
  while (now != NULL) {
    num_stat++;
    temp[4] = num_stat;
    g_node_traverse (now->data, G_IN_ORDER, G_TRAVERSE_ALL, -1, go_in_action, min_max); 
    now = g_list_next(now);
  }

  return 0;
}


void
check_vars_in_or_combination(GNode *tree, int min, int max)
     /* funzione utilizzata sull'evento di destra di   event OR event 
	per controllare che non vi compaiano variabili non definite */
{
  int min_max[5];
  min_max[0] = min;
  min_max[1] = max;
  min_max[2] = 0; /* conterra' il numero dell'evento */
  min_max[3] = 0; /* conterra' il contesto nel quale si trova la var da verificare */
  min_max[4] = 0; /* conterra' il numero sequenziale delle azioni di un evento (utilizzato solo nel caso delle azioni) */
  
  g_node_traverse (tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, search_vars, min_max);
  
}




/******************************* modify_macro_num_events *********************************/


/* variabili globali utilizzate dalla funzione change_mme() */
extern int in_declarations;
extern int counter_in_macro;
extern int counter;

static int
change_mme(GNode *node, gpointer data)
     /* rinumera l'evento countenuto nel nodo correntemente esaminato
	con il criterio descritto di seguito in corrispondenza della funzione 
	chiamante modify_macro_num_events() */
{
  event_node_t *event;
  int num;

  event = node->data;
  switch (event->op) 
    {
    case   none:
      
      assert(event->simple_event != NULL);
      
      /* assumo per ogni evento in entrata ci sia il corrispondente in uscita */
      if ( event->simple_event->context == intro )
	num = in_declarations ? -(++counter_in_macro) : ++counter;
      else 
	num = in_declarations ? -counter_in_macro : counter;
      
      event->simple_event->num_event = num;
      break;
    default:
      ;
    }
  
  return 0;
}


void
modify_macro_num_events(GNode *macro_tree)
     /* rinumera gli eventi contenuti nell'albero di una macro passatogli:
	se siamo nella dichiarazione di una macro li rinumera a partire da -(counter_in_macro + 1)
	e counter_in_macro assume il valore assoluto dell'ultimo evento, 
	altimenti li rinumera a partire da counter + 1, e counter assume il valore dell'ultimo evento */
{
  g_node_traverse(macro_tree, G_IN_ORDER, G_TRAVERSE_LEAFS, -1, change_mme, NULL);
}




/***************************** find_simple_event *****************************************/


gboolean
find_simple_event(GNode *node, gpointer data) /* node contiene un elmento di tipo event_node_t
						 data e' un puntatore ad un GList per restituire 
						 la lista di simple_event trovati */
{
  if ( (* (event_node_t *) node->data).op == none )
    if ( (* (event_node_t *) node->data).simple_event != NULL )
      * (GList **) data = g_list_prepend( * (GList **) data, (* (event_node_t *) node->data).simple_event);

  return FALSE;
}





/************************** append_group_condition ***************************************/


static gboolean 
find_handler(GNode *node, gpointer data)
     /* controlla che il nodo correntemente esaminato contenga un handler, ed in caso affermativo
	lo inserisce nella lista cui punta *data;
        NOTA: funzione utilizzata da append_group_condition() */
{
  if ( (* (arg_node_t *) node->data).arg != NULL )
    if ( (* (arg_node_t *) node->data).arg->type_content == HANDLER )  
      * (GList **) data = g_list_prepend( * (GList **) data, (* (arg_node_t *) node->data).arg->content.handler);
  
  return FALSE;
}


/* variabili globali utilizzate dalla funzione append_group_condition() */
extern int in_macro;
extern label_table_t macro_label_table;
extern label_table_t label_table;

GList *
append_group_condition(GList *a_single_condition, simple_event_t *se, follow_array_t afs)
     /* controlla che le variabili e i label contenuti in a_single_condition siano correttamente
	definiti e in caso affermativo appende a_single_condition alle condizioni del simple event se */
{  
  GList *a_condition_clone;
  GList *hel, *handler_list_clone;
  label_value_t *fe;
  int ne;
  short error = 0;

  GNode *tree_cond;
  arg_node_t *argnode;
  
  hel = handler_list_clone = NULL;
  ne = se->num_event;

  a_condition_clone = clone_condition_or_action(a_single_condition);
  tree_cond = (GNode *) a_condition_clone->data;
  argnode = (arg_node_t *)tree_cond->data;
  
  g_node_traverse((GNode *) a_condition_clone->data, G_POST_ORDER, 
		  G_TRAVERSE_LEAFS, -1, find_handler, &handler_list_clone);
 
  for ( hel = handler_list_clone; hel; hel = g_list_next(hel) )
    {
      if ( ( * (handler_t *) hel->data).var_name != NULL )
	{
	  /* variabile */ 
	  scope_el_t *scope, *first_scope;
	  int num;
	  
	  if (in_macro)
	    num = -num_macro;
	  else
	    num = num_seq;
	  
	  first_scope = firstscope(((handler_t *) hel->data)->var_name, num);

	  if ( first_scope )
	    {
	      	      
	      if ( abs(first_scope->num_event) >= abs(afs.follow_array[1].evid.num_event) )
		{
		  /* la variabile e' definita per la prima volta all'interno della combinazione di eventi */
		  error = 1;
		  scope = lastscope(((handler_t *) hel->data)->var_name, num);
		  for ( ; scope; scope = prevscope() )
		    if ( is_descendant(afs, scope->num_event, scope->context, ne, se->context) )
		      {
			error = 0;
			break;
		      }
		}
	      /* else
		 -- la variabile e' definita prima della combinazione di eventi in esame:
		 -- assumo che la variabile sia definita anche all'interno della combinazione */
	    }
	  else 
	    {
	      assert( firstscope(((handler_t *) hel->data)->var_name, 0) != NULL );
	      /* se la variabile non e' definita ne' localmente ne' globalmente vi  un errore nel programma */
	    }
	  
	  if ( error )
	    {
	      char *context = se->context == intro ? "intro" : "exit";
	      char *syscall_name = get_sysc_name(se->n_sysc);

	      yyerror("ERROR: variable %s undefined in %s condition of %s", ( * (handler_t *) hel->data).var_name, context, syscall_name);
	      break;
	    }
	  
	}
      else 
	{
	  /* segnaposto */
	  if (in_macro)
	    {
	      assert( ne < 0 ); 
	      fe = get_label(macro_label_table, ((handler_t *) hel->data)->label_ref.label_name, ne);
	    }
	  else
	    {
	      assert( ne > 0 );
	      fe = get_label(label_table, ((handler_t *) hel->data)->label_ref.label_name, ne);
	    }
	  
	  if (fe)
	    {
	      ((handler_t *) hel->data)->label_ref.n_arg = fe->n_arg;
	      ((handler_t *) hel->data)->type_label = fe->type_label;
	    }
	  else
	    {
	      char *context = se->context == intro ? "intro" : "exit";
	      char *syscall_name = get_sysc_name(se->n_sysc);
	      
	      yyerror("ERROR: label %s undefined in %s condition of %s", ( * (handler_t *) hel->data).label_ref.label_name, context, syscall_name );
	      
	      error = 1;
	    }
	}
    }  /* end for */
  
  g_list_free(handler_list_clone);

  if (error)
    {
      free_condition_or_action(a_condition_clone);
      return NULL;
    }

  /* appende la condition clone alla simple_event */
  se->condition = g_list_concat(se->condition, a_condition_clone);
  
  argnode->funct = (*argnode->funct)(g_node_first_child(tree_cond)->data, g_node_last_child(tree_cond)->data, num_macro);
  return a_condition_clone;
}





/************************* append_group_action ****************************************/


static gboolean
find_right_handler(GNode *node, gpointer data)
     /*  controlla che il nodo correntemente esaminato contenga un handler e che sia il figlio destro
	 di un nodo contenente la funzione di assegnamento; in caso affermativo	lo inserisce 
	 nella lista cui punta *data;
	 NOTA: funzione utilizzata da append_group_action() */
{
  if ( ((arg_node_t *) node->data)->arg != NULL )
    if ( ((arg_node_t *) node->data)->arg->type_content == HANDLER )
      /* if ( !(node->parent->children == node && ((arg_node_t *) node->parent->data)->funct == (void *(*)()) assign) ) */
      /* quando un nodo si trova a profondita' 2 riguarda un assegnamento, a profondita' > 2 e' un arg di funzione */
      if ( !(node->parent->children == node && g_node_depth(node) == 2) )
	{
	  /* node non e' il membro a sinistra dell'assegnamento */
	  * (GList **) data = g_list_prepend( * (GList **) data, (* (arg_node_t *) node->data).arg->content.handler);
	}
  
  return FALSE;
}


static gboolean
find_left_handler(GNode *node, gpointer data) 
     /*  controlla che il nodo correntemente esaminato contenga un handler e che sia il figlio sinistro
	 di un nodo contenente la funzione di assegnamento; in caso affermativo	lo inserisce 
	 nella lista cui punta *data;
	 NOTA: funzione utilizzata da append_group_action() */
{
  if ( ((arg_node_t *) node->data)->arg != NULL )
    if ( ((arg_node_t *) node->data)->arg->type_content == HANDLER )
      /* if (node->parent->children == node && ((arg_node_t *) node->parent->data)->funct == (void *(*)()) assign) */
      /* quando un nodo si trova a profondita' 2 riguarda un assegnamento, a profondita' > 2 e' un arg di funzione */
      if (node->parent->children == node && g_node_depth(node) == 2)
	{
	  /* node e' il membro a sinistra dell'assegnamento */
	  * (GList **) data = g_list_prepend( * (GList **) data, (* (arg_node_t *) node->data).arg->content.handler);
	}
  
  return FALSE;
}



/* variabili globali utilizzate dalla funzione append_group_action() */
extern int in_macro;
extern label_table_t macro_label_table;
extern label_table_t label_table;

GList *
append_group_action(GList *a_single_action, int num_statement, simple_event_t *se, follow_array_t afs)
     /* controlla che le variabili e i label contenuti in a_single_action siano correttamente
	definiti e in caso affermativo appende a_single_action alle azioni del simple event se */
{
  GList *an_action_clone;
  GList *hel, *right_handler_list_clone, *left_handler_list_clone;
  label_value_t *fe;
  int ne;
  short error = 0;
  int num;

  GNode *tree_cond;
  arg_node_t *argnode;
  
  if (in_macro)
    num = -num_macro;
  else
    num = num_seq;

  hel = left_handler_list_clone = right_handler_list_clone = NULL;
  ne = se->num_event;
  
  an_action_clone = clone_condition_or_action(a_single_action);
  tree_cond = (GNode *) an_action_clone->data;
  argnode = (arg_node_t *)tree_cond->data;
  
  g_node_traverse((GNode *) an_action_clone->data, G_POST_ORDER, G_TRAVERSE_LEAFS, -1, find_right_handler, &right_handler_list_clone);
  
  g_node_traverse((GNode *) an_action_clone->data, G_POST_ORDER, G_TRAVERSE_LEAFS, -1, find_left_handler, &left_handler_list_clone);
  
  /* controllo che le variabili e i segnaposti che eventualmente compaiono a destra di 
     assegnamenti nell'istruzione in esame risultino definite all'evento cui vogliamo
     appendere l'istruzione */
  for ( hel = right_handler_list_clone; hel; hel = g_list_next(hel) )
    {
      if ( ( * (handler_t *) hel->data).var_name != NULL )
	{
	  /* variabile */ 
	  scope_el_t *scope, *first_scope;
	  int num;
	  
	  if (in_macro)
	    num = -num_macro;
	  else
	    num = num_seq;
	  
	  first_scope = firstscope(((handler_t *) hel->data)->var_name, num);

	  if ( first_scope )
	    {
	      /* la variabile e' definita localmente */
	      if ( abs(first_scope->num_event) >= abs(afs.follow_array[1].evid.num_event) )
		{
		  /* la variabile e' definita per la prima volta all'interno della combinazione di eventi */
		  error = 1;
		  scope = lastscope(((handler_t *) hel->data)->var_name, num);
		  for ( ; scope; scope = prevscope() )
		    if ( (scope->num_event == ne && scope->context == se->context) || is_descendant(afs, scope->num_event, scope->context, ne, se->context) )
		      {
			/* casi in cui la variabile risulta definita:
			   1) la variabile e' definita nello stesso evento: allora deve essere che 
			      la variabile e' definita nella stesso gruppo di istruzioni a cui vogliamo
			      appendere l'istruzione in esame
			   2) la variabile e' definita in un evento da cui discende quello in esame
			*/
			error = 0;
			break;
		      }
		}
	      /* else
		 -- la variabile e' definita prima della combinazione di eventi in esame:
		 -- assumo che la variabile sia definita anche all'interno della combinazione */
	    }
	  else
	    {
	      assert( firstscope(((handler_t *) hel->data)->var_name, 0) != NULL );
	      /* se la variabile non e' definita ne' localmente ne' globalmente vi  un errore nel programma */
	    }
	  
	  if ( error )
	    {
	      char *context = se->context == intro ? "intro" : "exit";
	      char *syscall_name = get_sysc_name(se->n_sysc);

	      yyerror("ERROR: variable %s undefined in %s statements of %s", ( * (handler_t *) hel->data).var_name, context, syscall_name );
	      break;
	    }
	  
	}
      else
	{
	  /* segnaposto */
	  if (in_macro)
	    {
	      assert( ne < 0 );
	      fe = get_label(macro_label_table, ((handler_t *) hel->data)->label_ref.label_name, ne);
	    }
	  else
	    {
	      assert( ne > 0 );
	      fe = get_label(label_table, ((handler_t *) hel->data)->label_ref.label_name, ne);
	    }
	  
	  if (fe)
	    {
	      ((handler_t *) hel->data)->label_ref.n_arg = fe->n_arg;
	      ((handler_t *) hel->data)->type_label = fe->type_label;
	    }
	  else
	    {
	      char *context = se->context == intro ? "intro" : "exit";
	      char *syscall_name = get_sysc_name(se->n_sysc);
	      
	      yyerror("ERROR: label %s undefined in %s statements of %s", ( * (handler_t *) hel->data).label_ref.label_name, context, syscall_name );
	      error = 1;
	      break;
	    }
	}
      
    }  /* end for */
  
  if (!error)
    for ( hel = left_handler_list_clone; hel; hel = g_list_next(hel) )
      if  ( firstscope(((handler_t *) hel->data)->var_name, 0) == NULL )
	{
	  /* la variabile non e' definita globalemente */
	  scope_el_t *scope;	  
	  
	  scope = putscope(((handler_t *) hel->data)->var_name, num);
	  
	  scope->num_event = ne;
	  scope->context = se->context;
	  scope->num_statement = g_list_length(se->action) + 1;
	  
	  sort_current_scope();
	  
	}
  
  g_list_free(right_handler_list_clone);
  g_list_free(left_handler_list_clone);

  if (error)
    {
      free_condition_or_action(an_action_clone);
      return NULL;
    }
  
  /* appende la action clone alla simple_event */
  se->action = g_list_concat(se->action, an_action_clone);

  argnode->funct =  (void *(*)()) assign(g_node_first_child(tree_cond)->data, g_node_last_child(tree_cond)->data, num_macro);
  return an_action_clone;
}






/************************ check_event_for_not ********************************************/

/* NOTA: non tutti i possibili tipi di sequenze eventi si possono negare : 
   si possono negare:
   -- un singolo evento;
   -- una combinazione di eventi della forma { e1 || e2 || ... || eN }
*/

int
check_event_for_not (GNode *node, gpointer data)
     /* controlla che il nodo correntemente esaminato faccia parte di un albero costituito da un OR di eventi */
{
  event_node_t *event;
  int sysc_num;
  /* minus_sysc e' var. globale */

  event = node->data;

  if (G_NODE_IS_LEAF(node))
    {
      assert(event->op == none);
      /* non e' permesso inserire ANY nell'evento negato */
      if (event->simple_event->n_sysc == ANY)
	{
	  yyerror("ERROR: syntax error ( !@ not permitted)");
	  exit(-1);				     
	}
      if (event->simple_event->action != NULL)
	{
	  yyerror("ERROR: syntax error ( actions in ! event not permitted)");
	  exit(-1);				     
	}
      
      /* costruisco la lista che contiene i numeri di syscall da togliere dall'ANY */
      sysc_num = event->simple_event->n_sysc;

      if (!g_list_find(minus_sysc, GINT_TO_POINTER(sysc_num)))
	{
	  minus_sysc = g_list_append(minus_sysc, GINT_TO_POINTER(sysc_num));
	}
      

    }
  else /* !G_NODE_IS_LEAF(node) */
    {
      switch (event->op) 
	{
	case  sequencing:
	  if (!G_NODE_IS_LEAF(g_node_first_child(node)))
	    {
	      yyerror("ERROR:  ! { ..; .. ; ..} not implemented");
	      exit(-1);				     
	    }
	  break;
	case  repetition:
	  yyerror("ERROR:  ! { .. * ..} not permitted");
	  exit(-1);
	break;
	default:
	  ;
	}
    }
  
  return 0;
}




/***************************** make_any_for_not *******************************************/


GNode *
make_any_for_not(GList *l)
     /* crea una coppia {intro; exiting} per l'evento speciale any 
	e restituisce il puntatore alla radice dell'albero  */
{
  event_node_t *father, *left_child, *right_child;
  GNode *tree;

  father = (event_node_t *) malloc(sizeof(event_node_t));
  father->op = sequencing;
  
  left_child = (event_node_t *) malloc(sizeof(event_node_t));
  left_child->op = none;
  left_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
  (*left_child->simple_event).n_sysc = ANY;
  (*left_child->simple_event).num_event = 0;
  (*left_child->simple_event).context = intro;
  (*left_child->simple_event).fl.minus_sysc = l;
  (*left_child->simple_event).empty_follow = 0;
  (*left_child->simple_event).condition = NULL;
  (*left_child->simple_event).action = NULL;
  
  right_child = (event_node_t *) malloc(sizeof(event_node_t));
  right_child->op = none;
  right_child->simple_event = (simple_event_t *) malloc(sizeof(simple_event_t));
  (*right_child->simple_event).n_sysc = ANY;
  (*right_child->simple_event).num_event = 0;
  (*right_child->simple_event).context = exiting;
  (*right_child->simple_event).fl.minus_sysc = l; /* basta che assegni il puntatore alla lista all'ANY intro */
  (*right_child->simple_event).empty_follow = 0;
  (*right_child->simple_event).condition = NULL;
  (*right_child->simple_event).action = NULL; 
  
  tree = g_node_new(father);
  g_node_append_data(tree, left_child);
  g_node_append_data(tree, right_child);     
  
  return tree;
}




/*************************** get_negated_event *******************************************/


/* la variabile globale utilizzata da cut_and_buid_event */
extern GList *minus_sysc;

static int
cut_and_build_event(GNode *node, gpointer data)
     /* se il nodo correntemente esaminato contiene l'operatore di sequencing 
        e almeno uno dei figli contiene un simple_event con la lista delle condizioni 
	non vuota, duplica il ramo dell'albero, nega la lista delle condizioni e 
	imposta opportunamente il flag empty_follow dei simple event */
{
  GNode *a_clone, *root;
  event_node_t *event, *intro, *exit , *father;

  event = node->data;
  
  switch (event->op) 
    {
    case  sequencing:
      intro = g_node_first_child(node)->data; assert(intro->op == none);
      exit = g_node_last_child(node)->data;   assert(exit->op == none);
      
      if ((intro->simple_event->condition != NULL) || (exit->simple_event->condition != NULL))
	{
	  intro->simple_event->fl.negated_condition = 1;
	  exit->simple_event->fl.negated_condition = 1;
	  
	  g_node_unlink(node); /* NOTA BENE: se node e' la radice dell'albero non succede niente */         
	  assert(node != NULL);
	  
	  a_clone = clone_tree_with_event_node(node);
	  
	  intro = g_node_first_child(a_clone)->data; assert(intro->op == none);
	  exit = g_node_last_child(a_clone)->data;   assert(exit->op == none);
	  
	  intro->simple_event->fl.negated_condition = 0;
	  intro->simple_event->empty_follow = 0;
	  exit->simple_event->fl.negated_condition = 0;
	  exit->simple_event->empty_follow = 1;
	  
	  father = (event_node_t *) malloc(sizeof(event_node_t)); 
	  father->op = alternation;
	  
	  root = g_node_new(father);
	  
	  g_node_append(root, a_clone);
	  g_node_append(root, node);
	}
      else
	{
	  g_node_unlink(node);        
	  assert(node != NULL);
	  intro->simple_event->empty_follow = 0;
	  exit->simple_event->empty_follow = 1;
	}
      
      *((GNode **)data) = node;  
      return 1; /* interrompo la visita */
      
      break;
    default:
      ;
    }
  
  return 0;
}


GNode *
get_negated_event(GNode *tree)
     /* isola il primo nodo contenente una coppia di simple event {intro; exiting} dell'abero passatogli 
	e lo restituisce negandone le condizioni se esistono; per ottenere i nodi successivi riutilizzare
	la funzione get_negated_event() con argomento uguale a NULL */
{
  GNode *branch = NULL;
  static GNode *local_tree;
  static int end;

  if (tree)
    {
      end = 0;
      local_tree = tree;
    }
  else
    if ( end )
      {
	end = 0;
	return NULL;
      }
  
  g_node_traverse (local_tree, G_IN_ORDER, G_TRAVERSE_NON_LEAFS, -1, cut_and_build_event, &branch);  
  
  if ( branch == local_tree ) /* e' il caso in cui tree e' sysc_intro ; sysc_exit */
    end = 1;
  else 
    end = 0; 

  if (branch == NULL)
    free_parse_tree(local_tree);

  return branch == NULL ? NULL : g_node_get_root(branch);
}




/************************ rename_var_handler *********************************************/


/* variabili globali utilizzate da rename_var_handler() */
extern int in_macro;
extern int num_macro_in_event;
extern int num_macro_in_macro;

gboolean
rename_var_handler(GNode *node, gpointer data)
     /* rinomina la variabile eventualmente contenuta nel nodo correntemente esaminato */
{
  
  if ( (* (arg_node_t *) node->data).arg != NULL )
    if ( (* (arg_node_t *) node->data).arg->type_content == HANDLER )  
      if ( (* (arg_node_t *) node->data).arg->content.handler->var_name != NULL )
	{
	  char **name = &((* (arg_node_t *) node->data).arg->content.handler->var_name);
	  
	  if ( getvar(var_table, *name) == NULL )
	    {
	      /* la variabile non  globale */
	      
	      char num[5];
	      char *name_var;
	      
	      snprintf(num, 5, "%+d", !in_macro ? num_macro_in_event : num_macro_in_macro);
	      name_var = (char *) malloc( strlen(num) + strlen(*name)+1);
	      memcpy((char *) name_var, num, strlen(num));
	      memcpy((char *) name_var+strlen(num), *name, strlen(*name)+1);
	      
	      free(*name);
	      
	      *name = name_var;
	      printf("\nDEBUG: rename_var_handler %s", (* (arg_node_t *) node->data).arg->content.handler->var_name);
	    }
	}
  
  return FALSE;
}



/*********************************************
 * Gestione degli errori durante il parsing  *
 *********************************************/


/* variabili globali utilizzate da yyerror() */
extern char *parsed_file; 
extern int num_lines;
extern int num_error;
extern int in_declarations;
extern int num_seq;

int 
yyerror(char *s, ...)
{
  va_list args;

  va_start(args, s);

  if (num_error == 0)
    fprintf(stderr, "In specification file '%s':\n", parsed_file);
  
  num_error++;

  fprintf(stderr, "  %s: line %d: ", parsed_file, num_lines);
  vfprintf(stderr, s, args);
  if (!in_declarations)  fprintf (stderr, " in event sequence n. %d\n", num_seq);
  else fprintf(stderr, " in declarations\n");
  
  va_end(args);

  return 0;
}

/* end */
