/*
 * Copyright (c) 2001-2003 Davide Fais, Stefano Frassi, Vincenzo Di Stefano, Filippo Giuntini, Luca Pugliese <frassi@iit.cnr.it>
 * for adding wIDSard code
 * Copyright (c) 1991, 1992 Paul Kranenburg <pk@cs.few.eur.nl>
 * Copyright (c) 1993 Branko Lankester <branko@hacktic.nl>
 * Copyright (c) 1993, 1994, 1995, 1996 Rick Sladkey <jrs@world.std.com>
 * Copyright (c) 1996-1999 Wichert Akkerman <wichert@cistron.nl>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *	$Id: widsard.c,v 1.3 2003/02/17 15:57:40 frassi Exp $
 */

#include <sys/types.h>

#include <signal.h>
#include <errno.h>
#include <sys/param.h>
#include <fcntl.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <pwd.h>
#include <grp.h>
#include <string.h>
//#include <stdlib.h>
#include <limits.h>

//#include "threads.h"
//#include "queue.h"
#include "defs.h"


/*****************************************************************
 *** dichiarazioni per il parsing e la costruzione dell'automa ***
 *****************************************************************/

#include "internal_syscall.h"
#include "util.h"

#include <assert.h>
#include "parser.h"

extern char *strsignal (int sig);

extern int yyparse ();

/* dichiarata in wids.lex */
extern int num_error;
extern FILE *yyin;
extern char *parsed_file;

syntax_tree events_tree = NULL;

/* TABELLE DEI SIMBOLI */
extern var_table_t var_table;
extern macro_var_table_t macro_var_table;

extern label_table_t label_table;
extern label_table_t macro_label_table;

extern macro_entry *macro_table;

extern funct_table_t funct_table;
extern module_table_t module_table;

/* strutture automa e follow_array_t definite in automa.h */
automa atm;
follow_array_t arf;
automa_state process_state;	/* era state */
process_state_table_t process_state_table;
begin_of_subseqs_ll subseqs_ll = NULL;

/* stato corrente */
struct tcb *current;

/**********************************************************************
 *** fine dichiarazioni per il parsing e la costruzione dell'automa ***
 **********************************************************************/

/**************************************************************
 ** dichiarazioni per la creazione e la gestione dei threads **
 **************************************************************/
/* pthread_t pro, con, clienter;	//widsard, produttore e consumatore */
/* int base = 10;			//per strtol */
/* int debug_level = 0;		//inizializzata con un flag, puo' essere modificata dal c_thread */
/* final_state_table_t final_event = NULL; */
/* int num_log = 0, *canale_log = NULL; */
/* msg_queue *mqueue = NULL;	//condivisa da trace_syscall e dal producer */
/* str_queue *squeue = NULL;	//condivisa dal producer e dal consumer */
/* double_queue *dq; */
/* int root = 1; */
/* int string_pipe[2];		//serve al getPushMessage per ottenere la stringa con la syscall */
/* FILE *pipeOut, *pipeIn; */
/* FILE *logfile; */
/*******************************************************************
 ** fine dichiarazioni per la creazione e la gestione dei threads **
 *******************************************************************/

int pflag_seen = 0;

uid_t run_uid;
gid_t run_gid;

int nprocs;
char *progname;
extern char version[];
extern char **environ;

/*****************************
 * OTHER GLOBAL DECLARATIONS *
 *****************************/

struct
{
  int outffp;			/* tell to create a output file for each traced process */
  char *progname;		/* program name */
}
line_command;

int max_strlen = DEFAULT_STRLEN;	/* max number of string chars to print */
int nprocs;			/* number of traced processes */
FILE *outf;

/* struct tcb tcbtab[MAX_PROCS];	 processes state table */
GHashTable *tcbtab;
int in_cleanup = 0;             /* quando entra nella funzione cleanup viene settato ad 1 */


extern char version[];
extern char **environ;

struct tcb *pid2tcb (int pid);
//static int trace (msg_queue * mq);
static int trace ();
//Chiede al gestore del server la password, per verificarla in seguito
//static char *ask_password ();
//static int verify_my_password (char *pwd, int *l, int n);
static void cleanup (void);
static void interrupt (int sig);
static sigset_t empty_set, blocked_set;


#ifdef HAVE_SIG_ATOMIC_T
static volatile sig_atomic_t interrupted;
#else /* !HAVE_SIG_ATOMIC_T */
#ifdef __STDC__
static volatile int interrupted;
#else /* !__STDC__ */
static int interrupted;
#endif /* !__STDC__ */
#endif /* !HAVE_SIG_ATOMIC_T */


extern int set_personality (int personality);
//extern int trace_syscall (struct tcb *, msg_queue *);
extern int trace_syscall (struct tcb *);

/* static void */                   /* orig usage... */
/* usage (ofp, exitval) */
/*      FILE *ofp; */
/*      int exitval; */
/* { */
/*   fprintf (ofp, "\ */
/* usage: widsard [-hV] ... \n\ */
/*               -f rules_filename [-p pid] ... [-s strsize] [command [arg ...]]\n\ */
/* -h -- print help message\n\ */
/* -V -- print version\n\ */
/* -f rules_filename -- the file that contain the rules\n\ */
/* -p pid -- trace process with process id PID, may be repeated\n\ */
/* -s strsize -- limit length of print strings to STRSIZE chars (default %d)\n\ */
/* -o file -- send trace output to FILE \n\ */
/* -O -- send trace output to a file named \"command-PID\" foreach process created\n\ */
/* ", DEFAULT_STRLEN); */
/*   exit (exitval); */
/* } */

static void
usage (ofp, exitval)
     FILE *ofp;
     int exitval;
{
  fprintf (ofp, "\
usage: widsard [-hV] -f rules_filename [-s strsize] command [arg ...]\n\
-h -- print help message\n\
-V -- print version\n\
-f rules_filename -- the file that contain the rules\n\
-s strsize -- limit length of print strings to STRSIZE chars (default %d)\n\
", DEFAULT_STRLEN);
  exit (exitval);
}

/**************** PARTI TOLTE DA usage() *******************************************
-t trace_level -- if 0 no messages will be sent to the client,\n\
                  if 1 only final messages will be sent,\n\
                  if 2 all messages will be sent\n\
*/

int
main (argc, argv)
     int argc;
     char *argv[];
{
  extern int optind;		/* vedi getopt */
  extern char *optarg;		/* vedi getopt */
  struct tcb *tcp;
  int c, pid = 0;
  struct sigaction sa;
  static char buf[BUFSIZ];

  /* queste erano globali */
  int pflag_seen = 0;
  uid_t run_uid=0;
  gid_t run_gid=0;
  char *progname=NULL;
  char *outfname = NULL;	/* output file name */
  //char *mypasswd;
  int f_flag_found = 0;

  //Si cerca di aprire in append il file di log ma,
  //se esso non esiste, si cerca di crearlo.
  //Se anche questo dovesse fallire, ci si accontenta
  //di loggare sullo stderr
/*   num_log = 1; */
/*   canale_log = calloc (num_log, sizeof (int)); */
/*   if (canale_log != NULL) */
/*     { */
/*       if ((canale_log[0] = open ("consumer_log", O_APPEND)) == -1) */
/* 	if ((canale_log[0] = open ("consumer_log", O_CREAT | O_WRONLY)) == -1) */
/* 	  canale_log[0] = 2; */

/*       if (canale_log[0] != 2) */
/* 	fchmod (canale_log[0], S_IRUSR | S_IWUSR | S_IXUSR); */
/*     } */

/*    if ((mypasswd = ask_password ()) == NULL) */
/*      { */
/*        log_event */
/*  	("Errore leggendo la password del superuser!!!\nImpossibile proseguire.\n", */
/*  	 canale_log, num_log); */
/*        exit (1); */
/*      } */

/*    log_event ("Verifica password superuser in corso...\n", canale_log, */
/*  	     num_log); */
/*    if (verify_my_password (mypasswd, canale_log, num_log) == -1) */
/*      { */
/*        log_event ("Password superuser errata! Impossibile proseguire.\n", */
/*  		 canale_log, num_log); */
/*        exit (1); */
/*      } */
/*    free (mypasswd); */
/*    mypasswd = NULL; */
/*    log_event ("Verifica superata con successo!\n", canale_log, num_log); */

  progname = argv[0];
  outf = stderr;
  set_personality (DEFAULT_PERSONALITY); /* si trova in syscall.c e crea diverse strutture dati */

  /* Allocate the initial tcbtab.  */
  /* creo la tabella hash per contenere le struct tcb */
  tcbtab = g_hash_table_new(g_int_hash, g_int_equal);


  while ((c = getopt (argc, argv, "+hVOf:p:s:o:t:")) != EOF)
    {
      switch (c)
	{
	case 'o':
	  outfname = strdup (optarg);
	  break;

	case 'O':
	  line_command.outffp = 1;
	  break;

	case 'h':
	  usage (stdout, 0);
	  break;

	case 'V':
	  printf ("%s\n", version);
	  exit (0);
	  break;

/* 	case 't': */
/* 	  { */
/* 	    char *end_p = NULL; */
/* 	    debug_level = (int) strtol (optarg, &end_p, base); */
/* 	    if (end_p == optarg) */
/* 	      usage (stderr, 1); */
/* 	  } */
/* 	  break; */

	case 'f':
	  /***** creo l'albero degli eventi e successivamente il relativo automa *****/

	  f_flag_found = 1;
	  parsed_file = clone_string (optarg);
	  if ((yyin = fopen (optarg, "r")) == NULL)
	    {
	      perror ("");
	      exit (-1);
	    }
	  /***************************************************/
	  /* Inizializzazione strutture dati per il parsing  */
	  /***************************************************/
	  var_table = NEW_VAR_TABLE;
	  macro_var_table = NEW_MACRO_VAR_TABLE;

	  label_table = NEW_LABEL_TABLE;
	  macro_label_table = NEW_LABEL_TABLE;

	  macro_table = (macro_entry *) NULL;

	  init_funct_table (&funct_table);

	  if (g_module_supported ())
	    module_table = NEW_MODULE_TABLE;
	  else
	    {
	      module_table = NULL;
	      printf ("Runtime module unsupported");
	    }

	  init_var_scope_table ();

	  /**********************************/
	  /* Parsing file di configurazione */
	  /**********************************/
	  if (yyparse ())	/* ritorna 0 se nel file di configurazione non ha trovato errori */
	    {
	      printf("\nCorrect rules file before run widsard\n");
	      exit(-1);
	    }
	  printf ("\n*******************************");
	  printf ("\n%d error detected in rules file\n", num_error);
	  printf ("*******************************\n");
	  
	  assert (events_tree != NULL);

	  /**********************/
	  /* Costruzione automa */
	  /**********************/
	  atm = AQE_gen (events_tree);
	  print_automa (atm);

	  /* calcola e stampa la lista di liste delle sottosequenze */
	  //subseqs_ll = compute_subranges_rec (events_tree, subseqs_ll);
	  //print_subseqs_ll (subseqs_ll);
	  
	  break;

	case 'p':
	  if ((pid = atoi (optarg)) == 0)
	    {
	      fprintf (stderr, "%s: Invalid process id: %s\n",
		       progname, optarg);
	      break;
	    }
	  if (pid == getpid ())
	    {
	      fprintf (stderr,
		       "%s: I'm sorry, I can't let you do that, Dave.\n",
		       progname);
	      break;
	    }
	  if ((tcp = alloctcb (pid)) == NULL)
	    {
	      fprintf (stderr,
		       "%s: tcb table full, please recompile widsard\n",
		       progname);
	      exit (1);
	    }
	  tcp->flags |= TCB_ATTACHED;
	  pflag_seen++;
	  break;

	case 's':
	  max_strlen = atoi (optarg);
	  break;
	default:
	  usage (stderr, 1);
	  break;
	}
    }

  if (!f_flag_found)
    {
      fprintf (stderr, "ERROR: no rules file specified\n");
      usage (stderr, 1);
    }

  run_uid = getuid ();
  run_gid = getgid ();
  setreuid (geteuid (), getuid ());
  
  if (outf == stderr)
    setvbuf (outf, buf, _IOLBF, BUFSIZ);

  /********************************************************************************************************
   ** in questo for widsard si attacca a processi gia' esistenti passati come parametro : es. widsard -p 610    CONTROLLARE....
  for (c = 0, tcp = tcbtab; c < MAX_PROCS; c++, tcp++)
    {
       Reinitialize the output since it may have changed. 
      tcp->outf = outf;		 se il proc. non e' nostro figlio 
      if (!(tcp->flags & TCB_INUSE) || !(tcp->flags & TCB_ATTACHED))
	continue;
      if (ptrace (PTRACE_ATTACH, tcp->pid, (char *) 1, 0) < 0)
	{			 mi attacco al proc. che gli ho passato 
	  perror ("attach: ptrace(PTRACE_ATTACH, ...)");	 con il -p 
	  droptcb (tcp);
	  continue;
	}
      fprintf (stderr, "Process %u attached - interrupt to quit\n", tcp->pid);
    }
  **********************************************************************************************************/

  /**************************************************************************************/
  /* verra' eseguito l'if se c'e' un comando passato come parametro : es. widsard httpd */
  /**************************************************************************************/
  if (optind < argc)
    {

      struct stat statbuf;
      char *filename, *tmp;
      char pathname[MAXPATHLEN];

      /* se argv contiene il path lo copio direttamente in pathname */
      filename = argv[optind];

      /* set program name to trace */
      tmp = strrchr (filename, '/');
      tmp = tmp ? tmp++ : filename;
      line_command.progname = strdup (tmp);

      /* look for complete program name to exec */
      if (strchr (filename, '/'))
	strcpy (pathname, filename);
#ifdef USE_DEBUGGING_EXEC
      /*
       * Debuggers customarily check the current directory
       * first regardless of the path but doing that gives
       * security geeks a panic attack.
       */
      /* se non gli passo il path lo cerco nella dir corrente */
      else if (stat (filename, &statbuf) == 0)
	strcpy (pathname, filename);
#endif /* USE_DEBUGGING_EXEC */
      else
	{
	  char *path;
	  int m, n, len;
	  /* se nella dir corrente non c'e' analizzo la var PATH di sistema */
	  for (path = getenv ("PATH"); path && *path; path += m)
	    {
	      if (strchr (path, ':'))
		{
		  n = strchr (path, ':') - path;
		  m = n + 1;
		}
	      else
		m = n = strlen (path);
	      if (n == 0)
		{
		  getcwd (pathname, MAXPATHLEN);
		  len = strlen (pathname);
		}
	      else
		{
		  strncpy (pathname, path, n);
		  len = n;
		}		/* estrapolato il path che mi interessa lo copio */
	      if (len && pathname[len - 1] != '/')
		pathname[len++] = '/';
	      /* in fondo ci attacco argv iniziale privo del path */
	      strcpy (pathname + len, filename);
	      if (stat (pathname, &statbuf) == 0)
		break;
	    }
	}
      if (stat (pathname, &statbuf) < 0)
	{
	  fprintf (stderr, "%s: %s: command not found\n", progname, filename);
	  exit (1);
	}

      switch (pid = fork ())
	{
	case -1:
	  perror ("widsard: fork");
	  cleanup ();
	  exit (1);
	  break;
	case 0:
	  { /* caso processo figlio */

	    if (ptrace (PTRACE_TRACEME, 0, (char *) 1, 0) < 0)
	      {
		perror ("widsard: ptrace(PTRACE_TRACEME, ...)");
		return -1;
	      }

	    if (geteuid () == 0)
	      {
		uid_t run_euid = run_uid;
		gid_t run_egid = run_gid;

		if (statbuf.st_mode & S_ISUID)
		  run_euid = statbuf.st_uid;
		if (statbuf.st_mode & S_ISGID)
		  run_egid = statbuf.st_gid;

		/*
		 * It is important to set groups before we
		 * lose privileges on setuid.
		 */
		if (setregid (run_gid, run_egid) < 0)
		  {
		    perror ("setregid");
		    exit (1);
		  }
		if (setreuid (run_uid, run_euid) < 0)
		  {
		    perror ("setreuid");
		    exit (1);
		  }
	      }
	    else
	      setreuid (run_uid, run_uid);

	    execv (pathname, &argv[optind]);
	    perror ("widsard: exec");
	    _exit (1);
	    break;
	  }
	default: /* caso processo tracciante */
	  
	  printf("process created %d\n", pid);
	  if ((tcp = alloctcb (pid)) == NULL)
	    {
	      fprintf (stderr, "tcb table full\n");
	      cleanup ();
	      exit (1);
	    }
	  
	  /* Inizializzo lo stato del processo creato */
	  tcp->process_state = init_automa_state (tcp->process_state, var_table);
	  break;
	}
    }
  else if (pflag_seen == 0)	/* errore : non sono stati passati a widsard processi da tracciare */
    usage (stderr, 1);
  
  /* inizializzo i due insiemi di segnali come vuoti */
  sigemptyset (&empty_set);
  sigemptyset (&blocked_set);

  sa.sa_handler = SIG_IGN;
  /* inizializzo la maschera */
  sigemptyset (&sa.sa_mask);
  /* specifico il flags */
  sa.sa_flags = 0;
  /* associo a questi segnali l'azione relativa all'handler di &sa (li ignoro) */
  sigaction (SIGTTOU, &sa, NULL);
  sigaction (SIGTTIN, &sa, NULL);
  /* inserisco questi segnali nel gruppo blocked_set */
  sigaddset (&blocked_set, SIGHUP);
  sigaddset (&blocked_set, SIGINT);
  sigaddset (&blocked_set, SIGQUIT);
  sigaddset (&blocked_set, SIGPIPE);
  sigaddset (&blocked_set, SIGTERM);
  /* ai segnali delle sigation associo l'azione interrupt (vedi funzione) */
  sa.sa_handler = interrupt;

  sigaction (SIGHUP, &sa, NULL);
  sigaction (SIGINT, &sa, NULL);
  sigaction (SIGQUIT, &sa, NULL);
  sigaction (SIGPIPE, &sa, NULL);
  sigaction (SIGTERM, &sa, NULL);
  
 /* Make sure SIGCHLD has the default action so that waitpid
   definitely works without losing track of children.  The user
   should not have given us a bogus state to inherit, but he might
   have.  Arguably we should detect SIG_IGN here and pass it on
   to children, but probably noone really needs that.  */
  sa.sa_handler = SIG_DFL;       // CONTROLLARE: AGGIUNTO
  sigaction(SIGCHLD, &sa, NULL); // CONTROLLARE: AGGIUNTO


  //viene inizializzata una hashtable per capire se un determinato numero
  //di evento si trova in un sottointervallo finale
  /*
  final_event = initFinalEventsTable (subseqs_ll);

  if (final_event == NULL)
    {
      fprintf (stderr, "main: impossibile inizializzare le SD\n");
      cleanup ();
      exit (1);
    }

  //viene creata una coda condivisa
  mqueue = (msg_queue *) init_msg_queue ();

  if (mqueue == NULL)
    {
      fprintf (stderr, "main: creazione coda non riuscita.\n");
      cleanup ();
      exit (1);
    }

  //viene creata un'altra coda condivisa
  squeue = (str_queue *) init_str_queue ();

  if (squeue == NULL)
    {
      fprintf (stderr, "main: creazione coda non riuscita.\n");
      cleanup ();
      exit (1);
    }

  dq = (double_queue *) malloc (sizeof (double_queue));

  if (dq == NULL)
    {
      fprintf (stderr, "main: creazione doppia coda non riuscita.\n");
      cleanup ();
      exit (1);
    }

  dq->msg = mqueue;
  dq->str = squeue;

  //Si cerca di aprire in append il file di log ma,
  //se esso non esiste, si cerca di crearlo.
  //Se anche questo dovesse fallire, ci si ricorda di cio'.
  if ((logfile = fopen (FILENAME, "a")) == NULL)
    if ((logfile = fopen (FILENAME, "w")) == NULL)
      log_event ("main: impossibile aprire il file di log delle syscall\n",
		 canale_log, num_log);

  if (logfile != NULL)
    {
      chmod (FILENAME, S_IRUSR | S_IWUSR | S_IXUSR);
      setlinebuf (logfile);
    }

  pipe (string_pipe);
  pipeIn = fdopen (string_pipe[0], "r");
  pipeOut = fdopen (string_pipe[1], "w");
  setlinebuf (pipeOut);

  //vengono creati i threads
  //pthread_create (&pro, NULL, producer, dq);

  //pthread_create (&con, NULL, consumer, squeue);
  */

  printf("\nApplication started\n");

  /* richiamo trace(), la funzione principale */
  /*
  if (trace (mqueue) < 0)
    {
      pthread_cond_signal (mqueue->notFull);
      pthread_cond_signal (mqueue->notEmpty);
      pthread_cond_signal (squeue->notFull);
      pthread_cond_signal (squeue->notEmpty);
      pthread_cancel (pro);
      pthread_cancel (con);
      pthread_cancel (clienter);
      delete_msg_queue (mqueue);
      delete_str_queue (squeue);
      mqueue = NULL;
      squeue = NULL;
      if (logfile != NULL)
	fclose (logfile);
      fclose (pipeIn);
      fclose (pipeOut);
      close (string_pipe[0]);
      close (string_pipe[1]);
      free (dq);
      _exit (1);
    }
  */
  if (trace() < 0)
    {
      fprintf(stderr,"main: trace() < 0\n");
      exit(1);
    }
  /*
  pthread_cond_signal (mqueue->notFull);
  pthread_cond_signal (mqueue->notEmpty);
  pthread_cond_signal (squeue->notFull);
  pthread_cond_signal (squeue->notEmpty);
  pthread_cancel (pro);
  pthread_cancel (con);
  pthread_cancel (clienter);
  delete_msg_queue (mqueue);
  delete_str_queue (squeue);
  mqueue = NULL;
  squeue = NULL;
  if (logfile != NULL)
    fclose (logfile);
  fclose (pipeIn);
  fclose (pipeOut);
  close (string_pipe[0]);
  close (string_pipe[1]);
  free (dq);
  */
  cleanup();
  exit (0);
}

struct tcb *
alloctcb (pid) /* si alloca un nuovo tcb, inizializzando i vari campi di tale struttura */
int pid;
{
  //int i;
  int j;
  struct tcb *tcp = malloc(sizeof(struct tcb));
  long *pid_ptr;

  tcp->pid = pid;
  tcp->process_state = NULL;
  tcp->parent = NULL;
  tcp->nchildren = 0;
  tcp->flags = /* TCB_INUSE | */ TCB_STARTUP; // TCB_INUSE non mi serve piu' con la tabella hash

  /* set output file */
  if (line_command.outffp)
    {
      char filename[1024];
      sprintf (filename, "%s-%d", line_command.progname, pid);
      tcp->outf = fopen (filename, "w");
    }
  else
    tcp->outf = outf;	/* Initialize to current out file */
  
  nprocs++;
  
  for (j = 0; j < MAX_ARGS; j++)
    tcp->arg_state[j] = EMPTY;
  
  tcp->arg_state[RVAL] = EMPTY;
  
  pid_ptr = (long *) malloc(sizeof(long));
  *pid_ptr = pid;
  
  g_hash_table_insert(tcbtab, pid_ptr, tcp);
  //g_hash_table_insert(tcbtab, clone_long(&pid), tcp);
  
  return tcp;
  
  /* 
  for (i = 0, tcp = tcbtab; i < MAX_PROCS; i++, tcp++)
    {
      if ((tcp->flags & TCB_INUSE) == 0)
	{
	  tcp->pid = pid;
	  tcp->process_state = NULL;
	  tcp->parent = NULL;
	  tcp->nchildren = 0;
	  tcp->flags = TCB_INUSE | TCB_STARTUP;

	
	  if (line_command.outffp)
	    {
	      char filename[1024];
	      sprintf (filename, "%s-%d", line_command.progname, pid);
	      tcp->outf = fopen (filename, "w");
	    }
	  else
	    tcp->outf = outf;	

	  nprocs++;

	  for (j = 0; j < MAX_ARGS; j++)
	    tcp->arg_state[j] = EMPTY;

	  tcp->arg_state[rval] = EMPTY;

	  return tcp;
	}
    }
  
  
    return NULL; */
}


struct tcb *
pid2tcb (pid)
int pid;
{
  //int i;
  //struct tcb *tcp;

  return (struct tcb *) g_hash_table_lookup(tcbtab, &pid);

  /*
  for (i = 0, tcp = tcbtab; i < MAX_PROCS; i++, tcp++)
    {
      if (pid && tcp->pid != pid)
	continue;
      if (tcp->flags & TCB_INUSE)
	return tcp;
    }
  return NULL;
  */
}


void
droptcb (struct tcb *tcp)
/* 
   azzero tutti i campi di tcp 
   e comando l'eliminazione della struct_for_monitor associata al processo
*/
{
  /* long *ab = malloc(sizeof(long));  CONTROLLARE se puo' essere fatto senza malloc: SI */
  long *ab;
  
  struct tcb *pro;
  long pid = tcp->pid;
  
  /*
  if (mqueue != NULL)
    {
      msg_building_info *mbi = malloc (sizeof (msg_building_info));

      if (mbi == NULL)
	log_event
	  ("droptcb: impossibile creare un messaggio di eliminazione\n",
	   canale_log, num_log);
      else
	{
	  mbi->pid = tcp->pid;
	  mbi->erase = 1;

	  pthread_mutex_lock (mqueue->mut);

	  while (mqueue->full)
	    {
	      log_event ("droptcb: message queue PIENA.\n", canale_log,
			 num_log);
	      pthread_cond_wait (mqueue->notFull, mqueue->mut);
	    }

	  push_msg_queue (mqueue, mbi);
	  pthread_cond_signal (mqueue->notEmpty);

	  pthread_mutex_unlock (mqueue->mut);
	}
      }
  */

  if (tcp->pid == 0)
    return;
  
  nprocs--;
  //tcp->pid = 0;
  //tcp->flags = 0;
  
  free_atm_state (tcp->process_state);
  //tcp->process_state = NULL;
  
  /* se tcp <> NULL ha gia' il padre quindi ----> azzero il campo padre e
     decremento quello del figlio */
  if (tcp->parent != NULL)
    {
      tcp->parent->nchildren--;
      //tcp->parent = NULL;
    }
  
  //tcp->outf = 0;
  //tcp->ev_set = NULL;
  //tcp->send_anyway = 0;

  g_hash_table_lookup_extended (tcbtab, &pid, (void **)&ab, (void **)&pro);
  g_hash_table_remove (tcbtab, ab);
  //printf("--%x - %d\n", ab, *ab);
  free(ab); // dealloco la key pid allocata in alloctcb
  //assert(tcp == pro);
  free(tcp); // dealloco la struct tcb allocata in alloctcb

}


static int
resume (tcp)			/* risveglio un processo */
struct tcb *tcp;
{
  if (tcp == NULL)
    return -1;
  /* se il processo non era sospeso */
  if (!(tcp->flags & TCB_SUSPENDED))
    {
      fprintf (stderr, "PANIC: pid %u not suspended\n", tcp->pid);
      return -1;
    }
  tcp->flags &= ~TCB_SUSPENDED;
  /* per farlo ripartire , gli faccio la sys_call 
     sblocando la wait su cui si e' fermato il proc. */
  if (ptrace (PTRACE_SYSCALL, tcp->pid, (char *) 1, 0) < 0)
    {
      perror ("resume: ptrace(PTRACE_SYSCALL, ...)");
      return -1;
    }

  fprintf (stderr, "Process %u resumed\n", tcp->pid);
  return 0;
}



/* detach traced process; continue with sig */
int
detach (tcp, sig)  /* termina la tracciatura di un processo */
struct tcb *tcp;
int sig;
{
  int error = 0;
  int status;


  if (tcp->flags & TCB_BPTSET)
    sig = SIGKILL;

  /*
   * Linux wrongly insists the child be stopped
   * before detaching.  Arghh.  We go through hoops
   * to make a clean break of things.
   */
  if ((error = ptrace (PTRACE_DETACH, tcp->pid, (char *) 1, sig)) == 0)
    {
      /* On a clear day, you can see forever. */
    }
  else if (errno != ESRCH)
    {
      /* Shouldn't happen. */
      perror ("detach: ptrace(PTRACE_DETACH, ...)");
    }
  else if (kill (tcp->pid, 0) < 0)
    {
      if (errno != ESRCH)
	perror ("detach: checking sanity");
    }
  else if (kill (tcp->pid, SIGSTOP) < 0)
    {
      if (errno != ESRCH)
	perror ("detach: stopping child");
    }
  else
    {
      for (;;)
	{
	  if (waitpid (tcp->pid, &status, 0) < 0)
	    {
	      if (errno != ECHILD)
		perror ("detach: waiting");
	      break;
	    }			/* se il figlio non e' fermo */
	  if (!WIFSTOPPED (status))
	    {
	      /* Au revoir, mon ami. */
	      break;
	    }
	  if (WSTOPSIG (status) == SIGSTOP)
	    {
	      if ((error = ptrace (PTRACE_DETACH, tcp->pid, (char *) 1, sig)) < 0)
		{
		  if (errno != ESRCH)
		    perror ("detach: ptrace(PTRACE_DETACH, ...)");
		  /* I died trying. */
		}
	      break;
	    }
	  if ((error = ptrace (PTRACE_CONT, tcp->pid, (char *) 1,
			       WSTOPSIG (status) == SIGTRAP ?
			       0 : WSTOPSIG (status))) < 0)
	    {
	      if (errno != ESRCH)
		perror ("detach: ptrace(PTRACE_CONT, ...)");
	      break;
	    }
	} // end for
    }				

  /* se il padre e' sospeso e il pid del processo che aspetta il padre = tcp.pid (<=0) 
     e cioe' quello che e' stato appena ucciso */
  if (waiting_parent (tcp))
    error = resume (tcp->parent);
  
  fprintf(stderr, "Process %u detached\n", tcp->pid);
  
  if (!in_cleanup) droptcb (tcp);
  
  return error;
}

static void
free_tcbtab(gpointer key, gpointer value, gpointer user_data)
{
  struct tcb *tcp = (struct tcb *) value;
  
  /* if (tcp->flags & TCB_INUSE)  CONTROLLARE: controllo inutile? SI */
  // {
      
      if (tcp_last && (tcp_last == tcp))
	{
	  tprintf (" <unfinished ...>\n");
	  tcp_last = NULL;
	}			 
      if (tcp->flags & TCB_ATTACHED) {
	detach (tcp, 0);
      }
      else
	{			
	  kill (tcp->pid, SIGCONT);
	  kill (tcp->pid, SIGTERM);
	}
  // }
}

static void
cleanup ()			/* reinizializzo tutti i processi della tcb */
{
  //int i;
  //struct tcb *tcp;
  //printf("\nintro_cleanup\n");fflush(NULL);
  in_cleanup = 1;
  g_hash_table_foreach(tcbtab, free_tcbtab, NULL);
  in_cleanup = 0;
  //printf("\nexit_cleanup\n");fflush(NULL);
  /*
  for (i = 0, tcp = tcbtab; i < MAX_PROCS; i++, tcp++)
    {
      if (!(tcp->flags & TCB_INUSE))
	continue;

      if (tcp_last && (tcp_last == tcp))
	{
	  tprintf (" <unfinished ...>\n");
	  tcp_last = NULL;
	}			 se non e' nostro figlio 
      if (tcp->flags & TCB_ATTACHED)
	detach (tcp, 0);
      else
	{			 altrimenti faccio il resume e poi li uccido 
	  kill (tcp->pid, SIGCONT);
	  kill (tcp->pid, SIGTERM);
	}
    }
  */
}


static void
interrupt (sig)
int sig;
{
  interrupted = 1;
}


#define WNOHANG         0x00000001
#define WUNTRACED       0x00000002

#define __WNOTHREAD     0x20000000      /* Don't wait on children of other threads in this group */
#define __WALL          0x40000000      /* Wait on all children, regardless of type */
#define __WCLONE        0x80000000      /* Wait only on non-SIGCHLD children */



static int
//trace (msg_queue * mq)
trace ()
{
  int pid;
  int wait_errno;
  int status;
  struct tcb *tcp;

#ifdef LINUX22222222 /* ATTENZIONE non funge bene */
  static int wait4_options = __WALL ;
#endif

  while (nprocs != 0)
    {

      /* prima dell'esecuzione blocco tutti i segnali del gruppo blocked_set */
      sigprocmask (SIG_SETMASK, &empty_set, NULL);
      
#ifdef LINUX22222222 /* ATTENZIONE non funge bene (aggiornare con l'ultima versione di strace */
		pid = wait4(-1, &status, wait4_options, NULL);
		if ((wait4_options & __WALL) && errno == EINVAL) {
		  /* this kernel does not support __WALL */ /* Wait for all children (clone and not) if __WALL is set;
		                                             * otherwise, wait for clone children *only* if __WCLONE is
		                                             * set; otherwise, wait for non-clone children *only*. (Note:
		                                             * A "clone" child here is one that reports to its parent
		                                             * using a signal other than SIGCHLD.) */
		  
		  fflush(NULL);
		        wait4_options &= ~__WALL;
			errno = 0;
			pid = wait4(-1, &status, wait4_options, NULL);
		}
		if (!(wait4_options & __WALL) && errno == ECHILD) {
			/* most likely a "cloned" process */
		  
			pid = wait4(-1, &status, 0x80000000, NULL);
			if (pid == -1) {
				fprintf(stderr, "strace: clone wait4 "
						"failed: %s\n", strerror(errno));
			}
		}
#else
		pid = wait4(-1, &status, 0, NULL);
#endif /* __WALL */
      
      wait_errno = errno;
      sigprocmask (SIG_BLOCK, &blocked_set, NULL);

      if (interrupted)
	return 0;

      if (pid == -1)
	{
	  switch (wait_errno)
	    {
	    case EINTR:
	      continue;
	    case ECHILD:
	      return 0;
	    default:
	      errno = wait_errno;
	      perror ("widsard: wait");
	      return -1;
	    }
	}

      /* Look up `pid' in our table. */
      if ((current = tcp = pid2tcb (pid)) == NULL)
	{
	  fprintf (stderr, "unknown pid: %u\n", pid);
	  if (WIFSTOPPED (status))
	    ptrace (PTRACE_CONT, pid, (char *) 1, 0);
	  return -1;
	}

      /* set current output file */
      outf = tcp->outf;

      if (tcp->flags & TCB_SUSPENDED)
	{			/* ha fatto una wait4 */
	  /*
	   * Apparently, doing any ptrace() call on a stopped
	   * process, provokes the kernel to report the process
	   * status again on a subsequent wait(), even if the
	   * process has not been actually restarted.
	   * Since we have inspected the arguments of suspended
	   * processes we end up here testing for this case.
	   */
	  continue;
	}

      if (tcp->flags & TCB_STOPPED) /* AGGIUNTA PER LE AZIONI */
	{			/* ho fatto una freeze. Anche se mi arriva una SIGCONT
				   non riparte ma mi entra qui e poi ritorna alla wait */
	  printf ("the process is stopped...\n");
	  continue;
	}

      /* se il figlio e' uscito a causa di un segnale che non e' stato preso */
      if (WIFSIGNALED (status))
	{

	  /* printleader(tcp); */
	  tprintf ("+++ killed by %s +++", signame (WTERMSIG (status)));
	  /* printtrailer(tcp); */

	  droptcb (tcp);
	  continue;
	}

      /* il figlio e' uscito normalmente */
      if (WIFEXITED (status))
	{
	  if (tcp->flags & TCB_ATTACHED)
	    fprintf (stderr, "PANIC: attached pid %u exited\n", pid);
	  droptcb (tcp);
	  continue;
	}

      /* il figlio non e' fermo */
      if (!WIFSTOPPED (status))
	{
	  fprintf (stderr, "PANIC: pid %u not stopped\n", pid);
	  droptcb (tcp);
	  continue;
	}

      if (tcp->flags & TCB_STARTUP)
	{
	  /*
	   * Questo flag indica che e' stata effettuata una ptrace(PTRACE_ATTACH,...), 
	   * ma ancora non abbiamo iniziato a tracciare le syscall. La prossima volta
	   * il processo monitorato si dovrebbe effettivamente fermare all'entrata 
	   * di una syscall.
	   */

	  tcp->flags &= ~TCB_STARTUP;

	  if (tcp->flags & TCB_ATTACHED)
	    {
	      /*
	       * Interestingly, the process may stop
	       * with STOPSIG equal to some other signal
	       * than SIGSTOP if we happen to attach
	       * just before the process takes a signal.
	       */

	      if (!WIFSTOPPED (status))
		{		/* il processo monitorato non e' fermo */
		  fprintf (stderr, "pid %u not stopped\n", pid);
		  detach (tcp, WSTOPSIG (status));
		  continue;
		}
	    }

	  if (tcp->flags & TCB_BPTSET)
	    {

	      if (clearbpt (tcp) < 0)	/* Pretty fatal */
		{
		  droptcb (tcp);
		  cleanup ();
		  return -1;
		}
	    }
	  goto tracing;
	}

      if (WSTOPSIG (status) != SIGTRAP)
	{
	  /* il segnale che ha fermato il figlio non e' una SIGTRAP */
	  if (WSTOPSIG (status) == SIGSTOP && (tcp->flags & TCB_SIGTRAPPED))
	    {
	      /*
	       * Trapped attempt to block SIGTRAP
	       * Hope we are back in control now.
	       */
	      tcp->flags &= ~(TCB_INSYSCALL | TCB_SIGTRAPPED);
	      if (ptrace (PTRACE_SYSCALL, pid, (char *) 1, 0) < 0)
		{
		  perror ("trace: ptrace(PTRACE_SYSCALL, ...)");
		  cleanup ();
		  return -1;
		}
	      continue;
	    }

	  /* AGGIUNTA tprintf */
	  //tprintf("--- %s (%s) @ %lx (%lx) ---\n", signame(WSTOPSIG(status)), strsignal(WSTOPSIG(status)), 0L, 0L);
	  

	  if ((tcp->flags & TCB_ATTACHED) &&
	      !sigishandled (tcp, WSTOPSIG (status)))
	    {
	      detach (tcp, WSTOPSIG (status));
	      continue;
	    }

	  if (ptrace (PTRACE_SYSCALL, pid, (char *) 1, WSTOPSIG (status)) < 0)
	    {
	      perror ("trace: ptrace(PTRACE_SYSCALL, ...)");
	      cleanup ();
	      return -1;
	    }

	  tcp->flags &= ~TCB_SUSPENDED;
	  continue;
	}

      //if (trace_syscall (tcp, mq) < 0)
      if (trace_syscall (tcp) < 0) /*********** vedi syscall.c ************/
	{
	  if (tcp->flags & TCB_ATTACHED)
	    detach (tcp, 0);
	  else
	    {
	      ptrace (PTRACE_KILL, tcp->pid, (char *) 1, SIGTERM);
	      droptcb (tcp);
	    }
	  continue;
	}

      if (tcp->flags & TCB_EXITING)
	{
	  if (tcp->flags & TCB_ATTACHED)
	    detach (tcp, 0);
	  else if (ptrace (PTRACE_CONT, pid, (char *) 1, 0) < 0)
	    {
	      perror ("widsard: ptrace(PTRACE_CONT, ...)");
	      cleanup ();
	      return -1;
	    }
	  continue;
	}

      if (tcp->flags & TCB_KILLING) /* AGGIUNTA PER LE AZIONI */
	{
	  if (tcp->flags & TCB_ATTACHED)
	    detach (tcp, SIGKILL);
	  else if (ptrace (PTRACE_CONT, pid, (char *) 1, 0) < 0)
	    {
	      perror ("widsard: ptrace(PTRACE_CONT, ...)");
	      cleanup ();
	      return -1;
	    }
	  continue;
	}

      if (tcp->flags & TCB_SUSPENDED)
	{
	  fprintf (stderr, "Process %u suspended\n", pid);
	  continue;
	}

      if (tcp->flags & TCB_STOPPED) /* AGGIUNTA PER LE AZIONI */
	{			/* ho fatto una freeze. Anche se mi arriva una SIGCONT
				   non riparte ma mi entra qui e poi ritorna alla wait */
	  printf ("the process is stopped...\n");
	  continue;
	}

    tracing:
      if (ptrace (PTRACE_SYSCALL, pid, (char *) 1, 0) < 0)
	{
	  perror ("trace: ptrace(PTRACE_SYSCALL, ...)");
	  cleanup ();
	  return -1;
	}

    } // end while (nprocs != 0)
  
  return 0;
}

//**********************************************************************************************
//** ask_password **
//******************
/*
static char *
ask_password ()
{
  char *pwd = calloc (30, sizeof (char));
  int lun;

  if (write (1, "Inserisci la password di superuser: ", 36) == -1)
    {
      free (pwd);
      return NULL;
    }
  if ((lun = read (0, pwd, 30 * sizeof (char))) == -1)
    {
      free (pwd);
      return NULL;
    }
  pwd[lun - 1] = '\0';
  return pwd;
}				//ask_password
*/
//**********************************************************************************************
//** verify_my_password **
//************************
/*
static int
verify_my_password (char *pwd, int *l, int n)
{
  if (verify_password ("superuser", pwd, root, l, n) == NO_PWD)
    return -1;
  return 0;
}				//verify_my_password
*/
